//=============================================================================
///\file: BitVector.h Variable Length Bit Vector Functions
///
/// The CBitVector class represents sets of 16-bit unsigned integers.
/// Used for indexes into transactions array, tablesInUse.
// (C) Copyright IBM Corp. 2001, 2011  All Rights Reserved.
//=============================================================================

#ifndef BITVECTOR_INCLUDED
#define BITVECTOR_INCLUDED

#ifndef NZSTD_INCLUDED
# include "nzstd.h"
#endif

#ifndef NZTYPES_INCLUDED
# include "nztypes.h"
#endif

#ifndef ERROR_INCLUDED
# include "error.h"
#endif

class CCommBuf;
class CBitVector {
public:
    typedef uint16  index_type;

 public:

    enum {
        // 16-bit short words (don't ask).
        MAX_WORDS = 0x1000,
        // "For technical reasons" (limitations in getPositionOfNextBit, etc.),
        // can only accommodate indexes/bits 0 through 65,534.
        MAX_BITS = 0xFFFF                 // MAX_WORDS*16 -1
    };

    // for the STUPID diab 'compiler' that suffers from  C++ envy    
    void * operator new(size_t) {return (void*)1; }
    void * operator new[](size_t) {return (void*)1; }
    void operator delete(void*) { }
    void operator delete[](void*){ }
        
    CBitVector( ); 
    CBitVector( const CBitVector& that );
    CBitVector& operator=( const CBitVector& that );
    ~CBitVector( );

    void Serialize(CCommBuf* cbuf);
    void Deserialize(uint16 cardinality, uint16 firstBit, void* words);

    bool test( uint16 index ) const;
    void set( uint16 index );
    void unset( uint16 index );
    void clear( );

    // "next" includes bit at position "start"
    // these return getMaxBits (one past the last bit) if no next bit is set
    uint16 getPositionOfNextBit( uint16 start=0 ) const;

    // returns bits newly set
    uint16 orInWord( uint16 wordIndex, uint16 value );

    uint16 operator[]( uint16 wordIndex ) const {  // return the index-th m_word
        if (i_isSingleton()) {
            if (wordIndex == m_firstBit/16)
                return 1 << (m_firstBit % 16);
            else
                return 0;
        }
        else if (wordIndex >= m_numWords)
            return 0;
        else
            return m_words[wordIndex];
    }

    // validate cardinality against actual bits
    void validateCount ( ) const;
    // no bits should be set at "maxBits" or higher
    void validateMax ( uint16 maxBits ) const;

    // is m_firstBit the only bit
    bool i_isSingleton () const {
        return ( m_cardinality == 1 && m_words == NULL );
    }

#ifdef NOT_USED
    bool trim ( uint16 maxBits );
    void grow( uint16 index );  // grow to include this bit position
#endif
    void growNumWords( uint16 newNumWords ); // grow to this # words

    uint16 i_getCardinality() const { return m_cardinality; }

    // "logical" number of words (not physically stored if singleton)
    uint16 i_getNumWords() const {
        return i_isSingleton() ? (m_firstBit / 16) + 1 : m_numWords;
    }

    //return this field for serializing over the commbuf
    uint16 i_serializeNumWords() const {
      return m_numWords;
    }

    //return the words for serializing over the commbuf
    uint16* i_serializeWords() const {
      return i_isSingleton() ? NULL : m_words;
    }

    // CAREFUL - "MaxBits" is really highest possible index plus one
    uint16 i_getMaxBits() const {
        uint16 numWords = i_getNumWords ();
        return (numWords >= MAX_WORDS) ? MAX_BITS : numWords * 16;
    }

 private:
    enum {
        // Bit 65535 (bit 15 in word MAX_WORDS-1) cannot be set.
        LAST_WORD_INVALID_BITS = 1 << 15
    };

    union {
        uint16  m_firstBit;     // cache of only bit if m_words is NULL and cardinality is 1
        uint16  m_numWords;     // otherwise, size of m_words
    };
    uint16  m_cardinality;  // number of bits set
    uint16* m_words;

    // abort() if data members not consistent
    void    validate () const;
};

class CBitVectorIterator {
 public:
    explicit CBitVectorIterator( const CBitVector* theVector ) {
        reset (theVector);
    }

    CBitVectorIterator( const CBitVector* theVector, uint16 position )
        : m_theVector( theVector )
        , m_currentIndex( position )
        { }

    ~CBitVectorIterator( ) { }

    void reset (const CBitVector* theVector) {
        m_theVector = theVector;
        m_currentIndex = 0;
        // Position the iterator at the first bit (if any).
        m_currentIndex = m_theVector->getPositionOfNextBit( m_currentIndex );
    }

    operator bool( ) const { return m_currentIndex < m_theVector->i_getMaxBits(); }
    uint16 operator*( ) const { return m_currentIndex; }

    CBitVectorIterator& operator++( ) {
        m_currentIndex = m_theVector->getPositionOfNextBit( m_currentIndex+1 );
        return *this;
    }

    CBitVectorIterator  operator++( int ) {
        CBitVectorIterator result = *this;
        operator++();
        return result;
    }

 protected:
    const CBitVector* m_theVector;
    uint16 m_currentIndex;
};

#endif
