#ifndef TXID_INCLUDED
#define TXID_INCLUDED
/// @file TXID.h
/// @brief DBOS Transaction ID class and operators
///
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2008, 2014  All Rights Reserved.
//////////////////////////////////////////////////////////////////////

#ifndef TXIDCOMMON_INCLUDED
# include "TXIDCommon.h" // Common definitions for transaction id
#endif

// Not really a "mask" - misleading!
#define SUB_TXID_MASK (0x0001000000000000LL)
#define SUB_TXID_SHIFT (48)


#ifndef  __cplusplus

# error "TXID.h can only be #include'd from a C++ source file"

#else //  __cplusplus

// Forward declaration
class AtomicTXid;

class TXid {
    // the only data element!
    TXidValue value;

public:
    TXid(const TXid& t) { value = t.value; }
    TXid(const TXidValue& i) { value = i; }

#if FOR_CLI || WIN32
    // AtomicTXid does not exist for clients or windows build.
#else // (FOR_CLI || WIN32)
    /// \brief Construct a copy from an atomic TX ID 
    TXid( const AtomicTXid& id );
#endif // (FOR_CLI || WIN32)

    TXid() : value(TXid_NULL) { }

    int32 subxid() const { return (int32)(value >> SUB_TXID_SHIFT); }
    int64 xid() const { return (value & K_TXID_MAX); }
    int64 val() const { return value; }
    uint64 const& ref() const { return value; }

    void incsubxid() { value += SUB_TXID_MASK; }

    // Operators so I don't have to rewrite every piece of code that used xid as 64bit int!

    TXid& operator= (const TXid& rhs_scalar) { value = rhs_scalar.value; return *this;}
    TXid& operator= (const int64& rhs_scalar) { value = rhs_scalar; return *this;}

#if FOR_CLI || WIN32
    // AtomicTXid does not exist for clients or windows build.
#else // (FOR_CLI || WIN32)
    /// \brief Copy from an atomic TX ID 
    TXid& operator= ( const AtomicTXid& rhs );
#endif // (FOR_CLI || WIN32)

    TXid& operator += (int inc) { value += inc; return *this;}
    TXid& operator -= (int inc) { value -= inc; return *this;}
    TXid& operator ++ () { ++value; return *this;}   // prefix operator
    TXid operator ++ (int) { TXid ret(*this); value++; return ret; } // postfix operator  int param is a dummy to ke
    TXid& operator -- () { --value; return *this; }   // prefix operator
    TXid operator -- (int) { TXid ret(*this); value--; return ret; } // postfix operator  int param is a dummy to ke

    // The ppc compiler seems to need this operator...Without it I get errors like:
    // mirrortarget.cpp, line 349: error (dplus:1007): ambiguous conversion -- can't cast operand
    // mirrortarget.cpp, line 349: error (dplus:1012): '!' TXid, argument must be of pointer or arithmetic type

    bool operator !() {return (value?false:true);}

    // These conversions are required for places in the code
    // where txids are declared as either uint64's or int64's.
    // Some are in c modules and some are in places where
    // a class couldn't be used because of a union.
    operator uint64() const {return value;}
    operator  int64() const {return value;}
    // The bool conversion is needed for the lines such as
    // if (txid) .....
    operator   bool() const {return (value?true:false);}
};

#  define TXID_MASK(val) ((val) & K_TXID_MAX)

// Binary operators involving TXid.
// All of these comparison operators only operate on the lower 48 bits.
inline static bool operator==(const TXid& lhs, const TXid& rhs)   {return (lhs.xid() == rhs.xid());}
inline static bool operator==(const uint64& lhs,const TXid& rhs)  {return (lhs == rhs.xid());}
inline static bool operator==(const TXid& lhs, const int64& rhs)  {return (lhs.xid() == TXID_MASK(rhs));}
inline static TXid operator+(const TXid& lhs, int rhs)            {return TXid(lhs.xid() + TXID_MASK(rhs));}
inline static bool operator!=(const TXid& lhs, const TXid& rhs)   {return lhs.xid() != rhs.xid(); }
inline static bool operator!=(const TXid& lhs, const int64& rhs)  {return (lhs.xid() != TXID_MASK(rhs));}
inline static TXid operator& (const TXid& lhs,const TXid& rhs)    {return TXid(lhs.xid() & rhs.xid());}
inline static bool operator<(const TXid& lhs,const TXid& rhs)     {return (lhs.xid() < rhs.xid());}
inline static bool operator>(const TXid& lhs,const TXid& rhs)     {return (lhs.xid() > rhs.xid());}
inline static bool operator>(const TXid& lhs,const int64& rhs)    {return (lhs.xid() > TXID_MASK(rhs));}
inline static TXid operator-(const TXid& lhs, int rhs)            {return TXid(lhs.xid() - TXID_MASK(rhs));}
inline static bool operator||(bool lhs, const TXid& rhs)          {return (lhs || (rhs.xid() != 0));}
inline static bool operator&&(const TXid& lhs,bool rhs)           {return ((lhs.xid() != 0) && rhs);}
inline static int64 operator>>(const TXid& lhs,int rhs)           {return lhs.xid() >> TXID_MASK(rhs);}
inline static int64 operator&(const TXid& lhs,const int64& rhs)   {return lhs.xid() & TXID_MASK(rhs);}
inline static bool operator>=(const TXid& lhs,const TXid& rhs)    {return lhs.xid() >= rhs.xid();}
inline static bool operator<=(const TXid& lhs,const TXid& rhs)    {return lhs.xid() <= rhs.xid();}

//!FIX-sks: OBSOLETE, update callers.
inline static int64 TXid_MaskTop(const TXid& input) {return input.xid();}
inline static int64 TXid_MaskTop(const int64& val) {return val & K_TXID_MAX;}

struct TxInfo {
    TXid xId;               ///< The transaction's effective xid.
                            /// The effective xid will get used when making modifications (CTID, DTID, etc,.)
                            /// The effective xid will be the same as identifying xid for all transactions,
                            /// except for child transactions in a PMT. Child transactions' effective xid will
                            /// be their parent transaction's identifying xid.
    TXid identifyingXid;    ///< The transaction's identifying xid.

    TxInfo(const TXid &txid,const TXid &identifyingTxid): xId(txid), identifyingXid(identifyingTxid) { }
    explicit TxInfo(const TXid &txid): xId(txid), identifyingXid(txid) { }
    TxInfo(const TxInfo &other): xId(other.xId), identifyingXid(other.identifyingXid) { }

    bool isChild() const { return xId != identifyingXid; }
};

#define TxInfo_INVALID (TxInfo(TXid_NULL,TXid_NULL))


#if FOR_CLI || WIN32
// AtomicTXid does not exist for clients or windows build.
#else // (FOR_CLI || WIN32)

/// \brief Wraps the TXid class to provide atomic operations
class AtomicTXid
{
public:
    /// \brief Default constructor initializing to zero
    AtomicTXid( )
        : m_id()
        { __sync_fetch_and_and( &m_id, 0 ); }

    /// \brief Atomic Copy constructor 
    AtomicTXid( const AtomicTXid& rhs )     ///< Atomic TX ID to copy
        : m_id()
        { this->set( rhs.get() ); }

    /// \brief Atomic Assignment operator
    AtomicTXid& operator= ( const AtomicTXid& rhs ) ///< Atomic TX ID to copy
        {
            this->set( rhs.get() );
            return *this;
        }

    /// \brief Set the ID using an atomic operation
    void set( const TXid& xId )     ///< TX ID to set
        {
            while ( true ) {
                if ( this->cas( this->get(), xId.val() ) )
                    break;
            }
        }

    /// \brief Retrieve the ID using an atomic operation
    TXid get( ) const
        {
            return TXid( __sync_fetch_and_add( const_cast<volatile TXidValue*>( &m_id ), 0 ) );
        }

private:
    /// \brief Compares the ID to oldval and, if equal, sets to newval.
    /// \details If the ID was not equal to oldval then no change occurs.
    /// This compare and exchange is done atomically.
    bool cas( TXidValue oldval  ///< The value to compare with
            , TXidValue newval  ///< The value to conditionally replace with
            )
        {
            return __sync_bool_compare_and_swap( &m_id, oldval, newval );
        }

private:
    volatile TXidValue m_id;    ///< Atomic TX ID value
};

inline
TXid::TXid
    ( const AtomicTXid& id
    )
    : value( id.get().val() )
{
}

inline
TXid&
TXid::operator=
    ( const AtomicTXid& rhs
    )
{
    value = rhs.get().val();
    return *this;
}

#endif // (FOR_CLI || WIN32)

#endif // __cplusplus

#endif // TXID_INCLUDED
