/*
**      TXIter.h  -  iterate over transactions in CHostTXMgr or CSpuTXMgr
**
**      Copyright: (c) 2008 by Netezza
**      All rights reserved.
*/

#ifndef TXITER_INCLUDED
#define TXITER_INCLUDED

#ifndef ITERATOR_INCLUDED
# include <iterator>
# define ITERATOR_INCLUDED
#endif

#ifndef SYS_TYPES_INCLUDED
# include <sys/types.h>
# define SYS_TYPES_INCLUDED
#endif

#ifndef TXID_INCLUDED
# include "TXID.h"
#endif

#ifndef ERROR_INCLUDED
# include "error.h"
#endif

#include "TXMTypes.h"
#include <stddef.h>

// "indexes" 0x3FFFF (by CBitVector) and 0x3FFFE are reserved
const int TX_MAX_TRANS = 0x3FFFE;
const int TXINDEX_MAX = TX_MAX_TRANS - 1;

//!FIX-sks: These should be inline member functions?
#define TXINDEX_HIGH     (max_txs - 1)
#define TX_FREELIST_END  (max_txs)
#define TX_END_OF_LIST(idx) ((idx == TXINDEX_NONE.val()) || (idx == TX_FREELIST_END))

// Check invariants in DEBUG build.
#define CHECK_CUR_IDX()  NZ_ASSERT (checkCurIndex())
#define CHECK_PREV_IDX() NZ_ASSERT (checkPrevIndex())
#define CHECK_CUR_TX()   NZ_ASSERT (checkCurTx())                   // subsumes checkCurIndex()
#define CHECK_PREV_TX()  NZ_ASSERT (checkPrevTx())                  // subsumes checkPrevIndex()
#define CHECK_ALL()      NZ_ASSERT (checkCurTx() && checkPrevTx())

template <class TX, class TXMgr>
class CTXIter
: public std::iterator<TX *, ptrdiff_t>
{
 private:
  const int max_txs;
  unsigned int m_tx_returned;
  TxIndex m_cur_index;
  TxIndex m_prev_index;
  TX*   m_cur_tx;
  const TXMgr* txmi;

  void set_cur_tx ();

 public:
  CTXIter (const TXMgr *, int max_txs, TxIndex start_index, TxIndex prev_index = TXINDEX_UNKNOWN);
  CTXIter& operator++ ();
  CTXIter  operator++ (int);

  bool operator!= (const CTXIter& rhs) const { return m_cur_tx != rhs.m_cur_tx; };

  TX* operator* () { return m_cur_tx; };
  const TX* operator* () const { return (const TX*) m_cur_tx; };

  unsigned int tx_returned () const { return m_tx_returned; }

  TxIndex cur_index () const { return m_cur_index; }
  TXid cur_txid () const { return m_cur_tx ? m_cur_tx->i_getTID() : TXid_NULL; };
  
  TxIndex prev_index () const { return m_prev_index; }
// Do we need a const version of this?
  TX* prev_tx () const;

// Advance iterator prior to removal from list:
  CTXIter& skip ();

// Test invariants:
  bool checkCurIndex() const
      { return (m_cur_index >= TXINDEX_NONE) && (m_cur_index.val() <= TX_FREELIST_END); }
  bool checkPrevIndex() const
      { return (m_prev_index >= TXINDEX_UNKNOWN) && (m_prev_index.val() <= TXINDEX_HIGH); }

  bool checkCurTx () const
  { return checkCurIndex()
          && ((!m_cur_tx && TX_END_OF_LIST(m_cur_index.val()))
              || (m_cur_tx == txmi->i_getTX(m_cur_index)))
          && ((!m_cur_tx) || (m_cur_tx->nextTxIndex() == TXINDEX_NONE)
              || (txmi->i_getTX(m_cur_tx->nextTxIndex())->prevTxIndex() == m_cur_index)); }

  bool checkPrevTx () const
  { return checkPrevIndex()
          && (!prev_tx() || prev_tx()->nextTxIndex() == m_cur_index)
          && (!m_cur_tx || m_cur_tx->prevTxIndex() == m_prev_index); }
};

template <class TX, class TXMgr>
CTXIter<TX, TXMgr>::CTXIter (const TXMgr* mgr, int mts, TxIndex start_index, TxIndex prev_index)
  : txmi (mgr), max_txs (mts), m_cur_tx (NULL),
    m_cur_index (start_index), m_prev_index (prev_index), m_tx_returned (0)
{
  NZ_ASSERT (max_txs <= TX_MAX_TRANS);
  set_cur_tx ();
  CHECK_ALL();
}

// Used by constructor and by skip().
template <class TX, class TXMgr>
void CTXIter<TX, TXMgr>::set_cur_tx ()
{
  if (TX_END_OF_LIST(m_cur_index.val())) {
    m_cur_tx = NULL;
  }
  else {
    //m_tx_returned counts the number of txs from the beginning
    ++m_tx_returned;
    m_cur_tx = txmi->i_getTX (m_cur_index);
  }
}

template <class TX, class TXMgr>
TX* CTXIter<TX, TXMgr>::prev_tx () const
{
    if (m_prev_index != TXINDEX_NONE && m_prev_index != TXINDEX_UNKNOWN)
    {
        return txmi->i_getTX (m_prev_index);
    }
    else {
        return NULL;
    }
}

template <class TX, class TXMgr>
CTXIter<TX, TXMgr> &
CTXIter<TX, TXMgr>::operator++ ()
{
  m_prev_index = m_cur_index;
  skip ();
  CHECK_PREV_IDX();
  CHECK_PREV_TX();
  return *this;
}

template <class TX, class TXMgr>
CTXIter<TX, TXMgr>
CTXIter<TX, TXMgr>::operator++ (int)
{
  CTXIter<TX, TXMgr> t (*this);
  operator++();
  return t;
}

// Advance the cur_index without advancing prev_index.
//!NOTE-sks: does not take care of updating m_oldestTxIndex / m_newestTxIndex in the caller.
// Caller must remove m_cur_tx & update prev->m_nextTxIndex AFTER skip(), because it is accessed here:
//  TX* txToRemove = *iter;
//  iter.skip ();
//  int cur_index = iter.cur_index ();
//  TX* prevTx = iter.prev_tx ();
//  if (prevTx) {
//    prevTx->m_nextTxIndex = cur_index;
//  }
//  else {
//    // update the head of the list
//    m_oldestTxIndex = cur_index;
//  }
//  // also update m_oldestUnfinTxIndex, m_newestTxIndex if appropriate
//  // put txToRemove on the free list
//  // etc...
template <class TX, class TXMgr>
CTXIter<TX, TXMgr> &
CTXIter<TX, TXMgr>::skip ()
{
  if (m_cur_tx != NULL) {
    if (m_tx_returned >= max_txs) {
// Indicate max exceeded (m_cur_tx should have become NULL at this point).
//!FIX-sks: This should THROW_ERROR instead?!
        ++m_tx_returned;
        m_cur_index = TXINDEX_NONE;
        m_prev_index = TXINDEX_UNKNOWN;
    }
    else {
      // Advance to the next one.
      m_cur_index = m_cur_tx->nextTxIndex ();
    }
  }
  CHECK_CUR_IDX();
// This increments m_tx_returned if we're not at the end.
// (If we are, last element has already been counted.)
  set_cur_tx ();
  CHECK_CUR_TX();
  return *this;
}

#endif  // TXITER_INCLUDED
