/*
**      TXMgs.h  
**
** (C) Copyright IBM Corp. 2001, 2015  All Rights Reserved.
*/


#ifndef TXMSGS_INCLUDED
#define TXMSGS_INCLUDED

#include "nztypes.h"
#include "nzlimits.h"
#include "storagerefs.h"
#include "TXID.h"
#include "messages.h"
#include "TXIter.h"

#include "TruncateWMMsgArray.h"

// Over-the fabric structures use __i386__ packing rules
#include "nzpack_x86_on.h"

// Options bits for CCliStartTxMsg
enum {
    STARTTX_REALLYEXPLICIT      = 1 << 0,  // explicit BEGIN
    STARTTX_INVISIBLE           = 1 << 1,  // aka "read-only"
    STARTTX_SERIALIZABLE        = 1 << 2,  // what it says (off = snapshot isolation)
    STARTTX_QUEUEIFFULL         = 1 << 3,  // begin_queue_if_full
    STARTTX_CTASVISIBLE         = 1 << 4,  // for backup
    STARTTX_ISPARENT            = 1 << 5,  // parent of parallel multi-stream transaction (PMT)
    STARTTX_OVERRIDEVISIB       = 1 << 6,  // for replication
    STARTTX_OVERRIDEREUSETXID   = 1 << 7,  // for replication
    STARTTX_BLOCK_EARLIER_APPLY = 1 << 8,  // for replication, RTC 97711
    STARTTX_ISCTAS              = 1 << 9,  // for backup/ctas, RTC 91977
};

struct TableIDArray {
    int32 count;
    table_id_t array[1];
};
#define TableIDArraySize(numtables) (sizeof(int32) + (sizeof(table_id_t) * numtables) )

// message sent from client to dbos to start a transaction
struct CCliStartTxMsg {
    TXid prevXid;           // Applicable only during transaction restart
    TXid parentXid;
    uint32 startTxOptions;  // See enum bits above
    bool parent;  //!FIX-pmt: Add to startTxOptions...
    TXid overrideTXid;
    uint32 overrideInvisTXidCount;
    TXid overrideInvisTXids[1]; // variable-length array

    // replTablesModified is a variable length struct. It's used by replication.
    // Since there are 2 variable length arrays in this struct, we should NOT
    // reference replTablesModified directly.
    //
    // To get the correct offset, use the method getReplTablesModified().
    TableIDArray replTablesModified; // used for replication, variable length struct

    // replTablesModified is a variable array of TableIDArray pointers.
    // Even, it's preceding member overrideInvisTXids also a
    // variable member. So, to find the actual offset of 
    // replTablesModified, we need to find the end of overrideInvisTXids.

    // replTablesModified depends on if there is more than one
    // overrideInvisTXidCount or not. If there is only one
    // overrideInvisTXidCount, then pReplTablesModified can be
    // accessed directly. Otherwise, we need to go to end of
    // overrideInvisTXids array to get the correct offset of 
    // pReplTablesModified.

    TableIDArray *
    getReplTablesModified() const
    {
    
        if (overrideInvisTXidCount > 1)
            return ( (TableIDArray *) (overrideInvisTXids +
                                       overrideInvisTXidCount) );
        else
            return  (TableIDArray *)&replTablesModified;

    }
};

// message sent from host to spu to start a transaction
struct CStartTxMsg {
    TXid   tid;
    TxIndex::value_type slot;
    nzbool wait;
    nzbool isExplicit;
    nzbool isReadOnly;
    uint32 startTime;    // seconds
    uint32 startTimeMicros;
};

struct CCliCommitTxMsg {
    TXid    tid;
    nzbool  deferredExtReclamation;
};

// message sent from host to spu to commit or rollback a transaction
struct CCompleteTxMsg {
    TXid tid;
    bool deferredExtReclamation;
    uint16 txState;    // should be eCommitted or eAborted
    uint32 endTime;    // seconds
    uint32 endTimeMicros;
    TruncateWMMsgArray wmEntriesArray;

    CCompleteTxMsg()
        : wmEntriesArray( ) {}

    void init( TXid id,
               uint16 state,
               uint32 endtime,
               uint32 endtime_micros,
               uint16 numTruncates = 0,
               bool deferredTruncate = false )
    {
        tid = id;
        txState = state;
        endTime = endtime;
        endTimeMicros = endtime_micros;
        deferredExtReclamation = deferredTruncate;
        wmEntriesArray.init( numTruncates );
    }

    uint32 getNumTruncateWMEntries() const
    {
        return wmEntriesArray.getNumTruncateWMEntries();
    }

    int CommitMsgBufferSize() const
    {
        return sizeof(CCompleteTxMsg) + wmEntriesArray.getTruncateWMMsgSize();
    }

    void addTruncateWaterMark( table_id_t tblId,
                               int dsId,
                               uint64 waterMark )
    {
        wmEntriesArray.addTruncateWaterMark( tblId, dsId, waterMark );
    }

    TruncateWMMsg* getTruncateWMMsg( table_id_t id )
    {
        return wmEntriesArray.getTruncateWMMsg( id );
    }

    uint64 getWaterMark( table_id_t id, int dslice )
    {
        return wmEntriesArray.getWaterMark( id, dslice );
    }
};

// Force reclamation of truncated extents (NZ_FORCE_EXTENT_RECLAIM).
// One or the other member must be zero...
struct ForceExtentReclaimMsg {
    TXidValue  throughTXid;
    table_id_t tableId;
};

// message to copy visibility from source tx to dest tx
// (NZ_COPY_VISIBILITY; nzbackup now uses PMTs instead)
struct CCopyTxVisMsg {
    TXidValue srctxid;
    TXidValue desttxid;
};

struct CRestartTxMsg {
// Note: The fields in this struct come from the CTransaction class
// which needs to be passed to spus to restart a transaction after a reboot.
    TXid                m_tid;          // unique transaction identifier
    TxIndex::value_type m_slot;         // The Host TxEntry Memory Slot
    uint32              m_startTime;    // time this transaction started
    uint32              m_startTimeMicros;
                                        // used for victim selection
    int32               m_activityCount; // number of active plans/interactions
    uint16              m_state;        // e.g. active, committing, etc.
    int16               m_abortTries;   // how many attempts to abort
    nzbool              m_isExplicit;   // true if user issued BEGIN WORK
    nzbool              m_isReadOnly;   // true if user issued SET TRANSACTION READ ONLY
    nzbool              m_hasMadeVisibleChanges; // true if this transaction has modified any non-temporary data
    uint32              m_flags;                // special-handling flags
    int32               m_rollbackIntId;


// note: each of the TableIDArrays below ARE in the message, but since they are variable
// length only 1 actually appears in this structure.  They represent bits set in bit vectors in
// the host's version of the transaction.
  TableIDArray          m_tables;
  //    TableIDArray         m_tablesUpdated;    // tables updated
  //    TableIDArray         m_tablesRead;      /// tables read
  //    TableIDArray         m_tablesDeleted;   // DELETE or UPDATE
  //    TableIDArray         m_tablesInserted;  // INSERT or UPDATE
  //    TableIDArray         m_tablesTruncated; // TRUNCATE TABLE
  //    TableIDArray         m_tablesCreated;   // CREATE TABLE
  //    TableIDArray         m_tablesDropped;   // DROP TABLE
  //    TableIDArray         m_tablesGroomed;   // GROOM TABLE
  //    TableIDArray         m_sequencesChanged;  // sequence stuff
};

// Load continuation.

struct checkpoint_msg_t {
    int64 txId;
    uint32 tableId;
};

struct partial_rollback_msg_t {
    int64    txid; // the last checkpointed transaction
    int32    tblid;// table being loaded
    uint64   EOTs; // place holder for the end of tables for each dataslice
};

struct partial_commit_ack_t {
    int32 numTables;            // always 1
    uint64 eot;                 // degenerate array of length numTables
};


// For txdetail and rollback.

#define NULL_PAGEREF (NULL_PID.rep())
#define UNKN_PAGEREF (0xFFFFFFFFFFFFULL)


enum {
    TdjBad = -1,
    TdjInvalid = 0,
    TdjDelete = 1,
    TdjInsert = 2,
    TdjInsertEnd = 3,// only message interpreted by host
    TdjAbort = 4
};

// The blks per msgs is set to 24 to match the size of an extent although
// no attempt is made to do anything on an extent boundary yet.  That might
// be an improvement that gets made someday.
const int32 TxBlksPerMsg = MAX_GENC_DELETE_PAGES;

enum TxLogAckStatus {
    TX_LOG_ACK_SUCCESS = 0,
    TX_LOG_ACK_FAILURE = 1
};

struct TxLogAckRespMsg {
    dbos_msg_header_t   hdr;
    int32               status;
};

// message returned for groom and nzbackup.  All transactions prior to the stableXid are complete
// and are not on the invisibility list of any currently executing transaction.
struct TxStableXidInfoMsg {
    TXid stableXid;
    time64_t stableXidStartTime;
    TXid oldestUnfinishedXid;
    time64_t oldestUnfinishedStartTime;
};

// next 2 structures are used in place of "blocks" in the above message
// when the host is requesting the "End of Table"
struct TxGetEotRequest {
    int32 tableId;
    int64 offset; // offset in the tdj, not interpreted by the spu, just returned
};

struct TxGetEotResponse {
    int32 tableId;
    int64 offset;     // simply returned
    uint64 eot;       // current end of table pageid
    uint64 pageCount;
};

/// \brief Maintain the EOT information that gets stored in shared memory
struct TxEot {
    table_id_t tableId;
    uint32 dsId;
    uint64 eot;
    uint64 count;
};

struct TxDetailJnlHead {
    int64 tdjPrOffset;              // only used on the host
    int64 tdjTxid;
    int32 tdjType;
    int32 tdjDataSlice;
    union {
        int32 tdjTableId;
        int32 tdjNumTables;
    };

    TxDetailJnlHead ()
        : tdjPrOffset(-1), tdjTxid(K_TXID_MAX), tdjType(TdjBad), tdjDataSlice(-1), tdjTableId(-1)
        {}
    void init() {
        tdjPrOffset = -1;
        tdjTxid = K_TXID_MAX;
        tdjType = -1;
        tdjDataSlice = -1;
        tdjTableId = -1;
    }
};

struct TxDetailJnlData : TxDetailJnlHead {
    uint48 tdjBlocks[TxBlksPerMsg];  // list of page ids

    TxDetailJnlData () {
        for (int ix=0; ix<TxBlksPerMsg; ++ix) {
            tdjBlocks[ix] = tdjPack48 (UNKN_PAGEREF);
        }
    }
    void init() {
        TxDetailJnlHead::init();
        for (int ix=0; ix<TxBlksPerMsg; ++ix) {
            tdjBlocks[ix] = tdjPack48 (UNKN_PAGEREF);
        }
    }
};


// Indexes into tdjBlocks for TdjInsert entries.
int const StartEot = 0;
int const EndEot = 1;
int const StartCount = 2;
int const EndCount = 3;

#include "nzpack_x86_off.h"

#endif
