//========================================================================
///\file: TruncateWMMsgArray.h Truncate Water Mark Message Array class
//
// (C) Copyright IBM Corp. 2012, 2013  All Rights Reserved.
//========================================================================

#ifndef TRUNCATEWMMSGARRAY_INCLUDE
#define TRUNCATEWMMSGARRAY_INCLUDE

#include "TruncateWMMsg.h"

/// \brief Truncate Water Mark Message Array class
class TruncateWMMsgArray
{
public:
    TruncateWMMsgArray()
        : numTruncateWMEntries (0) {}

    TruncateWMMsg& operator[] (int wmIndex)
    {
        // It is "<= numTruncateWMEntries" because when adding a new truncate
        // water mark entry, we add it to the end of the array and then
        // increment numTruncateWMEntries
        NZ_CHECK( wmIndex >=0
                  && wmIndex <= numTruncateWMEntries,
                  NZ_ERROR_ILLVAL,
                  "invalid TruncateWMMsgArray index",
                  wmIndex );

        TruncateWMMsg *e = &truncEntries[0];
        int offset = TruncateWMMsg::TruncateWMMsgSize() * wmIndex;
        return *(TruncateWMMsg*)((char*)e +  offset);
    }

    void init( uint32 numTruncates )
    {
        numTruncateWMEntries = numTruncates;
        for ( int idx=0; idx < numTruncateWMEntries; idx++ )
        {
            (*this)[idx].init( NEVER_EID.rep() );
        }
    }

    void addTruncateWaterMark( table_id_t tblId,
                               int dsId,
                               uint64 waterMark )
    {
        TruncateWMMsg* entry = getTruncateWMMsg( tblId );

        if ( !entry )
        {
            // If a truncate water-mark entry doesn't exist for the tblId,
            // add it to the end of truncate water-mark array.
            entry = &((*this)[numTruncateWMEntries]);
            entry->addWaterMark( tblId, dsId, waterMark );
            numTruncateWMEntries++;
        }
        else
        {
            NZ_ASSERT( tblId == entry->getTableId() );
            // If the table already has a TruncateWMMsg, we will use it
            entry->addWaterMark( dsId, waterMark );
        }
    }

    TruncateWMMsg* getTruncateWMMsg( table_id_t id )
    {
        for (int idx=0; idx<numTruncateWMEntries; idx++)
        {
            if ( (*this)[idx].getTableId() == id )
                return &((*this)[idx]);
        }
        return NULL;
    }

    uint64 getWaterMark( table_id_t id, int dslice )
    {
        TruncateWMMsg *entry = getTruncateWMMsg( id );
        if (entry)
        {
            return entry->getWaterMark(dslice-1);
        }
        return NULL_EID.rep();
    }

    size_t getTruncateWMMsgSize() const
    {
        return (TruncateWMMsg::TruncateWMMsgSize() * numTruncateWMEntries);
    }

    uint32 getNumTruncateWMEntries() const
    {
        return numTruncateWMEntries;
    }

private:
    uint32 numTruncateWMEntries;
    TruncateWMMsg truncEntries[0];
};

#endif // TRUNCATEWMMSGARRAY_INCLUDE

