//========================================================================
///\file: TxIdList.h Transaction ID Container
//
// Copyright (c) 2010 Netezza Corporation.  All rights reserved.
//========================================================================

#ifndef TXIDLIST_INCLUDE
#define TXIDLIST_INCLUDE

#include <algorithm>
#include <vector>

#include "TXID.h"
#include "error.h"
#include "nzshmalloc.h"


/// \brief Transaction ID List Template class
///
/// A template base class that contains a list of transaction ids.
/// The typename \a _Alloc provides support for different memory allocators.
template<typename _Alloc = std::allocator<TXid> >
class TxIdListTemplate
{
public:
    /// \brief STL vector which holds the TXids for the TxIdListTemplate
    typedef typename std::vector<TXid, _Alloc> TxIdVector;
    /// \brief Iterator for the TxIdListTemplate
    typedef typename TxIdVector::iterator iterator;
    /// \brief Const Iterator for the TxIdListTemplate
    typedef typename TxIdVector::const_iterator const_iterator;

public:
    TxIdListTemplate();

public:
    bool        append( const TXid& element );
    bool        insertUnique( const TXid& element );
    void        copy( const TxIdListTemplate &src );
    void        remove( const TXid& element );
    iterator    erase( iterator first, iterator last );
    void        reserve( unsigned numElements );
    void        clear();
    bool        empty() const;
    unsigned    length() const;
    bool        contains( const TXid& element ) const;
    iterator    find( const TXid& element );
    iterator    findGreaterThan( const TXid& element );
    bool        findAscending( const TXid& element,
                               const TxIdListTemplate::const_iterator& startIt,
                               TxIdListTemplate::const_iterator& outIt ) const;

    iterator                begin();
    const_iterator          begin() const;
    iterator                end();
    const_iterator          end() const;

private:
    /// \brief Disabling copy constructor
    TxIdListTemplate( const TxIdListTemplate & );
    /// \brief Disabling copy operator
    TxIdListTemplate& operator= ( const TxIdListTemplate & );

private:
    TxIdVector m_list;  ///< \brief Storage for the list of TXids
};


class TxIdList : public TxIdListTemplate<>
/// \brief Default Memory Transaction ID List class
///
/// This class uses the default memory allocator for TxIdListTemplate
{
    // Use Standard Allocator
};

class TxIdListShm : public TxIdListTemplate<NzShmAlloc<char, mt_txidlist>::stl_alloc>
/// \brief Shared Memory Transaction ID List class
///
/// This class uses a shared memory allocator for TxIdListTemplate
{
    // Use NZ Shared Memory Allocator
public:
    /// \brief Allocates the number of \a bytes for the TxIdListShm Object
    void*   operator new( size_t bytes )
    { return mem::allocBlock(bytes, mem::AllocContext(mt_txidlist )); }
    void    operator delete( void* ptr )
    { mem::freeBlock(ptr); }
};


template<typename _Alloc>
inline
TxIdListTemplate<_Alloc>::TxIdListTemplate()
/// \brief Default constructor
    : m_list()
{
}

template<typename _Alloc>
inline void
TxIdListTemplate<_Alloc>::clear()
/// \brief Clears the list of TXids
{
    m_list.clear();
}

template<typename _Alloc>
inline void
TxIdListTemplate<_Alloc>::reserve( unsigned numElements )
/// \brief Reserves memory for a certain number of elements.
{
    m_list.reserve( numElements );
}

template<typename _Alloc>
inline bool
TxIdListTemplate<_Alloc>::empty() const
/// \brief Is the list empty?
{
    return ( 0 == m_list.size() );
}

template<typename _Alloc>
inline unsigned
TxIdListTemplate<_Alloc>::length() const
/// \brief Returns the number of TXids in the list
{
    return m_list.size();
}

template<typename _Alloc>
inline typename TxIdListTemplate<_Alloc>::iterator
TxIdListTemplate<_Alloc>::begin()
/// \brief Returns an iterator to the beginning in the list
{
    return m_list.begin();
}

template<typename _Alloc>
inline typename TxIdListTemplate<_Alloc>::const_iterator
TxIdListTemplate<_Alloc>::begin() const
/// \brief Returns a const iterator to the beginning in the list
{
    return m_list.begin();
}

template<typename _Alloc>
inline typename TxIdListTemplate<_Alloc>::iterator
TxIdListTemplate<_Alloc>::end()
/// \brief Returns an iterator just past the last element in the list
{
    return m_list.end();
}

template<typename _Alloc>
inline typename TxIdListTemplate<_Alloc>::const_iterator
TxIdListTemplate<_Alloc>::end() const
/// \brief Returns a const iterator just past the last element in the list
{
    return m_list.end();
}

template<typename _Alloc>
inline bool
TxIdListTemplate<_Alloc>::contains
    ( const TXid& element   ///< [in] The TX ID to find
    ) const
/// \brief Returns true if the element exists in the list
{
    return std::binary_search( this->begin(), this->end(), element );
}

template<typename _Alloc>
inline bool
TxIdListTemplate<_Alloc>::append
    ( const TXid& element   ///< [in] The TX ID to add
    )
/// \brief Add an unique element to the end of the list.
///
/// Optimization assumptions:
///   - The element is added with ascending values.
///   - The element must be unique in the list.
{
    if ( ( 0 == m_list.size() )          // Empty list test
      || ( m_list.back() < element ) ) { // Ascending value
        // Add the element to the end
        m_list.push_back( element );
        return true;
    }
    return false;
}

template<typename _Alloc>
inline bool
TxIdListTemplate<_Alloc>::insertUnique
    ( const TXid& element   ///< [in] The TX ID to insert
    )
/// \brief Insert an unique element in ascending order.
{
    std::pair<iterator, iterator> pairIt;
    pairIt = std::equal_range( this->begin(), this->end(), element );
    // If the lower_bound result is equal to the upper_bound result
    // then the element doesn't exist in the vector.
    if ( ( pairIt.first == this->end()   )
      || ( pairIt.first == pairIt.second ) ) {
        // Insert the element using the lower bound iterator
        m_list.insert( pairIt.first, element );
        return true;
    }
    return false;
}

template<typename _Alloc>
inline void
TxIdListTemplate<_Alloc>::copy
    ( const TxIdListTemplate& src   ///< [in] The TxIdListTemplate to copy
    )
/// \brief Makes a copy from the source list.
{
    m_list.assign( src.begin(), src.end() );
}

template<typename _Alloc>
inline void
TxIdListTemplate<_Alloc>::remove
    ( const TXid& element   ///< [in] The TX ID to remove
    )
/// \brief Removes the element from the list
{
    // Binary search remove
    iterator foundIt = std::lower_bound( this->begin(), this->end(), element );
    // Erase the element if found
    if ( (  foundIt != this->end() )
      && ( *foundIt == element     ) ) {
        m_list.erase( foundIt );
    }
}

template<typename _Alloc>
inline typename TxIdListTemplate<_Alloc>::iterator
TxIdListTemplate<_Alloc>::erase
    ( TxIdListTemplate::iterator first,     ///< [in] Iterator pointing to the first element to be erased
      TxIdListTemplate::iterator last       ///< [in] Iterator pointing to one past the last element to be erased
    )
/// \brief Remove a range of elements
/// \return An iterator pointing to the element pointed to by @a last prior to erasing (or end())
///
/// This method will erase the elements in the range and shorten the vector accordingly.
{
    return m_list.erase( first, last );
}

template<typename _Alloc>
inline typename TxIdListTemplate<_Alloc>::iterator
TxIdListTemplate<_Alloc>::find
    ( const TXid& element ///< [in] The TX ID to find
    )
/// \brief Find the element in the list using a binary search
/// \return If found, returns the iterator position. Otherwise, returns the end iterator.
{
    // Binary Search
    iterator it = std::lower_bound( this->begin(), this->end(), element );
    if ( ( it != this->end() ) && ( *it == element ) )
        return it;
    return this->end();
}

template<typename _Alloc>
inline typename TxIdListTemplate<_Alloc>::iterator
TxIdListTemplate<_Alloc>::findGreaterThan
    ( const TXid& element ///< [in] The TX ID to find
    )
/// \brief Find the first TX ID that is greater than @a element
/// \return If found, returns the iterator position. Otherwise, returns the end iterator.
{
    // Binary Search
    iterator it = std::lower_bound( this->begin(), this->end(), element );
    if ( ( it != this->end() ) && ( *it == element ) )
        ++it;   // Goto the next position
    return it;
}

template<typename _Alloc>
inline bool
TxIdListTemplate<_Alloc>::findAscending
    ( const TXid& element                               ///< [in] The TX ID to find
    , const TxIdListTemplate::const_iterator& startIt   ///< [in] starting position
    , TxIdListTemplate::const_iterator& outIt           ///< [out] element's position
    ) const
/// \brief Optimized find for a sorted list and finding by ascending values.
/// \return Returns true if the element exists in list.
///
/// The output iterator points the element in the list.
{
    // Sanity Check
    NZ_VERIFY( startIt >= this->begin() );
    NZ_VERIFY( startIt <= this->end() );
    outIt = startIt;

    // Adjust remove list iterator
    // The element at the iterator must be greater than
    // or equal to the input element.
    while ( ( outIt != this->end() ) && ( *outIt < element ) ) {
        ++outIt; // Increase iterator
    }

    if ( (  outIt != this->end() )
      && ( *outIt == element     ) ) {
        // Found element in list
        return true;
    }
    // Element not found
    return false;
}

#endif // TXIDLIST_INCLUDE
