/* (C) Copyright IBM Corp. 2008, 2011  All Rights Reserved. */

#ifndef BLOCKHEADER_INCLUDED
#define BLOCKHEADER_INCLUDED

#include "nzstd.h"
#include "nztypes.h"
#include "storagedefs.h"

//a pageheader starts each block on disk.  it is followed immediately
//by data records
struct blockheader_t {
    uint32 magic;               // 0xFEEDFACE
    uint32 crc;                 //                                   [omit from CRC]
    union {
        uint32 pageIdLo;        // forensic hint
        uint32 ssv1_blockNum;   // previously table relative page number
    };
    uint32 lba;                 // forensic hint                     [omit from CRC]
    uint32 length;              // data length contained in block
    uint32 layout;              // version                           [omit from CRC]
    uint32 tblId;               // table this belongs to
    uint32 nrSectNrRec;         // high order 8 bits: number of sectors, 0 means 256

    // would it kill us to use some inline accessors?
    ELayoutID getCompressedType() const {
        return (ELayoutID)(layout & COMPRESSED_TYPE_MASK);
    }

    bool getCompressedTypeStable() const {
        return layout & COMPRESSED_TYPE_STABLE_MASK;
    }

    bool getCompressedTypeColumnar() const {
        return (layout & COMPRESSED_TYPE_NONSTABLE_MASK) == COMPRESSED_TYPE_COLUMNAR;
    }

    bool getDtidClean() const {
        return layout & LAYOUT_FACT_DTID_CLEAN;
    }
    
    char * fence() const {
        return (char*)this + length;
    }

    void fixLayoutParity() {
        bool odd = false;
        for (uint32 clearingBits = layout; clearingBits; clearingBits &= clearingBits - 1)
            odd = !odd;
        
        if (!odd)
            layout ^= LAYOUT_BIT_PARITY;
    }

    void setCrcTypeFullBlock() {
        layout = (layout & ~CRC_TYPE_MASK) | CRC_TYPE_FULL_BLOCK;
        fixLayoutParity();
    }

    void setDtidDirty() {
        layout &= ~LAYOUT_FACT_DTID_CLEAN;
        fixLayoutParity();
    }
};

typedef blockheader_t pageheader_t;

struct diskPage_t : pageheader_t {
    char data[131072 - sizeof(pageheader_t)];
};

uint32 dataPageCrc(diskPage_t const *buf, const bool useFastCRC = false);

#define BH_PAYLOAD_START(_pbh) (((diskPage_t*)(_pbh))->data)
#define BH_PAYLOAD_LENGTH(_pbh) (sizeof ((diskPage_t*)(_pbh))->data)

#define BH_BYTE_LENGTH(_pbh, _fixedRecSize) (sizeof(blockheader_t) + (BH_NRECORDS(_pbh) * _fixedRecSize))

#define BH_PAGEID_LO(_pbh) ((_pbh)->pageIdLo)

#define BH_SET_PAGEID_LO(_pbh, _pid) do { \
    pageheader_t* _bh = (_pbh); \
    _bh->pageIdLo = (int32)_pid.rep(); \
} while (0)

#define BH_NRECORDS(_pbh) ((_pbh)->nrSectNrRec & 0x00ffffff)

#define BH_SET_NRECORDS(_pbh, _nr) do { \
    pageheader_t* _bh = (_pbh); \
    _bh->nrSectNrRec = (_bh->nrSectNrRec & 0xff000000) | ((_nr) & 0x00ffffff); \
} while (0)

#define BH_SET_NRECORDS_CLEAR_NSECTORS(_pbh, _nr) do { \
    pageheader_t* _bh = (_pbh); \
    _bh->nrSectNrRec = ((_nr) & 0x00ffffff); \
} while (0)

#define BH_INCR_NRECORDS(_pbh) do { \
    pageheader_t* _bh = (_pbh); \
    _bh->nrSectNrRec = (_bh->nrSectNrRec & 0xff000000) | ((_bh->nrSectNrRec + 1) & 0x00ffffff); \
} while (0)

#define BH_NSECTORS(_pbh) (((_pbh)->nrSectNrRec & 0xff000000) == 0 ? 256 : (_pbh)->nrSectNrRec >> 24)

#define BH_SET_NSECTORS(_pbh, _nsectors) do { \
    pageheader_t* _bh = (_pbh); \
    unsigned _n = (_nsectors); \
    NZ_ASSERT(_n <= 256); \
    _bh->nrSectNrRec = (((uint32) ((uint8) _n)) << 24) | (_bh->nrSectNrRec & 0x00ffffff); /* truncate to 8 bits, 0 means 256 sectors */ \
} while (0)

#endif // BLOCKHEADER_INCLUDED
