#ifndef CLIENT_INCLUDED
#define CLIENT_INCLUDED
/// @file client.h
/// @brief DBOS client interface
///
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2001, 2011  All Rights Reserved.
//////////////////////////////////////////////////////////////////////
///
/// Description:
///
//////////////////////////////////////////////////////////////////////

#include "error.h"
#include "nzstd.h"
#include "TXID.h"
#include "nztypes.h"

class CCommBuf;
struct client_result_msg_t;
class CDistSet;
class CTable;
struct refptr_t;
struct CreateSchedRuleStmt;
struct EnableSchedRuleStmt;
struct TxStableXidInfoMsg;

// --------------------- System management

extern void NzAbortOnError( // Set the abort error code - used for debug
    int errcode, bool noCrash)
        throw (CError);

extern void NzBtOnError( // Set the btOnError code - used for debug
    int errcode)         // see error.cpp to understand the values
        throw (CError);

// includeReserved (which applies to subordinate node replicated transactions)
// needs to be conservative if not specified.
extern void NzGetStableTXidInfo(TxStableXidInfoMsg&, bool includeReserved) throw (CError);
extern TXid NzGetStableTXid(bool includeReserved=true) throw (CError);

extern TXid NzGetReplStableTXid() throw (CError);
extern void NzSetReplStableTXid(const TXid& txid) throw (CError);
extern void NzSetDbReplicated(char const* dbnames, int bufsiz, bool replicated) throw (CError);

extern TXid NzGetTXid() throw (CError);
extern TXid NzReserveTXid() throw (CError);
extern TXid NzGetLastTXid() throw (CError);

extern TXid NzStartTransaction(  // Begin transaction, return xId
    bool isExplicit,         // true  for BEGIN WORK statement
                             // false for SQL statements outside a transaction scope
    bool isReadOnly=false,   // true = invisible
    bool serializable=true,
    bool queueIfFull=false,  // true = queue if 63 concurrent updaters
                             // false=force to read-only if 63 concurrent updaters
    bool ctasVisible=false,  // true = add CTASs to invisibility list
    bool isCtas=false,       // true = singleton CTAS
    const TXid& prevXid=TXid_NULL,
    const TXid& parentXid=TXid_NULL,
    bool parent = false,
    bool replBlockEarlierApply=true, // false = abort plan if earlier conflicting replicated apply arrives
    int32 numReplTablesModified = 0,
    Oid *replTablesModified = NULL,
    bool overrideVisibility=false,
    bool overrideAndReuseTXid=false,
    TXid overrideTXid = TXid(),
    uint32 overrideInvisTXidCount = 0,
    TXid *overrideInvisTXids = NULL)
        throw (CError);

extern void NzPrecommitTransaction( // Precommit work done so far
    const TXid& xId,    // transaction id obtained from NzStartTransation()
                        // dbos can figure this out from the client id, but
                        // it can be provided explicitly as well
    bool deferredTruncate)
        throw (CError);

extern void NzCommitTransaction( // Commit work done so far
    const TXid& xId,    // transaction id obtained from NzStartTransation()
                        // dbos can figure this out from the client id, but
                        // it can be provided explicitly as well
    bool deferredTruncate)
        throw (CError);

extern void NzAbortTransaction(  // Explicit rollback request
    const TXid& xId)    // transaction id obtained from NzStartTransation()
                        // dbos can figure this out from the client id, but
                        // it can be provided explicitly as well
        throw (CError);

extern void NzSessionAbortTx( // DBA request to kill transaction
    int sessionId)            // for the given session id
        throw (CError);

extern void NzSetTransactionReadOnly(
    bool readOnly)
        throw (CError);

extern void NzSetTransactionSerializable(
    bool serializable)
        throw (CError);

extern void NzSetTransactionReplicated(
    bool replicated)
        throw (CError);

extern void NzSetTruncStats(
    TXid txid,
    int32 tblid,
    int64 numrows,
    int numpages)
        throw (CError);


extern bool NzTriggerTransactionAbort_r() // Async transaction aborter
    throw ();                             // Reentrant version, ok to call from signal handler

extern void NzTriggerTransactionAbort() // Async transaction aborter
    throw (CError);                     // NON-REENTRANT

extern void NzTriggerTransactionAbortWithErr(int sessionId, const CError& err) // Async transaction aborter
    throw (CError);

extern int64 * NzGetInvisibilityList(int64 reftxn) // Get invisibility list for reference txn
    throw (CError);

extern void NzCopyVisibility(TXidValue srctxid, TXidValue desttxid ) // copy visibility from src txid to dest txid
    throw (CError);

extern void NzSetVisibility(TXidValue srctxid, TXidValue desttxid, TXidValue *invistxids, size_t count) // Set visibility
    throw (CError);

extern unsigned int NzGetSysInstanceId(CError *errptr) throw ();

// --------------------- Inserting records

extern void NzTableAppendRecord(   // Add record to table
    unsigned int tblId,
    void **pairs, unsigned char sbspace)   // array of val ptrs
        throw (CError);

extern void NzTableFlushAppendedRecords(  // Get appended records on disk
    unsigned int tblId)
        throw (CError);

// --------------------- Uploading (groundwork for streaming queries)

extern void NzUploadNextResponse( // Get next part of multi-message response
    client_result_msg_t *ack)  // OUT - new result set
        throw (CError);

// discard response and free response payload
extern void NzDiscardResponse(
    const client_result_msg_t *ack,
    int log_fd) // file descriptor to log messages, or -1 for none
	throw ();

// drain leftover responses on process exit
extern void NzDrainResponses ();

// --------------------- Utilities

extern void NzReorderHashCols(
    int streamTblId,        //tblId for stream
    int numHashCols,        //number of cols to be hashed
    int streamBaseList[],   // list of stream cols from the base table
    int streamHashList[],   //list of cols to be hashed in stream
    int tmpHashList[],      //list of cols to be hashed in temp
    bool nullsCompareList[])
        throw (CError);

extern int NzCliGetDbosClientId() throw ();
extern int NzCliGetSessionId() throw ();

extern void NzSetClientPriority( // raise/lower client's priority
    int sessionId, int newpri) throw (CError);

/////////////////////////////////
// Notify DBOS to change the parameters for a running RSG.
// Sends NZ_SET_RSG_DEFINITION message to Dispatch with the given fields.
// If warranted, this data propagates all the way to the SPUs.
// @param groupOID: RSG oid
// @param resourceMin
// @param resourceMax
// @param jobMax
// @param groupName
/////////////////////////////////
extern void NzSetRsgDefinition (
    int groupOID, int resourceMin, int resourceMax, int jobMax, char * groupName, bool delimIdent)  throw (CError);

extern void NzRenameUser(int userOID, const char * newname, bool delimIdent) throw (CError);

extern void NzRenameDatabase(int dbOID, const char * newname, bool delimIdent) throw (CError);

extern void NzSchedRuleCreate(CreateSchedRuleStmt* stmt) throw (CError);

extern void NzSchedRuleRename(int, const char * newname, bool delimIdent) throw (CError);

extern void NzSchedRuleDrop (int) throw (CError);

extern void NzSchedRuleEnable(int id, bool enable) throw (CError);

extern void NzCloseJob() throw (CError);

// send request, wait for response
// relinquishes ownership of store
extern void CliCallNz(
    int cmdCode,
    size_t sizeStore,
    void *store,
    int memtype,
    client_result_msg_t *ack)
        throw (CError);

extern void CliCallNz(
    int cmdCode,
    CCommBuf *cbuf,
    client_result_msg_t *ack)
        throw (CError);

extern void CliCallNzNoRelease(
    int cmdCode,
    size_t sizeStore,
    refptr_t *store,
    client_result_msg_t *ack)
        throw (CError);

extern void CliRegisterDistSet(int clientId, CDistSet *distSet);
extern void CliUnregisterDistSet(int clientId, CDistSet *distSet);

#endif // CLIENT_INCLUDED
