#ifndef DICTIONARY_INCLUDED
#define DICTIONARY_INCLUDED

#include "nztypes.h"
#include "TXID.h"
#include "nzlog.h" // for NzDumpDetail
#include "locktracker.h"
#include "storagedefs.h"
#include "sdlist.h"
#include "nzglobals.h"
#include "table.h"
#include "memNewDelete.h"

class CMutex;
class CTable;

// mutex names
LKTRKR_DECLARE_NAME(MTX_TBLDICT);
LKTRKR_DECLARE_NAME(MTX_TBLDICT_TR);

class CDictionary : public mem::newDelete<mt_dictionary>
{
    dlink_t m_tableList;     // non-transient tables
    dlink_t m_transientList; // transient tables
    dlink_t **m_table_hash;
    dlink_t **m_transient_hash;
    pthread_rwlock_t m_lock;             // non-transient list concurrency protection
    pthread_rwlock_t m_transient_lock;    // transient list concurrency protection
    CTable **m_vt_list;      // virtual tables
    unsigned int m_vt_cnt;

public:
    void *operator new(size_t);
    void operator delete(void *);

    CDictionary();
    ~CDictionary();

    int getAllDists( // Get all distributions for passed table id
        CTable *baseTable, // one of the tables
        CTable **tables,   // returned array
        int maxTables) const; // how many slots in returned array (must be MAX_TABLE_COLUMNS)

    CTable *getTable( // lookup table and increment table's ref count
        int id) const;

    CTable *lookupTableNoAddRef( // Fetch table description from dictionary
        int id) const;

    CTable *getPersistentTable(int id) const; // lookup table and increment table's ref count
    CTable *getTransientTable(int id) const; // lookup table and increment table's ref count

    CTable *lookupTransientTableLocked(int id) const;

    CTable *lookupVirtualTable(int id) const;

    void addTableDesc(CTable *table);
    void addVirtualTable(CTable *table);

    bool unlinkTable(CTable *tbl);
    CTable *unlinkTableById(int id);

    CTable *lookupTableByNameHack(const char *name) const;  // Don't use!, just for Debug logging!

    void DownloadTableDefs() const;

    CTable *FindNextChildTable(int parentId,
                               int lastId) const;  // ID of last table already seen

    void UnlinkChildTables(dlink_t *list, int parentId);

#if !FOR_SPU
    void RenameCatalog(const char *dbname, const char *newname);
    void RenameSchema(const char *dbname, const char *scname, const char *newname);
    void ReplicateCatalog(const char *dbnames, bool replicated);
#endif

    void DropTxTempTables(const TXid& xId);

    void Dump(const char *title=NULL,int detail=NzDumpDetail, FILE *fp=NzDumpFile, int indent=NzDumpIndent, int dumpTblFields=0) const;

    const dlink_t *tableList() const {
        return &m_tableList;
    }

    dlink_t *tableList() {
        return &m_tableList;
    }

    const dlink_t *transientList() const {
        return &m_transientList;
    }

    dlink_t *transientList() {
        return &m_transientList;
    }

    pthread_rwlock_t *rwlock() {
        return &m_lock;
    }

    pthread_rwlock_t *transientRwlock() {
        return &m_transient_lock;
    }

#define _VT_(oid,name) CTable *name;
#include "virtab_names.h"
#undef _VT_
};

class TableRef
{
    CTable *m_table;

    // not implemented
    TableRef(const TableRef&);
    TableRef& operator=(const TableRef&);

protected:

    void _attach(CTable *table) { m_table = table; }
    void _detach() { if (m_table) m_table->removeTableReference(); }

public:

    TableRef() { _attach(0); }
    TableRef(int id) { _attach(Nz->dictionary->getTable(id)); }
    TableRef(CTable *table) { _attach(table); }
    ~TableRef() { _detach(); }

    void lookupTable(int id) { attach(Nz->dictionary->getTable(id)); }
    void lookupPersistentTable(int id) { attach(Nz->dictionary->getPersistentTable(id)); }
    void lookupTransientTable(int id) { attach(Nz->dictionary->getTransientTable(id)); }

    CTable *operator->() const { return m_table; }
    operator CTable *() const { return m_table; }
    CTable *get() const { return m_table; }

    void attach(CTable *table) {
        NZ_VERIFY(m_table == 0);
        m_table = table;
    }
    CTable *detach() {
        CTable *const t = m_table;
        if (t) {
            t->removeTableReference();
            m_table = 0;
        }
        return t;
    }
    CTable *release() { CTable *const t = m_table; m_table = 0; return t; }
};

struct PersistentTableRef : public TableRef
{
    PersistentTableRef(int id) : TableRef(Nz->dictionary->getPersistentTable(id)) { }
};

struct TransientTableRef : public TableRef
{
    TransientTableRef(int id) : TableRef(Nz->dictionary->getTransientTable(id)) { }
};

#endif          // CDictionary_INCLUDED
