// distKey.h
//
// Classes and other support for distribution key calculation.
//
// (C) Copyright IBM Corp. 2001, 2011  All Rights Reserved.
//

#ifndef DIST_KEY_INCLUDED
#define DIST_KEY_INCLUDED

#ifndef NZTYPES_INCLUDED
# include "nztypes.h" // for record_t
#endif

#ifndef NZLIMITS_INCLUDED
# include "nzlimits.h"
#endif

class CTable;

#define DIST_KEY_MAP_SLOTS      65521

#if !FOR_SPU

int NzRealSpuId( int32 *pDiskKeyMap, uint8 *pBuf, int len );

//========================================================================
// Class: DistKeyFunctor - abstract base class for calculating distribution
// keys.

class DistKeyFunctor {
public:
    DistKeyFunctor (CTable* table) : m_table (table) {};
    virtual ~DistKeyFunctor () {};

    static
    DistKeyFunctor* newInstance (CTable* table);            // Factory
    CTable*         getTable () const { return m_table; }

    virtual int     calcSpuId (const record_t* rec) const = 0;
    virtual void*   makeKey (void* dst, const record_t* rec) const = 0;

    struct FieldInfo {
        FieldInfo (CTable* tbl) : table (tbl) {};

        CTable* table;
        int     fieldId;
    };

    typedef void* CopyRtn (void* dst, const void* src, FieldInfo* fieldInfo);
    
    static void* copy1 (void* dst, const void* src, FieldInfo*);
    static void* copy2 (void* dst, const void* src, FieldInfo*);
    static void* copy4 (void* dst, const void* src, FieldInfo*);
    static void* copy8 (void* dst, const void* src, FieldInfo*);
    static void* copy12 (void* dst, const void* src, FieldInfo*);
    static void* copy16 (void* dst, const void* src, FieldInfo*);
    static void* copyFixedN (void* dst, const void* src, FieldInfo* fInfo);
    static void* copyVarN (void* dst, const void* src, FieldInfo*);

private:
    CTable* m_table;
};

//========================================================================
// Class: SpuFieldDistKeyFtor - a functor class that creates a distribution
// key from a single field that contains an SPU ID.

class SpuFieldDistKeyFtor : public DistKeyFunctor {
public:
    SpuFieldDistKeyFtor (CTable* table, int spuFieldId);
    ~SpuFieldDistKeyFtor () {};

    int     calcSpuId (const record_t* rec) const;
    
private:
    void*   makeKey (void* dst, const record_t* rec) const;     // Not impl'd
    void    initFieldInfo (CTable* table, int spuFieldId);
    
    int m_fieldId;
    int m_fieldType;
};

//========================================================================
// Class: MultiFieldDistKeyFtor - a functor class that creates a distribution
// key from multiple record fields.

class MultiFieldDistKeyFtor : public DistKeyFunctor {
public:
    MultiFieldDistKeyFtor (CTable* table);
    ~MultiFieldDistKeyFtor () {};

    int     calcSpuId (const record_t* rec) const;
    void*   makeKey (void* dst, const record_t* rec) const;
    
private:
    void        initFieldInfo (CTable* table, int nFields, int16* distFieldIds);
    CopyRtn*    copyRtn (CTable* table, int fieldId);

    uint16*      m_distKeyMap;

    size_t      m_nFields;
    int16       m_fieldIds[MAX_DIST_KEY_FIELDS];
    CopyRtn*    m_copyRtns[MAX_DIST_KEY_FIELDS];
};

#endif

#endif
