#ifndef TABLEDIST_INCLUDED
#define TABLEDIST_INCLUDED

#if !FOR_SPU

#include "distKey.h"
#include "TXID.h"
#include "nzlimits.h"
#include "nzstd.h"
#include "nztypes.h"
#include "table.h"     // for ESimpleHashType


namespace rc { class Err; }

class CDistRows;
class COSInteraction;
class CTable;

struct TMemblockInfo;
struct TUseSpecialFields;
struct dbos_open_close_channel_t;

class CDistSet {
PRIVATE:
    CDistRows **m_dists;  // null terminated array

    TXid m_txid;
    int m_clientId;
    COSInteraction *m_distrowsInteraction; // current interaction
    COSInteraction *m_closeChnlInteraction; // current interaction
    dbos_open_close_channel_t *m_closeChnlMsgP; // content for close channel interaction

    record_t *m_recbuf;    // temp holding buffer

    // not implemented
    CDistSet(const CDistSet&);
    CDistSet& operator=(const CDistSet&);

    bool m_fromRealSPU;  // records constructed on SPUs (vs. host or sim SPU)

    bool m_clientCreated;           // true if created by a client (ex: loader)
    bool m_aborted; // true if cancelled or aborted

public:
    void * operator new (size_t);   // Shared memory constructor helper
    void operator delete (void*);   // Shared memory constructor helper

    CDistSet(       // constructor
        int baseId,
        const TXid& txid,
        int clientId,
        bool suppressSend=false,
        bool compressed = false,    //if src is compressed data
        bool clientCreated=false ); // Created by a client

    ~CDistSet();

    void AddRec(    // Add record to output buffers, send when buffer fills
        record_t *pRec, int size);

    void AddRecBuf( // Add records to output buffers, send as buffer fills
        record_t *pBuf, int size, bool assignRowIds);

    CTable *Finish(   // wrap-up and return status table
        bool returnErrors);

    void getAllSpusToSameLevel(); // Get each to same interaction id level
    inline CDistRows* GetUniqueDistRows() const { return m_dists[0]; }

    TXid txid() const { return m_txid; }

    void initDistrowsInteraction();
    void cleanupDistrowsInteraction();
    COSInteraction* i_interaction() const {
        return m_distrowsInteraction;
    }
    void setUserCancel() {
        m_aborted = true;
    }

    void setSourceLocus(    // indicate recs from SPU vs. host or sim SPU
        bool recsFromRealSPU );
};

class CDistStats;

class CDistRows {
friend void OsHandleOneSpuMessage(spu_msg_header_t *hdr);
friend bool HandleLoadCommand();
PRIVATE:
    CDistSet *m_distSet; // dist set to which this CDistRows belongs
    CTable *m_table;  // table that controls the distribution
    TMemblockInfo **m_buffers; // allocated recstore blocks, one for each spu
    int m_msgFormat;  // Host/Spu/HostCompressed
    uint64 *m_lastTime;   // to spoon-feed until comms allows higher data rates
    CDistStats* m_stats; // if non-null pointer to array of stats objects
    uint m_statsMemDiskAdjust; // delta between memory and disk record sizes
    int m_cliId;       // slot # in client table
    int m_numDistKeyFields; // from table
    int m_distKeyFields[MAX_DIST_KEY_FIELDS]; // from table
    ESimpleHashType m_simpleDistKey; // true if can self-hash dist col
    bool m_suppressSend;
    bool m_bUseStrmNet; //true to use strmnet
    int m_minDistBufSize; //minimum size of distribute buffer
    bool m_bCompressed; //true if source data is compressed recstore blocks
    bool m_bMemoryTableDist; //if true doing a memory table distribute

    void sendBlock( // Send recordstore block to SPU, may create new Interaction
        int spuId,          // 1-based spu id
        uint8 msgFlags=0); // triggers the 'lastCall' logic of Immed streaming plans

    int  getSpuIdFromRec(record_t *pRec);
    int  getSpuIdFromRecBe(record_t *pRec);

    void addRec(    // Add record to output buffers, send when buffer fills
        record_t *pRec, // packed record buffer
        int size);      // as returned from table->RecordSize(pRec)

    int m_appChannelVal;

public:
    void * operator new (size_t);   // Shared memory constructor helper
    void operator delete (void*);   // Shared memory constructor helper
    void flushLast(bool partial=false);   // Send out last block

    CDistRows(                // constructor
        CDistSet *distSet,
        CTable *table,
        int tblId,
        int msgFormat,
        bool suppressSend=false,
        bool bCompressed=false);

    ~CDistRows();             // destructor

    void addRec(    // Add record to output buffers, send when buffer fills
        record_t *pRec, // packed record buffer
        int size,      // as returned from table->RecordSize(pRec)
        int spuId);    // pre-calc'd target


    void enableStatistics();

    bool getStats(  // Retrieve stats for next SPU; returns TRUE until done
        uint32* spuId,
        uint32* nRecords,
        uint64* xmitBytes,
        uint32* xmitMsgs,
        uint64* diskBytes,
        uint32* diskRecBlks,
        uint64* diskExtentBytes);

    // FIXME - jsy: should both be methods in CTable
    bool hashRequiresVaryings();
    bool isUniqueDistKeyField(uint32 col) const;

    uint32 distKeyMapSlots() const { return DIST_KEY_MAP_SLOTS; }
    void cleanUp(); //free memory

    int allocRec( // Alloc space in output buffers, send when buffer fills
        record_t *pRec, // packed record buffer
        uint size,      // as returned from table->RecordSize(pR ec)
        void** recptr,  // start of record pointer in buffer
        rc::Err** insertErr);


};

class CDistStats {
public:
    uint static memDiskAdjust(CTable* table);
    void accumulateRecord(uint size, uint memDiskAdjust);
    void noteSendToSpu();
    uint32 nRecords() const;
    uint64 xmitBytes() const;
    uint32 xmitMsgs() const;
    uint64 diskBytes() const;
    uint32 diskRecBlks() const;
    uint64 diskExtentBytes() const;
private:
    enum { DISK_RECORD_BLOCK_BYTES = ((1<<17)-32) };

    // quantities tracked as words (= bytes>>2)
    uint m_nRecords;
    uint m_nMessages;
    uint m_xmittedRecordWords;  // memory format
    uint m_diskRecordWords;     // disk format
    int m_diskRecordBlockBytesAvailable;
    uint m_nDiskRecordBlocks;
};

inline bool
CDistRows::isUniqueDistKeyField(uint32 col) const {
    return (m_numDistKeyFields == 1) && (col == m_distKeyFields[0]);
}

inline uint
CDistStats::memDiskAdjust(CTable* table) {
    TableShape const* shape = table->Shape();
    return (shape->i_diskLengthOfSpecials() - shape->i_lengthOfSpecials()) >> 2;
}

inline void
CDistStats::accumulateRecord(uint size, uint memDiskAdjust) {
    m_nRecords +=1;
    size >>= 2; // sizes in word to avoid int64 accumulators
    m_xmittedRecordWords += size;
    size += memDiskAdjust;
    m_diskRecordWords += size;
    m_diskRecordBlockBytesAvailable -= size;
    if ( m_diskRecordBlockBytesAvailable < 0 ) {
        m_diskRecordBlockBytesAvailable = (DISK_RECORD_BLOCK_BYTES >> 2) - size;
        m_nDiskRecordBlocks += 1;
        }
}

inline void
CDistStats::noteSendToSpu() {
    m_nMessages += 1;
}

inline uint32
CDistStats::nRecords() const {
    return m_nRecords;
}

inline uint64
CDistStats::xmitBytes() const {
    return ((uint64)m_xmittedRecordWords)<<2;
}

inline uint32
CDistStats::xmitMsgs() const {
    return m_nMessages + (m_nRecords != 0);
}

inline uint64
CDistStats::diskBytes() const {
    return ((uint64)m_diskRecordWords)<<2;
}

inline uint32
CDistStats::diskRecBlks() const {
    return m_nDiskRecordBlocks;
}

inline uint64
CDistStats::diskExtentBytes() const {
    return ((uint64)((m_nDiskRecordBlocks+23)/24))*(24*(1<<17));
}

#endif // if !FOR_SPU

#endif
