#ifndef ERROR_INCLUDED
#define ERROR_INCLUDED

/*

    Error handling support, most especially the CError object

Throwing Errors:

THROW_ERROR(c,s,o)  This is the base error throwing macro.  It throws an
                    error code 'c', a string 's' that describes the error, and
                    a string 'o' that describes the principal object of
                    the error (e.g. spuname for failed spu).

NZ_CHECK(b,c,s,o)   This macro tests a boolean condition (assert-like)
                    and throws error on failure.

NZ_CHECK_LOGNZ_CHECK(b,c,s,o) The same as NZ_CHECK. Additionally
                    it logs the error message with LOG_PR on failure
                    before the error is thrown.

NZ_VERIFY(b)        This macro is used just to verify internal consistency.
                    Like NZ_CHECK, it tests a boolean condition, but it
                    is used for checking of array boundaries, valid
                    internal parameters, invalid members of switch statements.
                    The error report only includes sourcefile and line number.

NZ_VERIFY(b,s)      Alternative invocation available only in gcc environments
                    that allows a more descriptive diagnostic string.

NZ_VERIFY_STR(b,s)  This macro is identical to NZ_VERIFY in functionality with
                    the ability to report a run-time computed diagnostic string.

NZ_ERRNO_CHECK(b,o) This macro tests a boolean condition (assert-like)
                    and throws error on failure.  The error code and
                    strign are derived from 'errno' and the errno-based
                    strings table.

NZ_ASSERT(b)        Like NZ_VERIFY, but removed via condition compilation
                    from the production build.  Usually, you should use
                    NZ_VERIFY instead, but if the test is very compute
                    intensive or in an otherwise time-critical area,
                    this is appropriate.  This is never used for resource
                    errors that could occur at customer sites.  It is just
                    used to verify internal consistency.

Note that if the Nz->abortOnError is set, all of the aborters above
terminate the porgram, dropping core.

*/

#ifndef NZSTD_INCLUDED
# include "nzstd.h"
#endif

#ifndef TERROR_INCLUDED
# include "terror.h"
#endif

#ifndef RCSEVERITY_INCLUDED
# include "rcSeverity.h"
#endif

#ifndef ERRNO_INCLUDED
# include <errno.h>
# define ERRNO_INCLUDED
#endif

#define ABORT_ON_ERROR_ALL -1
#define ABORT_ON_ERROR_OFF 0

#define _NZ_ERRNO_LOG(l,c,o) CError_ErrnoLog((l),(c),(o),__LINE__,__FILE__)
#define NZ_ERRNO_LOG(l,o) _NZ_ERRNO_LOG((l),errno,(o));
#define THROW_ERROR(c,s,o) CError_Throw((c),false,(s),(o),__LINE__,__FILE__)
#define ERRNO_ERROR(c,o) CError_ErrnoThrow((o),(c),__LINE__,__FILE__)

void _dumpandgo();
#define DUMPANDGO _dumpandgo();

/* The following three macros were used during initial NPS development
 * because we hadn't yet figured out multi-thread C++ exception
 * handling with DIAB.  What you see here is support for the remaining
 * sites.  They are DEPRECATED. [Note that the no-op versions are long
 * gone.] -dk April 2006
 */
#define ERROR_HANDLER try {
#define ERROR_CATCH } catch ( CError &err )
#define ERROR_RETHROW throw

#define NZ_CHECK(b,code,str,obj) \
    do { \
        if (NZ_UNLIKELY(!(b))) \
            THROW_ERROR((code),(str),(obj)); \
    } while (0)

#define NZ_CHECK_DUMP(b,code,str,obj) \
    do { \
        if (NZ_UNLIKELY(!(b))) { \
            DUMPANDGO \
            THROW_ERROR((code),(str),(obj));\
        } \
    } while (0)

#define NZ_CHECK_LOG(b,code,str,obj) \
    do { \
        if (NZ_UNLIKELY(!(b))) { \
            LOG_PR((RC_ERROR, "%s\n", str));\
            THROW_ERROR((code),(str),(obj));\
        } \
    } while (0)

#define NZ_ERRNO_CHECK(b,obj) \
    do { \
        if (NZ_UNLIKELY(!(b))) \
            ERRNO_ERROR(errno, (obj)); \
    } while (0)

#ifdef WIN32
#define NZ_VERIFY(b) \
    do { \
        if (NZ_UNLIKELY(!(b))) {\
            ASSERT(false); \
        } \
    } while (0)
#define NZ_VERIFY_STR(b, str) \
    do { \
        if (NZ_UNLIKELY(!(b))) {\
            ASSERT(false); \
        } \
    } while (0)
#elif defined(__GNUC__)
#define NZ_VERIFY(b, str...)                           \
    do { \
        if (NZ_UNLIKELY(!(b))) \
            CError_AssertFailed(__LINE__,__FILE__ "\0" str); \
    } while (0)
#define NZ_VERIFY_STR(b, str) \
    do { \
        if (NZ_UNLIKELY(!(b))) \
            CError_AssertFailed(__LINE__,__FILE__ "\0", str); \
    } while (0)
#else
#define NZ_VERIFY(b) \
    do { \
        if (NZ_UNLIKELY(!(b))) \
            CError_AssertFailed(__LINE__,__FILE__ "\0"); \
    } while (0)
#define NZ_VERIFY_STR(b, str) \
    do { \
        if (NZ_UNLIKELY(!(b))) \
            CError_AssertFailed(__LINE__,__FILE__ "\0", str); \
    } while (0)
#endif


#if NZDEBUG || NZASSERT

# define NZ_ASSERT(b) NZ_VERIFY(b)
# define NZ_ASSERT_STR(b, str) NZ_VERIFY_STR(b, str)
# define ASSERT(b) NZ_ASSERT(b)
# define NZ_VERIFY_LOG(b, str...) \
    do {\
        if (NZ_UNLIKELY(!(b))) {\
            LOG_PR((RC_ERROR, str));\
            NZ_VERIFY(b);\
        }\
    } while (0)

# define NZ_ASSERT_LOG(b, str...) NZ_VERIFY_LOG(b, str)

#else // NZDEBUG || NZASSERT

# define NZ_ASSERT(b)
# define NZ_ASSERT_STR(b, str)
# define ASSERT(b)
# ifdef WIN32
# define NZ_ASSERT_LOG(b) NZ_ASSERT(b)
# elif defined(__GNUC__)
# define NZ_VERIFY_LOG(b, str...) \
    do {\
        if (NZ_UNLIKELY(!(b))) {\
            LOG_PR((RC_ERROR, str));\
            NZ_VERIFY(b); \
        }\
    } while (0)

# define NZ_ASSERT_LOG(b, str...) \
    do {\
        if (NZ_UNLIKELY(!(b))) {\
            LOG_PR((RC_ERROR, str));\
        }\
    } while (0)

#endif // WIN32
#endif // NZDEBUG || NZASSERT

#define ERROR_MARSHALL_BUFSIZE (sizeof(CError))


// Produce a standard string that describes the current process or thread
class CWhoami {
public:
    CWhoami();      // Default constructor (should be only constructor needed)
    const char *TextPtr() const { return m_buf; }
private:
    char m_buf[60];
};
#define whoami() (CWhoami().TextPtr())

class CError {

    int m_code;
    bool m_restartable;
    char m_Str[MAX_ERROR_STR_LEN+1];
    char m_Obj[MAX_ERROR_OBJ_LEN+1];
    char m_src[ASRCLEN+1];
    int m_Line;
    char m_who[MAX_ERROR_WHO_LEN+1];

    static CError s_abortErr;

public:

    enum { ERRNO_BIAS = 10000 };

    CError();
    explicit CError(int code,
           bool isErrnoError=false,
           const char *str="",
           const char *object="",
           int line=0,
           const char *src="",
           const char *who="",
           bool restartable=false);

    void Print() const;
    void PrintToBuffer(char *buf, size_t bufLen) const;
    char *Text(char *pBuf, size_t size) const;

    size_t MarshallIntoBuf(void *buf, size_t len) const;
    void unmarshallFromBuf(const void *buf);

    void *operator new(size_t);
    void operator delete(void *);

    int errorCode() const;
    bool isErrnoError() const;
    bool hasNzErrorCode(int code) const;
    bool hasErrnoCode(int err_no) const;
    const char *str() const;
    const char *obj() const;
    const char *src() const;
    int line() const;
    const char *who() const;
    bool isRestartable() const;

    static void setAbortErr(const CError& err);

    static const char *errorCodeStr(int code, bool restartable);
    const char *errorCodeStr() const;

    static int packErrorCode(int code, bool is_errno);
    static void unpackErrorCode(int *code, bool *is_errno, int raw_code);
};

inline int
CError::errorCode() const
{
    return m_code;
}

inline bool
CError::isErrnoError() const
{
    return m_code >= ERRNO_BIAS;
}

inline bool
CError::hasNzErrorCode(int code) const
{
    return m_code == code && !isErrnoError();
}

inline bool
CError::hasErrnoCode(int err_no) const
{
    return m_code - ERRNO_BIAS == err_no && isErrnoError();
}

inline const char *
CError::str() const
{
    return m_Str;
}

inline const char *
CError::obj() const
{
    return m_Obj;
}

inline const char *
CError::src() const
{
    return m_src;
}

inline int
CError::line() const
{
    return m_Line;
}

inline const char *
CError::who() const
{
    return m_who;
}

inline bool
CError::isRestartable() const
{
    return m_restartable;
};

inline const char *
CError::errorCodeStr() const
{
    return CError::errorCodeStr(m_code, m_restartable);
}

inline void
CError::setAbortErr(const CError& err)
{
    s_abortErr = err;
}

extern void CError_ErrnoLog(RcSeverity severity, int code, const char *obj, int line, const char *src);
extern void CError_ErrnoLog(RcSeverity severity, int code, const int64& obj, int line, const char *src);

#if defined(__GNUC__)
extern bool CError_read_from_file(CError *err, pid_t pid);
extern void CError_write_to_file(const CError& err);

extern void CError_AssertFailed(int line, const char *src) __attribute__((noreturn));
extern void CError_AssertFailed(int line, const char *file, const char* errorStr) __attribute__((noreturn));
extern void CError_Throw(const CError& err) throw (CError) __attribute__((noreturn));
extern void CError_Throw(int code, bool isErrnoError, const char *str,
                         const char *object, int line, const char *src) throw (CError) __attribute__((noreturn));
extern void CError_Throw(int code, bool isErrnoError, const char *str, int object,
                         int line, const char *src) throw (CError) __attribute__((noreturn));
extern void CError_Throw(int code, bool isErrnoError, const char *str, unsigned object,
                         int line, const char *src) throw (CError) __attribute__((noreturn));
extern void CError_Throw(int code, bool isErrnoError, const char *str, const int64& object,
                         int line, const char *src) throw (CError) __attribute__((noreturn));
extern void CError_Throw(int code, bool isErrnoError, const char *str, const uint64& object,
                         int line, const char *src) throw (CError) __attribute__((noreturn));
extern void CError_ErrnoThrow(const char *object, int code, int line, const char *src) throw (CError) __attribute__((noreturn));
extern void CError_ErrnoThrow(int object, int code, int line, const char *src) throw (CError) __attribute__((noreturn));
extern void CError_abort(const CError& err) __attribute__((noreturn));
#elif defined(__DCC__)
# pragma no_return CError_AssertFailed
extern void CError_AssertFailed(int line, const char *src);
extern void CError_AssertFailed(int line, const char *file, const char *errorStr);
# pragma no_return CError_Throw
extern void CError_Throw(const CError& err) throw (CError);
extern void CError_Throw(int code, bool isErrnoError, const char *str,
                         const char *object, int line, const char *src) throw (CError);
extern void CError_Throw(int code, bool isErrnoError, const char *str, int object,
                         int line, const char *src) throw (CError);
extern void CError_Throw(int code, bool isErrnoError, const char *str, unsigned object,
                         int line, const char *src) throw (CError);
extern void CError_Throw(int code, bool isErrnoError, const char *str, const int64& object,
                         int line, const char *src) throw (CError);
# pragma no_return CError_ErrnoThrow
extern void CError_ErrnoThrow(const char *object, int code, int line, const char *src) throw (CError);
extern void CError_ErrnoThrow(int object, int code, int line, const char *src) throw (CError);
# pragma no_return CError_abort
extern void CError_abort(const CError& err);
#elif defined(_MSC_VER)
# pragma warning(disable : 4290)
__declspec(noreturn) extern void CError_AssertFailed(int line, const char *src);
__declspec(noreturn) extern void CError_AssertFailed(int line, const char *file, const char *errorStr);
__declspec(noreturn) extern void CError_Throw(const CError& err) throw (CError);
__declspec(noreturn) extern void CError_Throw(int code, bool isErrnoError, const char *str,
                         const char *object, int line, const char *src) throw (CError);
__declspec(noreturn) extern void CError_Throw(int code, bool isErrnoError, const char *str,
                         int object, int line, const char *src) throw (CError);
__declspec(noreturn) extern void CError_Throw(int code, bool isErrnoError, const char *str,
                         unsigned object, int line, const char *src) throw (CError);
__declspec(noreturn) extern void CError_Throw(int code, bool isErrnoError, const char *str,
                         const int64& object, int line, const char *src) throw (CError);
__declspec(noreturn) extern void CError_ErrnoThrow(const char *object, int code, int line, const char *src) throw (CError);
__declspec(noreturn) extern void CError_ErrnoThrow(int object, int code, int line, const char *src) throw (CError);
__declspec(noreturn) extern void CError_abort(const CError& err);
#else
# error unsupported compiler
#endif

#if !FOR_SPU
#if defined(__GNUC__)
# define _ATTR_PRINTF(a,b) __attribute__((format(printf, a, b)))
#else
# define _ATTR_PRINTF(a,b)
#endif
extern void OsLogError(RcSeverity severity, const CError& err, const char *fmt, ...) throw () _ATTR_PRINTF(3,4);
#endif

#endif
