#ifndef _EXTENT_DROPXID_MAP_H_
#define _EXTENT_DROPXID_MAP_H_
//======================================================================
///\file: extdropxidmap.h [ExtentId,ExtentId] ==> 'dropxid' map
//
// (C) Copyright IBM Corp. 2014  All Rights Reserved.
//======================================================================

#include <map>

#include "nzshmalloc.h"     //< for NzShmAlloc
#include "dpm_common.h"     //< for RawSid
#include "TXIDCommon.h"     //< for TXidValue

typedef std::pair<RawSid,RawSid> ExtRange;

// ASSUMPTION: This map works under the assumption that no two extent id ranges
//             in our map can contain the same extent id. This is a fair assumption
//             because an extent will never have two dropxids associated with it (only
//             the first truncate/drop transaction's id gets registered in the extent's
//             metadata).
//
// As one would expect, the range [a, b] will be considered to be 'less' than
// [c, d] if 'b' (the last extent id in the first range) is less than 'c' (the
// first extent id in the second range).
//
// The comparison logic in ExtRangeCompare will allow us to 'find' the extent
// range that contains a particular extent id. This works because find() essentially
// does two comparisons (see (1) and (2) below) to figure out whether a key exists
// in the map (note that first and second are the same in keyToBeLookedUp):
//      (1) less     = keyInMap.second < keyToBeLookedUp.first
//      (2) greater  = keyToBeLookedUp.second < keyInMap.first
//      (3) foundKey = !( less || greater ),
//                     i.e., keyInMap.second >= keyToBeLookedUp.first = keyToBeLookedUp.second >= keyInMap.first
//
// For example, if we have:
// [1000,2000] => 0x1234
// [2001,3000] => 0x5678
// [3001,4000] => 0x9ABC
// if extentId=1500, we find [1000,2000] because !( 2000 < 1500 || 1500 < 1000 )
// if extentId=3000, we find [2001,3000] because !( 3000 < 3000 || 3000 < 2001 )
// if extentId=100, we find no matches including the first range because ( 2000 < 100 || 100 < 1000 )
// if extentId=5000, we find no matches including the last range because ( 4000 < 5000 || 5000 < 3001 )
struct ExtRangeCompare
{
    bool operator() (const ExtRange& lhs, const ExtRange& rhs) const
        { return lhs.second < rhs.first; }
};

// A map with a range of extent ids as the 'key', and the dropxid associated with them as
// the 'value'.
struct ExtRangeDropXidMap : public std::map< ExtRange, TXidValue, ExtRangeCompare,
                                             NzShmAlloc<std::pair<ExtRange, TXidValue>,
                                                        mt_ext_dropxid_map>::stl_alloc >
{
    void insertRange( RawSid minExtentId, RawSid maxExtentId, TXidValue dropxid )
    {
        insert( make_pair(make_pair(minExtentId, maxExtentId), dropxid) );
    }

    TXidValue lookupDropXid( RawSid extentId ) const
    {
        if ( empty() ) {
            return 0;
        }

        const_iterator lb = find( make_pair(extentId,extentId) );
        return (lb != end()) ? lb->second : 0;
    }
};


#endif
