#if !defined( FPGAPREPAPI_INCLUDED )
#define FPGAPREPAPI_INCLUDED

#include "nzstd.h"

// fpgaprepapi.h - structs, etc.

// The following structure is used to loosely describe the physical layout
//  of a table to the FPGA. Entries must match physical field ordering rules.
// Order is approximately (see CTable::finishCreate):
//  16-byte numeric
//  12-byte structured (intervals or time-tz)
//   8-byte numeric
//   8-byte integer, time, timestamp
//   8-byte double-float
//   4-byte numeric
//   4-byte integer, date
//   4-byte float
//   4-byte structured date (unused)
//   2-byte integer
//   2-byte structured date (unused)
//   1-byte integer, bool
//   1-byte fixed-char
//   2-byte fixed-char
//  ...
//  16-byte fixed-char (16 total, whether 1-byte, 2-byte, or 4-byte characters)
//  varyings
// for a total of 12 + 16 + 1 entries or 29 max.

const static int MAX_TABLE_DESCRIPTION_ENTRIES = 32;
const static int MAX_INSTRUCTION_TABLE_ENTRIES = 32;

// Following values must match FPGA expectations!
// Used in FPGA's table descriptor
enum ESchemaType {
    SCHEMA_TYPE_INVALID     = -1,
    SCHEMA_TYPE_NUMERIC     =  0,
    SCHEMA_TYPE_INT         =  1,
    SCHEMA_TYPE_FLOAT       =  2,
    SCHEMA_TYPE_TIME        =  3,
    SCHEMA_TYPE_FIXEDCHAR   =  4,
    SCHEMA_TYPE_VARCHAR     =  5,
/*  SCHEMA_TYPE_BLOB        =  6,  (a) Not understood by FPGA (b) OBSOLETE */
    SCHEMA_TYPE_DATE        =  9,  // No longer understood by FPGA
    SCHEMA_TYPE_DTIDBITADDR = 11,
    SCHEMA_TYPE_RECADDR     = 12,
    SCHEMA_TYPE_DTID        = 13,
    SCHEMA_TYPE_CTID        = 14,
    SCHEMA_TYPE_ROWID       = 15
};

typedef struct {
    ESchemaType schemaType;
    uint8 size;    // field size - 4 bits required since largest fixed-length
                   //  field is 16 bytes long
    uint16 count;  // number of fields of that size - 12 bits required to
                   //  allow for 4096 fields of same type
    bool compiled; // this type/size/count corresponds to compiled columns
                   //  (compression only)
    bool lengthAndNullsNext; // true => length is next in on-disk format
                             // (compression only)
    bool endOfTids; // no TID data types occur after this descriptor entry
} TableDescriptionEntry;

typedef struct {
    int fieldID;
    int instructionCount;
} InstructionTableEntry;

#define IS_SPECIAL_FIELD(x) \
 ((x) == PROJECT_ROWNUM) || \
 ((x) == PROJECT_CREATETXN ) || \
 ((x) == PROJECT_DELETETXN ) || \
 ((x) == PROJECT_ROWADDR ) || \
 ((x) == PROJECT_DTIDBITADDR )
// PROJECT_BLOCKNUM has been superseded by PROJECT_PAGEID and is no longer used,
// should be removed from all code that checks for it.
// ((x) == PROJECT_BLOCKNUM )

// ### Beginning of FPGA_BOOLS bit settings ---------------------------------------

// Flag bits controlling fpga code paths.           Used in:
// --------------------------------------           --------
  // tell fpga to ignore xids
const static int FPGA_BIT_IGNORE_XIDS    = 0x000001;  // Setup
  // tell system to use 128KB fpga buffers
const static int FPGA_BIT_128K_MIN       = 0x000002;  // GetFpgaBufferSize
  // disable bv joins in FPGA
const static int FPGA_BIT_DISABLE_BVJ    = 0x000004;  // restrictWalker
  // always report number of bytes delivered by fpga
//const static int FPGA_BIT_FPGA_BYTES     = 0x000010;  // processFpgaRecordsBackwards
  // verbosity in GetFpgaBufferSize
const static int FPGA_BIT_GFBS           = 0x000020;  // GetFpgaBufferSize
  // display "first" data from fpga
const static int FPGA_BIT_DUMP_DATA      = 0x000040;  // processFpgaRecordsBackwards
  // display ALL data from fpga
const static int FPGA_BIT_DUMP_ALL       = 0x000080;  // processFpgaRecordsBackwards
  // display fpga setup details
const static int FPGA_BIT_DUMP_SETUP     = 0x000100;  // fpgawrap.cpp
  // display fpga command info
const static int FPGA_BIT_DUMP_COMMANDS  = 0x000200;  // diskDriverTask (fpgaDisk.cpp)
  // return rowAddr instead of rowNum - kludge related to bug 2649
const static int FPGA_BIT_ROWNUM_AS_ADDR = 0x000400;  // genc.cpp, plan.cpp
  // tell fpga to validate table IDs
const static int FPGA_BIT_IGNORE_TBLID   = 0x000800;  // CFpgaWrap::CFpgaWrap
  // allow fpga to do varchar comparisons - bug 9608
const static int FPGA_BIT_ENABLE_VARCHAR = 0x001000;  //
  // dump entire fpga data on errors
const static int FPGA_BIT_MEM_RANGE_ALL  = 0x002000;  // SpuDumpMemRange
  // skip raw read on FFT/FWT generation
const static int FPGA_BIT_SKIP_RAW       = 0x004000;  // SpuDumpEiffelCore (fpgawrap.cpp)
  // perform record-count sanity check prior to compiled-page writes
const static int FPGA_BIT_COMP_PAGE_CHECK= 0x008000;  // SanityCheck (recordmgr.cpp)
  // display fpga setup details on errors
const static int FPGA_BIT_DUMP_ERRORS    = 0x010000;  // SpuDumpFpgaDiagInfo
  // core on SanityCheck failures
const static int FPGA_BIT_CORE_IF_INSANE = 0x020000;  // SanityCheck
  // core on disk write errors
const static int FPGA_BIT_FWT_ON_WR_ERR  = 0x040000;  // processDiskJob (diskmgr.cpp)
  // core on disk read errors
const static int FPGA_BIT_CORE_ON_RD_ERR = 0x080000;  // processDiskJob (diskmgr.cpp)
  // exclude 'prior' read info from FFTs
const static int FPGA_BIT_SKIP_PRIOR     = 0x100000;  // spuDumpEiffelCore (fpgawrap.cpp)
  // create FFT on successful retries
const static int FPGA_BIT_DUMP_RETRIES   = 0x200000;  // SpuDumpFpgaDiagInfoWrap
  // perform record-count sanity check at exit from appendToCompressed
const static int FPGA_BIT_COMP_PG_CHK_ATC= 0x400000;  // SanityCheck (recordmgr.cpp)
  // perform record-count sanity check after each compressRows
const static int FPGA_BIT_COMP_PG_CHK_CR = 0x800000;  // SanityCheck (recordmgr.cpp)

// ### End of FPGA_BOOLS bit settings ---------------------------------------------

#endif // FPGAPREPAPI_INCLUDED
