#ifndef GENINL_INCLUDED
#define GENINL_INCLUDED

/**
 * geninl.h - include file for UDX and Generated Code
 * (C) Copyright IBM Corp. 2007, 2014  All Rights Reserved.
 */

typedef uintptr_t Datum;
typedef int (*TCmpRecFn)(void *, record_t**, record_t**); //from recordcompare.h


#ifndef FLOAT4_MAX
#define FLOAT4_MAX  3.40282347E38F
#endif

#ifndef FLOAT4_MIN
#define FLOAT4_MIN  1.17549435E-38F
#endif

#include<math.h>

#include "sqlbool.h"
#include "terror.h"
#include "fmthelpers.h"
#include "bitarch.h"

#ifndef REPLICATE_SIGN32
// copy REPLICATE_SIGN32 and HIGH_32_BITS from magic header cnumeric.h
#define REPLICATE_SIGN32(val) (((signed)(val))>>31)
#define HIGH_32_BITS(val64) (((uint64)(val64))>>32)
#endif

// macros to test for 0 divisor for divide and mod
//#define IDIV(n,d) (d == 0) ? throwError( NZ_ERROR_OVERFLOW, "Divide by 0",0) : n/d !FIX-jpb not used?
#define IMOD(n,d) ((d) == 0) ? throwError( NZ_ERROR_OVERFLOW, "Divide by 0",0) : (n)%(d)
#define NMOD(n,d) ((d) == 0) ? throwError( NZ_ERROR_OVERFLOW, "Divide by 0",0) : (n)%(d)
#define FDIV(n,d) ((d) == 0.0) ? throwError( NZ_ERROR_OVERFLOW, "Divide by 0",0) : (n)/(d)

// Overflow date: January 1, 10000 is 8000 years beyond the NPS day-zero date of
// January 1, 2000.
// The timestamp of January 1, 10000 can be computed as days per 400 years
//  * number of 400 year cycles in 8000 years (20)
//  * seconds/day (60*60*24=86400)
//  * usecs/sec (1000000)
// The number of days in 400 years is ((25*4-3)*366) + ((75*4+3)*365)
// which is to say there are 97 leap years and 303 non leap years in 400
// years. This is because only 1 of every 4 century years is a leap year.
// Since both our day-zero date and the "overflow" date are the beginning of
// 400-year cycles, more complex logic to deal with a partial cycle is
// unnecessary.
#define JANUARY_1_10000_DAY ((((25*4-3)*366) + ((75*4+3)*365)) * 20)
#define JANUARY_1_10000_TS (JANUARY_1_10000_DAY * 86400000000LL)

#define MINDAYS -539354503/* minimum number of days possible before
                           * '2000-1-1' */

#define IS_NZ_VALID_DATE(result) \
 (result < JANUARY_1_10000_DAY && result > MINDAYS)

class NzNumeric {
 public:
    int32 d[4];
    char  scale;
    char  rscale;
    char  rprecision;

    NzNumeric() { }
    inline NzNumeric(int precision_val, int scale_val, int32 n32) {
        setValue(precision_val, scale_val, n32);
    }

    inline NzNumeric(int precision_val, int scale_val, int64 n64) {
        setValue(precision_val, scale_val, n64);
    }
    inline NzNumeric(int precision_val, int scale_val, CNumeric128 const& n128) {
        setValue(precision_val, scale_val, n128);
    }

    inline void setValue(int precision_val, int scale_val, int32 n32) {
        d[0] = d[1] = d[2] = REPLICATE_SIGN32(n32);
        d[3] = n32;
        scale = rscale = scale_val;
        rprecision = precision_val;
    }

    inline void setValue(int precision_val, int scale_val, int64 n64) {
        d[0] = d[1] = REPLICATE_SIGN32(HIGH_32_BITS(n64));
        d[2] = n64 >> 32;
        d[3] = n64 & 0xffffffff;
        scale = rscale = scale_val;
        rprecision = precision_val;
    }

    inline void setValue(int precision_val, int scale_val, CNumeric128 const& n128) {
        d[0] = n128.digit[0];
        d[1] = n128.digit[1];
        d[2] = n128.digit[2];
        d[3] = n128.digit[3];
        scale = rscale = scale_val;
        rprecision = precision_val;
    }
};


//  The following items must be in sync:
//  class NzVarlena in nde/misc/geninl.h
//  class NzVarlena in nde/expr/pgwrap.h
//  struct varlena in pg/include/c.h
//  struct varattrib in pg/include/postgres.h
//  struct varattrib in udx-source/udx-impls/v2/UDX_Varargs.cpp
class NzVarlena {
 private:
    int32 vl_len;
    int32 vl_fixedlen;
    union {
    	char *vl_ptr;
    	int64 va_align;
    };

	char  vl_dat[1];

 public:
    NzVarlena() { }
    inline explicit NzVarlena(vlen_t* p) {
        setValue(p, 0);
    }

    inline NzVarlena(vlen_t* p, int len) {
        setValue(p, len);
    }

    inline NzVarlena(char* p, int len) {
        setValue(p, len);
    }

    inline void setValue(vlen_t *p, int len = 0) {
        vl_fixedlen = len;
        if ( p ) {
            vl_len = *p-sizeof(vlen_t)+VARHDRSZ;
            vl_ptr = (char*)(p+1);
        } else {
            vl_len = VARHDRSZ;
            vl_ptr = &(vl_dat[0]);
            vl_dat[0] = 0;
        }
    }

    inline void setValue(char *p, int len) {
        vl_fixedlen = 0;
        vl_len = len+VARHDRSZ;
        if ( p ) {
            vl_ptr = p;
        } else {
            vl_ptr = &(vl_dat[0]);
            vl_dat[0] = 0;
        }
    }

    inline char *VarData() { return vl_ptr ? vl_ptr : vl_dat; }
};

//  The following items must be in sync:
//  class NzDatum in nde/misc/geninl.h
//  class NzDatum in nde/expr/pgwrap.h
class NzDatum {
 private:
	//  due to several cast pointers, we use this as a base value.
	//  for example, as_int64(), as_float(), as_double(), as_CNumericXX(), ...
    int64 val;
 public:
    NzDatum() { }
    inline explicit NzDatum(Datum d) { val = d; }
    inline int8 as_int8() const { return val; }
    inline int16 as_int16() const { return val; }
    inline int32 as_int32() const { return val; }
    inline int64 as_int64() const { return val ? *(int64*)val : throwError( NZ_ERROR_ILLVAL, "NzDatum int64 error", 0 ); }
    inline bool as_bool() const { return val != 0; }
    inline float as_float() const { return val ? *(float*)val : throwError( NZ_ERROR_ILLVAL, "NzDatum float error", 0 ); }
    inline double as_double() const { return val ? *(double*)val : throwError( NZ_ERROR_ILLVAL, "NzDatum double error", 0 ); }
    inline char * as_charptr() const { return ((NzVarlena*)val)->VarData(); }
    inline TimeTzADT as_TimeTzADT() const { if (val==0) throwError( NZ_ERROR_ILLVAL, "NzDatum TimeTzADT error", 0 ); return *(TimeTzADT*)val; }
    inline Interval as_Interval() const { if (val==0) throwError( NZ_ERROR_ILLVAL, "NzDatum Interval error", 0 ); return *(Interval*)val; }
    inline CNumeric32 as_CNumeric32() const {
        if (val==0) throwError( NZ_ERROR_ILLVAL, "NzDatum CNumeric32 error", 0 );
        return CNumeric32(((NzNumeric*)val)->d[0], ((NzNumeric*)val)->d[1], ((NzNumeric*)val)->d[2], ((NzNumeric*)val)->d[3]);
    }
    inline CNumeric64 as_CNumeric64() const {
        if (val==0) throwError( NZ_ERROR_ILLVAL, "NzDatum CNumeric64 error", 0 );
        return CNumeric64(((NzNumeric*)val)->d[0], ((NzNumeric*)val)->d[1], ((NzNumeric*)val)->d[2], ((NzNumeric*)val)->d[3]);
    }
    inline CNumeric128 as_CNumeric128() const {
        if (val==0) throwError( NZ_ERROR_ILLVAL, "NzDatum CNumeric128 error", 0 );
        return CNumeric128(((NzNumeric*)val)->d[0], ((NzNumeric*)val)->d[1], ((NzNumeric*)val)->d[2], ((NzNumeric*)val)->d[3]);
    }
};

struct DateParts {
    int month;
    int day;
    int year;
    inline DateParts( void ) { };
    inline DateParts( int date ) {
        int	i, j, l, n;

        l = date + 68569 + 2451545;
        n = l < 0 ? ((4 * l)-146096) / 146097 :(4 * l) / 146097;
        l -= (146097 * n + 3) / 4;
        i = (4000 * (l + 1)) / 1461001;
        l += 31 - (1461 * i) / 4;
        j = (80 * l) / 2447;
        day = l - (2447 * j) / 80;
        l = j / 11;
        month = (j + 2) - (12 * l);
        year = 100 * (n - 49) + i + l;
    }
    inline DateParts( int imonth, int iday, int iyear ) {
        month=imonth; day=iday; year=iyear;
    }
    inline int as_date(void) {
        int m12 = (month - 14) / 12;

        return ((1461 * (year + 4800 + m12)) / 4 + (367 * (month - 2 - 12 * (m12))) / 12
                - (3 * ((year + 4900 + m12) / 100)) / 4 + day - 32075 - 2451545);
    }
};

// inline to convert timestamp to date
inline static int32 ts_to_dt( int64 ts ) {
    if ( ts >= 0 )
        return ts / 86400000000LL;
    else
        return (ts - 86399999999LL) / 86400000000LL;
}

// inline to convert timestamp to time
inline static int64 ts_to_tm2( int64 ts, bool bEnabledTime24 ) {
	int64 result = ts;

    if ( ts >= 0 ) {
        result = ts % 86400000000LL;
    } else {
        int64 val = (-ts) % 86400000000LL;
        result = val ? 86400000000LL - val : val;
    }

    if (bEnabledTime24){
    	// is an encoded 24:00:00 timestamp value ?
       if ((result +1) == 86400000000LL){
    	  // remove encoding
    	  result += 1;
       }
    }

    return result;
}

// inline to convert timestamp to time
inline static int64 ts_to_tm( int64 ts ) {

    return ts_to_tm2(ts, false);
}

// inline to truncate timestamp to days
inline static int64 trunc_ts_day( int64 ts ) {
    return ((ts >= 0 ? ts : (ts - 86399999999LL) ) / 86400000000LL) * 86400000000LL;
}

// inline to truncate timestamp to year
inline static int64 trunc_ts_year( int64 ts ) {
    return (ts >= 0 ? ts : (ts - 86399999999LL) ) / 86400000000LL;
}

// inline to truncate timestamp to hour
inline static int64 trunc_ts_hour( int64 ts, bool bEnableTime24 ) {
	if (bEnableTime24) {
		// if 24:00:00 return 24:00:00
		if ((ts + 1) % 86400000000LL == 0) {
			return ts;
		}
	}
    return ((ts >= 0 ? ts : (ts - 3599999999LL) ) / 3600000000LL) * 3600000000LL;
}

// inline to truncate timestamp to minute
inline static int64 trunc_ts_minute( int64 ts , bool bEnableTime24 ) {
	if (bEnableTime24) {
		// if 24:00:00 return 24:00:00
		if ((ts + 1) % 86400000000LL == 0) {
			return ts;
		}
	}
    return ((ts >= 0 ? ts : (ts - 59999999LL) ) / 60000000LL) * 60000000LL;
}

// inline to truncate timestamp to second
inline static int64 trunc_ts_second( int64 ts , bool bEnableTime24 ) {
	if (bEnableTime24) {
		// if 24:00:00 return 24:00:00
		if ((ts + 1) % 86400000000LL == 0) {
			return ts;
		}
	}
    return ((ts >= 0 ? ts : (ts - 999999LL) ) / 1000000LL) * 1000000LL;
}

// inline to truncate timestamp to millisecond
inline static int64 trunc_ts_millisecond( int64 ts , bool bEnableTime24 ) {
	if (bEnableTime24) {
		// if 24:00:00 return 24:00:00
		if ((ts + 1) % 86400000000LL == 0) {
			return ts;
		}
	}
    return ((ts >= 0 ? ts : (ts - 999LL) )  / 1000LL) * 1000LL;
}

// inline to truncate interval to days
inline static Interval trunc_iv_day( Interval &iv ) {
    return Interval((iv.time/86400000000LL) * 86400000000LL, iv.month);
}

// inline to truncate interval to hours
inline static Interval trunc_iv_hour( Interval &iv ) {
    return Interval((iv.time/3600000000LL) * 3600000000LL, iv.month);
}

// inline to truncate interval to minutes
inline static Interval trunc_iv_minute( Interval &iv ) {
    return Interval((iv.time/60000000LL) * 60000000LL, iv.month);
}

// inline to truncate interval to seconds
inline static Interval trunc_iv_second( Interval &iv ) {
    return Interval((iv.time/1000000LL) * 1000000LL, iv.month);
}

// inline to truncate interval to milliseconds
inline static Interval trunc_iv_millisecond( Interval &iv ) {
    return Interval((iv.time/1000LL) * 1000LL, iv.month);
}

// inline to add interval to timetz
inline static TimeTzADT timetz_pl_interval( TimeTzADT timeTz, Interval interval, bool enableTime24 ) {
    int64 hold = (timeTz.time + interval.time) % 86400000000LL;
	// check whether time is 00:00:00 or 24:00:00
	if (enableTime24 && hold == 0){
		// case 1: 24 + 0 = 24
		// case 2: 24 + 24 = 24
		if ((timeTz.time == 86400000000LL && interval.time == 0LL)
			||(timeTz.time == 86400000000LL && interval.time == 86400000000LL))
		{
			// set the time 24:00:00
			hold = 86400000000LL;
		}
	}

    return TimeTzADT(hold < 0 ? hold+86400000000LL : hold, timeTz.zone);
}

// inline to subtract interval from time
inline static TimeADT time_mi_interval( TimeADT time, Interval interval, bool enableTime24  ) {
    int64 hold = (time - interval.time) % 86400000000LL;
	// check whether time is 00:00:00 or 24:00:00
	if (enableTime24 && hold == 0){
		// case 1: 24 - 0 = 24
		if ((time == 86400000000LL && interval.time == 0LL))
		{
			// set the time 24:00:00
			hold = 86400000000LL;
		}
	}
    return TimeADT(hold < 0 ? hold+86400000000LL : hold);
}

// inline to subtract interval from timetz
inline static TimeTzADT timetz_mi_interval( TimeTzADT timeTz, Interval interval, bool enableTime24  ) {
    int64 hold = (timeTz.time - interval.time) % 86400000000LL;
	// check whether time is 00:00:00 or 24:00:00
	if (enableTime24 && hold == 0){
		// case 1: 24 - 0 = 24
		if ((timeTz.time == 86400000000LL && interval.time == 0LL))
		{
			// set the time 24:00:00
			hold = 86400000000LL;
		}
	}

    return TimeTzADT(hold < 0 ? hold+86400000000LL : hold, timeTz.zone);
}

// inline to negate interval
inline static Interval negate_interval( Interval interval ) {
    return Interval(-interval.time, -interval.month);
}

// inline to extract epoch from interval
inline static int64 interval_epoch( Interval interval ) {
	if (interval.time < 0)
       return ((interval.time-500000)/1000000)+(31557600*((int64)interval.month/12))+(30*86400*((int64)interval.month%12));
	else
       return ((interval.time+500000)/1000000)+(31557600*((int64)interval.month/12))+(30*86400*((int64)interval.month%12));
}

// inline to NOT a sql boolean
inline static eSqlBool bool_not( eSqlBool sqlBool) {
    return (sqlBool == eSqlTrue ? eSqlFalse : sqlBool == eSqlFalse ? eSqlTrue : eSqlUnknown);
}

extern "C" {
    int nzMonthDays(int year, int month); // needed by ts_pl_iv
    void SpuPrintMemoryRange(void const* bufP,size_t count, int bias, int hex);
}

// inline to add interval to timestamp
inline static int64 ts_pl_iv( int64 ts, Interval iv, bool bEnableTime24 ) {
    int resultMonth;
    int resultYear;
    int64 result;
    int64 tstmp = ts;
    bool is24h = false;

	if (bEnableTime24) {
		if (iv.time != 0) {
			// if 24:00:00 add 1 microsecond to ensure correct computations
			if ((ts + 1) % 86400000000LL == 0) {
				tstmp += 1;
				is24h = true;
			}
		}
	}

    DateParts dp(ts_to_dt(tstmp));
    int resMonth = (dp.month-1+iv.month)%12;
    if (resMonth < 0) {
        resultMonth = resMonth + 13;
        resultYear = dp.year + ((dp.month-1+iv.month)/12) - 1;
    } else {
        resultMonth = resMonth + 1;
        resultYear = dp.year + ((dp.month-1+iv.month)/12);
    }
    int monthDays = nzMonthDays(resultYear, resultMonth);
    int resultDay = dp.day > monthDays ? monthDays : dp.day;
    DateParts dpResult(resultMonth, resultDay, resultYear);
    result = (dpResult.as_date() * 86400000000LL) + ts_to_tm(tstmp) + iv.time;

	if (bEnableTime24) {
		// add offset to ensure correct encoding
		// case 1:  XX:24:00:00 + 24:00:00    offset = -1

		// if  24:00:00
		if (is24h) {
			if (iv.time > 0 && (iv.time % 86400000000LL) == 0) {
				result += -1;
			}
		}
	}

    //validating for Timestamp overflow
    if (result >= JANUARY_1_10000_TS)
        throwError( NZ_ERROR_ILLVAL, "Timestamp overflow while adding an interval", 0 );

    return result;
}

inline static int32 dt_pl_in( int32 date, int32 val)  {
    int32 result = date + val;
    if (!IS_NZ_VALID_DATE(result))
        throwError( NZ_ERROR_ILLVAL, "Date overflow while adding an integer", 0 );
    return result;
}

inline static int32 dt_mi_in( int32 date, int32 val)  {
    int32 result = date - val;
    if (!IS_NZ_VALID_DATE(result))
        throwError( NZ_ERROR_ILLVAL, "Date undeflow while subtracting an integer", 0 );
    return result;
}

inline static vlen_t *setvlen(vlen_t *start, char *end) {
    *start = end - (char*)start;
    return start;
}

// check to see if a float4 val is outside of
// the FLOAT4_MIN, FLOAT4_MAX bounds.
// raise an elog warning if it is.
static float CheckFloat4_Val(double val)
{
    if ( fabs(val) > FLOAT4_MAX )
        throwError( NZ_ERROR_ILLVAL, "Bad float4 input format -- overflow", 0 );
    if ( val != 0.0 && fabs(val) < FLOAT4_MIN )
        throwError( NZ_ERROR_ILLVAL, "Bad float4 input format -- underflow", 0 );
    return (float)val;
}

// integer downcast inlines
inline static int8 downcast1(int16 i16) {
    if ( i16 > 127 || i16 < -128 )
        throwError( NZ_ERROR_ILLVAL, "int1 overflow", 0 );
    return i16;
}

inline static int8 downcast1(int32 i32) {
    if ( i32 > 127 || i32 < -128 )
        throwError( NZ_ERROR_ILLVAL, "int1 overflow", 0 );
    return i32;
}

inline static int8 downcast1(int64 i64) {
    if ( i64 > 127 || i64 < -128 )
        throwError( NZ_ERROR_ILLVAL, "int1 overflow", 0 );
    return i64;
}

inline static int16 downcast2(int32 i32) {
    if ( i32 > 32767 || i32 < -32768 )
        throwError( NZ_ERROR_ILLVAL, "int2 overflow", 0 );
    return i32;
}

inline static int16 downcast2(int64 i64) {
    if ( i64 > 32767 || i64 < -32768 )
        throwError( NZ_ERROR_ILLVAL, "int2 overflow", 0 );
    return i64;
}

inline static int32 downcast4(int64 i64) {
    if ( i64 > 2147483647LL || i64 < -2147483648LL )
        throwError( NZ_ERROR_ILLVAL, "int4 overflow", 0 );
    return i64;
}

inline static int8 cgabsi(int8 in) {
    if (((uint8)in) == (((uint8)1)<<7))
        overflow_8_arith();
    return in >= 0 ? in : -in;
}

inline static int16 cgabsi(int16 in) {
    if (((uint16)in) == (((uint16)1)<<15))
        overflow_16_arith();
    return in >= 0 ? in : -in;
}

inline static int32 cgabsi(int32 in) {
    if (((uint32)in) == (((uint32)1)<<31))
        overflow_32_arith();
    return in >= 0 ? in : -in;
}

inline static int64 cgabsi(int64 in) {
    if (((uint64)in) == (((uint64)1)<<63))
        overflow_64_arith();
    return in >= 0 ? in : -in;
}

template <class Type>
inline static Type cgabs(Type in) {
    return Type(in < 0 ? -in : in);
}

template <class Type, class Scaled>
inline static Type cgceil(Type in, Scaled div, Scaled divminus1 ) {
    return Type(in <= 0 ? in/div : (in+divminus1)/div);
}

template <class Type, class Scaled>
inline static Type cgfloor(Type in, Scaled div, Scaled divminus1 ) {
    return Type(in >= 0 ? in/div : (in-divminus1)/div);
}

template <class Type>
inline static int cgsign(Type in) {
    return in > 0 ? 1 : in < 0 ? -1 : 0;
}

template <class lType, class rType>
inline static lType cgmodl(lType n, rType d) {
    return lType(((d) == 0) ? throwError( NZ_ERROR_OVERFLOW, "Divide by 0",0) : (n)%(d));
}

template <class lType, class rType>
inline static rType cgmodr(lType n, rType d) {
    return rType(((d) == 0) ? throwError( NZ_ERROR_OVERFLOW, "Divide by 0",0) : (n)%(d));
}

template <class lType, class rType>
inline static double cgfdiv(lType n, rType d) {
    return ((d) == 0.0) ? throwError( NZ_ERROR_OVERFLOW, "Divide by 0",0) : (n)/(d);
}

// "empty" values for some datatypes
static const Interval mtIV = Interval(0);    // Interval
static vlen_t mtVC = sizeof(vlen_t);   // Varying
static char mtFC[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

#if __GNUC__
# define __THROW throw ()
#else // old gcc or diab
# define __THROW
#endif

extern "C" {
    typedef struct FunctionCallInfoData* FunctionCallInfo;
    typedef Datum (*PGFunction) (FunctionCallInfo fcinfo);

    double fmod(double, double) __THROW;
#ifdef __GNUC__
#undef printf
#endif
#ifndef NZ_UDX_COMPILE
    int printf(char const*, ...) __THROW;
    int sprintf(char*, char const*, ...) __THROW;
    int spuloglv(int level, char const* format, ...);
    int spulog(char const*, ...);
    void *memmove(void *dest, void const* src, size_t n) throw();
    void *memset(void *dest, int c,  size_t n) throw();
    void *memcpy(void *dest, void const* src, size_t n) throw();
    double rint(double d) __THROW;
#endif
    char* TextCatFFF (char* dest, char *arg1, int len1, char *arg2, int len2);
    vlen_t* TextCatVFF (vlen_t *dest, char *arg1, int len1, char *arg2, int len2);
    vlen_t* TextCatVVF (vlen_t *dest, vlen_t *arg1, char *arg2, int len2);
    vlen_t* TextCatVVV (vlen_t *dest, vlen_t *arg1, vlen_t *arg2);
    vlen_t* TextCatVXF (vlen_t *dest, vlen_t *arg1, int len1, char *arg2, int len2);
    vlen_t* TextCatVFV (vlen_t *dest, char *arg1, int len1, vlen_t *arg2);
    vlen_t* TextCatVFX (vlen_t *dest, char *arg1, int len1, vlen_t *arg2, int len2);
    vlen_t* TextCatVXX (vlen_t *dest, vlen_t *arg1, int len1, vlen_t *arg2, int len2);
    vlen_t* TextCatVXV (vlen_t* dest, vlen_t *arg1, int len1, vlen_t *arg2);
    vlen_t* TextCatVVX (vlen_t *dest, vlen_t *arg1, vlen_t *arg2, int len2);
    vlen_t* trimFF (vlen_t *dest, char style, char *arg1, int len1, const char *arg2, int len2);
    vlen_t* trimFV (vlen_t *dest, char style, char *arg1, int len1, vlen_t *arg2);
    vlen_t* trimFX (vlen_t *dest, char style, char *arg1, int len1, vlen_t *arg2, int dlen2);
    vlen_t* trimVF (vlen_t *dest, char style, vlen_t *arg1, const char *arg2, int len2);
    vlen_t* trimVV (vlen_t *dest, char style, vlen_t *arg1, vlen_t *arg2);
    vlen_t* trimVX (vlen_t *dest, char style, vlen_t *arg1, vlen_t *arg2, int dlen2);
    vlen_t* trimXF (vlen_t *dest, char style, vlen_t *arg1, int dlen1, const char *arg2, int len2);
    vlen_t* trimXV (vlen_t *dest, char style, vlen_t *arg1, int dlen1, vlen_t *arg2);
    vlen_t* trimXX (vlen_t *dest, char style, vlen_t *arg1, int dlen1, vlen_t *arg2, int dlen2);
    void *padFF (void *dest, char style, char *arg1, int len1, int rlen, const char *arg2, int len2);
    void *padFV (void *dest, char style, char *arg1, int len1, int rlen, vlen_t *arg2);
    void *padFX (void *dest, char style, char *arg1, int len1, int rlen, vlen_t *arg2, int dlen2);
    void *padVF (void *dest, char style, vlen_t *arg1, int rlen, const char *arg2, int len2);
    void *padVV (void *dest, char style, vlen_t *arg1, int rlen, vlen_t *arg2);
    void *padVX (void *dest, char style, vlen_t *arg1, int rlen, vlen_t *arg2, int dlen2);
    void *padXF (void *dest, char style, vlen_t *arg1, int dlen1, int rlen, const char *arg2, int len2);
    void *padXV (void *dest, char style, vlen_t *arg1, int dlen1, int rlen, vlen_t *arg2);
    void *padXX (void *dest, char style, vlen_t *arg1, int dlen1, int rlen, vlen_t *arg2, int dlen2);
    vlen_t* SubstrFixed(vlen_t* dest, char *src, int srcsize, int start, int cnt);
    vlen_t* SubstrVarFixed(vlen_t* dest,  vlen_t* src, int srcsize, int start, int cnt);
    vlen_t* SubstrVar(vlen_t* dest,  vlen_t* src, int start, int cnt);
    vlen_t* num32ToVarchar(vlen_t* dest, int32 num, int scale);
    vlen_t* num64ToVarchar(vlen_t* dest, int64 num, int scale);
    vlen_t* num128ToVarchar(vlen_t* dest, TNumericData* numP, int scale);
    char* lowerF(char *dest, char *arg1, int len);
    vlen_t* lowerV(vlen_t *dest, vlen_t *arg1);
    vlen_t* lowerCLOB(vlen_t *dest, vlen_t *arg1);
    vlen_t* lowerCLOBascii(vlen_t *dest, vlen_t *arg1);
    char* upperF(char *dest, char *arg1, int len);
    vlen_t* upperV(vlen_t *dest, vlen_t *arg1);
    vlen_t* upperCLOB(vlen_t *dest, vlen_t *arg1);
    vlen_t* upperCLOBascii(vlen_t *dest, vlen_t *arg1);
    char *FixedToFixed(char *src, int srclen, char *dst, int dstlen, int truncmode);
    char *VarToFixed(vlen_t *src, char *dst, int dstlen, int truncmode);
    vlen_t *VarToVarFixed(vlen_t *src, vlen_t *dst,  int dstlen, int truncmode);
    vlen_t *FixedToVarFixed(char *src, int srclen, vlen_t *dst, int dstlen, int truncmode);
    vlen_t *VarcharToVarchar(vlen_t *src, vlen_t *dst,  int dstlen, int truncmode);
    vlen_t *FixedToVarchar(char *src, int srclen, vlen_t *dst, int dstlen, int truncmode);
    vlen_t *VarFixedToVarchar(vlen_t *src, int srclendecl, vlen_t *dst,  int dstlen, int truncmode);
    vlen_t *TextToVarchar(vlen_t *src, vlen_t *dst,  int dstlen, int truncmode);
    char *TextToFixedChar(vlen_t *srcin, char *dst,  int dstlen, int truncmode);
    vlen_t *VarFixedToClob(vlen_t* dest, vlen_t* src, int len);
    vlen_t *FixedToClob(vlen_t* dest, char* src, int len);
    vlen_t *VarcharToClob(vlen_t* dest, vlen_t* src);
    int ClobCompare(vlen_t* left, vlen_t* right);
    vlen_t* SubstrText(vlen_t* dest,  vlen_t* srcin, int start, int cnt);
    vlen_t* TextCat(vlen_t *dest, vlen_t *arg1, vlen_t *arg2);
    int DayOfYear(int y, int m, int d);
    vlen_t *ClobReverse(vlen_t *resultP, vlen_t *sourceP);
    vlen_t *ForwardTranslate(vlen_t *resultP, vlen_t *sourceP, int32 transNum, int32 frame);
    vlen_t *ClobTranslate(vlen_t *resultP, vlen_t *sourceP, vlen_t *fromStr, vlen_t *toStr);
    int32 TextCksum(vlen_t *srcP);
    int64 DatePart(int namelen, char *partname, int32 date);
    int64 TimestampPart(int namelen, char *partname, int64 ts);
    int TimePart(int namelen, char *partname, int64 time);
    int64 IntervalPart(int namelen, char *partname, Interval iv);
    int64 TimestampTrunc(int namelen, char *partname, int64 ts);
    Interval IntervalTrunc(int namelen, char *partname, Interval iv);
    Datum duration_date_sub_date(FunctionCallInfo fcinfo);
    Datum duration_date_sub_dur(FunctionCallInfo fcinfo);
    Datum duration_date_add_dur(FunctionCallInfo fcinfo);
    Datum duration_time_sub_time(FunctionCallInfo fcinfo);
    Datum duration_time_sub_dur(FunctionCallInfo fcinfo);
    Datum duration_time_add_dur(FunctionCallInfo fcinfo);
    Datum duration_ts_sub_ts(FunctionCallInfo fcinfo);
    Datum duration_ts_sub_dur(FunctionCallInfo fcinfo);
    Datum duration_ts_add_dur(FunctionCallInfo fcinfo);
    Datum duration_part(FunctionCallInfo fcinfo);
    sqlbool OverlapsTS(int64 l1, int64 l2, int64 r1, int64 r2, bool l1null, bool l2null, bool r1null, bool r2null);
    sqlbool OverlapsT(int64 l1, int64 l2, int64 r1, int64 r2, bool l1null, bool l2null, bool r1null, bool r2null);
    sqlbool OverlapsTTZ(TimeTzADT* l1, TimeTzADT* l2, TimeTzADT* r1, TimeTzADT* r2, bool l1null, bool l2null, bool r1null, bool r2null);
    int throwError(int code, char const* buf, char const* obj);
    vlen_t *nz_pgsql_version(vlen_t *dest);
    double nz_dround(double arg1);
    double nz_dtrunc(double arg1);
    double nz_dfloor(double arg1);
    double nz_dceil(double arg1);
    double nz_dsqrt(double arg1);
    double nz_dexp(double arg1);
    double nz_dlog1(double arg1);
    double nz_dlog10(double arg1);
    double nz_dpow(double arg1, double arg2);
    float  nz_fpow(float arg1, float arg2);
    double nz_drandom(void);
    double nz_dpi(void);
    int32  nz_setseed(double arg1);
    double nz_dasin(double arg1);
    double nz_dacos(double arg1);
    double nz_datan(double arg1);
    double nz_dsin(double arg1);
    double nz_dcos(double arg1);
    double nz_dtan(double arg1);
    double nz_dcot(double arg1);
    double nz_degrees(double arg1);
    double nz_radians(double arg1);
    double nz_datan2(double arg1, double arg2);
    NzDatum nz_float4_numeric(float arg1, int32 typmod);
    NzDatum nz_float8_numeric(double arg1, int32 typmod);
    NzDatum nz_float8_numericonearg(double arg1);
    vlen_t *nz_timeofday(vlen_t *dest);
    vlen_t *nz_date_varchar(vlen_t *dest, int32 arg1);
    vlen_t *nz_time_varchar(vlen_t *dest, int64 arg1);
    vlen_t *nz_timetz_varchar(vlen_t *buf, TimeTzADT ttz);
    vlen_t *nz_timestamp_varchar(vlen_t *dest, int64 arg1);
    vlen_t *nz_interval_varchar(vlen_t *buf, Interval ival);
    int64 nz_time_pl_span(int64 time, Interval *spanP);
    int64 nz_timestamp_pl_span(int64 time, Interval *spanP);
    vlen_t *nz_int1_varchar(vlen_t *dest, int8 arg1);
    vlen_t *nz_int2_varchar(vlen_t *dest, int16 arg1);
    vlen_t *nz_int4_varchar(vlen_t *dest, int32 arg1);
    vlen_t *nz_int8_varchar(vlen_t *dest, int64 arg1);
    vlen_t *nz_bool_varchar(vlen_t *dest, bool arg1);
    vlen_t *nz_float4_varchar(vlen_t *dest, float arg1);
    vlen_t *nz_float8_varchar(vlen_t *dest, double arg1);
    NzDatum nz_varchar_numeric (vlen_t *arg1, int32 typmod);
    int32 nz_clob_pos(vlen_t *arg1, vlen_t *arg2);
    int64 Factorial64(int32 i);
    void GencLimitReached(CPlan *planP);
    void GencAbortPlan(CPlan *planP);
    bool GencIsRecordVisible (CPlan *planP, int64 txCreated, int64 txDeleted);
    char *month_nameA(int mon);
    char *month_name8(int mon);
    char *Month_nameA(int mon);
    char *Month_name8(int mon);
    char *MONTH_nameA(int mon);
    char *MONTH_name8(int mon);
    int month_name_len(int mon);
    char *day_nameA(int dayno);
    char *day_name8(int dayno);
    char *Day_nameA(int dayno);
    char *Day_name8(int dayno);
    char *DAY_nameA(int dayno);
    char *DAY_name8(int dayno);
    int day_name_len(int dayno);
    int64 nzMonthsBetween(int64 ts1, int64 ts2);
    Interval nzTimestampAge(int64 ts1, int64 ts2);
    Interval nzDateAge(int32 dt1, int32 dt2);
    int32 nzAddMonthsDate(int32 dt, int32 mn);
    int64 nzAddMonthsTimestamp(int64 ts, int32 mn);
    int32 nzLastDay(int32 dt);
    int32 nzToDate1f(char *str, int len);
    char *XpUnloadDate(char *dst, int date, bool bForTimestamp);
    void XpUnpackDate(char **dst, char **src, int dateStyle, int dateDelim);
    char *XpUnloadTime(char *dst, char *halfFullP, int64 time, int tzSeconds);
    void XpUnpackTime(char **dst, char **src, int *nybUsedP, int timeStyle,
            int timeDelim, bool includezeroseconds, int decimalDelim, int meridianDelim);
    void XpUnpackTimeTz(char **dst, char **src, int *nybUsedP, int timeStyle,
            int timeDelim, bool includezeroseconds, int decimalDelim, int meridianDelim);
    char *XpUnloadTimestamp(char *dst, char *halfFullP, int64 timestamp);
    void XpUnpackTimestamp(char **dst, char **src, int *nybUsedP, int dateStyle,
            int DateDelim, int timeStyle, int timeDelim, bool includezerosecondsi, int decimalDelim, int dateTimeDelim, int meridianDelim);
    char *XpUnloadInterval(char *dst, char *halfFullP, Interval interval);
    void XpUnpackInterval(char **dst, char **src, int *nybUsedP, int timeDelim,
                          bool includezeroseconds, int decimalDelim, int meridianDelim);

    char *XpHostUnloadDate(char *dst, int date, int dateStyle, int dateDelim);
    char *XpHostUnloadTime(char *dst, int64 time, int timeStyle, int timeDelim,
                           bool includezeroseconds, int decimalDelim, int meridianDelim);
    char *XpHostUnloadTimeTz(char *dst, TimeTzADT timetz, int timeStyle,
                             int timeDelim, bool includezeroseconds, int decimalDelim, int meridianDelim);
    char *XpHostUnloadTimestamp(char *dst, int64 timestamp, int dateStyle, int dateDelim,
                                int timeStyle, int timeDelim, bool includezeroseconds, int decimalDelim, int dateTimeDelim, int meridianDelim);
    char *XpHostUnloadInterval(char *dst, Interval interval, int timeDelim,
                               bool includezeroseconds, int decimalDelim, int meridianDelim);
    void XpHandleBadChar(char delim, char ch, char const* nullValue, bool nullValFound);
    NzDatum nz_to_number(char *instr, int len, vlen_t *fmt, int scale);

    // new style pg wrap
    NzDatum NzDirectFunctionCallVarArgs(PGFunction func, int num_args, ...);
    NzDatum NzDirectFunctionCall0(PGFunction func);
    NzDatum NzDirectFunctionCall1(PGFunction func, Datum arg1);
    NzDatum NzDirectFunctionCall2(PGFunction func, Datum arg1, Datum arg2);
    NzDatum NzDirectFunctionCall3(PGFunction func, Datum arg1, Datum arg2, Datum arg3);
    NzDatum NzDirectFunctionCall4(PGFunction func, Datum arg1, Datum arg2, Datum arg3, Datum arg4);
    vlen_t *NzTextDirectFunctionCallVarArgs(vlen_t *dest, PGFunction func, int num_args, ...);
    vlen_t *NzTextDirectFunctionCall0(vlen_t *dest, PGFunction func);
    vlen_t *NzTextDirectFunctionCall1(vlen_t *dest, PGFunction func, Datum arg1);
    vlen_t *NzTextDirectFunctionCall2(vlen_t *dest, PGFunction func, Datum arg1, Datum arg2);
    vlen_t *NzTextDirectFunctionCall3(vlen_t *dest, PGFunction func, Datum arg1, Datum arg2, Datum arg3);
    vlen_t *NzTextDirectFunctionCall4(vlen_t *dest, PGFunction func, Datum arg1, Datum arg2, Datum arg3, Datum arg4);
    Datum bool_max(FunctionCallInfo fcinfo);
    Datum bpchar_max(FunctionCallInfo fcinfo);
    Datum nchar_max(FunctionCallInfo fcinfo);
    Datum float4_max(FunctionCallInfo fcinfo);
    Datum float8_max(FunctionCallInfo fcinfo);
    Datum numeric_max(FunctionCallInfo fcinfo);
    Datum int1_max(FunctionCallInfo fcinfo);
    Datum int2_max(FunctionCallInfo fcinfo);
    Datum int4_max(FunctionCallInfo fcinfo);
    Datum int8_max(FunctionCallInfo fcinfo);
    Datum varchar_max(FunctionCallInfo fcinfo);
    Datum nvarchar_max(FunctionCallInfo fcinfo);
    Datum date_max(FunctionCallInfo fcinfo);
    Datum time_max(FunctionCallInfo fcinfo);
    Datum timetz_max(FunctionCallInfo fcinfo);
    Datum timestamp_max(FunctionCallInfo fcinfo);
    Datum interval_max(FunctionCallInfo fcinfo);
    Datum varbinary_max(FunctionCallInfo fcinfo);
    Datum geometry_max(FunctionCallInfo fcinfo);
    Datum bool_min(FunctionCallInfo fcinfo);
    Datum bpchar_min(FunctionCallInfo fcinfo);
    Datum nchar_min(FunctionCallInfo fcinfo);
    Datum float4_min(FunctionCallInfo fcinfo);
    Datum float8_min(FunctionCallInfo fcinfo);
    Datum numeric_min(FunctionCallInfo fcinfo);
    Datum int1_min(FunctionCallInfo fcinfo);
    Datum int2_min(FunctionCallInfo fcinfo);
    Datum int4_min(FunctionCallInfo fcinfo);
    Datum int8_min(FunctionCallInfo fcinfo);
    Datum varchar_min(FunctionCallInfo fcinfo);
    Datum nvarchar_min(FunctionCallInfo fcinfo);
    Datum date_min(FunctionCallInfo fcinfo);
    Datum time_min(FunctionCallInfo fcinfo);
    Datum timetz_min(FunctionCallInfo fcinfo);
    Datum timestamp_min(FunctionCallInfo fcinfo);
    Datum interval_min(FunctionCallInfo fcinfo);
    Datum varbinary_min(FunctionCallInfo fcinfo);
    Datum geometry_min(FunctionCallInfo fcinfo);

    Datum int1fac(FunctionCallInfo fcinfo);
    Datum int2fac(FunctionCallInfo fcinfo);
    Datum int4fac(FunctionCallInfo fcinfo);
    Datum int8fac(FunctionCallInfo fcinfo);
    Datum varchar_int1(FunctionCallInfo fcinfo);
    Datum varchar_int2(FunctionCallInfo fcinfo);
    Datum varchar_int4(FunctionCallInfo fcinfo);
    Datum varchar_int8(FunctionCallInfo fcinfo);
    Datum float4_numeric(FunctionCallInfo fcinfo);
    Datum float8_numeric(FunctionCallInfo fcinfo);
    Datum float8_numericonearg(FunctionCallInfo fcinfo);
    Datum numeric_varchar(FunctionCallInfo fcinfo);
    Datum interval_div(FunctionCallInfo fcinfo);
    Datum interval_mul(FunctionCallInfo fcinfo);
    Datum mul_d_interval(FunctionCallInfo fcinfo);
    Datum timestamp_pl_span(FunctionCallInfo fcinfo);
    Datum timestamp_to_localtime(FunctionCallInfo fcinfo);
    Datum timestamp_mi_span(FunctionCallInfo fcinfo);
    Datum varcharpos(FunctionCallInfo fcinfo);
    Datum numeric_to_number(FunctionCallInfo fcinfo);
    Datum nnumeric_to_number(FunctionCallInfo fcinfo);
    Datum to_date(FunctionCallInfo fcinfo);
    Datum nto_date(FunctionCallInfo fcinfo);
    Datum float_round(FunctionCallInfo fcinfo);
    Datum to_timestamp(FunctionCallInfo fcinfo);
    Datum nto_timestamp(FunctionCallInfo fcinfo);
    Datum timestamp_to_char(FunctionCallInfo fcinfo);
    Datum timestamp_to_nchar(FunctionCallInfo fcinfo);
    Datum numeric_to_char(FunctionCallInfo fcinfo);
    Datum numeric_to_nchar(FunctionCallInfo fcinfo);
    Datum int1_to_char(FunctionCallInfo fcinfo);
    Datum int1_to_nchar(FunctionCallInfo fcinfo);
    Datum int2_to_char(FunctionCallInfo fcinfo);
    Datum int2_to_nchar(FunctionCallInfo fcinfo);
    Datum int4_to_char(FunctionCallInfo fcinfo);
    Datum int4_to_nchar(FunctionCallInfo fcinfo);
    Datum int8_to_char(FunctionCallInfo fcinfo);
    Datum int8_to_nchar(FunctionCallInfo fcinfo);
    Datum float4_to_char(FunctionCallInfo fcinfo);
    Datum float4_to_nchar(FunctionCallInfo fcinfo);
    Datum float8_to_char(FunctionCallInfo fcinfo);
    Datum float8_to_nchar(FunctionCallInfo fcinfo);
    Datum translate(FunctionCallInfo fcinfo);
    Datum unicodes(FunctionCallInfo fcinfo);
    Datum bytecodes(FunctionCallInfo fcinfo);
    Datum like_escape(FunctionCallInfo fcinfo);
    Datum nlike_escape(FunctionCallInfo fcinfo);
    Datum varchar_date(FunctionCallInfo fcinfo);
    Datum varchar_time(FunctionCallInfo fcinfo);
    Datum varchar_float8(FunctionCallInfo fcinfo);
    Datum varchar_float4(FunctionCallInfo fcinfo);
    Datum varchar_timetz(FunctionCallInfo fcinfo);
    Datum varchar_timestamp(FunctionCallInfo fcinfo);
    Datum varchar_interval(FunctionCallInfo fcinfo);
    Datum varchar_bool(FunctionCallInfo fcinfo);
    Datum to_hex32(FunctionCallInfo fcinfo);
    Datum to_hex64(FunctionCallInfo fcinfo);
    Datum initcap(FunctionCallInfo fcinfo);
    Datum chr(FunctionCallInfo fcinfo);
    Datum binary2hex(FunctionCallInfo fcinfo);
    Datum hex2binary(FunctionCallInfo fcinfo);
    Datum string_to_int(FunctionCallInfo fcinfo);
    Datum int_to_string(FunctionCallInfo fcinfo);
    Datum nc_int8(FunctionCallInfo fcinfo);
    Datum nc_int4(FunctionCallInfo fcinfo);
    Datum nc_int2(FunctionCallInfo fcinfo);
    Datum nc_int1(FunctionCallInfo fcinfo);
    Datum nc_float8(FunctionCallInfo fcinfo);
    Datum nc_float4(FunctionCallInfo fcinfo);
    Datum nc_numeric(FunctionCallInfo fcinfo);
    Datum nc_char(FunctionCallInfo fcinfo);
    Datum nc_varchar(FunctionCallInfo fcinfo);
    Datum nc_nchar(FunctionCallInfo fcinfo);
    Datum nc_nvarchar(FunctionCallInfo fcinfo);
    Datum nc_geometry(FunctionCallInfo fcinfo);
    Datum nc_varbinary(FunctionCallInfo fcinfo);
    Datum nc_unkbinary(FunctionCallInfo fcinfo);
    Datum nc_interval(FunctionCallInfo fcinfo);
    Datum nc_date(FunctionCallInfo fcinfo);
    Datum nc_time(FunctionCallInfo fcinfo);
    Datum nc_timetz(FunctionCallInfo fcinfo);
    Datum nc_timestamp(FunctionCallInfo fcinfo);
    Datum nc_bool(FunctionCallInfo fcinfo);
    Datum nc_unk(FunctionCallInfo fcinfo);
    Datum timestamp_zone2(FunctionCallInfo fcinfo);
    Datum timestamp_tz_izone(FunctionCallInfo fcinfo);
    Datum timestamp_tz_zone(FunctionCallInfo fcinfo);
    Datum timestamp_tz_zone2(FunctionCallInfo fcinfo);
    void SetCmpRecFnForSort( TSortNode *pNode, TCmpRecFn cmpRecFn, CTable *tbl);
    void XpSortTable(TSortNode *pNode);
    void XpSortGrowStore(TSortNode *pNode, bool lastCall);
    void overflow_64_arith();
    void ValidateRecord(CTable *tbl, void *rec);
    uint32 NzRoundRobinSpuId();
    uint32 NzChunkRandomDsId (CTable *tbl, int _rdc_size);
    int MatchText(unsigned char *t, int tlen, unsigned char *p, int plen, int tvarfixedlen, int spacepadlen_in, bool bUtf);
    int ClobMatchText(vlen_t* subjectClob, vlen_t* patternVarchar);
    void LogMaterializeNode( CPlan *pPlan, SnippetNode *pNode, int lastCall);
    void *PlanPoolAlloc(CPlan *plan, size_t size);
    vlen_t *FCharToUtf8(char *in, int len, vlen_t *out);
    vlen_t *VCharToUtf8(vlen_t *in, vlen_t *out, int colSize);
    vlen_t *Utf8ToVChar(vlen_t *in, vlen_t *out, int maxcharlen);
    vlen_t *Utf8FixedToVChar(vlen_t *in, vlen_t *out, int declen);
    char *Utf8ToFixed (vlen_t *in, char *out, int outlen);
    vlen_t *NChar(vlen_t *src, vlen_t *dst,  int dstlen, int truncmode);
    vlen_t *NCharWithLen(vlen_t *src, int srcfixedlen, vlen_t *dst,  int dstlen, int truncmode);
    int64 nz_abstime_timestamp(int64 arg1); // SWS-41266: Function similar to PG abstime_timestamp
    int64 nz_abstime_date(int64 arg1); // SWS-45332: Function similar to PG abstime_date
    int64 nznextval(CNzSequence *seq);
    int64 nznextvalues(CNzSequence *seq, int cnt);
    // For routines called from gened code, we need prototype declarations in this header.
    // Here we add prototypes for the 4 functions implementing nchar type concatenations
    vlen_t* NTextCatVV (vlen_t *dest, vlen_t *arg1, vlen_t *arg2);
    vlen_t* NTextCatXX (vlen_t *dest, vlen_t *arg1, int len1, vlen_t *arg2, int len2);
    vlen_t* NTextCatXV (vlen_t* dest, vlen_t *arg1, int len1, vlen_t *arg2);
    vlen_t* NTextCatVX (vlen_t *dest, vlen_t *arg1, vlen_t *arg2, int len2);

    vlen_t* Unichr  (vlen_t* dest, int codepoint);
    uint32 utf8ToCodepoint (unsigned char *utf8);

    vlen_t *NTrimV (vlen_t *dest, char style, vlen_t *arg1);
    vlen_t *NTrimVV (vlen_t *dest, char style, vlen_t *arg1,
                             vlen_t *arg2, int arg2Size);
    vlen_t *NTrimVF (vlen_t *dest, char style, vlen_t *arg1,
                             vlen_t *arg2, int arg2Size);
    vlen_t *NTrimF (vlen_t *dest, char style, vlen_t *arg1, int arg1Size);
    vlen_t *NTrimFFV (vlen_t *dest, char style, vlen_t *arg1, int arg1Size,
                              vlen_t *arg2, int arg2Size);
    vlen_t *NPadF (vlen_t *dest, char style, vlen_t *arg1, int arg1Size,
                       int arg2);
    vlen_t *NPadV (vlen_t *dest, char style, vlen_t *arg1, int arg2);
    vlen_t *NPadVF (vlen_t *dest, char style, vlen_t *arg1, int arg2,
                    vlen_t *arg3, int arg3Size);
    vlen_t *NPadVV (vlen_t *dest, char style, vlen_t *arg1, int arg2,
                    vlen_t *arg3);
    vlen_t *NPadFV (vlen_t *dest, char style, vlen_t *arg1, int arg1Size,
                    int arg2, vlen_t *arg3);
    vlen_t *NPadFF (vlen_t *dest, char style, vlen_t *arg1, int arg1Size,
                    int arg2, vlen_t *arg3, int arg3Size);
    vlen_t *NRepeatV (vlen_t *dest, vlen_t *src, int nrep);
    vlen_t *NRepeatF (vlen_t *dest, vlen_t *arg1, int arg1Size, int arg2);
    vlen_t *NSubStr (vlen_t *dest, vlen_t *arg1, int arg1Size, int arg2);
    vlen_t *NSubStrL (vlen_t *dest, vlen_t *arg1, int arg1Size, int arg2,
                     int arg3);
    vlen_t *NCaseChange (vlen_t *dest, bool up, vlen_t *arg1, int arg1Size);
    vlen_t *NInitCap (vlen_t *dest, vlen_t *arg1, int arg1Size);
    Datum ntranslate(FunctionCallInfo fcinfo);
    int NStrPos (vlen_t *arg1, vlen_t *arg2);
    int NStrPosF (vlen_t *arg1, int arg1Size, vlen_t *arg2, int arg2Size);
    int NInStr (vlen_t *arg1, int arg1Size, vlen_t *arg2, int arg2Size,
                int n, int m);
    int InStr (unsigned char *arg1, int len1, int fixedLen1,
               unsigned char *arg2, int len2, int fixedLen2, int n, int m);

    vlen_t *FormatColumnAccess (vlen_t *dest, uint32 usage);
    vlen_t *FormatTableAccess (vlen_t *dest, uint32 usage);

    // Fuzzy Search Functions
    vlen_t *nz_nysiis_f(vlen_t *dest, int len, char *data, CPlan *pPlan);
    vlen_t *nz_nysiis_v(vlen_t *dest, int declen, vlen_t *data, CPlan *pPlan);
    int32 nz_dbl_metaphone_f(bool null, int len, char *data, CPlan *pPlan);
    int32 nz_dbl_metaphone_v(bool null, int declen, vlen_t *data, CPlan *pPlan);
    vlen_t *nz_primary_metaphone(vlen_t *dest, int32 val);
    vlen_t *nz_secondary_metaphone(vlen_t *dest, int32 val);
    int32 nz_le_distance(bool arg1null, int arg1declen, int arg1len, char *arg1data,
                         bool arg2null, int arg2declen, int arg2len, char *arg2data, CPlan *pPlan);
    int32 nz_dle_distance(bool arg1null, int arg1declen, int arg1len, char *arg1data,
                          bool arg2null, int arg2declen, int arg2len, char *arg2data, CPlan *pPlan);
    int32 nz_score_metaphone(int32 arg1, int32 arg2, int32 strongVal,
                          int32 normalVal, int32 minorVal, int32 noneVal);
    //NOTE: this is also declared in nz_utf8.h and udxhelpers.h
    //This is also defined in nz_utf8.cpp, udxharness.cpp
    bool isValidUtf8(uchar const* bytes, int length, int* charLength);
    //\NOTE

    ////////////////////
    /// MLS security functions
    ////////////////////
    vlen_t* spu_tohexVV(vlen_t* outbin, char* inhex, uint16 actualLen, uint16 declLen);
    vlen_t* spu_fromhexVV(vlen_t* outhex, char* inbin, uint16 inlen);
    vlen_t* spu_bitmap_to_text(vlen_t* outtext, int intype, vlen_t* inbin);
    vlen_t* spu_make_label(vlen_t* outtext, int level, vlen_t* incat, vlen_t* incoh);
    vlen_t* spu_insert_category(vlen_t* outbin,
                                vlen_t* label,
                                bool labelNull,
                                vlen_t* curbits,
                                int permMask);
    vlen_t* spu_insert_cohort(vlen_t* outbin,
                              vlen_t* label,
                              bool labelNull,
                              vlen_t* curbits,
                              vlen_t* clobits,
                              bool    cloNull,
                              int permMask);
    int16   spu_insert_level(vlen_t* label, bool labelNull, int16 currentLevel, int permMask);
    vlen_t* spu_update_category(vlen_t* outbin,
                                vlen_t* label,
                                bool labelNull,
                                vlen_t* oldLabel,
                                bool    oldLabelNull,
                                int permMask);
    vlen_t* spu_update_cohort(vlen_t* outbin,
                              vlen_t* label,
                              bool labelNull,
                              vlen_t* oldLabel,
                              bool    oldLabelNull,
                              int permMask);
    int16   spu_update_level(vlen_t* label,
                             bool labelNull,
                             vlen_t* oldLabel,
                             bool    oldLabelNull,
                             int permMask);
    vlen_t* spu_combine_label(vlen_t* outlabel,
                              vlen_t* inlabel1,
                              bool    inlabel1Null,
                              vlen_t* inlabel2,
                              bool    inlabel2Null);
    int     spu_max_label(char* maxResult, vlen_t* accumLabel, vlen_t* srcLabel);
}

//NOTE: This definition is copied in nz_utf8.h, udxhelpers.h and udxharness.cpp
#define utf8len(s) (utf8Lengths[(*((uchar const*)s)) >> 4])
//\NOTE

//NOTE: This is also declared in nz_utf8.h, udxhelpers.h
//This is also defined in nz_utf8.cpp, udxharness.cpp
extern uchar utf8Lengths[];
//\NOTE

//NOTE: this definition is copied in nz_utf8.h
inline static int utf8CharCount (uchar *str, int slen)
{
    int clen = 0;
    for (uchar *send = str + slen; str < send ;clen++)  {
        str += utf8len (str);
    }
    return clen;
}
//\NOTE

// unions for storing float and double literals via 32- and 64-bit ints
union floatLiteral {
    int32  asint32;
    float  asfloat;
    inline floatLiteral(const int32 val) { asint32 = val; };
};
union doubleLiteral {
    int64  asint64;
    double asdouble;
    inline doubleLiteral(const int64 val) { asint64 = val; };
};

inline static double nz_fmod( double val, double base ) {
    if ( base == 0.0 )
        throwError(NZ_ERROR_MATH, "divide by zero", 0);
    return fmod( val, base );
}

inline static int32 factorial_arg(int16 i) {
    if ( i < 0 )
        throwError(NZ_ERROR_MATH, "factorial argument negative", 0);
    if ( i > 20 )
        throwError(NZ_ERROR_MATH, "factorial overflow", 0);
    return (int32)i;
}

inline static int32 factorial_arg(int32 i) {
    if ( i < 0 )
        throwError(NZ_ERROR_MATH, "factorial argument negative", 0);
    if ( i > 20 )
        throwError(NZ_ERROR_MATH, "factorial overflow", 0);
    return i;
}

inline static int32 factorial_arg(int64 i) {
    if ( i < 0 )
        throwError(NZ_ERROR_MATH, "factorial argument negative", 0);
    if ( i > 20 )
        throwError(NZ_ERROR_MATH, "factorial overflow", 0);
    return (int32)i;
}



// Coding idiom comment:
// ---------------------
//  The construct
//
//     do
//         diff = XXXX;
//     while ( --cnt != 0 && diff == 0 );
//
// is particularly well rendered in the PPC instruction set.  On other
// architectures it might be more efficient to express this as
//
//     do
//         if ( diff = XXXX ) break;
//     while ( --cnt != 0 );
//
// Similarly, pre-incrementing of pointers models PPC load-with-update
// and store-with-update addressing.


#if __GNUC__
#define BytewiseMemcpy(d,s,n) __builtin_memcpy(d,s,n)
#define ShortwiseMemcpy(d,s,n) __builtin_memcpy(d,s,n)
#define IntwiseMemcpy(d,s,n) __builtin_memcpy(d,s,n)
#else
inline static void BytewiseMemcpy(void *in_dst, void *in_src, int in_cnt) {
    register int cnt = in_cnt;
    register char *dst = (char*)in_dst-1;
    register char *src = (char*)in_src-1;
    if ( cnt ) {
        do
            *++dst = *++src;
        while ( --cnt != 0 );
        }
}
inline static void IntwiseMemcpy(void *in_dst, void *in_src, uint32 in_cnt) {
    register int cnt = (in_cnt+3)>>2;
    register uint32 *dst = (uint32*)in_dst-1;
    register uint32 *src = (uint32*)in_src-1;
    if ( cnt ) {
        do
            *++dst = *++src;
        while ( --cnt != 0 );
    }
}
inline static void ShortwiseMemcpy(void *in_dst, void *in_src, uint32 in_cnt) {
    register int cnt = (in_cnt+1)>>1;
    register uint16 *dst = (uint16*)in_dst-1;
    register uint16 *src = (uint16*)in_src-1;
    if ( cnt ) {
        do
            *++dst = *++src;
        while ( --cnt != 0 );
    }
}
#endif

/**********************************************************************
function:  ByteCopyNoSpaces - copy from src to dst skipping
the trailing spaces in src
Returns the number of bytes copied
*/
inline static int ByteCopyNoSpaces(void *in_dst, void *in_src, int in_cnt, unsigned char space) {
    //skip spaces
    register char *src = (char*)in_src+in_cnt;

    while (src >= in_src)   {
        if (*--src != space)
            break;
    }
    //src points to last non space character
    register int cnt = (uint8*)src - (uint8*)in_src+1;
    if (cnt)   {
        register char *dst = (char*)in_dst+cnt-1;
        do
            *dst-- = *src--;
        while ( src >= in_src );
    }
    return cnt;
}


inline static void BytewiseMemClear(void *in_dst, int in_cnt) {
    register int cnt = in_cnt;
    register char *dst = (char*)in_dst-1;
    if ( cnt ) {
        do
            *++dst = 0;
        while ( --cnt != 0 );
        }
}

inline static void BytewiseMemSet(void *in_dst, int val, int in_cnt) {
    register int cnt = in_cnt;
    register char *dst = (char*)in_dst-1;
    if ( cnt ) {
        do
            *++dst = val;
        while ( --cnt != 0 );
        }
}


inline static int BytewiseMemcmp(void const* in_s1, void const* in_s2, int in_cnt) {
    register int diff = 0;
    register unsigned char *s1 = (unsigned char*)in_s1-1;
    register unsigned char *s2 = (unsigned char*)in_s2-1;
    register int cnt = in_cnt;
    if ( cnt ) {
        do
            diff = *++s1 - *++s2;
        while ( --cnt != 0 && diff == 0 );
        }
    return diff;
}

// code generation guarantees that in_s1Extra is non-zero
inline static int BytewiseMemcmpS1Longer(void *in_s1, void *in_s2, int in_s2Len, int in_s1Extra, unsigned char space) {
    register int cnt = in_s2Len;
    register unsigned char *s1 = (unsigned char*)in_s1-1;
    register unsigned char *s2 = (unsigned char*)in_s2-1;
    register int diff = 0;

    if ( cnt ) {
        do
            diff = *++s1 - *++s2;
        while ( --cnt != 0 && diff == 0 );
        }
    if ( ! diff ) {
        cnt = in_s1Extra;
        do
            diff = *++s1 - space;
        while ( --cnt != 0 && diff == 0 );
        }
    return diff;
}


// code generation guarantees that in_s2Extra is non-zero
inline static int BytewiseMemcmpS2Longer(void *in_s1, void *in_s2, int in_s1Len, int in_s2Extra, unsigned char space) {
    register int cnt = in_s1Len;
    register unsigned char *s1 = (unsigned char*)in_s1-1;
    register unsigned char *s2 = (unsigned char*)in_s2-1;
    register int diff = 0;

    if ( cnt ) {
        do
            diff = *++s1 - *++s2;
        while ( --cnt != 0 && diff == 0 );
        }
    if ( diff == 0 ) {
        cnt = in_s2Extra;
        do
            diff = space - *++s2;
        while ( --cnt != 0 && diff == 0 );
        }
    return diff;
}

inline static int TextCmpFF(char *arg1, int len1, char *arg2, int len2) {
    if ( len1 == len2 )
        return(BytewiseMemcmp(arg1, arg2, len1));
    else if ( len1 > len2 )
        return(BytewiseMemcmpS1Longer(arg1, arg2, len2, len1-len2, SPACE)); //SBC
    else
        return(BytewiseMemcmpS2Longer(arg1, arg2, len1, len2-len1, SPACE)); //SBC
}

// either or both of s1Len and s2Len may be zero
inline static int BytewiseMemcmpVarying(void const*in_s1, void const*in_s2, int in_s1Len, int in_s2Len, unsigned char space) {
    register int cnt = in_s1Len;
    register int lenDiff = in_s2Len - cnt;
    register unsigned char *s1 = (unsigned char*)in_s1-1;
    register unsigned char *s2 = (unsigned char*)in_s2-1;
    register int diff = 0;

    if ( lenDiff >= 0 ) { // equal or s2Len is larger
        if ( cnt ) { // s1Len non-zero
            do
                diff = *++s1 - *++s2;
            while ( --cnt != 0 && diff == 0 );
            }
        if ( diff ) goto RESULT_IS_KNOWN;
        if ( (cnt = lenDiff) != 0 ) { // s1Len not equal s2Len
            do
                diff = space - *++s2;
            while ( --cnt != 0 && diff == 0 );
            }
        }
    else {
        if ( (cnt = in_s2Len) != 0 ) {
            do
                diff = *++s1 - *++s2;
            while ( --cnt != 0 && diff == 0 );
            }
        if ( diff ) goto RESULT_IS_KNOWN;
        if ( (cnt = - lenDiff) != 0 ) {
            do
                diff = *++s1 - space;
            while ( --cnt != 0 && diff == 0 );
            }
        }
RESULT_IS_KNOWN:
    return diff;
}

// either or both of s1Len and s2Len may be zero
inline int BytewiseMemcmpVaryingNoPad(void const*in_s1, void const*in_s2, int in_s1Len, int in_s2Len) {
    register int cnt = 0;
    register unsigned char *s1 = (unsigned char*)in_s1-1;
    register unsigned char *s2 = (unsigned char*)in_s2-1;
    register int diff = 0;
    while (cnt < in_s1Len && cnt < in_s2Len)
    {
        diff = *++s1 - *++s2;
        if ( diff ) return diff;
        else cnt++;
    }
    return (in_s1Len - in_s2Len);
}

// BytewiseMemcmpVarying wrappers to avoid repeated GetVarFieldPtr() calls
inline static int BytewiseMemcmpFV(void const*f1, void const*v2, int f1Len, unsigned char space) {
    return BytewiseMemcmpVarying(f1, 1+(vlen_t*)v2, f1Len, *(vlen_t*)v2-sizeof(vlen_t), space);
}

inline static int BytewiseMemcmpVF(void const*v1, void const*f2, int f2Len, unsigned char space) {
    return BytewiseMemcmpVarying(1+(vlen_t*)v1, f2, *(vlen_t*)v1-sizeof(vlen_t), f2Len, space);
}

inline static int BytewiseMemcmpVV(void const*v1,void const*v2,unsigned char space) {
    return BytewiseMemcmpVarying(1+(vlen_t*)v1, 1+(vlen_t*)v2, *(vlen_t*)v1-sizeof(vlen_t), *(vlen_t*)v2-sizeof(vlen_t),space);
}

inline int BytewiseMemcmpVVNoPad(void const*v1,void const*v2) {
    return BytewiseMemcmpVaryingNoPad(1+(vlen_t*)v1, 1+(vlen_t*)v2, *(vlen_t*)v1-sizeof(vlen_t), *(vlen_t*)v2-sizeof(vlen_t));
}

// use inline functions to bind the base addresses of string arguments
inline static int BytewiseCmp2Chars(char *in_s1, void *in_s2) {
    register unsigned char *s1 = (unsigned char*)in_s1;
    register unsigned char *s2 = (unsigned char*)in_s2;
    register int diff;
    diff = s1[0] - s2[0];
    if ( diff ) goto RESULT_IS_KNOWN;
    diff = s1[1] - s2[1];
RESULT_IS_KNOWN:
    return diff;
}

inline static int BytewiseCmp3Chars(char *in_s1, void *in_s2) {
    register unsigned char *s1 = (unsigned char*)in_s1;
    register unsigned char *s2 = (unsigned char*)in_s2;
    register int diff;
    diff = s1[0] - s2[0];
    if ( diff ) goto RESULT_IS_KNOWN;
    diff = s1[1] - s2[1];
    if ( diff ) goto RESULT_IS_KNOWN;
    diff = s1[2] - s2[2];
RESULT_IS_KNOWN:
    return diff;
}

inline static int BytewiseCmp4Chars(char *in_s1, void *in_s2) {
    register unsigned char *s1 = (unsigned char*)in_s1;
    register unsigned char *s2 = (unsigned char*)in_s2;
    register int diff;
    diff = s1[0] - s2[0];
    if ( diff ) goto RESULT_IS_KNOWN;
    diff = s1[1] - s2[1];
    if ( diff ) goto RESULT_IS_KNOWN;
    diff = s1[2] - s2[2];
    if ( diff ) goto RESULT_IS_KNOWN;
    diff = s1[3] - s2[3];
RESULT_IS_KNOWN:
    return diff;
}


// (jsy) this is REALLY nasty -- varying fields are declared as vlen_t Fn[0]
//       (i.e. an array of 2 byte objects) while this function returns a char*
//       (i.e. an array of 1 byte objects); both are used when we want the
//       address of a field but behave differently once we involve them in
//       address arithmetic (e.g. subscripting).
inline static char *GetVarFieldPtr(record_t *recP, int varDex, int firstVarOffset) {
    register char *lenP = (char *)recP + firstVarOffset;
    for (int ctr = 0; ctr < varDex; ctr++)
        lenP += ALIGN_VARLENS(*(vlen_t *)lenP);
    return lenP;
}

inline static int GetDiskLengthOfSpecials (TableShape const* shape) {
    return( shape->m_diskSpecials.rowId ? sizeof(TRowId) : 0 )
        + ( shape->m_diskSpecials.txCreated ? sizeof(TXid) : 0 )
        + ( shape->m_diskSpecials.txDeleted ? sizeof(TXid) : 0 );
}

inline static int GetLengthOfSpecials (TableShape const* shape) {
    return( shape->m_leaderSize - shape->m_specialsOffset );
}

inline static void *GetDiskFieldAt(TableShape const* shape, char *recordP, int fieldOffset, bool fixedSize)
{
    register char* lenP;
    register int ctr;
    recordP += GetDiskLengthOfSpecials(shape);
    int skipMe = GetLengthOfSpecials(shape);
    if ( fixedSize )
        return (char *)recordP + fieldOffset - skipMe;
    return GetVarFieldPtr(recordP, fieldOffset, shape->m_fixedFieldsSize-skipMe);
}

inline static bool allspace(char *s, int start, int end, unsigned char space)
{
    register int i = start;

    while (i < end)
        if (s[i++]!=space) return false;

    return true;
}

#if __PPC__
#define getUInt64(p) *((uint64*)(p))
#define intelFixup(p)
#define intelFixup128(p)
#else
#define getUInt64(p) (((uint64)*((uint32*)(p))) << 32) | ((uint64)*((uint32*)(p)+1))
inline static void intelFixup(void *vp) {
    uint32 *p, val;
    p = (uint32*)vp;
    val = *p;
    *p = *(p+1);
    *(p+1) = val;
}
// !FIX-bz? should fixup the high order 64 bits, too.  But we are assuming they are not overflowed into...
inline static void intelFixup128(void *vp) {
    uint32 *p, val;
    p = (uint32*)vp + 2;
    val = *p;
    *p = *(p+1);
    *(p+1) = val;
}
#endif
// BADPTR is used to initialize record pointers to:
//  1) avoid "might be used before set" warnings in generated code
//  2) trap at runtime if the hypothetical actually happens
// NOTE: There is a comment regarding this macro in nz/src/nde/firm/mmu_spu.cpp.
//       If you modify this macro, please update that comment as well.

#define BADPTR(x) (x)0xDEADBEEF

#define getUInt64of128(p) getInt64((uint32*)(p)+2)
#define getInt64(p) (int64)(getUInt64(p))
#define getInt64of128(p) (int64)(getUInt64of128(p))
#define putInt64(p) *((uint64*)(p))
#define putInt64of128(p) putInt64((uint32*)(p)+2)
#define signExtendInt64of128(p) if (*((int64*)((int32*)(p)+2)) < 0) putInt64(p) = 0xFFFFFFFFFFFFFFFFLL; else putInt64(p) = 0

#define LikeMatch(t,u,p,l,d, utf) (1 == MatchText((t),(u),(p),(l),(d), 0, (utf)))
#define ClobLike(subjectClob, patternVarchar) (1 == ClobMatchText((subjectClob), (patternVarchar)))

#define StrLit(i) (plan->m_stringLiterals[i] + sizeof(vlen_t))
#define VStrLit(i) (vlen_t*)(plan->m_stringLiterals[i])
#define I32Lit(i) (plan->m_intLiterals[i])
#define I8Lit(i) (plan->m_int8Literals[i])
#define I16Lit(i) (plan->m_int16Literals[i])
#define I64Lit(i) (plan->m_int64Literals[i])
#define F4Lit(i) (plan->m_float4Literals[i])
#define F8Lit(i) (plan->m_float8Literals[i])
#define N1Lit(i) (plan->m_intLiterals[i])
#define N2Lit(i) (plan->m_int64Literals[i])
#define TzLit(i) (plan->m_timetzLiterals[i])
#define IvLit(i) (plan->m_timetzLiterals[i])
#define N4Lit(i) (plan->m_numeric4Literals[i])
#define BLit(i) (plan->m_boolLiterals[i])
#define CbLit(i) (vlen_t*)(plan->m_clobLiterals[i])

// spubind.cpp
extern "C" {
    void * OpenParamTable(CPlan * pPlan, int subplanNum);
    void GotHere (int line);
    void LogGeneric(const CPlan *plan, int num_args, ...);
    void ResetExprPool (CPlan *plan);
    void ValidateExprPool (CPlan *plan);
    bool EvalExpression (SnippetNode *pNode,void **fields,int dex,
                         int exprType, void *valBuf, int valLen);
    bool ExecJoinQual (void *qual,SnippetNode *pNode,void **flowFields,
                       CTable *tempTbl,void **tempFields);
    bool ExecRestrict (List *qual, SnippetNode *node, CTable *pTable,
                       void **fields);
    void gencDropTable(TDropTableNode* dropTableNode);
    void planExecDone (CPlan *plan, int sentMadeLast);
};

#define GET_NEXT_REC(rs,rec,kind,block)                         \
    (((rec+1) < (kind*)MBI_WRITE_BYTE(block)) ?                 \
        (rec+1) :                                               \
        (block = rs->NextBlockInMem(block), (kind*)MBI_ZERO_BYTE(block)))

#define GET_NEXT_REC_BLOCKNUM(rs,rec,kind,block,setBlock,incBlockNum)       \
    (((rec+1) < (kind*)MBI_WRITE_BYTE(block())) ?                 \
        (rec+1) :                                               \
        (setBlock( rs->NextBlockNumInMem(block(), incBlockNum()) ), (kind*)MBI_ZERO_BYTE(block())))

/////////////////////////////////////////////////////////////////////////////
// GET_NEXT_MRG_REC -  keeps blocks lowBlockNum to blockNum in memory
/////////////////////////////////////////////////////////////////////////////

#define GET_NEXT_MRG_REC(rec,kind,block,store,lowBlock,blockNum) \
( ((rec+1) < (kind*)MBI_WRITE_BYTE(block)) ?  \
  (rec+1) :  \
  ( ++blockNum,  \
    block = store->UpdateWindow(lowBlock,blockNum,block),  \
    (kind*)MBI_ZERO_BYTE(block) ) )

#define SIZE_ZERO_ERROR(recP) \
( sprintf(str,"scan-loop: zero-length encountered (recP=%p)", recP), \
  throwError(NZ_FATAL_ASSERTION,str,""), \
  0)

#define SIZE_WORD_CHECK(recP) ( (recP->size)==0 ? SIZE_ZERO_ERROR(recP) : (recP->size) )

#define GET_NEXT_MRG_VAR_REC(rec,kind,block,store,lowBlock,blockNum)\
( ((kind*)((char*)rec + SIZE_WORD_CHECK(rec)) < (kind*)MBI_WRITE_BYTE(block)) ? \
  (kind*)((char*)rec + (rec->size)) :  \
  ( ++blockNum,  \
    block = store->UpdateWindow(lowBlock,blockNum,block),  \
    (kind*)MBI_ZERO_BYTE(block) ) )

// WindowAggr window trimmer
#define TRIM_STALE_WINDOW_BLOCK(store,was,setwas,now) \
    do { \
        if ((now) > (was)) { \
             (store)->DeleteWindowBlock((was),(now)); \
             (setwas)( (now) ); \
        } \
    } while (0)

#ifndef RECORDSTORE_INCLUDED

#define GET_NEXT_VAR_REC(rs,rec,kind,block)                     \
  (((kind*)((char*)rec + SIZE_WORD_CHECK(rec)) < (kind*)MBI_WRITE_BYTE(block)) ? \
        (kind*)((char*)rec + (rec->size)) :                     \
        (block = rs->NextBlockInMem(block), (kind*)MBI_ZERO_BYTE(block)))

#define GET_NEXT_VAR_REC_BLOCKNUM(rs,rec,kind,block,setBlock,incBlockNum)   \
  (((kind*)((char*)rec + SIZE_WORD_CHECK(rec)) < (kind*)MBI_WRITE_BYTE(block())) ? \
        (kind*)((char*)rec + (rec->size)) :                     \
        (setBlock( rs->NextBlockNumInMem(block(), incBlockNum()) ), (kind*)MBI_ZERO_BYTE(block())))

#define LAST_REC(kind) ((kind*)BYTES_IN_NULLSTORE) /* result of offsetting into null block */

#define LOOP_RECS_VARIABLE(rs,rec,kind,block)                   \
        if ( block && block->used )                             \
            for ( rec = (kind*)MBI_ZERO_BYTE(block);            \
                rec != LAST_REC(kind);                          \
                    rec = GET_NEXT_VAR_REC(rs,rec,kind,block) )

#endif // !RECORDSTORE_INCLUDED

#define LOOP_RECS(rs,rec,kind,block)                            \
        if ( block && block->used )                             \
            for ( rec = (kind*)MBI_ZERO_BYTE(block);            \
                rec != LAST_REC(kind);                          \
                    rec = GET_NEXT_REC(rs,rec,kind,block) )

#define FIRST_BLOCK(table) ((table->PlanRecordStore(plan)) ? \
                            (table->PlanRecordStore(plan)->i_beginBlock()) : 0)
#define FIRST_RSBLOCK(rs) ((rs) ? ((rs)->i_beginBlock()) : 0)

#define FIRST_REC(block,kind)   ((block && block->used) ? ((kind*)MBI_ZERO_BYTE(block)) : LAST_REC(kind))

#define JOINED_IN_BIT 1   /* set lsb in hash table when record 'used' */
#define LAST_RECP_BIT 2   /* set in bucket pointer to singleton or in recP vector to mark end */


#define UPDATE_WRITE_ADDR(m,in_cnt)    \
    do {                               \
        register int cnt = in_cnt;     \
        m->used += cnt;                \
        m->avail -= cnt;               \
        /*printf( "U_W_A: used=%d, avail=%d\n", m->used, m->avail );*/ \
        if ( m->avail < 0 ) {          \
            throwError(NZ_FATAL_ASSERTION,"Buffer overrun in UPDATE_WRITE_ADDR","" ); \
        }                              \
    } while (0)

// !NOTE-bmz get rid of silly 0xffffffff convention on genVerifySpace!
// !FIX-dlm: computation of "max disk record size" is bogus, but what
//           it says is that it's maxRecordSize - memory-specials-size + 24
//           where 24 is the assumed size of disk specials.
#define VERIFY_FSPACE(dst,kind,rs,m,maxSize,node,extendFunc)               \
    do {                                                                   \
        /*printf( "V_FS: maxSize=%d, avail=%d\n", maxSize, m->avail );*/   \
        if ( maxSize > m->avail ) {                                        \
            node->extendFunc(false);                                       \
            dst = (kind*)rs->VerifySpace(maxSize);                         \
            m = rs->i_lastBlock();                                         \
        } else                                                             \
            dst = (kind*)MBI_WRITE_BYTE(m);                                \
    } while (0)

#define VERIFY_SPACE(dst,kind,rs,m,maxSize)                                \
    do {                                                                   \
        if ( maxSize > m->avail ) {                                        \
            dst = (kind*)rs->VerifySpace(maxSize);                         \
            m = rs->i_lastBlock();                                         \
        } else                                                             \
            dst = (kind*)MBI_WRITE_BYTE(m);                                \
    } while (0)


#define GET_SPACE_FOR_AGGREGATE(dst,kind,rs,m,maxSize,node,hashslot,hashobj,hashnum,extendFunc) \
    do {                                                                    \
        if ( maxSize > m->avail ) {                                         \
            node->extendFunc(false, (void***) &hashslot, hashobj, hashnum); \
            dst = (kind*)rs->VerifySpace(maxSize);                          \
            m = rs->i_lastBlock();                                          \
            *hashslot = dst;                                                \
        } else {                                                            \
            dst = (kind*)MBI_WRITE_BYTE(m);                                 \
            *hashslot = dst;                                                \
        }                                                                   \
    } while (0)


// This is obsolete.  It has been superseded by GET_SPACE_FOR_AGGREGATE.
#define VERIFY_AGGR_FSPACE(dst,kind,rs,m,maxSize,node,extendFunc,dst2,kind2,rs2, m2)  \
    do {                                                                   \
        if ( maxSize > m->avail ) {                                        \
            node->extendFunc(false);                                       \
            dst = (kind*)rs->VerifySpace(maxSize);                         \
            m = rs->i_lastBlock();                                         \
            m2 = rs2->i_lastBlock();                                       \
            dst2 = (kind2*)MBI_WRITE_BYTE(m2);                             \
        } else                                                             \
            dst = (kind*)MBI_WRITE_BYTE(m);                                \
    } while (0)

#define MASK_JOINED_IN_BIT(x,kind) \
((kind *)(((unsigned long)x) & ~((unsigned long)JOINED_IN_BIT)))

#define SET_JOINED_IN_BIT(x, kind) \
((kind *)(((unsigned long)x) | ((unsigned long)JOINED_IN_BIT)))

#define IS_JOINED_IN_BIT_SET(x) \
(((unsigned long)x) & ((unsigned long)JOINED_IN_BIT))

#define GET_RS_BIT_VECTOR(rs)\
( (rs) ? ((uint8 *)(rs->RecordZeroP())) : 0)

#if 0 // unused?
#define VERIFY_FWRITEHASH_SPACE(dst,kind,stor,m, node, extendFunc) \
    do { \
        if (sizeof(kind) > m->avail) {              \
            node->extendFunc(false);                \
            dst = (kind*)stor->VerifySpace(m->used);\
            m = stor->i_lastBlock();                \
        } else                                      \
            dst = (kind*)MBI_WRITE_BYTE(m);         \
    } while (0)

//
// BV_SETBIT - get the byte offset
// set a bit in a bit vector in (byte swapped)
// for eg: if you need to set a bit for value 3, you would normally turn on
// bit 3 in byte 0. BV_SETBIT will turn on bit 3 in byte 3
//
// compute the byteoffset like you do normally (val/8) called this b
// let x be the remainder when you divide b by 4
// swapped byte is (b - x) + (3 - x)
//
#define BV_SWAPPED_BITSET(pBytes,val) \
( ((unsigned char *)(pBytes))[ (((val) >> 3)&~0x3) + 3 - (((val)>>3)&0x3) ]  |= (1 << ((val) & 0x7)))

#endif // #if 0 // unused?

///////////////////////////////////////////////////////////////////////////////
//  BYTE SWAPPING ROUTINES
///////////////////////////////////////////////////////////////////////////////

//The RSWAP_XXX routines return the swapped value
//The SWAP_XXX routines swap the passed value and write it to passed address

#if __PPC__    /* { */
#if __DCC__    /* { */
//return the byte swapped 16-bit word
asm short RSWAP_INT16(short& ref)
{
% reg ref
!
    lhbrx  r3,r0,ref
}

//return the byte swapped 32-bit word
asm int RSWAP_INT32(int& ref)
{
% reg ref
!
    lwbrx  r3,r0,ref
}

//return the byte swapped 64-bit word
asm int64 RSWAP_INT64(int64& ref)
{
% reg ref
! "r0"
    addi    r0,ref,4
    lwbrx   r4,r0,ref
    lwbrx   r3,r0,r0
}

//byteswap the passed 16-bit value and assign it to dst
asm void SWAP_INT16(short* dst, short value)
{
% reg dst, value
!
    sthbrx  value,r0,dst
}

asm void SWAP_UINT16(unsigned short* dst, unsigned short value)
{
% reg dst, value
!
    sthbrx  value,r0,dst
}

//byteswap the passed 32-bit value and assign it to dst
asm void SWAP_INT32(int* dst, int value)
{
% reg dst, value
!
    stwbrx  value,r0,dst
}

//byteswap the passed float value and assign it to dst
asm void SWAP_FLOAT(float* dst, float value)
{
% reg dst, value
!
    stwbrx  value,r0,dst
}

//byteswap the passed 64-bit value and assign it to dst
asm void SWAP_INT64(int64* dst, int64 value)
{
% reg dst, value
! "r0"
    stwbrx  value!L,r0,dst
    addi    r0,dst,4
    stwbrx  value!H,r0,r0
}

//byteswap the passed double value and assign it to dst
asm void SWAP_DOUBLE(double* dst, double value)
{
% reg dst, value
! "r0"
    stwbrx  value!L,r0,dst
    addi    r0,dst,4
    stwbrx  value!H,r0,r0
}
#endif  /* } */
#if __GNUC__    /* { */
//return the byte swapped 16-bit word
static __inline__ short RSWAP_INT16(short& ref)
{
    register short rx;
    __asm__ (
"   lhbrx  %[out],r0,%[refp]"
    : [out] "=r" (rx)
    : [refp] "r" (&ref), [refMem] "m" (ref)
    : );
    return rx;
}

//return the byte swapped 32-bit word
static __inline__ int RSWAP_INT32(int& ref)
{
    register int sx;
    __asm__ (
"   lwbrx  %[out],r0,%[refp]"
    : [out] "=r" (sx)
    : [refp] "r" (&ref), [refMem] "m" (ref)
    : );
    return sx;
}

//return the byte swapped 64-bit word
static __inline__ int64 RSWAP_INT64(int64& ref)
{
    int uh, lh;
    __asm__ (
"   lwbrx   %[rlh],0,%[refp]\n\
    lwbrx   %[ruh],%[refp],%[off]"
    : [ruh] "=r" (uh), [rlh] "=&r" (lh)
    : [refp] "b" (&ref), [off] "r" (4), [refMem] "m" (ref)
    : );
    return ((int64)uh << 32) | lh ;
}

//byteswap the passed 16-bit value and assign it to dst
static __inline__ void SWAP_INT16(short* dst, short value)
{
    __asm__ (
"   sthbrx  %[rv],0,%[rdst]"
    : [shortMem] "=m" (*dst)
    : [rv] "r" (value), [rdst] "r" (dst)
    : );
}

static __inline__ void SWAP_UINT16(unsigned short* dst, unsigned short value)
{
    __asm__ (
"   sthbrx  %[rv],0,%[rdst]"
    : [shortMem] "=m" (*dst)
    : [rv] "r" (value), [rdst] "r" (dst)
    : );
}

//byteswap the passed 32-bit value and assign it to dst
static __inline__ void SWAP_INT32(int* dst, int value)
{
    __asm__ (
"   stwbrx  %[rv],0,%[rdst]"
    : [intMem] "=m" (*dst)
    : [rv] "r" (value), [rdst] "r" (dst)
    : );
}

//byteswap the passed 64-bit value and assign it to dst
static __inline__  void SWAP_INT64(int64* dst, int64 value)
{
    register int lh = (int) value;  /* yes, chop it! */
    register int uh = (int) (value >> 32);
    __asm__ (
"   stwbrx %[rlh],0,%[rdst]\n\
    stwbrx %[ruh],%[rdst],%[off]"
    : [stMem] "=m" (*dst)
    : [rdst] "b" (dst), [off] "r" (4), [ruh] "r" (uh), [rlh] "r" (lh)
    : );
}

//!NOTE-dk theoretically need equivalents for "SWAP_FLOAT" & "SWAP_DOUBLE"
// in __GNUC__ syntax (First, grep for it in genc.cpp)
#endif  /* } */

#define SWAP_AND_SET_TIMETZ(pD,pS) {                                          \
SWAP_INT64( ((int64 *)(pD)), *((int64 *)(pS)) );                              \
SWAP_INT32( ((int32*)(((int64 *)(pD)) +1)), *((int32*)(((int64 *)(pS)) +1)) );\
}

#define SWAP_AND_SET_INTERVAL(pD,pS) {                                        \
SWAP_INT64( ((int64 *)(pD)), *((int64 *)(pS)) );                              \
SWAP_INT32( ((int32*)(((int64 *)(pD)) +1)), *((int32*)(((int64 *)(pS)) +1)) );\
}

#define SWAP_AND_SET_NUMERIC(pD, pS, n)     \
{                                           \
    int count = n;                          \
    while (count--) {                       \
        SWAP_INT32((int32 *)(pD), *(pS));   \
        (pS)++;                             \
        (pD)++;                             \
    }                                       \
}

#else  /* } {  */
//for LINUX the swap macros just assign the value without swapping

#define SWAP_INT16(pD, value)                                 \
*(pD) = value                                                 \

#define SWAP_UINT16(pD, value)                                  \
*(pD) = value                                                 \

#define SWAP_INT32(pD, value)                                 \
*(pD) = value                                                 \

#define SWAP_FLOAT(pD, value)                                 \
*(pD) = value                                                 \

#define SWAP_INT64(pD, value)                                 \
*(pD) = value                                                 \

#define SWAP_DOUBLE(pD, value)                                \
*(pD) = value                                                 \

#define SWAP_AND_SET_TIMETZ(pD,pS) {                                          \
*((int64 *)(pD)) =  *((int64 *)(pS)) ;                                        \
*((int32*)(((int64 *)(pD)) +1)) =  *((int32*)(((int64 *)(pS)) +1)) ;          \
}

#define SWAP_AND_SET_INTERVAL(pD,pS) {                                        \
*((int64 *)(pD)) =  *((int64 *)(pS)) ;                                        \
*((int32*)(((int64 *)(pD)) +1)) =  *((int32*)(((int64 *)(pS)) +1)) ;          \
}

#define SWAP_AND_SET_NUMERIC(pD, pS, n)   \
{                                         \
    int count = n;                        \
    while (count--) {                     \
        SWAP_INT32((int32 *)(pD), *(pS)); \
        (pS)++;                           \
        (pD)++;                           \
    }                                     \
}
#endif /* } */

//return -1, 0, 1 based on x is "less than", "equal to" or "greater than"
#define SCALAR_DIFF(x,y)       ( ((x) < (y)) ? -1 : (((x) > (y))  ? 1 : 0) )
#define FAST_SCALAR_DIFF(x,y)  ((int)(x) - (int)(y))

// Special handling for floats and doubles.  Need to cast to handle NaN values.
union DOUBLE_INT64 { double d; int64 i64; DOUBLE_INT64(double dd) : d(dd){} };
union  FLOAT_INT32 { float  f; int32 i32;  FLOAT_INT32(float  ff) : f(ff){} };
#define IEEE_AS_INT64(x) (DOUBLE_INT64(x).i64 == (int64)0x8000000000000000ULL ? (int64)0 : DOUBLE_INT64(x).i64)
#define IEEE_AS_INT32(x) (FLOAT_INT32(x).i32 == (int32)0x80000000UL ? (int32)0 : FLOAT_INT32(x).i32)

//get body portion of a varchar string
#define VBODY(x) ((void *)(((uint8*)(x))+sizeof(vlen_t)))
//get length of a varchar string
#define VLEN(x)  (*((vlen_t *)(x)))
#define VBODYLEN(x)  (*((vlen_t *)(x))-sizeof(vlen_t))

//bit test and set routines for recordstore based bitvectors
#define BITS_IN_BLOCK   ((DEFAULT_RECSTORE_SIZE - sizeof(TMemblockInfo))*8)
#define RECSTORE_SET_BIT(rs,xb,bk,setbk)    { \
setbk( rs->UpdateWindow((xb)/BITS_IN_BLOCK, (xb)/BITS_IN_BLOCK,bk) );\
NzSetBit(MBI_ZERO_BYTE(bk), (xb)%BITS_IN_BLOCK); }

#define RECSTORE_BIT_IS_SET(rs,xb,bk,setbk) \
( ((setbk( rs->UpdateWindow((xb)/BITS_IN_BLOCK, (xb)/BITS_IN_BLOCK,bk) ) ) )\
 ? NzBitIsSet(MBI_ZERO_BYTE(bk), (xb)%BITS_IN_BLOCK)\
 : throwError(17, "Missing BitVector block for stream outer join",0))

// Macro to reset pg expr pool memory context conditional on total allocation
#define RESET_EXPR_POOL(plan) \
if ( (plan->m_exprPool)->poolSizeUsed() > 16000 )   \
   ResetExprPool(plan);

#include "nz_orafuncs.h"

#endif  /* GENINL_INCLUDED */
