 /* Hash-on-a-CRecordStore
 * NOTE: the obvious file name was claimed first by
 *     /./main/src/pg/include/access/hash.h
 * Primary clientele:
       hash-join inner table
       aggregating group by
 * (C) Copyright IBM Corp. 2000, 2011  All Rights Reserved.
*/
#ifndef HASHRECS_INCLUDED
#define HASHRECS_INCLUDED

#ifndef NZTYPES_INCLUDED
# include "nztypes.h"
#endif

#ifndef HASHENUM_INCLUDED
# include "hashenum.h"
#endif

#ifndef NODECOMMON_INCLUDED
# include "nodecommon.h"
#endif

class CPlan;
class CRecordStore;
class CTable;
class THashJoinNode;
class TAggregateNode;
class CSorter;

#define SELECT_HASH_BUCKET(hp,val) \
     (((hp)->m_hashBuckets)[((unsigned)(val)) % (hp)->m_numBuckets])

struct THashInfo { // basic 'record' in an m_hashStore
    THashInfo *next;    // next collision record or NULL
    record_t *recP;     // ptr to the actual record
};

#define JOINED_IN_BIT 1   /* set lsb in hash table when record 'used' */
#define LAST_RECP_BIT 2   /* set in bucket pointer to singleton or in recP vector to mark end */

// These are global parameters for performance tuning and such.
// They are loaded from the registry.
struct HashTableParams_t {
    int hashFlags;                // Various.  See enumeration "hashflags".
    int hashTableDefaultPages;    // Aggregation ("stretchy") tables start out at this size.
                                  // They are doubled, on the fly, as needed.
    int hashTableRebuildMultiplier;// Run-time-expandable hash tables grow by this much (in percent)
                                   // when they are rebuilt.  100 makes it double, 200 makes it triple.
    int hashTableRehashMask;      // Mask of low bits of hash to be used for secondary
                                  // hash when a slot is already full.  (The registry
                                  // actually specifies the number of low bits in this.)
    int hashTableWarningPercent;  // When a table gets this percent full, it is rebuilt.
    int hashAggrForceDiskBased;   // Force premature disk-based when we see this (nonzero)
                                  // number of items.  5000 is a good number to use.
                                  // Making it go disk-based every 10 items is silly,
                                  // and it only queries this every 6000 or so.
    int hashAggrForceRebuildLimit;// Don't allow a rebuild to increase greater than this number of pages.
                                  // Used to force rebuilds to fail and go disk-based instead.
};



class CHash {
 public: //!NOTE-dk until spupage.cpp buried, at least

    enum EHashMode {
        eHSbasic,        // Used for exist joins.  Hash table made in advance,
                         // from the temp table.  Hash table has buckets listing all
                         // keys that collided there.  Since this is an exists join,
                         // nothing further is needed.

        eHSbasic2D,      // Used for real cross joins.  Hash table made in advance,
                         // from the temp table.  Hash table has buckets listing all
                         // keys that collided there, and associates with each key
                         // a list of records in the temp table that have that key.
                         // (Multiple records in the temp table might have the same key---
                         // this is what leads to "ballooning".)
                         // This 2-level structure of structures is turned into a more
                         // compact "vector" form after the hash table is created, to
                         // speed up operation at run time.

        eHSgroupAggr,    // Used for grouped aggregates.  Hash table made
                         // on the fly as the generated code runs.

        eHSgroupDistinct // Used for grouping with distinct aggregates.
                         // This is operationally identical to the one above
                         // (they differ only in the way the generated code
                         // uses them), and differs only in the documentation
                         // string appearing in the _vt_hash_join_temp_table.
    };


    void * operator new (size_t);
    void operator delete (void*);

    // Constructor
    CHash(
        CPlan *pPlan,       // executing context
        SnippetNode* pControllingNode,
        CRecordStore *pRS,  // RecordStore to create hash over (may be empty)
        THashJoinNode *pHJ, // Node controlling hashed join (may be 0 / missing)
        EHashMode hashStyle,
        uint32 estimatedNumRecords,  // For hash join; we know how big the hash table will be.
        uint64 numOfRec=0);          // For aggr / group by; we can sometimes estimate eventual size.
    ~CHash();            // vs DeleteHashStore()?

    void ResetHashStore();   // reset hash tables and hash store
    void DeleteHashBucketsArray();
    void DeleteHashCompactedVectors();
    void FinalVtDump();
    uint32 numHashBucketArrayPages();

    void SetDisplacementCount(uint64 count) { m_displacementCount = count; }

    // We only need the number, for making the virtual table.  Datatype is ultimately "EAggrStyles".
    // The aggregation styles (in xpaggregate.h) start at zero, the way all good
    // enumerations do.  But we use zero for the default "not an aggregation" state,
    // and so we start the real aggregation states at 1.  The text table is finagled
    // to take care of this.
    void SetAggrStyleIndex(int styleIndex) { m_aggrStyleIndex = styleIndex+1; }

    // Ditto.  Datatype is ultimately "EAggrTypes".
    // The aggregation subtypes start at one!!!!.
    // We use zero for the default "not an aggregation" state, and start the
    // real aggregation states at 1.
    void SetAggrOpIndex(int OpIndex) { m_aggrOpIndex = OpIndex; }


    static void HashPages(EHashMode type,
                          uint64 numRecords,
                          EPlanNodeLocus locus,
                          int &bucketArrayPages,
                          int &hashInfoPages,
                          int *tempMem=0);

    static bool normalizeRecord(int fieldType,
                                 void* srcFp,
                                 int64* int64P,
                                 int* lenP);

    // These flags show up in the registry variable "system.hashflags".
    // See nzcfgldr.cpp, where the default is set to HF_MOVE_TO_FRONT.
    // For historical reasons, don't put anything new in the low bits.  There are plenty of high bits.
    enum hashflags {
        HF_MOVE_TO_FRONT               = 0x00000100,     // Put 256 in the registry.  This one is on by default; the others are off.
        HF_COMPACT_LEAF_RECORDS        = 0x00000200,     // Put 512 in the registry.
        HF_REALLOCATE_FOR_NUM_DISTINCT = 0x00000400,     // ************************ Shouldn't we take this out?
        HF_ALL_HASH_JOINS_DISK_BASED   = 0x00001000,     // Put 4096 in the registry.
        HF_DONT_TAKE_AGGR_SIZE_HINT    = 0x00002000      // Put 8192 in the registry.
    };

    static void NotifyRegistryUpdated();

    void chooseHashModulus(uint64 numRecords, bool forAggregation);

    void HashTableFill(int numColsHash
                       , int* pColHash
                       , bool khash);


    // All CTable::HashTableFill_XXX functions MUST have same signature
    typedef void HashTableFill_XXX_sig(
        int numCols,    // how many columns in colNums array
        const int *colNums,   // which columns to hash on
        bool khash,
        ESimpleHashType simpleHasher);

    void HashTableFill_Basic(
        int numCols,    // how many columns in colNums array
        const int *colNums,   // which columns to hash on
        bool khash,
        ESimpleHashType simpleHasher);

    void HashTableFill_Join(
        int numCols,    // how many columns in colNums array
        const int *colNums,   // which columns to hash on
        bool khash,
        ESimpleHashType simpleHasher);

    void statsDisplay(char const* context = "");
    void WriteVirtualTable();

    // Called from generated code:
    int64 CalculateWarningLevel();
    bool HashTableBeginRebuild(TAggregateNode *nodeForCounting, CSorter *memtester);
    void HashTableEndRebuild();

    EHashMode m_hashStyle;
    THashJoinNode* m_node;

    // For the hashing logic:  (first few data items known to generated code)
    unsigned int m_numBuckets;  // This is the number of slots in the active table.  It is prime.  It is
                                // slightly smaller than the multiple of 16K that is in m_nominalHashSlots.
    uint64 m_nominalHashSlots;  // The "nominal" size is an integral multiple of 16384 slots, which is 65536 bytes.
                                // Memory will be allocated to this size.  The actual size of the table will be
                                // a prime number of slots slightly smaller than this, because finding slots
                                // from a hash number modulo a prime is a Good Thing.  (In fact, the way we do
                                // secondary hashing absolutely requires it.)
                                // The claim is that it is safe to have this as a 32-bit number, since the
                                // hash table is in memory.  That may not be true when we go to 64 bits.
    THashInfo **m_hashBuckets;  // The hash table itself.  In many cases these items are bucket list heads.
    THashInfo **m_hashBucketEnd;// end of said array

    CRecordStore *m_hashStore;  // bucket content in a linked representation
    int64 m_FullnessWarning;      // Counts down as table is filled; when reach zero, need to reallocate.
    uint32      m_hashDeltaMask;// To get the amount by which we move forward if hash slot is busy.
    // end items known to generated code

    int m_aggrStyleIndex;    // This is NOT the hash style.  It is the style passed down from the
                             // aggregation node, for putting info in the diagnostic virtual table.
    int m_aggrOpIndex;       // If under an aggregation, this says what type it was.
    int m_rebuildCount;      // Used for displaying stats in virtual table.  Aggregation nodes also have a count.
    uint64 m_displacementCount;  // Ditto.

    unsigned int m_numHashTablePages;
    unsigned int m_numDistinctKeys;
    unsigned int m_numFilledBuckets;
    unsigned int m_numSingletonBuckets;
    unsigned int m_numKeyCollidedBuckets;
    unsigned int m_numDupRows;
    unsigned int m_maxKeyCollisions;
    unsigned int m_vectorSpaceBytes;
    unsigned int m_maxVectorLength;
    uint64 m_populationDuringRebuild;
    int64 m_cpuTime;
    int32 m_hwId;
    int32 m_tableId;
    int32 m_planId;
    int32 m_dsId;
    int32 m_snippetId;
    int16 m_firstCol;
    int8 m_numCols;
    int8 m_device;

    CRecordStore * const m_dataStore;  // documenting what got a hash
    SnippetNode* const m_ControllingNode;  // So the virtual table can figure out the number.

    void* m_compactedHashVectors;
    nzotuid_t m_hashOwner;      // process controlling hash (process local memory)

    // Hash rebuild parameters
    CPlan * const m_pPlan;      // executing context


 private:
    CHash(const CHash&);
    CHash& operator=(const CHash&);
};

#endif // HASHRECS_INCLUDED
