/// @file hcHwThread.h
/// @brief  Base class for SPU threads.
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2001-2011 All rights reserved.
//////////////////////////////////////////////////////////////////////
#ifndef HCHWTHREAD_INCLUDED
#define HCHWTHREAD_INCLUDED

#include "nzstd.h"
#include "rcPgmAttrs.h"
#include "nzstate.h"
#include "locktracker.h"

namespace ac {
  class PgmInfo;
}

namespace hc {  // hc == spuComps

//==============================================================================
// Class: Runnable - pure virtual interface for a runnable object

struct Runnable {

    virtual ~Runnable();

    virtual void Run() = 0;
};

inline Runnable::~Runnable() { }

extern pid_t createThread(Runnable *runnable, const char *name, int dsidx,
                          unsigned int priority, unsigned int timeSlice);

// adapter function
inline static void
run_runnable(void *arg)
{
    Runnable *const r = static_cast<Runnable *>(arg);
    r->Run();
}

//==============================================================================
// Class: HwThread - thread class supporting system management extentions

class HwThread {

public:

    static HwThread    *s_instance;         // of current process

    ac::PgmInfo        *m_pgmInfo;          // Record for this program
    int                 m_pgmId;
    rc::PgmAttrs        m_pgmAttrs;         // Label, src prefix, ...

    HwThread();
    ~HwThread();

    // MANIPULATORS
    static HwThread*    instance ();        // Of current running thread

    static void         attachThread(HwThread *hwThread, pid_t pid, const char *name);
    static void         unattachThread(HwThread *hwthread);
    static void         reinitThreadName(const char *name);

    static HwThread*    lookupThread(pid_t pid);

    static void         clearLockTracker();
    static void         makeTlsLockTracker();

    void                removeFromPgmTbl();

    void                enableStateControl(bool enable);
    static void         makeThreadStateAware(hc::HwThread *thr = hc::HwThread::instance());

    void                setPgmState(NzSysState sysState,
                                    int sysStateVersion = -1); // -1 = don't set

    // ACCESSORS
    const char*         name() const;
    NzSysState          pgmState () const;
    int                 pgmStateVersion () const;
    ac::PgmInfo*        pgmInfo() const;
    bool                hasStateControl() const;
    const rc::PgmAttrs* pgmAttrs() const;
    rc::PgmAttrs*       pgmAttrs();

    static
    rc::PgmAttrs*       currentThreadAttrs();

    // public data
    atomic_t           m_numLockTrackers;   // number of threads per process with locktrackers.
    LockTracker       *m_lockInfo[MAX_LT_INDEXES];
    LockTracker        m_lt0;  // first locktracker for the process;
                               // subsequent threads dynamically allocate
                               // additional lock trackers
};

inline HwThread *
HwThread::instance()
{
    return s_instance;
}

inline const char* HwThread::name() const
{
    return m_pgmAttrs.pgmLabel();
}

inline ac::PgmInfo* HwThread::pgmInfo() const
{
    return m_pgmInfo;
}

inline rc::PgmAttrs* HwThread::pgmAttrs()
{
    return &m_pgmAttrs;
}

inline const rc::PgmAttrs* HwThread::pgmAttrs() const
{
    return &m_pgmAttrs;
}

} // end namespace hc

#endif /* HCHWTHREAD_INCLUDED */
