//------------------------------------------------------------------------------
// hcMsgsBase.h
//
// Definitions of Hwm messages sent to track spus
//
// (C) Copyright IBM Corp. 2007, 2011  All Rights Reserved.
//
// All pieces of messages (like the one below) MUST be uint64
// alligned. These structures have been surgically refactored because
// we need the hc::CoreRequestCall, hc::CoreRequestReply, 
// hc::CoreCompleteCall and hc::CoreCompleteReply in the Nucleus SPU,
// virtual SPU, cluster SPU, Host and above all the Linux SPU. We have 
// major linkage issues with the hccomps library especially if just a few
// messages have to be available across executables / processes and also 
// you want to make the process really lean and mean in size. Before
// this reorganization we would have needed the entire SPU image to be 
// built for the Linux SPU core dumper which actually needs to be a pretty
// lean image. Purists can ask why is it that we are not refactoring
// the entire hwcomps area like this. Simple, availability of time or 
// non availability thereof.


#ifndef HCMSGSBASE_INCLUDED
#define HCMSGSBASE_INCLUDED

#include <limits.h>

#ifndef HCENUMS_INCLUDED
# include "hcEnums.h"
#endif

#ifndef NZSTD_INCLUDED
# include "nzstd.h"
#endif

#ifndef HCMSGHDR_INCLUDED
# include "hcMsgHdr.h"
#endif

namespace hc {

// Forward declarations.
class MsgHdr;
class Msg;
class CoreReqCall;
class CoreReqReply;
class CoreCompleteCall;
class CoreCompleteReply;

struct MsgHdrBaseLocal;
struct MsgHdrBaseNetwork;
struct MsgBaseNetwork;
struct CoreReqCallBaseNetwork;
struct CoreReqReplyBaseNetwork;
struct CoreCompleteCallBaseNetwork;

void swapMsgHdr (struct MsgHdrBaseNetwork *);
void swapMsg (struct MsgBaseNetwork *);
void swapCoreReqCall (struct CoreReqCallBaseNetwork *);
void swapCoreReqReply (struct CoreReqReplyBaseNetwork *);
void swapCoreCompleteCall (struct CoreCompleteCallBaseNetwork *);

struct MsgHdrBaseLocal {
    size_t          (*fn_msgSize)();
    void            (*fn_msgSwap)(MsgHdr* msg);
    void            (*fn_msgDump)(MsgHdr* msg);
    const char*     (*fn_msgName)();
    bool            (*fn_equalTo)(MsgHdr* left, MsgHdr* right);
    char*           m_payloadData;
    int             m_references;
    int             m_isPayloadAllocated;
    int             m_pad;
};

// WARNING: All pieces of messages (like the one below) MUST be uint64
//          alligned.
struct MsgHdrBaseNetwork {
public:
    struct hc_msg_hdr   m_hdr;

public:
    MsgHdrBaseNetwork(int type = hc::PAYLOAD);
    ~MsgHdrBaseNetwork();

    void setType (int type);
};

// WARNING: All pieces of messages (like the one below) MUST be uint64
//          alligned.
struct MsgBaseNetwork {
public:
    struct hc_msg_payload_hdr   m_phdr;

public:
    MsgBaseNetwork();
    ~MsgBaseNetwork();
};

#define CORE_REQ_ERROR_BUF_SIZE 2048
// WARNING: All pieces of messages (like the one below) MUST be uint64
//          alligned.
struct CoreReqCallBaseNetwork {
public:
    int                 m_hwId;
    int                 m_spaId;
    int                 m_spaSlot;
    int32               m_errorType;
    char                m_errorBuf[CORE_REQ_ERROR_BUF_SIZE];


public:
    CoreReqCallBaseNetwork();
    ~CoreReqCallBaseNetwork();
};

// WARNING: All pieces of messages (like the one below) MUST be uint64
//          alligned.
struct CoreReqReplyBaseNetwork {
public:
    int32               m_isOkToSend;
    char                m_pathname[512];

public:
    CoreReqReplyBaseNetwork();
    ~CoreReqReplyBaseNetwork();
};

// WARNING: All pieces of messages (like the one below) MUST be uint64
//          alligned.
struct CoreCompleteCallBaseNetwork {
public:
    int32               m_resultCode;

public:
    CoreCompleteCallBaseNetwork(int code = hc::ET_UNKNOWN);
    ~CoreCompleteCallBaseNetwork();

    void setResultCode (int code);
};

class MsgHdrNetwork {
public:
    struct MsgHdrBaseNetwork msgHdrBase;

public:
    MsgHdrNetwork (int type);
    ~MsgHdrNetwork();

    static size_t   sizeofThis(MsgHdrNetwork* msg);
    static void     swapThis(MsgHdrNetwork* msg);

    void swap();
    size_t size() const;
    void *data();
};

class MsgNetwork : public MsgHdrNetwork {
public:
    struct MsgBaseNetwork msgBase;

public:
    MsgNetwork(int type);
    ~MsgNetwork();

    static size_t   sizeofThis(MsgHdrNetwork* msg);
    static void     swapThis(MsgHdrNetwork* msg);

    void swap();
    size_t size() const;
    void *data();

private:
    typedef hc::MsgHdrNetwork    inherited;
};

class CoreReqCallNetwork : public MsgNetwork {
public:
    CoreReqCallBaseNetwork coreReqCallBase;

public:
    CoreReqCallNetwork();
    ~CoreReqCallNetwork();

    static size_t   sizeofThis(MsgHdrNetwork* msg);
    static void     swapThis(MsgHdrNetwork* msg);

    void swap();
    size_t size() const;
    void *data();

private:
    typedef hc::MsgNetwork    inherited;
};

class CoreReqReplyNetwork : public MsgNetwork {
public:
    struct CoreReqReplyBaseNetwork coreReqReplyBase;

public:
    CoreReqReplyNetwork();
    ~CoreReqReplyNetwork();

    static size_t   sizeofThis(MsgHdrNetwork* msg);
    static void     swapThis(MsgHdrNetwork* msg);

    void swap();
    size_t size() const;
    void *data();

private:
    typedef hc::MsgNetwork    inherited;
};

class CoreCompleteCallNetwork : public MsgNetwork {
public:
    enum { FILENAME_SZ = 256, TYPE_SZ = 256 };
    enum Flags { TEST = 0x1, CLEANUP = 0x2 };

    struct CoreCompleteCallBaseNetwork coreCompleteCallBase;
    char   m_coreFile [FILENAME_SZ];
    char   m_coreType [TYPE_SZ];
    uint32 m_flags;

public:
    CoreCompleteCallNetwork();
    ~CoreCompleteCallNetwork();

    static size_t   sizeofThis(MsgHdrNetwork* msg);
    static void     swapThis(MsgHdrNetwork* msg);

    void swap();
    size_t size() const;
    void *data();

private:
    typedef hc::MsgNetwork    inherited;
};

class CoreCompleteReplyNetwork : public MsgNetwork {
public:
    CoreCompleteReplyNetwork();
    ~CoreCompleteReplyNetwork();

    static size_t   sizeofThis(MsgHdrNetwork* msg);
    static void     swapThis(MsgHdrNetwork* msg);

    void swap();
    size_t size() const;
    void *data();

private:
    typedef hc::MsgNetwork    inherited;
};

};

#endif /* HCMSGSBASE_INCLUDED */

