#ifndef I386_ATOMIC_INCLUDED
#define I386_ATOMIC_INCLUDED

typedef struct { int64 __attribute__((aligned(8))) counter64; } atomic64_t;

inline static void
atomic_add(int i, atomic_t *v)
{
    int xreg = i;

    // locked exchange and add
    asm volatile (
       "lock xaddl %0,%1"
       : /* %0 */ "=r" (xreg), /* %1 */ "=m" (v->counter)
       : "0" (xreg), "m" (v->counter)
       : "cc"
       );
}

inline static void
atomic64_add(int64 i, atomic64_t *v)
{
    __sync_add_and_fetch(&v->counter64, i);
}

inline static int
atomic_add_return(int i, atomic_t *v)
{
    int xreg = i;

    // locked exchange and add
    asm volatile (
       "lock xaddl %0,%1"
       : /* %0 */ "=r" (xreg), /* %1 */ "=m" (v->counter)
       : "0" (xreg), "m" (v->counter)
       : "cc"
       );

    // x86 xadd returns value prior to addition
    return xreg + i;
}

inline static int64
atomic64_add_return(int64 i, atomic64_t *v)
{
    return __sync_add_and_fetch(&v->counter64, i);
}

inline static void
atomic_sub(int i, atomic_t *v)
{
    int xreg = -i;

    // locked exchange and add
    asm volatile (
       "lock xaddl %0,%1"
       : /* %0 */ "=r" (xreg), /* %1 */ "=m" (v->counter)
       : "0" (xreg), "m" (v->counter)
       : "cc"
       );
}

inline static void
atomic64_sub(int64 i, atomic64_t *v)
{
    __sync_sub_and_fetch(&v->counter64, i);
}

inline static int
atomic_sub_return(int i, atomic_t *v)
{
    int xreg = -i;

    // locked exchange and add
    asm volatile (
       "lock xaddl %0,%1"
       : /* %0 */ "=r" (xreg), /* %1 */ "=m" (v->counter)
       : "0" (xreg), "m" (v->counter)
       : "cc"
       );

    // x86 xadd returns value prior to addition
    return xreg - i;
}

inline static int64
atomic64_sub_return(int64 i, atomic64_t *v)
{
    return __sync_sub_and_fetch(&v->counter64, i);
}

inline static int64
atomic64_xchg(atomic64_t *v, int64 val)
{
    int64 old;
    int64 t = v->counter64;

    do {
        old = t;
        t = __sync_val_compare_and_swap(&v->counter64, old, val);
    } while (t != old);

    return old;
}

#define atomic_inc(v)           atomic_add(1, (v))
#define atomic64_inc(v)         atomic64_add(1, (v))
#define atomic_inc_return(v)    atomic_add_return(1, (v))
#define atomic64_inc_return(v)  atomic64_add_return(1, (v))
#define atomic_dec(v)           atomic_sub(1, (v))
#define atomic64_dec(v)         atomic64_sub(1, (v))
#define atomic_dec_return(v)    atomic_sub_return(1, (v))
#define atomic64_dec_return(v)  atomic64_sub_return(1, (v))

inline static int64
atomic64_read(atomic64_t *v)
{
    int64 ret;
#if defined(__x86_64__)
    ret=__sync_add_and_fetch(&v->counter64,0);
#else
    asm volatile(
        "mov %%ebx, %%eax\n"
        "mov %%ecx, %%edx\n"
        "lock cmpxchg8b %1\n"
        : "=&A" (ret)
        : "m" (v->counter64)
        : "cc"
        );
#endif
    return ret;
}

inline static void
atomic64_set(atomic64_t *v, int64 val)
{
    atomic64_xchg(v, val);
}

inline static int
atomic_inc_if_nonzero(atomic_t *v)
{
    int val = atomic_read(v);

    while (val)
    {
        unsigned char c;

        asm volatile (
            "lock cmpxchgl %3,%1 ; setz %2"
            : "=a" (val), "=m" (v->counter), "=qm" (c)
            : "r" (val + 1), "0" (val), "m" (v->counter)
            : "memory");

        if (c != 0)
        {
            return 1;
        }
    }

    return 0;
}

#define atomic_set_mask(m, v) \
    asm volatile("lock orl %0,%1" \
                 : : "r" (m), "m" ((v)->counter) : "memory")

#define atomic_clear_mask(m, v) \
    asm volatile("lock andl %0,%1" \
                 : : "r" (~(m)), "m" ((v)->counter) : "memory")

inline static unsigned int
atomic_cmpxchg(atomic_t *v, unsigned int old, unsigned int noo)
{
    unsigned int prev;

    asm volatile (
        "lock cmpxchgl %1,%2"
        : "=a" (prev)
        : "r" (noo), "m" (v->counter), "0" (old)
        : "memory");

    return prev;
}

inline static unsigned int
atomic_xchg(atomic_t *v, unsigned int val)
{
    unsigned int t = val;

    asm volatile (
        "lock xchgl %0,%1"
        : "=r" (t), "=m" (v->counter)
        : "0" (t), "m" (v->counter)
        : "memory");

    return t;
}

#endif // I386_ATOMIC_INCLUDED
