#ifndef IDIOMS_H
#define IDIOMS_H

//======================================================================
///\file idioms.h
//
// (C) Copyright IBM Corp. 2008, 2011  All Rights Reserved.
//======================================================================


/// Provide public "get" and "set" functions without exposing storage
/// address.

#define EXPOSE_ACCESSORS(TYPE, MEMBER) \
/*!\{\name m_##MEMBER:*/ \
    public: TYPE MEMBER() const { return m_##MEMBER; } \
    public: void SET_##MEMBER(TYPE x) { m_##MEMBER = x; }
/*!\}*/ \



/// Make it impossible to instantiate a class by declaring both the
/// default constructor and the copy constructor to be private.

#define NEVER_INSTANTIATE(TYPE) \
/*!\{\name Private unimplemented (so that TYPE can never be instantiated)*/ \
private: \
    TYPE(); \
    ~TYPE(); \
/*!\}*/ \



/// Given an STL-style CONTAINER run a classic \e for loop from begin() to
/// end() using a newly introduced forward ITERATOR.

#define FOR_EACH(ITERATOR, CONTAINER) \
for ( typeof ((CONTAINER).begin()) ITERATOR = (CONTAINER).begin() \
    ,                           _##ITERATOR = (CONTAINER).end() \
    ; ITERATOR != _##ITERATOR \
    ; ++ITERATOR \
    )


/// Given an C-style vector (BASE and COUNT) run a classic \e for loop
/// over the vector's elements using a newly introduced forward ITERATOR.

#define FOR_VECTOR(ITERATOR, BASE, COUNT) \
for ( typeof (BASE)[0] *ITERATOR = (BASE) \
    ,              * _##ITERATOR = (BASE) + (COUNT) \
    ; ITERATOR != _##ITERATOR \
    ; ++ITERATOR \
    )


/// Given a C-style fixed-size ARRAY run a classic \e for loop over the
/// array's elements using a newly introduced forward ITERATOR.

#define FOR_ARRAY(ITERATOR, ARRAY) \
for ( typeof (ARRAY)[0] *ITERATOR = (ARRAY) \
    ,               * _##ITERATOR = (ARRAY) + (sizeof(ARRAY) / sizeof((ARRAY)[0])) \
    ; ITERATOR != _##ITERATOR \
    ; ++ITERATOR \
    )


/// Combined alloca / placement new idiom.  Determine the amount of
/// memory required, acquire it via alloca and then invoke an associated
/// constructor.  This idiom assumes that constructors and bytesRequired
/// functions come in matched pairs with identical signatures.  Thus
/// given a class C, with a constructor called as C(X,Y) one can write
///
///      C* cp = new ALLOCA(C,(X,Y));
///
/// Which is equivalent to
///
///      C* cp = new( alloca(C::bytesRequired(X,Y) ) C(X,Y);

#define ALLOCA(TYPE,ARGS) (alloca(TYPE::bytesRequired ARGS)TYPE ARGS


#define countBits(MASK) __builtin_popcount(MASK)

#endif // IDIOMS_H
