#ifndef JOBINTR_INCLUDED
#define JOBINTR_INCLUDED
/// @file jobintr.h
/// @brief Job Interrupt state
///
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2008, 2011  All Rights Reserved.
//////////////////////////////////////////////////////////////////////
///
/// Description:
///
/// The following is taken from the Linux SPU write-up by Andy Galasso and
/// provides the motivation for the interrupt state implementation in this
/// module:
///
///   Event Process Termination Changes
///
///   In order to support graceful recovery from crashes or hangs in
///   user-supplied UDX code, we no longer rely on gened code to
///   periodically poll a plan's "interrupted" flag. Instead we use a
///   mechanism based on Linux signal handling to recover from certain
///   crashes or hangs.
///
///   We distinguish two possible states for an event process: recoverable
///   or non-recoverable. In the recoverable state, a crash of the event
///   process can be tolerated by the system without any ill effects beyond
///   aborting the specific query that crashed. In the non-recoverable
///   state, abnormal event process termination is treated as a fatal error
///   requiring NPS restart. An example of a non-recoverable state would be
///   when an event process is holding a lock on some shared memory
///   data. Conversely, an event process would typically be in a recoverable
///   state while it is executing gened code.
///
///   To handle asynchronous query abort (e.g., Control-C or nzsession
///   abort), a query abort will cause a SIGINT signal to be sent to the
///   event process. If the event process is in the recoverable state, the
///   SIGINT will cause the event process to terminate immediately. If the
///   event process is not in the recoverable state when a SIGINT arrives,
///   for example, when it is holding a lock on some shared memory data, the
///   SIGINT processing will be deferred until the event process indicates
///   it can safely handle the SIGINT (e.g., after it has released the
///   lock).
///
///   Immediate SIGINT processing is enabled prior to calling gened code,
///   and disabled upon return from the gened code. Functions that are
///   called from gened code and that access some shared state (i.e., access
///   shared memory) are required to temporarily suspend immediate SIGINT
///   processing. In order to detect cases where new code fails to do this,
///   we have added asserts in the shared memory locking code paths to
///   verify that immediate SIGINT processing has been disabled when a lock
///   is being acquired. The API's for enabling/disabling immediate SIGINT
///   processing are declared and commented in a new interface file,
///   jobintr.h.
///
///   On the host, the dispatch process is notified by Linux when an event
///   process terminates. When a recoverable event process is found to have
///   been terminated by a signal (e.g., SIGINT), dispatch responds by
///   re-spawning a new event process for the same client session. The
///   re-spawned event process initiates transaction abort processing and
///   handles SPU responses for the abort interactions. Event processes that
///   terminate while immediate SIGINT processing is disabled
///   (non-recoverable state), for example, failing an NZ_VERIFY outside
///   gened code, are treated as a fatal errors and cause NPS to restart.
///
///   On the Linux SPU, an event process that is terminated by SIGINT
///   requires cleanup processing, for example, to delete the CPlan instance
///   from shared memory. The cleanup processing is performed in the context
///   of the job task cleanup task, "jbtaskgc". A terminated event process is
///   detected by the process manager process (nzprocmgr) which sends a
///   notification to jbtaskgc to do the cleanup.
///
///   One significant implication of the event process termination changes
///   relates to the transaction detail journal (TDJ). The TDJ records block
///   numbers for table blocks modified on the SPU to allow fast rollback of
///   deletes and updates. Prior to the dev-lspu changes, the TDJ code on
///   the SPU used the comm-level ack as a guarantee that the TDJ message
///   would be processed by the event process on the host. Since event
///   processes can now be terminated by SIGINT before an acked message is
///   handed off from comm to the event process, this assumption is no
///   longer valid. The TDJ code has been modified to use an
///   application-level message sent from the host event process to the SPU
///   as the acknowledgment of receipt of the TDJ message.
///
//////////////////////////////////////////////////////////////////////

#ifndef NZSTD_INCLUDED
# include "nzstd.h"
#endif

#ifndef RCBARRIER_INCLUDED
# include "rcBarrier.h"
#endif

// private data - jobtask internal use only
struct _jobintr_state_t
{
    volatile bool sigint_enabled;
    volatile unsigned int critsec_count;
    volatile int deferred_sig;
};
#ifdef WIN32
extern "C" {
	struct _jobintr_state_t _s_jobintr_state;
};
#else
extern struct _jobintr_state_t _s_jobintr_state;
#endif


#if defined(__DCC__) || defined(_WIN32)
extern void _jobintr_process_interrupt();
#else
extern void _jobintr_process_interrupt() __attribute__((noreturn));
#endif

//
// Job task SIGINT handling public interface
//

/**
 * Enable/disable immediate SIGINT processing. When enabled, SIGINT will
 * terminate the job task. When disabled, SIGINT processing will be deferred
 * until enabled, and receipt of signals that terminate the process, e.g.,
 * SIGABRT, will be fatal.
 */
extern bool jobintr_enable_immed_sigint(bool enable);

/**
 * Enter a critical section. SIGINT processing will be deferred until the
 * critical section is exited. Receipt of signals that terminate the
 * process, e.g., SIGABRT, will be fatal. Critical sections can nest.
 */
inline static void
jobintr_enter_critsec()
{
    ++_s_jobintr_state.critsec_count;
}

/**
 * Leave a critical section. SIGINT will be processed immediately, and
 * receipt of signals that terminate the process, e.g., SIGABRT, will be not
 * be fatal.
 */
inline static void
jobintr_leave_critsec()
{
    if (--_s_jobintr_state.critsec_count == 0)
    {
        if (NZ_UNLIKELY(_s_jobintr_state.deferred_sig != 0))
        {
            if (_s_jobintr_state.sigint_enabled)
            {
                _jobintr_process_interrupt(); // does not return
            }
        }
    }
}

/**
 * Return whether immediate SIGINT processing is in effect. SIGINT is
 * handled immediately when both:
 *    o  enableImmediateSigint(true) has been called, and
 *    o  the process is not in a critical section
 *       (enterCriticalSection()/leaveCriticalSection())
 */
inline static bool
jobintr_immed_sigint_enabled()
{
    return _s_jobintr_state.sigint_enabled && _s_jobintr_state.critsec_count == 0;
}

/**
 * When immediate SIGINT processing is not enabled, a process may want
 * to check whether a SIGINT has been received, and, if so, process the
 * interrupt.
 */
inline static void
jobintr_check_for_sigint()
{
    if (NZ_UNLIKELY(_s_jobintr_state.deferred_sig != 0))
    {
        if (_s_jobintr_state.critsec_count == 0)
        {
            _jobintr_process_interrupt(); // does not return
        }
    }
}

/**
 * test whether deferred signal is pending
 */
inline static int
jobintr_sigint_pending()
{
    return _s_jobintr_state.deferred_sig;
}

#define JOBINTR_ENTER_CRITSEC() do { \
    jobintr_enter_critsec(); \
    BARRIER(); \
} while (0)

#define JOBINTR_LEAVE_CRITSEC() do { \
    BARRIER(); \
    jobintr_leave_critsec(); \
} while (0)

#define JOBINTR_VERIFY_SIGINT_DISABLED() NZ_VERIFY(!jobintr_immed_sigint_enabled())


struct AutoCriticalSection
{
    AutoCriticalSection() { JOBINTR_ENTER_CRITSEC(); }
    ~AutoCriticalSection() { JOBINTR_LEAVE_CRITSEC(); }
};

#define JOBINTR_DECLARE_CRITSEC_SCOPE() AutoCriticalSection __critsec

// fixme-lspu revisit these sites --
// these critical sections may not be necessary as we move stuff out of
// global shared memory
#define JOBINTR_DECLARE_CRITSEC_SCOPE_revisit() JOBINTR_DECLARE_CRITSEC_SCOPE()

// NOTE: AutoDisableImmedSigint is for special cases, in most situations you should
// be using JOBINTR_DECLARE_CRITSEC_SCOPE instead
class AutoDisableImmedSigint
{
    bool m_prior;
public:
    AutoDisableImmedSigint() { m_prior = jobintr_enable_immed_sigint(false); }
    ~AutoDisableImmedSigint() { jobintr_enable_immed_sigint(m_prior); }
};

class AutoEnableImmedSigint
{
    bool m_prior;
public:
    AutoEnableImmedSigint() { m_prior = jobintr_enable_immed_sigint(true); }
    ~AutoEnableImmedSigint() { jobintr_enable_immed_sigint(m_prior); }
};

// Conditional critical section
struct AutoCriticalSectionIf
{
    bool m_critical;

    AutoCriticalSectionIf(bool crit) : m_critical(crit)
    { if (m_critical) { JOBINTR_ENTER_CRITSEC(); } };

    ~AutoCriticalSectionIf()
    { if (m_critical) { JOBINTR_LEAVE_CRITSEC(); } };
};

#endif  // JOBINTR_INCLUDED
