#ifndef KHASH_INCLUDED
#define KHASH_INCLUDED

#include "error.h" // for NZ_ASSERT

// Number representations
// ----------------------
// Equal values under differing representations must result in the same
// hash.  (This is necessary if such values are to be joined using a
// local-only join.)  For numbers (various width of integers and numerics)
// we achieve this by conceptually widening all values to a canonical 16
// bytes and returning the hash of that bit pattern.  In reality we only
// widen to the compilers builtin 64 bit integer type.  For text strings
// that differ only in the number of trailing spaces or nulls ('\0') must
// produce the same hash.  We achieve this by the brute force approach of
// scanning backward from the end of the string.

// Multiplicative hashing
// ----------------------
// The actual hash used here is the Multiplicative Hash discussed in Knuth
// V3, page 513...  That hash multiplies by the largest prime that fits in
// the machine's natural word, returning the low half of the double width
// product as the result.  Using this scheme computing the hash of a given
// value requires only a single multiply by a constant.  The most random
// bits then occur at the top of the register.  Thus to isolate N bits at
// from a hash value H, where typically one might be inclined to mask, e.g.
//
//      bucketIndex = H & ((1<<N)-1);
//
// with the multiplicative hash one does best to shift, e.g.
//
//      bucketIndex = ((uint32)H) >> (32 - N);
//
// Finally, notice that both the number zero and the empty string have hash
// values of zero.

// Folding
// -------
// Since we want to perform exactly one single precision integer multiply
// to evaluate the hash we need a protocol for dealing with numbers in our
// 16 byte canonical representation.  The current "simple hash" strategy is
// to simply discard the high order 12 bytes.  Though this is arguably
// correct it only delivers decent skew behavior if the low 4 bytes are
// adequately random.
//
// The strategy adopted here is to fold the four 4 byte pieces together via
// xor-ing.  This has the obvious virtue that all bits participate in the
// hash.  It also admits to a simple optimization: for values widened from
// representations of 8 bytes or less we can ignore the high order 8 bytes
// during folding.  This is because those bytes must be entirely filled
// with copies of the canonical sign bits.  (This is true even when the
// value that was widened was 64 bits unsigned.)  Since all 8 high order
// bytes will be filled entirely with either 0s or with 1s they will
// contribute nothing to the final folded xor.


// Replicate the sign of an int32 through an entire 32 bit register
// (In C++ the result of x >> y when x is negative is implementation
// dependent.  This macro assumes an implementation providing arithmetic
// shift semantics.  This is the case for DCC, GCC and MS Visual C++.)
#ifndef REPLICATE_SIGN32
# define REPLICATE_SIGN32(val) (((int32)(val))>>31)
#endif

// Endian independent access to high 32 bits of a 64 bit integer
#define HIGH_32_BITS(val64) ((uint32)(((uint64)(val64))>>32))

#ifdef WIN32
// disable warning:  C4146, unary minus operator applied to unsigned type, result still unsigned
# pragma warning(disable:4146)
#endif

// This hash is used for all representations up to 64 bits.  By having this
// function take an int64 argument the appropriate zero or sign extension
// of narrow types will occur, setting of the folding operation required.
//
inline static uint32 KHashInt(int64 val64, uint32 chain = 0)
{
    return 0x9e3779b9 * (chain ^ ((uint32)val64) ^ HIGH_32_BITS(val64));
}

inline static uint32 KHashInt(int32 val32, uint32 chain = 0)
{
    return 0x9e3779b9 * (chain ^ (val32 >> 31) ^ val32 );
}

// The CNumeric classes are not really available a to DBOS sources.
// They exist in cnumeric.h but in a form that expects preprocessing
// before being copied to the ../sys/include directory.  The numeric.h
// file provides a TNumericDigit type and a single TNumericData class
// wrapping a maximal array of TNumericDigits.  Since this offers
// essentially zero abstraction I have chosen not to include that
// header.
//
// The hash for numerics (CNumeric32, CNumeric64, and CNumeric128) must
// match the preceding hash for 16 canonical integers.  This is easy to
// arrange since the 32 bit "numeric digits" are simply integers in the
// endian representation of the current environment.  If there is only a
// single numeric digit then we must xor it with 32 copies of the sign bit.
// For CNumeric64 and CNumeric128 the final folded xor is insensitive to
// the ordering of the numeric digits in memory.
//
// NOTE: This function will work only for nDigits = 1, 2, or 4.
//
inline static uint32 KHashNumeric(const void *buf, uint32 nDigits, uint32 chain = 0)
{
    const uint32 *digits = (const uint32 *) buf;
    uint32 digit0 = digits[0];
    uint32 folded = chain ^ digit0;
    --nDigits;
    if ( nDigits == 0 )
        folded ^= REPLICATE_SIGN32(digit0);
    else {
        folded ^= digits[1];
        if ( nDigits > 1 ) {
            folded ^= digits[2];
            folded ^= digits[3];
            }
        }
    return 0x9e3779b9 * folded;
}


// This function computes the hash of a buffer of bytes.  For successive
// sets of 4 bytes (or less at the end) it xors the concatenated 4 bytes
// into runningHash and multiplies by the magic constant.  Thus this loop
// executes 1/4 as many times as the CRC inner loop.  It is able to do
// this by exploiting the ability of both the x86 and the ppc (at least
// those ppc implementations being used at Netezza) to perform unaligned
// accesses directly in hardware without triggering an alignment trap.
//
// To achieve identical results on both the little-endian x86 and on the
// big-endian ppc the ppc code uses lwbrx (Load Word Byte Reverse Indexed)
// instructions to reorder 4 bytes of data to match the value that would
// be loaded on an x86.
//
#if __PPC__

#if __DCC__
asm uint32 KHash_LoadLittleEndian(const void *address)
{
% reg address
!
      lwbrx r3,r0,address
}
#endif

#if __GNUC__
static __inline__ uint32 KHash_LoadLittleEndian(const uint32 *address)
{
    uint32 rv;

    __asm__ (
    "lwbrx %0,0,%1"
    : "=r"(rv)
    : "r"(address), "m"(*address)
    : );

    return rv;
}
#endif

#endif // __PPC__

inline static uint32 KHashBytes(const void *buf, uint32 len, uint32 chain = 0)
{
    const uint8 *bytes = (const uint8 *) buf;
    uint32 runningHash = chain;
    uint32* pu = (uint32*)bytes;
    uint32 loopCount = len >> 2;
    if ( loopCount ) {
        do {
#if __PPC__
            runningHash ^= KHash_LoadLittleEndian(pu);
#else
            runningHash ^= *pu;
#endif
            pu++;
            runningHash *= 0x9e3779b9;
            --loopCount;
            }
        while ( loopCount );
        }
    uint32 remainder = len & 3;
    if ( remainder ) {
        // left justify chars we want; zero-fill remainder of register:
        //
        //   Remainder    shift        post-shift bytes in trailer
        //      1          24           pu+0,    0,    0,    0
        //      2          16           pu+1, pu+0,    0,    0
        //      3           8           pu+2, pu+1, pu+0,    0
        //
        uint32 finalShift = -32 + (remainder*8); // x86 LEA instruction
          finalShift = -finalShift;                // i.e. 32 - (remainder*8)
        uint8* pb = ((uint8*)pu) + remainder;
        uint32 trailer = 0;
        do {
            trailer <<= 8;
            trailer |= *--pb;
        } while (--remainder);
        trailer <<= finalShift;
        runningHash ^= trailer;
        runningHash *= 0x9e3779b9;
    }
    return runningHash;
}

// This function computes the hash of a text string.  It trims trailing
// spaces then calls KHashBytes() on what is left.
//
inline static uint32 KHashText(const void *str, uint32 len, uint32 chain = 0, uint8 space = ' ' )
{
    const uint8 *pc = (const uint8 *) str + len;
    uint32 trimmedLen = len;

    // strip trailing spaces
    while ( trimmedLen != 0 ) {
        if ( *--pc != space ) break;
        --trimmedLen;
        }
    return KHashBytes(str, trimmedLen, chain);
}

inline static uint32
KHashIntBuf(  // Compute 32-bit Khash value on passed buffer
    const void *pData,// data to be added to Khash
    uint32 byteCount, // byte count of pData
    uint32 chain )   // current running value of Khash
{
    switch (byteCount)  {
        case 1:
            return KHashInt( (int32)(*((int8*)pData)), chain);
        case 2:
            return KHashInt( (int32)(*((int16*)pData)), chain);
        case 4:
            return KHashInt( (int32)(*((int32*)pData)), chain);
        case 8:
            return KHashInt( (int64)(*((int64*)pData)), chain);
        default:
            NZ_ASSERT(0);
    }
    return 0;
}

inline static uint32
KHashNumericBuf(  // Compute 32-bit Khash value on passed buffer
    register const void *pData, // data to be added to Khash
    register uint32 byteCount,  // byte count of pData
    register uint32 chain)      // current running value of Khash val
{
    switch (byteCount) {
        case 4:
        case 8:
        case 16:
            return KHashNumeric(pData, byteCount / 4, chain);
        default:
            NZ_ASSERT(0);
    }
    return 0;
}

#endif
