#ifndef LDRPARSER_INCLUDED
#define LDRPARSER_INCLUDED

// ldrParser.h
//
// Record parser. Parses ascii records using a table definition into a list
// of binary data values.
//
// (C) Copyright IBM Corp. 2001, 2011  All Rights Reserved.
//
// Changes:
//
// 03-05-02  tho        added desciption to all the datatypes and their official
//                      and canonical names/aliases.
// 01-07-02  tho        Use inline data conversion for simple types, complex
//                      datatypes still uses dbcomps.
// 11-09-01  dick       Initial version.


// parser for DB2 for z/OS internal format: BEGIN
#include <map>
#include "ICUConverter.h"
#include "unicodeProperties.h"
#include "etString.h"
// parser for DB2 for z/OS internal format: END

class CPlan;
namespace recordformat
{
class ReplParser;
};

namespace et {
class ETOptions;
class Log;
};


#define MAX_BYTES_PER_NCHAR 4  // worst-case bytes to represent a UTF-8 character
enum { MAX_ASSIGNED_UNICODE_CODEPOINT = 0x10FFFD };

extern "C" int is_NFC (unsigned int*, unsigned int*);
extern "C" uint8 const unicodeProperties[];

namespace ldr {

#define MAX_FIELDS_PER_RECORD   4095
#define MAX_RECORD_SIZE         0x10000

#define BAD_REC_MAX             8000
#define BAD_FIELD_MAX           256
#define MAX_FILENAME_SIZE       1024
        
//==============================================================================
// Class: Parser - Processor object that parses records into a list of data
// values.
//
// The Parser is only ever created in shared memory as part of the xpparser node
// so the new/delete methods provide only that functionality.

class Parser;

typedef void ParserStatusCCb (void* cbObj, int64 records, int64 bytes);

class Parser {
friend class CGenCode;
public:
    // CREATORS
    Parser ();
    ~Parser ();

    void* operator new (size_t nBytes);
    void  operator delete (void* object);

    // MANIPULATORS
    void checkpoint()
        {
            m_rowsToBeProcessedCkpt = m_rowsToBeProcessed;
            m_rowsToBeSkippedCkpt = m_rowsToBeSkipped;
            m_rowsConsumedCkpt = m_rowsConsumed;
            m_haltedCkpt = m_halted;
            m_syntheticDistKeyCkpt = m_syntheticDistKey;
        }

    void rollback()
        {
            m_rowsToBeProcessed = m_rowsToBeProcessedCkpt;
            m_rowsToBeSkipped = m_rowsToBeSkippedCkpt;
            m_rowsConsumed = m_rowsConsumedCkpt;
            m_halted = m_haltedCkpt;
            m_syntheticDistKey = m_syntheticDistKeyCkpt;
        }

    void logError(int         errCode,
                  uint64      rejectCnt,
                  uint64      consumedCnt,
                  uint32      bytePos,
                  int         colNum,
                  const char* typeName,
                  const char* typeArg,
                  const char* badField,
                  const char* badStr);
        
    void logBadRecord(uint8 const* buffer, size_t size, bool suppressLinefeed = false);

    bool isUniqueDistKeyField(CTable *table, uint32 colNum);

    void init(const et::ETOptions*   loadOpts,
              const TXid&      txId,
              void*            tablePtr,
              void*            cbObj,
              ParserStatusCCb* cbFunc);

    // parser for DB2 for z/OS internal format: BEGIN
    struct FieldParseControl;

    std::string getDB2TypeName(FieldParseControl fpc);
    void initDB2InternalFormat(et::ETOptions const* loadOpts, TableShape const* shape);
    void logStats();
    void setBadFieldData(std::ostringstream& ossBadField, std::ostringstream& ossBadStr, uint8 const* buffer, uint8 const* bufferEnd,
            size_t const badFieldOffset, size_t const badFieldLength, size_t const badStrOffset, size_t const badStrLength,
            uint8 const*& badField, uint16 inputFieldLength = 0);

    /**
     * Verifies whether a given byte sequence consists of valid UTF-8 sequences.
     *
     * @param[in] ptr          starting address of the byte sequence to be verified
     * @param[in] length       length of the byte sequence to be verified
     * @param[out] ptrError    error case: memory address of the first invalid UTF-8 sequence
     * @param[out] utf8Status  error case: reason why the byte sequence that starts at @c ptrError is an invalid UTF-8 sequence
     * @return true, if the given byte sequence consists of valid UTF-8 sequences
     */
    inline bool validateUTF8(uint8 const* ptr, uint16 length, uint8 const*& ptrError, int& utf8Status) {
        uint8 const* ptrEnd = ptr + length;
        ptrError = NULL;
        utf8Status = _UTF8_BAD_CONTINUATION_BYTE;

        while (ptr < ptrEnd) {
            if (*ptr <= 0x7F) { // 1 B
                ++ptr;

                continue;
            } else if ((*ptr >= 0xC2) && (*ptr <= 0xDF)) {
                if ((*(ptr + 1) >= 0x80) && (*(ptr + 1) <= 0xBF)) { // 2 B
                    ptr += 2;

                    continue;
                } else {
                    ptrError = ptr + 1;
                }

                break;
            } else if (*ptr == 0xE0) {
                if ((*(ptr + 1) >= 0xA0) && (*(ptr + 1) <= 0xBF)) {
                    if ((*(ptr + 2) >= 0x80) && (*(ptr + 2) <= 0xBF)) { // 3 B
                        ptr += 3;

                        continue;
                    } else {
                        ptrError = ptr + 2;
                    }
                } else {
                    ptrError = ptr + 1;
                }

                break;
            } else if ((*ptr >= 0xE1) && (*ptr <= 0xEC)) {
                if ((*(ptr + 1) >= 0x80) && (*(ptr + 1) <= 0xBF)) {
                    if ((*(ptr + 2) >= 0x80) && (*(ptr + 2) <= 0xBF)) { // 3 B
                        ptr += 3;

                        continue;
                    } else {
                        ptrError = ptr + 2;
                    }
                } else {
                    ptrError = ptr + 1;
                }

                break;
            } else if ((*ptr == 0xED)) {
                if ((*(ptr + 1) >= 0x80) && (*(ptr + 1) <= 0x9F)) {
                    if ((*(ptr + 2) >= 0x80) && (*(ptr + 2) <= 0xBF)) { // 3 B
                        ptr += 3;

                        continue;
                    } else {
                        ptrError = ptr + 2;
                    }
                } else {
                    ptrError = ptr + 1;
                }

                break;
            } else if ((*ptr >= 0xEE) && (*ptr <= 0xEF)) {
                if ((*(ptr + 1) >= 0x80) && (*(ptr + 1) <= 0xBF)) {
                    if ((*(ptr + 2) >= 0x80) && (*(ptr + 2) <= 0xBF)) { // 3 B
                        ptr += 3;

                        continue;
                    } else {
                        ptrError = ptr + 2;
                    }
                } else {
                    ptrError = ptr + 1;
                }

                break;
            } else if (*ptr == 0xF0) {
                if ((*(ptr + 1) >= 0x90) && (*(ptr + 1) <= 0xBF)) {
                    if ((*(ptr + 2) >= 0x80) && (*(ptr + 2) <= 0xBF)) {
                        if ((*(ptr + 3) >= 0x80) && (*(ptr + 3) <= 0xBF)) { // 4 B
                            ptr += 4;

                            continue;
                        } else {
                            ptrError = ptr + 3;
                        }
                    } else {
                        ptrError = ptr + 2;
                    }
                } else {
                    ptrError = ptr + 1;
                }

                break;
            } else if ((*ptr >= 0xF1) && (*ptr <= 0xF3)) {
                if ((*(ptr + 1) >= 0x80) && (*(ptr + 1) <= 0xBF)) {
                    if ((*(ptr + 2) >= 0x80) && (*(ptr + 2) <= 0xBF)) {
                        if ((*(ptr + 3) >= 0x80) && (*(ptr + 3) <= 0xBF)) { // 4 B
                            ptr += 4;

                            continue;
                        } else {
                            ptrError = ptr + 3;
                        }
                    } else {
                        ptrError = ptr + 2;
                    }
                } else {
                    ptrError = ptr + 1;
                }

                break;
            } else if (*ptr == 0xF4) {
                if ((*(ptr + 1) >= 0x80) && (*(ptr + 1) <= 0x8F)) {
                    if ((*(ptr + 2) >= 0x80) && (*(ptr + 2) <= 0xBF)) {
                        if ((*(ptr + 3) >= 0x80) && (*(ptr + 3) <= 0xBF)) {
                            ptr += 4;

                            continue;
                        } else {
                            ptrError = ptr + 3;
                        }
                    } else {
                        ptrError = ptr + 2;
                    }
                } else {
                    ptrError = ptr + 1;
                }

                break;
            } else {
                ptrError = ptr;
                utf8Status = _UTF8_ILLEGAL_INITIAL_BYTE;

                break;
            }
        }

        return ptrError == NULL;
    }
    // parser for DB2 for z/OS internal format: END

    int ParseBinaryFormat(const uint8** pRemaining, const uint8* pEnd, CPlan *plan);
    int ParseFixedFormat(const uint8** pRemaining, const uint8* pEnd);
    int ParseReplFormat(const uint8** pRemaining, const uint8* pEnd);
    int ParseReplFormatDistKeysOnly(const uint8** pRemaining, const uint8* pEnd,
            uint32 *replFormatRecSize, int numOfDistKeyFields,
            const int *sortedByFieldIdDistKeys);

    void closeLogFiles ();
    void compactParseTypes();

    // for generated code...
    void GenPlanDoParse1(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse2(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse3(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse4(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse5(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse6(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse7(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse8(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse9(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse10(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse11(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse12(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse13(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse14(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse15(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse16(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse17(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse18(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse19(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);
    void GenPlanDoParse20(CPlan* plan, const uint8** pRemainingText, const uint8* pEndOfText, bool lastCall);

    // ACCESSORS
    bool    halted() const;
    uint64  maxErrors() const;
    uint64  rowsConsumed() const;
    uint64  bytesConsumed() const;
    uint64  rowsRejected() const;
    uint    rowsToBeSkipped() const;
    uint    badRowsSkipped() const;
    void    setEnableTime24(const bool flag);

    char const* parseTypeString(uint fpcParseDataType) const;
    char const* sqlTypeString(uint fpcParseDataType) const;

    char const* logFileName () const { return m_logFile.c_str(); }
    char const* badFileName () const { return m_badFile.c_str(); }

    char const* dbName() const { return m_dbName.c_str(); }
    char const* tblName() const { return m_tblName.c_str(); }
    char const* schemaName() const { return m_schemaName.c_str(); }
    char const* dataFile() const { return m_dataFile.c_str(); }

    void trapUnsupportedEncoding(int encoding, int unsupportedEncoding) const;

    // DATA MEMBERS
    enum RecordParseFlags {
        RPF_HAS_SIZE_WORD   = 0x00000001,
        RPF_HASH_VARYINGS   = 0x00000002,
        RPF_ENABLE_ROW_NUM  = 0x00000004,
        RPF_FILL_RECORD     = 0x00000008,
        RPF_TRUNC_STRING    = 0x00000010,
        RPF_TIME_ROUND_NANOS= 0x00000020,
        RPF_SHOW_PROGRESS   = 0x00000040,
        RPF_REPORT_STATS    = 0x00000080
    };

    const et::ETOptions*  m_loadOpts;

    uint                m_recordParseFlags;
    CTable*             m_table;
    TRowId*             m_pRowNum;

    size_t              m_fixedSize;
    uint16              m_fixedLongs;
    uint16              m_leaderLongs;
    uint16              m_nVarFields;
    uint16              m_firstBlobInPVS;
    uint8               m_nBlobFields;
    uint8               m_fieldDelim;
    uint8               m_quoteChar;
    uint8               m_escapeChar;
    uint8               m_sbcspcChar;
    uint8               m_dateDelim;
    uint8               m_dateTimeDelim;
    uint8               m_timeDelim;
    uint8               m_meridianDelim;
    uint8               m_eor2;
    uint8               m_nullTextLength;
    bool                m_requireQuotes;
    bool                m_halted;
    bool                m_partialRecord;
    uint64              m_rowsConsumed;
    uint64              m_bytesConsumed;
    uint64              m_rowsRejected;
    uint                m_rowsToBeProcessed;
    uint                m_rowsToBeSkipped;
    uint                m_badRowsSkipped;
    uint64              m_maxErrors;
    uint                m_nullText_0_3;
    bool                m_nullCaseFold;
    uint                m_nullMask_0_3;
    uint                m_y2Base;
    uint                m_stopperCtrlChars;
    uint                m_syntheticDistKey;
    uint                m_adjustZeroWrap;
    uint*               m_blobBuffer;
    size_t              m_blobBufferSize;
    bool                m_enableTime24;


    // fields that need to be checkpointed
    uint    m_rowsToBeProcessedCkpt;
    uint    m_rowsToBeSkippedCkpt;
    uint64  m_rowsConsumedCkpt;
    bool    m_haltedCkpt;
    uint    m_syntheticDistKeyCkpt;
    // end of checkpoint fields

    union typeEx {
        uint16  declSize;
        uint16  precAndScale;
        uint8   tsTimeFormat;   // assumes PT_TIME_xxx all non-zero
        uint8   dateDelim;
        uint8   dateTimeDelim;
        uint8   timeDelim;
        uint8   meridianDelim;
    };

    union typeEx2 {
        uint16  ccsid;          // CCSID
        uint16  decimalControl; // 4 bits for each precision step (32, 64, 96 & 128 bits)
    };

    // FieldParseControl and the FPF_ mask symbols describe how to parse one
    // field in an input row
    struct FieldParseControl {
        typeEx  extra;              // declared string length (or scale & precision)
        typeEx2 extra2;             // CCSID for strings, decimalControl for DECIMAL (DB2Z_BRF, DB2Z_RRF)
        uint16  fixedLength;        // length of fixed-length value (DB2Z_BRF, DB2Z_RRF)
        uint16  fixedLengthNULL;    // length of fixed-length value including NULL header (DB2Z_BRF, DB2Z_RRF)
        uint16  offset;             // physical position in record or PVS array
        uint16  rrfOffset;          // offset from record start to varying-length value offset (DB2Z_RRF)
        uint16  truncations;        // number of truncated values (DB2Z_BRF, DB2Z_RRF)
        int16   nullIndex;          // negative if not nullable
        uint8   flags;              // random controls bits (see below)
        uint8   parseDataType;      // index for switch() statement
        bool    rrfLastVarColumn;   // indicator for last varying-length column (DB2Z_RRF)
    };

#define FPF_LAST_FIELD          0x01
#define FPF_NULLABLE            0x02
#define FPF_UTF8                0x04
#define  FPF_NEGATE             FPF_UTF8
#define FPF_DATE_TRUNC_TIME     0x08
#define  FPF_DECIMAL_POINT      FPF_DATE_TRUNC_TIME
#define FPF_TIMESTAMP           0x10
#define  FPF_EXPECT_DIGIT       FPF_TIMESTAMP
#define FPF_EMPTY_DEFAULT       0x20
#define FPF_EMPTY_NULL          0x40
#define FPF_SUPRESS_LOAD        0x80

#if FPF_EMPTY_DEFAULT  != ' '
#error FPF_EMPTY_DEFAULT must equal value of ASCII space (0x20)
#endif

    FieldParseControl   m_FPC[MAX_FIELDS_PER_RECORD]; // logical order
    // parser for DB2 for z/OS internal format: BEGIN
    struct DB2ZDecimalCvtTables{
        // Order is very sensitive.  See initialize of decimal converters.
        uint8 neg2[256];
        uint8 pos2[256];
        uint8 neg1[256];
        uint8 pos1[256];
        char direct[256];

    };

    static DB2ZDecimalCvtTables cvtTables;
    
    std::map<uint16, ICUConverter*>* m_converters;
    int32 m_db2zPartitionID;
    bool m_incompleteRowInBuffer;
    bool m_firstExecution;
    nzotuid_t m_nzotuid;
    ptrdiff_t m_previousRemainingTextLength;
    char m_statsFile[MAX_FILENAME_SIZE];
    // parser for DB2 for z/OS internal format: END

    // Space to format at least the fixed part of the record
    uint64              m_record[MAX_RECORD_SIZE/sizeof(uint64)];

    // Descriptor for a recognized a VAR or VAR_FIX string in the input line
    struct ParsedVarString {
        const uint8*    body;           // text in original record
        uint16          length;         // final length in binary record
        uint8           pureCopy;       // unscrutinize vs drop null, esc & qd
        uint8           quoteChar;      // to check for quote doubling
    };

#define COALESCED_BLOB        0xFF
#define COALESCED_BLOB_LENGTH sizeof(uint*)

    // Descriptor for a recognized a BLOB or derivative in the input line
    struct ParsedBlob {
        uint*           body;           // NULL or length word in blob buffer
        uint            spare;          // match size of ParsedVarString
    };

// mask to support cheap testing for whether an ascii character is in the set
// ['A', 'a', 'C', 'c', 'G', 'g', 'T', 't', 'U', 'u']
#define BASE_BIT(LETTER) ( 1 << ( LETTER & 0x1F ) )
    enum { BASES_MASK = BASE_BIT('A')
                      | BASE_BIT('C')
                      | BASE_BIT('G')
                      | BASE_BIT('T')
                      | BASE_BIT('U')
         };
    
    // Space to accumulate string and blob descriptors
    ParsedVarString     m_PVS[MAX_FIELDS_PER_RECORD];

    //
    uint64              m_defaults[MAX_RECORD_SIZE/sizeof(uint64)];

    uint8               m_badRec[BAD_REC_MAX+2];
    uint8               m_badField[BAD_FIELD_MAX+2];

    // same as the values stored in et::ETOptions
    et::ETString        m_badFile;
    et::ETString        m_logFile;
    FILE *              m_badFilePtr;
    FILE *              m_logFilePtr;
    et::ETString        m_dbName;
    et::ETString        m_tblName;
    et::ETString        m_schemaName;
    et::ETString        m_dataFile;
    int                 m_ownerLogPid;
    int64               m_maxLogSizeBytes;

    int                 m_distKeyFieldUnique;

// Conditional field MUST be last
    enum { EMPTY_SLOT = 0xFF };
    uint8               m_compactedToParseType[256];
    uint8               m_PT_TIME_TZ;

    // Replication format parser
    recordformat::ReplParser *m_replParser;

    // CREATORS
    Parser (const Parser& src);              // Unimplemented
    Parser& operator= (const Parser& src);   // Unimplemented


    // Internal implementation

    static char const* PT_String[];
    static char const* Sql_String[];

    // 10^+(2^n) for n = 1..9; 10^-(2^n) for n = 1..9

    struct packedLongDouble {
        uint16 words[5];
    };

    static packedLongDouble xp_binaryPowersOf10[18];

    static uint powerOfTen[10];

    // see PT_NUMERIC16 parser
    static uint8 decimalToEncoded[];
    static uint8 encodedToDecimal[];


    static struct MonthInfo {
        uint16 Normal[16];
        uint16 LeapYear[16];
    } monthInfo;

    enum { MAX_YEAR = 10000 };
    static int yearInfo[MAX_YEAR];  // (day to 12/31 of prev year)*2 + is_leap_year
};

//------------------------------------------------------------------------------
// INLINE ACCESSORS
//
inline bool Parser::halted() const
{
    return m_halted;
}
        
inline uint64 Parser::maxErrors() const
{
    return m_maxErrors;
}
        
inline uint64 Parser::rowsConsumed() const
{
    return m_rowsConsumed;
}

inline uint64 Parser::bytesConsumed() const
{
     return m_bytesConsumed;
}

inline uint64 Parser::rowsRejected() const
{
    return m_rowsRejected;
}

inline uint Parser::rowsToBeSkipped() const
{
    return m_rowsToBeSkipped;
}

inline uint Parser::badRowsSkipped() const
{
    return m_badRowsSkipped;
}

inline void Parser::setEnableTime24(const bool flag)
{
    m_enableTime24 = flag;
}


//----------------------------------------------------------------------
// INLINE MANIPULATOR
//
inline char const* Parser::parseTypeString(uint fpcParseDataType) const
{
    // For SQL loader we map compacted values back to original PT enum
    char const* ret = 0;
    uint pt_code = m_compactedToParseType[fpcParseDataType];
    if ( pt_code != EMPTY_SLOT )
        ret = PT_String[pt_code];
    return ret;
}

inline char const* Parser::sqlTypeString(uint fpcParseDataType) const
{
    // For SQL loader we map compacted values back to original PT enum
    char const* ret = 0;
    uint pt_code = m_compactedToParseType[fpcParseDataType];
    if ( pt_code != EMPTY_SLOT )
        ret = Sql_String[pt_code];
    return ret;
}

};

#endif // LDRPARSER_INCLUDED
