#ifndef MEMINTERFACE_INCLUDE
#define MEMINTERFACE_INCLUDE
/// @file memInterface.h
/// @brief Memory Allocation Interface
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corporation 2011, 2014.  All rights reserved.
//////////////////////////////////////////////////////////////////////
/// This module provides the interface for memory allocation.  It is
/// intended to meet both shared and local memory allocation requirements for
/// NDE and host services and external clients. It is intended to replace
/// NzAlloc and the rc memory allocator.  It is not intended to supplant the
/// allocators used by postgres, emu, or fcomm.
///
/// There are two implementations of this interface:
///
///  1. memInterfaceHeap.cpp
///     This is a subset of the interface.  It provides no-ops for all of the
///     region calls.  It uses malloc for all allocations.  It only allows
///     REGION_HEAP allocations.  It puts block headers onto allocated blocks
///     and does basic verification.  This interface can be used where you need
///     to share code between external client applications (those that run on
///     platforms outside of the NPS hardware) and internal applications running
///     on the NPS.
///
///  2. memInterfaceRegion.cpp
///     This is the region-based interface, described more fully below, that
///     uses the lock-free memory allocator.  It is only intended for use on
///     internal applications running on Linux.  It is a superset, providing the
///     REGION_HEAP functions as well as real region-based allocations.
///
/// I chose the model of global functions in a namespace rather than a singleton
/// class primarily because it lets me keep all users of the interface ignorant
/// of any inner workings of the allocator.
///
/// Here is some vocabulary:
///
/// Segment         - a range of the virtual address space
/// Region          - a set of segments
/// Region instance - a memory allocation context joining a region and a memory
///                   allocator, for a particular overlay.  A
///                   single shared region have exactly one instance.  An
///                   overlay shared region has multiple instances.
///
/// Sometimes we use the term region when we mean region instance
///
/// KNOWN OBJECTS
/// -------------
/// One of the problems we have with intermixing unit tests and the global
/// system is the Nz pointer.  It points to a data structure that references
/// a huge number of types and pollutes the unit test environment.
///
/// In this interface, I have created the notion of a well-known object.  The
/// memory allocator pre-allocates an array of void* in each memory region.
/// An enumeration is provided to name the slots in the array.
///
/// There is a hazard of getting mismatched types with this scheme.
/// I had hoped to use boost::any, but it has two problems.  First, whenever you
/// construct a boost::any, it does a new - but there is no way to override the
/// memory allocator.  So the shared array points at one process' heap.  Second,
/// it relies on typeid() to validate the casting of the any into a typed
/// variable.  That is not reliable across different executables.
///
/// TODO: I have a couple of ideas to help, but for now, the interface just
/// takes and returns void*.
//////////////////////////////////////////////////////////////////////

#include "memTypeDefs.h"
#include "memAllocContext.h"

namespace mem {
    class FaultInterface;   // forward reference

////////////////////
/// RegionDesc - structure describing a region
////////////////////
struct RegionDesc
{
    regionName_t    m_regionName;   ///< the "name" of the region
    regionType_t    m_regionType;   ///< type of the region
    size_t          m_regionSize;   ///< size of the region in bytes
    size_t          m_minSegmentSize; ///< smallest segment of shared memory to map in bytes
    regionOverlay_t m_overlayCount; ///< number of overlays (instances) of the
                                    ///< region; for PROCESS and SHARED_SINGLE, must be 1
                                    ///< for SHARED_OVERLAY, number of overlays.
    int             m_concurrency;  ///< tuning param for the allocator
    numaPolicy_t    m_numaPolicy;   ///< tuning param: numa policy for the allocator
    bool            m_clearOnFree;  ///< clear memory when freed
    char            m_clearChar;    ///< pattern to store into the memory
    unsigned        m_ioalign;      ///< required io alignment for this region
}; // RegionDesc

////////////////////
/// MappingDesc - describes the mapping to be use used for a region
////////////////////
struct MappingDesc
{
    enum { MAX_MAP_COUNT = 4 };

    int             m_segmentCount;                 ///< actual number of segments
    void*           m_base[MAX_MAP_COUNT];          ///< array of segment addresses
    size_t          m_length[MAX_MAP_COUNT];        ///< array of segment lengths
};


////////////////////////////////////////
/// Region Management Interface
////////////////////////////////////////

    ////////////////////
    /// configureRegions - global region configuration
    ///
    /// Configure regions should be called exactly once by one process.  All
    /// other processes should call initProcess to read the configuration and
    /// set up their per-process state.
    ///
    /// Throws is regions cannot be configured as requested; e.g., not enough
    /// virtual address space for size.
    ///
    /// @param[in]  prefixName  - small prefix string to use for region names
    /// @param[in]  descCount   - size of the RegionDesc array
    /// @param[in]  config      - array of RegionDesc describing the regions to be configured
    /// @param[in]  mappings    - array of MappingDesc describing how the regions should be mapped
    ///                           If no array is provided, the OS will pick addresses.
    ///                           If an array is provided, then mappings must be specified for
    ///                           all configs; we will obey the mappings and clobber anything
    ///                           in the way.  Note: memConfigure calculates these.
    /// @return none
    ////////////////////
    void configureRegions(const char* prefixName,
                          int descCount,
                          RegionDesc* config,
                          MappingDesc* mappings = NULL);

    ////////////////////
    /// configureOverlays - set the number of overlays for a region
    ///
    /// We need to configure regions (and get them addresses in memory) before
    /// we know how many overlays the per-data slice region needs. So we allow
    /// the overlay count to be increased after the regions have been configured.
    /// The overlay count can only be increased.
    ///
    /// @param[in]  regionName      - region to configure
    /// @param[in]  overlayCount    - count to be set
    /// @return none
    ////////////////////
    void configureOverlays(regionName_t regionName, regionOverlay_t overlayCount);

    ////////////////////
    /// initProcess - initialize a process with the system memory configuration.
    ///
    /// @param[in]  prefixName  - small prefix string to use for region names
    /// @param[in]  heapid      - optional: set affinity of the process to an allocator heap.
    ///                           If 0, the heap is chosen by the allocator.
    ///                           If not 0, the heap is chosen using the heapid.
    /// @return true if process initialized; false if it was already initialized
    ////////////////////
    bool initProcess(const char* prefixName, long heapid = 0);

    ////////////////////
    /// finiProcess - unmaps all regions and resets the process state
    ///
    /// @return none
    ////////////////////
    void finiProcess();

    ////////////////////
    /// openRegion - make a region available for allocation
    ///
    /// Until a region is opened it cannot be used.  You cannot open two
    /// overlays of the same region.  You cannot open an overlay beyond the
    /// overlay count specified in the region description.
    ///
    /// To iterate over overlays of the same regionName, you must open and close
    /// them one at a time.
    ///
    /// @param[in]  regionName      - enumerated value "name" of the region to open
    /// @param[in]  overlayIndex    - overlay index - defaults to 0
    /// @return none
    ////////////////////
    void openRegion(regionName_t regionName, regionOverlay_t overlayIndex = 0);

    ////////////////////
    /// closeRegion - stop using a region
    ///
    /// The region and the particular overlay must be open in this process.
    /// They are closed.
    ///
    /// @param[in]  regionName      - enumerated value "name" of the region to open
    /// @param[in]  overlayIndex    - overlay index - defaults to 0
    /// @return none
    ////////////////////
    void closeRegion(regionName_t regionName, regionOverlay_t overlayIndex = 0);

    ////////////////////
    /// closeRegionAll - close any open regions
    ///
    /// This method is typically used on error paths to unmap regions.
    ///
    /// @return none
    ////////////////////
    void closeRegionAll();

    ////////////////////
    /// validateOverlay - test that the desired overlay is mapped
    ///
    /// @param[in]  regionName      - enumerated value "name" of the region
    /// @param[in]  overlayIndex    - overlay index to test
    /// @return true if the requested overlay is mapped; false otherwise
    ////////////////////
    bool validateOverlay(regionName_t regionName, regionOverlay_t overlayIndex);

    ////////////////////
    /// setOverlayCount - set the number of overlays for a region
    ///
    /// It turns out we need to configure regions (and get them addresses in
    /// memory) before we know how many overlays the per-data slice region
    /// needs. So we allow the overlay count to be increased after the regions
    /// have been configured.  The overlay count can only be increased.
    ///
    /// This should be done exactly once for the whole system before other
    /// processes map any regions.
    ///
    /// @param[in]  regionName      - region to set
    /// @param[in]  overlayCount    - count to be set
    /// @return none
    ////////////////////
    void setOverlayCount(regionName_t regionName, regionOverlay_t overlayCount);

    ////////////////////
    /// refreshRegion - have region allocator refresh, i.e., consolidate memory
    ///     into largest sizes.
    /// The region and the particular overlay must be open in this process.
    /// Only Global refresh presently implemented.
    ///
    /// @param[in]  regionName      - enumerated value "name" of the region to open
    /// @param[in]  overlayIndex    - overlay index - defaults to 0
    /// @return none
    ////////////////////
    void refreshRegion(regionName_t regionName, regionOverlay_t overlayIndex = 0);

    ////////////////////
    /// getRegionSize - return the size and available size of the region
    ///
    /// Note that the available size is an approximation.  The current lock-free
    /// allocator tends to under-report the size.
    ///
    /// If regionName is HEAP, we return 0 size; we don't know the heap size.
    /// In the heap-only allocator, this always returns 0.
    ///
    /// @param[in]      regionName    - name of the region to query
    /// @param[in]      overlayIndex  - overlay index - must be currently mapped
    /// @param[out]     availableSize - pointer to size; if not NULL, gets the
    ///                 size in bytes remaining available in the region.
    /// @return size of the region in bytes
    ////////////////////
    size_t getRegionSize(regionName_t regionName,
                         regionOverlay_t overlayIndex = 0,
                         size_t* availableSize = NULL);

    ////////////////////
    /// getOverlay - returns the currently mapped overlay
    ///
    /// If no overlay is mapped, NULL_OVERLAY is returned.
    /// This method was added to support the AutoRegion object.
    /// @see memAutoRegion.h
    ///
    /// @param[in]  regionName  - region whose overlay is returned
    /// @return current open overlay or NULL_OVERLAY
    ////////////////////
    regionOverlay_t getOverlay(regionName_t regionName);

    ////////////////////
    /// getMaxAddr - returns highest addr occupied
    ///
    /// This method was added to support the LiteHeap/PadHeapFile object.
    /// @see padHeapFile.h
    ///
    /// @return highest address in use across all region/overlay/segment
    ////////////////////
    void* getMaxAddr();

    ////////////////////
    /// setFaultInterface - sets the callback functions for faults/tracing
    ///
    /// If we are in a fault build, this will copy the function pointers into
    /// the static storage for this process.  These pointers cannot be global,
    /// since the code might not be linked in the same place in all processes;
    /// just like virtual function tables...
    ///
    /// Note that, by design, the callbacks could be used for things other than
    /// fault insertion, but that is the immediate motivation.
    ///
    /// @param[in]  faults  - pointer to a fault interface object
    /// @return none
    ////////////////////
    void setFaultInterface(FaultInterface* faults);

    ////////////////////
    /// setHeapId - set the heap id for a process
    ///
    /// This method can be used to (re)set the heap id for a process. See
    /// initProcess for details.
    ///
    /// @param[in]  heapid      - optional: set affinity of the process to an allocator heap.
    /// @return none
    ////////////////////
    void setHeapId(long heapid = 0);

////////////////////////////////////////
/// Allocation Interface
////////////////////////////////////////

    ////////////////////
    /// allocBlock - get some memory from a region
    ///
    /// Throws if the memory cannot be allocated.
    ///
    /// @param[in]  size    - byte size to allocate
    /// @param[in]  ctx     - reference to allocation context
    /// @return pointer to allocated memory
    ////////////////////
    void* allocBlock(size_t size, const AllocContext& ctx);

    ////////////////////
    /// allocBlockNothrow - allocate; return null on failure
    ///
    /// @param[in]  size    - byte size to allocate
    /// @param[in]  ctx     - allocation context
    /// @return pointer to allocated memory or NULL on failure
    ////////////////////
    void* allocBlockNothrow(size_t size, const AllocContext& ctx);

    ////////////////////
    /// callocBlock - allocate and memset the allocated memory to zero
    ///
    /// Throws if the memory cannot be allocated.
    ///
    /// @param[in]  size    - byte size to allocate
    /// @param[in]  ctx     - allocation context
    /// @return pointer to allocated and cleared memory
    ////////////////////
    void* callocBlock(size_t size, const AllocContext& ctx);

    ////////////////////
    /// reallocBlock - allocate new, copy, free old
    ///
    /// The semantics of reallocBlock are that if the incoming pointer is NULL,
    /// it behaves like an allocBlock.  That requires passing an AllocContext to
    /// steer the allocation to the correct region.
    ///
    /// That presents a problem: what is the right behavior when the block is
    /// already allocated and the block header does not match the AllocContext?
    /// One alternative that would be to actually move the block to a different
    /// region.  However, at this layer, we do not know the size of the old
    /// block; that is only known to the base memory allocator.  So without
    /// adding more requirements to the base allocator, we cannot support that
    /// semantic.  Instead, we choose to make a mismatch an error.  If the
    /// realloc is called with a context that does not match the allocated
    /// block, we barf.
    ///
    /// Semantics notes:
    ///  1. reallocBlock does not perform a reallocate if the new size is less than
    ///     the existing size.
    ///  2. reallocBlock does not maintain io alignment.  A request to
    ///     reallocate an aligned block is an error.  Requesting an aligned,
    ///     reallocated block is also an error.
    ///
    /// @param[in]  ptr     - pointer to previously allocated memory
    /// @param[in]  size    - size of new memory to allocate
    /// @param[in]  ctx     - allocation context
    /// @return pointer to new memory
    ////////////////////
    void* reallocBlock(void* ptr, size_t size, const AllocContext& ctx);

    ////////////////////
    /// freeBlock - free memory
    ///
    /// NOTE: it is safe to pass NULL to freeBlock.  It checks for NULL and
    /// quietly returns.  This allows it to be used unchanged for C++ delete
    /// object semantics.
    ///
    /// @param[in]  ptr     - pointer to previously allocated memory or NULL
    /// @return none
    ////////////////////
    void freeBlock(void* ptr);

    inline void freeBlock(const void* ptr)
    { freeBlock(const_cast<void*>(ptr)); };

    ////////////////////
    /// validateBlock - check for a valid header on a block
    ///
    /// A block is valid if it is:
    ///  1. has a properly formed header
    ///  2. owned by the memory subsystem
    ///  3. is currently allocated
    ///  4. matched the incoming memory type
    ///
    /// @param[in]  p   - pointer to the block
    /// @param[in]  mt  - required memory type of the block
    /// @return true if valid, false otherwise
    ////////////////////
    bool validateBlock(const void* p, memType_t mt);

    ////////////////////
    /// getMemType    - validate the header and return the memory type of a block
    ///
    /// @param[in]  p   - pointer to the block in question
    /// @param[out] mt  - pointer to a memType_t where we return the memory type
    /// @return true if the pointer was allocated from the memory subsystem and the
    ///         block is currently allocated; false otherwise.
    ////////////////////
    bool getMemType(const void* p, memType_t* mt);

    ////////////////////
    /// getMemRegion    - given a block, return its region
    ///
    /// The incoming block pointer may not have been allocated by the memory
    /// subsystem.  It may be on the stack.  In that case, we do not want to
    /// look at the block header.  The header address (p - 8) may not be a valid
    /// virtual address.  Even if it is valid, it may not actually be a real
    /// header, but just other junk on the stack.  See what happened in SWS-71754...
    ///
    /// Therefore, in the memRegion case, this method inspects the pointer to see
    /// if it is in one of the memory regions.  If so, then it inspects the
    /// block header.  There are three possible returns:
    ///
    ///   1. REGION_HEAP if the pointer is not in our memory mapping.
    /// If the pointer is in our region mapping then:
    ///   2. The region in the block header if the header looks valid
    ///   3. REGION_MAX if the block header is not valid.
    ///
    /// So we return HEAP for pointers allocated by memHeap and for objects on
    /// the stack, without looking at their block headers.
    ///
    /// In the memInterfaceHeap allocator, we simply return HEAP every time.
    ///
    /// @param[in]  p   - pointer to the block in question
    /// @return resulting region name or REGION_MAX
    ////////////////////
    regionName_t getMemRegion(const void* p);

    ////////////////////
    /// getFriendlySize - compute an allocator-friendly size
    ///
    /// The returned size is big enough to hold the incoming size, but can be
    /// bigger, if that fits better for the allocator.
    ///
    /// @param[in]  sz  - required size
    /// @return friendly size, greater than or equal to sz
    ////////////////////
    size_t getFriendlySize(size_t sz);

    ////////////////////
    /// setTrackingId - sets the value for the allocation tracking id for this process
    ///
    /// Note: the allocation context overrides this value
    ///
    /// @param[in]  trackType   - type of id this is; one of the memTrack_t enumerations
    /// @param[in]  trackId     - a 32-bit value - semantics unknown to the memory subsystem.
    /// @return none
    ////////////////////
    void setTrackingId(memTrack_t trackType, uint32_t trackId);


////////////////////////////////////////
/// Known Object Interface
////////////////////////////////////////

    ////////////////////
    /// setKnownObject - set the value of a known object
    ///
    /// We provide a way to ensure the slot is empty before it is written and
    /// throw if it isn't empty.
    ///
    /// @param[in]  regionName      - enumerated value "name" of the region
    /// @param[in]  ko          - identifies the known object
    /// @param[in]  whatever    - the value for the object
    /// @param[in]  mustBeEmpty - if true, throw on non-empty; otherwise, overwrite
    ///                           defaults to false
    /// @return none
    ////////////////////
    void setKnownObject(regionName_t regionName,
                        memKnownObject_t ko,
                        void* whatever,
                        bool mustBeEmpty = false);

    ////////////////////
    /// getKnownObject - retrieves the value of a known object
    ///
    /// @param[in]  regionName      - enumerated value "name" of the region
    /// @param[in]  ko              - identifies the known object
    /// @return void* value from slot ko
    ////////////////////
    void* getKnownObject(regionName_t regionName, memKnownObject_t ko);

}; // namespace mem

#endif // MEMINTERFACE_INCLUDE
