#ifndef MEMNEWDELETE_INCLUDE
#define MEMNEWDELETE_INCLUDE
/// @file memNewDelete.h
/// @brief Template for operator new and delete using memory subsystem
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corporation 2011, 2012.  All rights reserved.
//////////////////////////////////////////////////////////////////////
/// These template can be used to easily make operator new and delete methods for
/// a class.  You simply derive your class from the template.  See the template
/// below for parameter definitions.
///
/// In all cases, the resulting new operator will allocate the incoming size.
/// As a consequence, you cannot use this in the case where you want to allocate
/// a size different than sizeof(MyClass).
///
/// Each template generates two new operators.  The first is a standard C++ operator
/// new that only takes the size to allocate.  The second is a non-standard
/// operator new that takes a region argument directing the allocation to a
/// specific region.
///
/// Examples:
///
///     class MyClass : public mem::newDelete<> { ... }
///         allocate anonymous memory from the global shared region
///
///     class MyClass : public mem::newDelete<mt_table> { ... }
///         allocate mt_table memory from the global shared region
///
///     class MyClass : public mem::newDelete<mt_table, REGION_PM>
///         allocate mt_table memory from the current overlay of the partition manager region
///
///     class MyClass : public mem::newDelete<mt_table, REGION_GLOBAL,
///                                                         MEM_ALIGN, MEM_SCOPE_PERMANENT>
///         allocate mt_table memory, io aligned, tracked as permanent from the
///         global shared region.
///
///     class MyClass : public mem::newDeleteClear<mt_table> { ... }
///         allocate mt_table memory from the global shared region and
///         initialize it to zero.
///
//////////////////////////////////////////////////////////////////////

#include "memInterface.h"

namespace mem {

////////////////////
/// newDeleteBase -- variants of delete for newDelete and newDeleteClear
////////////////////
template <memType_t         mt, ///< memory type from memTypes.h
          mem::regionName_t nm, ///< region for allocation
          mem::memAlign_t   al, ///< no io alignment required
          mem::memScope_t   sc> ///< scope of allocation
class newDeleteBase
{
public:
    static void operator delete(void* p)
    { mem::freeBlock(p); };

    static void operator delete[](void* p)
    { mem::freeBlock(p); };

    static void operator delete(void* p, mem::regionName_t regionName)
    { mem::freeBlock(p); };

    static void operator delete[](void* p, mem::regionName_t regionName)
    { mem::freeBlock(p); };
};

////////////////////
/// newDelete -- variants of new, using mem::allocBlock instead of malloc
////////////////////
template <memType_t         mt = mt_anonymous,             ///< memory type from memTypes.h
          mem::regionName_t nm = mem::REGION_GLOBAL,       ///< region for allocation
          mem::memAlign_t   al = mem::MEM_NOALIGN,         ///< no io alignment required
          mem::memScope_t   sc = mem::MEM_SCOPE_TEMPORARY> ///< scope of allocation
class newDelete : public newDeleteBase<mt, nm, al, sc>
{
public:
    static void* operator new(size_t sz)
    { return mem::allocBlock(sz, mem::AllocContext(mt, nm, al, sc)); };

    static void* operator new[](size_t sz)
    { return mem::allocBlock(sz, mem::AllocContext(mt, nm, al, sc)); };

    static void* operator new(size_t sz, mem::regionName_t regionName)
    { return mem::allocBlock(sz, mem::AllocContext(mt, regionName, al, sc)); };

    static void* operator new[](size_t sz, mem::regionName_t regionName)
    { return mem::allocBlock(sz, mem::AllocContext(mt, regionName, al, sc)); };
};

////////////////////
/// newDeleteClear -- variants of new, using mem::callocBlock instead of malloc
////////////////////
template <memType_t         mt = mt_anonymous,             ///< memory type from memTypes.h
          mem::regionName_t nm = mem::REGION_GLOBAL,       ///< region for allocation
          mem::memAlign_t   al = mem::MEM_NOALIGN,         ///< no io alignment required
          mem::memScope_t   sc = mem::MEM_SCOPE_TEMPORARY> ///< scope of allocation
class newDeleteClear : public newDeleteBase<mt, nm, al, sc>
{
public:
    static void* operator new(size_t sz)
    { return mem::callocBlock(sz, mem::AllocContext(mt, nm, al, sc)); };

    static void* operator new[](size_t sz)
    { return mem::callocBlock(sz, mem::AllocContext(mt, nm, al, sc)); };

    static void* operator new(size_t sz, mem::regionName_t regionName)
    { return mem::callocBlock(sz, mem::AllocContext(mt, regionName, al, sc)); };

    static void* operator new[](size_t sz, mem::regionName_t regionName)
    { return mem::callocBlock(sz, mem::AllocContext(mt, regionName, al, sc)); };
};

}; // namespace mem

#endif // MEMNEWDELETE_INCLUDE
