#ifndef MEMPOOL_INCLUDE
#define MEMPOOL_INCLUDE
/// @file memPool.h
/// @brief Memory "pool" layered on top of base memory allocation
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corporation 2011, 2012.  All rights reserved.
//////////////////////////////////////////////////////////////////////
/// The memory pool provides a container from which memory can be allocated.
/// There is no concurrency control for allocations within a memory pool;
/// synchronization is the responsibility of the caller.  Concurrent use can
/// cause corruption.
///
/// The entire pool can be destroyed in one operation, which is handy for
/// cleanup.  As a consequence, on pool destruction, no destructors are called
/// on objects allocated from within the pool.  Again, it is up to the caller to
/// ensure correct object behavior.
///
/// All allocations are aligned to sizeof(void*).  You cannot use objects
/// allocated from this pool implementation for doing IO.  [NOTE: that would be
/// fairly simple to add, but it is extra work and there are no cases where it
/// is done, so don't bother for the time being.]
///
/// Blocks allocated from the pool do not have block headers.  There is no free
/// function.
///
/// In production, this code never throws, but the underlying memory allocator
/// will throw on out-of-memory.
///
/// SIZING NEW CHUNKS
///
/// In general, we want to handle many small allocations.  But previous
/// experience is that we would occasionally get a huge allocation and waste
/// free space in previous chunks.
///
/// We use the following algorithm:
///
/// - if the request is bigger than 3/4 K_BIG_SIZE:
///      --> we just allocate it and link it after the current head
///      This avoids having large allocations waste space.
///
/// - request is bigger than 1/2 K_BIG_SIZE:
///      --> double it, allocate it, and link it as the new head
///      This accelerates the doubling process beyond K_BIG_SIZE, but makes
///      sure we have enough space beyond the request size to allocate more.
///
/// - take max of request and current chunk size and allocate that.
///      --> double allocate size; but max is K_BIG_SIZE
///      This accelerates doubling, but puts a lid on so we don't allocate
///      huge chunks.
///
/// K_BIG_SIZE is chosen to be a friendly size to the memory allocator when
/// prefixed with the PoolChunk block.
///
/// TODO: Is a C interface really needed??
/// This interface provides a C++-object implementation of a pool.  There is a
/// C-function adapter that can be used:
///@see memPoolC.h
//////////////////////////////////////////////////////////////////////

#include "memInterface.h"

class CPlan;

namespace mem {

////////////////////
/// class PoolChunk - class that operates on a single chunk
////////////////////
class PoolChunk
{
public:
    PoolChunk(size_t sz) :
        m_next(NULL),
        m_size(sz),
        m_used(0)
    {};

    ////////////////////
    /// chunkCreate - Create a chunk with (at least) sz available for allocation
    ////////////////////
    static PoolChunk* chunkCreate(size_t sz, AllocContext ctx);

    ////////////////////
    /// chunkAlloc - allocate size from this chunk; return NULL if it is unavailable
    ///
    /// @param[in]  sz  - size to allocate; already aligned.
    /// @return pointer to allocated space; NULL if there is not enough space left
    ////////////////////
    void* chunkAlloc(size_t sz);

#ifdef NZDEBUG
    ////////////////////
    /// chunkValidate - check chunk for memory corruption
    ///
    /// This is only performed in debug builds; see memPool.cpp for
    /// the logic of when it is included.  It aborts if it detects corruption.
    ///
    /// @return none
    ////////////////////
    void chunkValidate();
#endif

    ////////////////////
    /// chunkSize - available space in the chunk
    ////////////////////
    size_t chunkSize()
    { return m_size; };

    ////////////////////
    /// chunkSizeUsed - space allocated from the chunk
    ////////////////////
    size_t chunkSizeUsed()
    { return m_used; };

public:
    PoolChunk*  m_next;     ///< Next pointer in the chunk list; linked newest to oldest
    size_t      m_size;     ///< Byte size of chunk space
    size_t      m_used;     ///< Bytes used from this chunk

// Windows warns on zero length array.  Turn off the warning.
#ifdef WIN32
#pragma warning(push)
#pragma warning(disable:4200)
#endif
    char        m_data[0];  ///< Access to trailing data
#ifdef WIN32
#pragma warning(pop)
#endif

}; // class PoolChunk

////////////////////
/// class Pool
////////////////////
class Pool
{
public:
    ////////////////////
    /// Destroy a Pool - a normal delete and destructor will do fine.
    ////////////////////
    static void operator delete(void* p)
    { return mem::freeBlock(p); };

    ~Pool()
    { poolReset(); };

    ////////////////////
    /// poolCreate - factory method for creating a pool object in the right context
    ///
    /// This will allocate a pool object in the specified allocation context and
    /// save the context for future chunk allocations.
    ///
    /// @param[in]  initialSize     - usable size after pool object overheads
    /// @param[in]  ctx             - context for memory allocation
    /// @return Constructed pool object.
    ////////////////////
    static Pool* poolCreate(size_t initialSize, AllocContext ctx, CPlan *plan = NULL);

    ////////////////////
    /// poolAlloc - allocate a block
    ///
    /// @param[in]  sz      - size of block to allocate
    /// @return pointer to allocated block
    ////////////////////
    void* poolAlloc(size_t sz);

    ////////////////////
    /// poolCalloc - allocate and zero a block
    ///
    /// @param[in]  sz      - size of block to allocate
    /// @return pointer to allocated block
    ////////////////////
    void* poolCalloc(size_t sz);

    ////////////////////
    /// poolReset - release all pool resources back to initialSize
    ////////////////////
    void poolReset();

    ////////////////////
    /// poolStrdup - strdup allocating from the pool
    ////////////////////
    char* poolStrdup(const char* str);

    ////////////////////
    /// poolSize - total memory allocated for this pool, not including structure overhead
    ////////////////////
    size_t poolSize()
    { return m_totalSize; };

    ////////////////////
    /// poolSizeUsed - total memory allocate from this pool; that is, allocated
    /// with poolAlloc or poolCalloc
    ////////////////////
    size_t poolSizeUsed()
    { return m_totalSizeUsed; };

    ////////////////////
    /// poolSizeWaste - sum of unallocated space in chunks other than the head
    ////////////////////
    size_t poolSizeWaste();

#ifdef NZDEBUG
    ////////////////////
    /// poolValidate - check pool for memory stomping
    ///
    /// This is only performed in debug builds; see memPool.cpp for
    /// the logic of when it is included.  It aborts if it detects corruption.
    ///
    /// @return none
    ////////////////////
    void poolValidate();
#endif

    ////////////////////
    /// setCallerContext - setter for the caller context
    ///
    /// Caller context is a void* that allows a caller to store a tidbit of info
    /// inside the pool object.  See objpool for a usage example.
    ///
    /// @param[in]  v - context to set
    /// @return none
    ////////////////////
    void setCallerContext(void* v)
    { m_callerContext = v; };

    ////////////////////
    /// getCallerContext - return caller context
    ///
    /// @return caller context
    ////////////////////
    void* getCallerContext()
    { return m_callerContext; };

    void setPlan(CPlan* plan);
 
protected:
    // Only used by the poolCreate and derived classes
    Pool(size_t useSize, AllocContext ctx, CPlan *plan = NULL) :
        m_plan(plan),
        m_head(&m_firstChunk),
        m_allocContext(ctx),
        m_totalSizeUsed(0),
        m_callerContext(NULL),
        m_firstChunk(useSize)
    {
        m_totalSize = m_firstChunk.chunkSize();
    };

protected:
    CPlan*          m_plan;         ///< This is going to exist only for pool and exprPool on SPU's
    PoolChunk*      m_head;         ///< Head of list of chunks.  We always
                                    ///< allocate from the head chunk.
    AllocContext    m_allocContext; ///< Context to use for chunk allocation
    size_t          m_totalSize;    ///< memory allocate for this pool
    size_t          m_totalSizeUsed;///< memory allocated from this pool
    void*           m_callerContext;///< context for caller (supports objpool)
    PoolChunk       m_firstChunk;   ///< Embedded first chunk
}; // class Pool

}; // namespace mem

#endif // MEMPOOL_INCLUDE
