#ifndef NODECOMMON_INCLUDED
#define NODECOMMON_INCLUDED

#include "nztypes.h"
#include "xptimer.h"
#include "xpplannerflags.h"

class CPlan;
class TableShape;
class CTable;
class CRecordStore;
class CCommBuf;

enum EPlanNodeLocus {
    invalidLocus = -1,
    inSPU,
    inDBOS
};

enum EPlanNodeTableType {
    tableAtSPU = 0,
    externalTable,
    systemTable,
    virtualTable
};

enum ENodeTypes {
    InvalidNode = -1,
    RootNode = 0,
    // Dead nodes removed and explicit numbering removed May 2007 -dk
    ScanNode = 1,
    RestrictNode,
    ProjectNode,
    SortNode,
    MaterializeNode,
    JoinNode,
    ReturnNode,
    SaveTempNode,
    UniqueNode,
    AggregateNode,
    SortedAggregateNode,
    InsertNode,
    DeleteNode,
    DeleteScanListNode,
    DeleteMergeNode,
    UpdateNode,
    DoneNode,
    DownloadTableNode,
    DropTableNode,
    ExistsJoinNode,
    HashNode,
    CrossProductNode,
    LimitNode,
    HashJoinNode,
    MergeJoinNode,
    ExprJoinNode,
    BindIndexNode,
    BitVectorNodeObs,  // obsolete
    MergeNode,
    TableSizeNode,
    ParserNode,
    UnloadNode,
    InsertStatsNode,
    HashPartNode,
    WindowAggrNode,
    ScanPrepNode,
    CreateTempTableNode,
    ScanBlockCountNode,
    UngroupNode,
    GroupingSetNode,
    FuncJoinNode,
    StreamAppendNode,
    ReorgPrepNode,
    ReorgStatsNode,
    ReorgMergeNode,
    ReclaimNode,
    DataSliceFilterNode,
    DataSliceHashNode,
    ConditionalBranchNode,
    BranchEndNode,
    // These next few entries allow provisional creation of a new
    // Node without rebuilding the world.
    Next09Node,
    LastNodeMarker            // MUST BE LAST
};

enum ESerializeMode {
    SerModeAll,        // entire node
    SerModeDefOnly,    // definition only (T*Def members)
    SerModeFingerprint // fingerprint only (similar to DefOnly)
};

struct pcTTzInt {
    int64 p64;
    int32 p32;
};

class SnippetNodeDef {
protected:
    ENodeTypes m_type;
    EPlanNodeLocus m_place;   // spu/dbos
    int m_flags;              // hints - NfJoinSequence, ...
};

class SnippetNodeRuntimeInfo {
protected:
    // For node stats gathering (e.g. planStats, virtabs...)
    uint32 m_rtFlags;       // Non-definition flags; cf. m_flags
    int64 m_numRecs;        // number of records output by this node
    int64 m_byteCount;      // node-specific, but usually size of output records
    CTimer64 m_cpuTime;     // ticks spent in this node
    int16 m_srcFailureCode; // reason code for NfRtNotSnipResCacheable
    int m_srcFailureAux;    // object id associated with srcFailureCode

    // For node stats gathering (e.g. planStats, virtabs...)
    // All nodes with the same statsTag get combined stats (presented in last).
    // If the detailStats is also set, the stats are gathered on a per-node
    // basis for virtabs, but reported combined for the front-end
    int64 m_statsTag;     // front-end supplied id for this node
    bool m_detailStats; // whether to collect non-combined, super-detailed info

    TableShape *m_result;  // output table shape of this node (may be same as input)
    CTable *m_resultTable; // records container for nodes that materialize records

    CPlan *m_plan;  // plan that contains this node

    CRecordStore *m_pRS; // Record set for this node / plan / data-slice (if materialization)
    // The srcHack member is a duplicate of the precedent stream info
    // This gets tossed as soon as all nodes converted over to record
    // streaming operation.
    TableShape *m_srcHack;    // for phase-in !NOTE-bmz
    CTable *m_srcHackTable;

    // For the code generator:   !NOTE-bmz make storage conditional compiled
    const TableShape *m_tblShape;

    char *m_comment;    // Postgres supplied comment string, for dumping
    int64       m_numRecsEst;
    unsigned    m_plannerFlags;

};

// This is an abstract base class for all concrete node types.
class SnippetNode : private SnippetNodeDef, private SnippetNodeRuntimeInfo {

public:
    SnippetNode *m_next;
    SnippetNode *m_prev; // doubly threaded for easy manipulation

    SnippetNode();
    virtual ~SnippetNode();
    
    // Some node types' Init functions don't care about 'place' argument.
    virtual const TableShape* Init( TableShape const * src, EPlanNodeLocus place ) = 0;
// dlm-check    virtual ~SnippetNode() { /* nothing to do here */ }
    // Note: Initially SnippetNode::Serialize is only called for Snippet Result Cache Fingerprint.
    // Real marshalling mostly occurs in CPlan::LoadCommBuff, and de-marshalling in
    // CPlan:getHdrFromCommBuf, CPlan::getExprFromCommBuf, and CPlan::getNodeFromCommBuf
    // (search for marshall in plan.cpp.)
    // Any node we run into that doesn't override the Serialize and IsValidForSrc methods
    //  invalidates snippet-result-cache-ability.
    // Note: Serialize is non-const, because it can set a failure reason.
    virtual bool IsValidForSrc() { return false; }
    virtual bool Serialize ( CCommBuf * cbuf, ESerializeMode mode ) { return false; }
    virtual bool SerializeBase( CCommBuf* cbuf, ESerializeMode mode );
    virtual void Deserialize( CCommBuf const * cbuf, ESerializeMode mode );

    // Address-of accessors
    // Used for things like virtual-table materialization.
    inline int64*    AddrOfByteCount()    { return &m_byteCount; }
    // This is the only accessor for m_cpuTime.  Don't pass it by value;
    // that would call the constructor, resetting it to zero.
    inline CTimer64* AddrOfCpuTime()        { return &m_cpuTime; }
    inline int64*    AddrOfNumRecs()        { return &m_numRecs; }
    inline uint32*   AddrOfRtFlags()        { return &m_rtFlags; }
    inline int*      AddrOfSrcFailureAux()  { return &m_srcFailureAux; }
    inline int16*    AddrOfSrcFailureCode() { return &m_srcFailureCode; }
    inline int64*    AddrOfStatsTag()       { return &m_statsTag; }

    // Accessors
    inline int64          ByteCount()       const { return m_byteCount; }
    inline char*          Comment()         const { return m_comment; }
    inline bool           DetailStats()     const { return m_detailStats; }
    inline int            Flags()           const { return m_flags; }
    inline int64          NumRecs()         const { return m_numRecs; }
    inline EPlanNodeLocus Place()           const { return m_place; }
    inline CPlan*         Plan()            const { return m_plan; }
    inline CRecordStore*  RecStore()              { return m_pRS; }
    inline TableShape*    Result()          const { return m_result; }
    inline CTable*        ResultTable()     const { return m_resultTable; }
    inline uint32         RtFlags()         const { return m_rtFlags; }
    inline int            SrcFailureAux()   const { return m_srcFailureAux; }
    inline int16          SrcFailureCode()  const { return m_srcFailureCode; }
    inline TableShape*    SrcHack()         const { return m_srcHack; }
    inline CTable*        SrcHackTable()    const { return m_srcHackTable; }
    inline int64          StatsTag()        const { return m_statsTag; }
    inline const TableShape*  TblShape()          { return m_tblShape; }
    inline ENodeTypes         Type()        const { return m_type; }
    inline int64          NumRecsEst()      const { return m_numRecsEst; }
    inline int            PlannerFlags ()   const { return m_plannerFlags; }
    inline bool           PlannerRestrictedSlice() const { return (m_plannerFlags & PfRestrictSlices) == 1; }

    // Testers
    virtual inline bool IsExprStyle()       const { return false; }
    virtual inline bool IsJoinType()        const { return false; }
    virtual inline bool IsSpu2SpuDownload() const { return false; }

    // Mutators
    inline void AddFlag( int flag )                     { m_flags |= flag; }
    inline void ClrFlag( int flag )                     { m_flags &= ~flag; }
    inline void AddRtFlags( uint32 val )                { m_rtFlags |= val; }
    inline void IncrByteCount( int64 val )              { m_byteCount += val; }

    // Non-void setters
    inline TableShape* SetResult( TableShape* pTable )  { m_result = pTable; return m_result; }

    // Setters
    inline void SetDetailStats( bool bb)                { m_detailStats = bb; }
    inline void SetFlags( int flags )                   { m_flags = flags; }
    inline void SetNumRecs( int64 num )                 { m_numRecs = num; }
    inline void SetPlace( EPlanNodeLocus place )        { m_place = place; }
    inline void SetPlan( CPlan* plan)                   { m_plan = plan; }
    inline void SetResultTable( CTable* pTable )        { m_resultTable = pTable; }
    inline void SetRecStore( CRecordStore* pRecStore )  { m_pRS = pRecStore; }
    inline void SetTblShape( const TableShape* shape )  { m_tblShape = shape; }
    inline void SetSrcFailure( int16 reason, int aux )  { m_srcFailureCode = reason; m_srcFailureAux = aux; }
    inline void SetSrcHack( TableShape* hack )          { m_srcHack = hack; }
    inline void SetSrcHackTable( CTable* hack )         { m_srcHackTable = hack; }
    inline void SetStatsTag( int64 tag )                { m_statsTag = tag; }
    inline void SetType( ENodeTypes type )              { m_type = type ; }
    inline void SetNumRecsEst( int64 num )              { m_numRecsEst = num; }
    inline void SetPlannerFlags( int flags )            { m_plannerFlags = flags; }

    // Non-inlines
    void setComment( const char* cmt );
    void SetComment( char* cmt );
};


// SnippetNodeDefn flags
enum NodeFlags {
    NfNone = 0,
    NfJoinSequence          = (1<<0),
    NfRightNotIn            = (1<<1), // For NOT IN and relOp ALL (Right-Not-Exists case)
    NfLeftNotIn             = (1<<2), // For NOT IN and relOp ALL (Left-Not-Exists case)
    NfReturnNoData          = (1<<3), // For NOT IN and relOp ALL to return no data
    NfPassThroughJoin       = (1<<4)  // For NOT IN and relOp ALL to return all stream rows
};

// SnippetNodeRuntimeInfo flags (as opposed to SnippetNodeDefn flags)
enum NodeRtFlags {
    NfRtNone = 0,
    NfRtNotSnipResCacheable = (1<<0)
};


#endif // NODECOMMON_INCLUDED
