/// @file nzRecordFormat.h
/// @brief 
///
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2012  All Rights Reserved.
//////////////////////////////////////////////////////////////////////
///
/// Description: Internal API. Declares the Netezza SQL Replication log file data
///              types as written to the physical log by the PTS
///
/// For version-to-version compatibility, the record format is self-describing.
/// Reusable LogClass and LogField records are stored in the CommitStream.
/// These do not change for the life of any TransactionStream using these
/// definitions. When the CommitStream changes, new LogClass and LogField
/// declarations are loaded.
///////////////////////////////////////////////////////////////////////


#ifndef NZRECORDFORMAT_H_
#define NZRECORDFORMAT_H_

#include <stdint.h>
#include <string>
#include <list>
#include <vector>
#include <map>
#include <utility>

#include "nzRecordFormatAllocator.h"
#include "nzRecordFormatByteBuffer.h"
#include "nzRecordFormatException.h"

//FIXME: (RecordFormat) Add diagnostic information to RecordRegistry, e.g. user assigned name.

namespace recordformat
{

class FieldClass;
struct FieldData;

typedef std::map<rfstring, rfstring, std::less<rfstring>, RecordFormatAllocatorType> map_string_string;
typedef std::vector<FieldClass*, RecordFormatAllocatorType> vector_FieldClass;
typedef std::vector<FieldData, RecordFormatAllocatorType> vector_FieldData;
typedef std::vector<uint8_t, RecordFormatAllocatorType> vector_uint8;

/**
 * The physical, byte-stream storage model for the field data-type
 *
 * These are stored on-disk as needed (for nullable and variable length fields)
 * and in the FieldClass metamodel
 *
 * To preserve the ability to add new types in the future, RecordFieldType is a
 * metadata class like FieldClass and RecordClass.
 *
 * To solve static initialization order problems and enable version to version
 * differences each RecordRegistry defines its own collection of RecordFieldTypes;
 * there aren't any global instances.
 *
 */

class RecordObject;
class RecordRegistry;
class RecordFormatByteBuffer;

/**
 * Known field types at the time of coding - future code may expand this set
 */
enum RecordFieldTypeId
{
    FIELD_TYPE_BOOL   = 0,         // 1 byte, true or false value

    FIELD_TYPE_UINT8  = 1,
    FIELD_TYPE_UINT16 = 2,
    FIELD_TYPE_UINT32 = 3,
    FIELD_TYPE_UINT64 = 4,

    FIELD_TYPE_INT8   = 5,
    FIELD_TYPE_INT16  = 6,
    FIELD_TYPE_INT32  = 7,
    FIELD_TYPE_INT64  = 8,

    FIELD_TYPE_STRING = 9,       /** Variable length "C" zero terminated string type */

    //FIXME: (RecordFormat) Need test case for handling of unknown type fields
    //FUTURE: (RecordFormat) Need fieldType extension mechanism

    COUNT_FIELD_TYPE         /** Enumerated value of the count of field types */
};

enum StringFormat
{
    NULL_TERMINATED = 0,
    TWO_BYTES_LENGTH,
    FOUR_BYTES_LENGTH,
};

struct RecordFieldType
{
    RECORD_FORMAT_ALLOCATION(RecordFieldType);
    friend class RecordRegistry;
public:
    /**
     * Get the name of a field type
     */
    char const * getName() const {
        return m_name.c_str();
    }

    /**
     * Get the name of the field type as a C type
     */
    char const * getCType() const {
        return m_cType.c_str();
    }

    bool isFixedSize() const {
        return m_fixedSize;
    }

    /**
     * The length of a single item packed in a RecordObject
     *
     * If the type is itself variable length this returns zero (0).
     */
    uint32_t getItemPackedSize() const {
        return m_itemPackedSize;
    }

    /**
     * The length of a single item in-memory. For structures the compiler
     * will compute a layout which may contain padding, causing the in-memory
     * instance to be larger the its packed RecordObject form.
     */
    uint32_t getItemMemorySize() const {
        return m_itemMemorySize;
    }

    /**
     * The start alignment of a single item in memory
     */
    uint32_t getItemMemoryAlignment() const {
        return m_itemMemoryAlignment;
    }

    bool operator == (RecordFieldType const& rhs) const {
        return m_typeId == rhs.m_typeId;
    }

    bool operator != (RecordFieldType const& rhs) const {
        return m_typeId != rhs.m_typeId;
    }

    const RecordFieldTypeId m_typeId;

private:

    const bool     m_fixedSize;
    const uint32_t m_itemPackedSize;
    const uint32_t m_itemMemorySize;
    const uint32_t m_itemMemoryAlignment;
    const rfstring m_name;
    const rfstring m_cType;

    RecordFieldType(RecordRegistry & registry, RecordFieldTypeId typeId, std::string const& name, std::string const& cType,
            bool fixedSize, uint32_t itemPackedSize, uint32_t itemMemorySize, uint32_t itemAlignment);

    //Do not implement!
    RecordFieldType(RecordFieldType const&);
    RecordFieldType const& operator=(RecordFieldType const&);
};

/**
 * Interface for dependent components to serialize "bootstrap" record types
 *
 * Tag record definitions in a ".record" file with "@bootstrap true" to
 * configure them to be bootstrap types. Bootstrap types are used to
 * negotiate a conversation and exchange record metadata, are exchanged
 * before non-bootstrap records can be used, and may not change without
 * breaking backward and forward compatibility.
 *
 * At a minimum the following record classes are "bootstrap":
 *   - RecordFormatHeader   - a header record
 *   - FieldTypeDefinition  - serialized RecordFieldType
 *   - RecordDefinition     - serialized RecordClass
 *   - FieldDefinition      - serialized FieldClass
 */
struct RecordSerializer
{
    RECORD_FORMAT_ALLOCATION(RecordSerializer);
public:
    virtual ~RecordSerializer(){};

    /**
     * Get the buffer to read the metadata conversation
     */
    virtual RecordFormatByteBuffer &getReadBuffer() = 0;

    /**
     * Get the buffer to write the metadata conversation
     */
    virtual RecordFormatByteBuffer &getWriteBuffer() = 0;

    /**
     * Callback to notify a record has been placed in the buffer.
     * Perform any needed action (write serialized record to file or network).
     */
    virtual void recordWriteComplete() = 0;

    /**
     * Callback to request a complete new record be placed in the buffer for reading
     *
     * @return the record typeid of the received record
     */
    virtual uint32_t receiveCompleteRecord(RecordObject &recordObject) = 0;
};


enum RecordFieldRepeat
{
      REPEAT_EXACTLY_ONE = 0 // a scalar field
    , REPEAT_EXACTLY_N =  4// a fixed length array of scalar fields
    , REPEAT_ZERO_OR_EXACTLY_N =  5 // a fixed length array of scalar nullable fields
    , REPEAT_ZERO_OR_ONE = 1  // a field with either a null or a single value
    , REPEAT_ZERO_OR_MORE = 2 // a nullable, repeating, variable length field
    , REPEAT_ONE_OR_MORE = 3  // a repeating, variable length field
};

std::string toString(RecordFieldRepeat repeat);

/**
 * The compression performed on a particular field before it was written to the log
 */
enum CompressionOptions
{
    COMPRESSION_RAW = 0,
    COMPRESSION_SNAPPY = 1,
    COMPRESSION_BZIP = 2,
    COMPRESSION_RESERVED = 3
};

/**
 * FUTURE: (LogFormat) Encrption algorithm performed on a field
 */
enum EncryptionOptions {
    ENCRYPTION_NONE=0, ENCRYPTION_AES256=1
};

/**
 * This class contains instance specific variable and nullable field metadata
 */
struct FieldData {
    RECORD_FORMAT_ALLOCATION(FieldData);
public:
    FieldData() : m_isNull(true), m_itemCount(0), m_memoryOffset(0), m_memoryBytes(0) {}
    FieldData(bool isNull, uint32_t itemCount) : m_isNull(isNull), m_itemCount(itemCount),
            m_memoryOffset(0), m_memoryBytes(0) {}

    bool m_isNull;
    uint32_t m_itemCount;  // Base item can be variable length if RecordFieldType.getFixedLength() == false

    uint32_t m_memoryOffset; // Start of field in unpacked, aligned memory buffer
    uint32_t m_memoryBytes; // Length in bytes of all items in field in unpacked, aligned memory buffer
                     // Count of base items in field in stream. total length can be calculated as length*count

    std::string toString() const;
};

/**
 * This class defines the common fields for all records and supports reading backward, forward, and current record structures
 */
class PTSProtocol;
class RecordClass;
class RecordObject
{
    RECORD_FORMAT_ALLOCATION(RecordObject);
    friend class RecordClass;
    friend class FieldClass;
    friend class PTSProtocol;
private:
    /**
     * First field written to each log record; structure id of entry
     *
     * Value is one of RecordType; value is uint8_t on disk; values not understood
     * can be skipped by advancing stream m_packedByteLength bytes after reading the full
     * log record header.
     */
    uint32_t m_recordType;
    RecordClass const* m_ptrRecordClass;

    /**
     * Physical size of this record in the stream excluding static record header.
     *
     * Any single log record is limited to 4 GB length. RecordObjects are to be
     * designed such that larger records can be broken up to fit within the limit.
     */
    uint32_t m_packedByteLength;

    /**
     * All fields
     */
    vector_FieldData m_fieldData;

    /*
     * Null bits handling
     */
    uint32_t m_nullBitsSize;
    uint32_t m_nullBitsLength;
    uint8_t *m_ptrNullBits;

    //unpacked record data - potentially 4G - actual header size
    uint32_t m_dataCount;
    uint8_t *m_ptrData;

    //packed / iterative read API data
    uint8_t *m_ptrPackedData;
    uint32_t m_currentNullBitIndex;
    uint8_t *m_ptrCurrentPackedFieldData;

    /**
     * Test whether a nullable or variable length field has content
     *
     * Private; only safe for FieldClass to call. Public API is FieldClass.isFieldNull(record)
     */
    bool isFieldNull(uint16_t fieldId) const;

    /**
     * Get the FieldData for a nullable or variable length field
     *
     * Private; only safe for FieldClass to call.
     */
    FieldData const& getFieldData(uint16_t fieldId) const;
    FieldData & getFieldData(uint16_t fieldId);

    /**
     * Field ids of those fields that failed version-to-version mustUndestand test during unpack(ByteBuffer &)
     */
    vector_uint8 m_versionToVersionFailures;

    /**
     * Read nullBits from the provided buffer
     *
     * Users must call after readStaticHeader and before readNextField
     */
    void readDynamicHeader(RecordFormatByteBuffer &);

public:

    //High order bit in fieldFlags indicates whether remaining bits are either:
    // * Value-flag (NULL or
    // * Offset in fieldData array
    static const size_t MAX_RECORD_FIELDS = 0x7FF;

    // 4 bytes recordType (TODO: (RecordFormat) 1 byte minimum with "UTF8" compression)
    // 4 bytes record size (TODO: (RecordFormat) 1 byte minimum with "UTF8" compression)
    static const size_t MIN_HEADER_SIZE = 4 + 4;

    //   4096 bytes of fieldFlags data
    //   24564 bytes of variableFieldData
    static const size_t MAX_HEADER_SIZE = 4 + 4 + 2 + 4096 + 2 + 24564;

    RecordObject(RecordRegistry const& registry);

    virtual ~RecordObject();

    vector_uint8 const& getVersionToVersionFailures() const { return m_versionToVersionFailures; }
    uint32_t getRecordType() const { return m_recordType; }
    uint32_t getRecordPackedSize() const { return m_packedByteLength; }
    RecordClass const& getRecordClass() const;
    uint8_t *getPackedData() {return m_ptrPackedData;}
    void *getAndIncrementPackedFieldData(uint32_t size)
    {
        void *data = m_ptrCurrentPackedFieldData;
        m_ptrCurrentPackedFieldData += size;
        return data;
    }


    /**
     * Read the static header from the provided buffer
     */
    uint32_t readStaticHeader(RecordFormatByteBuffer &);

    /**
     * Read record type from the provided buffer
     */
    uint32_t readRecordType(RecordFormatByteBuffer &);

    /**
     * Unpack the record from the stream format to the in-memory format
     */
    void unpack(RecordFormatByteBuffer &);

    /**
     * After calling readStaticHeader(), skip the body and advance the byte buffer to the next record
     */
    void skipRecordBody(RecordFormatByteBuffer &readBuffer);

    /*
     * API to iteratively read fields without necessarily unpacking all fields
     *
     * Fields are read sequentially from the beginning. Caller can stop processing
     * fields and proceed to the next record any time by calling skipRecordBody().
     */
    void readFields(RecordFormatByteBuffer & readBuffer);

};

/**
 * FieldClass captures record fields in the log files themselves.
 * These are used to serialize a record to the log format, and to dynamically
 * read the serialized log record.
 *
 * Since the FieldClass and RecordClass are defined for all versions
 * of the new log format, old versions can read newer versions and visa-versa.
 *
 */

const rfstring & getBootstrapStr();

class FieldClass
{
    RECORD_FORMAT_ALLOCATION(FieldClass);
    friend class RecordClass;
public:

    /**
     * THROW_ASSERT if parameters are incompatible with field declaration
     */
    void validate(RecordFieldType const& fieldType, bool isNull, uint32_t itemCount) const;

    /**
     * Does the field have a fixed item count (fixed length field)?
     */
    bool isFixedItemCountField() const
    {
        return (m_repeats == REPEAT_EXACTLY_ONE || m_repeats == REPEAT_ZERO_OR_ONE || m_repeats == REPEAT_EXACTLY_N || m_repeats == REPEAT_ZERO_OR_EXACTLY_N);
    }

    /**
     * Is the field nullable?
     */
    bool isNullableField() const
    {
        return m_isNullable;
    }

    void cacheNullBitPosition(uint32_t nullBitIndex)
    {
        m_nullFlagOffset = nullBitIndex / 8;
        uint32_t bitPosition = 7 - (nullBitIndex % 8);
        m_nullBitMask = 1 << bitPosition;
        m_isNullable = true;
    }

    /**
     * Does the field have a variable item count?
     */
    bool isVariableArrayField() const
    {
        return (m_repeats == REPEAT_ONE_OR_MORE || m_repeats == REPEAT_ZERO_OR_MORE);
    }

    RecordFieldType const& getFieldType() const
    {
        return *m_fieldType;
    }

    RecordRegistry const& getRegistry() const
    {
        return *m_registry;
    }

    /**
     * Get the count of items for a fixed length field; 0 for variable length fields
     */
    uint32_t getFixedItemCount() const { return m_fixedItemCount; }

    char const * getName() const {
        return m_fieldName.c_str();
    }

    /**
     * Return field id; which is the position of the field in the record
     */
    uint16_t getFieldId() const
    {
        return m_fieldId;
    }

//    /**
//     * Get the position of the first byte of this field in the record
//     *
//     * May be known from record class (for fixed width fields or first variable length field)
//     * or may have to be dynamically calculated.
//     */
//    uint32_t getSerializedStartOffset(RecordObject const& record) const;

    //FIXME: (RecordFormat) Implement compression and encryption tracking for fields
    CompressionOptions getCompression(RecordObject const& record) const;
    EncryptionOptions getEncryption(RecordObject const& record) const;

//    /**
//     * Get the length of the field in bytes from the record
//     *
//     * May be statically known, or may have to be computed for nullable and variable length fields
//     */
//    uint32_t getSerializedLengthInBytes(RecordObject const& record) const;

    /**
     * Test whether a value for the field was placed in the log record
     */
    bool isNull(RecordObject const& record) const
    {
        // Short-circuit non-nullable fields
        if(NZ_UNLIKELY(!isNullableField()))
            return false;
        // Reflect this field from the record instance
        return isNullForNullableField(record);
    }

    // Optimized version that requires (but does not check) isNullableField()
    bool isNullForNullableField(RecordObject const& record) const
    {
        // Reflect this field from the record instance
        return ((record.m_ptrNullBits[m_nullFlagOffset] & m_nullBitMask) != (uint8_t) 0);
    }

    /**
     * Get the actual count of items in the record for this field
     */
    uint32_t getItemCount(RecordObject const& record) const;

    /**
     * Obtain the itemCount and void* to packed data in a byte buffer
     *
     * Use this API as an alternative to unpack, either to read part of a record
     * or to avoid the expense of copying the data to the RecordObject memory buffer.
     *
     * The calling sequence is:
     *
     *     RecordFormatByteBuffer readBuffer;
     *     RecordObject record;
     *     FieldClass const* field = record.readFields(readBuffer);
     *     while(field != NULL && !terminationCriteriaMet()) {
     *         pair<uint32_t,void*> data;
     *         field = field->packedData(record, data);
     *         //process the data ...
     *     }
     *     //To make sure that the readBuffer is positioned at the end of the record:
     *     record.skipRecordBody();
     *
     * Packed multibyte data is in network byte order (big endian) and needs to
     * be ntoh / bswap'ed into host oreder by the caller. Packed data is likely
     * to be mis-aligned.
     *
     * The return value is the next field in the record, or NULL if this is the
     * last field.
     */
    FieldClass const* packedData(RecordObject &record,
            RecordFormatByteBuffer & readBuffer, std::pair<uint32_t,void*> &data) const;


    /**
     * Unpack this field from source <readBuffer> to destination <ptrData>
     *
     * Does not check for NULL field, caller is expected to
     * In case of version-to-version failure, does not raise error. Returns true if
     * version-to-version error is detected, but still unpacks if possible so reflection
     * API can be used to process contents.
     */
    bool unpack(FieldData & fieldData, uint8_t *ptrData, size_t dataSize,
            uint8_t *ptrCurrent, RecordFormatByteBuffer &readBuffer) const;

    /**
     * Obtain pointer to a bool log record field in a record
     */
    bool const* readBool(RecordObject const& record) const;

    /**
     * Obtain pointer to an char log record field in a record
     */
    char const* readChar(RecordObject const& record) const;

    /**
     * Obtain pointer to an 8-bit, single byte, unsigned log record field in a record
     */
    uint8_t const* readUint8(RecordObject const& record) const;

    /**
     * Obtain pointer to a 16-bit, two byte, unsigned log record field in a record
     */
    uint16_t const* readUint16(RecordObject const& record) const;

    /**
     * Obtain pointer to a 32-bit, two byte, unsigned log record field in a record
     */
    uint32_t const* readUint32(RecordObject const& record) const;

    /**
     * Obtain pointer to a 64-bit, two byte, unsigned log record field in a record
     */
    uint64_t const* readUint64(RecordObject const& record) const;

    /**
     * Obtain pointer to an 8-bit, single byte, signed log record field in a record
     */
    int8_t const* readInt8(RecordObject const& record) const;

    /**
     * Obtain pointer to a 16-bit, two byte, signed log record field in a record
     */
    int16_t const* readInt16(RecordObject const& record) const;

    /**
     * Obtain pointer to a 32-bit, two byte, signed log record field in a record
     */
    int32_t const* readInt32(RecordObject const& record) const;

    /**
     * Obtain pointer to a 64-bit, two byte, signed log record field in a record
     */
    int64_t const* readInt64(RecordObject const& record) const;

    /**
     * Obtain pointer to an char log record field in a record
     */
    const char** readString(RecordObject const& record) const;

    std::string toString(int indent = 0) const;

protected:

    FieldClass(std::string const& fieldName,
            RecordFieldType const& fieldType, RecordClass const& recordClass,
            RecordFieldRepeat repeats, uint32_t fixedItemCount,
            bool mustUnderstand, bool dontUnderstand, bool compressible, bool encrypt);

    /**
     * Position in a 'complete' record class
     */
    uint16_t m_fieldId;

    /**
     * Next field in the record, if not the last field
     */
    FieldClass const* m_nextField;
public:

    /**
     * Whether an occurrence of an unknown field is to be treated as a version to version failure.
     */
    bool m_mustUnderstand;

    /**
     * Whether an occurrence of this field is treated as a version to version failure.
     *
     * This is calculated on the assumption that fields defined in the base registry
     * are "understood"; fields not in the base registry (either because the field or
     * the entire class is not in the base registry) are understood.
     *
     * All fields in the base registry are automatically m_dontUnderstand = false
     */
    bool m_dontUnderstand;

    /**
     * Whether the log field can be compressed before being written
     *
     * Actual compression method used is written to the individual log record
     */
    bool m_compressible;

    /**
     * Whether the log field must be stored encrypted
     *
     * Actual encryption method used is written to the individual log record
     */
    bool m_encrypt;

    /**
     * RecordFieldType enumerated type of field
     */
    const RecordFieldType *m_fieldType;
    RecordFieldTypeId m_fieldTypeId;

    /**
     * Container for this field class; used to
     */
    const RecordRegistry *m_registry;

    /**
     * Container for this field class; used to
     */
    const RecordClass *m_recordClass;

    bool m_isNullable;
    uint32_t m_nullFlagOffset;
    uint8_t m_nullBitMask;

    /**
     * Whether records with NULL entries for this field can be written to the log
     *
     * Nullable fields are a kind of variable length, i.e. 0 or 1.
     *
     * passed to FieldClass constructor
     */
    RecordFieldRepeat m_repeats;

private:

    /**
     * Clone constructor
     *
     * Used to initialize fields of bootstrap record classes to be identical
     */
    FieldClass(RecordRegistry const& registry, FieldClass const& rhs);

    /**
     * Field items count for m_repeats=REPEAT_EXACTLY_ONE and REPEAT_EXACTLY_N; 0 for nullable
     * and variable length fields
     *
     * Use the instance values for nullable and variable length fields instead of the class values
     */
    uint32_t m_fixedItemCount;

    /**
     * The field name, unique within the record and not permitted to change for compatibility reasons
     */
    rfstring m_fieldName;
};

/**
 * RecordClass captures record fields in the log files themselves. These
 * are used to dynamically interrogate the log structure of a file at run-time.
 *
 * New log record and field classes are demand stored into the commit stream
 * for a given session stream, preceding the commit record. To read a given
 * session stream, the commit stream is first interrogated to determine if
 * any additional Class definitions are declared, and if so these are loaded.
 */
class RecordClass
{
    RECORD_FORMAT_ALLOCATION(RecordClass);
private:
    const uint32_t m_recordType;
    const rfstring m_recordName;
    const RecordRegistry *m_registry;

    /**
     * If a field is not binary compatible with a field in the base registry then the field is "dontUnderstand".
     * The occurrence of a dontUnderstand field for which the mustUndersand flag is set is treated as a
     * version to version failure.
     * If any field of a record class is must-understand and has repeats EXACTLY_ONE, EXACTLY_N, or ONE_OR_MORE
     * then the entire record class is must-understand.
     * In this case an occurrence of the record class is treated as a version to version failure.
    */
    bool m_dontUnderstand;

    /**
     * Fields in id-order; order is not fixed until "complete"
     *
     * Supports writer subclass hierarchies as base classes and subclasses
     * can intermingle fixed-length and variable length field declarations
     */
    // Note: m_fields.size will be slow because it has to iterate through the list
    // Note: m_fields cannot be a vector because a vector may reposition the fields and there are pointers to them
    vector_FieldClass m_fields;
    size_t m_fieldsSize; // Caches m_fields.size() because list::size() is slow

    /**
     * Number of nullable fields in this log record class
     */
    uint32_t m_nullableFieldCount;

public:

    RecordClass(RecordRegistry &registry, std::string const& name, int32_t type, bool dontUnderstand);
    RecordClass(RecordRegistry &registry, RecordClass const& toClone);

    virtual ~RecordClass();

    bool verifyMustUnderstand() const { return m_dontUnderstand; }

    uint32_t getNullableFieldCount() const { return m_nullableFieldCount; }
    // divide by 8 and round up using integer arithmetic
    uint32_t getNullBitsSize() const { return (getNullableFieldCount() + 7) / 8; }

    vector_FieldClass const& getFields() const { return m_fields; }

    void allocateFieldClassVector (size_t newSize);

    //Don't use m_fields.size(); list implementation is to walk the list
    uint16_t getFieldCount() const { return m_fieldsSize; }

    RecordRegistry const& getRegistry() const { return *m_registry; }
    uint32_t getRecordType() const { return m_recordType; }
    char const * getName() const { return m_recordName.c_str(); }

    /**
     * Declare a scalar (single item) field in the record class
     *
     * All fields must be statically declared before first use of any field.
     * Fields are stored in the record instance in the following order:
     * 1. Fixed-length non-nullable fields in the static order
     * 2. Nullable and variable-length fields in the order written
     *
     */
    FieldClass const* declareScalarField(std::string const& fieldName, RecordFieldType const& type,
            bool mustUnderstand, bool dontUnderstand, bool compressible, bool encryptable);

    /**
     * Declare fixed length array of fields of the same type
     */
    FieldClass const* declareFixedSizeArrayOfFields(
            std::string const& fieldName, RecordFieldType const& type, uint32_t fixedCount,
            bool mustUnderstand, bool dontUnderstand, bool compressible, bool encryptable);

    /**
     * Declare nullable fixed length array of fields of the same type
     */
    FieldClass const* declareNullableFixedSizeArrayOfFields(
            std::string const& fieldName, RecordFieldType const& type, uint32_t count,
            bool mustUnderstand, bool dontUnderstand, bool compressible, bool encryptable);
    /**
     * Declare a nullable scalar (0 or 1 item) field in the record class
     */
    FieldClass const* declareNullableScalarField(std::string const& fieldName, RecordFieldType const& type,
            bool mustUnderstand, bool dontUnderstand, bool compressible, bool encryptable);
    /**

     * Declare an array (1 or more items) field of scalar values in the record class
     */
    FieldClass const* declareArrayField(std::string const& fieldName, RecordFieldType const& type,
            bool mustUnderstand, bool dontUnderstand, bool compressible, bool encryptable);

    /**
     * Declare a nullable array (0 or more items) field of scalar values in the record class
     */
    FieldClass const* declareNullableArrayField(std::string const& fieldName, RecordFieldType const& type,
            bool mustUnderstand, bool dontUnderstand, bool compressible, bool encryptable);

    /**
     * Get a field by name
     *
     * In forward compatible scenarios, an older version may not have a field,
     * in which case NULL will be returned
     */
    FieldClass const* getField(std::string const& fieldName) const;
    FieldClass const* getField(uint32_t fieldId) const;

    /**
     * Print the record class definition, including all nested content
     */
    std::string toString(int indent = 0) const;

    /** deserialize RecordClass fields in a stream */
    void deserializeField(
            RecordRegistry & registry,
            std::string const& name,
            uint16_t fieldId,
            RecordFieldType const& fieldType,
            bool mustUnderstand, bool dontUnderstand,
            bool compressible,
            bool encrypt,
            RecordFieldRepeat repeat,
            uint32_t fixedItemCount);

    /**
     * Subclasses/Serializers must call this method once they finish defining the fields
     *
     * May be called multiple times to support record class hierarchies; recomputes from scratch
     * every time
     */
    void completeInitialization();

private:

    //Not implemented; not permitted to copy-construct
    RecordClass(RecordClass const&);

    FieldClass const* declareFixedLengthField(
            std::string const& fieldName, RecordFieldType const& type, RecordFieldRepeat repeat,
            uint32_t fixedItemCount, bool mustUnderstand, bool dontUnderstand);

    FieldClass const* declareVariableLengthField(
            std::string const& fieldName, RecordFieldType const& type,
            RecordFieldRepeat repeat, uint32_t count, bool mustUnderstand, bool dontUnderstand,
            bool compressible, bool encrypt);

    void cacheNullBitPosition(FieldClass *field);

protected:

    /**
     * Storage for code generator attributes aka "pragmas"
     */
    map_string_string m_pragmas;

public:

    /**
     * API for accessing pragma attributes
     */
    bool existsPragma(rfstring const& pragmaName) const;
    std::string getScalarPragma(std::string const& pragmaName) const;
    map_string_string const& getPragmas() const;

};


class RecordWriter
{
    RECORD_FORMAT_ALLOCATION(RecordWriter);
public:
    virtual ~RecordWriter();

protected:

    RecordWriter(RecordClass const& recordClass);

    void init(RecordFormatByteBuffer *recDataBuffer);

    //Fixed length field types; called from writeRecord to output actual data
    void writeBoolField(FieldClass const& field, bool const* value, uint32_t len = 1);
//    void writeCharField(FieldClass const& field, const char * value, uint32_t len = 1);

    void writeUint8Field(FieldClass const& field, uint8_t const* value, uint32_t len = 1);
    void writeUint16Field(FieldClass const& field, uint16_t const* value, uint32_t len = 1);
    void writeUint32Field(FieldClass const& field, uint32_t const* value, uint32_t len = 1);
    void writeUint64Field(FieldClass const& field, uint64_t const* value, uint32_t len = 1);

    void writeInt8Field(FieldClass const& field, int8_t const* value, uint32_t len = 1);
    void writeInt16Field(FieldClass const& field, int16_t const* value, uint32_t len = 1);
    void writeInt32Field(FieldClass const& field, int32_t const* value, uint32_t len = 1);
    void writeInt64Field(FieldClass const& field, int64_t const* value, uint32_t len = 1);

    //Variable length types
    void writeStringField(FieldClass const& field, char const** value, StringFormat format, uint32_t itemCount = 1);
    void writeStringField(FieldClass const& field, std::string const* value, uint32_t len = 1);

    void writeScalarBoolField(FieldClass const& field, bool const* value);
//    void writeCharField(FieldClass const& field, const char * value, uint32_t len = 1);

    void writeScalarUint8Field(FieldClass const& field, uint8_t const* value);
    void writeScalarUint16Field(FieldClass const& field, uint16_t const* value);
    void writeScalarUint32Field(FieldClass const& field, uint32_t const* value);
    void writeScalarUint64Field(FieldClass const& field, uint64_t const* value);

    void writeScalarInt8Field(FieldClass const& field, int8_t const* value);
    void writeScalarInt16Field(FieldClass const& field, int16_t const* value);
    void writeScalarInt32Field(FieldClass const& field, int32_t const* value);
    void writeScalarInt64Field(FieldClass const& field, int64_t const* value);

    //Variable length types
    void writeScalarStringField(FieldClass const& field, char const** value, StringFormat format);
    void writeScalarStringField(FieldClass const& field, std::string const* value);

    void writeRecordLength();

    RecordClass const* m_ptrRecordClass;
    RecordFormatByteBuffer *m_recordDataBuffer;
    uint32_t m_fieldFlags_startOffset;

    uint8_t m_lastFieldId;
    uint32_t m_current_nullBit_index;
    uint8_t *m_nullBits;
    uint32_t m_nullBitsSize;

    void writeFieldHeader (FieldClass const& field, uint32_t count, bool isNull);

private:
    //Only for use by static writing code
    RecordFieldType const* m_boolFieldType;
    RecordFieldType const* m_uint8FieldType;
    RecordFieldType const* m_uint16FieldType;
    RecordFieldType const* m_uint32FieldType;
    RecordFieldType const* m_uint64FieldType;
    RecordFieldType const* m_int8FieldType;
    RecordFieldType const* m_int16FieldType;
    RecordFieldType const* m_int32FieldType;
    RecordFieldType const* m_int64FieldType;
    RecordFieldType const* m_stringFieldType;
};


struct RecordFormatHeaderReader;
struct FieldTypeDefinitionReader;
struct RecordDefinitionReader;
struct FieldDefinitionReader;
struct RecordFormatHeaderRecordClass;
struct FieldTypeDefinitionRecordClass;
struct RecordDefinitionRecordClass;
struct FieldDefinitionRecordClass;

typedef std::map<uint32_t, RecordFieldType const*, std::less<uint32_t>, RecordFormatAllocatorType> map_uint32_RecordFieldType;
typedef std::map<rfstring, RecordFieldType const*, std::less<rfstring>, RecordFormatAllocatorType> map_string_RecordFieldType;
typedef std::map<uint32_t, RecordClass const*, std::less<uint32_t>, RecordFormatAllocatorType> map_uint32_RecordClass;
typedef std::map<rfstring, RecordClass const*, std::less<rfstring>, RecordFormatAllocatorType> map_string_RecordClass;

/*
 * Record registry for log records or messages
 *
 * Pointers registered are not managed by this registry object. They may
 * be statically allocated (by generated code) or they may be dynamically
 * allocated by a protocol session or a log record reader (in which case
 * memory management is performed by the allocator and not by the registry
 * instance).
 */
class RecordRegistry {
    RECORD_FORMAT_ALLOCATION(RecordRegistry);
    friend class RecordClass;
    friend class RecordFieldType;
private:
    //FieldType registry so new field types can be added in a forward compatible way.
    map_uint32_RecordFieldType m_fieldTypesById;
    map_string_RecordFieldType m_fieldTypesByName;
    //Storage for field types
    uint32_t m_fieldTypesCount;
    RecordFieldType *m_ptrFieldTypes;

    //Record registry so new record definitions can be added in a forward compatible way.
    map_uint32_RecordClass m_recordClassesByType;
    map_string_RecordClass m_recordClassesByName;

    //Keeps track of dynamically allocated RecordClass ids
    uint32_t m_nextRecordId;

public:

    RecordRegistry();
    virtual ~RecordRegistry();

    std::string toString() const;

    /**
     * Map a type name to a defined RecordFieldType value
     *
     * @return pointer to matching type, or null if the name does not exist in the compiled code.
     */
    RecordFieldType const* getFieldType(std::string const& name) const;

    /**
     * Map a type id to a defined RecordFieldType value
     *
     * Throws ASSERT if the typeId is undefined
     */
    RecordFieldType const& getFieldType(uint32_t typeId) const;

    uint32_t getFieldTypeCount() const {
        return m_fieldTypesById.size();
    }

    map_uint32_RecordFieldType const& getFieldTypes() const {
        return m_fieldTypesById;
    }

    RecordClass const& getRecordClassByType(uint32_t type) const;
    RecordClass const& getRecordClassByName(std::string const& name) const;

    uint32_t getRecordClassCount() const {
        return m_recordClassesByName.size();
    }

    map_uint32_RecordClass const& getRecordClassesByType() const {
        return m_recordClassesByType;
    }

    map_string_RecordClass const& getRecordClassesByName() const {
        return m_recordClassesByName;
    }

    /**
     * Add a registry member RecordFieldType
     */
    void registerFieldType(RecordFieldType const& fieldType);

    /**
     * Method to deserialize and load a RecordFieldTypeDefinition as a RecordFieldType
     */
    uint32_t registerFieldType(RecordFieldTypeId typeId, std::string const& name, std::string const& cType,
            bool fixedSize, uint32_t itemPackedSize, uint32_t itemMemorySize, uint32_t itemAlignment);

    /**
     * Method to deserialize and load a RecordDefinition as a RecordClass
     */
    uint32_t registerRecordClass(std::string const& messageName, RecordClass const* recordClass, int32_t type);

    /**
     * Interface for dependent components to serialize all record types
     *
     * "Bootstrap" record definitions are asserted to be common to all record registries
     * past, present and future.
     * Tag record definitions with "@bootstrap true" to cause them to be bootstrap types
     * The following record classes are always "bootstrap":
     *   - RecordFormatHeader   - a header record
     *   - FieldTypeDefinition  - serialized RecordFieldType
     *   - RecordDefinition     - serialized RecordClass
     *   - FieldDefinition      - serialized FieldClass
     */
    void serializeRecordMetadata(RecordSerializer &serializer, bool serializeBootstrapped = true) const;

    /**
     * Deserialize record metadata into this generic RecordRegistry instance
     */
    void deserializeRecordMetadata(RecordRegistry const& baseRegistry,
            RecordSerializer & serializer,
            RecordObject &record);

    void allocateDynamicRecordClassArray(uint32_t recordClassCount);
    void deallocateDynamicRecordClassArray();
    void deallocateFieldTypesArray();
    RecordClass *addDynamicRecordClass(std::string const& name, int32_t type, bool dontUnderstand);

    void handleRecordFormatHeader(RecordFormatHeaderReader &header);
    void handleFieldTypeDefinition(FieldTypeDefinitionReader &fieldTypeReader,
            RecordRegistry const& baseRegistry);
    void handleRecordDefinition(RecordDefinitionReader &recordDefinitionReader,
            RecordRegistry const& baseRegistry, RecordClass const **baseRecordClass, RecordClass **dynamicRecordClass);
    void handleFieldDefinition(FieldDefinitionReader &fieldReader, uint16_t fieldId,
            RecordRegistry const& baseRegistry, RecordClass const *baseRecordClass, RecordClass *dynamicRecordClass);

    //Base field types: extensions can add new field types using registerFieldType

    //These members need to come at the end of the class definition, in order of
    //RecordFieldType to RecordClass, so that dependencies are correctly initialized
    //during construction
    const RecordFieldType m_boolFieldType;
//    const RecordFieldType m_charFieldType;
    const RecordFieldType m_uint8FieldType;
    const RecordFieldType m_uint16FieldType;
    const RecordFieldType m_uint32FieldType;
    const RecordFieldType m_uint64FieldType;
    const RecordFieldType m_int8FieldType;
    const RecordFieldType m_int16FieldType;
    const RecordFieldType m_int32FieldType;
    const RecordFieldType m_int64FieldType;
    const RecordFieldType m_stringFieldType;

    //Record classes must be declared after the RecordFieldTypes on which they depend
    const RecordFormatHeaderRecordClass *m_recordFormatHeaderClass;
    const FieldTypeDefinitionRecordClass *m_fieldTypeDefinitionRecordClass;
    const RecordDefinitionRecordClass *m_recordDefinitionRecordClass;
    const FieldDefinitionRecordClass *m_fieldDefinitionRecordClass;

    /**
     * Container for dynamic record classes stored in this registry
     * Used for non-bootstrap messages (i.e. those configured using a class member).
     * Space for all dynamic definitions is reserved in advance so pointers
     * into the collection can be used.
     */
    uint32_t m_dynamicRecordClassCount;
    RecordClass *m_dynamicRecordClasses;


private:
    // Not implemented - do not call
    RecordRegistry(const RecordRegistry &);
    RecordRegistry &operator=(const RecordRegistry &);
};

inline
RecordClass const& RecordObject::getRecordClass() const
{
    return *m_ptrRecordClass;
}

inline
FieldClass const* RecordClass::getField(uint32_t fieldId) const
{
    if (NZ_UNLIKELY(fieldId >= m_fields.size()))
    {
        return (FieldClass*) NULL;
    }
    else
    {
        return m_fields[fieldId] ;
    }
}

inline
void RecordObject::readFields(RecordFormatByteBuffer & readBuffer)
{
    readDynamicHeader(readBuffer);

    m_ptrCurrentPackedFieldData = readBuffer.getAddress() + readBuffer.getOffset();
    m_currentNullBitIndex = 0;
}

inline
uint32_t RecordObject::readRecordType(RecordFormatByteBuffer & readBuffer)
{
    //Preserve offsets in the read buffer relative to the start of the record
    uint32_t startOffset = readBuffer.getOffset();

    m_packedByteLength = readBuffer.getInt4();
    m_recordType = readBuffer.getInt4();

    return m_recordType;
}

inline
uint32_t RecordObject::readStaticHeader(RecordFormatByteBuffer & readBuffer)
{
    //Preserve offsets in the read buffer relative to the start of the record
    uint32_t startOffset = readBuffer.getOffset();

    m_packedByteLength = readBuffer.getInt4();
    m_recordType = readBuffer.getInt4();

//    PTS_TRACEALWAYS("DEBUG: readStaticHeader startOffset=" << startOffset << " readBuffer.getOffset()=" << readBuffer.getOffset() << " recordType=" << m_recordType << " packedByteLength=" << m_packedByteLength);

    RecordClass const* ptrPrevRecordClass = m_ptrRecordClass;
    if(NZ_UNLIKELY(ptrPrevRecordClass != NULL && ptrPrevRecordClass->getRecordType() != m_recordType))
    {
        //RecordObject being re-used for a different RecordClass instance;
        //Get the RecordClass from the previous RecordClasses registry
        m_ptrRecordClass = &ptrPrevRecordClass->getRegistry().getRecordClassByType(m_recordType);

        //m_nullBitsSize is the size of the null bits for the current record
        //If records with larger null bits were read then m_nullBitsLength will
        //be greater then m_nullBitsSize
        m_nullBitsSize = m_ptrRecordClass->getNullBitsSize();
        if(m_nullBitsSize > m_nullBitsLength)
        {
            RecordFormatAllocatorType::rebind<uint8_t>::other allocator;
            allocator.deallocate(m_ptrNullBits);
            // Over-allocate to minimize re-allocation
            m_nullBitsLength = m_ptrRecordClass->getNullBitsSize() + 16;
            m_ptrNullBits = allocator.allocate(m_nullBitsLength);
        }
    }

    m_ptrPackedData = readBuffer.getAddress() + startOffset;

    return m_packedByteLength - MIN_HEADER_SIZE;
}


inline
void RecordObject::readDynamicHeader(RecordFormatByteBuffer &readBuffer)
{
    if(NZ_UNLIKELY(!m_fieldData.empty()))
        m_fieldData.clear();

    //Clear any left-over version to version field IDs
    if(NZ_UNLIKELY(!m_versionToVersionFailures.empty()))
        m_versionToVersionFailures.clear();

    //Compute the size of the variable length header portion
    RecordClass const& recordClass = *(m_ptrRecordClass);

    //Calculate in-memory buffer size to unpack the entire record
    //Assume 8 bytes of alignment on every field
    uint32_t memoryDataSize = m_packedByteLength + 8 * recordClass.getFieldCount();
    if(NZ_UNLIKELY(memoryDataSize > m_dataCount))
    {
        RecordFormatAllocatorType::rebind<uint64_t>::other allocator;
        if (m_ptrData != NULL)
        {
            allocator.deallocate(reinterpret_cast<uint64_t*> (m_ptrData));
            m_ptrData = NULL; // Prevent duplicate delete if new fails
        }
        //Use uint64_t to force dword alignment of the data buffer; over-allocate to minimize re-allocation
        memoryDataSize += 1024; //Over-allocate to minimize reallocation
        uint32_t dwordDataSize = (memoryDataSize / 8 + ((memoryDataSize % 8 > 0) ? 1 : 0));
        m_ptrData = reinterpret_cast<uint8_t*> (allocator.allocate(dwordDataSize));
        m_dataCount = dwordDataSize * 8;
    }

    //Swizzle the null bits into an in-memory array
    if(NZ_LIKELY(m_nullBitsSize > 0))
    {
        readBuffer.getBytesArray(&m_ptrNullBits[0], m_nullBitsSize);
    }
}


} /* namespace recordformat */

#endif /* NZRECORDFORMAT_H_ */
