/// @file nzRecordFormatByteBuffer.h
/// @brief Handles the record format byte buffer.
///
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2011, 2013  All Rights Reserved.
//////////////////////////////////////////////////////////////////////
///
/// Description:
///
/// Handles the record format byte buffer.
///
//////////////////////////////////////////////////////////////////////

#ifndef NZRECORDFORMATBYTEBUFFER_H
#define NZRECORDFORMATBYTEBUFFER_H

#include <string>
#include <stdint.h>
#include <arpa/inet.h>
#include <vector>
#include <iostream>
#include <sstream>
#include <iomanip>
#include "nzstd.h"

#include "nzRecordFormatAllocator.h"

namespace recordformat {

class RecordFormatByteBuffer
{
    RECORD_FORMAT_ALLOCATION(RecordFormatByteBuffer);
public:
    RecordFormatByteBuffer(size_t size);
    RecordFormatByteBuffer(uint8_t *buffer);
    RecordFormatByteBuffer(uint8_t *buffer, size_t bufferSize);
    virtual ~RecordFormatByteBuffer();
    void replaceBuffer(uint8_t *buffer, size_t bufferSize);

	uint8_t *getAddress() const;
	size_t getOffset() const;
	size_t getAllocSize() const;
	size_t getDataSize() const;

	void setMessageStart() { m_msgStart = m_offset; }
	void setMessageSize(size_t msgSize) { m_msgSize = msgSize; }
	size_t getMessageStart() const { return m_msgStart; }
	size_t getMessageSize() const { return m_msgSize; }

	uint8_t getInt1();
	uint8_t getInt1(size_t offset);
	uint16_t getInt2();
	uint16_t getInt2(size_t offset);
	uint32_t getInt4();
	uint32_t getInt4(size_t offset);
	uint64_t getInt8();
	uint64_t getInt8(size_t offset);
	uint32_t getBytes(uint8_t *buffer, uint32_t bufferLength);
	uint32_t getBytes(uint8_t *buffer, uint32_t offset, uint32_t bufferLength);
	uint32_t getBytesArray(uint8_t *buffer, uint32_t bufferLength);
    uint32_t getBytesPointer(uint8_t **dataPtr, uint32_t bufferLength);

	void putInt1(uint8_t val);
	void putInt1(uint8_t val, size_t offset);
	void putInt2(uint16_t val);
	void putInt2(uint16_t val, size_t offset);
	size_t putInt2Array(size_t count, uint16_t const* val, size_t offset);
	void putInt4(uint32_t val);
	void putInt4(uint32_t val, size_t offset);
	size_t putInt4Array(size_t itemCount, uint32_t const* val, size_t offset);
	void putInt8(uint64_t val);
	void putInt8(uint64_t val, size_t offset);
	size_t putInt8Array(size_t count, uint64_t const* val, size_t offset);
	void putVectorString(std::vector<std::string> const &args);
	void putBytes(const uint8_t *buffer, uint32_t bufferLength);
	void putBytes(const uint8_t *buffer, uint32_t offset, uint32_t bufferLength);
	void putBytesFrom(RecordFormatByteBuffer &buffer, uint32_t size);
	void insertBytesFrom(RecordFormatByteBuffer &buffer, uint32_t offset = 0);
	void advance(size_t count);
	void compact();
	void shiftLeft(size_t offset, size_t numberOfBytes);

	void rewind();
	void rewind(size_t value);
	void reset();
	void advanceAddress(size_t size);
	void setDataSize(size_t size);
	void setOffset(size_t offset);

	std::string toString(int indent = 0) const;

	void checkGetSize(size_t size);
	void checkGetSize(size_t offset, size_t size);;
	void checkPutSize(size_t size);
	void checkPutSize(size_t offset, size_t size);

protected:
	RecordFormatByteBuffer(const RecordFormatByteBuffer &); // not implemented
	const RecordFormatByteBuffer & operator=(const RecordFormatByteBuffer &); // not implemented

	size_t m_offset;
	size_t m_allocSize;
	size_t m_msgStart;
	size_t m_msgSize;
	size_t m_dataSize;
	bool m_isBufAllocatedOutside;
	uint8_t *m_buffer;

private:
    void throwAssert(const std::string & file, int line, const std::string &message);
};

inline
RecordFormatByteBuffer::RecordFormatByteBuffer(size_t size)
{
    m_offset = 0;
    m_dataSize = 0;
    m_msgStart = 0;
    m_msgSize = 0;
    m_allocSize = size;
    //Force the buffer to 64-bit alignment
    size_t alloc64Bitsize = (m_allocSize / 8);
    alloc64Bitsize += (m_allocSize % 8 == 0) ? 0 : 1;
    RecordFormatAllocatorType::rebind<uint64_t>::other allocator;
    m_buffer = reinterpret_cast<uint8_t*> (allocator.allocate(alloc64Bitsize));
    m_isBufAllocatedOutside = false;
}

inline
RecordFormatByteBuffer::RecordFormatByteBuffer(uint8_t *buffer)
{
    m_offset = 0;
    m_dataSize = 0;
    m_msgStart = 0;
    m_msgSize = 0;
    m_allocSize = 0xFFFF; // FIXME: (LogFormat20) Implement shared memory buffer
    //Force the buffer to 64-bit alignment
    size_t alloc64Bitsize = (m_allocSize / 8);
    alloc64Bitsize += (m_allocSize % 8 == 0) ? 0 : 1;
    RecordFormatAllocatorType::rebind<uint64_t>::other allocator;
    m_buffer = reinterpret_cast<uint8_t*> (allocator.allocate(alloc64Bitsize));
    m_isBufAllocatedOutside = true;
}

// This constructor is used for creating ByteBuffer wrapper of allocated memory buffer
inline
RecordFormatByteBuffer::RecordFormatByteBuffer(uint8_t *buffer, size_t bufferSize)
{
    m_buffer = buffer;
    m_offset = 0;
    // assuming that passed buffer size is unlimited
    m_dataSize = m_allocSize = bufferSize;
    m_msgStart = 0;
    m_msgSize = m_dataSize;
    m_isBufAllocatedOutside = true;
}

inline
void RecordFormatByteBuffer::replaceBuffer(uint8_t *buffer, size_t bufferSize)
{
    if (NZ_UNLIKELY(!m_isBufAllocatedOutside))
    {
        throwAssert(__FILE__, __LINE__, "Cannot replace internally-allocated buffer");
    }

    m_buffer = buffer;
    m_offset = 0;
    // assuming that passed buffer size is unlimited
    m_dataSize = m_allocSize = bufferSize;
    m_msgStart = 0;
    m_msgSize = m_dataSize;
    m_isBufAllocatedOutside = true;
}

inline
uint8_t *RecordFormatByteBuffer::getAddress() const
{
    return m_buffer;
}

inline
size_t RecordFormatByteBuffer::getOffset() const
{
    return m_offset;
}

inline
size_t RecordFormatByteBuffer::getAllocSize() const
{
    return m_allocSize;
}

inline
size_t RecordFormatByteBuffer::getDataSize() const
{
    return m_dataSize;
}

inline
uint8_t RecordFormatByteBuffer::getInt1()
{
    checkGetSize(1);

    uint8_t val;
    ::memcpy(&val, m_buffer + m_offset, 1);
    m_offset += 1;

    return val;
}

inline
uint8_t RecordFormatByteBuffer::getInt1(size_t offset)
{
    checkGetSize(offset, 1);

    uint8_t val;
    ::memcpy(&val, m_buffer + offset, 1);
    return val;
}

inline
uint16_t RecordFormatByteBuffer::getInt2()
{
    checkGetSize(2);

    uint16_t nval;
    ::memcpy(&nval, m_buffer + m_offset, 2);
    m_offset += 2;

    uint16_t val = ntohs(nval);

    return val;
}

inline
uint16_t RecordFormatByteBuffer::getInt2(size_t offset)
{
    checkGetSize(offset, 2);

    uint16_t nval;
    ::memcpy(&nval, m_buffer + offset, 2);
    uint16_t val = ntohs(nval);
    return val;
}

inline
uint32_t RecordFormatByteBuffer::getInt4()
{
    checkGetSize(4); // FIXME 76134: comment this out to achieve 10% more performance

    uint32_t nval;
    ::memcpy(&nval, m_buffer + m_offset, 4);
    m_offset += 4;

    uint32_t val = ntohl(nval);

    return val;
}

inline
uint32_t RecordFormatByteBuffer::getInt4(size_t offset)
{
    checkGetSize(offset, 4);

    uint32_t nval;
    ::memcpy(&nval, m_buffer + offset, 4);
    uint32_t val = ntohl(nval);

    return val;
}

inline
uint64_t RecordFormatByteBuffer::getInt8()
{
    checkGetSize(8);

    uint32_t nval1;
    uint32_t nval2;
    ::memcpy(&nval1, m_buffer + m_offset, 4);
    m_offset += 4;
    ::memcpy(&nval2, m_buffer + m_offset, 4);
    m_offset += 4;

    uint64_t val = ((uint64_t) ntohl(nval1) << 32) | (uint64_t) ntohl(nval2);

    return val;
}


inline
uint64_t RecordFormatByteBuffer::getInt8(size_t offset)
{
    checkGetSize(offset, 8);

    uint32_t nval1;
    uint32_t nval2;
    ::memcpy(&nval1, m_buffer + offset, 4);
    offset += 4;
    ::memcpy(&nval2, m_buffer + offset, 4);
    uint64_t val = ((uint64_t) ntohl(nval1) << 32) | (uint64_t) ntohl(nval2);
    return val;
}

inline
uint32_t RecordFormatByteBuffer::getBytes(uint8_t *buffer, uint32_t bufferLength)
{
    uint32_t length = getInt4();
    if (NZ_UNLIKELY(length > bufferLength))
    {
        throwAssert(__FILE__, __LINE__, "caller's buffer is too small.");
    }

    checkGetSize(length);

    ::memcpy(buffer, m_buffer + m_offset, length);
    m_offset += length;

    return length;
}

inline
uint32_t RecordFormatByteBuffer::getBytes(uint8_t *buffer, uint32_t offset, uint32_t bufferLength)
{
    checkGetSize(offset, bufferLength);

    ::memcpy(buffer, m_buffer + offset, bufferLength);

    return bufferLength;
}

inline
uint32_t RecordFormatByteBuffer::getBytesArray(uint8_t *buffer, uint32_t bufferLength)
{
    checkGetSize(bufferLength);

    ::memcpy(buffer, m_buffer + m_offset, bufferLength);
    m_offset += bufferLength;

    return bufferLength;
}

inline
uint32_t RecordFormatByteBuffer::getBytesPointer(uint8_t **dataPtr, uint32_t bufferLength)
{
    uint32_t length = getInt4();
    if (NZ_UNLIKELY(length > bufferLength))
    {
        throwAssert(__FILE__, __LINE__, "caller's buffer is too small.");
    }

    checkGetSize(length);

    *dataPtr = m_buffer + m_offset;
    m_offset += length;

    return length;
}

inline
void RecordFormatByteBuffer::putInt1(uint8_t val)
{
    checkPutSize(1);

    ::memcpy(m_buffer + m_offset, &val, 1);
    m_offset += 1;
    if (m_offset > m_dataSize)
    {
        m_dataSize = m_offset;
    }
}


inline
void RecordFormatByteBuffer::putInt1(uint8_t val, size_t offset)
{
    checkPutSize(offset, 1);
    ::memcpy(m_buffer + offset, &val, 1);
    offset += 1;

    if(offset > m_dataSize)
    {
        m_dataSize = offset;
    }
}

inline
void RecordFormatByteBuffer::putInt2(uint16_t val)
{
    checkPutSize(2);

    uint16_t nval = htons(val);
    ::memcpy(m_buffer + m_offset, &nval, 2);
    m_offset += 2;
    if (m_offset > m_dataSize)
    {
        m_dataSize = m_offset;
    }
}

inline
void RecordFormatByteBuffer::putInt2(uint16_t val, size_t offset)
{
    checkPutSize(offset,2);
    uint16_t nval = htons(val);
    ::memcpy(m_buffer + offset, &nval, 2);
    offset += 2;

    if(offset > m_dataSize)
    {
        m_dataSize = offset;
    }
}

inline
size_t RecordFormatByteBuffer::putInt2Array(size_t count, uint16_t const* val, size_t offset)
{
    checkPutSize(offset, 2 * count);
    for(int i = 0; i < count; i++)
    {
        uint16_t nval = htons(val[i]);
        ::memcpy(m_buffer + offset, &nval, 2);
        offset += 2;
    }

    if(offset > m_dataSize)
    {
        m_dataSize = offset;
    }

    return offset;
}


inline
void RecordFormatByteBuffer::putInt4(uint32_t val)
{
    checkPutSize(4);

    uint32_t nval = htonl(val);
    ::memcpy(m_buffer + m_offset, &nval, 4);
    m_offset += 4;
    if (m_offset > m_dataSize)
    {
        m_dataSize = m_offset;
    }
}

inline
void RecordFormatByteBuffer::putInt4(uint32_t val, size_t offset)
{
    checkPutSize(offset,4);

    uint32_t nval = htonl(val);
    ::memcpy(m_buffer + offset, &nval, 4);
    offset += 4;

    if(offset > m_dataSize)
    {
        m_dataSize = offset;
    }
}

inline
size_t RecordFormatByteBuffer::putInt4Array(size_t itemCount, uint32_t const* val, size_t offset)
{
    checkPutSize(offset, 4 * itemCount);

    for (int i = 0; i < itemCount; i++)
    {
        uint32_t nval = htonl(val[i]);
        ::memcpy(m_buffer + offset, &nval, 4);
        offset += 4;
    }

    if(offset > m_dataSize)
    {
        m_dataSize = offset;
    }

    return offset;
}

inline
void RecordFormatByteBuffer::putInt8(uint64_t val)
{
    checkPutSize(8);

    uint32_t nval1 = htonl(val >> 32);
    uint32_t nval2 = htonl(val & 0xFFFFFFFF);
    ::memcpy(m_buffer + m_offset, &nval1, 4);
    m_offset += 4;
    ::memcpy(m_buffer + m_offset, &nval2, 4);
    m_offset += 4;
    if (m_offset > m_dataSize)
    {
        m_dataSize = m_offset;
    }
}

inline
void RecordFormatByteBuffer::putInt8(uint64_t val, size_t offset)
{
    checkPutSize(offset,8);

    uint32_t nval1 = htonl(val >> 32);
    uint32_t nval2 = htonl(val & 0xFFFFFFFF);

    ::memcpy(m_buffer + offset, &nval1, 4);
    offset += 4;
    ::memcpy(m_buffer + offset, &nval2, 4);
    offset += 4;

    if(offset > m_dataSize)
    {
        m_dataSize = offset;
    }
}

inline
size_t RecordFormatByteBuffer::putInt8Array(size_t count, uint64_t const* val, size_t offset)
{
    checkPutSize(offset, 8 * count);

    for(int i = 0; i < count; i++)
    {
        uint32_t nval1 = htonl(val[i] >> 32);
        uint32_t nval2 = htonl(val[i] & 0xFFFFFFFF);

        ::memcpy(m_buffer + offset, &nval1, 4);
        offset += 4;
        ::memcpy(m_buffer + offset, &nval2, 4);
        offset += 4;
    }
    if(offset > m_dataSize)
    {
        m_dataSize = offset;
    }

    return offset;
}

inline
void RecordFormatByteBuffer::putVectorString(std::vector<std::string> const &args)
{
    putInt4(args.size());
    for(int i = 0; i < args.size(); i++)
    {
        std::string const& arg = args[i];

        //Exclude the null terminator in the byte array containing the string
        putBytes((uint8_t *) arg.c_str(), arg.length());
    }
}

inline
void RecordFormatByteBuffer::putBytes(const uint8_t *buffer, uint32_t bufferLength)
{
    putInt4(bufferLength);
    checkPutSize(bufferLength);
    ::memcpy(m_buffer + m_offset, buffer, bufferLength);
    m_offset += bufferLength;

    if (m_offset > m_dataSize)
    {
        m_dataSize = m_offset;
    }
}

inline
void RecordFormatByteBuffer::putBytes(const uint8_t *buffer, uint32_t offset, uint32_t bufferLength)
{
    checkPutSize(offset,bufferLength);
    ::memcpy(m_buffer + offset, buffer, bufferLength);

    if(offset + bufferLength > m_dataSize)
    {
        m_dataSize = offset + bufferLength;
    }
}

inline
void RecordFormatByteBuffer::putBytesFrom(RecordFormatByteBuffer &buffer, uint32_t size)
{
    buffer.checkGetSize(size);
    checkPutSize(size);

    ::memcpy(m_buffer + m_offset, buffer.m_buffer + buffer.m_offset, size);
    m_offset += size;
    buffer.m_offset += size;

    if (m_offset > m_dataSize)
    {
        m_dataSize = m_offset;
    }
}

inline
void RecordFormatByteBuffer::insertBytesFrom(RecordFormatByteBuffer &buffer, uint32_t offset /* = 0 */)
{
    size_t size = buffer.getDataSize();
    checkPutSize(size, offset);
    ::memmove((m_buffer + offset + size), (m_buffer + offset), getDataSize());
    ::memcpy ((m_buffer + offset), buffer.m_buffer, size);
    m_offset += size;

    if (m_offset > m_dataSize)
    {
        m_dataSize = m_offset;
    }
}

inline
void RecordFormatByteBuffer::advance(size_t count)
{
    checkGetSize(count);
    m_offset += count;
}

inline
void RecordFormatByteBuffer::advanceAddress(size_t size)
{
    if (NZ_UNLIKELY(!m_isBufAllocatedOutside))
    {
        throwAssert(__FILE__, __LINE__, "advanceAddress is not allowed for internally allocated buffer");
    }

    if (NZ_UNLIKELY(m_allocSize < size))
    {
        throwAssert(__FILE__, __LINE__, "offset < numberOfBytes");
    }

    m_buffer += size;
    m_allocSize -= size;
}

inline
void RecordFormatByteBuffer::compact()
{
    size_t numUnusedBytes = getDataSize() - getOffset();
    if (numUnusedBytes > 0)
    {
        // Move unused memory to the start of the buffer
        ::memmove(getAddress(), getAddress() + getOffset(),
                getDataSize() - getOffset());
    }
    rewind();
    setDataSize(numUnusedBytes);
}


inline
void RecordFormatByteBuffer::shiftLeft(size_t offset, size_t numberOfBytes)
{
    if (NZ_UNLIKELY(offset < numberOfBytes))
    {
        throwAssert(__FILE__, __LINE__, "offset < numberOfBytes");
    }
    ::memmove((m_buffer + (offset - numberOfBytes)), (m_buffer + offset), (m_dataSize-offset));
    // adjust data size
    m_dataSize -= numberOfBytes;
}

inline
void RecordFormatByteBuffer::rewind()
{
    m_msgStart = m_offset = 0;
    m_msgSize = m_dataSize;
}

inline
void RecordFormatByteBuffer::rewind(size_t value)
{
    if (NZ_UNLIKELY(m_offset < value))
    {
        throwAssert(__FILE__, __LINE__, "offset < rewind value");
    }
    m_offset -= value;
}

inline
void RecordFormatByteBuffer::reset()
{
    m_offset = m_msgStart = m_msgSize = m_dataSize = 0;
}

inline
void RecordFormatByteBuffer::setDataSize(size_t size)
{
    //  m_offset must have been initialized correctly before calling
    if (NZ_UNLIKELY(size > m_allocSize))
    {
        throwAssert(__FILE__, __LINE__, "doesn't fit");
    }

    m_msgSize = m_dataSize = size;
}

inline
void RecordFormatByteBuffer::setOffset(size_t offset)
{
    if (NZ_UNLIKELY(offset > m_dataSize))
    {
        throwAssert(__FILE__, __LINE__, "doesn't fit");
    }

    m_offset = offset;
}


inline
void RecordFormatByteBuffer::checkGetSize(size_t size)
{
    if (NZ_UNLIKELY(m_offset + size > m_msgSize))
    {
        throwAssert(__FILE__, __LINE__, "no more data");
    }
}

inline
void RecordFormatByteBuffer::checkGetSize(size_t offset, size_t size)
{
    if (NZ_UNLIKELY(offset + size > m_msgSize))
    {
        throwAssert(__FILE__, __LINE__, "no more data");
    }
}

inline
void RecordFormatByteBuffer::checkPutSize(size_t size)
{
    if (NZ_UNLIKELY((m_offset + size) > m_allocSize))
    {
        throwAssert(__FILE__, __LINE__, "doesn't fit");
    }
}

inline
void RecordFormatByteBuffer::checkPutSize(size_t offset, size_t size)
{
    if (NZ_UNLIKELY((offset + size) > m_allocSize))
    {
        std::ostringstream out;
        out << "cannot put in byte buffer. offset = " << offset << " size = " << size
                << " m_allocSize = " << m_allocSize;
        throwAssert(__FILE__, __LINE__, out.str());
    }
}


} // namespace recordformat

#endif // NZRECORDFORMATBYTEBUFFER_H
