/*
 * (C) Copyright IBM Corp. 2012  All Rights Reserved.
 */

#ifndef _NZTABLELOGRECORD_H_
#define _NZTABLELOGRECORD_H_



/*
 * Log Format 2.0 Record Definition
 * 
 * Record field ids are assigned in-order. Never delete or reorder field 
 * declarations. 
 */
#include <string>
#include <vector>

struct TableShape;

namespace recordformat
{

typedef std::vector<FieldClass const*, RecordFormatAllocatorType> vector_FieldClass_Ptr;

struct TableLogRecordClass: public RecordClass
{
    RECORD_FORMAT_ALLOCATION(TableLogRecordClass);
public:
    //  record definition fixed size = uint16_t(field count ) + uint32_t (record type) + header(4 + 1) (see RecordDefinitionWriter.writeRecord() method)
    static const uint8_t RECORD_DEFINITION_FIXED_SIZE = 2 + 2 + 4 + 4 + 1;
    //  record definition fixed size = uint16_t(field count ) + uint32_t (record type) + header(4 + 4 + 1) (see FieldDefinitionWriter.writeRecord() method)
    static const uint8_t FIELD_DEFINITION_FIXED_SIZE = 1 + 4 + 4 + 2 + 1 + 1 + 1 + 2;

    TableLogRecordClass(RecordRegistry& registry, std::string const &tableName, int32_t tableId);
    void declareFields(TableShape const *nzTableShape);
    void declareNumericTypeField(bool nullable, uint16_t size, std::string const &name);
    void declareFloatTypeField(bool nullable, uint16_t size, std::string const &name);
    void declareIntTypeField(bool nullable, uint16_t size, std::string const &name);
    void serialize(RecordSerializer &serializer) const;
    vector_FieldClass_Ptr m_tableColumns;
    uint32_t m_maxSerializeSize;
};

struct TableLogRecordWriter: public RecordWriter
{
    RECORD_FORMAT_ALLOCATION(TableLogRecordWriter);
public:
	TableLogRecordWriter(TableLogRecordClass const & tableLogRecordClass);

    TableLogRecordClass const& getRecordClass() const
    {
    	return *static_cast<TableLogRecordClass const*>(m_ptrRecordClass);
    }
    void writeRecord(RecordFormatByteBuffer &recordBuffer, std::vector<void const*> const & data);
    void writeRecord(RecordFormatByteBuffer &recordBuffer, void const *data[], int dataSize);
private:
    uint32_t getFieldItemCount(FieldClass const *field, bool isNull);
};

struct TableLogRecordReader
{
    RECORD_FORMAT_ALLOCATION(TableLogRecordReader);
public:
    void bind(RecordObject const& record);
    /* Field commitType class properties */
    const void* getFieldData(int fieldId);
    std::string getLogRecordAsString();

private:
    RecordObject const* m_record;
    /* internal cache of field to member bindings */
    vector_FieldClass_Ptr m_tableColumns;
};

struct TableLogRecordRegistry: public RecordRegistry
{
    RECORD_FORMAT_ALLOCATION(TableLogRecordRegistry);
public:
	TableLogRecordRegistry();
};

struct TableLogRecordClassSerializer : public RecordSerializer
{
    RECORD_FORMAT_ALLOCATION(TableLogRecordClassSerializer);
public:
	static const uint64_t BUFFER_SIZE = 128000;
	TableLogRecordClassSerializer(RecordFormatByteBuffer *serializeBuffer);
	TableLogRecordClassSerializer(uint8_t *serializeBuffer, size_t serializeBufferSize);

	virtual ~TableLogRecordClassSerializer();

	RecordFormatByteBuffer m_logRecBuffer; // buffer for keeping one log record
	RecordFormatByteBuffer *m_serializeBuffer; // buffer for keeping fully serialized log record. Passed by called
	uint32_t m_totalSize;

	/**
	 * Get the buffer to read the metadata conversation
	 */
	virtual RecordFormatByteBuffer &getReadBuffer();

	/**
	 * Get the buffer to write the metadata conversation
	 */
	virtual RecordFormatByteBuffer &getWriteBuffer();

	/**
	 * Callback to notify a record has been placed in the buffer.
	 * Perform any needed action (write serialized record to file or network).
	 */
	virtual void recordWriteComplete();

	/**
	 * Callback to request a complete new record be placed in the buffer for reading
	 *
	 * @return the record typeid of the received record
	 */
	virtual uint32_t receiveCompleteRecord(RecordObject &recordObject);

	uint32_t getSerialziedSize(){
	    return m_totalSize;
	}
};


}

#endif /* _NZLOGRECORD_H_ */
