#ifndef NZAUTOMUTEX_INCLUDED
#define NZAUTOMUTEX_INCLUDED

#ifndef NZTYPES_INCLUDED
# include "nztypes.h"
#endif

/* Derived from Prem's CAutoMutex, but can be built into DBOS/SPU */

/*
   CNzAutoMutex - convenient class to Lock and Unlock mutexes.
   very useful when you have several exit points from your function and you
   want guaranteed mutex unlock on exit from the function

   eg:

   int doSomething(nzmutex_t *mutex)
   {
        //acquire mutex lock
        CNzAutoMutex autoMutex(mutex);

        //do some critical stuff

        if(error ...)
            return -1;  //mutex unlock is automatic

        //do more stuff
        ...

        return 0;       //mutex unlock is automatic
   }

*/

class CNzAutoMutex {
public:
    CNzAutoMutex(nzmutex_t *mutex, const char *type, int id = 0);   // lock and hold reference to mutex
    ~CNzAutoMutex();            //unlocks mutex when object goes out of scope
    void Lock();                //explicitly lock the mutex (after Unlock())
    void Unlock();              //explicity unlock mutex
    bool IsLocked() const;
    nzmutex_t *Mutex() const;

private:
    nzmutex_t *m_mutex;
    bool m_locked;

#if FOR_SPU
    const char *m_type;
    int m_id;
#endif
};

inline bool
CNzAutoMutex::IsLocked() const
{
    return m_locked;
}

inline nzmutex_t *
CNzAutoMutex::Mutex() const
{
    return m_mutex;
}

#endif
