#ifndef NZCRC_INCLUDED
#define NZCRC_INCLUDED
/// @file nzcrc.h
/// @brief CRC computation functions
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2001,2011 All rights reserved.
//////////////////////////////////////////////////////////////////////
#include "nzstd.h"

// Note: NzFastCrcBuf() is an order of magnitude faster than the standard
// NzCrcBuf(), since it uses native crc32 assembler instructions implemented
// on the CPU, rather than a software-only CRC mechanism using lookup tables.
// However, special care needs to be taken when switching from the existing
// CRC algorithm to the new one, in particular, we must be able to handle
// seeing both the old CRC or new CRC in case we do not migrate all on-disk
// CRCs, and we would also need to handle changing all the CRCs prior to
// a customer downgrade.  As well, if used for data pages, the FPGA code
// must be updated with the same CRC algorithm.  And, if the CRC/hash value
// for distribution keys are changed, data migration would also be required.
CEXTERN uint32 NzFastCrcBuf(   // Compute 32-bit CRC value on passed buffer
    register const void *pData,// data to be added to CRC
    register uint32 byteCount, // byte count of pData
    register uint32 crcVal );  // current running value of CRC

CEXTERN uint32 NzCrcBuf(  // Compute 32-bit CRC value on passed buffer
    register const void *pData,// data to be added to CRC
    register uint32 byteCount, // byte count of pData
    register uint32 crcVal );  // current running value of CRC

uint32 NzCrcFloatBuf(  // Compute 32-bit CRC value on passed buffer
    register const void *pData,     // data to be added to CRC
    register uint32 byteCount, // byte count of pData
    register uint32 crcVal );  // current running value of CRC

inline uint32
NzCrc32(const void *buf, size_t len, uint32 initial)
{
    return NzCrcBuf(buf, len, initial);
}

CEXTERN uint32 NzCrcCharBuf(  // Compute 32-bit CRC value on passed buffer
    register const void *pData,// data to be added to CRC
    register uint32 byteCount, // byte count of pData
    register uint32 crcVal,    // current running value of CRC
    uint8 space);              //character used for space comparison


uint32 NzCrcIntBuf(  // Compute 32-bit CRC value on passed buffer
    register const void *pData,// data to be added to CRC
    register uint32 byteCount, // byte count of pData
    register uint32 crcVal );  // current running value of CRC

uint32 NzCrcNumericBuf(  // Compute 32-bit CRC value on passed buffer
    register const void *pData,// data to be added to CRC
    register uint32 byteCount, // byte count of pData
    register uint32 crcVal );  // current running value of CRC

#if defined(__i386__) || defined(__x86_64__)  // little-endian
uint32 NzCrcIntBufBe(  // NzCrcIntBuf for a big-endian record (from spu)
    register const void *pData,// data to be added to CRC
    register uint32 byteCount, // byte count of pData
    register uint32 crcVal );  // current running value of CRC

uint32 NzCrcNumericBufBe(  // NzCrcNumericBuf for a big-endian rec (from spu)
    register const void *pData,// data to be added to CRC
    register uint32 byteCount, // byte count of pData
    register uint32 crcVal );  // current running value of CRC
#endif // little-endian

#endif
