#ifndef NZGLOBALS_INCLUDED
#define NZGLOBALS_INCLUDED

//========================================================================
///\file: nzglobals.h
//
// Common shared memory structure and a few often-used process globals.
//
// (C) Copyright IBM Corp. 2009, 2011  All Rights Reserved.
//========================================================================

#include "nzlimits.h"
#include "nzstd.h"
#include "nztypes.h"
#include "rcAtomic.h"

#if !FOR_CLI
# include "locktracker.h"
#endif

#if WIN32
typedef int pid_t;
#endif


struct NzState;
struct NzConfig;
class CDictionary;
class PlanDictionary;
class CNzSeqDictionary;
class CNzLibDictionary;
class CSpuSwapMgr;

class CRecordMgr;
class CSwapSpace;
class CPlan;
class CTable;
class CPlanCache;
class CRegen;
class CRegenTarget;
class CMutex;
class CFComm;
struct HashTableParams_t;
struct client_comms_t;
struct event_comms_t;
struct spool_acct_t;
class CNetEtherPacketLogger;
class CWorkSched;
class CFCommShared;
struct dispatch_queue_t;
class CMlsNameTable;
class PlanFilesQueue;
struct PlanArchiver;
class SchedRuleDefs;
class CSrcMgr; // Snippet Result Cache Manager
class TblIdMapper;
class RedrHostRedunInfo;
struct PageCacheAcct;
struct TempCompressOptions;
struct PGQueue;

namespace rc {
  class DbgCtrls;
  class Semaphore;
}
namespace hc {
  class SpuMgr;
}
namespace spu {
  class Queue;
}
namespace sc {
  class CSessionStore;
}
namespace alc {
   class ShData;
}
namespace repl {
   class ShData;
}
namespace cryptoc {
    class CHostKey;
}

namespace lf {
    class HazardManager;
}

#if !FOR_CLI

// mutex names
LKTRKR_DECLARE_NAME(MTX_SYS);
LKTRKR_DECLARE_NAME(MTX_TXMGR);
LKTRKR_DECLARE_NAME(MTX_QUTEX_RES);
LKTRKR_DECLARE_NAME(MTX_DATASLICETBL);

#define NZ_COMMON_MEM_MAGIC 0x421417ed



struct common_memory_t {

    unsigned commonMemInited; // set to NZ_COMMON_MEM_MAGIC when Nz inited

    // General
    int maxClients;  // = startup.maxConnections + NZ_INTERNAL_CLIENTS

    nzmutex_t txMgrMutex;  // concurrent access to TxJournal, xid, ...
    nzmutex_t sysMutex;    // general-purpose short-duration mutex
    nzmutex_t aggrTableMutex; // short message aggregation table.

    bool localSPU; // spu running locally on the same box as the host
    bool useFpgaPrep;   // use fpga setup generator
    bool realFpga;      // use "real" fpga (not just as disk controller)
    bool fpgaDump;      // do "diagnostic data dumping" in fpgawrap.cpp
    bool autoDumpandgo; // Enable system initiated dumpandgo (NZ_CHECK_DUMP)
    PlanDictionary *hostPlanDictionary;  // mapper of plan id to CPlan*
    int fpgaDumpCount;  // used in creating diagnostic dump files
    int fpgaFeatures;   // flags indicating current fpga features available
    int numSPUs;        // number of primary dataslices across all spus
    CDictionary *dictionary;
    CNzSeqDictionary *hostSeqDictionary;
    CNzLibDictionary *libDictionary;
    TblIdMapper *tblIdMapper;
    RedrHostRedunInfo *redrHostRedunInfoPtr;
    PageCacheAcct    *pageCacheAcct;           // parameters for CRecordStore
                                               // file page cache management
    TempCompressOptions *tempCompressOptions;  // parameters for CRecordStore
                                               // transient compression in file
    atomic_t tempTableCounter;  // create ids for class CTable job transients
    atomic_t host_rsid_cnt;     // create ids for class CRecordStore on host
    int16 abortOnError;
    int16 btOnError;
    bool  noCrash;    //true when cliqa -abort arg1 is called with -noCrash

    nzmutex_t quTexMutex; // Mutex guarding queue pool and mutex pool, etc

    // Queue pool
    int *availQueues; // list of allocated queues currently unused
    int sizeAvailQueues; // number of members in availQueues
    int availQueueDex;  // current LIFO pos in availQueues

    // Mutex pool
    int *availMutexes; // list of allocated mutexes currently unused
    int sizeAvailMutexes; // number of members in availMutexes
    int availMutexDex;  // current LIFO pos in availMutexes

    // comps mem layouts
    void       *rcMemLayout;        // Pointer to recomps memory structure
    void       *acMemLayout;        // Pointer to appcomps memory structure
    void       *scMemLayout;        // Pointer to syscomps memory structure
    void       *opstMemLayout;      // Pointer to opstats memory structure

    // system state
    NzState *nzstate;
    nzmutex_t dataslicetblMutex; // Mutex guarding nzstate->dataSliceTbl

    // system configuration settings (registry)
    NzConfig *syscfg;

    //springfield
    int maxSpringFieldSize; //env var controlled max size of spring field

    int streamBatchSize; // Return Size Size limit in bytes
    int maxOutstandingClientResults; // Number of In-flight return sets before spooling
    int mergeMaxWaitingBlocks; //Limit of Blocks per SPU before XOFF from CmergePlan

    // Temporary Development flags to ease integration/testing
    bool m_flagAllowDiskHashJoin;  // NZ_DISK_HASH_JOIN
    // Things to manipulate hash tables from the registry.
    HashTableParams_t *m_hashTableParams;

    // Debug
    bool stableDump;  // for autotesting
    CTable **nodeInfoVirtualTables;   // table of vt_plan
    int logResultsCnt;  // counter for results table logging

    bool alwaysUsePGEval;  // for QA
    bool forceFullNodeStats; // whether to get stats even if no plan reference

    // Cliqa access
    int maxSortBlocks;  // how many recordstore blocks we can materialize in sort
    bool udxStatsEnabled; // collection and use of UDX/AE child stats
    bool udxStatsDetailEnabled; // reporting of detailed UDX/AE child stats in plan files

    // Host Specific
    pid_t dispatcherProcess;
    dispatch_queue_t *dispatchQueue;

    rc::DbgCtrls *dbosDbgCtrls;

    int nextQueueKey;       // next SysV msg queue key

    atomic_t clientIdCounter;   // counter for assigning client id's
    atomic_t clientUidCounter;  // unique id
    client_comms_t *clients;    // client array

    PGQueue *pgq;               // posgres-side client queueing

    sc::CSessionStore *sessionStore;   // store of session context objects
    CMlsNameTable *mlsNameTable;
    atomic_t mlsTempTableId; // monotonically increasing temp table number

    spool_acct_t *spool_acct;   // shared spool accounting info

    event_comms_t *sysev;       // destination for internal, "system" events
                                // not associated with any client

    CPlanCache *planCache;      // keeps recent generated code & plans (for debug)

    PlanArchiver *archiver; // encompasses objects related to plan files archiving:
                            // shared memory queue and archive mutex

    SchedRuleDefs *schedRules;
    // host interaction table
    void *interactiontbl;
    atomic_t interactionIdCounter;
    // statically allocated pool of interactions
    void *interactionPool;

    void *msgPool; // pre-allocated pool of plan abort messages and dispatch-to-event

    CWorkSched *workSched;      // jobs of work to be scheduled

    void *joblist;    // list of active jobs

    CSrcMgr * srcMgr; // Snippet Result Cache Manager

    // SPU specific

    CRecordMgr *recordmgr;
    pid_t spuDispatchThread;

    hc::SpuMgr *spumgr; // the object that interafaces with the system mgr

    //Acid specific
    //  the following hold the pointers to various
    //  transaction related managers (singleton pattern)
    //  include-file ordering problems make it easier
    //  to define these pointers as void*
    //  the various ACID include files cast appropriately
    void* p_theExtReclamationMgr; // really ExtentReclamationMgr*
    void* p_theJournalMgr;  // really CJournalMgr*
    void* p_theHostTXMgr;   // really CHostTXMgr*
    void* p_theTXidMgr;     // really CTXidMgr*
    void* p_rollbackEstimates;   // really TxDsRollbackEstimates*

    // switches to control unsafe mirror error handling
    bool abortOnErrorAfterCommit; // default true

    // Regen specific
    CRegen *regenPrimaryP;      // source object for regen of some primary partition
    CRegen *regenMirrorP;       // source object for regen of some mirror partition
    CRegenTarget *regenTargetP; // regen-target object

    // Distribution Key Support
    uint16 *distKeyMap;  // Shared distribution key map (hash value->spu id)
    uint16 *redrDistKeyMap;  // Shared distribution key map (hash value->spu id)
    uint16 numRedrDslices; // count of redr dslices 

    // Final size of snippet on SPU, given prior failovers, choice of
    // spuDistBucketSize and other lifetime parameters.
    int snippetBufferCap;
    // !NOTE-dk "spu DownloadTableNode" should get msg pool dynamically
    // in snippet, not from a "system lifetime" allocation.
    uint snippetCapForDHJ; //snippet blocks threshold for DHJ
    uint snippetCapForWindowAggr; //snippet blocks limit for Window Aggregation

    uint snippetCapForDHJOverride; //cliqa override snippetCapForDHJ (currently not used)
    uint snippetCapForWindowAggrOverride; //cliqa override snippetCapForWindowAggr
    uint snippetCapVirtualPages;  // event proc data's upper bound from virtual addressing
    
    // Packet Logger
    CNetEtherPacketLogger *packetLogger; // packet logger object
    int packetLoggerLevel; // 0=off, 1=TxRx, 2=Full

    // COSInteraction logging parameter
    int m_osiLogDepth;     // last (m_osiLogDepth) no-more-to-come get logged

    // network_config
    void *network_config;

    CFComm *fComm;
    CFCommShared *fCommShared;

    // memory usage for data reader.
    int CheckpointMegsInUse;           // Total Megabytes of memory used for checkpoints
    nzmutex_t CheckpointLock;      // synchronizes access to CheckpointMegsInUse
    void *CheckpointQueue;             // queue of loads waiting for memory to run

    // the shared state for alcapp and alloder
    alc::ShData *alcShdata;

    // shared config/state info for Postgres, replcapture, replapplymgr
    repl::ShData *replShdata;

    // Shared Host Key -- AES 256 Bit key
    cryptoc::CHostKey *theHostKey;

    // Hazard Pointer Manager
    lf::HazardManager *hazardManager;

    // useful for runtime instrumentation, for example, with sched
    int reserved[56];

    // Moved to end to minimize number of large instruction data offsets relative to Nz
    char kitDir[STRPATH+1];     // just for development version
    char dataDir[STRPATH+1];    // just for development version
    char planDir[STRPATH+1];    // where we log plans and gen'd code to
    char planHistDir[STRPATH+1];// plan history directory
    char transDir[STRPATH+1];   // root of tree storing temporary files for SPU/Host
    char padfileDir[STRPATH+1]; // root of directories for file-backed SPUPAD
};

/* ------------- Globals --------------- */

extern common_memory_t *Nz;

#endif // !FOR_CLI

extern int NzMyProcessId;
extern const char *NzMyProcessName;
extern nzotuid_t NzMyOwnerId;

CEXTERN int NetezzaNDE;     // defined in postgres
CEXTERN bool NetezzaNoPad;  // defined in postgres
CEXTERN int FastUpgradeEnabled;

/* ------------- Prototypes ------------- */

extern void NzInitCommonMemT();

// this is called after startupsvr has read startup.maxConnections:
extern void NzGrowCommonMemT(unsigned numClients);

#endif // NZGLOBALS_INCLUDED
