#ifndef NZNUMBERS_INCLUDED
#define NZNUMBERS_INCLUDED

#ifndef STDIO_INCLUDED
# include <stdio.h>
# define STDIO_INCLUDED
#endif

#ifndef LIMITS_INCLUDED
# include <limits.h>
# define LIMITS_INCLUDED
#endif

#if !defined(LLONG_MAX) && !defined(LLONG_MIN)
/* GCC may have these @ limits.h [with dialect dependencies]; DIAB does not */
#define LLONG_MAX (0x7FFFFFFFFFFFFFFFLL)
#define LLONG_MIN (-LLONG_MAX - 1)
#endif

extern "C" int throwError(int code, const char *buf, const char *obj);


// This is a double float, with a "low part" for extra precision during aggregation operations.
// This extra precision should remove the nondeterminacy that would otherwise be observed
// when summing doubleprecision float columns over multiple SPU's.
struct CSuperDouble {
    double highpart;
    double lowpart;

    // Constructors/converters.
    //
    // Declaring the up-converters "explicit" makes the system "superdouble-averse".  When in doubt
    // about a plain double vs. a superdouble, choose the plain double.  Sometimes generated code
    // comes up with
    //  <double-var> = <boolean> ? <double> : <superdouble>;
    // To do the conversion across the colon, the compiler needs to know which way to do the convert.
    // We want it to do a downward convert of the superdouble to a plain double.
    // In general, we only want superdoubles to exist as the actual target of summations.
    CSuperDouble() { highpart = lowpart = 0.0; }
    explicit CSuperDouble(int32 val) { highpart = (double) val; lowpart = 0.0; }
    explicit CSuperDouble(int64 val) { highpart = (double) val; lowpart = 0.0; }
    explicit CSuperDouble(float val) { highpart = (double) val; lowpart = 0.0; }
    explicit CSuperDouble(double val) { highpart = val; lowpart = 0.0; }
    CSuperDouble(const CSuperDouble & val) { highpart = val.highpart; lowpart = val.lowpart; }

    // Assignment.
    CSuperDouble const& operator = (int32 val) { highpart = (double) val; lowpart = 0.0; return *this; };
    CSuperDouble const& operator = (int64 val) { highpart = (double) val; lowpart = 0.0; return *this; };
    CSuperDouble const& operator = (float val) { highpart = (double) val; lowpart = 0.0; return *this; };
    CSuperDouble const& operator = (double val) { highpart = (double) val; lowpart = 0.0; return *this; };
    CSuperDouble const& operator = (CSuperDouble const& rhs) { highpart = rhs.highpart; lowpart = rhs.lowpart; return *this; };

    // Converter out.
    operator double () const
    { return highpart+lowpart; }

    // This is the super-precise add from double.  The summing in the SPUs does this.
    CSuperDouble operator + (double const rhs) const;
    CSuperDouble const& operator += (double const rhs);

    // This is the super-precise add from another CSuperDouble.
    // The summing of the SPU results in the host does this.
    CSuperDouble operator + (CSuperDouble const & rhs) const;
    CSuperDouble const& operator += (CSuperDouble const & rhs);

    // Comparisons.
    bool operator == (CSuperDouble const & rhs) const;
    bool operator > (CSuperDouble const & rhs) const;
    bool operator < (CSuperDouble const & rhs) const;

    bool operator != (CSuperDouble const & rhs) const { return !(*this == rhs); }
    bool operator >= (CSuperDouble const & rhs) const { return !(*this < rhs); }
    bool operator <= (CSuperDouble const & rhs) const { return !(*this > rhs); }
    // An efficient strcmp-like function.
    static int compareLikeStrcmp(CSuperDouble const & lhs, CSuperDouble const & rhs);
};


inline CSuperDouble CSuperDouble::operator + (double const rhs) const
{
    CSuperDouble ttt(*this);
    double t1 = highpart;
    ttt.highpart += rhs;
    double t2 = ttt.highpart - t1;
    ttt.lowpart += rhs-t2;
    return ttt;
}

inline CSuperDouble const& CSuperDouble::operator += (double const rhs)
{
    double t1 = highpart;
    highpart += rhs;
    double t2 = highpart - t1;
    lowpart += rhs-t2;
    return *this;
}

inline CSuperDouble CSuperDouble::operator + (CSuperDouble const & rhs) const
{
    CSuperDouble ttt(*this);
    double t1 = highpart;
    ttt.highpart += rhs.highpart;
    double t2 = ttt.highpart - t1;
    ttt.lowpart += rhs.highpart-t2;
    ttt.lowpart += rhs.lowpart;
    return ttt;
}

inline CSuperDouble const& CSuperDouble::operator += (CSuperDouble const & rhs)
{
    double t1 = highpart;
    highpart += rhs.highpart;
    double t2 = highpart - t1;
    lowpart += rhs.highpart-t2;
    lowpart += rhs.lowpart;
    return *this;
}

inline bool CSuperDouble::operator == (CSuperDouble const & rhs) const
{
    // Do a quick purification of the high part to be sure we don't miss something.
    double xl = highpart+lowpart;
    double xr = rhs.highpart+rhs.lowpart;
    if (xl != xr) return false;
    // They seem to be the same; get the "pure" low parts.
    double reml = lowpart + (highpart-xl);
    double remr = rhs.lowpart + (rhs.highpart-xr);
    return reml == remr;
}

inline bool CSuperDouble::operator > (CSuperDouble const & rhs) const
{
    // Do a quick purification of the high part to be sure we don't miss something.
    double xl = highpart+lowpart;
    double xr = rhs.highpart+rhs.lowpart;
    if (xl > xr) return true;
    if (xl < xr) return false;
    // They seem to be the same; get the "pure" low parts.
    double reml = lowpart + (highpart-xl);
    double remr = rhs.lowpart + (rhs.highpart-xr);
    return reml > remr;
}

inline bool CSuperDouble::operator < (CSuperDouble const & rhs) const
{
    // Do a quick purification of the high part to be sure we don't miss something.
    double xl = highpart+lowpart;
    double xr = rhs.highpart+rhs.lowpart;
    if (xl < xr) return true;
    if (xl > xr) return false;
    // They seem to be the same; get the "pure" low parts.
    double reml = lowpart + (highpart-xl);
    double remr = rhs.lowpart + (rhs.highpart-xr);
    return reml < remr;
}

// This is used in recordcompare.cpp; we want it to be efficient.
// It has the usual -1/0/+1 semantics.
inline int CSuperDouble::compareLikeStrcmp(CSuperDouble const & lhs, CSuperDouble const & rhs)
{
    // Do a quick purification of the high part to be sure we don't miss something.
    double xl = lhs.highpart+lhs.lowpart;
    double xr = rhs.highpart+rhs.lowpart;
    if (xl > xr) return 1;
    if (xl < xr) return -1;
    // They seem to be the same; get the "pure" low parts.
    double reml = lhs.lowpart + (lhs.highpart-xl);
    double remr = rhs.lowpart + (rhs.highpart-xr);
    if (reml > remr) return 1;
    if (reml < remr) return -1;
    return 0;
}


inline float const& CheckFloat(float const& value)
{
    uint32 static const EXP  = 0xff << 23;
    uint32 static const FRAC = (1 << 23) -1;

    if ((*(uint32*)&value & EXP) == EXP) {
        // A saturated exponent is either a NaN or an infinity
        if (*(uint32*)&value & FRAC)
            THROW_ERROR(NZ_ERROR_MATH, "Float4 result is a NaN", 0);
    }
    return value;
}

inline double const& CheckDouble(double const& value)
{
    uint64 static const EXP  = 0x7ffULL << 52;
    uint64 static const FRAC = (1ULL << 52) -1;
    if ((*(uint64*)&value & EXP) == EXP) {
        // A saturated exponent is either a NaN or an infinity
        if (*(uint64*)&value & FRAC)
            THROW_ERROR(NZ_ERROR_MATH, "Float8 result is a NaN", 0);
    }
    return value;
}

inline CSuperDouble CheckSuperDouble(CSuperDouble const& value)
{
    CheckDouble(value.highpart);
    return value;
}

inline double const& CheckForRangeErrorInt8(double const& value)
{
    //Checks whether the value is within a range
    //of 'int1' (-128 to 127).

    if (value < SCHAR_MIN || value > SCHAR_MAX) {
        char buffer[50];
        sprintf (buffer, "'%le' out of range", value);
        throwError (NZ_ERROR_OVERFLOW, buffer, "CheckForRangeErrorInt8");
    }

    return value;
}

inline double const& CheckForRangeErrorInt16(double const& value)
{
    //Checks whether the value is within a range
    //of 'int2' (-32768 to 32767).

    if (value < SHRT_MIN || value > SHRT_MAX) {
        char buffer[50];
        sprintf (buffer, "'%le' out of range", value);
        throwError (NZ_ERROR_OVERFLOW, buffer, "CheckForRangeErrorInt16");
    }

    return value;
}

inline double const& CheckForRangeErrorInt32(double const& value)
{
    //Checks whether the value is within a range
    //of 'int4' (-2147483648 to 2147483647).

    if (value < INT_MIN  || value > INT_MAX) {
        char buffer[50];
        sprintf (buffer, "'%le' out of range", value);
        throwError (NZ_ERROR_OVERFLOW, buffer, "CheckForRangeErrorInt32");
    }

    return value;
}

inline double const& CheckForRangeErrorInt64(double const& value)
{
    //Checks whether the value is within a range
    //of 'int8' (-9223372036854775808 to 9223372036854775807).

    if (value < LLONG_MIN || value > LLONG_MAX) {
        char buffer[50];
        sprintf (buffer, "'%le' out of range", value);
        throwError (NZ_ERROR_OVERFLOW, buffer, "CheckForRangeErrorInt64");
    }

    return value;
}

#endif
