#ifndef NZPTHREAD_INCLUDED
#define NZPTHREAD_INCLUDED

#ifdef __cplusplus
# include "error.h"
# include "locktracker.h"
extern "C" {
#else // for C / Postgres code
# include <pthread.h>
# include <errno.h>
# define NZ_VERIFY(b)    Assert((b))
# define LKTRKR_SETWAITING(a, b)  {}
# define LKTRKR_CLRWAITING(a, b)  {}
# define LKTRKR_ADDLOCK(a, b)     {}
# define LKTRKR_REMOVELOCK(a, b)  {}
#endif

#include <pthread.h>

#define _NZPTRETURN0(funcall) do { \
    int const ret = funcall; \
    NZ_VERIFY(ret == 0); \
} while (false)


/*  We must use PTHREAD_PROCESS_SHARED mutexes and condition variables.  For
    convenience we initialize static pthread mutex and condition attributes
    to be used for mutex and condition variable initialization.
 */
extern pthread_once_t _nzpthread_init_once;
extern void _nzpthread_init();
extern pthread_mutexattr_t _nzpthread_process_shared_mutexattr;
extern pthread_rwlockattr_t _nzpthread_rwlockattr_process_shared;
extern pthread_rwlockattr_t _nzpthread_rwlockattr_process_shared_prefer_writer;
extern pthread_condattr_t _nzpthread_process_shared_condattr;
inline static pthread_mutexattr_t *
nzpthread_process_shared_mutexattr()
{
    _NZPTRETURN0(pthread_once(&_nzpthread_init_once, &_nzpthread_init));
    return &_nzpthread_process_shared_mutexattr;
}
inline static pthread_rwlockattr_t *
nzpthread_rwlockattr_process_shared()
{
    _NZPTRETURN0(pthread_once(&_nzpthread_init_once, &_nzpthread_init));
    return &_nzpthread_rwlockattr_process_shared;
}
inline static pthread_rwlockattr_t *
nzpthread_rwlockattr_process_shared_prefer_writer()
{
    _NZPTRETURN0(pthread_once(&_nzpthread_init_once, &_nzpthread_init));
    return &_nzpthread_rwlockattr_process_shared_prefer_writer;
}
inline static pthread_condattr_t *
nzpthread_process_shared_condattr()
{
    _NZPTRETURN0(pthread_once(&_nzpthread_init_once, &_nzpthread_init));
    return &_nzpthread_process_shared_condattr;
}
#define NZ_PTHREAD_SHARED_MUTEX (nzpthread_process_shared_mutexattr())
#define NZ_PTHREAD_RWLOCK_SHARED (nzpthread_rwlockattr_process_shared())
#define NZ_PTHREAD_RWLOCK_SHARED_PREFER_WRITER (nzpthread_rwlockattr_process_shared_prefer_writer())
#define NZ_PTHREAD_SHARED_COND (nzpthread_process_shared_condattr())

inline static void
nzpthread_cond_init(pthread_cond_t *cond, pthread_condattr_t *cond_attr)
{
#if NZDEBUG
    if (cond_attr) {
        // Right now we do not use threads, so all condition variables
        // should be process-shared
        int is_process_shared;
        _NZPTRETURN0(pthread_condattr_getpshared(
                         cond_attr,
                         &is_process_shared));
        NZ_VERIFY(is_process_shared);
    }
#endif

    _NZPTRETURN0(pthread_cond_init(cond, cond_attr));
}

inline static void
nzpthread_cond_signal(pthread_cond_t *cond)
{
    _NZPTRETURN0(pthread_cond_signal(cond));
}

inline static void
nzpthread_cond_broadcast(pthread_cond_t *cond)
{
    _NZPTRETURN0(pthread_cond_broadcast(cond));
}

inline static void
nzpthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{
    _NZPTRETURN0(pthread_cond_wait(cond, mutex));
}

inline static void
nzpthread_cond_wait_trk(pthread_cond_t *cond, pthread_mutex_t *mutex, const char *name, int id)
{
    LKTRKR_SETWAITING(name, id);
    nzpthread_cond_wait(cond, mutex);
    LKTRKR_CLRWAITING(name, id);
}

inline static bool
nzpthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex, const struct timespec *abstime)
{
    int const ret = pthread_cond_timedwait(cond, mutex, abstime);
    switch (ret)
    {
    case 0: return true;
    case ETIMEDOUT: return false;
    default: NZ_VERIFY(false);
    }
}

inline static bool
nzpthread_cond_timedwait_trk(pthread_cond_t *cond, pthread_mutex_t *mutex, const struct timespec *abstime,
                             const char *name, int id)
{
    LKTRKR_SETWAITING(name, id);
    bool const ret = nzpthread_cond_timedwait(cond, mutex, abstime);
    LKTRKR_CLRWAITING(name, id);
    return ret;
}

inline static void
nzpthread_cond_destroy(pthread_cond_t *cond)
{
    _NZPTRETURN0(pthread_cond_destroy(cond));
}

inline static void
nzpthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *mutexattr)
{
#if NZDEBUG
    if (mutexattr) {
        // Right now we do not use threads, so all mutexes
        // should be process-shared
        int is_process_shared;
        _NZPTRETURN0(pthread_mutexattr_getpshared(
                         mutexattr,
                         &is_process_shared));
        NZ_VERIFY(is_process_shared);
    }
#endif
    _NZPTRETURN0(pthread_mutex_init(mutex, mutexattr));
}

inline static void
nzpthread_mutex_lock(pthread_mutex_t *mutex)
{
    _NZPTRETURN0(pthread_mutex_lock(mutex));
}

inline static void
nzpthread_mutex_lock_trk(pthread_mutex_t *mutex, const char *name, int id)
{
    LKTRKR_SETWAITING(name, id);
    nzpthread_mutex_lock(mutex);
    LKTRKR_ADDLOCK(name, id);
}

inline static bool
nzpthread_mutex_trylock(pthread_mutex_t *mutex)
{
    int const ret = pthread_mutex_trylock(mutex);
    switch (ret)
    {
    case 0:
        return true;
    case EBUSY:
    case EDEADLK:
        return false;
    default: NZ_VERIFY(false);
    }
}

inline static void
nzpthread_mutex_unlock(pthread_mutex_t *mutex)
{
    _NZPTRETURN0(pthread_mutex_unlock(mutex));
}

inline static void
nzpthread_mutex_unlock_trk(pthread_mutex_t *mutex, const char *name, int id)
{
    LKTRKR_REMOVELOCK(name, id);
    nzpthread_mutex_unlock(mutex);
}

inline static void
nzpthread_mutex_destroy(pthread_mutex_t *mutex)
{
    _NZPTRETURN0(pthread_mutex_destroy(mutex));
}

inline static void
nzpthread_rwlock_init(pthread_rwlock_t *l, const pthread_rwlockattr_t *__restrict__ attr)
{
#if NZDEBUG
    if (attr) {
        // Right now we do not use threads, so all rwlocks should be
        // process-shared
        int is_process_shared;
        _NZPTRETURN0(pthread_rwlockattr_getpshared(
                         attr,
                         &is_process_shared));
        NZ_VERIFY(is_process_shared);
    }
#endif

    _NZPTRETURN0(pthread_rwlock_init(l, attr));
}

inline static void
nzpthread_rwlock_destroy(pthread_rwlock_t *l)
{
    _NZPTRETURN0(pthread_rwlock_destroy(l));
}

inline static void
nzpthread_rwlock_rdlock(pthread_rwlock_t *l)
{
    _NZPTRETURN0(pthread_rwlock_rdlock(l));
}

inline static void
nzpthread_rwlock_wrlock(pthread_rwlock_t *l)
{
    _NZPTRETURN0(pthread_rwlock_wrlock(l));
}

inline static void
nzpthread_rwlock_unlock(pthread_rwlock_t *l)
{
    _NZPTRETURN0(pthread_rwlock_unlock(l));
}

inline static void
nzpthread_rwlock_rdlock_trk(pthread_rwlock_t *l, const char *name, int id)
{
    LKTRKR_SETWAITING(name, id);
    nzpthread_rwlock_rdlock(l);
    LKTRKR_ADDLOCK(name, id);
}

inline static void
nzpthread_rwlock_wrlock_trk(pthread_rwlock_t *l, const char *name, int id)
{
    LKTRKR_SETWAITING(name, id);
    nzpthread_rwlock_wrlock(l);
    LKTRKR_ADDLOCK(name, id);
}

inline static void
nzpthread_rwlock_unlock_trk(pthread_rwlock_t *l, const char *name, int id)
{
    LKTRKR_REMOVELOCK(name, id);
    nzpthread_rwlock_unlock(l);
}

inline static void
nzpthread_rwlockattr_init(pthread_rwlockattr_t *a)
{
    _NZPTRETURN0(pthread_rwlockattr_init(a));
}

inline static void
nzpthread_rwlockattr_setpshared(pthread_rwlockattr_t *a, int ps)
{
    _NZPTRETURN0(pthread_rwlockattr_setpshared(a, ps));
}

inline static void
nzpthread_rwlockattr_destroy(pthread_rwlockattr_t *a)
{
    _NZPTRETURN0(pthread_rwlockattr_destroy(a));
}

inline static void
nzpthread_create(pthread_t *thread, pthread_attr_t *attr,
                 void * (*start_routine)(void *), void *arg)
{
    _NZPTRETURN0(pthread_create(thread, attr, start_routine, arg));
}

inline static void
nzpthread_join(pthread_t thread, void **retval)
{
    _NZPTRETURN0(pthread_join(thread, retval));
}

inline static void
nzpthread_setspecific(pthread_key_t key, const void *value)
{
    _NZPTRETURN0(pthread_setspecific(key, value));
}

inline static void
nzpthread_key_create(pthread_key_t *key, void (*destructor)(void *))
{
    _NZPTRETURN0(pthread_key_create(key, destructor));
}

inline static void
nzpthread_once(pthread_once_t *once_control, void (*init_routine) (void) )
{
    _NZPTRETURN0(pthread_once(once_control, init_routine));
}

inline static void
nzpthread_mutexattr_init(pthread_mutexattr_t *attr)
{
    _NZPTRETURN0(pthread_mutexattr_init(attr));
}

inline static void
nzpthread_mutexattr_settype(pthread_mutexattr_t *attr, int kind)
{
    _NZPTRETURN0(pthread_mutexattr_settype(attr, kind));
}

inline static void
nzpthread_mutexattr_setpshared(pthread_mutexattr_t *attr,
                               int pshared)
{
    _NZPTRETURN0(pthread_mutexattr_setpshared(attr, pshared));
}

#ifdef  __cplusplus
} // extern "C"
#endif

#endif
