/*
 *  nzrep.h
 *  Netezza Remote Execution Protocol
 *  main client include file for nzrep system
 * (C) Copyright IBM Corp. 2008, 2011  All Rights Reserved.
 */

#ifndef __NZREP_H__
#define  __NZREP_H__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include "nzrepcommands.h"

/* size of error message buffer */
#define NZREP_ERROR_MESSAGE_LENGTH 1024

/* maximum data size that can be sent is four hundred thousand bytes */
#define NZREP_MAXIMUM_BUFFER_SIZE 50000
#define NZREP_MAXIMUM_DATA_SIZE 2*NZREP_MAXIMUM_BUFFER_SIZE

/*
    struct NzrepInitialization

    Used to call function nzrepCreateConnection in the parent process
    and nzrepInitializeChild in the child process.

    Variables handle and errorMessage are always out parameters.

    The handle is opaque and holds the state of the connection.
    It is used in subsequent nzrep calls.

    Variable parameters is an output parameter for the parent and
    an input parameter for the child.  It could be passed to a
    forked only child via memory and to an exec'd child using
    command line parameter(s) or environment variables(s)
*/


typedef struct NZREP_HANDLE_S {} * NZREP_HANDLE;

typedef struct NzrepInitialization
{
    NZREP_HANDLE handle;
    char errorMessage[NZREP_ERROR_MESSAGE_LENGTH];
    int parameters[2];
    int nonBlocking;
    char *tempFileName;
    const char *tempDir;
    int childMode;
} NzrepInitialization;

/*
    struct NzrepData

    Holds the command and data payload parameters.

    This is an in parameter to function nzrepSend and
    an out parameter from nzrepReceive.

    Length may be zero if only a command is being sent
    or received. Similarly data may be NULL.

    For nzrepSend, data should point to the data you are sending
    and length should be that data's length.

    After a nzrepReceive the data is owned
    by the handle, do not free it !!!
*/
typedef struct NzrepData
{
    int32_t command;
    char * data;
    int32_t length;
} NzrepData;

/*
    Create new connection, call in parent before forking.
    Specify 1 for nonBlocking, 0 for blocking.
    Returns 0 on success, -1 if error
*/
extern int nzrepCreateConnection(
    NzrepInitialization * args);

/*
    Initialize connection, call in parent after forking
    Returns 0 on success, -1 if error
    The in handle is from nzrepCreateConnection
*/
extern int nzrepIntializeParent(
    NZREP_HANDLE in_handle);

/*
    Initialize connection, call in child
    Specify 1 for nonBlocking, 0 for blocking.
    Returns 0 on success, -1 if error
    Clears the handle error state.
*/
extern int nzrepInitializeChild(
    NzrepInitialization * args);

/* Cancel a hung nzrep connection.
   This connection will be created via
   nzrepInitializeChild generally and be
   outside the normal parent -> child
   nzrep relationship.
 */
extern void nzrepCancel(NZREP_HANDLE in_handle);

/* Cancel a hung nzrep connection.
   This will take care of calling 
   nzrepInitializeChild and nzrepClose.
   This is intended to be called by a third
   party process not part of the normal
   parent -> child nzrep relationship
*/
extern void nzrepAbort(const char* strFile);


/*
    Send a command with optional data
    Returns 0 on success, -1 if error
    Clears the handle error state.
    Overwrites the data received from the last nzreReceive call (if any).
*/
extern int nzrepSend(
    NZREP_HANDLE in_handle,
    const NzrepData * in_data,
    int in_timeout_milliseconds);

/*
    Receive a command and associated data (if any)
    Returns 0 on success, -1 if error
    Clears the handle error state.
    The memory holding the data returned in out_data
    belongs to the handle and should not be freed my the caller.
    This memory will be overwritten by the next nzrepReceive
    or nzrepSend call.
*/
extern int nzrepReceive(
    NZREP_HANDLE in_handle,
    NzrepData * out_data,
    int in_timeout_milliseconds);

/*
  Tell nzrep that either the parent or child
  has exited so pending sends or receives should
  be interrupted
*/
extern void nzrepSetExit(NZREP_HANDLE in_handle);

/*
    Returns a true value if the an error has occured,
    false value if it did not.
*/
extern int nzrepIsError(
    NZREP_HANDLE in_handle);

/*
    Explicitly clears the handle error condition.
    The handle error state also includes the last error message text,
    the last errno, and the last timed-out state.
*/
extern void nzrepClearError(
    NZREP_HANDLE in_handle);

/*
    Returns a true value if the last opertion timed out,
    false value if it did not
*/
extern int nzrepIsTimeout(
    NZREP_HANDLE in_handle);

/*
    Return the last error message encountered by nzrep
*/
extern char * nzrepGetLastErrorText(
    NZREP_HANDLE in_handle);

/*
    Return the last errno value encountered by nzrep
    An errno is only produced by OS system calls
*/
extern int nzrepGetLastErrno(
    NZREP_HANDLE in_handle);

/*
    Close the the nzrep handle
*/
extern void nzrepClose(
    NZREP_HANDLE in_handle);

/*
    Resize the internal handle buffer to the default and clear all
    error conditions.  Useful if a very large data item was sent or received
    and you want to release that memory.
    !!! Invalidates the last NzrepData returned from nzrepReceive !!!
    This function is optional.
    Returns 0 on success, -1 if error
*/
extern int nzrepReset(
    NZREP_HANDLE in_handle);

/* version for testing parent-child compatibility */
extern int nzrepGetVersion();


#ifdef __cplusplus
} /* closing brace for extern "C" */
#endif

#endif
