/*
 *  nzsds.h
 *  Netezza System Data serialization
 *  main client include file for nzsds utility
 * (C) Copyright IBM Corp. 2008, 2011  All Rights Reserved.
 */

#ifndef __NZSDS_H__
#define __NZSDS_H__


#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stdbool.h>
#include "nzuds.h"

/*
    struct NzsdsDataType
    datatype definitions
*/
typedef enum NzsdsDataType
{
    NZSDS_INT8 = 0,
    NZSDS_INT16,
    NZSDS_INT32,
    NZSDS_INT64,
    NZSDS_FLOAT,
    NZSDS_DOUBLE,
    NZSDS_STRING,
    NZSDS_UDS,
    NZSDS_MAX
} NzsdsDataType;

/*
    struct NzsdsData
    field data to serialize / deserialize
*/
typedef struct NzsdsData
{
    union
    {
        const int8_t * pInt8;
        const int16_t * pInt16;
        const int32_t * pInt32;
        const int64_t * pInt64;
        const float * pFloat;
        const double * pDouble;
        const char * pString;
        const char * pUds;
    } data;
    NzsdsDataType type;
    int32_t numElements;
    bool isArray;
} NzsdsData;


/*
    serialization of a NZSDS data stream
*/
typedef struct NzsdsBuffer
{
    const char * buffer;
    int32_t length;
} NzsdsBuffer;


typedef struct NZSDS_HANDLE_S {} * NZSDS_HANDLE;



/*
    create and return a new NZSDS handle
    if arg is NULL return an empty handle otherwise
    the handle is initialized with the content of arg.
    arg must be the output of nzsdsGetBufferPointer.
    A copy of arg will be created and owned by the handle.
*/
NZSDS_HANDLE nzsdsCreate(const NzsdsBuffer * arg);

/*
   Reinitialize handle with contents of arg.
   returns 0 is success, -1 otherwise.
*/
int nzsdsSetData(NZSDS_HANDLE handle, const NzsdsBuffer * arg);

/*
   Reinitialize handle with contents of arg. (but don't
   create copy of arg. The contents of arg must
   continue to be valid until done with the handle.
   returns 0 is success, -1 otherwise.
*/
int nzsdsSetDataNoCopy(NZSDS_HANDLE handle, const NzsdsBuffer * arg);



/*
    serialize a field into the handle at the end
    returns 0 if sucess, -1 if memory allocation error
*/
int nzsdsWriteInt8(NZSDS_HANDLE handle, int8_t data);
int nzsdsWriteInt8Array(NZSDS_HANDLE handle, const int8_t * data, int numElements);
int nzsdsWriteInt16(NZSDS_HANDLE handle, int16_t data);
int nzsdsWriteInt16Array(NZSDS_HANDLE handle, const int16_t * data, int numElements);
int nzsdsWriteInt32(NZSDS_HANDLE handle, int32_t data);
int nzsdsWriteInt32Array(NZSDS_HANDLE handle, const int32_t * data, int numElements);
int nzsdsWriteInt64(NZSDS_HANDLE handle, int64_t data);
int nzsdsWriteInt64Array(NZSDS_HANDLE handle, const int64_t * data, int numElements);
int nzsdsWriteFloat(NZSDS_HANDLE handle, float data);
int nzsdsWriteFloatArray(NZSDS_HANDLE handle, const float * data, int numElements);
int nzsdsWriteDouble(NZSDS_HANDLE handle, double data);
int nzsdsWriteDoubleArray(NZSDS_HANDLE handle, const double * data, int numElements);

/*
 * String must be null terminated and numElements must count the null
 * terminator
*/
int nzsdsWriteString(NZSDS_HANDLE handle, const char * data, int numElements);

/*
 * UDS will be written in the UDS compressed format
*/
int nzsdsWriteUds(NZSDS_HANDLE handle, NZUDS_HANDLE udsHandle);

/*
    serialize an NZSDS handle into a NzsdsBuffer
    The handle owns this data

    The uncompressed version compresses only if more than
    NZSER_BLOCKSIZE can be saved.

    The compressed version removes *all* free space.
    This is generally not useful if the packet
    is going to be sent as an nzrepSend payload.
*/
int nzsdsGetBufferPointer(NZSDS_HANDLE handle, NzsdsBuffer * arg);
int nzsdsGetBufferPointerCompressed(NZSDS_HANDLE handle, NzsdsBuffer * arg);

/*
    get next serialized field
    return a true value if more data, a false value at end of data
*/
int nzsdsReadNext(NZSDS_HANDLE handle, NzsdsData * data);

/*
    returns number of serialized fields
*/
int nzsdsGetNumElements(NZSDS_HANDLE handle);

/*
   Reset read element position for nzsdsReadNext
   return 0 on succexs, -1 out of range
*/
int nzsdsSetReadElementPosition(NZSDS_HANDLE handle, int arg);

/*
   remove elements with position >= nulElements
   if read position >= arg read position set to arg (end of data)
*/
int nzsdsShrinkTo(NZSDS_HANDLE handle, int numElements);

/*
   release handle
*/
void nzsdsClose(NZSDS_HANDLE handle);


#ifdef __cplusplus
} /* closing brace for extern "C" */
#endif

#endif
