//-----------------------------------------------------------------------------
// nzstate.h
//
// System state enumeration plus global system state and data slice address
// information.
//
// (C) Copyright IBM Corp. 2002, 2011  All Rights Reserved.

#ifndef NZSTATE_INCLUDED
#define NZSTATE_INCLUDED

#ifndef NZSTD_INCLUDED
# include "nzstd.h"
#endif

#ifndef NZLIMITS_INCLUDED
# include "nzlimits.h"
#endif

#ifndef RCATOMIC_INCLUDED
# include "rcAtomic.h"
#endif

#if !FOR_SPU
# include <fstream>
#endif

namespace rc {
  class TxtTbl;
  class ByteStrm;
}

#define INVALID_SPA_IDX ((uchar) -1)

//-----------------------------------------------------------------------------
// NzSysState is an enum of system state values.

enum NzSysState {
    SS_NO_STATE = 0,

    // The order of these states matter.  The code currently assumes that
    // the system can legally transisiton to any state with a larger value
    // with a few exceptions
    //
    // Any changes made to this must be reflected in the following files:
    // main/src/webclient/admin/js/spulist.js
    // main/src/nde/nz/nzstate.h

    SS_DISCOVERING              = 1,
    SS_DISCOVERED               = 2,
    SS_INITIALIZING             = 3,
    SS_INITIALIZED              = 4,
    SS_PREONLINING              = 5,
    SS_PREONLINE                = 6,
    SS_RESUMING                 = 7,
    SS_ONLINE                   = 8,
    SS_RESTRICTED_RESUMING      = 9,
    SS_RESTRICTED_ONLINE        = 10,
    SS_MAINTAINING              = 11,
    SS_MAINTAIN                 = 12,
    SS_SYNCING                  = 13,
    SS_SYNCED                   = 14,
    SS_SYNCING_NOW              = 15,
    SS_SYNCED_NOW               = 16,
    SS_RECOVER                  = 17,
    SS_RECOVERING               = 18,
    SS_FAILING_BACK             = 19,
    SS_FAILED_BACK              = 20,
    SS_PAUSING                  = 21,
    SS_PAUSED                   = 22,
    SS_PAUSING_NOW              = 23,
    SS_PAUSED_NOW               = 24,
    SS_OFFLINING                = 25,
    SS_OFFLINE                  = 26,
    SS_OFFLINING_NOW            = 27,
    SS_OFFLINE_NOW              = 28,
    SS_STOPPING                 = 29,
    SS_STOPPED                  = 30,
    SS_STOPPING_NOW             = 31,
    SS_STOPPED_NOW              = 32,

    // SS_STARTING is used only by waitForState, and only when the system
    // appears to be coming up (i.e., startupsvr is running), but there is
    // no shared memory with an initialized systemState value.  That is,
    // systemState can never contain this value.
    SS_STARTING                 = 99,

    SS_DOWN                     = 100,
    SS_UNREACHABLE              = 101,
    SS_BAD_STATE                = 102
};

enum {
    SS_MAX_DATASLICES = MAX_DATASLICES
};

extern "C" {
    const char* nzSysStateSymbol (int sysStateId);
    const char* nzSysStateLabel (int sysStateId);
    NzSysState  nzTransitionState (int expectedState);
    NzSysState  nzExpectedState (int transitionState);
    NzSysState  nzSymbolToSysState (const char* symbol, bool expectedOnly);
    bool        nzIsSameBaseSysState (int sysState1, int sysState2);
    bool        nzIsNowSysState (int sysState);
    bool        nzIsOnlineSysState (int sysState);
    bool        nzIsRecoverySysState (int sysState);
    bool        nzIsQuiescingSysState (int sysState);
    bool        nzIsQuiescedSysState (int sysState);
    bool        nzIsPausedOrOfflineSysState(int sysState);
    bool        nzIsRunnableSysState (int sysState);
    bool        nzIsQueueableSysState (int sysState);
    bool        nzIsDownOrGoingSysState (int sysState);
    bool        nzIsStopOrGoingSysState (int sysState);
    bool        nzIsPriorToInitializing (int sysState);    
    bool        nzIsPriorToInitialized (int sysState);    
    bool        nzIsPriorToOnlineSysState (int sysState);    
    bool        nzIsPaus_ (int sysState);
}

#if __cplusplus
void nzListSysStates (rc::TxtTbl* txtTbl, bool waitForStatesOnly = false);
#endif

//-----------------------------------------------------------------------------
// NzDataSliceAddrs is a structure of data slice IP address/partition pairs
// for the primary, mirror, and regeneration partitions.

struct NzDataSliceAddrs {
    uint    primaryIpAddr;
    uint    mirrorIpAddr;
    uint    regenIpAddr;

    uint    regenId;

    uchar   primaryPartitionId;
    uchar   mirrorPartitionId;
    uchar   regenPartitionId;

    uchar   pad1;

#if __cplusplus
    void    saveOn (rc::ByteStrm &s);
    void    restoreFrom (rc::ByteStrm &s);
#endif
};

//-----------------------------------------------------------------------------
// NzDataPartitionInfo is a structure which defines the state of a partition

struct NzDataPartitionInfo {
    uint    myIp;
    uint    dataPartitionId;  // 0-based data partition index
    uint    dataSliceId;
    uint    regenId;

    uint    isPrimary;
    uint    isRegenSrc;
    uint    isRegenTgt;

    uint    lastRegenToBlockAcked;
    uint    totalRegenToBlocks;
    uint    regenToAlmostDone;
    uint    regenToDone;
    uint    regenToNodeId;
    uint    lastRegenToBlockSent;
    uint    regenFromNodeId;


    // stats

#if __cplusplus
    void    clear ();
#endif
};

//-----------------------------------------------------------------------------
// NzSpuLoc is a structure which defines the location of a Spu and the amount
// of usable memory.

struct NzSpuLoc {
    uint    ipAddr;
    uchar   slot;
    uchar   spaId;

#if __cplusplus
    void    saveOn (rc::ByteStrm &s) const;
    void    restoreFrom (rc::ByteStrm &s);
#endif
};


//-----------------------------------------------------------------------------
// NzState is a structure of globals system state variables. It generally lives
// in NzShm->state.

struct NzState {
    NzSysState          systemState;            // Global, current system state
    int                 systemStateVersion;     // Global version of system state
    NzSysState          expectedState;          // State to transition to
    int                 expectedStateVersion;   // Version of expected state
    char                reasonForDown[256];     //Reason why the system is down

    int                 numJobsRestarting;      // number of jobs being restarted
                                                // (valid on host only)

    int                 failoverCount;          // On Mustang, number of Spus running in
                                                // failover
    int                 regenCount;             // On Mustang, number of Spus performing
                                                // regen
    int                 newSystem;              // 1 = yes 
    int                 expandSystem;           // 1 = yes (allows sysmgr data slice expansion)
    int                 vacuumCount;            // number of vacuum jobs
                                                // outstanding
    int                 blockReclaimsForSync;   // 1 => block new reclaims to allow sync
    int                 topologyVersion;        // incremented whenever topology changes
    int                 devMapVersion;          // incremented whenever devmap changes
    int                 numDataSlices;          // Data slices in the system
    NzDataSliceAddrs   *dataSliceTbl;           // Addresses for each data slice
                                                //  indexed by (dataSliceId-1).
    int                 numSpus;                // ALL Spus seen by sysmgr
    int                 numSfis;                // ALL Sfi seen by sysmgr

    unsigned            numSpas;                // number of SPA's
    uchar              *spaId;                  // array SPA id's
    uchar               spaIdx[256];            // inverse map of SPA id's
                                                // spaIdx[spaId] = spa index, or INVALID_SPA_IDX
    atomic_t            numCommRx;              // number of comm rx processes

    // SPU only:
    unsigned int        numDataPartitions;      // number of data
                                                // partitions per SPU
    NzDataPartitionInfo **dataPartition;        // Info about each
                                                // data partition

    int                spuLocSize;              // size of spuLocTbl
    NzSpuLoc           *spuLocTbl;              // mapping between ip
                                                // and spu/slot number

    int                sysShutdown;            // 1 =>  the system is going down and won't get up.
    long               sysUptimeFirst;          // NPS uptime at startup
    atomic_t           sysReachedOnline;        // Did system reach the online state   

#ifdef __cplusplus
    static NzState*     instance ();            // ShMem instance ()

    static bool         systemIsNotRunnable();
    static bool         systemIsFrontChannelUp();
    static bool         systemIsDownOrGoing();
    static bool         systemIsPaus_();

    void                initNzState (NzSysState initialState, int nDataSlices);
    NzDataSliceAddrs*   dataSliceAddrs (uint dataSliceId);
    bool                isFailedOverDSlice(int id) const;

    // These "set" routines are for the system manager and startup server.

    void                setSysState (int sysState);
    void                setSysState (int sysState, int expectedState,
                                     int expectedStateVersion);
    void                setExpectedState (int expectedState,
                                          int expectedStateVersion);
    void                setReasonForDown (const char* reason);

    void dump();

    // These are for SPU only
    void                updatePartition (NzDataPartitionInfo* info);
    void                updatePartitions ();

    // SPU marshalling support
    void            saveOn (rc::ByteStrm& s, bool includeDetails) const;
    void            restoreFrom (rc::ByteStrm& s);

#endif
};

#ifdef __cplusplus

inline NzDataSliceAddrs* NzState::dataSliceAddrs (uint dataSliceId)
{
    if (!dataSliceTbl)
        return NULL;

    return &dataSliceTbl[dataSliceId-1];
}

#endif

#endif /* NZSTATE_INCLUDED */
