// nzstd.h
//
// (C) Copyright IBM Corporation 2001, 2017.  All rights reserved.
//
// C/C++ extension definitions that are product independent. This header
// file is included in almost all other include files, so any additions to
// it must be applicable to a very wide variety of situations.
//
// Specific 'nz' includes files are:
//
//  nzstd.h         - basic C/C++ defs that are product independent
//  nzlimits.h      - system-wide resource limits, array sizes, etc.
//  nzrecord.h      - basic record and field definitions
//  nzcomm.h        - high-level communication types and messages
//  nzglobals.h     - common memory structure and a few often-used globals
//
// NOTE TO DEVELOPERS:
//
// Please be careful about what you add to this file. Changes to this file
// impact almost every other source file in the system.
//

#ifndef NZSTD_INCLUDED
#define NZSTD_INCLUDED

// For C++ we need these defined in order to get everything
#if defined(__cplusplus)
# ifndef __STDC_CONSTANT_MACROS
#  define __STDC_CONSTANT_MACROS
# endif
# ifndef __STDC_LIMIT_MACROS
#  define __STDC_LIMIT_MACROS
# endif
# ifndef __STDC_FORMAT_MACROS
#  define __STDC_FORMAT_MACROS
# endif
#endif

#include "nzasm.h"

#include <sys/types.h>
#include <string.h>                     // for STREQ*/STRNEQ* macros
#include <stdint.h>

#include <inttypes.h>

// Standard defines and macros

#ifndef TRUE
# define TRUE   1
# define FALSE  0
#endif

#define PTRS_SAME_SIZE_AS_INTS TRUE // will change soon
#define INTS_ARE_FOUR_BYTES TRUE    // probably true for next 10 years

#ifndef NULL
# define NULL   0
#endif

#ifndef EOS
# define EOS    '\0'
#endif

#ifndef NL
# if WIN32
#  define NL "\r\n"
# else
#  define NL "\n"
# endif
#endif

#ifndef MIN
# define MIN(a,b)   (((a)<(b))?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b)   (((a)>(b))?(a):(b))
#endif

#ifndef STREQ
# define STREQ(a,b) (!strcmp(a,b))
#endif

#ifndef STREQNC    // compare string with case insensitive

#ifdef WIN32
# define STREQNC(a,b)    (!_stricmp(a,b))    // NT - compare using lowercase
# define STRNEQNC(a,b,n) (!_strnicmp(a,b,n)) // compare using lc with size
#else
# define STREQNC(a,b)    (!strcasecmp(a,b))
# define STRNEQNC(a,b,n) (!strncasecmp(a,b,n))
#endif

#endif


#ifndef T_OR_F
// also defined in execdebug.h
# define T_OR_F(b)				((b) ? "true" : "false")
#endif
#define T_OR_F_UC(b)				((b) ? "TRUE" : "FALSE")

#define BRKCASE     break; case

#define streq(a,b)  (!strcmp(a,b))       // deprecated (not upper case)
#define brkcase     break; case

#if defined(__cplusplus)
# define CEXTERN extern "C"
# define CBEGIN  CEXTERN {
# define CEND    }
#else
# define CEXTERN
# define CBEGIN
# define CEND
#endif


/**
 * String splicer.  Allows for #define values to be "spliced" into strings.
 * Examples:
 * ------------------------------------------------
 * #define SOME_VALUE 400
 * const char* someString = "Value is " STRING_SPLICE(SOME_VALUE) ". I rock!";
 * someString -> "Value is 400. I rock!"
 * ------------------------------------------------
 * enum foo { bar = 1 };
 * const char* someOtherString = STRING_SPLICE(bar);
 * someOtherString -> "bar"
 * ------------------------------------------------
 *
 * Does not resolve the value of enums/variables - #defines only. Otherwise
 * - will splice the name of the variable/enum - not the value.
 *
 * Note: the HELPER is NOT redundant. It's necessary.
 */
#define STRING_SPLICE_HELPER(a) # a
#define STRING_SPLICE(a) STRING_SPLICE_HELPER(a)


//
// NOTE!  These types should be avoided and we should start using
//        the ISO C standard types (see <stdint.h>)
//

CBEGIN

typedef unsigned int    uint;
typedef unsigned char   uchar;
typedef uint8_t         byte;

// Ideally we'd use stdint.h types here, but these types need to be
// identical to the typedefs in pg/include/c.h
typedef signed char     int8;
typedef unsigned char   uint8;
typedef signed short    int16;
typedef unsigned short  uint16;
typedef signed int      int32;
typedef unsigned int    uint32;

// For some reason, int64/uint64 are based on standard types, not POD types
typedef int64_t         int64;
typedef uint64_t        uint64;

// to avoid bool size issues between GCC intel and ppc targets
typedef uint8_t         nzbool;

// Integer types large enough to contain pointers
typedef uintptr_t       uint_ptr;
typedef intptr_t        int_ptr;

// This doesn't belong here but where else to put it?
// Somewhere in /./main/compat/w32 maybe?
#if defined(WIN32) || defined(NT_X86)
typedef int pid_t;
#endif

CEND

// Note: FOR_POSTGRES is defined when building postgres sources

#if !FOR_POSTGRES
# ifndef __cplusplus
   typedef int bool;
# endif
#endif

// Bit constants

#define BIT0    0x0001
#define BIT1    0x0002
#define BIT2    0x0004
#define BIT3    0x0008
#define BIT4    0x0010
#define BIT5    0x0020
#define BIT6    0x0040
#define BIT7    0x0080
#define BIT8    0x0100
#define BIT9    0x0200
#define BIT10   0x0400
#define BIT11   0x0800
#define BIT12   0x1000
#define BIT13   0x2000
#define BIT14   0x4000
#define BIT15   0x8000

#define BIT16   0x00010000
#define BIT17   0x00020000
#define BIT18   0x00040000
#define BIT19   0x00080000
#define BIT20   0x00100000
#define BIT21   0x00200000
#define BIT22   0x00400000
#define BIT23   0x00800000
#define BIT24   0x01000000
#define BIT25   0x02000000
#define BIT26   0x04000000
#define BIT27   0x08000000
#define BIT28   0x10000000
#define BIT29   0x20000000
#define BIT30   0x40000000
#define BIT31   0x80000000

#define BIT32   INT64_C(0x0000000100000000)
#define BIT33   INT64_C(0x0000000200000000)
#define BIT34   INT64_C(0x0000000400000000)
#define BIT35   INT64_C(0x0000000800000000)
#define BIT36   INT64_C(0x0000001000000000)
#define BIT37   INT64_C(0x0000002000000000)
#define BIT38   INT64_C(0x0000004000000000)
#define BIT39   INT64_C(0x0000008000000000)
#define BIT40   INT64_C(0x0000010000000000)
#define BIT41   INT64_C(0x0000020000000000)
#define BIT42   INT64_C(0x0000040000000000)

#if defined(__GNUC__)
# if __GNUC__ == 2 && __GNUC_MINOR__ < 96
#  define NZ_LIKELY(cond)       (cond)
#  define NZ_UNLIKELY(cond)     (cond)
# else
#  define NZ_LIKELY(cond)       __builtin_expect(!!(cond),1)
#  define NZ_UNLIKELY(cond)     __builtin_expect(!!(cond),0)
# endif
#elif defined(_MSC_VER)
# define NZ_LIKELY(cond)        (cond)
# define NZ_UNLIKELY(cond)      (cond)
# ifndef __attribute__
#  define __attribute__(x)
# endif
#else
# error "unsuported compiler"
#endif

// Alignment control

#define ALIGN_RECORDS ALIGN_VAL64

// NOTE: ALIGN_VAL64 and ALIGN_INT128 macro definitions depend on
//       whether we care about alignment on boundaries beyond 32-bit.
//       At present we don't, so ALIGN_64 and ALIGN_128 are not defined.
//       However, we still want to treat 128-bit, 64-bit, and 32-bit
//       entities as logically distinct, and order them as such in records.
//       For the moment, 96-bit entities will be "ALIGNed_128".
//#define ALIGN_64 1

#define ALIGN_SHORT(s)    ((((s)+1)   >> 1) << 1)
#define ALIGN_INT32(i)    ((((i)+3)   >> 2) << 2)
#if ALIGN_64
# define ALIGN_VAL64(l)   ((((l)+7)   >> 3) << 3)
# if ALIGN_128
#  define ALIGN_INT128(x) ((((x)+0xF) >> 4) << 4)
# else
#  define ALIGN_INT128(x) ((((x)+7)   >> 3) << 3)
# endif
#else
# define ALIGN_VAL64(l)   ((((l)+3)   >> 2) << 2)
# define ALIGN_INT128(x)  ((((x)+3)   >> 2) << 2)
#endif

//#define ALIGN_SHORT(s) ((((s)+1) >> 1) << 1)
//#define ALIGN_INT32(i) ((((i)+3) >> 2) << 2)
//#define ALIGN_VAL64(l) ((((l)+7) >> 3) << 3)

// A few 'very global' build controls

#if !BUST_PROTECTION
# define PROTECTED protected         // !FIX-FR-dw - can these be deleted?
# define PRIVATE private
#else
# define PROTECTED public
# define PRIVATE public
#endif

#endif // NZSTD_INCLUDED
