/*
 *  nzuds.h
 *  Netezza User Data Serialization
 *  main client C language include file for nzuds utility
 * (C) Copyright IBM Corp. 2008, 2011  All Rights Reserved.
 */

#ifndef __NZUDS_H__
#define __NZUDS_H__


#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include "nzudsudxtypes.h"
#include "nzudsdata.h"

/*
    Serialization of a NZUDS data stream
    The memory that buffer points to is owned by the handle / object
    and should not be freed by the user.
*/
typedef struct NzudsBuffer
{
    const char * buffer;
    int32_t length;
} NzudsBuffer;

/*
    Used to validate data types when deserializing a UDS buffer.
    The user owns the memory pointed to by fieldTypes.
*/
typedef struct NzudsUserTypes
{
    int length;
    NzudsDataType * fieldTypes;
} NzudsUserTypes;


typedef struct NZUDS_HANDLE_S {} * NZUDS_HANDLE;

/*
    create and return a new NZUDS handle
    if arg is NULL return an empty handle otherwise
    the handle is initialized with the content of arg.
    arg must be the output of nzudsGetBufferPointer.
    A copy of arg will be created and owned by the handle.

    If arg is NULL then dataTypes must must be null.
    Argument dataTypes is used to verify the number and types
    of the NzudsBuffer argument.
*/
NZUDS_HANDLE nzudsCreate(const NzudsBuffer * arg, const NzudsUserTypes * dataTypes);



/*
   Reinitialize handle with contents of arg.
   returns 0 is success, -1 otherwise.
*/
int nzudsSetData(NZUDS_HANDLE handle, const NzudsBuffer * arg);

/*
   Reinitialize handle with contents of arg. (but don't
   create copy of arg. The contents of arg must
   continue to be valid until done with the handle.
   returns 0 is success, -1 otherwise.
*/
int nzudsSetDataNoCopy(NZUDS_HANDLE handle, const NzudsBuffer * arg);

/*
  Fill in dataType for specified index from buffer. Return -1 on error, 0 on success.
*/
int nzudsGetType(NZUDS_HANDLE handle, int idx, NzudsDataType * dataType);

/*
  Fill in dataTypes from buffer. Return -1 on error, 0 on success.
  The caller must free the fieldTypes member which we be allocated
  by this function.
*/
int nzudsGetTypes(NZUDS_HANDLE handle, NzudsUserTypes * dataTypes);
/*
    serialize a field into the handle at the end
    returns 0 if sucess, -1 if memory allocation error
*/
int nzudsWriteInt8(NZUDS_HANDLE handle, int8_t data);
int nzudsWriteInt8Null(NZUDS_HANDLE handle);
int nzudsWriteInt16(NZUDS_HANDLE handle, int16_t data);
int nzudsWriteInt16Null(NZUDS_HANDLE handle);
int nzudsWriteInt32(NZUDS_HANDLE handle, int32_t data);
int nzudsWriteInt32Null(NZUDS_HANDLE handle);
int nzudsWriteInt64(NZUDS_HANDLE handle, int64_t data);
int nzudsWriteInt64Null(NZUDS_HANDLE handle);

int nzudsWriteFloat(NZUDS_HANDLE handle, float data);
int nzudsWriteFloatNull(NZUDS_HANDLE handle);
int nzudsWriteDouble(NZUDS_HANDLE handle, double data);
int nzudsWriteDoubleNull(NZUDS_HANDLE handle);
int nzudsWriteBool(NZUDS_HANDLE handle, int8_t data);
int nzudsWriteBoolNull(NZUDS_HANDLE handle);

int nzudsWriteDate(NZUDS_HANDLE handle, int32_t data);
int nzudsWriteDateNull(NZUDS_HANDLE handle);
int nzudsWriteTime(NZUDS_HANDLE handle, int64_t data);
int nzudsWriteTimeNull(NZUDS_HANDLE handle);
int nzudsWriteTimeStamp(NZUDS_HANDLE handle, int64_t data);
int nzudsWriteTimeStampNull(NZUDS_HANDLE handle);

int nzudsWriteTimeTz(NZUDS_HANDLE handle, const NzudsTimeTz * data);
int nzudsWriteTimeTzNull(NZUDS_HANDLE handle);
int nzudsWriteInterval(NZUDS_HANDLE handle, const NzudsInterval * data);
int nzudsWriteIntervalNull(NZUDS_HANDLE handle);

int nzudsWriteNumeric32(NZUDS_HANDLE handle, const NzudsNumeric32 * data);
int nzudsWriteNumeric32Null(NZUDS_HANDLE handle);
int nzudsWriteNumeric64(NZUDS_HANDLE handle, const NzudsNumeric64 * data);
int nzudsWriteNumeric64Null(NZUDS_HANDLE handle);
int nzudsWriteNumeric128(NZUDS_HANDLE handle, const NzudsNumeric128 * data);
int nzudsWriteNumeric128Null(NZUDS_HANDLE handle);


int nzudsWriteFixedString(NZUDS_HANDLE handle, const char * data, int length);
int nzudsWriteFixedStringNull(NZUDS_HANDLE handle);
int nzudsWriteVariableString(NZUDS_HANDLE handle, const char * data, int length);
int nzudsWriteVariableStringNull(NZUDS_HANDLE handle);
int nzudsWriteNationalFixedString(NZUDS_HANDLE handle, const char * data, int length);
int nzudsWriteNationalFixedStringNull(NZUDS_HANDLE handle);
int nzudsWriteNationalVariableString(NZUDS_HANDLE handle, const char * data, int length);
int nzudsWriteNationalVariableStringNull(NZUDS_HANDLE handle);
int nzudsWriteGeometryString(NZUDS_HANDLE handle, const char * data, int length);
int nzudsWriteGeometryStringNull(NZUDS_HANDLE handle);
int nzudsWriteVarbinaryString(NZUDS_HANDLE handle, const char * data, int length);
int nzudsWriteVarbinaryStringNull(NZUDS_HANDLE handle);

/*
    serialize an NZUDS handle into a NzudsBuffer
    the handle owns this data

    The uncompressed version compresses only if more than
    NZSER_BLOCKSIZE can be saved.

    The compressed version removes *all* free space.
    This is generally not useful if the packet
    is going to be sent as an nzrepSend payload.
*/
int nzudsGetBufferPointer(NZUDS_HANDLE handle, NzudsBuffer * arg);
int nzudsGetBufferPointerCompressed(NZUDS_HANDLE handle, NzudsBuffer * arg);

/*
    get next serialized field
    return a true value if more data, a false value at end of data
*/
int nzudsReadNext(NZUDS_HANDLE handle, NzudsData * data);

/*
    write this variant serialized field
    returns 0 if sucess, -1 if memory allocation error
*/
int nzudsWrite(NZUDS_HANDLE handle, NzudsData * data);

/*
    returns number of serialized fields
*/
int nzudsGetNumElements(NZUDS_HANDLE handle);

/*
   Reset read element position for nzudsReadNext
   return 0 on success, -1 out of range
*/
int nzudsSetReadElementPosition(NZUDS_HANDLE handle, int arg);

/*
  Return read element position for nzudsReadNext
*/
int nzudsGetReadElementPosition(NZUDS_HANDLE handle);

/*
   remove elements with position >= nulElements
   if read position >= arg read position set to arg (end of data)
*/
int nzudsShrinkTo(NZUDS_HANDLE handle, int numElements);

/*
   release handle
*/
void nzudsClose(NZUDS_HANDLE handle);


#ifdef __cplusplus
} /* closing brace for extern "C" */
#endif

#endif
