/*
 *  nzudscpp.h
 *  Netezza User Data Serialization
 *  main client C++ language include file for nzuds utility
 * (C) Copyright IBM Corp. 2008, 2011  All Rights Reserved.
 */

#ifndef __NZUDSCPP_H__
#define __NZUDSCPP_H__

extern "C" {
#include "nzuds.h"
}

#include "udx.h"


/*
    Serialization of a NZUDS data stream.
    The memory that buffer points to is owned by the NzudsCppData object
    and should not be freed by the user.
*/
struct NzudsCppBuffer
{
    const char * buffer;
    int32_t length;

    void fillInBuffer(NzudsBuffer *pBuffer) const;
    void fromBuffer(const NzudsBuffer *pBuffer);
};

/*
    Used to validate data types when deserializing a UDS buffer.
    The user owns the memory pointed to by fieldTypes.
*/
struct NzudsCppUserTypes
{
    int length;
    nz::udx::UdxBase::DataType * fieldTypes;

    
    void fillInTypes(NzudsUserTypes *pTypes) const;
    void fromTypes(NzudsUserTypes *pTypes);
    
};

/*
    class NzudsCppData
    field data to serialize / deserialize
    The memory that this class points to belongs
    to the class.

    The data in this class is valid until it is used in another
    readNext, any write operation is performed, or the handle is closed.
*/
class NzudsCppData
{
public:
    union
    {
        const int8_t * pInt8;
        const int16_t * pInt16;
        const int32_t * pInt32;
        const int64_t * pInt64;
        const float * pFloat;
        const double * pDouble;
        const int32_t * pDate;
        const int64_t * pTime;
        const int64_t * pTimeStamp;
        const Interval * pInterval;
        const bool * pBool;
        const TimeTzADT * pTimeTz;
        const char * pFixedString;
        const char * pVariableString;
        const char * pNationalFixedString;
        const char * pNationalVariableString;
        const CNumeric32 * pNumeric32;
        const CNumeric64 * pNumeric64;
        const CNumeric128 * pNumeric128;
        const char * pGeometryString;
        const char * pVarbinaryString;
    } data;
    nz::udx::UdxBase::DataType type;
    int32_t length;
    int8_t isNull;

    void fromData(const NzudsData *pData);
private:
    CNumeric32 numeric32Val;
    CNumeric64 numeric64Val;
    CNumeric128 numeric128Val;
    TimeTzADT timeTz;
    Interval interval;
};

class Nzuds
{
public:
    /* create empty instance */
    Nzuds();
    /* create instance from buffer */
    Nzuds(const NzudsCppBuffer * arg);
    /* create instance from buffer verifying fields match dataTypes */
    Nzuds(const NzudsCppBuffer * arg, NzudsCppUserTypes * dataTypes);
    ~Nzuds();

    /*
      Reinitialize handle with contents of arg. 
      returns 0 is success, -1 otherwise.
    */
    int setData(const NzudsCppBuffer * arg);

    /*
      Reinitialize handle with contents of arg. (but don't
      create copy of arg. The contents of arg must
      continue to be valid until done with the handle.
      returns 0 is success, -1 otherwise.
    */
    int setDataNoCopy(const NzudsCppBuffer * arg);


    /*
        serialize a field into the serialization object at the end
        returns 0 if sucess, -1 if memory allocation error
    */
    int writeInt8(int8_t data);
    int writeInt8Null();
    int writeInt16(int16_t data);
    int writeInt16Null();
    int writeInt32(int32_t data);
    int writeInt32Null();
    int writeInt64(int64_t data);
    int writeInt64Null();

    int writeFixedString(const char * data, int length);
    int writeFixedStringNull();
    int writeVariableString(const char * data, int length);
    int writeVariableStringNull();
    int writeNationalFixedString(const char * data, int length);
    int writeNationalFixedStringNull();
    int writeNationalVariableString(const char * data, int length);
    int writeNationalVariableStringNull();
    int writeGeometryString(const char * data, int length);
    int writeGeometryStringNull();
    int writeVarbinaryString(const char * data, int length);
    int writeVarbinaryStringNull();
    int writeFloat(float data);
    int writeFloatNull();
    int writeDouble(double data);
    int writeDoubleNull();
    int writeBool(bool data);
    int writeBoolNull();

    int writeDate(int32_t data);
    int writeDateNull();
    int writeTime(int64_t data);
    int writeTimeNull();
    int writeTimeStamp(int64_t data);
    int writeTimeStampNull();
    int writeTimeTz(const TimeTzADT * data);
    int writeTimeTzNull();
    int writeInterval(const Interval * data);
    int writeIntervalNull();

    int writeNumeric32(const CNumeric32 * data);
    int writeNumeric32Null();
    int writeNumeric64(const CNumeric64 * data);
    int writeNumeric64Null();
    int writeNumeric128(const CNumeric128 * data);
    int writeNumeric128Null();

    /*
        Serialize an NZUDS object into a NzudsCppBuffer.
        The object owns this data.
    */
    int getBufferPointer(NzudsCppBuffer * arg);
    int getBufferPointerCompressed(NzudsCppBuffer * arg);

    /*
      Fill in dataTypes from buffer. Return -1 on error, 0 on success.
      The caller must free the fieldTypes member which will be allocated
      by this function.
    */
    int getTypes(NzudsCppUserTypes * dataTypes);

    /*
      Fill in dataType from buffer. Return -1 on error, 0 on success.
    */
    int getType(int idx, nz::udx::UdxBase::DataType * dataType);

    /*
        get next serialized field
        the data returned is only valid until the next read next
        return a true value if more data, a false value at end of data
    */
    int readNext(NzudsCppData * data);

    /*
        returns number of serialized fields
    */
    int getNumElements();

    /*
       Reset read element position for readNext
       return 0 on success, -1 out of range
    */
    int setReadElementPosition(int arg);

    /*
      Return read element position for nzudsReadNext
    */
    int getReadElementPosition();

    /*
       remove elements with position >= nulElements
       if read position >= arg read position set to arg (end of data)
    */
    int shrinkTo(int numElements);

private:
    NZUDS_HANDLE handle;
};

#endif
