/*
 * Copyright IBM Corporation 2001, 2014
 */

// pcodegen.h - definitions relating to expression p-code

// (C) Copyright IBM Corp. 2005, 2012.  All Rights Reserved.

/*
  !FIX-dlm: Add convenience methods for generating this stuff.
   Add function support.
   Single literals array of pointers to literals?
   Ability to specify initial capacities up-front?
*/

/* Notes re gened code calling DirectFunctionCallN:
   1) need these c linkage extern in geninl.h:
    Datum DirectFunctionCall1(PGFunction func, Datum arg1);
    Datum int8fac(FunctionCallInfo fcinfo);

   2) also need above in BINDNAMEs in spubindsymbols.h

   3) not using direct call, but wrappers in spubind.cpp,
      as this facilitates pg arg passing
*/
   


#if !defined( INCLUDED_PCODEGEN )
#define INCLUDED_PCODEGEN

#ifndef NZSTD_INCLUDED
# include "nzstd.h"
#endif

#ifndef NZTYPES_INCLUDED
# include "nztypes.h"
#endif

#ifndef STRING_INCLUDED
# include <string.h> // for memset
# define STRING_INCLUDED
#endif

#include "codefragment.h"
#include "memPool.h"

class TStreamInfo;
class CGenCode;
class TableShape;
class CPlan;
struct CaseExpr;
struct Expr;
struct pcArg;
struct Var;
struct Param;
struct Const;
struct RelabelType;
struct TNumericData;
struct List;
struct Node;
struct MultiColExpr;

const static int PCODE_END = 0xFFFF;    // magic, end-of-pcode indicator

#define MAXSYSOID 200000 // !REWORK - don't define inline - where?

// define some buffer sizes parametrically
#define SMALLSIZE           40
#define NULLTESTARRAYSIZE  200
#define MEDIUMSIZE        2000 // big enough for field name string + "Null" or + (int64) extractor, patternbuf needs

#if 0
 #define BIGSIZE         200000 // really big 
 #define NULLTERMSIZE    100000
 #define ARGSTRSIZE        5000 
#else
 #define BIGSIZE         1000
 #define NULLTERMSIZE    1000
 #define ARGSTRSIZE      1000
#endif


enum EpcArgCode {
    epcCodeField = 0,   // field reference by logical field ID
    epcCodeTemp,        // reference to temporary
    epcCodeLiteral,     // literal reference
    epcCodeResult,      // goes to result table
    epcCodeTossResult,  // goes to the bit bucket (elimination of duplicate subexprs)
    epcCodeParam        // subquery Param reference
};

enum EpcInstrCode {
    epcInstrFunc = 0,
    epcInstrCase,
    epcInstrPostgres
};

struct pcArg {
    EpcArgCode code;      // field, temp, or literal
    int type;             // data type
    union {
        int charsize;     // char field size
        struct {
            int domaintype; // numeric domain type
            int precision;  // numeric precision
            int declaredPrecision; // declared precision of the numeric
            int scale;      // numeric scale
        } num;
    } tmod;
    bool nullable;        // true if field can be null
    bool nullConst;       // true if arg is a null constant
    bool caseRelated;      // true if arg is direct result of CASE expr
    bool caseWhenRelated; // true if this is an arg in a case stmt when expr
    unsigned int table;         // INNER, OUTER, or index into range table
    int index;          // field number, temp number, or literal offset (signed for magic like rownum)
    unsigned int scaleIndex;    // index to literal scale multiplier value

    pcArg() { memset(this, 0, sizeof(pcArg)); } // for clean local pcArg's mostly
    void Dump( void ) const;
    // !NOTE-jpb Better to have an "argMatch" function than this == op that only tests some members
    bool operator==(pcArg& that) {
        return (code==that.code && type==that.type && table==that.table && index==that.index);
    }
};

struct pcCaseWhen {
    pcArg* exprP;
    pcArg* valueP;
};

struct pcCaseInstruction {
    int whencnt;
    pcCaseWhen *whenP;  // points to array of pcCaseWhen structs
    pcArg* defvalueP;   // default return val

    void Dump( void ) const;
};

struct pcPostgresInstruction {
    void Dump( void ) const;
};

struct pcFuncInstruction { // for functions and opertators
    int argcnt;
    pcArg *argP;      // points to array of pcArg structs
    uint32 op;        // operation
    int argIndex;     // current argument number in current instruction (0 or 1)

    void Dump( void ) const;
};

struct pcInstruction {
    EpcInstrCode type;
    pcArg* resultP;        // where to put result
    int hintFlags;
    int caseNum;
    int caseWhenNum;
    union pcInst {
        pcCaseInstruction     caseInst;
        pcFuncInstruction     funcInst;
        pcPostgresInstruction pgInst;
    } pcInst;

    void Dump( void ) const;
};

// pcode instruction hint flags
#define HINT_CASE_RESULT_CHILD   0x0001
#define HINT_NUMERIC_CAST_RESULT 0x0002
#define HINT_MATERIALIZED_CASE   0x0004
#define HINT_CASE_WHEN_CHILD     0x0008
#define HINT_UDF_ARG             0x0010
#define HINT_UDF_COMPLETE        0x0020
        // pcode entry for UDF is complete and must have corresponding entry in UDF info array

// used for literal values of the two structured postgres temporal types
// i.e. TimeTz and Interval
struct pcTTzInt; 

struct pcFacts {
    pcInstruction* pPcode;      // pointer to pcode instructions

    int8* pLiterals8;           // pointer to memory containing literals
    int16* pLiterals16;         // pointer to memory containing literals
    int32* pLiterals32;         // pointer to memory containing literals
    int64* pLiterals64;         // pointer to memory containing literals
    pcTTzInt* pLiterals96;      // pointer to memory containing literals
    TNumericData* pLiterals128; // pointer to memory containing literals
    char** pLiteralsChar;       // pointer to memory containing literals
    char** pLiteralsClob;       // pointer to memory containing literals
    bool* pLiteralsBool;        // pointer to memory containing literals
    float* pLiteralsFlt;        // pointer to memory containing literals
    double* pLiteralsDbl;       // pointer to memory containing literals
    unsigned char** pLiteralsVarbinary; // pointer to memory containing literals
    
    int literals8Count;         // number of literals in pLiterals8
    int literals16Count;        // number of literals in pLiterals16
    int literals32Count;        // number of literals in pLiterals32
    int literals64Count;        // number of literals in pLiterals64
    int literals96Count;        // number of literals in pLiterals96
    int literals128Count;       // number of literals in pLiterals128
    int literalsCharCount;      // number of literals in pLiteralsChar
    int literalsClobCount;      // number of literals in pLiteralsClob
    int literalsBoolCount;      // number of literals in pLiteralsBool
    int literalsFltCount;       // number of literals in pLiteralsFlt
    int literalsDblCount;       // number of literals in pLiteralsDbl
    int literalsVarbinaryCount; // number of literals in pLiteralsVarbinary
    
    int pcodeCount;           // number of pcode instructions
    unsigned int temp8Index;          // current temp8-result index
    unsigned int temp16Index;         // current temp16-result index
    unsigned int temp32Index;         // current temp32-result index
    unsigned int temp64Index;         // current temp64-result index
    unsigned int temp96Index;         // current temp96-result index
    unsigned int temp128Index;        // current temp128-result index
    unsigned int tempCharIndex;       // current tempChar-result index
    unsigned int tempBoolIndex;       // current tempBool-result index
    unsigned int tempFltIndex;        // current tempFlt-result index
    unsigned int tempDblIndex;        // current tempDbl-result index
    unsigned int tempVarbinaryIndex;  // current tempVarbinary-result index

    void Dump( void ) const;
};

struct PcodeContext {
    int pcodeCapacity;        // current number of slots in pPcodeStack

    int pcodeGenCounter;      // used for generated temp name uniqueness
    int literals8Capacity;    // current capacity of pLiterals8
    int literals16Capacity;   // current capacity of pLiterals16
    int literals32Capacity;   // current capacity of pLiterals32
    int literals64Capacity;   // current capacity of pLiterals64
    int literals96Capacity;   // current capacity of pLiterals96
    int literals128Capacity;  // current capacity of pLiterals128
    int literalsCharCapacity; // current capacity of pLiteralsChar
    int literalsClobCapacity; // current capacity of pLiteralsClob
    int literalsBoolCapacity; // current capacity of pLiteralsBool
    int literalsFltCapacity;  // current capacity of pLiteralsFlt
    int literalsDblCapacity;  // current capacity of pLiteralsDbl
    int literalsVarbinaryCapacity; // current capacity of pLiteralsVarbinary
    pcFacts facts;            // data attached to project node

    void Dump( void ) const;
};

enum EpcGenexStyleCode {
    epcGenexStyleExpr = 0,
    epcGenexStyleRestrict,
    epcGenexStyleConditionalBranch,
    epcGenexStyleHashJoinQual,
    epcGenexStyleFilter
};

struct NullableField { // used in code gen to gen null test
    EpcArgCode code;
    int tblnum;
    int colnum;
};

enum EpcSourceTempCode {
    epcSourceTempLastEntry = -4,
    epcSourceTempSubExpr   = -3,
    epcSourceTempField     = -2,
    epcSourceTempLiteral   = -1, // negs signal subexpr type rather than temp type
    epcSourceTempCharPtr   =  0,
    epcSourceTempInt,
    epcSourceTempInt64,
    epcSourceTempVarArray,
    epcSourceTempCharArray,
    epcSourceTempDateParts,
    epcSourceTempBool,
    epcSourceTempDbl
};

struct sourceTemp {
    EpcSourceTempCode  code;
    int                size;
    int                generation;
    pcArg              arg;
    int                targetOverride;
    int                caseNum;
    int                pointerIndex;
    char              *pNullTestStr;
    char              *pCtorArgs;
    char	          *pMarshal;
    bool	           isstrict;
    bool               unmaterializedWhenArg;
    bool               isLateMaterializedWhenArg;
};

struct paramSpec {
    int         subplanNum;
    TStreamInfo *flowInfo;
    int         ownerNodePos;
};

struct pcState {
    pcFacts facts;
    int paramEnd;
};

class CPcodeGen {

 public:
    void *operator new(size_t);
    void operator delete(void*);

    CPcodeGen();
    ~CPcodeGen();

    bool GenerateQual(
        List* pQual,      // pointer to expression node
        const TableShape* pStreamTable,
        const TableShape* pTempTable,
        int joinType = 0);
    bool Generate(
        Node** const pNodeP,    // pointer to expression node pointer
        const int resultIndex,  // fieldID in result table if non-negative
        const bool bCanUsePostgres,
        const TableShape* const pStreamTable,
        const TableShape* constpTempTable,
        int* const varStrBufSizeP,
        const int joinType = 0);
    bool GenerateMultiColExpr(
        const MultiColExpr* const mce,
        const int resultIndex,  // fieldID in result table if non-negative
        const TableShape* const pStreamTable,
        const TableShape* const pTempTable,
        const int joinType,
        Node* placeholderTargets[],
        int varStrBufSizes[]);
    bool PcodeExpr(
        const Node* const pNode,        // pointer to expression node
        pcArg* pArg,        // pointer to place to store arg info (or NULL)
        int resultIndex,    // fieldID in result table if non-negative
        int* pScaleFactor,  // scale result up by 10**scaleFactor (for intermediate numerics)
        int hintFlags,      // hint flags
        int caseNum,
        int caseWhenNum,
        bool relabelTextOk=false); // used internal for explicit cast text to varchar support
    void GenEndInstruction( void );
    void InitContext( CPlan* pPlan );
    void ResetPool();
    void Dump( void ) const;
    bool NeedsPostgresEval(
        int pcodeStart,
        int pcodeCount );
    bool NeedsPostgresMemoryContext(
        int pcodeStart,
        int pcodeCount );
#if 0    // unused
    int ExprVarStrBufferSize(
        int exprno,
        int codeStart,
        int codeCount);
#endif    
    void SetNullableFlags(
        int codeStart,
        int codeSize);
    bool ExprNullable(
        const int exprno,
        const int codeStart,
        const int codeCount);
    bool PcodeToCcode(
        int codeStart,
        int codeSize,
        bool declareResults,
        EpcGenexStyleCode style,
        TStreamInfo *outStreamP,
        TStreamInfo *flowInStreamP,
        TStreamInfo *tempInStreamP,
        const char *pVarName = NULL,
         bool aggrJoin = false,
        int * snipResCacheableP = 0); // reason not cacheable
    void InitTemps( void );
    void InitParamSpecList(
        void);
    void EmitParamStructs(
        CGenCode * const genCodeP);
    void EmitParamRecInits(
        int nodePos,
        CGenCode * const genCodeP);

    // accessors:

    inline int i_pcodeCount( void ) {
        return m_PcodeContext.facts.pcodeCount;
    }
    inline pcFacts* i_pPcodeFacts( void ) {
        return &(m_PcodeContext.facts);
    }
    inline int i_tempOffsetsCount( void ) {
        return m_tempOffsetsCount;
    }
    inline int* i_tempOffsetsP( void ) {
        return m_tempOffsetsP;
    }
    inline int i_seqCount( void ) {
        return m_seqCount;
    }
    inline int* i_sequenceP( void ) {
        return m_sequenceP;
    }

    CodeFragment *m_workterm;
    CodeFragment *m_marshall;
    CodeFragment *m_nullterm;
    CodeFragment *m_nullterm2;
    CodeFragment *m_nulleqterm;
    CodeFragment *m_term;
    CodeFragment *m_prefix;
    
  private:
    
    bool pcodeCase(
        const CaseExpr* const pCaseExpr,
        pcArg* const pArg,
        const int pcodeIndex,
        const int resultIndex,
        const int hintFlags,
        const int caseNum,
        const int caseWhenNum );
    bool pcodeFunc(
        Expr* pExpr,
        pcArg* pArg,
        int pcodeIndex,
        int resultIndex,
        int hintFlags,
        int caseNum,
        int caseWhenNum );
    bool pcodeOper(
        Expr* pExpr,
        pcArg* pArg,
        int pcodeIndex,
        int resultIndex,
        int hintFlags,
        int caseNum,
        int caseWhenNum );
    bool deriveNullFldName(
        pcInstruction* pPcode,
        char *nullfldname,
        char *fldname,
        bool bCaseExpr,
        const char *pVarName);
    bool isValidPcodeOp(
        Oid op );
    bool isSupportedPcodeOp(
        Oid op );
    bool pcodeVar(
        Var* pVar,
        pcArg* pArg);
    bool pcodeParam(
        const Param* const pParam,
        pcArg* const pArg);
    bool pcodeConst(
        Const* pConst,
        pcArg* pArg,
        int* pScaleFactor );
    bool pcodeRelabelCopyArg(
        RelabelType* pRelabelType,
        Node* pNode,
        pcArg* pArg,
        int pcodeIndex,
        int resultIndex,
        int* pScaleFactor,
        int hintFlags,
        int caseNum,
        int caseWhenNum );    
    bool pcodeRelabelType(
        RelabelType* pRelabelType,
        pcArg* pArg,
        int resultIndex,
        int* pScaleFactor,
        bool relabelTextOk,
        int hintFlags,
        int caseNum,
        int caseWhenNum );
    bool pcodeCopyArgToClob(
        Node* pNode,
        pcArg* pArg,
        int pcodeIndex,
        int resultIndex,
        int* pScaleFactor,
        int hintFlags,
        int caseNum,
        int caseWhenNum);
    void setupCaseInst(
        pcInstruction* const pInstr,
        const int hintFlags, const int caseNum, const int caseWhenNum,
        const int argcnt,
        const bool allocCaseExpr = true);
    void setupCaseResult(
        pcArg* const pResult,
        const Node* const expr,
        pcArg* const arg,
        const int resultIndex,
        const int col= -1); // only needed for multi-col expr
    bool generateOneCaseValue(
        const Node* const inputValExpr,
        pcArg* const arg,
        const int hintFlags,
        const int caseNum,
        const int caseWhenNum,
        const int postgresType,
        const int pcodeType,
        const int caseResultScale,
        int* const scaleFactor,
        const int fixedCharResultSize);
    bool PcodeMultiColExpr(
        const MultiColExpr* const mce,
        const int resultIndex,
        Node* placeholderTargets[],
        int varStrBufSizes[]);
    int addToLiterals(
        Const* pConst,
        int* pScaleFactor,
        int* pDataType,
        int* pScale);
    int mapType(
        const int type,
        const Node* const pNode,
        const bool bDomain,
        const int col = -1); // only needed for multi-col expr
    void setTMod(
        pcArg* const pArg,
        const Node* const pNode,
        const int col = -1); // only needed for multi-col expr
    void validateSizing(
        const Node* const pNode,
        const int col = -1); // only needed for multi-col expr
    bool numericModFunc(
        int funcid);
    bool floatModFunc(
        int funcid);
    void* growShPoolAllocedArray(
        void* pOld,     // pointer to old object
        int* pCap,      // address of current capacity
        int increment,  // number by which to increase capacity
        int elemSize ); // size of array element
    void growPcode(const int numMissingInstrs = 1);
    void growLiterals8( int newMin );
    void growLiterals16( int newMin );
    void growLiterals32( int newMin );
    void growLiterals64( int newMin );
    void growLiterals96( int newMin );
    void growLiterals128( int newMin );
    void growLiteralsChar( int newMin );
    void growLiteralsClob( int newMin );
    void growLiteralsBool( int newMin );
    void growLiteralsFlt( int newMin );
    void growLiteralsDbl( int newMin );
    void growLiteralsVarbinary( int newMin );
    void growSubexprTemps( int newMin );
    void growSourceTemps( int newMin );
    void growNullfields( int newMin );
    void growParamSpecs( int newMin );
    bool pcodeAndOrNot(
        Expr* pExpr,
        pcArg* pArg,
        int pcodeIndex,
        int resultIndex,
        int hintFlags,
        int caseNum,
        int caseWhenNum );
    int getResultScale(
        const Node * const pNode,
        int* const precisionP = 0,
        int* const declaredPrecisionP = 0,
        const int col = -1); // only for multi-col expr
    int getResultNumericDigitCount(
        const Node * const pNode,
        const bool bDomain,
        const int col = -1); // only for multi-col expr
    int getDataType(
        const Node *const pNode,
        const bool bDomain = false,
        const int col = -1); // only for multi-col expr
    int getDataSize(
        const Node *const pNode,
        const int col = -1); // only for multi-col expr
    int getOpCode(
        Expr *pExpr);
    const char *opcodepattern(
        pcInstruction *pPcode,
        char *buf);
    int ndeResultType(pcInstruction *pPcode);
    bool ndeResultTypeArray(pcInstruction *pPcode);
    bool pcodeArgCcode(
        pcArg *pArg,
        CodeFragment *codeFrag,
        bool bForceUseTempSource = false,
        bool bOverride=false,
        bool bWithPrefix=true,
        int caseNum=0,
        bool strictop=false);
    bool opIsNumMul(
        int op);
    void saveState(
        pcState *);
    void restoreState(
        pcState *);
    void addNullableField(
        EpcArgCode code,
        int tblnum,
        int colnum,
        int opcode);
    int addParamSpec(
        int subplanNum);
    int firstParamInstr(
        int subplanNum);
    int instrOwnerNode(
        int instrNum,
        CGenCode * genCodeP);
    int addTemp(
        EpcSourceTempCode code,
        int size,
        pcArg *pArg = 0,
        const char *nullteststrP = 0,
        int targetOverride = 0,
        const char *ctorArgsP = 0,
        int caseNum = 0);
    int getSubexpressionNum(pcArg *pArg,
        bool bOverride=false,
        bool *bNotFoundP=NULL);
    void setSubexpressionGeneration(
        pcArg *pArg,
        bool bOverride=false,
        int caseNum=0);
    void getTempName(
        bool bSubexpr,
        int tempNo,
        char *buf);
    char* getCharBufName(
        pcInstruction *pPcode,
        char *nameBuf);
    void putTempDecls(
        bool forJoin);
    void materializeIsNullArgs(
        int codeStart,
        int codeSize);
    void materializeWrapPointerArgs(
        pcInstruction *pPcode,
        int codeStart,
        int codeSize);
    void eliminateRepeatedSubExpressions(
        int codeStart,
        int codeSize);
    void eliminateRepeatedCaseSubExpressions(
        int codeStart,
        int codeSize);
    int tempIsMaterialized(
        const EpcSourceTempCode code,
        pcArg*const pArg,
        const bool bOverride=false,
        const bool bIncludeCaseRelated=false,
        const bool bIgnoreLateMaterializedWhenArgs=false,
        const int caseNum=0);
    void materializeCommonSubExpressions(
        int codeStart,
        int codeSize);
    bool cseEligible(
        pcInstruction *pCode);
    void materializeCaseResults(
        int codeStart,
        int codeSize);
    void materializeOneNullLiteral(
        pcArg* const arg);
    void materializeNullLiterals(
        const int codeStart,
        const int codeSize);
    void insureTempMaterialized(
        pcArg *pArg);
    bool implementationRequiresMaterialization(
        pcInstruction *pPcode1,
        pcInstruction *pPcode2 );
    void replaceTemp(
        pcInstruction *pOrigInst,
        pcInstruction *pReplacementInst,
        int codeStart,
        int codeSize);
    bool argsMatch(
        pcArg *pArg1,
        pcArg *pArg2);
    bool tempsMatch(
        pcInstruction *pInstr1,
        pcInstruction *pInstr2);
    int usedCount(
        pcInstruction *pInstr1,
        pcInstruction *pInstr2);
    bool isCoercionRequired(
        pcArg *pArg, 
        int resType);
    bool genInstrPostgres(
        pcInstruction *pPcode);
    bool genInstrFunc(pcInstruction *pPcode,
                      int codeStart,
                      int codeCount,
                      int * snipResCacheableP);
    
    bool genInstrCase(pcInstruction *pPcode,
                      int codeStart,
                      int codeCount,
                      char *nullfldname);

    void emitInlineCharComparison(char *cRef,
                                  int charOffset, 
                                  const char *literalP,
                                  int literalSize);
    bool numericToFloat(
        bool bDouble,
        pcArg *pArg,
        CodeFragment *codeFrag);
    pcInstruction *findTempSource(
        pcArg *pArg,
        int codeStart,
        int codeSize);
    void getArgNullList(
        pcArg *pArg,
        int codeStart,
        int codeSize,
        int opcode,
        bool forNullEq=false,
        bool bIgnoreLateMaterializedWhenArgs=false,
        bool bIgnoreStrictness=false,
        bool bIsNullValue=false);
    bool buildNullTestStr(
        CodeFragment *nulltern,
        bool forNullEq=false);
    void getNullTest(
        char *buf,
        int bufsize,
        pcArg *pArg,
        char *tempName);
    const char *simpleWrap(int opcode);
    const char *simplerWrap(int opcode);
    bool datumIsPointer( pcArg* pArg );
    bool datumRequiresPgFormatTemp( pcArg* pArg );
    void addToOffsetsArray(
        int tempSize );
    char *tsFmtString (
        int type,
        const char *str );

    int getSeqOffset (
        int seqOid);
    
    bool instrIsLiteral(
        pcInstruction *pPcode);

    bool nondeterministicUDF(
        pcInstruction *pPcode);

    void putMarshal (
        CodeFragment *workbuf,
        const char *nullfldnmP);

    Node* getPlaceholderForNode(
        Node* const node,
        const int resultInstrIndex,
        const int col = -1);

    int nextTempIndexForType(const int type);
    int prepareNextPCodeInstruction(const int numNewInstrsNeeded = 1);

    TStreamInfo *m_outStreamP;
    TStreamInfo *m_flowInStreamP;
    TStreamInfo *m_tempInStreamP;

    bool m_declareResult;
    EpcGenexStyleCode m_style;
    bool m_newCase;
    bool m_caseCSE;
    bool m_numOptimization;

    char** m_temp8;
    char** m_temp16;
    char** m_temp32;
    char** m_temp64;
    char** m_temp96;
    char** m_temp128;
    char** m_tempchar;
    char** m_tempbool;
    char** m_tempdbl;
    char** m_tempflt;
    unsigned char** m_tempvarbinary;

    sourceTemp* m_sourceTemps;
    sourceTemp* m_subexprTemps;
    NullableField* m_nullfields;
    paramSpec* m_paramSpecs;

    int m_temp8CurCnt;
    int m_temp16CurCnt;
    int m_temp32CurCnt;
    int m_temp64CurCnt;
    int m_temp96CurCnt;
    int m_temp128CurCnt;
    int m_tempcharCurCnt;
    int m_tempboolCurCnt;
    int m_tempdblCurCnt;
    int m_tempfltCurCnt;
    int m_tempvarbinaryCurCnt;

    int m_sourceTempsCurCnt;
    int m_subexprTempsCurCnt;
    int m_nullfieldsCurCnt;
    int m_paramSpecsCurCnt;

    bool m_bNullTestingEnabled;

    int m_tempOffsetsCount;    // number of entries in m_pTempOffsets
    int m_tempOffsetsCapacity; // capacity of m_pTempOffsets
    int* m_tempOffsetsP;       // offsets of vlen_t temp buffers for expressions

    int m_seqCount;            // number of entries in m_sequences
    int m_seqCapacity;         // capacity of m_sequences array
    int* m_sequenceP;          // pointer to array of sequence ids / object pointers

    const TableShape* m_StreamTable;
    const TableShape* m_TempTable;

    CPlan*  m_pPlan;
    int m_joinType;
    bool m_aggrJoin;

    PcodeContext m_PcodeContext;
    mem::Pool *m_pcodePool;
};


// functions we've overloaded in pg_proc.h (e.g. date_part), will not have
// entries for all overloads in fmgroids.h.  Unforseen consequence of overloads.
// So we have to define some here, which is less that ideal.  !FIX-jpb
#define PCF_DATE_PART 1384
#define PCF_ABSTIME_PART 1382
#define PCF_ABSTIME_TIME 1364

#define PCF_SUBSTR3F1    936  // 3-arg form of substring for fixed char arg 
#define PCF_SUBSTR2F1    937  // 2-arg form of substring for fixed char arg
#define PCF_SUBSTR3F2    877  // 3-arg form of substr    for fixed char arg 
#define PCF_SUBSTR2F2    883  // 2-arg form of substr    for fixed char arg

#define PCF_SUBSTR3V2    864  // 3-arg form of substring for var char arg 
#define PCF_SUBSTR2V2    865  // 2-arg form of substring for var char arg 
#define PCF_SUBSTR3V1    862  // 3-arg form of substr    for var char arg 
#define PCF_SUBSTR2V1    863  // 2-arg form of substr    for var char arg 

#define PCF_SUBSTR3T1    248  // 3-arg form of substring for text arg 
#define PCF_SUBSTR2T1    249  // 2-arg form of substring for text arg 
#define PCF_SUBSTR3T2    246  // 3-arg form of substr    for text arg 
#define PCF_SUBSTR2T2    247  // 2-arg form of substr    for text arg 

#define PCF_SUBSTR3B1    473  // 3-arg form of substring for blob arg 
#define PCF_SUBSTR2B1    566  // 2-arg form of substring for blob arg 
#define PCF_SUBSTR3B2    583  // 3-arg form of substr    for blob arg 
#define PCF_SUBSTR2B2    676  // 2-arg form of substr    for blob arg 

/* Removed until SubstNucl implemented in spubind.cpp
 * Search for SUBSTR_NUCL_SUPPORT to find related code to enable
 *
 * These defines must be in synch with the relevant pg_proc.h definitions
 *
 * #define PCF_SUBSTR3N1    712  // 3-arg form of substring for nucl arg 
 * #define PCF_SUBSTR2N1    767  // 2-arg form of substring for nucl arg 
 * #define PCF_SUBSTR3N2    143  // 3-arg form of substr    for nucl arg 
 * #define PCF_SUBSTR2N2    239  // 2-arg form of substr    for nucl arg

#define PCF_SUBSTR3P1    949  // 3-arg form of substring for prot arg 
#define PCF_SUBSTR2P1    973  // 2-arg form of substring for prot arg 
#define PCF_SUBSTR3P2    773  // 3-arg form of substr    for prot arg 
#define PCF_SUBSTR2P2    886  // 2-arg form of substr    for prot arg 
*/


#define PCF_NUMERIC_MOD_OP    1729  // % for numerics (distinct from 1728 F_NUMERIC_MOD function!)

// "abs" funcs from pg_proc
#define PCF_FLT4_ABS 1394
#define PCF_FLT8_ABS 1395
#define PCF_INT8_ABS 1396
#define PCF_INT4_ABS 1397
#define PCF_INT2_ABS 1398
#define PCF_INT1_ABS 6076
#define PCF_NUM_ABS  1705

// alternative mod funcs from pg_proc
#define PCF_INT1_MOD  2519
#define PCF_INT2_MOD  940
#define PCF_INT4_MOD  941
#define PCF_INT24_MOD 942
#define PCF_INT42_MOD 943
#define PCF_INT8_MOD  947
#define PCF_INT14_MOD 1586
#define PCF_INT41_MOD 1587
#define PCF_INT18_MOD 1588
#define PCF_INT81_MOD 1589
#define PCF_INT28_MOD 1590
#define PCF_INT82_MOD 1591
#define PCF_INT48_MOD 1592
#define PCF_INT84_MOD 1593
#define PCF_INT12_MOD 1594
#define PCF_INT21_MOD 1595
#define PCF_FLOAT4_MOD 1148
#define PCF_FLOAT8_MOD 1149

#define PCF_NMIXEDMOD_1 580
#define PCF_NMIXEDMOD_2 903
#define PCF_NMIXEDMOD_3 906
#define PCF_NMIXEDMOD_4 909
#define PCF_NMIXEDMOD_5 912
#define PCF_NMIXEDMOD_6 915
#define PCF_NMIXEDMOD_7 918
#define PCF_NMIXEDMOD_8 921
#define PCF_NMIXEDMOD_9 924
#define PCF_NMIXEDMOD_10 927
#define PCF_NMIXEDMOD_11 930
#define PCF_NMIXEDMOD_12 933
#define PCF_NMIXEDMOD_13 512
#define PCF_NMIXEDMOD_14 496
#define PCF_NMIXEDMOD_15 499
#define PCF_NMIXEDMOD_16 502

#define PCF_FMIXEDMOD_1 570
#define PCF_FMIXEDMOD_2 571
#define PCF_FMIXEDMOD_3 572
#define PCF_FMIXEDMOD_4 573
#define PCF_FMIXEDMOD_5 574
#define PCF_FMIXEDMOD_6 575
#define PCF_FMIXEDMOD_7 576
#define PCF_FMIXEDMOD_8 577
#define PCF_FMIXEDMOD_9 578
#define PCF_FMIXEDMOD_10 579
#define PCF_FMIXEDMOD_11 901
#define PCF_FMIXEDMOD_12 902
#define PCF_FMIXEDMOD_13 904
#define PCF_FMIXEDMOD_14 905
#define PCF_FMIXEDMOD_15 907
#define PCF_FMIXEDMOD_16 908
#define PCF_FMIXEDMOD_17 910
#define PCF_FMIXEDMOD_18 911
#define PCF_FMIXEDMOD_19 913
#define PCF_FMIXEDMOD_20 914
#define PCF_FMIXEDMOD_21 916
#define PCF_FMIXEDMOD_22 917
#define PCF_FMIXEDMOD_23 919
#define PCF_FMIXEDMOD_24 920
#define PCF_FMIXEDMOD_25 922
#define PCF_FMIXEDMOD_26 923
#define PCF_FMIXEDMOD_27 925
#define PCF_FMIXEDMOD_28 926
#define PCF_FMIXEDMOD_29 928
#define PCF_FMIXEDMOD_30 929
#define PCF_FMIXEDMOD_31 931
#define PCF_FMIXEDMOD_32 932
#define PCF_FMIXEDMOD_33 934
#define PCF_FMIXEDMOD_34 511
#define PCF_FMIXEDMOD_35 513
#define PCF_FMIXEDMOD_36 495
#define PCF_FMIXEDMOD_37 497
#define PCF_FMIXEDMOD_38 498
#define PCF_FMIXEDMOD_39 500
#define PCF_FMIXEDMOD_40 501
#define PCF_FMIXEDMOD_41 503
#define PCF_FMIXEDMOD_42 504

// 3-arg shift funcs
#define PCF_INT2SHL 2544
#define PCF_INT2SHR 2545
#define PCF_INT4SHL 2546
#define PCF_INT4SHR 2547
#define PCF_INT8SHL 2548
#define PCF_INT8SHR 2549
#define PCF_INT1SHL 2550
#define PCF_INT1SHR 2551

// upper and lower for varchars (fixed in fmgroids.h)
#define PCF_LOWER_VARCHAR 888
#define PCF_UPPER_VARCHAR 889

// sized string LENGTH variants
#define PCF_BPCHARLEN  1372
#define PCF_VARCHARLEN 1373

// TRIM variants
#define PCF_LTRIM1 881
#define PCF_LTRIM2 892
#define PCF_LTRIM3 896
#define PCF_RTRIM1 882
#define PCF_RTRIM2 893
#define PCF_RTRIM3 897
#define PCF_BTRIM  885

// PAD variants
#define PCF_LPAD1 879
#define PCF_LPAD2 890
#define PCF_LPAD3 894
#define PCF_RPAD1 880
#define PCF_RPAD2 891
#define PCF_RPAD3 895

// translate variant
#define PCF_TRANSLATE 878

// extra like/nlike opcodes from pg_proc.h
#define PCF_LIKE  1569
#define PCF_NLIKE 1570

// noop coercions:
#define PCF_I8TOI8   1573
#define PCF_I4TOI4   1405
#define PCF_I2TOI2   1403
#define PCF_I1TOI1   1581
#define PCF_DTOD     1404
#define PCF_FTOF     1402
#define PCF_TSTOTS   1368
#define PCF_IVTOIV   1369
#define PCF_DTTODT   1371
#define PCF_TTZTOTTZ 1387

// misc aliases
#define PCF_DLOG10 1340
#define PCF_DLN 1341
#define PCF_DROUND 1342
#define PCF_DTRUNC 1343
#define PCF_DSQRT 1344
#define PCF_DPOW 1346
#define PCF_DEXP 1347
#define PCF_TEXTPOS 868
#define PCF_TRANSLATE 878
#define PCF_INITCAP 898
#define PCF_OVERLAPS_TIMESTAMP1 1305
#define PCF_OVERLAPS_TIMESATMP2 1306
#define PCF_OVERLAPS_TIMESTAMP3 1307
#define PCF_OVERLAPS_TIME1 1309
#define PCF_OVERLAPS_TIME2 1310
#define PCF_OVERLAPS_TIME3 1311
#define PCF_DATETIMETZ_TIMESTAMP 1359
#define PCF_TIMESTAMP_AGE 1386
#define PCF_DATE_AGE 1187
#define PCF_DATETIMETZ_TIMESTAMP 1359
#define PCF_DATETIME_PL 1272
#define PCF_TIMEDATE_PL 1296
#define PCF_TIMETZDATE_PL 1298
#define PCF_TIME_PART 1385
#define PCF_STRPOS 868
#define PCF_NUMERIC_ROUND 1708
#define PCF_NUMERIC_TRUNC 1710
#define PCF_INT2FAC 1391
#define PCF_INT4FAC 1392
#define PCF_INT8FAC 1393
#define PCF_NAMELIKE 1571
#define PCF_NAMENLIKE 1572
#define PCF_BYTECODES2 725
#define PCF_CLOB_POS 344
#define PCF_BINARY2HEX 6390
#define PCF_HEX2BINARY 6441

// we need our own typecodes for the numeric variation
#define FIRST_PCODE_OID 20000    // stay above the 16384 max internal Postgres oid (transam.h)

// first, the various numeric data sizes
#define PC_NUMERIC_1 20000  // numeric held in 1 32-bit digit
#define PC_NUMERIC_2 20001  // numeric held in 2 32-bit digits
#define PC_NUMERIC_4 20002  // numeric held in 4 32-bit digits

// pcode instruction to invoke the postgres evaluator:
#define PC_EXEC_EVAL  20003  // call pg exec eval

#define PC_OR         20004  // or
#define PC_AND        20005  // and

// datatype for fixedvar field
#define PC_VARFIXED   20006
#define PC_FIXED_N    20007

// magic function to return a simple const
#define PCF_CONSTX    20008

// magic type codes for materialization of a pg format temps for wrapped pg func
#define PC_VARLENA    20009
#define PC_NUMERICVAR 20010

// pcode function code meaning copy arg0
#define PC_COPY_ARG0  20011

// pcode funcs for current timestamp stuff
#define PC_CUR_DATE         20012
#define PC_CUR_TIME         20013
#define PC_CUR_TIMESTAMP    20014
#define PC_TSTAMP_CHAR_FAST 20015 // a format string we can handle in gened code

// pcode function code meaning convert type
#define PC_CVT_TYPE      20016

// pcode function to do "not"
#define PC_NOT           20017

#define PC_CHAR_DATE_1   20018 // to_date with format string 'yyyymm'

#define PC_COPY_TO_CLOB  20019 // relabel of char/varchar to clob uses this


#define IS_PCODE_VARYING_TYPE(t) (t==PGTYPE_text || \
                            t==PGTYPE_nchar || \
                            t==PGTYPE_nvarchar || \
                            t==PGTYPE_geometry || \
                            t==PGTYPE_varbinary || \
                            t==PGTYPE_varchar || \
                            t==PC_VARFIXED)
#define IS_PCODE_STR_OR_BINARY(t) (t==PGTYPE_bpchar || t==PC_FIXED_N || IS_PCODE_VARYING_TYPE(t))
#define IS_NUMERIC(t) (t==PC_NUMERIC_1 || t==PC_NUMERIC_2 || t==PC_NUMERIC_4)
#define IS_INT(t) (t==PGTYPE_int1 || t==PGTYPE_int2 || t==PGTYPE_int4 || t==PGTYPE_int8)

enum FtId {
    TSFMT_HH12 = 1,
    TSFMT_HH24 = 2,
    TSFMT_MI = 3,
    TSFMT_SS = 4,
    TSFMT_SSSS = 5,
    TSFMT_ucMERIDIAN = 6,
    TSFMT_ucdMERIDIAN = 7,
    TSFMT_lcMERIDIAN = 8,
    TSFMT_lcdMERIDIAN = 9,
    TSFMT_Y4C = 10,
    TSFMT_Y4 = 11,
    TSFMT_Y3 = 12,
    TSFMT_Y2 = 13,
    TSFMT_Y1 = 14,
    TSFMT_ucERA = 15,
    TSFMT_ucdERA = 16,
    TSFMT_lcERA = 17,
    TSFMT_lcdERA = 18,
    TSFMT_MONTH = 19,
    TSFMT_Month = 20,
    TSFMT_month = 21,
    TSFMT_MON = 22,
    TSFMT_Mon = 23,
    TSFMT_mon = 24,
    TSFMT_MM = 25,
    TSFMT_DAY = 26,
    TSFMT_Day = 27,
    TSFMT_day = 28,
    TSFMT_DY = 29,
    TSFMT_Dy = 30,
    TSFMT_dy = 31,
    TSFMT_DDD = 32,
    TSFMT_DD = 33,
    TSFMT_D = 34,
    TSFMT_W = 35,
    TSFMT_WW = 36,
    TSFMT_IW = 37,
    TSFMT_CC = 38,
    TSFMT_J = 39,
    TSFMT_Q = 40,
    TSFMT_RM = 41,
    TSFMT_rm = 42,
    TSFMT_MS = 43,
    TSFMT_US = 44,
    TSFMT_IY3 = 45,
    TSFMT_IY2 = 46,
    TSFMT_IY1 = 47,
    TSFMT_I1 = 48
};

// format allows FM modifier
enum FtFM {
    fmN = 0,
    fmY
};

// format allows TH or th suffix
enum FtTH {
    thN = 0,
    thY
};

// format string is case-sensitive
enum FtCase {
    caseN = 0,
    caseY
};

// format supported in gened code
enum FtSupported {
    ftN = 0,
    ftY
};

// format output is fixed length
enum FtFixed {
    fxN = 0,
    fxY
};

// format flags
#define FtFlagFixedLen  0x01 // if clear, varying (may be fixed with FM)
#define FtFlagTH        0x02
#define FtFlagth        0x04
#define FtFlagAlwaysSet 0x80 // always set to avoid a null byte in fmt string


// part flags
#define fHH12     0x00000001
#define fHH24     0x00000002
#define fMI       0x00000004
#define fSS       0x00000008
#define fSSSS     0x00000010
#define fMERIDIAN 0x00000020 // needs fSSSS
#define fDATE     0x00000040 // gets MM DD and YYYY
#define fXX1      0x00000080

#define fYYY      0x00000100 // needs fDATE
#define fYY       0x00000200 // needs fDATE
#define fY        0x00000400 // needs fDATE
#define fERA      0x00000800
#define fMONTH    0x00001000 // needs fDATE
#define fDAY      0x00002000
#define fDDD      0x00004000
#define fXX2      0x00008000

//#define fD        0x00010000 // unused
#define fW        0x00020000 // needs fDATE
#define fWW       0x00040000 
#define fIW       0x00080000 // not supported
#define fCC       0x00100000 // needs fDATE
#define fJ        0x00200000 // not supported
#define fQ        0x00400000 // not supported
#define fXX3      0x00800000

#define fMS       0x10000000
#define fUS       0x20000000
#define fXX4      0x80000000

struct tsFmtToken {
    int ft_len;
    FtFM ft_FMable;
    FtTH ft_THable;
    FtSupported ft_supported;
    FtCase ft_CaseSensitive;
    FtFixed ft_fixed;
    int64 ft_parts;
    FtId ft_tokenId;
    const char *ft_tokenString;
    int ft_maxoutlen;
    tsFmtToken *ft_next;
};


// flags for host.gencLiteralTypeMask bit layout
#define fLit_int1 0x0001
#define fLit_int2 0x0002
#define fLit_int4 0x0004
#define fLit_int8 0x0008
#define fLit_num1 0x0010
#define fLit_num2 0x0020
#define fLit_num4 0x0040
#define fLit_bool 0x0080
#define fLit_flt4 0x0100
#define fLit_flt8 0x0200
#define fLit_ival 0x0400
#define fLit_ttz  0x0800
#define fLit_scmp 0x1000


#endif // #if !defined( INCLUDED_PCODEGEN )
