/*
 * Copyright IBM Corporation 2001, 2016.  All Rights Reserved.
 */

#ifndef PLAN_INCLUDED
#define PLAN_INCLUDED

/*
    Execution Plan object

When a record store is delivered to an execution plan, the plan executes
itself, sequencing on the directives for the record store.

*/

#include "nzstd.h"
#include "TXID.h"
#include "pcodegen.h"   // CPcodeGen used by value
#include "nzlog.h"
#include "messages.h"  // for ALL_SPUS_LID
#include "rcAtomic.h"
#include "rcBarrier.h"
#include "rcStringRef.h"
#include "scHistoryDataEnums.h" // for EPlanStatus
#include "nztimestamp.h"
#include "groupingsetinfo.h"
#include "groomstmt.h" // for GroomMode
#include "keyvalue.h" // for EZMCombiner
#include "sdlist.h"  // m_storeList types, methods
#include "fcommchannel.h"  // required before xpreturn.h
#include "xpreturn.h" // for ERetStrMode
#include "xpunload_enums.h"
#include "xpscan.h" // for EScanType
#include "xpdownload.h" // for EDownloadType
#include "xpdataslice.h" // for EDSlice types
#include "planfiles.h"
#include "vector" // std::vector
#include "BitVector.h"
#include "error.h"

#ifndef CHARSETMAPS_INCLUDED
# include "charsetmaps.h"
#endif

#if FOR_SPU
# include "mirrortopo.h"
# include "spujob.h"
#endif


class CGenCode;
class CFpgaWrap;
struct fpgaPortablePacket;
class TStreamInfo;
class CGenCode;
class PlanDictionary;
struct TMemblockInfo;
class TJoinNode;
class TMergeJoinNode;
class TProjectNode;
class CRecordStore;
class CCommBuf;
namespace mem { class Pool; };
class DSliceInteraction;
class COSInteraction;
struct client_result_msg_t;
struct client_comms_t;
struct pcFacts;
struct ExprContext;
struct SubPlan;
struct client_to_dbos_msg;
struct NumericVar;
struct Interval;
struct varlena;
class SnippetNode;
class CUnloadCtrl;
class TUnloadNode;
struct TZoneMapUnion;
struct TZoneMapConj;
struct TZoneMapExpr;
struct TInsertStatsNode;
class CWorkItem;
struct TtmpTableJoinInfo;
class TCreateTempTableNode;
struct TCreateTempTableMsg;
class TScanBlockCountNode;
class CExtTableInterface;
struct scan_state_t;
struct plan_shared_t;
class CNzSequence;
class CNzLibraryHelper;
namespace zm { class Accumulator; }
struct DataSliceFilter;
struct DataSliceHash;
struct Fuzzy_buffers;
struct Func;
struct List;
struct Node;
struct TNumericData;
struct TKeyValuePacket;
class  CFingerprint; // for Snippet Result Cache
class  CSnipResInfo; // for Snippet Result Cache

#if FOR_SPU
class SpuJob;
typedef SpuJob PlanJob;
struct GroomSpuState;
#else
class OsJob;
typedef OsJob PlanJob;
#endif

#define KEY_EXTRACTING  // define to enable key-extraction code
#define NullArg -1

/* Number of parameters allowed in a plan.  Verified in in ndelink.cpp */
#define MAX_PARAM_EXEC 200

enum EPlanTypes {
    PlanSelect,
    PlanInsertInto,
    PlanDelete,
    PlanUpdate,
    PlanMerge
};

class TCrossProductNode;
class TExprJoinNode;
class TExistsJoinNode;
class THashNode;
class TSortNode;
class TScanNode;
class THashJoinNode;
class TJoinNode;
class TUniqueNode;
class TSaveTempNode;
class TBindIndexNode;
class TDownloadTableNode;
class TDropTableNode;
class TLimitNode;
struct TAggrField;
struct TimeTzADT;
class TAggregateNode;
class TSortedAggregateClntNode;
class TInsertNode;
class TDeleteNode;
class TMergeNode;
class TStreamAppendClntNode;
class TTableSizeNode;
class TParserNode;
class TRestrictNode;
class TReturnNode;
class TMaterializeNode;
class THashPartNode;
struct TWindowAggrField;
struct TWindowFrame;
class TWindowAggrNode;
class TScanPrepNode;
class TUngroupNode;
class TDoneNode;
class TGroupingSetNode;
class TFuncJoinNode;
struct TFuncField;
class TReorgPrepClntNode;
class TReorgMergeClntNode;
class TReorgStatsClntNode;
class TReclaimClntNode;
class TDataSliceFilterNode;
class TDataSliceHashNode;
class TConditionalBranchNode;
class TBranchEndNode;
class TNext09Node;

namespace sc { class CHistoryData; };
namespace zm { class Column; };

Datum NzExecQual(List *qual,
                 ExprContext *econtextP,
                 CPlan *pPlan,
                 bool resultForNull,
                 bool forSubplan);

void
PgExprInitConfgFromPlan(CPlan *plan);

#define MAX_SUBPLAN_NESTING 32


enum EJoinQualType {
    joinQualSimple = 1, // all non-cartesian have this
    joinQualPostJoin,   // used by exists join
    joinQualOther       // used by hash, exists, and merge (?)
};

struct TSubplanSpec {
    Oid m_tableId;
    int m_sublinkType;
    CTable *m_table;
    void *m_record;
    bool m_isTableEmpty;
};

// ?UDF? start
enum EUdfType {
    udfCSource = 1,   // C source file
    udfCObject,       // C object file
    udfJavaClass      // Java Class file (byte codes)
};

#include "nzpack_x86_on.h" // packing on

struct UdfInfo {
    Oid       funcid;
    int       flags;
    EUdfType  type;
    char     *funcname;
    // !REWORK - returntype, arglist currently unused (tho set up).  Remove them and classname, funcsource?
    char     *returntype; // !NOTE-jpb  It may be better to make this and arglist be (lists) of type oids
    char     *arglist;
  //char     *classname;  // if function is a class member
  //char     *funcsource;
    int       retTypmod;
    int      *argTypmods;
    int       nargs;
    void     *infoPtr;
    bool      fenced;
    bool      nullcall;
    bool      determ;
    bool      stub;
    int       log;
    char     *memory;
    char     *hostcksum;
    char     *spucksum;
    char     *datadir;
    char     *name;
    int       snippetNum;
    int       numenvironment;
    void     *environment;
    int       numparameter;
    char    **parameters;
    int       version;
};

struct UdtfInfo : public UdfInfo {

    int joinType;
    int numFields;
    TFuncField *returns;
    bool lastcall;
    int correlated;
};

struct UdaInfo {
    Oid       funcid;
    int       flags;
    EUdfType  type;
    char     *funcname;
    // !REWORK - returntype, arglist currently unused (tho set up).  Remove them and classname, funcsource?
    char     *returntype; // !NOTE-jpb  It may be better to make this and arglist be (lists) of type oids
    char     *arglist;
  //char     *classname;  // if function is a class member
  //char     *funcsource;
    int       retType;
    int       retTypmod;
    int      *argTypmods;
    int      *intTypmods;
    int      *intTypes;
    int       nints;
    int       nargs;
    void     *infoPtr;
    void     *nodePtr;
    bool      fenced;
    bool      stub;
    int       log;
    char     *memory;
    char     *hostcksum;
    char     *spucksum;
    char     *datadir;
    char     *name;
    int       snippetNum;
    int       numenvironment;
    void     *environment;
    int       numparameter;
    char    **parameters;
    int       version;
    bool      grouped;
};

#define UdfInfoFlag_NewSource 0x01
#define UdfInfoFlag_InstanceRemovedByCSE 0x02
// ?UDF? end

struct PlanTimeStats
{
    // duration of postgres-side queuing
    uint64      pgQueueDuration;
    //timestamp when plan was submitted to parser.
    time64_t    parserSubmitTime;
    //timestamp when plan was submitted to analyze and rewrite phase.
    time64_t    rewriteSubmitTime;
    //timestamp when plan was submitted to analyze and rewrite phase
    time64_t    plannerSubmitTime;
    //timestamp when plan was submitted from postgres to DBOS
    time64_t    dbosSubmitTime;
    //timestamp when plan was added to Gatekeeper Prep
    time64_t    prepTime;
    //timestamp when plan was added to GRA Queue
    time64_t    graQueueTime;
    //timestamp when plan submitted to Snippet Scheduler
    time64_t    startTime;
    //timestamp when the snippet scheduler ran the first snippet
    time64_t    execStartTime;
    //timestamp when plan ended
    time64_t    endTime;

    // used for snippet time calculations
    time64_t    curSnippetStartTime;
    int64       workAtSnippetStart;

    int64       planWork;
    int64       planWorkMicros;

    int         maxHostMemBlocks;
    int         maxSpuMemBlocks;
    int         maxChannels;

    PlanTimeStats():
        pgQueueDuration(0),
        parserSubmitTime(0),rewriteSubmitTime(0),
        plannerSubmitTime(0), dbosSubmitTime(0),
        prepTime(0), graQueueTime(0), startTime(0),
        execStartTime(0), endTime(0), curSnippetStartTime(0),
        workAtSnippetStart(0), planWork(0), planWorkMicros(0),
        maxHostMemBlocks(0), maxSpuMemBlocks(0), maxChannels(0)
    {}

    uint64 getPGQDur() const {
         return pgQueueDuration;
    }

    uint64 getTXQDur() const {
         if(dbosSubmitTime == 0){
             return 0;
         }
         if (prepTime == 0){
             return time64() - dbosSubmitTime;
         } else {
             NZ_ASSERT(prepTime >= dbosSubmitTime);
             return prepTime - dbosSubmitTime;
         }
    }

    uint64 getPrepDur() const {
        if(prepTime == 0){
            return 0;
        }
        if (graQueueTime == 0){
            return time64() - prepTime;
        } else {
            NZ_ASSERT(graQueueTime >= prepTime);
                return graQueueTime - prepTime;
        }
    }

    uint64 getGRAQDur() const {
        if(graQueueTime == 0){
            return 0;
        }
        if (startTime == 0){
            return time64() - graQueueTime;
        } else {
            NZ_ASSERT(startTime >= graQueueTime);
            return startTime - graQueueTime;
        }

    }

    /**
     * This method ensures that all timing values get the current time assigned.
     * This is necessary in case of a system-generated sample scan where some
     * timing values like plannerSubmitTime do not get set. If that is the case,
     * this method sets them to the current time to keep reporting consistent.
     */
    void ensureSubmitTimesNonZero(){
        time64_t now = time64();

        if(parserSubmitTime == 0){
            parserSubmitTime = now;
        }

        if(rewriteSubmitTime == 0){
            rewriteSubmitTime = now;
        }

        if(plannerSubmitTime == 0){
            plannerSubmitTime = now;
        }

        // since GK has been replaced by the scheduler rules, set the
        // GK queue time to the same value as the submit time. LIMIT based
        // queuing is not performed in the GRA and must be measured there.
        if (!dbosSubmitTime){
            dbosSubmitTime = now;
        }
    }
};

#include "nzpack_x86_off.h" // packing off

enum PlanSessionVarType
{
    SESVAR_CAST_EMPTY_STR_ZERO,
    SESVAR_LEGACY_CAST_STR_TO_NUM,
    SESVAR_ANSI_SUBSTRING,
    SESVAR_FORCE_PG_EVAL,
    SESVAR_ENABLE_ISAO,
    SESVAR_STRICT_TEMPORALS_FORMAT,
    SESVAR_DATESTYLE_SQL_EURO,
    SESVAR_ENABLE_NEW_CASE,
    SESVAR_ENABLE_CASE_CSE,
    SESVAR_ENABLE_TIME24,
    SESVAR_ENABLE_NUMERIC_OPTIMIZATION,
    SESVAR_DISABLE_GENC_GCC_OPT,
    SESVAR_DISABLE_SRC,
    SESVAR_ENABLE_EXTEND_MAX_SEGMENT_SIZE,
    SESVAR_VALIDATE_DATASLICE_FILTER,
    SESVAR_DATESTYLE_ISO_YMD,
    SESVAR_ENABLE_SRC_INSTR,
    SESVAR_ENABLE_RANDOM_DIST_CHUNK,
    SESVAR_ENABLE_CPP_COLLECTION

};

enum ESnipResTtid {                 // Snippet Result Transient Table ID.
    eSnipResTtid_UNKNOWN = -1,      // have not checked current snippet yet
    eSnipResTtid_NONE = 0           // current snippet does not produce a result transient table.
                                    // >0 is the actual TTID.
};
// June 2012: table_id_t is int32 but the code base treats it as unsigned int sometimes,
// to allow more base tables.  See NZ_BASE_DBOS_TEMP_TABLE_ID in nzlimits.h and table.h.
// CPlan::m_SRC_candidate_TTID is only usable if it is *NOT* one of the special values in ESnipResTtid,
// and it *IS* a valid transient ID:
#define IS_USABLE_SRC_TTID(id)    ( (id != eSnipResTtid_UNKNOWN) && (id != eSnipResTtid_NONE) && (CTable::idIsTransient(id) ) )


typedef std::vector<
             table_id_t,
             NzShmAlloc<CSnipResInfo*, mt_src_info>::stl_alloc
        > SRC_TableIdVector;
typedef SRC_TableIdVector::iterator SRC_TableIdVector_iterator;

class CPlan {
public:

    // WARNING! The following members MUST BE FIRST.
    // They are replicated inline for generated code!
    SnippetNode **m_nodeArray;     // assembled from linked list
    volatile bool  m_limitReached;  // set true by the LimitNode, looked at by ScanNode & *Looper
    int m_oopsCount;    // number of records erroneously returned by FPGA
#if FOR_SPU
    TRecLoc m_curRecOrdLoc;   // only for virtual spus, loop entry addr
    PageId m_curTablePageId;        // Osr Page ID of the currently scanned table
#endif
    CTable *m_fabricTable;  // snippet response (same as interaction->m_Result)
    TxInfo m_xInfo;     // transaction info
    int m_sessionId; // session id
    int m_subplanTblCnt; // number of subplan tables
    TSubplanSpec *m_subplanSpecs;
    char **m_stringLiterals;
    char **m_clobLiterals;
    int *m_intLiterals;
    int8 *m_int8Literals;
    int16 *m_int16Literals;
    int64 *m_int64Literals;
    float *m_float4Literals;
    double *m_float8Literals;
    pcTTzInt *m_timetzLiterals;
    TNumericData *m_numeric4Literals;
    bool *m_boolLiterals;
    TSpuId m_dataSliceId;  // for gen'd code's support of PROJECT_SPUID field
    int m_dsliceIndex;  // on spu: spu-relative dataslice index
    volatile bool m_interruptRequested; // abort query now
    void (*m_nextGenFunc)(CPlan *plan, char *recordBuf, char *bufEnds, bool lastCall); //next function called by base GenPlanXXX function
    SnippetNode *m_pNextGenNode; //materializing node
    vlen_t **m_tempPointersP;  // pointers into big temp vlen_t buffer
    int m_tempsCount;          // temp vlen_t buffer count 
    nzotuid_t m_tempsOwner;      // pid of process that will de allocate m_tempPointersP
    CNzSequence **m_sequenceP;  // pointers to referenced sequence objects
    CUnloadCtrl *m_unload; // non-null if "unload-shunting" the results
    bool m_unloadAlterTblCompressedNoNull; // set if the compressed unload is on altered table with all fields as non nulls
#if FOR_SPU
    nzotuid_t m_exprPoolOwner;
#endif
    mem::Pool *m_exprPool;   // per-tuple allocation pool for expression evaluations
    void **m_udaObject;
    void **m_udfObject;     // holds pointers to instantiated UDF classes
    void **m_udtfObject;
    bool m_unTermFlg;
    bool m_unTermFlg2;

    vlen_t **m_charObjects;
    int     m_nCharCount;
    int     m_nCharCapacity;
    void **m_FuncObjects;
    int     m_nFuncCount;
    int     m_nFuncCapacity;
    void **m_AggObjects;
    int     m_nAggCount;
    int     m_nAggCapacity;
    void **m_TableFuncObjects;
    int    m_nTableFuncCount;
    int    m_nTableFuncCapacity;

    /* Note about m_reserved:
     * LogGeneric() can be used to instrument generated code and collect
     * the results in a virtual table. m_reserved is useful to maintain
     * state across the calls to generated code. See _vt_reserved and
     * LogGeneric().
     */
    int m_reserved[16];

    // End of special generated code visible member data

    CNzLibraryHelper *m_libHelper;
   //for extending max segment size in codefragment
    int m_extend_max_segment_size ;
    bool m_udxStub;
    bool m_udxStubIndividual;
    int m_nCharReserved;
    int m_random_dist_chunk_size;

    int m_sizeNodeArray;    // number of members in nodeArray

    int m_snippetNum;   // counter for which snippet this is
    int m_totalSnippets;

    int m_numScanPasses;   // For progress tracking. How many scans; 1 for initial scan, +1 for sort, +n for DHJ, ...
    int m_curScanPass;     // For progress tracking. Current scan, 1..m_numScanPasses; -1 means unknown.
    int m_dataTableScanPass; // For progress tracking. Which scan pass is the real table scan pass?
    double m_curScanPassProgress; // For progress tracking. 0.0 .. 1.0. -1 means unknown.
    int getCurScanPass() const { return m_curScanPass; }
    int getDataTableScanPass() const { return m_dataTableScanPass; }
    double getCurScanPassProgress() const { return m_curScanPassProgress; }

    scan_state_t *m_scanState; // SPU: scan state

    plan_shared_t *m_plan_shared;

#if FOR_SPU
    zm::Accumulator *m_zmAccumSeq[PAGES_PER_EXTENT]; // Generate statistics: zone map accumulators
    PageId m_accumulatingPageId[PAGES_PER_EXTENT];   // Generate statistics: pageId for each entry in m_zmAccumSeq
    int m_curAccumIx;                                // Generate statistics: current index into the arrays above.
#endif

    int m_blocksInRecovery; // number of blocks in recovery for this plan
    CTable *m_planStats;    // table that holds per-node exec-time statistics

    bool m_txRegistered;    // set iff registerPlan() succeeded
    bool m_usedSwTxFiltering;

    bool m_closeCbWillNotify; // true => close channel callback will notify scheduler of plan completion

    bool m_enableZMapDbg;//for passing session variable to certain spu-side VTs
                         //in CPlan::LoadCommBuf (ZMap Dbg Info)

    int m_groomMode;        // 0 for non-groom plans; GroomMode for groom plans
    TXidValue m_stableTXid; // for groom plans

    void *m_pgPlan;   // top level postgres plan (needed by ndelink only)
    unsigned int m_tickCount; // for GENC logging
    pcFacts *m_pPcodeFacts;   // pcode facts

    //!NOTE-dk need a more effective scheme.  But what abort machinery, etc other
    //than the CPlan destructor do we have with the event-loop architecture?
    dlink_t m_storeList;    // List of "active" CRecordStore (was: CSwapStore)

    // execution image for this snippet
    int (*m_genCodeEntry)(CPlan *plan, char *recordBuf, char *bufEnds, bool lastCall);
    void (*m_genCodeUdfCleanup)(CPlan *plan);
    void *m_genCodeBase;  // base address of loaded generated code
    int m_genCodeSize;  // bytecount of m_genCodeBase
    bool m_genCodeLoaded;  // true after generated code fully loaded

    // support for debugging generated code
    bool m_bLoggedEval;      // set to true when snippet excessive expr EVAL recursion induces warning printf

    //used by PCSampler to find location of the .cpp and .o files in the cache
    int m_genCodeDirIdx; // if >=0 then directory where code was cached
    int m_genCodeFileIdx;

    int  m_rollbackIntId;    // set when the rollback interaction is actually allocated  MBC

    dlink_t m_cleanupList; // list of cleanup callbacks
    nzotuid_t m_dlopenOwner;      // pid (dbosDispatch or dbosEvent) of process
                            // that should unload the machine code

    CTable *m_resultTableToDestroy;  // Interaction handed result to plan; plan responsible for destruction
    int m_channelOpenItpCount;

    CBitVector m_dataSlicesToRunOn;
    CBitVector m_dataSlicesToRunOrig;
    std::vector<int,
                NzShmAlloc<int, mt_plan>::stl_alloc
                > droppedTableIdsForThisSnippet;
    std::vector<TSaveTempNode*,
                NzShmAlloc<TSaveTempNode*, mt_plan>::stl_alloc
                >createdTransientsForThisSnippet;


    // ============= Methods ==============

    void *operator new(size_t);
    void operator delete(void*);

    CPlan( // Create
        EPlanTypes type, // PlanSelect,PlanInsertInto,PlanDelete,PlanUpdate...
        const TXid& xId,
        const SBCSmap *CSmapP);

    CPlan( // Create plan object from a Comm buffer
        int clientId,       // clientId to associate this plan with
        int cliInteractId,  // client request number (only relevant if DBOS)
        int dsid,           // datasliceid (spu only)
        const void *marshalledData, // data in IO buffer
        size_t lenMarshalledData,  // size of marshalledData
        mem::Pool *pool);

    ~CPlan();

    void GenCodeLoop(bool lastCall);

    void ExecGenCode(    // Run the generated code against the passed record buf
        char *pBufStart,        // first record
        char *pBufFence,        // byte after last record
        bool lastCall,          // true if last call - no data at bufXXX
        bool forHost,           // just used in logging logic
        bool logIt);            // logging logic

    static const char *NodeTypeStr( //return a string representation for node type
        ENodeTypes nodeType);
    static void BufferAllowanceCalculator(void);

    typedef void (cleanup_callback_t)(const void *, size_t size);
    void *RegisterCleanupCallback(cleanup_callback_t *callback,
                                  const void *cb_data, size_t cb_data_len);
    void UnregisterCleanupCallback(void *cookie) throw ();
    void doCleanupCallbacks() throw ();

    TSortNode *AddSortNode( // Create Sort Node into plan
        EPlanNodeLocus place,
        int numFields,
        int *fieldList, // must stay stable until SendToNDE()
        int *ascendingList,
        int *nullsLoList,
        int mergeType = FALSE,
        SnippetNode *addPos = NULL);   // optional insert (after) position

    TMaterializeNode *AddMaterializeNode( // Create Materialize Node into plan
        EPlanNodeLocus place);

    TRestrictNode *AddRestrictNode( // Create Restrict Node into plan
        EPlanNodeLocus place,
        List *qual,
        bool bForHaving = false,
        ExprContext *econtext = NULL,
        int pcodeStart=0,           // number of first pcode instruction
        int pcodeCount=0);           // pcode instruction count

    TJoinNode *AddJoinNode( // Create Join Node into plan
        EPlanNodeLocus place,
        int tblId,
        int tblField,
        int joinField,
        int joinType,
        List *joinQual);    // NULL if simple col==col

    TSaveTempNode *AddSaveTempNode( // Create SaveTemp Node into plan
        EPlanNodeLocus place,
        int  tblId,           // temp table id
        bool transient,       // whether to destroy when plan deleted
        bool append,          // whether to append to existing tblId
        bool saveInFpgaStyle=false, //true to save in fpga style format
        bool usedInMultipleJoins=false, //true if used in multiple joins
        bool isPrebroadcastTbl=false,
        bool isBlockList=false);   // true if temp table for mview pivoting

    TProjectNode *AddProjectNode( // Create Project Node into plan
        EPlanNodeLocus place,
        int numFields,
        int *fieldsList,   // must stay stable until SendToNDE()
        Node **exprList=NULL,   // expressions only if fieldList[ix] == -1
        int *varStrSizeList=NULL, // for varchar expr result sizes
        int pcodeStart=0,           // number of first pcode instruction
        int pcodeCount=0,           // pcode instruction count
        int physIDCount=0,            // number of entries in physIDs
        int *physIDs=0,              // phys IDs for project table when done by fpga
        bool handledByFpga=false,     // temp, real code prunes node out
        SnippetNode *addPos=NULL,   // optional place to insert node
        char **dbgColNames=NULL,    // optional - colnames for dbos plan dumper
        int reproject=0,            // base off previous stream
        int projectFlags=0,        // m_projectFlags member
        bool suppressMaxCheck=false, // This should only be done for table
        bool useStruct=false,       // funcs for now. May be doable
                                     // everywhere eventually.
        bool forFuncJoin=false);

    // Removes a project node that was created by the fpga
    // see comments in CFpgaPrep::compensateForModifications
    void RemoveFpgaCreatedProjectNode(TProjectNode* pNode);

    TScanNode *AddScanNode( // Create Scan Node into plan
        // Client -> DBOS -> SPU values
        EPlanNodeLocus place,
        table_id_t tblId,   // table to scan
        int flags=SfNone,   // scan control flags
        EPlanNodeTableType tableType = tableAtSPU,
        int spuId=ALL_SPUS_LID,      // for single spu queries (?? ever used)
        EScanType scanType=eScanTypeBoth, // read ordered stuff, new stuff or both
        int blockCost=24,            // 128KB blocks to scan (via stats, optimizer)
        int blockCostSource=eBlockCostFromPgPlanner,
        int estNumSlices=0,
        // not client, DBOS created to SPU values (clients all default)
        fpgaPortablePacket *fpgaMicrocode=0, // fpga's setup bytes
        TKeyValue *pKeyValues=0,     // key values for index scan
        int numKeyValues=0,          // number thereof
        TZoneMapUnion *pZoneMapInfo=0, // for zone-map restricted scans
        int XIDMode=0,               // fpga:EXIDMode
        int totalXIDs=0,             // transaction visibility list info
        int invisibleXIDs=0,
        TXid *xIds=NULL,
        double sampleGrowthPower =0.0,
        table_id_t indexTable=0,     // table id holding blockList
        TUseSpecialFields *extraSpecialsP=0);

    TReturnNode *AddReturnNode( // Create Return Node into plan
        EPlanNodeLocus place,
        int shapeTable=0,   // deprecated, we always derive this
        int batchSize=0,    // if >0, size before returning
        int forBackup=false,    // true if no host-side plan
        ERetStrMode retStrMode=RetRecords, // non-zero for formatted strings method
        SnippetNode *pBefore=0);

    TReturnNode *AddReturnNode( // Create Return Node into plan
        EPlanNodeLocus place,
        int batchSize,          // if >0, size before returning
        int forBackup,          // true if no host-side plan
        ERetStrMode retStrMode, // non-zero for formatted strings method
        int numPackets,       // number of send-side packet buffers
        int sizeReassembly,   // maximum size of reassembly buffers
        int numReassembly,    // number of reassembly buffers
        TChannel *channel=NULL, // Fcomm channel
        SnippetNode *pBefore=0);   // allows insertion for plan changes

    TUniqueNode *AddUniqueNode(EPlanNodeLocus place);

    TAggregateNode *AddAggregateNode( // Create Aggregate (groupby) node in plan
        EPlanNodeLocus place,
        int numGroups,          // how many grouping (not projecting) fields
        int *inputGroupingList, // grouping fields referring to input stream
        int numAggrs,           // how many aggregating fields
        TAggrField *grps,       // the grouping field info type/expr/distinct...
        int aggrStyle=0,        // residual support for obsolete aggr scheme
        int groupingSetID=-1,   // handle to relevant grouping set id
        bool handledByFpga=false,    // temp, real code prunes node out
        int *ordercolsList=0,   // to force aggr-result order other than by group
        int *ascendingList=0,    // ascending/descending flags for ordercolsList
        int *nullsLoList=0,      // null collation flags for ordercolsList
        int numFilters=0,       // number of aggregate filters
        Node **filterList=NULL, // array of aggregate filters
        int pcodeStart=0,           // number of first pcode instruction
        int pcodeCount=0,           // pcode instruction count
        bool forExpressStats=false,
        uint64 approxRows=0,
        bool bRollup=false);

    TSortedAggregateClntNode *AddSortedAggregateNode( // Create SortedAggregate (within group ( order by ..) ) node in plan
        EPlanNodeLocus place,
        int numAggrs,
        TAggrField *aggrs,
        int numGroups,
        int *groupingList,
        int numOrderCols,
        int *orderColList,
        int *ascendingList,
        int *nullsLoList );

    TInsertNode *AddInsertNode( // Create Insert Records Node into plan
        EPlanNodeLocus place,
        table_id_t tblId,       // which table to insert records into
        unsigned int insertFlags = 0); // Insert flags which provide hints to insert logic

    TDeleteNode *AddDeleteNode( //create delete node in plan
        EPlanNodeLocus place,
        int tblId);

    TDownloadTableNode *AddDownloadTableNode( // broadcast temp table
        EPlanNodeLocus place,
        EDownloadType distMode, // broadcast/distribute
        int linkId,   // temp table id to use in SPU
        int numDistKeys=0,    // how many distKey fields (MAX_DIST_KEY_FIELDS)
        int *distKeys=NULL,   // array of distKeys
        int spuId=ALL_SPUS_LID,   // for non-broadcast target
        int keepStore=FALSE,  // if true keep store on dbos after download
        bool usedInMultipleJoins = false,   //true if used in multiple joins
        bool isPrebroadcastTbl = false,
        unsigned int downloadTableFlags = 0, // flags which provide hints to logic
        int compressed=FALSE,   // where to compress
        TChannel *channel=NULL, // spu2spu: Fcomm channel
        int interactionId=0,    // spu2spu: download interaction id
        int syncId=0,           // spu2spu: sync interaction id
        int numPackets=0,       // spu2spu: number of send-side packet buffers
        int sizeReassembly=0,   // spu2spu: maximum size of reassembly buffers
        int numReassembly=0);    // spu2spu: number of reassembly buffers


    TDropTableNode *AddDropTableNode(
        EPlanNodeLocus place,
        int linkId);  // temp table id to use in SPU

    TDoneNode *AddDoneNode (
        EPlanNodeLocus place,
        int numPackets=0,
        int sizeReassembly=0,
        int numReassembly=0,
        TChannel *channel=NULL,
        SnippetNode *pBefore=0 );

    TExistsJoinNode *AddExistsJoinNode( // existance joining
        EPlanNodeLocus place,
        int tblId,          // already existant temp table
        int notJoin,        // if TRUE, is a non-existance check
        int alreadySorted,  // if TRUE, don't need to sort the temp table
        int numFields,      // how many refs in tblField/joinField
        int *tblField,      // field in temp table
        int *joinField,     // field in pipeline table
        List *joinQual,     // if not null, use pg expression evaluator
        int leftMode,       // if TRUE, return 'temp' recs rather than 'stream' recs
        List *nullTestQualInner=NULL, // null test qual for not exists joins (inner)
        List *nullTestQualOuter=NULL, // null test qual for not exists joins (outer)
        int joinFlags=0,    // join flags
        bool collectstats=false,
        int64 tempTblSize=0,
        int64 tempTblNumRecords=0);

    THashNode *AddHashNode(EPlanNodeLocus place,
        int numFields,      // how many fields
        int *fieldsList,    // field nums
        int tblId=0,        // table to pick fields from.  0 means stream
        int hashFlags=0);   // extra hints to dbos/spu

    TCrossProductNode *AddCrossProductNode(EPlanNodeLocus place,
        int tblId);          // already existant temp table

    TLimitNode *AddLimitNode(EPlanNodeLocus place,
        const int64& limit,    // max # rows
        const int64& offset);  // limit beginning from row #

    THashJoinNode *AddHashJoinNode(
        EPlanNodeLocus place,
        int tblId,          // already existant temp table
        int numHashCols,    // number of colums to hash
        int *tmpHashCols,   // columns to hash for temp table
        int *streamHashCols,// columns to hash for stream
        List *joinQual,     // post join qual
        List *otherQual,    // post join qual
        int joinType,       // join operation (inner/outer/both...)
        int pcodeStart,
        int pcodeCount,
        int pcodeStartOtherQual,
        int pcodeCountOtherQual,
        bool collectstats=false,
        int64 tempTblSize=0,
        int64 tempTblNumRecords=0);

    TMergeJoinNode *AddMergeJoinNode(EPlanNodeLocus place,
        int tblId,          // already existant temp table
        int tblField,       // field in temp table
        int joinField,      // field in pipeline table
        int joinType,       // join operation
        List *joinQual,     // NULL if simple col==col
        List *otherQual,    // post join qual
        bool preSorted);    // true if temp table already sorted

    TExprJoinNode *AddExprJoinNode(EPlanNodeLocus place,
        int tblId,          // already existant temp table
        List *joinQual,     // PG expression for matching
        List *otherQual,    // post join qual
        int joinType,       // join operation
        int pcodeStart,     // number of first pcode instruction
        int pcodeCount);    // number of pcode instructions

    TBindIndexNode *AddBindIndexNode(EPlanNodeLocus place,
        int id,                // table to convert to index status
        int indexType,         // NZ_INDEX_XXX
        int parentId,          // base table
        int numIndexFields,    // number of order-by fields
        int *indexFieldsList,  // order by
        int *indexAscendList,  // sort up?
        int *indexNullsLoList, // null collation
        int numProjectFields,  // thin-table column rowcount
        int *indexProjectList, // thin-table column subset
        int sortType);

    TMergeNode *AddMergeNode(  //merge node for thin tables
        EPlanNodeLocus place,
        int tblId,              //table contained new stuff
        int nCols,              //number of order by cols in thin table
        int *pCols ,            //array of order by cols in thin table
        int *pAsc,              //asc or desc indicator list
        int *pNullsLo);         //nulls collation indicator list

    TStreamAppendClntNode *AddStreamAppendNode(  //for union all
        EPlanNodeLocus place,
        int tblId);

    TTableSizeNode *AddTableSizeNode( //add the table size node
        EPlanNodeLocus place,
        int tblId,              //table id in question
        bool getSize,           //true to get size, false to set size
        bool forMergeJoin=false,//true if table is for MergeJoins
        int  mergeCol=-1);      //mergeCol if forMergeJoin

    TParserNode *AddParserNode( // add record parser node
        EPlanNodeLocus place,
        const TParserNode *source); // optional source template

    TUnloadNode *AddUnloadNode(
        EPlanNodeLocus place,       // must be inDBOS
        EUnloaderMode mode,         // file/socket...
        const char *filename,       // name of file/socket/program
        bool mlsTable,
        const int64& tableSource,   // source query contains a table modeled as a view if this is non-zero.
        char delim,                 // character that seperates fields
        bool trailingDelim,         // true if there is a delim before the newline
        char dateDelim,             // char between mon, day, year
        char dateTimeDelim,         // char between date and time
        char timeDelim,             // char between hour, min, sec
        char meridianDelim,         // char between sec and meridian
        char tokEscape,             // if non-zero, char to use as escape char (usually \)
        ParserBools boolStyle,      // ParserBool_1_0, ParserBool_T_F, ...
        ParserDates dateStyle,      // ParserDate_YMD, ParserDate_MDY, ...
        ParserTimes timeStyle,      // ParserTime_12, ParserTime_24
        EUnloaderFormat dataFormat, // UnloadFormatText...
        const char *nullValue,            // null string
        bool crInString,            // accept unescaped CR if it's true
        bool includezeroseconds,    // true if 00 to be included for seconds
        char decimalDelim,          // decimal delimiter
        uint64 recordDelim,         // record delimiter for load/unload
        bool includeheader,         // add column header in unloaded file
        int relId = 0,              // relId of the external table
        int timeout = 0);           // timeout for load/unload

    TUnloadNode *AddUnloadNode(
        EPlanNodeLocus place,       // we'll require inDBOS
        CUnloadCtrl *unload,        // ptr to unload ctrl
        EUnloaderMode mode,         // file/socket...
        const char *filename,       // name of file/socket/program
        const int64& tableSource,   // source query contains a table modeled as a view if this is non-zero.
        bool mlsTable,
        int timeout = 0);           // timeout for load/unload

    TInsertStatsNode *AddInsertStatsNode(
        const EPlanNodeLocus place,
        const bool consolidating,           // host consolidation of spu stats
        bool noMinMax = false,              // not tracking min/max (e.g. for thintable or reclaim)
        const bool forReclaim = false,      // inserting for reclaim
        const bool forMerge = false);       // stats for MERGE statement

    THashPartNode *insertHashPartNode( EPlanNodeLocus place,
                                       SnippetNode *addPos = NULL );
    bool insertHashPartNodes(SnippetNode *pFirst, SnippetNode *pFence);

    TWindowAggrNode *AddWindowAggrNode(EPlanNodeLocus place,
                                       int numAggrs,            // number of aggregates
                                       TWindowAggrField *aggrs, // list of agrregates
                                       int numPartitionCols,    // number of partition columns
                                       int *partitionColList,   // list of partition columns
                                       int numOrderCols,        // number of order columns
                                       int *ordercolsList,      // list of order columns
                                       int *ascendingList,      // list of order direction flags
                                       int *nullsLoList,        // list of null collation flags
                                       int numFrames,           // number of frames in partition/order window
                                       TWindowFrame *frames);   // list of window frames

    TScanPrepNode *AddScanPrepNode(EPlanNodeLocus place,
                                   TScanNode *pRef,
                                   SnippetNode *addPos=NULL);

    TCreateTempTableNode *AddCreateTempTableNode(EPlanNodeLocus place,
                                      TDownloadTableNode *pDown,
                                      TCreateTempTableMsg *pMsg,
                                      int msgSize,
                                      SnippetNode *addPos=NULL);

    TScanBlockCountNode *AddScanBlockCountNode(EPlanNodeLocus place,
                                      TScanNode *pRef,
                                      bool getSize,
                                      SnippetNode *addPos=NULL);

    TUngroupNode *AddUngroupNode(EPlanNodeLocus place, int countFldDex);

    TGroupingSetNode *AddGroupingSetNode(EPlanNodeLocus place,
                                         int groupingSetID);

    int AddGroupingSetInfo(
        int numGroups,               // number of group columns
        int *groups,                 // array of source group col numbers
        int numGrpSets,              // number of grouping sets
        unsigned char **grpVectors,  // array of grouping set vector pointerss
        int numGroupingArgs,         // number of distinct grouping() args
        int *groupingArgList);       // array of m_groups entry numbers
    TFuncJoinNode *AddFuncJoinNode(EPlanNodeLocus place,
                                   bool correlated,
                                   int joinType,
                                   int numargs,
                                   int *argcols,
                                   int numfields,
                                   TFuncField *fieldList,
                                   Func *pFunc,
                                   List *qual,
                                   bool over,
                                   bool order,
                                   bool partition,
                                   int numPart,
                                   int *partCols,
                                   int numNulls,
                                   bool *nulls
                                   );

    TReorgPrepClntNode *AddReorgPrepNode(
        int nKeys,
        int tempTableId,
        int *normAlgs,
        int *normParams,
        const uint64& maxSpuMem);

    TReorgMergeClntNode *AddReorgMergeNode();
    TReorgStatsClntNode *AddReorgStatsNode(const int64& groomId);

    TReclaimClntNode *AddReclaimNode(TXidValue stableXid,
                                     GroomMode groomMode,
                                     uint32 reclaimFlags);

    TDataSliceFilterNode *AddDataSliceFilterNode(EPlanNodeLocus place,
                                                 EDSliceFilterType filterType,
                                                 List *baseFilterValues,
                                                 Oid tblOid,
                                                 SnippetNode* pAfter=NULL);



    TDataSliceHashNode *AddDataSliceHashNode(EPlanNodeLocus place,
                                             int numFields, int *fieldsList,
                                             Oid tblOid);

    TConditionalBranchNode *AddConditionalBranchNode(EPlanNodeLocus place, List *qual, bool isElse = false);
    TBranchEndNode *AddBranchEndNode(EPlanNodeLocus place);

    TNext09Node *AddNext09Node(EPlanNodeLocus place);

    void AddStatsTable(); // Create the query stats reciever table

    void appendNodeStats(); // Add node execution time/size stats to m_planStats
    CTable *GetNodeStatsTblIfEnabled(); //get plan node stats table

    void AddUnloadInfo( // supplies ascii-unload control info
        CUnloadCtrl *info);  // mode, filename, delims...

    void executePlanInit();  // Prepare nodes (SPU)
    void closePlanData(); // disconnect from data storage (SPU)

    EPlanNodeLocus ScanLocus() const;

    int GetSnippetNum() const {
        return m_snippetNum;
    }

    int doneSnippets() const {
        return m_snippetNum > 0 ? m_snippetNum - 1 : 0;
    }

    int totalSnippets() const {
        return m_totalSnippets;
    }

    void GetSpecialsForThisPlan(
        TUseSpecialFields *pSpecials,
        TScanNode *pNode);

    void LoadCommBuf( // Load passed comm buffer with plan bytes
        CCommBuf *cbuf,             // buffer to create
        SnippetNode *firstNode,    // where to start marshalling (inclusive)
        SnippetNode *stopNode,     // where to stop marshalling (exclusive)
        void *objectCodeP,      // machine code image
        int objectCodeSize);    // bytecount of machine code image

    void addQualToCommBuf( // Add a qual list, used for restrict/join/exists...
        CCommBuf *cbuf, List *qual, ExprContext *econtext, int *invalidFuncP);

    List *getQualFromCommBuf( // Extract a PG qual list item from comm buf
        CCommBuf *cbuf,
        const bool restoreMode = false);

    void addProjectionFieldsToCommBuf(CCommBuf* const cbuf,
                                      TProjectNode* const pProjectNode,
                                      const bool adjustSrcInvalidFunc = false,
                                      const int srcCode = 0,
                                      const int srcAux = 0);
    void getProjectionFieldsFromCommBuf(CCommBuf* const cbuf,
                                        int* const numFields,
                                        int** const fieldsList,
                                        Node*** const exprList,
                                        const bool restoreMode = false);

    void createCode( // Generate machine code buffer
        const TableShape *entryTable, // null for spu, m_Result for dbos
        SnippetNode *firstNode,    // where to start marshalling (inclusive)
        SnippetNode *stopNode,     // where to stop marshalling (exclusive)
        void **objectCodeP,      // machine code image
        size_t *objectCodeSize);    // bytecount of machine code image

    void SpuBindObjectImage( // load to memory, set plan members
        const void *objectCodeP,       // machine code image
        size_t objectCodeSize);     // bytecount of machine code image

    void SendToNDE(client_result_msg_t *ack);

    void dumpPlanBuf(const char *commbuf, size_t len, int planid);

    bool ExtractKeyValues(
        TKeyValuePacket **ppPacket,
        TScanNode *pScanNode,   // ScanNode at start of snippet
        SnippetNode *pLastSnippetNode );

    void SpuDoneReporting(  // Notification of spu finishing its snippet
        int dsId);         // valid range from 1 to Nz->numSPUs

    void setBaseSlicesToRunOn(TDataSliceFilterNode *DSFN, bool* isInitialized);
    void setJoinSlicesToRunOn(TDataSliceFilterNode *DSFN, bool* isInitialized);
    void setSlicesToRunOn();

    COSInteraction* singleSliceAccountingInteraction();

    void Dump(const char *title=NULL,int detail=NzDumpDetail, FILE *fp=NzDumpFile, int indent=NzDumpIndent) const;
    void dump() const;   // For DBG_DUMP_OBJ ()
    void Dumper() const;  // for debugger access
    void dumpSnippet(SnippetNode *first, SnippetNode *last, int detail, FILE *fp, int indent,
             bool skipcomment = false, bool wholePlan = false) const;
	void SrcCmprSnippetResult(int snippetNumToDump) const;             
    void dumpNode(SnippetNode *pNode, int detail, FILE *fp, int indent, int snippetNum, int nodeNum,
          bool skipcomment = false) const;
    void dumpPcode() const;

    SnippetNode *NextNode(SnippetNode *fromNode) const {
        return fromNode ? fromNode->m_next : m_nodeList;
    }

    void ProcessScanTable(bool includeLastCall=true);
    void ProcessPrepScan();

    //remove record stores for all nodes before this node
    void ReleaseRecordStores( SnippetNode *fence );
    void ReleaseStoreForTable(CTable *pTable, bool willReuseRecordStore); //release store for this table

    void DbosNotifySrcOnSnippetDone();
    void DbosProcess(   // Run the generated code on passed mesage
                     TMemblockInfo *block);  // data from spus/loader
    void DbosProcess(   // Run the generated code on table's data
                     CTable *table);     // data from loader/virtab

    void DbosPrepScanProcessDone();
    void DbosProcessDone();
    void DbosSnippetDone();
    void DbosLastSnippetDone();
    void DbosLogPlanDone();

    void DbosNotePlanAbort( const CError& err, int dataslice ) throw ();

    void DbosNewPlan();     // once-per-plan init in dbos

    bool DbosHostSetup();   // host-only snippets
    bool DbosHostAlloc();
    bool DbosHostExec();

    // host control work on behalf of spu snippets
    bool DbosSpuSetup();
    void DbosEstimateSnippet();
    void DbosAllocateSnippetResources(SnippetNode *first, SnippetNode *fence);
    bool DbosHostAllocateSnippetResources(SnippetNode *first, SnippetNode *fence);
    void DbosSpuDataExec();
    void DbosPrepScanExec();
    unsigned DbosGetPrepScanTotal() const;
    void DbosNewSnippet(SnippetNode *firstNodeInSnippet);
    void correctEstimatesIfDRO();

    bool DbosIsPlanMemoryOnly(int maxMemTables) const;

    bool cacheSnippetResult();

    bool isAllDBOS() const;
    bool hasParserNode() const;
    bool hasUnloadNode() const;

    bool hasDownloadTableNode() const;
    bool hasInsertNode() const;

    bool scansTable(Oid tableOid) const;

    bool hasSnippetNode(int, ENodeTypes) const; // is a node of specific type contained in a given snippet

    bool hasCurrentSnippetNode(ENodeTypes type) const; // is a node of specific type contained in the current snippet

    void getKeyword(const char *, char *, int) const; // generate a keyword for plan's SQL (SEL, UPD, ...)

    bool isPrebroadcast() const; // is this a pre-broadcast plan

    int getPriorPlansNum() const; // get the number of prior plans (including jit plans)

    bool exprNullable(
        TProjectNode *pProjector,
        int exprNo);

    unsigned int pagesToBeScanned() const; // total pages in base table after zm benefit
    unsigned int pagesToBeScannedBeforeJoin() const; // total pages in base table after zm benefit
    unsigned int pagesSampled() const;     // num pages sampled during jitstats
    double extentsToBeScanned() const; // total extents in base table after zm benefit
    double extentsToBeScannedBeforeJoin() const; // total extents in base table after zm benefit
    double extentsSampled() const;     // num extents sampled during jitstats
    void allocateUdfStorage(int nCount);
    void allocateUdaStorage(int nCount);
    void allocateUdtfStorage(int nCount);
    vlen_t *allocateCharBuf(int size, int id);
    void *allocateFuncInfo(int id, int ver);
    void *allocateAggInfo(int id, int ver);
    void *allocateTableFuncInfo(int id, int ver);
    void setInsertTypeInfo(bool);

    void SetSubplanTableCount (int cnt);
    void SetSubplanSpec (int specnum, Oid tableOid, int type);

    static bool IsSchedEventLogOn();
    void LogSchedEvent(         // Add an event to the event log virtab
            int16 id,           // Scheduler event id from nzsched_events.h
            int64 time,  // Time, in ticks, event occured (zero: get it please)
            int64 more1, // Event-specific information (or zero)
            int16 more2, // Event-specific information (or zero)
            int32 tableId=0,      // Table ID, 0 if NA
            int64 data1=0,        // Event-specific information
            int64 data2=0,        // Event-specific information
            int64 data3=0,        // Event-specific information
            int64 data4=0,        // Event-specific information
            int64 data5=0,        // Event-specific information
            int64 data6=0,        // Event-specific information
            int64 data7=0,        // Event-specific information
            int64 data8=0) const; // Event-specific information
    static void SchedEventLogger( // Add an event to the event log virtab
            int16 id,           // Scheduler event id from nzsched_events.h
            int spuId,          // Dataslice id of the execution plan
            const TXid& xid,    // Transaction id
            int planId,         // Plan id
            int snip,           // Snippet number within plan
            int64 time,         // Time, in ticks, event occured
            int64 more1,        // Event-specific information
            int16 more2,        // Event-specific information
            int32 tableId=0,    // Table ID, 0 if NA
            int64 data1=0,      // Event-specific information
            int64 data2=0,      // Event-specific information
            int64 data3=0,      // Event-specific information
            int64 data4=0,      // Event-specific information
            int64 data5=0,      // Event-specific information
            int64 data6=0,      // Event-specific information
            int64 data7=0,      // Event-specific information
            int64 data8=0);     // Event-specific information

    //note: m_planTimeStats.m_submitTime is set in setStmtTime()
    void markPrepTime()
    {
        m_planTimeStats.prepTime = time64();
    }

    void markGraQueueTime()
    {
        m_planTimeStats.graQueueTime = time64();
    }

    //slightly more complex and has side effects, so not inlined
    void markGraStartTime();

    void markAbortTime()
    {
        m_planTimeStats.endTime = time64();

        // back fill timestamps not yet set
        if (m_planTimeStats.execStartTime == 0) {
        	m_planTimeStats.execStartTime = m_planTimeStats.endTime;
        }
        if (m_planTimeStats.startTime == 0) {
        	m_planTimeStats.startTime = m_planTimeStats.endTime;
        }
        if (m_planTimeStats.graQueueTime == 0) {
        	m_planTimeStats.graQueueTime = m_planTimeStats.endTime;
        }
        if (m_planTimeStats.prepTime == 0) {
        	m_planTimeStats.prepTime = m_planTimeStats.endTime;
        }
    }

    void markEndTime()
    {
        m_planTimeStats.endTime = time64();
    }

    void markSnippetStartTime(const time64_t& snippetStartTime, const int64& planWork, bool isPrep)
    {
        //do nothing if we already have a start time
        if (m_planTimeStats.curSnippetStartTime != 0)
            return;

        // Update plan exec start if this is the first non-prep snippet
        if (m_planTimeStats.execStartTime == 0 && !isPrep) {
            m_planTimeStats.execStartTime = snippetStartTime;
        }

        m_planTimeStats.workAtSnippetStart = planWork;
        m_planTimeStats.curSnippetStartTime = snippetStartTime;
    }

    //Get a const reference to the time stats of this plan.
    const PlanTimeStats& getTimeStats() const
    {
        return m_planTimeStats;
    }

    // return max and avg transient result table pages and rows
    void getTransResultTablePageAndRowCounts(float* avgPages, float* maxPages, float* avgRows, float* maxRows) const;

    /////////////////////////////////
    /// Tell the plan to open the return data channel to the host
    /// at the start of every snippet. Used for reclaim progress
    /// reporting and intermediate resource accounting.
    /// @param val true means open the channel when snippet starts.
    ///            false means open the channel whenever gencode
    ///            feels like it.
    /////////////////////////////////
    void setOpenChannelUpfront(bool val)
    {
        m_openChannelUpfront = val;
    }

    /////////////////////////////////
    /// Inquire whether this plan opens the return data channel
    /// at the start of every snippet, or waits until gencode decides to.
    /// @return true if channel is opened at start of snippet.
    ///         False otherwise.
    /////////////////////////////////
    bool isUpfrontChannel() const {
        return m_openChannelUpfront;
    }

    void i_setState(int state) {
        m_state = state;
    }
    void i_setQueue(int queue) {
        m_gkQueueIx = queue;
    }
    int i_state() const {
        return m_state;
    }
    int i_queue() const {
        return m_state;
    }

    int i_queueIx() const {
        return m_gkQueueIx;
    }

    bool isCurrentSnippetRateLimited() const;
    bool isCurrentSnippetCompressed() const;

    int64 i_resultRows() const {
        return m_resultRows;
    }
    int64 i_resultBytes() const {
        return m_resultBytes;
    }
    void i_incResultRows(unsigned int inc) {
        m_resultRows += inc;
    }
    void i_incResultBytes(int64 inc) {
        m_resultBytes += inc;
    }
    bool i_detectedAsRunaway() const {
        return m_detectedAsRunaway;
    }
    int i_runawayLimit() const {
        return m_runawayLimit;
    }
    void i_setDetectedAsRunaway(bool tf) {
        m_detectedAsRunaway = tf;
    }
    void i_setEnableZMapDbg(bool flag) {
        m_enableZMapDbg = flag;
    }
    bool i_enableZMapDbg() const {
        return m_enableZMapDbg;
    }

    void i_setGroomMode (GroomMode groomMode, TXidValue stableTXid) {
        m_groomMode = groomMode;
        m_stableTXid = stableTXid;
    }
    GroomMode i_groomMode() const {
        return (GroomMode) m_groomMode; // 0 for non-groom plans; GroomMode for groom plans
    }
    TXidValue stableTXid() const {
        return m_stableTXid;
    }

    const CError& abortReason() const {
        return m_abortReason;
    }
    void setAbortReason(const CError& err) {
        m_abortReason = err;
    }
    void i_setismainplan(bool ismain) {
        m_ismainplan = ismain;
    }
    bool i_ismainplan() const {
        return m_ismainplan;
    }

    void i_initPcodeTemps() {
        m_PcodeGen->InitTemps();
    }
    void i_initParamSpecList() {
        m_PcodeGen->InitParamSpecList();
    }
    void i_emitParamStructs(CGenCode *genCodeP) {
        m_PcodeGen->EmitParamStructs(genCodeP);
    }
    void i_emitParamRecInits(int nodePos, CGenCode *genCodeP) {
        m_PcodeGen->EmitParamRecInits( nodePos, genCodeP );
    }
    CTable *i_scanTable() const {
        return m_curScan ? m_curScan->ScanTable() : 0;
    }
    const TableShape *i_fpgaResultTable() const {
        return m_curScan ? m_curScan->FpgaResultTable() : 0;
    }
    SnippetNode *i_nodeList() const {
        return m_nodeList;
    }
    mem::Pool *i_pool() const {
        return m_pool;
    }
    const TxInfo & i_xInfo() const {
        return m_xInfo;
    }
    void i_setXinfo(const TxInfo& xInfo) {
        m_xInfo = xInfo;
    }
    const TXid& i_xId() const {
        return m_xInfo.xId;
    }
    const TXid& i_identifyingXid() const {
        return m_xInfo.identifyingXid;
    }
    void i_setXid(const TXid& tid) {
        m_xInfo.xId = tid;
    }
    int i_sessionId() const {
        return m_sessionId;
    }
    SBCSmap *i_CSmapP() const {
        return m_CSmapP;
    }
    void i_setSessionid(int sid) {
        m_sessionId = sid;
    }
   
    int  i_extendMaxSegmentSize() const {
        return m_extend_max_segment_size;
    }
    void i_setExtendMaxSegmentSize(int newsize) {
        m_extend_max_segment_size = newsize;
    }
    
    int i_randomChunkSize() const {
        return m_random_dist_chunk_size;
    }
    void i_setRandomDistChunkSize(int randomChunkSize) {
        m_random_dist_chunk_size = randomChunkSize;
    }
   

    TUseSpecialFields i_specialTableFields() const {
        return m_specialTableFields;
    }
    EPlanTypes i_planType() const {
        return m_planType;
    }

    ExprContext *i_econtext() const {
        return m_econtext;
    }
    void setExprContext(const ExprContext *econtext);

    void i_noteStartOfJoinSequence() {
        NZ_VERIFY(!(m_nodeCreationFlags & NfJoinSequence));
        m_nodeCreationFlags |= NfJoinSequence;
    }

    void i_noteEndOfJoinSequence() {
        NZ_VERIFY(m_nodeCreationFlags & NfJoinSequence);
        m_nodeCreationFlags &= ~NfJoinSequence;
    }

    int i_getNodeCreateFlag() const {
        return m_nodeCreationFlags;
    }

    void i_setNodeCreateFlag(int flag) {
        m_nodeCreationFlags |= flag;
    }

    CFpgaWrap *i_fpgaWrapper() const {
        return m_fpgaWrapper;
    }

    TScanNode *i_scanNode() const {
        NZ_VERIFY(m_nodeList && m_nodeList->Type() == ScanNode);
        return (TScanNode *) m_nodeList;
    }
    int i_blocksInRecovery() const {
        return m_blocksInRecovery;
    }
    int i_incBlocksInRecovery() {
        return ++m_blocksInRecovery;
    }
    int i_decBlocksInRecovery() {
        NZ_VERIFY( m_blocksInRecovery > 0 );
        return --m_blocksInRecovery;
    }
    TScanNode *i_curScan() const {
        return m_curScan;
    }
    int planDsid() const {
#if FOR_SPU
        return m_dataSliceId;
#else
        return 0;
#endif
    }
    int planDsliceIndex() const {
#if FOR_SPU
        return m_dsliceIndex;
#else
        return 0;
#endif
    }
#if FOR_SPU
    DSliceInteraction *planDSliceInteraction() const {
        return m_dsinteraction;
    }
#else
    COSInteraction *i_getOsInteraction() const {
        return m_interaction;
    }
#endif
    int i_id() const { // plan id used with PlanDictionary
        return m_id;
    }
    CPlan *i_next() const {
        return m_next;
    }

    int i_planid() const { // unique id visible to outside world (not PlanDictionary)
        return m_planid;
    }
    CWorkItem *i_workItem() const {
        return m_workItem;
    }
    int i_clientId() const {
        return m_clientId;
    }
    int i_cliInteractId() const {
        return m_cliInteractId;
    }

#if FOR_SPU
    const char *i_userName() const {
        return m_userName;
    }
#else
    const char *i_userName() const {
        return m_userName.c_str();
    }
    void i_setUserName(const char *userName) {
        m_userName = userName;
    }
#endif

#if FOR_SPU
    const char *i_queryStr() const {
        return m_queryStr;
    }
#else
    void i_setQueryStr(const char *queryStr) {
        m_queryStr = queryStr;
    }
    const char *i_queryStr() const {
        return m_queryStr.c_str();
    }
    const rc::StringRef& queryStrRef() const {
        return m_queryStr;
    }
#endif

    const char *i_planFileName() const { //get user specified plan file name
        if (!m_planFileName || (*m_planFileName != 'F'))
            return 0;
        return m_planFileName + 1; // skip the first letter indicating File/Dir
    }
    const char *i_planDirName() const { //get user specified plan dir name
        if (!m_planFileName || (*m_planFileName != 'D'))
            return 0;
        return m_planFileName + 1; // skip the first letter indicating File/Dir
    }
    void i_setPlanFileName(char *planFileName) {
        //set user specified plan file name
        m_planFileName = planFileName;
    }

    void setStmtTime(AbsoluteTime secs, int usecs);
    Timestamp i_stmtTime() const {
        return m_stmtTime;
    }

    // Client Info Fields
    void setClientUserId(const char *client_user_id) {
        m_client_user_id = client_user_id;
    }
    const char * getClientUserId() const {
        return m_client_user_id;
    }
    void setClientApplicationName(const char *client_application_name) {
        m_client_application_name = client_application_name;
    }
    const char * getClientApplicationName() const {
        return m_client_application_name;
    }
    void setClientWorkstationName(const char *client_workstation_name) {
        m_client_workstation_name = client_workstation_name;
    }
    const char * getClientWorkstationName() const {
        return m_client_workstation_name;
    }
    void setClientAccountingString(const char *client_accounting_string) {
        m_client_accounting_string = client_accounting_string;
    }
    const char * getClientAccountingString() const {
        return m_client_accounting_string;
    }

    void SetPGQueueDur(const uint64& duration) {
        m_planTimeStats.pgQueueDuration = duration;
    }

    void setParserSubmitTime(time64_t time){
        m_planTimeStats.parserSubmitTime = time;
    }

    void setRewriteSubmitTime(time64_t time){
        m_planTimeStats.rewriteSubmitTime = time;
    }

    void setPlannerSubmitTime(time64_t time){
        m_planTimeStats.plannerSubmitTime = time;
    }

    void setSessionVar(PlanSessionVarType which) {
        m_sessionVarFlags |= (1U << which);
    }
    bool sessionVar(PlanSessionVarType which) const {
        return m_sessionVarFlags & (1U << which);
    }

    void i_setTotalDisk(const int64& totalDisk) {
        m_totalDisk = totalDisk;
    }
    int64 i_totalDisk() const {
        return m_totalDisk;
    }
    void i_setPeakMem(const int64& peakMem) {
        m_peakMem = peakMem;
    }
    int64 i_peakMem() const {
        return m_peakMem;
    }
    void i_setTotalCost(const double& totalCost) {
        m_totalCost = totalCost;
    }
    double i_totalCost() const {
        return m_totalCost;
    }
    void i_setRetries(int retries) {
        m_retries = retries;
    }
    int i_retries() const {
        return m_retries;
    }
    void i_setParentOpid(const uint64& opid) {
        m_parentOpid = opid;
    }
    uint64 i_parentOpid() const {
        return m_parentOpid;
    }
    void i_setOpid(const uint64& opid) {
        m_opid = opid;
    }
    uint64 i_opid() const {
        return m_opid;
    }
    bool i_snippetSuccess () const {
        return m_snippetSuccess;
    }

    void i_setSnippetSuccess (bool success) {
        m_snippetSuccess = success;
    }

    void i_setCurMergeNode(TSortNode *nodeP) {
        NZ_ASSERT(!m_curMergeNode);
        m_curMergeNode = nodeP;
    }

    void i_setForGenStats(bool forGenStats) {
        m_forGenStats = forGenStats;
    }
    bool i_forGenStats() const {
        return(m_forGenStats);
    }
    void i_setForJitStats(bool forJitStats) {
        m_forJitStats = forJitStats;
    }
    bool i_forJitStats() const {
        return(m_forJitStats);
    }
    void i_setAdviseNotShortQuery(bool adviseNotShortQuery)  {
         m_adviseNotShortQuery = adviseNotShortQuery;
    }
    bool i_adviseNotShortQuery() const {
         return (m_adviseNotShortQuery);
    }
    void i_setForRestore(bool forRestore) {
        m_forRestore = forRestore;
    }
    bool i_forRestore() const {
        return(m_forRestore);
    }
    bool isCompressedUnload() const;
    void i_setunloadAlterTblCompressedNoNull(bool unloadAlterTblCompressedNoNull) {
        m_unloadAlterTblCompressedNoNull = unloadAlterTblCompressedNoNull;
    }
    bool i_unloadAlterTblCompressedNoNull() const {
        return m_unloadAlterTblCompressedNoNull;
    }
    bool i_isPrepSnippet() const {
        return m_firstSnippetNode && (m_firstSnippetNode->Type() == ScanPrepNode);
    }

    void setDbCharset(int charset);
    int i_dbCharset() const {
        return m_dbCharset;
    }

    void setUdxStub(bool stub);
    bool i_udxStub() const {
        return m_udxStub;
    }
    void setUdxStubIndividual(bool stub);
    bool i_udxStubIndividual() const {
        return m_udxStubIndividual;
    }

    void setIsSliceRestricted(bool dro) {
        m_snipRestrictSlices = dro;
    }

    bool i_isSliceRestricted() const {
        return m_snipRestrictSlices;
    }

    void setExpectedResultRows(int rows) {
        m_snipExpectedResultRows = rows;
    }
    int i_expectedResultRows() const {
        return m_snipExpectedResultRows;
    }

    bool stopDownloadChannel();     // returns false if no download node
    bool restartDownloadChannel();  // returns false if no download node

    PlanJob *job() const { return m_job; }
    void setJob(PlanJob *job) { m_job = job; }

    bool morePlansInJob() const { return m_morePlansInJob; }
    void setMorePlansInJob(bool val) { m_morePlansInJob = val; }

    SnippetNode *i_firstSnippetNode() const { return m_firstSnippetNode; }
    SnippetNode *findLastSnippetNode(SnippetNode *pNode) const;
    SnippetNode *i_firstDbosNode()    const { return m_firstDbosNode; }
    SnippetNode *i_nextSnippetNode()  const { return m_nextSnippetNode; }

    CExtTableInterface *i_exttable() const { return m_extTableInterface; }

    Fuzzy_buffers *i_fuzzyBuffers() const { return m_fuzzyBuffersP; }

    int i_udfCount(int snippetNum) const;
    int i_udaCount(int snippetNum) const;
    int i_udtfCount(int snippetNum) const;


    int i_udfCount() const { // ?UDF?
        return m_udfCount;
    }

    UdfInfo *i_getUdfInfo(int i) { // ?UDF?
        return &m_udfInfoP[i];
    }

    int i_udtfCount() const { // ?UDF?
        return m_udtfCount;
    }

    UdtfInfo *i_getUdtfInfo(int i) { // ?UDF?
        return &m_udtfInfoP[i];
    }

    int i_udaCount() const { // ?UDF?
        return m_udaCount;
    }

    UdaInfo *i_getUdaInfo(int i) { // ?UDF?
        return &m_udaInfoP[i];
    }

    bool i_isLastSnippet() const {
        return m_nextSnippetNode == 0;
    }

    bool getInsertTypeValues() {
        return m_InsertTypeValues;
    }

    /////////////////////////////////
    /// Determine whether the current snippet involves running a UDF, UDA or UDTF.
    /// Not deterministic until the plan is constructed and begins to execute.
    /// @return true if the current snippet runs a UDX, false otherwise.
    /////////////////////////////////
    bool isUDXPlan() const;

    void mapUDXPads(); // ?UDF?
    
    int registerCharObject();

    int i_groupingSetInfoCount() const {
        return m_groupingSetInfoCount;
    }

    GroupingSetInfo *i_getGroupingSetInfo(int i) {
        if (i<m_groupingSetInfoCount)
            return &m_groupingSetInfoP[i];
        else
            return NULL;
    }

#if FOR_SPU

    /////////////////////////////////
    /// Get the memory size (128KB pages) of the significant allocations for
    /// this plan & slice
    /// @return tracked recstore + comm + recptrs + hashbuckets
    /////////////////////////////////
    int getMemPageCount() const;          // total shared + non shared
    int getSharedMemPageCount() const;    // shared record store part
    int getNonSharedMemPageCount() const; // other memory
    int getFileBackedStorePageCount() const; // file backed recstore pages

    double getScanStateProgress() const;
    int getScanTableTotalBlocks() const;
    int getScanTableScanBlocks() const;
    unsigned int getScanPGConsidered() const;
    unsigned int getScanPGChanged() const;
    unsigned int getScanPGConserv() const;

    // Start a new scan. Increment the scan pass. Reset the scan percentage.
    void noteProgressNewScanPass() {
        if (m_curScanPass < m_numScanPasses) {
            m_curScanPass++;
        }
        m_curScanPassProgress = 0.0;
    }
    // Note scan progress. "current " is the counter and "total" is the max value.
    void noteProgress(int current, int total) {
        if (current >= 0 && total > 0 && current <= total) {
            m_curScanPassProgress = (double)current / total;
        }
    }

    void updateCommMemoryInUse(int pageCntDelta) {
        atomic_add(pageCntDelta, &m_commMemoryAlloced);
        if (pageCntDelta > 0)
            SpuJob::update_resacct_mem_hwm ( getMemPageCount() );
    }
    void updateRecPtrArrPageCnt(int pageCntDelta) {
        atomic_add(pageCntDelta, &m_recPtrArrPageCnt);
        if (pageCntDelta > 0)
            SpuJob::update_resacct_mem_hwm ( getMemPageCount() );
    }
    void updateHashBucketArrPageCnt(int pageCntDelta) {
        atomic_add(pageCntDelta, &m_hashBucketArrPageCnt);
        if (pageCntDelta > 0)
            SpuJob::update_resacct_mem_hwm ( getMemPageCount() );
    }
    void updateStoreListPageCount(int pageCntDelta) {
        atomic_add(pageCntDelta, &m_storeListPageCnt);
        if (pageCntDelta > 0)
            SpuJob::update_resacct_mem_hwm ( getMemPageCount() );
    }
    void updateFileBackedStorePageCount(int newFileBackedStorePageCntDelta) {
        atomic_add(newFileBackedStorePageCntDelta, &m_fileBackedStorePageCnt);
        if (newFileBackedStorePageCntDelta > 0)
            SpuJob::update_resacct_temp_disk_hwm( getFileBackedStorePageCount() );
    }
    void updateSharedPageCount(int pageCntDelta) {
        atomic_add(pageCntDelta, &m_sharedPageCnt);
        if (pageCntDelta > 0)
            SpuJob::update_resacct_mem_hwm ( getMemPageCount() );
    }

    GroomSpuState *groomState() const
    {
        return m_job->groomState(m_dsliceIndex);
    }

    TCHand preservedTCHand()
    {
        return m_job->preservedTCHand(m_dsliceIndex);
    }
    void setPreservedTCHand(TCHand handle)
    {
        m_job->setPreservedTCHand(m_dsliceIndex, handle);
    }

#endif

    // how many scan passes in this snippet?
    void determineScanPasses(SnippetNode* firstNodeOfSnippet, SnippetNode* lastNodeOfSnippet);

    void setGsZmapRecompute(bool recompute) {
        m_gsZmapColumns = (zm::Column *) recompute;
    }
    void setGsZmapRecompute(zm::Column *columns) {
        m_gsZmapColumns = columns;
    }
    zm::Column *gsZmapColumns() const {
        return m_gsZmapColumns;
    }

    bool isGroomPlan() const { return m_groomMode != 0; }
    bool isGroomVersions() const  { return m_groomMode == GROOM_VERSIONS; }

    bool isMviewCreateOrderedPlan() const;

    void execUdfCleanupFunc() throw ();

    int getSpuSnipFabricBlockEstimate() const { return m_spuSnipFabricBlockEstimate; }
    int getSpuSnipMemEstimate() const { return m_spuSnipMemEstimate; }

    PlanFilesContainer *i_planFiles() const {
        return m_planFiles;
    }

    void i_setPlanFiles(PlanFilesContainer *pfc) {
        m_planFiles = pfc;
    }

    COSInteraction *i_reconstituteTransientItp() const { return m_reconstituteTransientItp; }

#if !FOR_SPU
    SnippetNode* i_parserNode() const { return m_parserNode;}
    void setParserNode(SnippetNode* node) { m_parserNode = node;}
#endif

  PRIVATE:
    mem::Pool *m_pool;       // allocation pool for nodes
    ExprContext *m_econtext;// pg expr context for pg eval code mem allocs
    SnippetNode *m_nodeList;   // executable node chain
    SnippetNode *m_lastNode;   // last node in entire node list
    EPlanTypes m_planType;      // Select, Delete, ...

    SnippetNode *m_firstSnippetNode; // First node in snippet;
    SnippetNode *m_firstDbosNode;    // First DBOS node in snippet; if none, same as nextSnippetNode
    SnippetNode *m_nextSnippetNode;  // First node of next snippet; if none, NULL

#if FOR_SPU
    DSliceInteraction *m_dsinteraction; // associated dslice interaction
#else
    COSInteraction *m_interaction;  // last dbos snippet
#endif
    int m_clientId;             // for dbos opstat warts
    int m_cliInteractId;        // client request sequence number

    TUseSpecialFields m_specialTableFields;

    TScanNode *m_curScan;

#if FOR_SPU
    const char *m_queryStr;
#else
    rc::StringRef m_queryStr; // text of the SQL query that built this plan
#endif

    rc::StringRef m_colList; // list of column names
    rc::StringRef m_colLengthList; // list cf length of each column

    char *m_planFileName; //user specified plan file name

    PlanFilesContainer *m_planFiles;

#if FOR_SPU
    const char *m_userName;
#else
    rc::StringRef m_userName;
#endif

    /// Current datetime of statement (plan) - in Netezza format.
    /// Used for "current date" and "current time" in-SQL calculations.
    /// Does not belong in PlanTimeStats.
    Timestamp m_stmtTime;

    /// Plan time measures.
    PlanTimeStats m_planTimeStats;

    /// True if the plan should open the return channel to the host
    /// at the start of every snippet. False otherwise.
    bool m_openChannelUpfront;

    unsigned int m_sessionVarFlags; // session variable flags

    const char *m_client_user_id;           // Client info field
    const char *m_client_application_name;  // Client info field
    const char *m_client_workstation_name;  // Client info field
    const char *m_client_accounting_string; // Client info field

    int64 m_totalDisk; // bytes of disk scanned by each spu
    int64 m_peakMem;   // peak mem in bytes required from each spu
    double m_totalCost; // total run cost of a plan in milliseconds
    bool m_ismainplan; // if current plan is main/primary plan
    int m_retries;
    uint64 m_parentOpid;
    uint64 m_opid;
    sc::CHistoryData *m_history;
    CError m_abortReason;   // plan error code (0 if success)

    // For query, query hist op stats


    // While producing plans, the client tags join sections to give a hint
    // to the DBOS code that the nodes can be dynamically massaged into
    // an SPU based join.  These cannot be nested.
    int m_nodeCreationFlags; // hints - NfJoinSequence, ...

    CFpgaWrap *m_fpgaWrapper; // If a snippet ever support multiple scans, this would need to move

    bool m_keepObject;      // is compiler product to be kept?
    int m_id;               // plan id used with PlanDictionary
    CPlan *m_next;          // next in PlanDictionary

    CWorkItem *m_workItem;  // This plan's DBOS scheduler control block (DBOS)

    CPcodeGen *m_PcodeGen;   // expression-pcode generator
    void *m_dlopenHandle;   // handle from loader of 80x86 machine code
    int m_planid;           // unique id visible to outside world (not PlanDictionary)

    bool m_snippetSuccess;  // Snippet pass/fail indicator
    bool m_forGenStats;     // TRUE if plan for statistics generation
    zm::Column *m_gsZmapColumns; // Info to compile genstats zm accumulator sequence
    bool m_forRestore;      // TRUE if plan is part of restore operation

    bool m_forJitStats;           // TRUE is plan is for JIT stats generation
    bool m_adviseNotShortQuery;  //So planner/optimizer can signal to GRA that a plan shouldn't be short.
    //!NOTE-dk generalize(?) for nodes which need to be cleaned up at end
    // of current snippet-cycle on Host?
    TSortNode *m_curMergeNode;  // 0 or DBOS side merge node
    bool m_modified; //true if decision maker modified snippet for materializedNode
    bool m_DHJmodified; //true if DHJ decision maker modified snippet

    // Snippet Result Cache bookkeeping:  (TODO:  encapsulate this better!)

       // This section is filled in *before* the SPU snippet runs:
    int              m_SRC_candidate_TTID; // UNKNOWN==(-1), NONE==0, or TTID>0 of result.
                                           // The remaining m_SRC_* attributes are only valid if this is > 0.
    int              m_SRC_type;           // Type of result: SaveTemp, DownloadDistribute, or DownloadBroadcast.
                                           // Really ESnipResType in srcmgr.h; 'int' here since ESnipResTtid
                                           // is unknown to most plan.h includers.
    int              m_SRC_baseTableId;    // Base table from which transient is derived.  For primary snippet, this is enough;
                                           // for dependent snippets will need to track multiple tables.
    SnippetNode    * m_SRC_pFirstNode;     // first node of snippet.  Snippet may be "compound" (spu Return + dbos DownloadTable/Broadcast etc.)
    SnippetNode    * m_SRC_pLastNode;      // fence after (possibly compound) snippet, or NULL if end of plan.
    SnippetNode    * m_SRC_pResultNode;    // node that produces result: spu SaveTemp, spu DownloadTable/Distribute, dbos DownloadTable/Broadcast
    CFingerprint   * m_SRC_pFingerprint;   // compute/save fingerprint before a snippet is run so it is canonical

       // This section is filled in *after* the SPU snippet runs.
    SnippetNode    * m_SRC_pEntryNode;     // start next snip here if reuse cached result and skip snippet (may skip DBOS TableSizeNode etc.)
    bool	     m_SRC_setEntryNode;   // if true, set CPlan::m_entryNode from CPlan::m_SRC_pEntryNode in SrcReconstituteCb.


       // If reconstituted current result from Snippet Result cache:
    int              m_SRC_reconstituted_TTID;         // ttid of table that was reconstituted
    int              m_SRC_reconstituted_from_TTID;    // ttid of cached entry from which ttid was reconstituted
    int              m_SRC_reconstituted_from_planid;  // id of plan that produced transient that was cached.
    int              m_SRC_reconstituted_from_snipid;  // id of snippet that produced transient that was cached.
    uint64           m_SRC_nTransPages;                 // number of pages   in transient table when cached
    uint64           m_SRC_nTransRecords;               // number of records in transient table when cached
    uint64           m_SRC_nTransBytes;                 // number of bytes   in transient table when cached
    uint64           m_SRC_nPagesInLargestSlice;        // number of pages   in largest slice of transient table when cached
    uint64           m_SRC_nRecordsInLargestSlice;      // number of records in largest slice of transient table when cached
    uint64           m_SRC_nBytesInLargestSlice;        // number of records in largest slice of transient table when cached

       // If caching current result:
    int              m_SRC_caching_TTID;            // ttid of result we will cache
    int              m_SRC_cached_TTID;             // if caching succeeded, same as m_SRC_caching_TTID, else 0.

    // If NOT caching result:
    int              m_SRC_actual_TTID;             // ttid we'd be caching if we WERE caching

    SRC_TableIdVector m_SRC_ttids2Drop;     // list of TTIDs to drop if skip snippet that had DropTableNode(s)

    // create temp table msg
    TDownloadTableNode *m_pSpu2SpuDownloadNode;
    const TableShape *m_resultShape;

    // reconstitute transient interaction for SnippetResultCache
    // to make transient from cache instead of running snippet; 0 when done.
    COSInteraction *m_reconstituteTransientItp;

    // create temp table interaction for inDBOS distribute DownloadTableNode
    // of current snippet, or 0 when none
    COSInteraction *m_createTempTableItp;

    // channel open interaction used for opening an fcomm channel on spus
    COSInteraction *m_channelOpenItp;

    // Interface to External data reader (if any)
    CExtTableInterface *m_extTableInterface;
#if !FOR_SPU
    SnippetNode *m_parserNode;
#endif

    int64       m_resultRows;
    int64       m_resultBytes;
    int64       m_snipExpectedResultRows; // from planner; 0 means unknown
    int         m_state;
    int         m_gkQueueIx;
    bool        m_snipRestrictSlices; // do we expect significant slice restrictions?

    // runaway query status
    bool         m_detectedAsRunaway;
    unsigned int m_runawayLimit;

    CGenCode *m_pGenCode;

    PlanJob *m_job;
    bool m_morePlansInJob;

    // db charset info
    unsigned int m_dbCharset; // 8- bit char set oid (currently either LATIN9OID or UNDECLAREDOID)
    SBCSmap *m_CSmapP;   // ptr to SBCS map structure

    int     m_seqCount;             // number of sequences used in plan ( = m_sequenceP entry count)

    Fuzzy_buffers *m_fuzzyBuffersP;   // for Fuzzy Search Function buffers
    int64          m_UdfSizeBytes;
    int64          m_UdaSizeBytes;
    int64          m_UdtfSizeBytes;

#if FOR_SPU
    atomic_t    m_commMemoryAlloced; // 128K pages alloced for comm channel pools
    atomic_t    m_storeListPageCnt;  // 128k pages alloced for record stores +
                                     // gencodesize + plan's pool+ plan's expression pool
    atomic_t    m_sharedPageCnt;     // 128k pages alloced for record stores shared on SPU
    atomic_t    m_recPtrArrPageCnt;  // 128k pages alloced for record pointer arrays.
    atomic_t    m_hashBucketArrPageCnt; // 128k pages alloced for hash bucket arrays.
    atomic_t    m_fileBackedStorePageCnt; // 128k pages alloced for file backed record stores
#endif

    bool     m_InsertTypeValues; // Will be true if INSERT has VALUES clause.

    // ?UDF? start
    UdfInfo *m_udfInfoP;      // list of UdfInfo structs for Udfs used in this snippet
    int      m_udfCount;      // count of entries in m_udfInfoP list
    int      m_udfCapacity;   // capacity of UdfInfo list
    UdtfInfo *m_udtfInfoP;      // list of UdtfInfo structs for Udtfs used in this snippet
    int      m_udtfCount;      // count of entries in m_udtfInfoP list
    int      m_udtfCapacity;   // capacity of UdtfInfo list

    UdaInfo *m_udaInfoP;      // list of UdfInfo structs for Udfs used in this snippet
    int      m_udaCount;      // count of entries in m_udfInfoP list
    int      m_udaCapacity;   // capacity of UdfInfo list
    // ?UDF? end

    GroupingSetInfo *m_groupingSetInfoP;          // list of GroupingSetInfo structs for GroupingSetNodes used in this snippet
    int              m_groupingSetInfoCount;      // count of entries in m_groupingSetInfoP list
    int              m_groupingSetInfoCapacity;   // capacity of UdfInfo list

    int m_snippetNumForNode;

    int m_spuSnipFabricBlockEstimate;  // Estimated number of blocks SPU will send
    int m_spuSnipMemEstimate;          // Estimated memory SPU will use for this snippet (blocks)

    friend class CFpgaWrap;     // for execScan()

    friend class COSInteraction;
    friend class CSpuSwapMgr;
    friend class TRestrictNode;
    friend class TReturnNode;
    friend class TProjectNode;
    friend class TAggregateNode;
    friend class PlanDictionary;
    friend class CPcodeGen;

    int getPlanSubPath(char *buf) const; // <planid>
    int getPlanBasePath(char *buf) const; // /nz/kit/log/plans
    int getPlanPath(char *buf) const; // /nz/kit/log/plans/<planid>
    int getPlanName(char *buf) const; // <planid>.pln
    int getPlanHistBasePath(char *buf) const; // /nz/kit/log/planshist

    void DumpPlanToFile() const; //dump plan into file
    void ReDumpPlanToFile(int snippetNum, bool aborted = false); //redump plan if was modified by DHJ or aborted

    /////////////////////////////////
    /// Print snippet's stats - times, estimates, resources, ... - to plan
    /// file.
    /// @param snippetNum: the number of snippet
    /// @param dumpPlanTotals: dump the plan stats (true for last snippet)
    /////////////////////////////////
    void dumpSnippetStatsToPlanFile(int snippetNum, bool dumpPlanTotals = false);
    void dumpChildStatsToPlanFile(FILE* plnFp, const CWorkItem *pwi);

    void dumpToPlanResourceVt(bool aborted = false) const;
    void dumpToSnipResourceVt(bool aborted = false) const;

    void getExprFuzzySearchWorkBufNeeds(Node *pExpr);
    void getQualFuzzySearchWorkBufNeeds(List *pQual);
    void getSnippetFuzzySearchWorkBufNeeds(SnippetNode *pFirst,SnippetNode *pFence);

    void getExprUdfWorkBufNeeds(Node *pExpr, int64 *pnUdfSize);
    void getQualUdfWorkBufNeeds(List *pQual, int64 *pnUdfSize);
    void getSnippetUdxWorkBufNeeds(SnippetNode *pFirst,SnippetNode *pFence);

    TMaterializeNode *insertMaterializeNode( //insert a materializing node
        EPlanNodeLocus place,
        SnippetNode *addPos=NULL);

    void deriveTblShapes (
        SnippetNode *pFirstNode,    // beginning of snippet
        SnippetNode *pLastNode);    // end (exclusive) of snippet

    const TableShape *findFirstTblShape(
        SnippetNode *pFirstNode,    // beginning of snippet
        SnippetNode *pLastNode);    // end (exclusive) of snippet

    void scanSnippetForDHJ(
        SnippetNode *pFirstNode,    // first node in snippet
        SnippetNode *pLastNode);    // last node in snippet

    int getPlanFixedMemory (   // return the amount of fixed memory for plan
        SnippetNode *firstNode,
        SnippetNode *lastNode);

    void setMicrocode(
        TScanNode *pNode,        // first node for this snippet
        fpgaPortablePacket *fpgaMicrocode);

    void setKeyValues(
        TScanNode *pNode,
        TKeyValue *pKeys,
        int numValues);

    void getConstFromCommBuf(
        CCommBuf *cbuf,
        int *consttype,
        int *constlen,
        Datum *constvalue);

    void addConstToCommBuf(
        CCommBuf *cbuf,        // comm buffer pointer
        int consttype,
        int constlen,
        Datum constvalue);

    TKeyValue *getKeyValuesFromCommBuf(
        CCommBuf *cbuf,
        int numKeyValues);

    void addKeyValuesToCommBuf(
        CCommBuf *cbuf,         // comm buffer pointer
        int numKeyValues,       // number of key values
        TKeyValue *pKeyValues); // array of key values

    TZoneMapUnion *getZoneMapUnionFromCommBuf(
        CCommBuf *cbuf);

    void addZoneMapUnionToCommBuf(
        CCommBuf *cbuf,         // comm buffer pointer
        TZoneMapUnion *pZoneMapUnion);

    void getZoneMapConjFromCommBuf(
        CCommBuf *cbuf,
        TZoneMapConj *pzmConj );

    void addZoneMapConjToCommBuf(
        CCommBuf *cbuf,         // comm buffer pointer
        const TZoneMapConj& zmConj );

    void getZoneMapExprFromCommBuf(
        CCommBuf *cbuf,
        TZoneMapExpr *pzmExpr );

    void addZoneMapExprToCommBuf(
        CCommBuf *cbuf,         // comm buffer pointer
        const TZoneMapExpr& zmExpr );

    TZoneMapUnion *getZoneMapExprArrayFromCommBuf(
        CCommBuf *cbuf,
        int exprCount,
        EZMCombiner zmCombiner );

    void addZoneMapExprArrayToCommBuf(
        CCommBuf* cbuf,
        TZoneMapUnion* pZoneMapUnion );

    int childListLength(
        TZoneMapUnion* pZoneMapUnion );

    int simpleExprListLength(
        TZoneMapUnion *pZoneMapUnion );

    void rectifyPhysIDs(
        int *pIDs,        // physical field IDs to be rectified
        int oldIDCount,   // number of IDs in list
        int newIDCount ); // number of IDs to remain

    void initNodesState( // Spin thru plan nodes, initing internals
        EPlanNodeLocus place, const TableShape *src,
        SnippetNode *pFirst, SnippetNode *pFence);

    void initHostEventer(SnippetNode *firstDbosNode, SnippetNode *nextSnippetNode); // Create generated code to recieve snippet results

    void computeDHJPartitionSizes(   //determine partition sizes for DHJ nodes
        SnippetNode *pFirstNode,   //first node in snippet
        SnippetNode *pFence);      //node beyond last node in snippet

    void verifyDHJPartitionSizes(   //determine partition sizes for DHJ nodes
        SnippetNode *pFirstNode,   //first node in snippet
        SnippetNode *pFence);      //node beyond last node in snippet

    const TableShape *initScanNode(TScanNode *pNode, EPlanNodeLocus place);
    const TableShape *initProjectNode(const TableShape *src, TProjectNode *pNode);
    const TableShape *initRestrictNode(const TableShape *src, TRestrictNode *pNode);

    void addExprToCommBuf(CCommBuf *cbuf, Node *expression,
                          ExprContext *econtext, int *invalidFuncP);
    void addExprExprToCommBuf(CCommBuf *cbuf, Expr *expr,
                              ExprContext *econtext, int *invalidFuncP);
    void addIntsListToCommBuf(CCommBuf *cbuf, List *list);
    void addExprContextToCommBuf(CCommBuf *cbuf, ExprContext *econtext);
    void addTableOpHdr(CCommBuf *cbuf, SnippetNode *node, int *codeP = 0, int *auxP = 0);
    void addEndOfSubplanMarker (CCommBuf *cbuf);
    void threadIn( // Add node to nodeList
        SnippetNode *node,     // node to add
        SnippetNode *addPos=NULL); // node after which to insert this (NULL for end)
    void threadInBefore( // Add node to nodeList
        SnippetNode *node,     // node to add
        SnippetNode *beforePos=NULL); // node which to insert this in front of (NULL for end)
    void unThread( // Remove node from nodeList
        SnippetNode *node);

    pcFacts *loadPcodeFromCommBuf(
        CCommBuf *cbuf );
    void addPcodeToCommBuf(
        CCommBuf *cbuf,
        pcFacts *facts);

    Node *getExprFromCommBuf(CCommBuf *cbuf, const bool restoreMode = false);
    SubPlan *getSubPlanFromCommBuf(CCommBuf *cbuf);
    List *getIntsListFromCommBuf(CCommBuf *cbuf);
    Expr *getExprExprFromCommBuf(CCommBuf *cbuf);

    List *getDSFVListFromCommBuf (CCommBuf *cbuf);
    void addDSFVListToCommBuf (CCommBuf *cbuf, List *dsfvList, int *invalidFuncP);

    void getHdrFromCommBuf (
        CCommBuf *cbuf,
        int *typeP,         // OUT
        int *placeP,        // OUT
        int *flagsP,        // OUT
        int64 *statsTagP,   // OUT
        int *detailStatsP,  // OUT
        uint32 *rtFlagsP,   // OUT
        int16 *srcFailureCodeP, // OUT
        int *srcFailureAuxP,    // OUT
        char **commentP,    // OUT
        int *plannerFlags,  // OUT
        int64 *numRecsEst); // OUT


    bool GetNodeFromCommBuf(
        CCommBuf *cbuf);     // 'elf' array of data

    uint32 ComputeFingerprintIDFromFingerprint(
        CCommBuf const * pBuf ) const;

    bool CreateFingerprintFromSnippet(
        SnippetNode * pStart,
        SnippetNode * pFence,
        CCommBuf * pBuf,
        table_id_t resultTblid );


    int countSnippets() const; // Return number of snippets in this plan
    int getNodeIdx(const SnippetNode *pNode) const;
    void setSnippetNumForNode(int lastLocus);
    void i_setSnippetNumForNode(int num) {
        m_snippetNumForNode = num;
    }
    int i_snippetNumForNode() const {
        return m_snippetNumForNode;
    }
    void setUsedSwTxFiltering() {
        m_usedSwTxFiltering = true;
    }
    bool usedSwTxFiltering() {
        return m_usedSwTxFiltering;
    }

    int i_reconPlanId() const {
            return m_SRC_reconstituted_from_planid;
    }

    int i_reconSnipId() const {
              return m_SRC_reconstituted_from_snipid;
    }

    int i_cachedId() const {
    	return m_SRC_cached_TTID;
    }

    bool i_DHJmodified() const {
    	return m_DHJmodified;
    }

    bool createCcodeForNode(
        TProjectNode *pProjector,
        TStreamInfo *outStreamP,
        TStreamInfo *flowInStreamP,
        TStreamInfo *tempInStreamP,
        CGenCode *genCodeP);

    bool createCcodeForNode(
        TJoinNode *pJoin,
        EJoinQualType qualType,
        TStreamInfo *outStreamP,
        TStreamInfo *flowInStreamP,
        TStreamInfo *tempInStreamP,
        CGenCode *genCodeP,
        bool aggrJoin);

    bool createCcodeForNode(
        TFuncJoinNode *pJoin,
        TStreamInfo *outStreamP,
        TStreamInfo *flowInStreamP,
        TStreamInfo *tempInStreamP,
        CGenCode *genCodeP);

    bool createCcodeForNode(
        TRestrictNode *pRestrict,
        TStreamInfo *outStreamP,
        TStreamInfo *flowInStreamP,
        TStreamInfo *tempInStreamP,
        CGenCode *genCodeP);

    bool createCcodeForNode(
        TConditionalBranchNode *pBranch,
        TStreamInfo *outStreamP,
        TStreamInfo *flowInStreamP,
        TStreamInfo *tempInStreamP,
        CGenCode *genCodeP);

    bool createCcodeForNode(
        TAggregateNode *pAggregate,
        TStreamInfo *outStreamP,
        TStreamInfo *flowInStreamP,
        TStreamInfo *tempInStreamP,
        CGenCode *genCodeP);

    bool createCcodeForFrame(
        SnippetNode* pNode,
        TWindowFrame *pWindowFrame,
        TStreamInfo *outStreamP,
        TStreamInfo *flowInStreamP,
        TStreamInfo *tempInStreamP,
        CGenCode *genCodeP);

    bool createCcodeForNode(
        SnippetNode* pNode,
        List *qual,
        int pcodeStart,
        int pcodeEnd,
        const char *pVarName,
        TStreamInfo *outStreamP,
        TStreamInfo *flowInStreamP,
        TStreamInfo *tempInStreamP,
        CGenCode *genCodeP);

    bool needsPostgresEval(
        TProjectNode *pProjector );

    bool needsPostgresEval(
        TAggregateNode *pAggregate );

    bool needsPostgresEval(
        THashJoinNode *pHashJoin );

    bool needsPostgresEval(
        TRestrictNode *pRestrict );

    bool createPcodeForNode(
        TProjectNode *pProjector,
        const TableShape *pTempTbl);

    bool createPcodeForNode(
        TRestrictNode *pRestrict);

    bool createPcodeForNode(
        TJoinNode *pJoin,
        const TableShape *pTempTbl);

    bool createPcodeForNode(
        TAggregateNode *pAggregate );

    bool createPcodeForNode(
        TWindowAggrNode *pWindowAggr );

    bool createPcodeForNode(
        TFuncJoinNode *pJoin,
        const TableShape *pTempTbl);

    bool createPcodeForNode(
        TConditionalBranchNode *pRestrict);

    bool createPcodeForNode( TSaveTempNode *pSaveTempNode );

    bool createPcodeForNode( TDownloadTableNode *pDownloadTableNode );

    void initPcodeContext();

    void initPcodeVarAndSeqPtrs();

    // ?UDF? start
    void addUdfInfo(pcInstruction *pInstr, struct Func *func,
                char **returntypeP, char **argListP,  int nargs, int snippetNum);
    UdfInfo *getUdfInfo(pcInstruction *pInstr, Oid funcid, int snippetNum);
    int getUdfEntryNo(pcInstruction *pInstr, Oid funcid, int snippetNum);

    void addUdtfInfo(TFuncJoinNode *node, int snippetNum);
    UdtfInfo *getUdtfInfo(TFuncJoinNode *node, int snippetNum);
    int getUdtfEntryNo(TFuncJoinNode *node, int snippetNum);

    void addUdaInfo(TAggrField *pInstr, void *pNode, int snippetNum);
    UdaInfo *getUdaInfo(TAggrField *pInstr, void *pNode, int snippetNum);
    int getUdaEntryNo(TAggrField *pInstr, void *pNode, int snippetNum);
    void addUdaInfo(TWindowAggrField *pInstr, void *pNode, int snippetNum);
    UdaInfo *getUdaInfo(TWindowAggrField *pInstr, void *pNode, int snippetNum);
    int getUdaEntryNo(TWindowAggrField *pInstr, void *pNode, int snippetNum);
    // ?UDF? end

    void propagatePhysIDs(
        const TableShape *pScanTable,   // possible source of physIDs
        TProjectNode *pProj ); // project node for which we are making list

    const TableShape *deriveSnippetProjection( // Create tableDesc based on snippet's projection
        SnippetNode *pScanNode,    // beginning of snippet
        SnippetNode *pLastNode);     // end (exclusive) of snippet

    void extractIndexKeysAndMicrocode( // getMicKeys
        TScanNode *pNode,   // scan node of snippet
        SnippetNode *lastSnippetNode, // last node of snippet
        bool isPrepSnippet ); // true -> prep snippet

    void scanForSPUPreps();
    void createPrepSnippet(
        TScanNode *pScan,
        TDownloadTableNode *pNewTable);
    TScanNode *scanNeedsPrep(TScanNode *pScan);  //TScanNode class??

    static TScanNode *GetRealScanNode(SnippetNode *pNode);

    int getRealSnippetNum(
        SnippetNode *prepNodeP,      // first node of prep snippet
        TScanNode *scanNodeP,         // first node of corresponding real snippet
        int snipNum) const;          // current snippet number

    void execSpuSnippet(
        SnippetNode *firstNode,
        SnippetNode *lastNode,
        const TableShape *resultShape);
    int spuSnippetSchedCosts (
        SnippetNode *pFirstNode,   // first (inclusive) node
        SnippetNode *pFenceNode);  // last  (exclusive) node
    void hostSnippetSchedCosts (
        SnippetNode *pFirstNode,
        int nominalReturnSize);

    friend int lazyLoadMachineCode(
        CPlan *plan,
        char *recordBuf,
        char *bufEnds,
        bool lastCall);

    void cleanUpPlanNodes(bool finalInstance);

    void updateVirtab();    // Add execution info to plan's virtab

    void writeToVtPlanLog();  // Add plan record to _vt_plan_log

    void unloadGenCode();

    void planHistLog       (EPlanStatus status);
    void planHistLogHelper (sc::CHistoryData *historyData, EPlanStatus status);
    void fillPlanHist      (sc::CHistoryData *historyData, EPlanStatus status);

    // Snippet Result Cache helpers:
    int setSrcState( table_id_t ttid, int srt, SnippetNode* pNode );
    int setSrcErrorState( table_id_t ttid, int srt, table_id_t realTblid, SnippetNode* pNode, int code, int size );
    table_id_t adjustEntryNode(
        SnippetNode* pDLTNode,
        table_id_t ttid );
    table_id_t analyzeDbosSnippet(
        SnippetNode* pHostNode,
        SnippetNode* pFirstNode,
        SnippetNode* pFence,
        int numResults );
    int analyzeSnippetResult(
        SnippetNode *pFirstNode,
        SnippetNode *pFenceNode );
    CSnipResInfo * reuseSnippetResult();
    bool checkSpecialActions(
        SnippetNode *pFirstNode,
        SnippetNode *pFenceNode );
    void cleanupSRCobjects();
};

extern "C" void GencAbortPlan (CPlan *planP); // see also geninl.h

#define PLAN_CHECK_INTERRUPT(plan) do { \
    CPlan *const _plan = (plan); \
    bool const interrupted = _plan->m_interruptRequested; \
    SMP_RMB(); \
    if (NZ_UNLIKELY(interrupted)) \
        GencAbortPlan(_plan); \
} while (false)

//Check the time and send a Snippet Resource Accounting message, if one is due.
#if FOR_SPU
#define SEND_RESACCT_IF_DUE(plan) plan->planDSliceInteraction()->dsiSendResAcctIfDue(plan->job());
#else
#define SEND_RESACCT_IF_DUE(plan)
#endif

//!FIX-dlm  should be in a "Scan Object"
// These are bits. They may be ORed together.
// In particular PROCESS and either LIMITED or FINAL can occur together.
// If PROCESS is set, the generated code is called. If either LIMITED or
//  FINAL is also set, the size is set to 0 and the isFinal flag is set.
// If LIMITED is set without PROCESS this means that we've already done
//  the 'final' processing and are now letting enqueued scans play out.
// When FINAL is set, we can delete the plan.
enum {
    FINALITY_PROCESS = 1,    // normal scan
    FINALITY_LIMITED = 2,    // first scan after limit reached
    FINALITY_FINAL = 4       // final scan, whether limited or not
};

#define SYSTEM_SCAN_PREFIX "#"

/* Bit masks used for setting the bit for m_sessionVar
 * A maximum of 32 session variables can be accommodated
*/
#define CAST_EMPTY_STR_ZERO 0
#define LEGACY_CAST_STR_TO_NUM 1
#define ANSI_SUBSTRING 2
#define ENABLE_ISAO 3
#define STRICT_TEMPORALS_FORMAT 4
#define DATESTYLE_SQL_EURO 5
#define FORCE_PG_EVAL 6
#define ENABLE_NEW_CASE 7
#define ENABLE_TIME24 8
#define DISABLE_GENC_GCC_OPT 8
#define DISABLE_SRC 9

#endif // !defined( PLAN_INCLUDED )
