/*
 * Classes and routines for writing and accessing plan files to/from shared memory buffer.
 * IPC between dbos event and archiver through shared memory queue.
 *
 * (C) Copyright IBM Corp. 2011, 2012  All Rights Reserved.
 */

#ifndef PLANFILES_INCLUDED
#define PLANFILES_INCLUDED

#include <vector>
#include <string>
#include <cstdio>
#include "../include/regex.h"

#include "rcShmAllocBase.h"
#include "sdlist.h"
#include "nzautomutex.h"
#include "locktracker.h"
#include "nzmutex.h"

LKTRKR_DECLARE_NAME(MTX_PLANFILES);
LKTRKR_DECLARE_NAME(MTX_PLANFILES_QUEUE);

class PlanFilesWriter {
public:
    virtual void write(const char * pathname, const char * data) = 0;
protected:
    PlanFilesWriter() {}
};

class PlanFilesContainer : public rc::ShmAllocBase<PlanFilesContainer, mt_planfiles> {

    typedef shmstring<char>::type File;
    typedef shmstring<char>::type Name;
    typedef container<std::vector, File>::type FileVec;
    typedef container<std::vector, Name>::type NameVec;

public:
    PlanFilesContainer();
    ~PlanFilesContainer();

    File& openSrc(const char *);
    File& openPln(const char *);
    void closeSrc();

    PlanFilesContainer * appendSrc(const char *);
    PlanFilesContainer * appendPln(const char *);

    void writePln(PlanFilesWriter *);
    void writeSrc(PlanFilesWriter *);
    void writeAll(PlanFilesWriter *);

    void writeSrc(const char *, PlanFilesWriter *);
    void writePln(const char *, PlanFilesWriter *);

    inline nzmutex_t * getMutex()
    {
        return &m_mutex;
    }

    const char * getSrcC();
    const char * getPlnC();
    int getPlnSize();


private:

    Name m_plnName;
    Name m_srcName;

    File m_pln;
    File m_src;

    FileVec m_sources;
    NameVec m_srcNames;

    bool m_isNew;

    nzmutex_t m_mutex; // Secure all the operations that want to change or read
                       // .pln buffer as it may be accessed from different processes.
                       // This may cause reading a buffer which is not EOL terminated.
                       // So far we don't need to secure the operations on source code
                       // buffer, as there's no concurrent access to it.
                       // If it's to be accessed from different processes, don't forget
                       // to use the lock.
};

class PlanFileStream {
    FILE * m_fp;
    char * m_buf;
    size_t m_bufL;

public:
    PlanFileStream();

    ~PlanFileStream();

    FILE * getStream();

    const char * c_str() const;

    size_t size() const;

    void copyInto(FILE *);
};

class PlanFileFdStream {
    static const char * const s_fpattern;

    char m_fname[32];
    FILE * m_fp;

public:
    PlanFileFdStream();

    ~PlanFileFdStream();

    FILE * getStreamRewind();

    FILE * getStream();
};

#if !FOR_SPU

struct plan_files_t {
        dlink_t link;
        int planId;
        PlanFilesContainer * planFiles;
};

// here, to keep control over memory type it goes to
plan_files_t * create_plan_files();
void destroy_plan_files(plan_files_t *);

class PlanFilesQueue {

        dlink_t m_queue;
        nzmutex_t m_queue_lock;
        uint64 m_numAvail;

public:
        PlanFilesQueue();
        ~PlanFilesQueue();
        void * operator new(size_t);
        void operator delete(void *);

        void put(plan_files_t * planFiles);
        plan_files_t * get(bool * full);
        plan_files_t * getLocked();

        inline nzmutex_t * getMutex()
        {
            return &m_queue_lock;
        }

        inline dlink_t * getQueue()
        {
            return &m_queue;
        }

        inline uint64 getSize()
        {
            return m_numAvail;
        }
};
#endif

#endif // PLANFILES_INCLUDED
