/// @file ptsException.h
/// @brief PTS exception classes.
///
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2011, 2012  All Rights Reserved.
//////////////////////////////////////////////////////////////////////
///
/// Description:
///
/// PTS exception classes.
///
//////////////////////////////////////////////////////////////////////

#ifndef PTSEXCEPTION_H
#define PTSEXCEPTION_H

#include <stdexcept>
#include <string>
#include <errno.h>
#include "ptsTrace.h"

namespace pts
{

// Base exception class with file and line number
class PTSBaseException : public std::exception
{
public:
    virtual const char* what() const throw();
    const std::string& getFile() const;
    int getLine() const;
    const std::string& getStacktrace() const;

protected:
    PTSBaseException() : m_line(0) { }
    PTSBaseException(const std::string &file, int line, const std::string &errorMessage, const std::string &stacktrace);
    PTSBaseException(const PTSBaseException &other);
    virtual ~PTSBaseException() throw();

private:
    std::string m_message;
    std::string m_file;
    int m_line;
    std::string m_stacktrace;
};

// Exception base class for system API failures
class PTSSystemException : public PTSBaseException
{
public:
    const std::string &getFunctionName() const;
    int getErrno() const;
    virtual const char* what() const throw();

protected:
    PTSSystemException() : m_errno(0) { }
    PTSSystemException(const std::string &file, int line,
            const std::string &functionName, int errnum, const std::string &objectName,
            const std::string &errorMessage,
            const std::string &stacktrace);
    PTSSystemException(const PTSSystemException &other);
    virtual ~PTSSystemException() throw();

    void setWhat();

    std::string m_functionName;
    int m_errno;
    std::string m_objectName;
    std::string m_what;
};

// Exception class for system API failures that set errno
class PTSErrnoException : public PTSSystemException
{
public:
    PTSErrnoException() { }
    PTSErrnoException(const std::string &file, int line,
            const std::string &functionName, int errnum, const std::string &objectName,
            const std::string &stacktrace);
    virtual ~PTSErrnoException() throw();

    bool isFileExistenceException() const;
    bool isFileAccessException() const;
};

// Exception class for system API failures that set h_errno
class PTSHErrnoException : public PTSSystemException
{
public:
    PTSHErrnoException(const std::string &file, int line,
            const std::string &functionName, int errnum, const std::string &objectName,
            const std::string &stacktrace);
    virtual ~PTSHErrnoException() throw();
};

// Exception class for mis-configuration issues
class PTSConfigException : public PTSBaseException
{
public:
    PTSConfigException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSConfigException() throw();
};

// Exception class for invalid command-line arguments
class PTSParamException : public PTSBaseException
{
public:
    PTSParamException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);

    PTSParamException(const std::string &file, int line,
            const std::string &paramName,
            const std::string &paramValue,
            const std::string &stacktrace);

    PTSParamException(const std::string &file, int line,
            const int paramPosition,
            const std::string &paramValue,
            const std::string &stacktrace);

    virtual ~PTSParamException() throw();

    virtual const char* what() const throw();

private:
    void setWhat();

    int m_paramPosition;
    std::string m_paramName;
    std::string m_paramValue;
    std::string m_what;
};

// Exception class for invalid network messages
class PTSNetworkException : public PTSBaseException
{
public:
    PTSNetworkException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSNetworkException() throw();
};

// Exception class for when network connection is broken
class PTSNetworkDisconnectException : public PTSBaseException
{
public:
    PTSNetworkDisconnectException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSNetworkDisconnectException() throw();
};

// Exception class for when network connection is broken for a network alias
class PTSNetworkAliasException : public PTSBaseException
{
public:
    PTSNetworkAliasException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSNetworkAliasException() throw();
};

// Exception class for when sender and receiver are at incompatible software revisions
class PTSNetworkRevisionException : public PTSBaseException
{
public:
    PTSNetworkRevisionException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSNetworkRevisionException() throw();
};

// Exception class for when database row is missing
class PTSInvalidRowException : public PTSBaseException
{
public:
    PTSInvalidRowException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSInvalidRowException() throw();
};

// Exception class for invalid conditions
class PTSAssertException : public PTSBaseException
{
public:
    PTSAssertException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSAssertException() throw();
};

// Exception class for inotify/file system monitoring problems
class PTSMonitorException : public PTSBaseException
{
public:
    PTSMonitorException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSMonitorException() throw();
};

// Exception class to wrap rc::Err
class PTSReusableComponentException : public PTSBaseException
{
public:
    PTSReusableComponentException(const std::string &file, int line,
    		const std::string &errMsg, const std::string &stacktrace);
    virtual ~PTSReusableComponentException() throw();
};


// Exception class for SQL Log reader
class PTSSqlLogReaderException : public PTSBaseException
{
public:
    PTSSqlLogReaderException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSSqlLogReaderException() throw();
};

// Exception class for invalid conditions
class PTSVersionException : public PTSBaseException
{
public:
    PTSVersionException(const std::string &file, int line,
            const std::string &errorMessage,
            const std::string &stacktrace);
    virtual ~PTSVersionException() throw();
};
} //namespace pts

#define THROW_ERRNO(f, e, o) throw pts::PTSErrnoException(__FILE__, __LINE__, f, e, o, PTSCallStack::getCallStack(__LINE__))
#define THROW_ERRNO2(f, e) throw pts::PTSErrnoException(__FILE__, __LINE__, f, e, "", PTSCallStack::getCallStack(__LINE__))
#define THROW_H_ERRNO(f, e, o) throw pts::PTSHErrnoException(__FILE__, __LINE__, f, e, o, PTSCallStack::getCallStack(__LINE__))
#define THROW_CONFIG(m) throw pts::PTSConfigException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_PARAM(m) throw pts::PTSParamException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_PARAM2(p, v) throw pts::PTSParamException(__FILE__, __LINE__, p, v, PTSCallStack::getCallStack(__LINE__))
#define THROW_NETWORK_ALIAS(m) throw pts::PTSNetworkAliasException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_NETWORK(m) throw pts::PTSNetworkException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_NETWORK_REVISION(m) throw pts::PTSNetworkRevisionException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_DISCONNECT(m) throw pts::PTSNetworkDisconnectException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_NOROW(m) throw pts::PTSInvalidRowException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_ASSERT(m) throw pts::PTSAssertException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_MONITOR(m) throw pts::PTSMonitorException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_RC(e, m) throw pts::PTSReusableComponentException(e.fileName(), e.lineNo(), m, PTSCallStack::getCallStack(__LINE__))
#define THROW_SQLLOGREADER(m) throw pts::PTSSqlLogReaderException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))
#define THROW_VERSION(m) throw pts::PTSVersionException(__FILE__, __LINE__, m, PTSCallStack::getCallStack(__LINE__))

#define PTS_UNLIKELY(cond)   __builtin_expect(!!(cond),0)
#define PTS_LIKELY(cond)     __builtin_expect(!!(cond),1)

#define PTS_ERRNO_CHECK(cond, op, eno) do { \
    if (PTS_UNLIKELY(!(cond)))              \
        THROW_ERRNO2((op), (eno));          \
} while (false)

#define PTS_CHECK(cond, msg) do { \
    if (PTS_UNLIKELY(!(cond)))    \
        THROW_ASSERT(msg);        \
} while (false)

#endif // PTSEXCEPTION_H
