/// @file ptsTrace.h
/// @brief Tracing facility.
///
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2011  All Rights Reserved.
//////////////////////////////////////////////////////////////////////
///
/// Description:
///
/// Tracing facility.
///
//////////////////////////////////////////////////////////////////////

#ifndef PTSTRACE_H
#define PTSTRACE_H

#include <stdint.h>
#include <string>
#include <iostream>
#include <sstream>
#include <syslog.h>
#include <stdarg.h>

namespace pts
{


enum TraceStatus
{
	DETAIL_TRACE_OFF = 0,
	DETAIL_TRACE_ON = 1,
	DETAIL_TRACE_FULL = 2
};

class PTSTrace : public std::ostringstream
{
public:
	// Constructor is only for PTS_TRACE macros
	PTSTrace(const char *file, int line);
	virtual ~PTSTrace();

	static void enableTracing(TraceStatus traceLevel = DETAIL_TRACE_ON);
	static void disableTracing();
	static void setTraceDir(const std::string& dirPath);
	static void setTraceFile(const std::string& appName, const std::string& instanceName);
	static void setTraceFileSize(int64_t fileSize);
	static void setTraceFileCount(int32_t fileCount);
	static void traceAlways(int priority, std::ostream& os);
	static bool traceFileEmpty();
	static void deleteTraceFile();
	static void setThreadPrefix(const std::string& prefix);

	static const char* getBasename(const char *filePath);

	// Inline for performance
	static bool isOn()
	{
		return m_traceOn & (DETAIL_TRACE_FULL | DETAIL_TRACE_ON);
	}

	static bool isFull()
	{
		return m_traceOn & DETAIL_TRACE_FULL;
	}

	static std::string getTraceDir()
	{
		return m_traceDir;
	}
	static int getTraceFileSize()
	{
		return m_fileSize;
	}

	static int getTraceFileCount()
	{
		return m_fileCount;
	}

	static int getTraceOn()
	{
		return m_traceOn;
	}

private:
	static void checkFileSize();
	static void checkFileCount();

	static TraceStatus m_traceOn;
	static std::string m_traceDir;
	static std::string m_identity;
	static std::ostream *m_traceFile;
	static int64_t m_fileSize;
	static int32_t m_fileCount;
    static pthread_mutex_t m_mutex;

	static std::string combinePath(const std::string& rootDir, const std::string& path);
};

class PTSCallStack
{
	static PTSCallStack __thread *g_bottomOfStack;

public:

	static std::string getCallStack(int lineNumber = 0);

	PTSCallStack(const char* cname, const char* mname)
	: m_className(cname), m_methodName(mname), m_next(g_bottomOfStack)
	{
		g_bottomOfStack = this;

		if(PTSTrace::isFull())
			traceEntry();
	}

	~PTSCallStack()
	{
		if(PTSTrace::isFull())
			traceExit();

		g_bottomOfStack = (PTSCallStack*) m_next;
	}

private:

	void traceEntry();
	void traceExit();

	const char *m_className;
	const char *m_methodName;
	const PTSCallStack *m_next;
};

} //namespace pts

#define PTS_TRACEALWAYS(args) \
	do { \
		pts::PTSTrace _trace(__FILE__, __LINE__); \
		pts::PTSTrace::traceAlways(LOG_ERR, _trace << args); \
	} while (false)

#define PTS_TRACE(args) \
	do { \
		if (pts::PTSTrace::isOn()) \
		{ \
			pts::PTSTrace _trace(__FILE__, __LINE__); \
			pts::PTSTrace::traceAlways(LOG_DEBUG, _trace << args); \
		} \
	} while (false)

extern void PTS_LOG(const char *fmt, ...) __attribute__((format(printf, 1, 2)));

#endif // PTSTRACE_H
