//=============================================================================
///\file: rcAtomicSmp.h - Implementation of an Atomic SMP safe class
//
// (C) Copyright IBM Corp. 2012, 2014  All Rights Reserved.
//=============================================================================

#ifndef RCATOMICSMP_INCLUDED
#define RCATOMICSMP_INCLUDED

#include "rcAtomic.h"
#include "rcBarrier.h"

#ifdef __cplusplus
#if FOR_CLI || WIN32
    // assume irrelevant when neither potentially-SMP host nor SPU
#else // otherwise, assume Linux

namespace rc
{

/// \brief An atomic class wrapper of atomic_t that is safe in a SMP environment
/// \details Note this class is safe for use in the following models:
/// \li Multi-threaded SMP Environment
/// \li Shared Memory Multi-process SMP Environment
/// \li Multi-threaded Single Processor
/// \li Single-threaded Single Processor
class AtomicSmpInt
{
private:
    atomic_t    m_val;      ///< atomic data member (volatile in struct)

public:
    /// \brief Destructor
    ~AtomicSmpInt( )
        { }

    /// \brief Default constructor
    AtomicSmpInt()
        {
            SMP_WMB();
            this->set( 0 );
        }

    /// \brief Construct with value
    AtomicSmpInt( int value )
        {
            SMP_WMB();
            this->set( value );
        }

    /// \brief Returns the value
    int get( ) const
        {
            int temp = atomic_read( &m_val );
            SMP_RMB();
            return temp;
        }

    /// \brief Sets a value
    void set( int i )
        {
            SMP_WMB();
            atomic_set( &m_val, i );
        }

    /// \brief Clear the value
    void clear( )
        {
            SMP_WMB();
            atomic_set( &m_val, 0 );
        }

    /// \brief Adds \a i to the value
    /// \return Returns the value after the addition
    int add( int i )
        { return atomic_add_return( i, &m_val ); }

    /// \brief Subtracts \a i from the value
    /// \return Returns the value after the subtraction
    int sub( int i )
        { return atomic_sub_return( i, &m_val ); }

    /// \brief Value is Bitwise AND against \a mask
    /// \return Returns the value after the AND
    int bit_and( int mask )
        {
            // Invert mask...
            atomic_clear_mask( ~mask, &m_val );
            return this->get();
        }

    /// \brief Value is Bitwise OR against \a mask
    /// \return Returns the value after the OR
    int bit_or( int mask )
        {
            atomic_set_mask( mask, &m_val );
            return this->get();
        }

    /// \brief Compare and exchange old versus new value
    /// \return Previous value before the exchange
    int cmpxchg( int oldValue, int newValue )
        { return atomic_cmpxchg( &m_val, oldValue, newValue ); }

    /// \brief Exchange value
    /// \return Previous value before the exchange
    int xchg( int value )
        { return atomic_xchg( &m_val, value ); }

    /// \brief Type conversion
    operator int() const
        { return this->get(); }

    /// \brief Value assignment operator
    int operator= ( int value )
        {
            this->set( value );
            return this->get();
        }

    /// \brief Overload post-increment operator
    int operator++ ( int )
        { return this->add( 1 ) - 1; }

    /// \brief Overload pre-increment operator
    int operator++ ( )
        { return this->add( 1 ); }

    /// \brief Overload post-decrement operator
    int operator-- ( int )
        { return this->sub( 1 ) + 1; }

    /// \brief Overload pre-decrement operator
    int operator-- ( )
        { return this->sub( 1 ); }

    /// \brief Overload += operator
    int operator+= ( int i )
        { return this->add( i ); }

    /// \brief Overload -= operator
    int operator-= ( int i )
        { return this->sub( i ); }

    /// \brief Overload &= operator
    int operator&= ( int mask )
        { return this->bit_and( mask ); }

    /// \brief Overload -= operator
    int operator|= ( int mask )
        { return this->bit_or( mask ); }

private:
    /// \brief Disabled Copy Constructor
    AtomicSmpInt( const AtomicSmpInt& rhs );

    /// \brief Disabled Assignment Operator
    AtomicSmpInt& operator= ( const AtomicSmpInt& rhs );
}; // End of AtomicSmpInt

}; // End of Namespace rc

#endif // End FOR_CLI || WIN32
#endif // End __cplusplus

#endif /* RCATOMICSMP_INCLUDED */
