//------------------------------------------------------------------------------
// rcDbg.h
//
// Utility class that supports conditional debug output by module and flag.
//
// (C) Copyright IBM Corp. 2001, 2011  All Rights Reserved.
//
// NOTE: changed this class to use default new/delete.  Having it use the memory
// subsystem is not a useful dependency.  It only allocates from heap.

#ifndef RCDBG_INCLUDED
#define RCDBG_INCLUDED

#include <string.h>
#include "nzstd.h"
#include "rcStdString.h"
#include "rcBool.h"
#include "rcRefCnt.h"
#include <stdarg.h>

// Usage:
//
// Used to conditionally print debug output based on a module and debug flag.
// The latter is similar to a debug level. The difference is that you can choose
// subsets of output at better granularity.
//
// Typical output would look like:
//
//  [i,sps ] opening file 'abc' ...
//
// The debug output contains the flag and module that caused that line to be
// printed.
//
// The simplest form of the code to make this happens would look like:
//
//  DBG_PR ((DM_SPS, DF_INIT, "opening file '%s' ...\n", filename));
//
// This is really a macro that expands to the following:
//
//  if (RC_DBG_ANY_ON)
//      RC_DBG->print (DM_SPS, DF_INIT, "opening file '%s' ...\n", filename);
//
// This latter form is important, as it can be used when multiple prints are
// grouped or when the print arguments incur non-trivial overhead. For instance:
//
//  if (RC_DBG_ANY_ON) {
//      RC_DBG->print (DM_SPS, DF_INIT, "opening file '%s' ...\n", filename);
//      if (RC_DBG->isOn (DM_SPS, DF_INIT_DATA))
//          file->dump ();
//  }
//
// The global RC_DBG_ANY_ON is non-zero if any flags are on. This simple
// check makes the general check for debug very efficient. (For the standard
// case where debug printing is off.) That way, debug print statements can
// be left in the code for production product.
//
// Also note the pairs of flags. The lower case is for single line output ('i'
// corresponding to DF_INIT). The upper case version is for extended details
// ('I' for DF_INIT_DATA).

// If using explicit "va_list" mechanism instead of an ordinary arglist, use
//  DBG_VPR instead of DBG_PR, as in:
//
//    void foo(const char *fmt, ...)
//    {
//        va_list ap;
//        va_start(ap, fmt);
//        DBG_VPR((DM_SPS, DF_INIT, ap, fmt));
//        va_end(ap);
//    }

// Macro for convenience, ie RC_DBG->print (...)

#define RC_DBG (rc::Dbg::instance())
#define RC_DBG_ANY_ON (NZ_UNLIKELY(rc::Dbg::isAnyOn ()))
#define RC_DBG_ON(module,flag) (RC_DBG_ANY_ON && RC_DBG->isOn (module, flag))

#define DBG_PR(body) do { if (RC_DBG_ANY_ON) RC_DBG->print body; } while (0)
#define DBG_VPR(body) do { if (RC_DBG_ANY_ON) RC_DBG->vprint body; } while (0)

#define DBG_DUMP_OBJ(mod,flg,msg,obj) do { \
    if (RC_DBG_ANY_ON && RC_DBG->print (mod, flg, "%s\n", msg)) (obj)->dump (); } \
    while (0)

void rcInitDbg (int* argc, char** argv);

const char* const DBG_MODULES_ENV_VAR   = "DBG_MODULES";
const char* const DBG_FLAGS_ENV_VAR     = "DBG_FLAGS";

const char* const SPU_DBG_MODS_ENV_VAR  = "SPU_DBG_MODULES";
const char* const SPU_DBG_FLAGS_ENV_VAR = "SPU_DBG_FLAGS";

//------------------------------------------------------------------------------
// RcDbgModule is an enum of module IDs. Modules represent a subsystem or major
// area and are enabled with the -Dm <module list> command line argument (or
// the DBG_MODULES environment variable). Module IDs index into an array, with
// slot 0 unused.
//
// RcDbgModule is a global enum (rather than class scoped) in order to shorten
// the symbol name for convenience - since it's a highly used enum.

// Note: Changes to this list of modules must be reflected in
//       Dbg::moduleSymbol () and Dbg::moduleLabel ().

enum RcDbgModule {
    DM_ALL = -1, DM_NONE = 0,
    DM_AC = 1,  /* App Comps        */  DM_ADC,     /* Admin cmds       */
    DM_BTS,     /* Boot Svr         */
    DM_CLM,     /* Client Mgr       */  DM_COMM,    /* Comm layer       */
    DM_DBOS,    /* DBOS             */  DM_DC,      /* DB Comps         */
    DM_ERR,     /* Error Handling   */  DM_EVM,     /* Event Mgr        */
    DM_FPGA,    /* FPGA             */  DM_HC,      /* HW Comps         */
    DM_IPC,     /* Inter-proc Comm  */  DM_LDR,     /* Loader           */
    DM_IPS,     /* IPC Svr          */
    DM_MEM,     /* Memory subsystem */  DM_NLOG,    /* NZ Log () func   */
    DM_NTF,     /* Notifier         */  DM_PLAN,    /* Plan             */
    DM_PG,      /* Postgres         */  DM_RC,      /* Recomps          */
    DM_SC,      /* Sys comps        */  DM_NET,     /* Net comps        */
    DM_SIG,     /* signal handling  */  DM_SNMP,    /* SNMP             */
    DM_SPS,     /* Startup Svr      */  DM_SPU,     /* SPU proper       */
    DM_SQST,    /* SQL Stats        */
    DM_STM,     /* Stats Mgr        */  DM_SYS,     /* sys              */
    DM_TMP,     /* "For now" cases  */  DM_TPM,     /* Topology Mgr     */
    DM_TXM,     /* transaction mgr  */  DM_UTL,     /* util, misc       */
    DM_SM,      /* System Mgr       */  DM_GENC,    /* gen code         */
    DM_SFI,     /* SFI              */  DM_STOR,    /* storage          */
    DM_SES,     /* Session manager  */  DM_EXPR,    /* expressions      */
    DM_BNR,     /* backup, restore  */  DM_OPST,    /* op stats         */
    DM_VAC,     /* nzreclaim        */  DM_CC,      /* ctrlcomps        */
    DM_REGEN,   /* regen            */  DM_SMACK,   /* Smack            */
    DM_ZMAP,    /* zone map         */  DM_SWAP,    /* Swapper          */
    DM_SQLOLAP, /* sqlolap analytics*/  DM_SEQ,     /* Sequences        */
    DM_LDRSHM,  /* ldr dbos sh mem  */  DM_TXRB,    /* tx  roll back    */
    DM_SCHED,   /* Scheduler        */  DM_SCHEDRS, /* Sched Resources  */
    DM_NBCLIB,  /* Netezza Backup Conector Interface  */
    DM_UDX,     /* UDF and UDA      */  DM_LIB,     /* Libraries        */
    DM_RECLAIM_RECOV, /* SPU Reclaim Recovery mgr */
    DM_SESSION, /* Session Context  */  DM_GROOM,   /* Groom/DOTs       */
    DM_TST,     /* Unit tests       */  DM_AMS,     /* AMM Server       */
    DM_MLS,     /* Multi-Level Security */
    DM_DPM,     /* Data Partition Manager (aka OSR) */
    DM_SCSI,    /* SCSI Disk */
    DM_RATES,   /* host distribute rate limiting*/
    DM_PHY,     /* Phy error count mornitoring */
    DM_REPL,    /* Replication Capture and Apply agents */
    DM_SRC,	/* Snippet Result Cache */
    DM_SRULE,   /* Scheduler Rules */

    DM_END, DM_BEGIN = DM_AC, NUM_DBG_MODULES = DM_END
};

//------------------------------------------------------------------------------
// RcDbgFlags is an enum of bits, one or more of which can be on at a time.
// Flags are 'levels' or 'cases' that can apply to any module and are enabled
// with the -Df <flags> command line argument (or the DBG_FLAGS environment
// variable).
//
// RcDbgFlags is a global enum (rather than class scoped) in order to shorten
// the symbol name for convenience - since it's a highly used enum.

// Note: Changes to these flags must be reflected in Dbg::flagToChar(),
//       Dbg::charToFlag (), and Dbg::printUsage ().

enum RcDbgFlags {
    DF_NONE     = 0,        DF_ALL          = 0x7fffffff,
    DF_MSG      = BIT0,     DF_MSG_DATA     = BIT1,     // c, C (comm msgs)
    DF_DEBUG    = BIT2,     DF_DEBUG_DATA   = BIT3,     // d, D
    DF_INIT     = BIT4,     DF_INIT_DATA    = BIT5,     // i, I
    DF_MEMORY   = BIT6,     DF_MEMORY_DATA  = BIT7,     // m, M
    DF_PROFILE  = BIT8,     DF_PROFILE_DATA = BIT9,     // p, P
    DF_SQL      = BIT10,    DF_SQL_DATA     = BIT11,    // s, S
    DF_MINOR    = BIT12,    DF_MINOR_DATA   = BIT13,    // z, Z
    DF_TIME     = BIT14,    DF_TIME_DATA    = BIT15,    // t, T
    DF_FAULT    = BIT16,    DF_FAULT_DATA   = BIT17,    // f, F
    DF_STATE    = BIT18,    DF_STATE_DATA   = BIT19,    // x, X
    DF_RB       = BIT20,    DF_RB_DATA      = BIT21,    // r, R
    DF_AEK      = BIT22,    DF_AEK_DATA     = BIT23,    // k, K

    DF_END, DF_BEGIN = DF_MSG,

    DF_ALL_BUT_MEM  = (DF_ALL & ~(DF_MEMORY | DF_MEMORY_DATA)),
    DF_BASIC_FLAGS  = (DF_MSG | DF_DEBUG | DF_INIT | DF_SQL)
};

namespace rc {

class TmpBuf;
class ByteStrm;
class OutDevice;

//==============================================================================
// Class: DbgCtrls - print 'controls' for debug output facility.

class DbgCtrls {
public:
    enum FieldFlags {
        FF_NONE = 0, FF_HAS_FLAGS = BIT0, FF_HAS_MODULES = BIT1,

        FF_ALL = (FF_HAS_FLAGS | FF_HAS_MODULES)
    };

private:
    // NOTE: This class is used as a payload from the system manager to SPUs.
    // So, be careful of adding to the fields because the wire footprint
    // matches the object one.

    int         m_fieldFlags;               // For mergeFrom ()

    RcDbgFlags  m_flags;                    // Current debug flags turned on

    RcBool      m_outputToFile;
    RcBool      m_prSrc;
    RcBool      m_prTimestamp;
    RcBool      m_teeToHost;                // For embedded apps (SPU, SFI).
    RcBool      m_teeToLog;
    RcBool      m_teeToSysLog;

    char        m_modules[NUM_DBG_MODULES]; // Current module flags turned on

public:
    // CREATORS
    DbgCtrls (RcBool initialValue = RC_B_NO);
    DbgCtrls (const DbgCtrls& src);
    ~DbgCtrls ();

    // MANIPULATORS
    void        clear ();
    void        swap ();                                // For Host <-> SPU
    void        restoreFrom (rc::ByteStrm& s);
    void        mergeFrom (const DbgCtrls& src);
    DbgCtrls&   operator= (const DbgCtrls& src);

    void        processArgs (int* argc, char** argv);
#if !FOR_SPU
    string      composeArgList();
#endif
    void        clearFlag (RcDbgFlags flag);
    void        setFlag (RcDbgFlags flag);
    void        setFlags (RcDbgFlags flags);

    void        clearModule (RcDbgModule module);
    void        setModule (RcDbgModule module);
    void        setModules (const char* modules);

    void        setOutputToFile (RcBool val);
    void        setPrSrc (RcBool val);
    void        setPrTimestamp (RcBool val);
    void        setTeeToHost (RcBool val);
    void        setTeeToLog (RcBool val);
    void        setTeeToSysLog (RcBool val);

    // Each process*Arg method below throws rc:Err on error

    void        processFlagsArg (const char* arg, bool allowNull = false);
    void        processModulesArg (const char* arg, bool allowNull = false);

    // ACCESSORS
    size_t      binarySize () const;
    void        saveOn (rc::ByteStrm& s) const;
    void        dump () const;
    void        dumpObj (int indentLv = 0, const char* label = NULL) const;

    char*       getModuleFlags (rc::TmpBuf *tmpBuf, bool *allModules) const;
    char*       getDebugFlags (rc::TmpBuf *tmpBuf) const;


    FieldFlags  fieldFlags () const;
    RcDbgFlags  flags () const;
    const char* modules () const;           // Vector of MAX_DBG_MODULES bools

    RcBool      outputToFile () const;
    RcBool      prTimestamp () const;
    RcBool      prSrc () const;
    RcBool      teeToHost () const;
    RcBool      teeToLog () const;
    RcBool      teeToSysLog () const;

private:
    friend class Dbg;
};

//==============================================================================
// Class: Dbg - a general purpose conditional debug facility.

class Dbg : public rc::RefCnt {
private:
    typedef rc::RefCnt inherited;

    static bool s_anyOn;                // Anything turned on?

    DbgCtrls    m_ctrlsBuf;             // Buffer for non-alloc'd ctrls.
    DbgCtrls*   m_ctrls;                // Current flags, modules, etc.
    OutDevice*  m_stdOutDevice;         // Std output device to use
    OutDevice*  m_sysLogDevice;         // Optional sysLog device

    bool        m_usingFile;

public:
    // CREATORS
    Dbg (int references = 0);
    Dbg (rc::OutDevice* stdOutDevice, int references = 0);
    Dbg (rc::OutDevice* stdOutDevice, DbgCtrls* ctrls, int references = 0);
    ~Dbg ();

    static Dbg* instance ();                                // Default instance
    static void setInstance (Dbg* dbg);
    static Dbg* defaultInstance (bool allocIfNull = true);

    // MANIPULATORS
    void        addReference ();
    int         removeReference ();

    void        init (int* argc, char** argv);              // Throws rc::Err

    void        vprint(RcDbgModule module, RcDbgFlags flag,
		       const char* fmt, va_list args);
    bool        print (RcDbgModule module, RcDbgFlags flag,
                        const char* fmt, ...);
    bool        print (RcDbgModule module, RcDbgFlags flag, int indentLvl,
                        const char* fmt, ...);
    void        print (int indentLvl, const char* fmt, ...);
    void        print (const char* fullLine, bool addNewline = false);

    void        setCtrls (const DbgCtrls& ctrls);
    void        mergeCtrls (const DbgCtrls& ctrls);

    void        setCtrlsPtr (DbgCtrls* ctrls);      // For shMem case
    DbgCtrls*   ctrlsPtr ();                        // For shMem case

    void        clearFlag (RcDbgFlags flag);
    void        setFlag (RcDbgFlags flag);
    void        setFlags (RcDbgFlags flags);

    void        clearModule (RcDbgModule module);
    void        setModule (RcDbgModule module);

    void        setStdOutDevice (rc::OutDevice* device);
    void        setPrSrc (bool enable);
    void        setTeeToHost (bool enable);
    void        setTeeToLog (bool enable);
    void        setTeeToSysLog (bool enable);

    void        updateAnyOn ();

#if !FOR_SPU
    string      composeArgList();
#endif

    // ACCESSORS
    static bool         isAnyOn ();
    bool                isOn (RcDbgModule module, RcDbgFlags flag) const;

    void                dump () const;
    void                dumpObj (int indentLv = 0,
                                 const char* label = NULL) const;
    void                printUsage (bool forDbgCli = false) const;

    const DbgCtrls&     ctrls () const;
    RcDbgFlags          flags () const;
    const char*         modules () const;   // Vector of MAX_DBG_MODULES bools
    rc::OutDevice*      stdOutDevice () const;
    rc::OutDevice*      sysLogDevice () const;

    static char         flagToChar (RcDbgFlags flag);
    static RcDbgFlags   charToFlag (char c);

    static RcDbgModule  moduleId (const char* moduleSymbol);
    static const char*  moduleSymbol (RcDbgModule module);
    static const char*  moduleLabel (RcDbgModule module);

private:
    // NOT IMPLEMENTED
    Dbg (const Dbg& src);
    Dbg& operator = (const Dbg& src);

    // PRIVATE MANIPULATORS
    void        processArgs (int* argc, char** argv);
#if !SPU_OS_NUCLEUS
    string      outFilePath ();
#endif
    void        pr (const char* prefix1, const char* line);
    void        pr (const char* fullLine, bool addNewline);

    int         fmtPrefixes (char* prefix1Buf, char* prefix2Buf,
                             const DbgCtrls& ctrls, RcDbgModule module,
                             RcDbgFlags flag, int indentLvl);

    void        updateOutFile ();
    void        updateSysLogTee ();
};

//==============================================================================
// Class: DbgCtrlsStack - a utility class that uses the ctor and dtor to save
// the current dbg ctrls, set new ones, and then restore the original.

class DbgCtrlsStack {
private:
    Dbg*        m_dbg;
    DbgCtrls    m_savedCtrls;

public:
    // CREATORS
    DbgCtrlsStack (Dbg* dbg = NULL);        // NULL == use RC_DBG
    DbgCtrlsStack (RcDbgModule module, RcDbgFlags flags, Dbg* dbg = NULL);
    ~DbgCtrlsStack ();

    // ACCESSORS

    const Dbg*      dbg () const;
    const DbgCtrls& savedCtrls () const;
};

//==============================================================================
// Class: DbgFdStack - a utility class that uses the ctor and dtor to save
// the current dbg fd, set a new one, and then restore the original.

class DbgFdStack {
private:
    Dbg*            m_dbg;
    rc::OutDevice*  m_outDevice;
    void*           m_savedFd;
    bool            m_teeToLog;
    bool            m_teeToSysLog;

public:
    // CREATORS
    DbgFdStack (void* fd, Dbg* dbg = NULL);
    ~DbgFdStack ();

    // ACCESSORS

    const Dbg*              dbg () const;
    const rc::OutDevice*    outDevice () const;
    const void*             savedFd () const;
    bool                    teeToLog () const;
    bool                    teeToSysLog () const;

private:
    // NOT IMPLEMENTED
    DbgFdStack ();
    DbgFdStack (const DbgFdStack* src);
    DbgFdStack& operator= (const DbgFdStack* src);
};

//------------------------------------------------------------------------------
// rc::DbgCtrls INLINE MANIPULATORS

inline void DbgCtrls::clearFlag (RcDbgFlags flag)
{
    m_flags = (RcDbgFlags) (m_flags & ~flag);
}

inline void DbgCtrls::setFlag (RcDbgFlags flag)
{
    m_flags       = (RcDbgFlags) (m_flags | flag);
    m_fieldFlags |= FF_HAS_FLAGS;
}

inline void DbgCtrls::setFlags (RcDbgFlags flags)
{
    m_flags       = flags;
    m_fieldFlags |= FF_HAS_FLAGS;
}

inline void DbgCtrls::setOutputToFile (RcBool val)
{
    m_outputToFile = val;
}

inline void DbgCtrls::setPrSrc (RcBool val)
{
    m_prSrc = val;
}

inline void DbgCtrls::setPrTimestamp (RcBool val)
{
    m_prTimestamp = val;
}

inline void DbgCtrls::setTeeToHost (RcBool val)
{
    m_teeToHost = val;
}

inline void DbgCtrls::setTeeToLog (RcBool val)
{
    m_teeToLog = val;
}

inline void DbgCtrls::setTeeToSysLog (RcBool val)
{
    m_teeToSysLog = val;
}

//------------------------------------------------------------------------------
// rc::DbgCtrls INLINE ACCESSORS

inline DbgCtrls::FieldFlags DbgCtrls::fieldFlags () const
{
    return (FieldFlags) m_fieldFlags;
}

inline RcDbgFlags DbgCtrls::flags () const
{
    return m_flags;
}

inline const char* DbgCtrls::modules () const
{
    return m_modules;
}

inline RcBool DbgCtrls::outputToFile () const
{
    return m_outputToFile;
}

inline RcBool DbgCtrls::prSrc () const
{
    return m_prSrc;
}

inline RcBool DbgCtrls::prTimestamp () const
{
    return m_prTimestamp;
}

inline RcBool DbgCtrls::teeToHost () const
{
    return m_teeToHost;
}

inline RcBool DbgCtrls::teeToLog () const
{
    return m_teeToLog;
}

inline RcBool DbgCtrls::teeToSysLog () const
{
    return m_teeToSysLog;
}

//------------------------------------------------------------------------------
// rc::Dbg INLINE MANIPULATORS

inline DbgCtrls* Dbg::ctrlsPtr ()
{
    return m_ctrls;
}

//------------------------------------------------------------------------------
// rc::Dbg INLINE ACCESSORS

inline bool Dbg::isAnyOn ()
{
    return s_anyOn;
}

inline const DbgCtrls& Dbg::ctrls () const
{
    return *m_ctrls;
}

inline RcDbgFlags Dbg::flags () const
{
    return m_ctrls->flags ();
}

inline const char* Dbg::modules () const
{
    return m_ctrls->modules ();
}

inline rc::OutDevice* Dbg::stdOutDevice () const
{
    return m_stdOutDevice;
}

inline rc::OutDevice* Dbg::sysLogDevice () const
{
    return m_sysLogDevice;
}

//------------------------------------------------------------------------------
// rc::DbgCtrlsStack INLINE ACCESSORS

inline const Dbg* DbgCtrlsStack::dbg () const
{
    return m_dbg;
}

inline const DbgCtrls& DbgCtrlsStack::savedCtrls () const
{
    return m_savedCtrls;
}

//------------------------------------------------------------------------------
// rc::DbgFdStack INLINE ACCESSORS

inline const Dbg* DbgFdStack::dbg () const
{
    return m_dbg;
}

inline const rc::OutDevice* DbgFdStack::outDevice () const
{
    return m_outDevice;
}

inline const void* DbgFdStack::savedFd () const
{
    return m_savedFd;
}

inline bool DbgFdStack::teeToLog () const
{
    return m_teeToLog;
}

inline bool DbgFdStack::teeToSysLog () const
{
    return m_teeToSysLog;
}

} // end namespace rc

#endif /* RCDBG_INCLUDED */
