//------------------------------------------------------------------------------
// rcPgmAttrs.h
//
// A per program or thread object that holds high-level program attributes
// (name, log dir, src string, ...).
//
// (C) Copyright IBM Corp. 2002, 2011  All Rights Reserved.
//
// Changes:
//
// 07-15-03  dick       Added startMsgLogged boolean.
// 06-21-02  dick       Initial version.

#ifndef RCPGMATTRS_INCLUDED
#define RCPGMATTRS_INCLUDED

#ifndef NZSTD_INCLUDED
# include "nzstd.h"
#endif

#include "rcPgmClass.h"

#define MAX_LT_INDEXES  64  // max of 64 threads per process.

namespace rc {

//==============================================================================
// Class: PgmAttrs - a global or per thread object that holds high-level pgm
// attributes. Typically used by rc::Dbg, rc::Log, ...

class PgmAttrs {
public:
    enum { PATH_BUF_SIZE = 256, NAME_BUF_SIZE = 48, STD_SRC_PREFIX_LEN = 13 };

    typedef PgmAttrs*           InstanceRtn ();
    typedef rc::PgmClass::Val   PgmClassVal;

private:
    static PgmAttrs*    s_instance;
    static InstanceRtn* s_instanceRtn;
    static bool         s_initialized;

    static PgmClassVal  s_pgmClass;
    static char         s_arg0[PATH_BUF_SIZE];
    static char         s_pgmExecDir[PATH_BUF_SIZE];
    static char         s_pgmName[NAME_BUF_SIZE];

    static int          s_hwId;
    static char         s_logName[NAME_BUF_SIZE];
    static int          s_pid;

    char m_pgmLabel[NAME_BUF_SIZE];
    char m_srcPrefix[NAME_BUF_SIZE+16];     // '(<label>[.hwId][.identity]) '
    int  m_identity;                        // Actual value used
    int  m_hwIdInSrcPrefix;
    bool m_startMsgLogged;

public:
    // CREATORS
    PgmAttrs ();
    ~PgmAttrs ();

    // MANIPULATORS
    static
    PgmAttrs*           instance ();                // For pgm or thread

    void                init (int argc, char** argv, int pgmClass,
                              const char* pgmLabel = NULL,
                              const char* logName = NULL);
    void                reinit (const char* pgmLabel, const char* logName);

    static void         setHwId (int hwId);
    static void         setInstanceRtn (InstanceRtn* rtn);
    static void         setLogName (const char* logName);

    void                setIdentity (int identity);
    void                setPgmLabel (const char* pgmLabel, int identity = 0);
    void                setStartMsgLogged (bool val);

    // ACCESSORS
    static bool         isInitialized ();
    void                prAttrs () const;

    static PgmClassVal  pgmClass ();
    static const char*  arg0 ();
    static const char*  pgmExecDir ();
    static const char*  pgmName ();
    
    static int          hwId ();
    static const char*  logName ();
    static int          pid ();

    static bool         isDaemon ();            // Pass-throughs to rc::PgmClass
    static bool         isSingleDaemon ();

    const char*         pgmLabel () const;
    const char*         srcPrefix () const;     // For debug, log output
    int                 identity () const;
    bool                isStartMsgLogged () const;

private:
    // NOT IMPLEMENTED
    PgmAttrs (const PgmAttrs& src);
    PgmAttrs& operator= (const PgmAttrs& src);

    // MANIPULATORS
    void    processArgs (int argc, char** argv);
    void    buildSrcPrefix ();
};

inline bool PgmAttrs::isInitialized ()
{
    return s_initialized;
}

inline rc::PgmClass::Val PgmAttrs::pgmClass ()
{
    return s_pgmClass;
}

inline const char* PgmAttrs::arg0 ()
{
    return (s_arg0[0] == EOS) ? NULL : s_arg0;
}

inline const char* PgmAttrs::pgmExecDir ()
{
    return (s_pgmExecDir[0] == EOS) ? NULL : s_pgmExecDir;
}

inline const char* PgmAttrs::pgmName ()
{
    return (s_pgmName[0] == EOS) ? NULL : s_pgmName;
}

inline int PgmAttrs::hwId ()
{
    return s_hwId;
}

inline const char* PgmAttrs::logName ()
{
    return (s_logName[0] == EOS) ? NULL : s_logName;
}

inline int PgmAttrs::pid ()
{
    return s_pid;
}

inline bool PgmAttrs::isDaemon ()
{
    return rc::PgmClass::isDaemon (s_pgmClass);
}

inline bool PgmAttrs::isSingleDaemon ()
{
    return rc::PgmClass::isSingleDaemon (s_pgmClass);
}

inline const char* PgmAttrs::pgmLabel () const
{
    return (m_pgmLabel[0] == EOS) ? NULL : m_pgmLabel;
}

inline int PgmAttrs::identity () const
{
    return m_identity;
}

inline bool PgmAttrs::isStartMsgLogged () const
{
    return m_startMsgLogged;
}

} // end namespace rc

#endif /* RCPGMATTRS_INCLUDED */
