/// @file rcShmAllocBase.h
/// @brief A base class for objects wishing to be allocated in NzShm
///
//////////////////////////////////////////////////////////////////////
/// Copyright 2008 Netezza Corporation.  All rights reserved.
//////////////////////////////////////////////////////////////////////
///
/// Description:
///
/// A simple base class to provide a mechanism of creating objects in
/// netezza shared memory. It also includes typedefs to make the
/// process of creating STL containers easy
///
/// Usage:
///
/// class MyClass : public ShmAlloc<MyClass, mt_MyClass>  {
/// public:
///    // Declare standard containers of MyClass in shared memory
///    // like this -
///    typedef container<std::vector>  MyClassVector;
///    typedef container<std::list>    MyClassList;
///
///    // The SHM allocator can be accessed by MyClass::allocator_type
///
///    // Define a shared memory allocator of different kind
///    typedef typename allocator_type::rebind<Foo, mt_Foo>::type FooAllocator;
/// };
///
/// ask - 2008-06-04
//////////////////////////////////////////////////////////////////////

#ifndef RCSHMALLOCBASE_INCLUDED
#define RCSHMALLOCBASE_INCLUDED

#ifndef STL_STRING_INCLUDED
# include <string>
# define STL_STRING_INCLUDED
#endif

#ifndef STL_MAP_INCLUDED
# include <map>
# define STL_MAP_INCLUDED
#endif

#if !defined(NZSHMALLOC_INCLUDED) && !defined(TEST_STANDALONE)
# include "nzshmalloc.h"     //< for NzShmAlloc
#endif

#if !defined(NZGLOBALS_INCLUDED) && !defined(TEST_STANDALONE)
# include "nzglobals.h"
#endif

#ifdef TEST_STANDALONE
#include <memory>
#define memType_t int
template <class T, int mt>
struct NzShmAlloc {
    typedef std::allocator<T>   stl_alloc;
};
#endif 

namespace rc {

template <typename T, memType_t mt, template <class, class> class Cont>
struct ShmContainer  {
    typedef typename NzShmAlloc<T, mt>::stl_alloc   allocator_type;
    typedef Cont<T, allocator_type>                 type;
};

template <typename K, typename V, memType_t mt>
struct ShmMap {
    typedef std::pair<const K, V>                           value_type;
#ifdef UNIT_TEST
    typedef typename std::allocator<value_type>             allocator_type;
#else
    typedef typename NzShmAlloc<value_type, mt>::stl_alloc  allocator_type;
#endif
    typedef std::map<K, V, std::less<K>, allocator_type>    type;
};

template <typename T, memType_t mt>
struct ShmAllocBase {

    typedef typename NzShmAlloc<T, mt>::stl_alloc    allocator_type;

    template <typename C, int newMt=mt>
    struct rebind  {
        typedef typename NzShmAlloc<C, newMt>::stl_alloc type;
    };

    template <int newMT>
    struct rebindMt {
        typedef typename NzShmAlloc<T, newMT>::stl_alloc type;
    };

    template <template <class, class> class Container, class Member=T>
    struct container  {
        typedef typename ShmContainer<Member, mt, Container>::type   type;
    };

    template <typename K, typename V>
    struct shmmap  {
        typedef typename ShmMap<K, V, mt>::type          type;
    };

    template <class C=char>
    struct shmstring {
        typedef typename rebind<C, mt>::type  allocator_type;
        typedef std::basic_string<C, std::char_traits<C>, allocator_type>  type;
    };

    template <class C=char>
    struct shmostringstream {
        typedef typename rebind<C, mt>::type  allocator_type;
        typedef std::basic_ostringstream<C, std::char_traits<C>, allocator_type>  type;
    };

#ifndef TEST_STANDALONE
    void *operator new (size_t nbytes, void *hint=0)
    { return mem::callocBlock(nbytes, mem::AllocContext(mt)); }

    void operator delete (void *pointer)
    { return mem::freeBlock(pointer); }
#endif
};

} // namespace rc

#endif
