//------------------------------------------------------------------------------
// rcStlAllocAdaptor.h
//
// SGI-STL-style Allocation Adaptor that uses rc::Allocator.
//
//      Copyright 2003 Netezza Corporation
//
// Changes:
//

#ifndef RCSTLALLOCADAPTOR_INCLUDED
#define RCSTLALLOCADAPTOR_INCLUDED

#ifdef __GNUC__

#ifndef STL_MEMORY_INCLUDED
# include <memory>
# define STL_MEMORY_INCLUDED
#endif

namespace rc {  // recomps

// -----------------------------------------------------------------------------
// Class StlAllocAdaptor - An adaptor to convert SGI style allocator to std C++
// allocator.  Template parameter SgiAlloc should be the SGI style allocator and
// T should be the type of objects that are to be allocated. 
// 
// NOTE: SGI style allocator allocates bytes. So one should not pass anything as
// the template parameter T; char is implicitly assumed. 
//
// NOTE: Passing void as the template parameter T will run into problems as 
// reference and const_reference would not be defined 

template <typename T, typename SgiAlloc>
struct StlAllocAdaptor {
private:    
    SgiAlloc    m_underlyingAlloc;

public:
    typedef size_t          size_type;
    typedef ptrdiff_t       difference_type;
    typedef T*              pointer;
    typedef const T*        const_pointer;
    typedef T&              reference;
    typedef const T&        const_reference;
    typedef T               value_type;

    template <typename U>
    struct rebind {
        typedef StlAllocAdaptor<U, SgiAlloc> other;
    };

    StlAllocAdaptor ()  throw ()                                            {}
    StlAllocAdaptor (const StlAllocAdaptor& src)  throw ()
        : m_underlyingAlloc (src.m_underlyingAlloc)                         {}
    template <typename U>
        StlAllocAdaptor (const StlAllocAdaptor<U, SgiAlloc>& src) throw ()
            : m_underlyingAlloc (src.underlyingAlloc ())                    {}
    StlAllocAdaptor (const SgiAlloc& alloc) : m_underlyingAlloc (alloc)     {}
    ~StlAllocAdaptor () throw ()                                            {}

    // ACCESSORS

    pointer         address (reference x)       const { return &x; }
    const_pointer   address (const_reference x) const { return &x; }

    size_type       max_size () const throw () 
    { 
        // Return how many Ts can fit into 4GB 
        return size_t (-1) / sizeof (T);
    }

    // ALLOCATORS

    void construct (pointer p, const T& val) { ::new (p) T (val); }
    void destroy   (pointer p)               { p->~T(); }

    pointer allocate (size_t n, std::allocator<void>::const_pointer hint = 0)
    {
        return static_cast<pointer> 
            (m_underlyingAlloc.allocate (n * sizeof (T)));
    }

    void deallocate (std::allocator<void>::pointer p, size_t n)
    {
        m_underlyingAlloc.deallocate (p, n * sizeof (T));
    }

    // Note: These are extra methods that are not a part of the std allocator
    // interface; we support them for our own needs

    void setUnderlyingAlloc (const SgiAlloc& newAlloc)
    {
        // Change the underlying allocator; valid copy assignment should be
        // defined for it
        m_underlyingAlloc = newAlloc;
    }

    const SgiAlloc& underlyingAlloc () const 
    {
        return m_underlyingAlloc;
    }
};

// Equality (and UnEquality) for StlAllocAdaptor ... should simply resolve to
// comparison of the underlying SGI style allocator object (for same T)

template <typename T, typename SgiAlloc>
inline bool operator== (const StlAllocAdaptor<T, SgiAlloc>& a1, 
                        const StlAllocAdaptor<T, SgiAlloc>& a2)
{ 
    return a1.underlyingAlloc() == a2.underlyingAlloc(); 
}

template <typename T, typename SgiAlloc>
inline bool operator!= (const StlAllocAdaptor<T, SgiAlloc>& a1, 
                        const StlAllocAdaptor<T, SgiAlloc>& a2)
{ 
    return a1.underlyingAlloc() != a2.underlyingAlloc(); 
}

} // end namespace rc

#endif // __GNUC__

#endif /* RCSTLALLOCADAPTOR_INCLUDED */
