#ifndef RCSTRINGREF_INCLUDED
#define RCSTRINGREF_INCLUDED
/// @file rcStringRef.h
/// @brief Reference-counted shared memory string
//////////////////////////////////////////////////////////////////////
/// (C) Copyright IBM Corp. 2011 All rights reserved.
//////////////////////////////////////////////////////////////////////
/// Not really counted as much as a method of tracking references that
/// share a single allocation of a string.
///
/// The reference counting is implemented by maintaining a doubly linked list of
/// StringRef objects - all pointing to the same string.
///
/// The memory is allocated from REGION_GLOBAL.  When we link with the full
/// memory library, that is real global shared memory.  When we link with the
/// client/standalone memory library, that is heap.  But it doesn't matter to
/// the algorithm.
//////////////////////////////////////////////////////////////////////

#include "memInterface.h"

namespace rc {

class StringRef
{
    char *s;
    StringRef *prev;
    StringRef *next;

    void _unbind();
    void _assign(StringRef& rhs);
    void _assign(const char *_s);

public:

    StringRef();
    StringRef(const char *str); // copies str
    StringRef(const StringRef&);
    ~StringRef();
    StringRef& operator=(const StringRef&);
    StringRef& operator=(const char *str); // copies str

    StringRef& assign(const char *str, size_t len); // copies len bytes

    const char *c_str() const;
};

inline void
StringRef::_unbind()
{
    if (s == 0)
        return;

    if (next == this)
    {
        mem::freeBlock(s);
        s = NULL;
    }
    else
    {
        next->prev = prev;
        prev->next = next;
    }
}

inline void
StringRef::_assign(StringRef& rhs)
{
    if (rhs.s)
    {
        s = rhs.s;
        rhs.prev->next = this;
        prev = rhs.prev;
        next = &rhs;
        rhs.prev = this;
    }
    else
    {
        s = 0;
        prev = next = 0;
    }
}

inline void
StringRef::_assign(const char *_s)
{
    if (_s)
    {
        size_t len = strlen(_s);
        s = (char *) mem::allocBlock(len + 1, mem::AllocContext(mt_string_ref));
        memcpy(s, _s, len + 1);
        prev = next = this;
    }
    else
    {
        s = 0;
        prev = next = 0;
    }
}

inline
StringRef::StringRef()
{
    s = 0;
    prev = next = 0;
}

inline
StringRef::StringRef(const char *_s)
{
    _assign(_s);
}

inline
StringRef::StringRef(const StringRef& rhs)
{
    _assign(const_cast<StringRef&>(rhs));
}

inline
StringRef::~StringRef()
{
    _unbind();
}

inline StringRef&
StringRef::operator=(const StringRef& rhs)
{
    if (&rhs != this)
    {
        _unbind();
        _assign(const_cast<StringRef&>(rhs));
    }
    return *this;
}

inline StringRef&
StringRef::operator=(const char *_s)
{
    _unbind();
    _assign(_s);

    return *this;
}

inline StringRef&
StringRef::assign(const char *_s, size_t len)
{
    _unbind();

    if (_s == 0 || len == 0)
    {
        _assign(0);
    }
    else
    {
        s = (char *) mem::allocBlock(len + 1, mem::AllocContext(mt_string_ref));
        memcpy(s, _s, len);
        s[len] = 0;
        prev = next = this;
    }
    return *this;
}

inline const char *
StringRef::c_str() const
{
    return s;
}

} // namespace rc

#endif
