#ifndef RECLOC_H
#define RECLOC_H

//======================================================================
///\file recloc.h
//
// (C) Copyright IBM Corp. 2009, 2011  All Rights Reserved.
//======================================================================

#include "dpm_common.h"
#include "nztypes.h"
#include "storagerefs.h"

// Record locator packing:
//
//        Extent ID        Page   RecNum / DTID bit offset
//  +---------------------+-----+-------------------------+
//  |         39          |  5  |           20            |
//  +---------------------+-----+-------------------------+
//          (512B)        (0..24)          (1M)
//
// So long as every row includes an 8 byte Deleter TxID there cannot
// be more that 2^14 (= 128KB / 8B) rows on a page.  20 bits is also
// sufficient to hold the maximum DTID bit offset.
//
// But there's more! The Deleter TID field returned by the FPGA now
// includes a 21st bit. For an unstable page (the append tip, generally),
// the FPGA returns the 21st bit set, and a row number in place of
// a dtid bit address. This allows us to find the correct record even
// though it may have moved.
//
// We can't quite encode this information in 20 bits.
// But we can take over the page number field, in the unstable case,
// and put the page number into the available extra 6 low bits:
//
//        Extent ID        Flag   Page X      RecNum
//  +---------------------+-----+-----+-+-----------------+
//  |         39          |11111|  5  |x|       14        |
//  +---------------------+-----+-----+-+-----------------+
//          (512B)              (0..24)       (16383)
//
// The records within an unstable page still all sort together.
// Within an extent, any unstable page will always be the last page
// containing data (since it is the append tip, or the result of a
// groom, and groom will re-read and re-groom any extent containing
// an unstable page, because this is a partially-full extent from
// a previous groom operation), the sort order will actually be
// correct for the pages within an extent, too. (It is not clear
// that this is required).

inline
TRecLoc EncodeRecLoc
    /// Given a page ID and an indication of a record within that page
    /// combine the two to return a TRecLoc.
    ( PageId pid
    , unsigned rec
    )
{
    uint64 pidRep = pid.rep();
    return (rec & (1 << 20))
            ? ((pidRep | 0x1F) << 20) | ((pidRep & 0x1F) << 15) | (rec & ((1<<14)-1))
            : ( pidRep         << 20) | (rec & ((1<<20)-1));
}

inline
unsigned RecLocRec
    /// Given a TRecLoc return the record indicator component.
    ( TRecLoc recLoc
    )
{
    return (((unsigned)recLoc >> 20) & 0x1F) == 0x1F
            ? (1 << 20) | ((unsigned)recLoc & ((1<<14)-1))
            :             ((unsigned)recLoc & ((1<<20)-1));
}

inline
PageId RecLocPage
    /// Given a TRecLoc return the page ID component.
    ( TRecLoc recLoc
    )
{
    return (((unsigned)recLoc >> 20) & 0x1F) == 0x1F
            ? toPageId(((recLoc >> 20) & ~0x1F) | ((recLoc >> 15) & 0x1F))
            : toPageId(recLoc >> 20);
}

inline
bool RecLocsSamePage
    /// Given two TRecLocs return true IFF they denote different pages.
    ( TRecLoc left
    , TRecLoc right
    )
{
    return ((left ^ right) >> 15) == 0
       || (((left ^ right) >> 20) == 0 && (((left >> 20) & 0x1F) != 0x1F));
}

inline
ExtentId RecLocScanStart
    /// Given a watermark TRecLoc return an ExtentId appropriate for
    /// initializing a "watermark" scan.  This requires that the scan
    /// include the page previously captured in a watermark.
    ///
    /// TCHand_::scanListCreateFilled's lowThreshold parameter represents
    /// the lower bound and highThreshold the upper bound.  The created 
    /// scan list will contain only extents whose serial numbers are stricly
    /// greater than the lowThreshold and strictly smaller than 
    /// the highThreshold.  There is no requirement that the thresholds
    /// ExtentIds denote a existing extent nor that that extent belong to
    /// the table being scanned.
    ///
    /// Supplying TCHand_::scanListCreateFilled a value formatted as an
    /// ExtentId with an extent serial strictly less than that of the
    /// page captured in watermark ensures that that containing extent
    /// (and hence the watermarked page) is included in the scanlist.
    ( TRecLoc recLoc
    )
{
    return makeExtentId(RecLocPage(recLoc).extentSerial() - (recLoc != 0));
}

#endif // RECLOC_H
