/* (C) Copyright IBM Corp. 2001, 2014  All Rights Reserved. */

#ifndef RECORDMGR_INCLUDED
#define RECORDMGR_INCLUDED

/* Copyright (c) 2001-2005 Netezza Corporation.  All rights reserved. */

#include "blockheader.h"
#include "locktracker.h"
#include "nztypes.h"
#include "dpm_common.h"
#include "storagedefs.h"
#include "storagerefs.h"
#include "TXID.h"
#include "xpdelete.h"
#include "zmval8.h"
#include "memNewDelete.h"   // for memNewDelete template

// mutex name(s)
LKTRKR_DECLARE_NAME(MTX_DATAMGR);

class CMutex;
class CStackMutex;
class  TProjectNode;
struct TMemblockInfo;
struct dbos_drop_table_t;
struct dbos_trunc_table_t;
struct dbos_reclaim_extents_t;
struct dbos_create_table_t;
struct TIoStream;
class CTable;
class CPlan;
class SnippetNode;
struct TableDslice;
struct TDSLock;
class TInsertNode;
struct ExprContext;
class ZLib;
struct TReceiverInfo;

namespace VcompCodec { class Compress; }

enum EAppendMode {
    COMPRESSED = 0,
    UNCOMPRESSED
};

/*
 CSPUStorage Object

The storage manager provides record and table management services.
Some of the methods in this class are "glue" only (for instance, InsertRecords)
and supply a "partition" argument (i.e. PART_A or Part_B) to lower level
services.
*/


struct AppendInfo {
    EAppendMode appendMode;  // COMPRESSED or UNCOMPRESSED
    TInsertNode* insertNodeP;
    TXid Xnum;
    CPlan* planP;
    bool isGroomPlan;
    blockheader_t* blockHeaderP;
    char* firstRecP;
    char* nextRecP;
    const char* fenceP;
    TDSLock* tableDsliceLockP;
    TIoStream* iStreamP;
    VcompCodec::Compress* codecP;
    int tableNum;
    int dsid;
    CTable* tableP;
    bool inMemoryFormat;
    int recStyle;
    char* workP;
    int recordSize;
    int recordSizeCorrection;
    int uncompressedByteCount;
    int correctedSize;
    int bytesUsed;
};

class CRecordMgr : public mem::newDelete<mt_storage>
{
    const static int ACTIVE_BUFFERS_IN_INSERT_STREAM = 4;
    // Actually, we're using the spare buffer when necessary to reprime
    // a codec, so for the moment, this is 1, too.
    const static int SPARE_BUFFERS_IN_INSERT_STREAM = 1;
    friend class CStorageMgr;
 public:
    void checkForAbort(
        const TXid& Xnum,
        CPlan *plan);

    CRecordMgr();
    ~CRecordMgr();

    void DropListOfTables( dbos_drop_table_t *dropMsg,
                           int dsId,
                           int interactionId);

    ExtentId TruncateTable (
        int tblid,
        const TXid& xid,
        int dsid,
        ExtentId highestExtent,
        int interactionId,  // identifies interaction from host
        bool markExtentEmpty,
        bool commit);

    void ReclaimExtents( dbos_reclaim_extents_t *reclaimMsg,
                         int dsId );

    TIoStream *initIStreamForAppend(CTable *tableP,
                                    table_id_t redrTblId,
                                    int dsid,
                                    bool forGroom,
                                    const TXid& Xnum,
                                    TCHand tableConnection,
                                    bool startFromNewPage,
                                    bool trackNoDeletes,
                                    TIoStream **bufferToFreeP,
                                    int *numRecordsP,
                                    char **workP,
                                    blockheader_t **retHdrP,
                                    bool *isNewCodecP);

    void computeZoneMapMinMax (AppendInfo* appendInfoP,
                               char *recP,
                               bool reformat,     // true->recP is memory-format
                               const int64& rowid);

    blockheader_t *FlushBuffer ( AppendInfo* appendInfoP );

    void resetForNextBuffer(
        AppendInfo* appendInfoP );
    void appendDiskFormatToUncompressed(
        AppendInfo* appendInfoP );
    void appendMemoryFormatToUncompressed(
        AppendInfo* appendInfoP );
    void appendToCompressed(
        AppendInfo* appendInfoP );

    void CleanupTableConnection (
        unsigned dsid,
        table_id_t tableId,
        const TXid& txid,
        bool commit
        );

    int AbortAllExistingTableConnections (
        unsigned dsid,
        const TXid& txid
        );

    void AppendData (         //adds data to end of table
        const TXid& Xnum,    //transaction number
        int tableId,         //table number
        int dsid,            // dataslice id
        void * data,         //data to append (records)
        int NumBytes,        //length of data
        bool trackNoDeletes, //mark extents that have no deleted records
        bool txidspresent,   //if txids precede data
        bool reformat,       //true if "in storage format" recs
        int recStyle, // only relevent if reformat
        SnippetNode* planNode=NULL );      //

    void FlushAppendBuffer ( // flush append buffer to disk
        int tableNum,         // table to flush
        int dsid,             // data slice id
        TXid xid = TXid_NULL, // transaction id (for scan only)
        CPlan *planP = NULL,         // plan object
        PageId pageId = NULL_PID, // which page (NULL_PID=any)
        bool updateNewStuff = false, // update newstuff pointer
        int parentId = 0);    // thin table's base table

    void DiscardAppendBuffer( //toss append buffer (mostly for reclaim)
        TableDslice *tds,     //table dslice with stream to flush
        bool forInsert);      // true=>flush insert stream, false=>flush groom stream

    void DeleteRecords(         // marks records deleted
        const TXid& Xnum,       // transaction num
        CTable *tblP,           // table object
        int dsid,
        TDeleteNode *nodePtr,
        bool forDelete,         // false=UPDATE stmt
        CPlan* plan=NULL );

    void FetchRecordsByRecordAddr(
        TRecLoc* recLocs,       // records to fetch
        int numrecs,            // number of records
        CTable *srcP,           // disk table object read from
        int dsid,               // dataslice id
        int srcTblId,           // since srcP is a clone, we need to pass the tableId
        CTable *dstP,           // memory table object to write to
        TProjectNode *pNode,    // which columns to get
        ExprContext *econtextP, // pg expr context for pg eval code mem allocs
        const TXid& xId);       // for CURRENT_TIME, etc, evaluation

    void InsertRecords (        // inserts records at end of table
            const TXid& xId,    // transaction id
            table_id_t tblId,          // table number
            int dsid,           // dataslice id
            bool trackNoDeletes,//mark extents that have no deleted records
            TMemblockInfo *memblock,
            TReceiverInfo *receiverInfoP = NULL);

    void InsertDiskFormRecords (//inserts disk-ready records at end of table
            const TXid& xId,    // transaction id
            table_id_t tblId,          //table number
            int dsid,           // dataslice id
            bool trackNoDeletes,//mark extents that have no deleted records
            int numBytes,       //number of bytes
            void *records,      //record data
            SnippetNode* planNode=NULL );

    void PreCommit(             // Phase 1 of two-phase commit
        const TXid& xId,        // transaction id to commit
        int dsid,               // dataslice id
        int numTables=0,        // number of tables to FlushAppendBuffer
        const int *tables=NULL, // the table ids
        uint64 *EOTs = NULL);   // where to return the end of table block for each table

    void MarkTxAbort(        // remove aborting tx from append buffer
        const TXid& xId,
        bool hasGroomTableEffect,
        int tblId );

    void MarkTxComplete(        // remove aborted/committed tx from append buffer
        const TXid& xId,
        int tblId );

    PageId RollbackInserts ( // Rollback inserts between 'start' and 'end'
        TXid const& txid,
        int tableId,
        int dsid,
        PageId const start,
        PageId const end);

    void RollbackPages(  // Undo any updates to a table that were part of this transaction
        const TXid& txid,     // transaction number
        int tableId,          // table id
        int dsid,             // dataslice id
        const PageId *pages,
        int numPages);

    // bulk rollback on host restore
    void RollbackTransactionsSince(
        int tableId,
        int dsid,             // dataslice id
        const TXid& baseXid,
        int interactionId);   // identifies interaction from host

    void GetEofBlockAndRec(CTable *table,      //table to return values for
                           int dsid,           // dataslice id
                           int *block,         //(output) eof block
                           int *rec);          //(output) recnum of last rec

    void CreateTable (        //creates a table
        dbos_create_table_t *crtMsg, //message from dbos
        int dsid,
        int interactionId);   // identifies interaction from host

    CMutex *GetDataMutex(int dsid);

    TRowId GetPrevRowid(int dsid) const;    // m_prevRowid accessors
    void SetPrevRowid(int dsid, TRowId val);
    TRowId * GetPrevRowidP(int dsid);

    int partialRollbackBuffer(const TXid& badTX, CTable *table, blockheader_t *blockp);
    void partialRollback(struct partial_rollback_msg_t *rollback, int interactionId);

    void discardIOStreams(int tableNum, int dsid, bool forGroom);

    void LazyFreeCodec(
        VcompCodec::Compress* codecP,
        int tableId,
        int dsid);

    void RefreshLockingProtocol(void);

 private:
/********************* PRIVATE ****************************************/

    TIoStream* getAppendResources (
        int TableNum,
        int dsid,
        bool forGroom,
        const TXid& Xnum,
        CStackMutex& dsLock,
        TDSLock& tdsLock);

    void sanityCheck(
        TIoStream* iStreamP,
        CTable* tableP,
        int dsid);

    void initCodecForAppend(
        CTable *tableP,            // table with which this codec is associated
        int dsid,                  // dataslice id
        bool forGroom,             // true if append is for groom
        void* dataP,               // source buffer for records
        size_t byteCount,          // bytes in source buffer
        bool emptyBlock,
        blockheader_t *hdrP,       // pointer to append buffer for table/partition
        const TXid& Xnum,
        bool mustReprime,
        int* reprimedByteCountP ); // uncompressed byte equivalent returned by reprime (or 0)

    VcompCodec::Compress* getCodec(
        int tableId,
        int dsid,
        bool *isNewCodecP );      // true -> allocated a new codec

    void removeTablesFromPartition (
        dbos_drop_table_t* dropMsg,
        int dsid,
        int interactionId);

    void showInserters (
        CTable *table,
        int dsid);

    void checkAllocation(
        CTable *table,
        int dsid,
        int block,
        void *stream);

    bool rollbackAppendBuffer(
        CTable *table,       // table to rollback
        int dsid,            // dataslice id
        const TXid& xId);    // transaction ids to roll back

    bool rollbackRecordsInBuffer(
        CTable *table,           // table to rollback
        blockheader_t *blockP,   // buffer pointer
        const TXid& xId,         // transaction ids to roll back
        bool verbose );          // describe results for this buffer in printf

    // uninsert/undelete one record
    bool undoRecordInsertDelete (
        const TXid& xId,         // transaction ids to roll back
        CTable *table,           // table object
        char *rechdr);           // pointer to start of record on block

    bool rollbackBufferSince(
        CTable *table,           // table to rollback
        blockheader_t *blockP,   // buffer pointer
        const TXid& baseXid);    // roll back all transactions > baseXid

    int addRecordToBlockBuffer(
        int recStyle,       // describes decoding style (host/spu/compressed...)
        CTable* pTable,     // table object associated with record and block buffer
        record_t* pRec,     // array of bytes making up the record
        char* pBuf,         // block buffer into which to put record
        int availableBytes, // bytes available in buffer
        const TXid& Xnum,   // current transaction ID
        int* pBytesInInputRecord, // bytes consumed in input buffer
        int dsid);          // current dataslice id

    int addCompressedRecordToBlockBuffer(
        CTable* pTable,     // table object associated with record and block buffer
        record_t* pRec,     // array of bytes making up the record
        char* pBuf,         // block buffer into which to put record
        int availableBytes, // bytes available in buffer
        const TXid& Xnum,   // current transaction ID
        int* pBytesInInputRecord); // bytes consumed in input buffer

    /// \brief Allocate and initialize the rollback buffers
    ///
    /// This method is called from the constructor to preallocate
    /// the rollback buffers. The buffers are preallocated to avoid
    /// error in the rollback path that could cause a failure.
    void initRollbackBuffer();

    /// \brief Free the rollback buffers
    /// \sa CRecordMgr::initRollbackBuffer
    void freeRollbackBuffer();

    /// \brief Retrieve the buffer to use for rolling back data changes
    /// \return A preallocated buffer at the specificed dataslice index.
    ///
    /// Ensure one user of a given rollback buffer at a time!
    /// These used preallocated rollback buffers are used by the SpuBatch process
    /// \sa CRecordMgr::initRollbackBuffer
    diskPage_t* getRollbackBuffer(unsigned int dsdex);  ///< Dataslice index

    /// \brief Reestablish the buffer to use for rolling back data changes
    /// Ensure one user of a given rollback buffer at a time!
    /// These used preallocated rollback buffers are used by the SpuBatch process
    /// \sa CRecordMgr::initRollbackBuffer
    void putRollbackBuffer( unsigned int dsdex          ///< Dataslice index
                          , diskPage_t* bufP            ///< Pointer to rollback buffer
                          );

    /// \brief Determine whether fine-grained or legacy locking protocol should be used
    /// \sa CRecordMgr::refreshLockingProtocol
    bool useFineGrainedLocking(void) const { return m_useFineGrainedLocking; }

private:
    CMutex *m_sliceLock;    ///< array of locks, one per data slice
    TRowId *m_prevRowId;    ///< used for compressing out rowNums; entry per-dataslice
    diskPage_t** m_rollbackBuffer;  ///< array of preallocated buffers for rollback, one entry per data slice
    bool   m_useFineGrainedLocking; ///< true (>0) if locking protocol should use table dataslice lock instead of datamgr lock
};

void uncompressRecStoreBlock ( // uncompress zeropacked or Zlib'ed block
    uint32 blockFormat,
    char *outBlock,
    uint32 *outBlockSize,
    char *inBlock,
    uint32 inBlockSize);

// Uncompressed Disk Record Header Structure
struct diskRecHdr {
    uint64 rownum;
    uint64 creatxid;
    uint64 deletexid;
};

#endif
