#ifndef RECORDSTORE_INCLUDED        // don't include multiple times
#define RECORDSTORE_INCLUDED

#include "nzglobals.h"
#include "messages.h" // for dbos_msg_header_t and spu_msg_header_t
#include "error.h"
#include "sdlist.h"     // dlink_t and tools
#include "rcAtomic.h"
#include "sparse_bitmap.h"
#include "locktracker.h"

/*
 * Record stores are a programmable collection of data pages.  The data
 * pages actually hold the records. The record format is defined by the
 * associated CTable object.  Some specialized record stores contain
 * special purpose data not usually associated with a CTable (e.g.,
 * THashInfo).  Others are only vaguely connected to a CTable describing
 * the record shape.  After many years of creative extension of the
 * concept, entropy is fierce.
 *
 * When the goal is extraction of fields and other record shape related
 * information, you probably need both (1) a RecordStore method to find a
 * data buffer AND (2) a CTable method to supply the record format
 * information to locate and process the byte(s) in the data buffer.  In
 * most high-frequency use cases, this original call-per-datum interface
 * has been replaced by "generated code" that generates custom struct
 * declarations for the table record shape, and in-line C++ code for
 * manipulating data fields.
 *
 * The access methods have evolved quite a bit from the initial days of
 * Netezza.  This is the third major generation of the
 * CRecordStore. Understanding the history may permit one to understand
 * the quirks in the interface.  Also, please remember that the popular
 * APIs in this header file have rarely been changed.
 *
 * The zero-th development version of the CRecordStore assumed that the
 * data set could be accomodated 100% in global shared memory. There was
 * no way to use disk storage to create a larger data set.  This
 * limitation was not merchantable, and lead to two developments:
 *
 * (a) "Stream processing", where algorithms were taught to only look at
 * a 'point' in the file, often only a very few records, or a very few
 * blocks of records.
 *
 * (b) SWAP storage on the SPUs. This was a primitive, non-hierarchical
 * file system that permitted on-disk storage of pages of record data.
 * Every transient data page was assigned a location in SWAP storage on
 * the disk.
 *
 * There was a cooperatively scheduled, non-preempting swap system that
 * could remove to disk some or all of a query`s data pages from SPU
 * memory temporarily. This was done to mitigate cases where physical
 * memory was over-committed to the aggregate needs of all queries in
 * progress.
 *
 * The marriage of stream processing and swap storage resulted in "disk
 * based" data base operations where large volumes of intermediate data
 * were stored on the disk to wait for future processing steps.  Said
 * steps only requiring a relatively small portition of the conceptual,
 * entire data set to be in memory at any one moment.
 *
 * The way the swap storage worked also caused some objects to be
 * creatively stuffed into a CRecordStore / TMemblockInfo envelope so
 * that the swap system could recover the memory for said objects during
 * periods of memory over-commitment.
 *
 * Hashed organizations (see CHash class) also used CRecordStore /
 * TMemblockInfo as a "sub- allocator" that did need to interact with
 * global shared memory well too often.
 *
 * The third and current version of the CRecordStore is designed to be
 * hosted on Linux.  Linux provides demand paging, data caching, real
 * file systems, and security controls; none of which were available from
 * the Nucleus RTOS formerly used on the SPU.  We have chosen to use the
 * memory-mapped style of file access to bring data pages into the query
 * process memory for operations.
 *
 * That is the ideal -- an object and methods for a single dataslice of
 * but one query.  Yet somewhere along the line, that was stretched into
 * the stored virtual table with record store, which can collect system
 * operational statistics from any process wishing to contribute.  This
 * seemed easy to do at the time, as the underlying storage was Global
 * Shared Memory.  This usage persists, though many parts of CRecordStore
 * support are now specialized for stored virtual tables.  Global sharing
 * and access by any executable also require that the implementation not
 * use virtual functions and object layering to avoid the executable-
 * specific C++ Virtual Function Table (VFT), with compile-time pointers
 * to methods.
 *
 * The CRecordStore system has three types of storage it knows how to use:
 * (a) global shared memory
 * (b) pure per-process memory [usually not used]
 * (c) mapped pages from operating system files.
 *
 * [There is tension in disk volume space management between (b) and
 * (c). The former is backed by space in Linux swap partitions (or swap
 * files). The latter uses backing space managed by file systems.  It is
 * relatively hard to shift disk space between file system and swap
 * partition usages.  NPS prefers to use Linux large (>> 2GB) files. We
 * have more space and controlled viewing of the file data.  Paging
 * partition space is typically reserved for structures managed by the
 * operating system, like the task stack.]
 *
 * There are also two access paradigms:
 * (a) the entire data set is visible, as with a hashed store
 * (b) the streaming model, only a point in the stream being visible.
 * OK, I lied, there is a third,
 * (c) a pseudo-random access method. Its use must be carefully
 * programmed lest one create an O(N**2) [or worse] algorithm where the
 * units are multi-millisecond disk I/Os.  Locality of reference counts
 * for a lot in performance evaluations.
 *
 * And finally, usage models.
 *
 * (a) intermediate data set storage used by one snippet / event process
 * (or other actor) at a time.
 *
 * (b) query shared. The complication is that the data may be wanted
 * concurrently by two cooperating parts of the execution.  For example,
 * the "dbs ReturnNode" has data produced (in DBOS Event) and consumed
 * (in Postgres) in parallel.
 *
 * (c) the stored- virtual model, where any process may contribute to a
 * globally shared resource.  There are a number of legacy hacks of
 * CRecordStore that follow the stored- virtual model for now, as that is
 * the closest to the original all-in-memory picture.
 *
 * Focus on one use case at a time to gain understanding.
 *
 * The use case to start with is the create - fill - close // re-open -
 * read - remove model which is used for what you probably think of as a
 * transient intermediate result.  There are two famous instances of
 * "create-fill- close", being (1) a locally produced data set being
 * absorbed by a SaveTempNode AND (2) a "distributed" data set landing on
 * a SPU for later use.  "re-open - read - delete" is the data set`s
 * reason for being created.  The data can be read for statistics
 * gathering, a join`s inner tables, a fabric scan, etc.
 */

#define NZ_BEST_REC_STORE_SIZE 131072   // a/k/a NZ_PAGE_SIZE

class CShMemFile;
class CPlan;
class CTable;
class CHash;
struct TDPIBlock;
struct CTransTableCtx;
class SpuJob;
struct job_dslice_t;

extern "C" {
    int throwError(int code, const char *buf, const char *obj);
}

// The following constants are the valid values for the m_blockFormat
// member of the TMemblockInfo struct.  They are used mostly for compressed
// unloading and reloading.
#define RECBLOCK_FORMAT_NONE 0      // ordinary binary
#define RECBLOCK_FORMAT_ZEROPACK 1  // zero packed bytes[]
#define RECBLOCK_FORMAT_LZW 2       // lzw bytes[] (NOT CURRENTLY SUPPORTED)
#define RECBLOCK_FORMAT_ZLIB 3      // zlib bytes[]
#define RECBLOCK_FORMAT_BZ2 4       // bzip2 bytes[] (NOT CURRENTLY SUPPORTED)
#define RECBLOCK_FORMAT_REPL 5       // replication format
#define RECBLOCK_FORMAT_MAX 5       // for reload corruption check

/**
 * TMemblockInfo is the fundamental container for internal records
 * that (may) travel on wires or move to temporary storage devices.
 *
 * To test data integrity in SPU temporary storage and in-RAM caching,
 * a checkSum may be maintained on all fields except 'writeAddr' and
 * 'next'. The checksum implementation believes these fields are
 * contiguous. -dk Jan 2004
 *
 * WARNING: structure saved in compressed external files.  If it changes,
 *          bump EXTTAB_HEADER_VERSION and assign a bug to Scott to write
 *                   code for backward compatibility.
 */

#include "nzpack_x86_on.h"

struct page_header_t {
    int m_jobId;
    union {
        uint32 m_rsId;
        int m_tblId;
    } m_storageId;
    uint32 m_rsDsId;
    uint32 m_pageId;
};

struct TMemblockInfo {
    union { // this must be first (56 bytes - see BYTES_IN_NULLSTORE below)
        dbos_msg_header_t toSpu;    //!< from host
        spu_msg_header_t toDbos;    //!< from spu
        page_header_t toDisk;
    } msghdr;
    uint8 m_blockFormat;   //!< format control info...
    uint8 m_specialUseByte;  // this byte is used for special case when SPU is parsing replication format mem blocks
    uint8 unused[2];       //!< alignment and future use
    int m_size;   //!< size allocated (DEFAULT_RECSTORE_SIZE typical)
    int used;     //!< bytes used
    int avail;    //!< bytes left in store
        // begin checksum holiday
        // The next 8 bytes of this object have an awkward layout
#if OLD_RS_NEXT_NEED
# if __LP64__
    TMemblockInfo *next;// 64-bit (8-byte) pointer for block list linking
# else

    int XwriteAddrX;    // DISUSED (was: where new records get written to)
    TMemblockInfo *next;// 32-bit pointer for block list linking
# endif // __LP64__
#else
    // Normal CRecordStore environments do not believe in pointers inside TMemblockInfo.
    // However, there are some "Detached Duty" situations where chaining thru next
    // is hanging on with tenacity.  GROAN! --dk 4/2008
    int XwriteAddrX;    // DISUSED (was: where new records get written to)
    int XnextX;         // DISUSED (was: next in chain)
#endif // OLD_RS_NEXT_NEED
        // end checksum holiday
    int m_numRecords;   //!< how many record in this block
//  char bytes[0];      // Not ISO C++ ; use MBI_ZERO_BYTE

    int i_actualSize() const {
        return sizeof(TMemblockInfo) + used;
    }
};

#include "nzpack_x86_off.h"

// Find byte zero of the data area of the TMemblockInfo page.
#define MBI_ZERO_BYTE(m) ( (char*)((TMemblockInfo *)(m) + 1) )

// Find the next-to-be-written byte in the data area of the
// TMemblockInfo page; a/k/a fence byte when the page is being read.
#define MBI_WRITE_BYTE(m) ( MBI_ZERO_BYTE(m) + (m)->used )

struct TNominalVarRec {
    uint16 size;
    uint8 nulls[2];
    uint32 stuff;
};

enum { DPIBlockDim=80 };      // Size of the "m_item" array in a TDPIBlock.

enum EDPIUsage {
    eFDPmagic= 0xFD504147,    // FD-PAG (2 hex, 3 ASCII) for m_FDPMagic
    eDPvirgin =0x80000000,    // Initialized but unassigned
    eDPassigned=0,            // allocated, but info being worked on
    eDPdisused=0x20000000,    // Once was used, but now pining for new assignment
    eDPshared=0x08000000,     // Global shared
    eDPpure=  0x02000000,     // Local memory only
    eDPfile=  0x01000000,     // File backed (m_filePageNum valid), subtypes:
    eDPmapped=  0x1,          // . mmap-ped into memory
    eDPhotpage= 0x2           // . hot page for additions, but not yet committed to file
};

struct TDataPageInfo {
    TDataPageInfo *m_nextDPI;// Link to next TPageInfo (of same ilk)

    // Current in-memory location (or 0).  This points to a 128K area of memory.
    // It could be in various memory classes -- local, global, mapped from disk.
    // (If mapped from disk, it is of course on a 4K page boundary.)  This is
    // given as a "TMemblockInfo", but that's just the header.
    TMemblockInfo *m_memoryAddr;

    // Page number of the VIRTUAL page (128K) on disk.  If the file is
    // compressed, this is the "virtual" number.  Not the same as the physical
    // location of the bits on the disk.
    uint32 m_filePageNum;

    int m_DPIusage;       // Notations about usage of this DataPage (EDPIUsage)
};

#define INIT_THIS_MRG (0xFFFFFFFFUL)

enum {
    BYTES_IN_NULLSTORE = sizeof(TMemblockInfo)
};

#define GET_NEXT_REC_SIZED(rs,rec,kind,block,size)                        \
  (((kind*)((char*)rec + (size)) < (kind*)MBI_WRITE_BYTE(block)) ? \
        (kind*)((char*)rec + (size)) :                     \
        (block = (rs)->NextBlockInMem(block), (kind*)MBI_ZERO_BYTE(block)))

#define GET_NEXT_VAR_REC(rs,rec,kind,block)                        \
  (((kind*)((char*)rec + (((TNominalVarRec*)rec)->size)) < (kind*)MBI_WRITE_BYTE(block)) ? \
        (kind*)((char*)rec + (((TNominalVarRec*)rec)->size)) :                     \
        (block = (rs)->NextBlockInMem(block), (kind*)MBI_ZERO_BYTE(block)))

#define GET_NEXT_VAR_REC_BLOCKNUM(rs,rec,kind,block,setBlock,incBlockNum)   \
  (((kind*)((char*)rec + SIZE_WORD_CHECK(rec)) < (kind*)MBI_WRITE_BYTE(block())) ? \
        (kind*)((char*)rec + (rec->size)) :                     \
        (setBlock( rs->NextBlockNumInMem(block(), incBlockNum()) ), (kind*)MBI_ZERO_BYTE(block())))

#define LAST_REC(kind) ((kind*)BYTES_IN_NULLSTORE) // result of offsetting into null block

// in genc LOOP_RECS is the equivalent to LOOP_RECS_SIZED here
#define LOOP_RECS_SIZED(rs,rec,kind,block,size)                 \
        if ( block && block->used )                             \
            for ( rec = (kind*)MBI_ZERO_BYTE(block);            \
                rec != LAST_REC(kind);                          \
                    rec = GET_NEXT_REC_SIZED(rs,rec,kind,block,size) )

#define LOOP_RECS_VARIABLE(rs,rec,kind,block)                   \
        if ( block && block->used )                             \
            for ( rec = (kind*)MBI_ZERO_BYTE(block);            \
                rec != LAST_REC(kind);                          \
                    rec = GET_NEXT_VAR_REC(rs,rec,kind,block) )

#define GET_NEXT_NZREC(rs,rec,kind,block,fixedSizeElseZero) \
  ( (kind*) ( ( ( ((char*)(rec)) + ((fixedSizeElseZero)?(fixedSizeElseZero):((TNominalVarRec*)rec)->size) ) < MBI_WRITE_BYTE(block) ) ? \
                ( ((char*)(rec)) + ((fixedSizeElseZero)?(fixedSizeElseZero):((TNominalVarRec*)rec)->size) ) : \
                ( block = (rs)->NextBlockInMem(block), MBI_ZERO_BYTE(block) ) \
            ) \
  )

#define LOOP_NZRECS(rs,rec,block,fixedSizeElseZero)        \
    record_t *rec = 0;                                  \
    if (block && block->used)                           \
      for (rec = (record_t *)MBI_ZERO_BYTE(block);      \
        rec != LAST_REC(record_t);                      \
          rec = GET_NEXT_NZREC(rs,rec,record_t,block,fixedSizeElseZero))

#define UPDATE_WRITE_ADDR(m,in_cnt)    \
    do {                               \
        register int cnt = in_cnt;     \
        m->used += cnt;                \
        m->avail -= cnt;               \
        if ( m->avail < 0 ) {          \
            throwError(NZ_FATAL_ASSERTION,"Buffer overrun in UPDATE_WRITE_ADDR","" ); \
        }                              \
    } while (0)

struct pgacct_t
{
    pthread_mutex_t pg_lock;     // mutex
    unsigned int pg_nr_resident; // number of resident pages
    unsigned int pg_nr_dirty;    // number of dirty pages
    sparse_bitmap pg_resident;   // resident pages bitmap
    sparse_bitmap pg_dirty;      // dirty pages bitmap
    sparse_bitmap pg_readahead;  // readahead pages bitmap
    int pg_punch_pos;            // when != -1, last page num that was hole punched
};

// broadcast table recordstore page accounting
struct bcast_rs_pgacct_t
{
    pgacct_t brs_pgacct;
    unsigned int brs_nr_ds;   // number of per-dataslice entries
    unsigned int brs_lo_pos;  // lowest mapped pos
    struct bcast_ds {
        sparse_bitmap brs_ds_mapped; // per-dataslice mapped pages bitmaps
    } brs_ds[0];
};

//! \brief Missing doxygen comment.
class CRecordStore {

    friend class CTable;
    friend class CVirtualTableScan;
    friend class SnippetResultCache;

public:

    enum ERStype {
        eRSunknown = -2,
        eRSderiveIt=0,              // Generic call is passing table...
                                    // Look at table and pick...
        eRStransientTable = 1,
        eRSHash,
        eRSvirtualTable,
        eRSBitVector,
        eRSSortSpill,
        eRSTwoParty,
        eRStempWorkWindow           // one snippet scratchpad storage
    };

    enum EStorePolicy {
        eSPunknown = -2,
        eSPsame=-1,                 // same policy as before
        eSPclosed=0,                // store detached from plan, to be reused?
        eSPrandomGrow=1,            // traditional "all-in" (growing on end?)
        eSPsequentialGrow=2,        // e.g. table loader (needs only 'hot page')
        eSPwindowView=3,            // snippet-controlled reading view on some/all rows
        eSPsequentialRead=4,
        eSPrandomRead=5,            // For disk-based merging, etc. via xxxStoreBlock()
        eSPMustBeLast
    };

    enum EFilePolicySetting {
        eSFPnone=0,                 // Give no file advice (starting state)
        eSFPdefault=1,              // Apply registry defaults for m_currentPolicy
        eSFPdoAll=2                 // Apply enhanced "all pages" policy
    };

 public:

    record_t **m_pRecords;      // individual record pointers (optional)
    uint64 m_NumRecords;        // records stored
    int m_allocatedRecSlots;    // allocated number of record pointers
    int m_meteredRecSlots;      // when m_NumRecords reaches m_meteredRecSlots then m_meteredNumOfPages will be reported for recPtrArr
    CHash *m_hash;              // hash for finding individual records (?vs node)
    CTable *m_ModelTable;       // source/description of 'records' in store
    int m_pageCnt;              // number of pages in visible window

    // m_transBytes is maintained only for download and savetemp types of stores.
    uint64 m_transBytes;        // Byte count for transients.

    // page accounting data
    union {
        // for non-broadcast recordstores
        struct {
            pgacct_t m_rsPgacctS;     // resident/dirty page info
            sparse_bitmap m_rsMapped; // mapped pages bitmap
        };
        // for broadcast recordstores
        struct {
            bcast_rs_pgacct_t *m_rsBcastRsPgacct; // broadcast recstore shared page accounting
            unsigned int m_rsBcastRsPgacctIdx;    // broadcast recstore dataslice index
        };
    };
    unsigned int m_rsNrAnonPages;     // number of anonymous pages
    pgacct_t *m_rsPgacct;
    struct { // page steal state
        pthread_mutex_t m_rsStealLock; // protects page steal state
        pthread_cond_t m_rsStealCond;  // for signaling page steal waiters
        unsigned int m_rsStealCnt;     // count of outstanding page stealing ops
        unsigned int m_rsStealWaitCnt; // count of page steal waiters
    };
    atomic_t m_rsNrSync;              // for virtab logging: total number of syncs

    int m_rsJobLinked; // 0 = not linked, 1 = linked to job, 2 = linked to job and job dslice
    int m_rsStealableList; // stealable list index, or -1 if not linked
    unsigned int m_rsId;  // recordstore id, unique within job-dataslice
    SpuJob *m_rsJob;
    job_dslice_t *m_rsJobDslice;
    dlink_t m_rsJobLink; // spu job recstore list link
    dlink_t m_rsRsStealableLink; // stealable recstore list link
    bool m_rsFinalUse;     // when true, this record store will not be reopened after closed
    bool m_rsWillDrop;     // when true, this record store will be dropped
    bool m_rsHolePunchEnabled;
    bool m_rsJobReferenced; // If made a reference then release while destroying

 private:

    ERStype m_storeType;          //  what type of storage for 'records'?
    volatile EStorePolicy m_currentPolicy; // current activity vs this store
    int m_pageBytes;              // standard allocation size for this store
    int m_pageCapacity;           // usually same value as m_pageBytes
                                  // but may be lower; used to "size" TMemblockInfos for this store
    nzotuid_t m_recArrOwner;      // process that allocated m_pRecords Array
    nzotuid_t m_fileOwner;        // process handling m_file
    uint32         m_extendPageNum; // Next page for extending m_file
    uint32         m_firstSeqNum; // Sequence number of first page in window
    uint32         m_lastSeqNum;  // Sequence number of last page in window
    int            m_RSdsid;      // slice of m_ModelTable this RS is for
    TDataPageInfo *m_firstWindow; // Begin window ; first on chain
    TDataPageInfo *m_lastWindow;  // End window ; last on chain
    TDataPageInfo *m_loopCursor;  // page a loop-thru is pointing at

    CShMemFile    *m_file;        // File containing backing pages
    bool           m_fileCompressed; // File content is compressed
    unsigned int   m_readFlushCount;
    unsigned int   m_readPunchCount;
    atomic64_t     m_rs_shm_physFileSize; // Mirror value from m_file, accessible when the file is closed
    atomic64_t     m_rs_shm_fstatTestSize; // Test value to determine that physFilesize is the same as current physical file
                                           // determined by performing fstat on on disk file

    TDPIBlock     *m_DPIBList;    // bulk allocations of arrays of TDataPageInfo (for cleanup)
    TDataPageInfo *m_unusedDPI;   // LIFO of TDataPageInfo that are pure empty / never-assigned
    dlink_t        m_storeLink;   // special list for "lazy cleanup" of active CRecordStore objects
    CPlan         *m_storeLinkPlan; // plan pointer if this recordstore is linked to a plan's storeList
    nzotuid_t      m_spuImplicitClose; // flag of process wanting earlier
                                       // lazy close
    void *operator new(size_t);
    void operator delete(void *);
    void cleanUpResources (int storeLinkPlanId, int storeLinkPlanSnipNum) throw ();
    static CRecordStore * makeRecStore(int size, int capacity, CTable *table, CPlan *plan, ERStype type, EStorePolicy policy);
 public:

    static void RecordStoreInitPreFork();
    static void RecordStoreInitConfig();

    unsigned int RsPageBytes() const { return m_pageBytes; }
    nzotuid_t RsFileOwner() const { return m_fileOwner; }
    bool RsFileCompressed() const { return m_fileCompressed; }
    EStorePolicy RsPolicy() const { return m_currentPolicy; }
    void RsIncrementReadFlushCount(int n) { m_readFlushCount += n; }
    void RsIncrementReadPunchCount(int n) { m_readPunchCount += n; }

    void _vt_transtbl_iostat_log(const CShMemFile *pSHM, int planIdArg, int snippetNumArg) const;

    static CRecordStore *CreateRecordStore(   // Call for any new CRecordStore object sub-type
                 ERStype type,
                 CTable *parentP,
                 CPlan *planP,
                 int size,
                 int capacity,
                 EStorePolicy policy ,
                 const CTransTableCtx& transTableCtx);
    static CRecordStore *CreateRecordStoreNoPlan(   // variant for plan-less environment (datarcv)
                 ERStype type,
                 int dsId,
                 CTable *associatedTable,
                 EStorePolicy policy,
                 const CTransTableCtx& transTableCtx);
#if FOR_SPU
    static void _make_rs_not_stealable(CRecordStore *rs);
#endif
    static CRecordStore *Reopen(CPlan *plan, CRecordStore *pRS, bool forceImplicitClose, bool readOnly = true);
    static void DestroyRecordStore(CRecordStore *pRS);

    static CRecordStore *CreateBitVectorStore(CPlan *plan, const uint64& numBits);

 private:
    // "Private" is trying to remind others to use Factory/Destroy.
    // Alas, most calls are from a friend, CTable, so the compiler is rarely upset.
    CRecordStore(int size,
                 int capacity,
                 CTable *parentP,
                 ERStype type,
                 EStorePolicy policy);
    ~CRecordStore ();

    // weird copy constructor for CloseDownloadRecordStore (HACK)
    CRecordStore(const CRecordStore&);
    // not implemented
    CRecordStore& operator=(const CRecordStore&);

 public:
    char *VerifySpace( // Increase size of store if needed
        int need);

    char *AllocateSpace( // Increase size of store if needed
        int need);

    char *VerifyRingSpace( // Verify space in ring buffer or grow buffer
        int need,       // record size
        size_t ringSize);  // total public size of ring buffer

    void AdjustRingSpace ( // preallocate or grow/shrink vt ring buffer
        size_t ringSize);

    void ReclaimRingSpace (); // reclaim empty virtab blocks

    void setPageRecCount(uint cnt);

    void updateWriteAddr( // move used/avail forward
        int cnt);

    int64 totalSize() const;    // return total used in all blocks

    int totalAllocatedSize() const;    // return total allocated in all blocks

    void Reset();  // Make record store empty, without deleting it

    TMemblockInfo *DetachData(ERStype fromType);

    void EndPageAccounting(); // credit current page and stop any further accounting

    // This is called from far away, so we make it a static method to give it a quasi-namespace.
    static void InitRecBlock(TMemblockInfo *m, int size);

    void SetWindow(     // Set the window to comprise indicated pages
        int lowBlock,   // first page we need to keep in memory
        int highBlock); // last page to keep in memory

    TMemblockInfo *UpdateWindow(
        int lowBlock,   // first block we need to keep in memory
        int blockNum,   // (high?) block to return a memory address for
        TMemblockInfo *oldBlock);  // an appendix (validated with NZ_VERIFY, otherwise unused)

    void DeleteWindowBlock(  //WindowAggrNode special, DEPRECATED
        int lowBlock,
        int newLowBlock);

    TMemblockInfo *RewindToFirstBlock(int numBlocks);

    void CloseStore();

    record_t *gencVerifySpace(int size, rdex_t recDex);


    TDataPageInfo *i_firstWindow() const { return m_firstWindow; }

    TMemblockInfo *i_firstBlock() const;   // no longer trivial inline

    TMemblockInfo *i_lastBlock() const;   // no longer trivial inline

    // Iterator-style versions that interact with m_loopCursor
    TMemblockInfo *i_beginBlock();
    TMemblockInfo *i_beginBlock(TMemblockInfo *trackBlock);
    TMemblockInfo *i_endBlock() const;  //wink-nudge: will be zero/NULL

    TMemblockInfo *NextBlockInMem(      // next mem block in begin->end list
        TMemblockInfo *pBlock);

    // Get next block @ cursor (without changing window, as would UpdateWindow())
    TMemblockInfo *NextBlockNumInMem(
        TMemblockInfo *pBlock,
        int blockNum);

    // Is the RS already open for reading?
    // This is used when duplicate single-row subqueries are elimitated
    bool isOpenForReading() const;

    ERStype i_type() const {
        return m_storeType;
    }

    //! \brief Accessor for m_ModelTable.
    CTable *modelTable() const {
        return m_ModelTable;
    }

    int i_blockCnt() const {
        return m_pageCnt;
    }

    // CTable refugees begin

    uint64 i_NumRecords() const {
        return (this==0) ? 0 : m_NumRecords;
    }
    int i_allocatedRecSlots () const {
        return m_allocatedRecSlots;
    }

    record_t *i_recordPtr (int recNum) const {
        NZ_ASSERT(m_pRecords);
        NZ_ASSERT((recNum < m_NumRecords) && (recNum >= 0));
        return m_pRecords[recNum];
    }

    int DataPages() const;

    uint64 i_transBytes () const {
        if ( this != 0 ) {
            NZ_ASSERT (m_storeType == eRStransientTable);
            return m_transBytes;
        }
        return 0;
    }

    record_t *RecordZeroP() const; // Address of the existing zero-th record

    void *FieldAt( // Return address of field given recDex & fdex
        int recNum,     // record pos in m_pRecords array (0-N)
        int field) const;     // field dex

    void DoneWithExecThings(); // temps that stick on during plan

    void FinishAddingRecord( // Bind in record buf gotten by VerifySpace call
        record_t *rec,      // the record buffer as returned by VerifySpace
        int recLen);    // the length of the buffer

    // called by generated code (genc.cpp,spubindsymbols.h)
    void VerifyFreeRecordSlots( // Make sure records array is big enough
        int freeSlots);     // number of additional slots desired

    void updateMemRecPtrArr ();
    void LoadRecptrsArray(); // Load the m_pRecords array from record store
    void RefillRecptrs();    // reconstruct recptrs after swap-in
    void DeleteRecptrArray();

    void DeleteHash();       // Delete hash on record store

    void SortStore( // Rebuild m_pRecords in a sorted manner
        int numCols, int *sortCols, int *sortAscend);

    void SortStore(int sortAscend);// Sort ascending on all columns

    void AppendContent(TMemblockInfo *m);

    void AppendRawContent(
        char* pData,   // pointer to content bytes
        size_t bytes,  // number of bytes of content
        size_t recs ); // number of records corresponding to content

    // CTable refugees end

    // Disk-based hash joining special
    void DiskHashPartitionSettings(int outBufferPages);
    // Merge-sort specials
    int MergeSettings(int readAheadPages);

    static void VetRecBlock(    // check/fix "fast path" TMemblockInfo
        TMemblockInfo *m);

    void ResetStore();

    static void SnippetCloseDown(bool inDtor, dlink_t *storeList);  // special for plan end

    static void DestroyJobRecordStores(SpuJob *job) throw ();

    void SetFilePolicy(EFilePolicySetting filePolicy,
                       EStorePolicy newStorePolicy = eSPsame);

    void SetSpuImplicitClose(bool val);  // Hack to extend m_storeType concept

#if FOR_SPU
    static unsigned int totalPagesInUse(dlink_t *storeList);
#endif
    CPlan *i_storeLinkPlan () const {
        return m_storeLinkPlan;
    }


    int RsDsid() const {
        return m_RSdsid;
    }

    CShMemFile *RsFile() const;
    void RsFileName(char *buf, size_t len) const;

    void ReleaseRsPgacct() throw ();

    // page accounting notification methods
    void PhysPagesPrepareWrite(unsigned int pagenr, unsigned int cnt);
    void PhysPagesWritten(unsigned int pagenr, unsigned int cnt);
    void PhysPagesPrepareReadAhead(unsigned int pagenr, unsigned int cnt);
    void PhysPagesPrepareMap(unsigned int pagenr, unsigned int cnt);
    void PhysPagesMapped(unsigned int pagenr, unsigned int cnt);
    void PhysPagesUnmapped(unsigned int pagenr, unsigned int cnt,
                           unsigned int dismiss_cnt = (unsigned int) -1) throw ();
    void AnonPagesPrepareMap(unsigned int numPages);
    void AnonPagesMapped(const void *p, unsigned int numPages);
    void AnonPagesUnmapped(const void *p, unsigned int numPages) throw ();

    // notification for non-recordstore physical memory
    // allocation/deallocation. Typically you would want to call this after
    // you do an mmap().
    static void NotifyPhysMemUse(ssize_t deltaBytes);
    static void NotifyPadMemUse(ssize_t deltaBytes, int deltaPages);

    // notification for heap memory allocation/deallocation
    static void NotifyAnonHeapUse(ssize_t deltaBytes);

    void updatePhysFileSize(uint64 newSize);
    uint64 getAdjustedPhysFileSize();
    void updateTestFileSize(uint64 newSize);
    uint64 getFstatTestSize();

 private:

    void verifyTotalRecordSlots( // Make sure records array is big enough
        int totalSlots);         // total number of slots desired

    // This is a local wrapper for InitRecBlock, the latter being intended for external use.
    void InitializeRecBlock(TDataPageInfo& dpi);

    void attachRecBlock(TDataPageInfo& dpi);

    void freeRecBlock(TDataPageInfo& dpi); // Free a recordstore block

    TDataPageInfo *allocDataPageInfo();

    void setUpFirst(int block);

    TDataPageInfo *addBlock(     // Add a new block to a recordstore
        int need);               // record size requested

    void trimLowEnd(int lowBlock, int newLowBlock);
    bool growHighEnd(int hiBlock, int newHiBlock);

    void cleanUpHotPage();

    void cleanUpWindow();

    void cleanUpDPI();

    void freeDPI(TDataPageInfo *p);

#if FOR_SPU
    void updatePlanPageCount(int count);
    void updatePlanFileBackedPageCount(int count);
#endif
    void setPageHeader(TDataPageInfo& dpi);
    void verifyPageHeader(TDataPageInfo& dpi);

#if FOR_SPU
    void linkToPlanList(CPlan *plan);


    static dlink_t *link_not_hash(dlink_t *list, dlink_t *link);
#endif

    void openBackingFile(int mode, int dsId);

    void implicitSnippetClose(bool inDtor);
};

#if !FOR_SPU
inline void
CRecordStore::NotifyAnonHeapUse(ssize_t deltaBytes)
{
}
#endif // !FOR_SPU

enum {
    RSST_CREATE,
    RSST_UNLINK,
    RSST_EXTEND,
    RSST_MAP_ANON,
    RSST_MAP_FILE,
    RSST_PGIN,
    RSST_STEAL1,
    RSST_STEAL2,
    RSST_STEAL_INSUF,
    RSST_STEAL_SYNC,
    RSST_STEAL_INVAL,
    RSST_HOLE_PUNCH,
    RSST_INVAL_UNLINK,
    RSST_NONJOB,
    RSST_UDXPAD,
    RSST_FSTRIM,

    _RSST_LAST
};

struct StealableList {
    dlink_t list;
    pthread_rwlock_t lock;
};

struct PageCacheAcct {

    // Linux page cache page count estimates, units of 128k pages
    atomic_t globalResidentPages;
    atomic_t globalDirtyPages;
    atomic_t globalAnonPages;

    atomic_t nextFsTrimWriteCount;  // trim when value of rsstat[RSST_STEAL_SYNC] crosses this
    atomic_t rsstat[_RSST_LAST];

    unsigned int usableMemoryMB;           // total memory available to NPS processes

    // global page limits
    unsigned int globalResidentPageLimit1; // max nr resident 128k pages, excluding SQB reserve
    unsigned int globalResidentPageLimit2; // max nr resident 128k pages, including SQB reserve
    unsigned int globalResidentPageLimit3; // resident page limit oversubscription error threshold
    bool residentLimitOomFatal;            // true=oversubscrition dumps core

    unsigned int jobResidentPageLimitMax; // max resident pages per job, non-sqb jobs
    atomic_t     jobResidentPageLimitCur; // current per-job resident page limit
    unsigned int jobResidentPageLimitSQB; // max resident pages per job, sqb jobs
    unsigned int rsWriteFlushMinPages;    // min number of pages to flush
    unsigned int rsWriteFlushMaxPages;    // max number of pages to flush
    unsigned int jobLimitGrowPages;       // job limit growth increment
    unsigned int jobLimitShrinkPages;     // job limit shrink increment

    unsigned int rsSpuSqbJobQuotaPages;   // page quota for sqb jobs to avoid page stealing

    // Settings for fadvise & madvise calls
    unsigned int rsSerialReadAheadPages;
    unsigned int rsBulkReadAheadPages;
    unsigned int rsRandomReadAheadPages;

    bool nztmpfsHolePunchEnabled;

    unsigned int rsSpuWriteFlushMaxPages; // max number of pages to flush it is used by host to estimate aggr and sort mem

    unsigned int fsTrimThreshPages; // NZCFG->system->getInt("rsNzlocalTrimThreshPages")
    bool         fsTrimSupported;

#if FOR_SPU
    StealableList *rsStealableList; // per-datatslice array of stealable recordstore lists
#endif

    atomic64_t nonJobBytes;
};

LKTRKR_DECLARE_NAME(MTX_RSSTEALLIST);
LKTRKR_DECLARE_NAME(MTX_RSSTEALCOND);

// Environment options controlling transient (temp) table compression.
struct TempCompressOptions {
    uint32 ttcCrossOverPoint; // default page to cut over to compressed storage
    uint32 ttcTransientCompressOptions;
    uint32 ttcTransientCompressChecksumTrigger;
    uint32 ttcOsPageMask; // Conveniently set to 4095, or whatever, from
                          // querying the OS.

    // The following are derived from ttcTransientCompressOptions whenever
    // it is loaded.  This is for the convenience of the diagnostic code in
    // the compressor.

    uint32 ttcDebugLevel;
    uint32 ttcChecksumAction;
    uint32 ttcHundredsBits;
    uint32 ttcThousandsBits;
    uint32 ttcTenthousandsBits;
};

inline static void
inc_rs_stat(int stat)
{
    atomic_inc(&Nz->pageCacheAcct->rsstat[stat]);
}

inline static void
dec_rs_stat(int stat)
{
    atomic_dec(&Nz->pageCacheAcct->rsstat[stat]);
}

inline static void
add_rs_stat(int stat, int val)
{
    atomic_add(val, &Nz->pageCacheAcct->rsstat[stat]);
}

#endif          // RECORDSTORE_INCLUDED
