#ifndef SDLIST_INCLUDED
#define SDLIST_INCLUDED

#ifndef STDDEF_INCLUDED
# include <stddef.h>
# define STDDEF_INCLUDED
#endif

#ifdef __cplusplus
extern "C" {
#endif

// ============ doubly-linked list =============

#define dlcontainer_of(ptr, type, member) \
    ((type *)((char *)(ptr) - offsetof(type,member)))

// doubly-linked list link
typedef struct _dlink_t
{
    struct _dlink_t *next;
    struct _dlink_t *prev;
} dlink_t;

#define LIST_INITIALIZER(dl) { &(dl), &(dl) }
#define DECLARE_LIST(dl) dlink_t dl = LIST_INITIALIZER(dl)

#define LIST_IS_EMPTY(dl) ((dl)->next == (dl))

#define LIST_INIT(dl) do { \
    (dl)->next = (dl); \
    (dl)->prev = (dl); \
} while (0)

// insert item before pos
#define LIST_INSERT_BEFORE(item, pos) do { \
    dlink_t *const _item = (item); \
    dlink_t *const _pos = (pos); \
    _pos->prev->next = _item; \
    _item->prev = _pos->prev; \
    _item->next = _pos; \
    _pos->prev = _item; \
} while (0)

// insert item after pos
#define LIST_INSERT_AFTER(item, pos) do { \
    dlink_t *const _item = (item); \
    dlink_t *const _pos = (pos); \
    _pos->next->prev = _item; \
    _item->next = _pos->next; \
    _item->prev = _pos; \
    _pos->next = _item; \
} while (0)

#define LIST_PUSH_BACK_LINK(l,link) do { \
    /* insert to tail of list (before head) */ \
    (l)->prev->next = (link); \
    (link)->prev = (l)->prev; \
    (link)->next = (l); \
    (l)->prev = (link); \
} while (0)

#define LIST_PUSH_FRONT_LINK(l,link) do { \
    /* insert to front of list */ \
    (l)->next->prev = (link); \
    (link)->next = (l)->next; \
    (link)->prev = (l); \
    (l)->next = (link); \
} while (0)

#define LIST_PUSH_FRONT(l,req,link) LIST_PUSH_FRONT_LINK(l,&(req)->link)

#define LIST_PUSH_BACK(l,req,link) LIST_PUSH_BACK_LINK(l,&(req)->link)

// append dst list to src list and empty src list
#define LIST_APPEND(dst, src) do { \
    dlink_t *const _d = (dst); \
    dlink_t *const _s = (src); \
    _d->prev->next = _s->next; \
    _s->next->prev = _d->prev; \
    _s->prev->next = _d; \
    _d->prev = _s->prev; \
    LIST_INIT(_s); \
} while (0)

#define LIST_DELETE_LINK(link) do { \
    (link)->prev->next = (link)->next; \
    (link)->next->prev = (link)->prev; \
} while (0)

#define LIST_DELETE(req,link) LIST_DELETE_LINK(&(req)->link)

#define LIST_RESET_LINK(link) do { \
    (link)->next = 0; \
    (link)->prev = 0; \
} while (0)

#define LIST_FRONT(l) ((l)->next)
#define LIST_END(l) (l)

#define LIST_FOR_EACH(pos, l) \
    for (pos = LIST_FRONT(l); pos != LIST_END(l); pos = pos->next)

#define LIST_FOR_EACH_REVERSE(pos, l) \
    for (pos = (l)->prev; pos != (l); pos = pos->prev)

#define LIST_FOR_EACH_ENTRY(p, l, link) \
    for (p = dlcontainer_of(LIST_FRONT(l), typeof(*p), link); \
         &p->link != LIST_END(l); \
         p = dlcontainer_of(p->link.next, typeof(*p), link))

// ============ singly-linked list =============

typedef struct _slink_t
{
    struct _slink_t *next;
} slink_t;

typedef struct
{
    slink_t *head;
    slink_t *tail;
} fifo_t;

#define FIFO_INITIALIZER {0,0}

#define FIFO_INIT(f) do { \
    fifo_t *const _f = (f); \
    _f->head = _f->tail = 0; \
} while (0)

#define FIFO_PUSH_BACK_LINK(f, pos) do { \
    fifo_t *const _f = (f); \
    slink_t *const _e = (pos); \
    if (_f->tail == 0) \
        _f->head = _f->tail = _e; \
    else \
    { \
        _f->tail->next = _e; \
        _f->tail = _e; \
    } \
    _e->next = 0; \
} while (0)

#define FIFO_APPEND(_f1, _f2) do { \
    fifo_t *f1 = (_f1); \
    fifo_t *f2 = (_f2); \
    if (f1->tail == 0) { \
        *f1 = *f2; \
    } else if (f2->tail) { \
        f1->tail->next = f2->head; \
        f1->tail = f2->tail; \
    } \
    FIFO_INIT(f2); \
} while (0)

#define FIFO_PUSH_BACK(f, elt, link) FIFO_PUSH_BACK_LINK((f), (&(elt)->link))

#define FIFO_FRONT(f) ((f)->head)

#define FIFO_FOR_EACH(pos, f) \
    for (pos = FIFO_FRONT(f); pos; pos = pos->next)

#define FIFO_IS_EMPTY(f) ((f)->head == 0)

#define __FIFO_POP_FRONT(f, h) do { \
    fifo_t *const _f = (f); \
    h = _f->head; \
    _f->head = h->next; \
    if (_f->head == 0) \
        _f->tail = 0; \
} while (0)

#define _FIFO_POP_FRONT(f) do { \
    slink_t *h; \
    __FIFO_POP_FRONT((f), h); \
} while (0)

#define FIFO_POP_FRONT(f) do { \
    slink_t *h; \
    __FIFO_POP_FRONT((f), h); \
    h->next = 0; \
} while (0)

// slow O(N) count the number of items in fifo
inline static unsigned int
fifo_numitems(const fifo_t *f)
{
    unsigned int n = 0;
    const slink_t *pos;
    FIFO_FOR_EACH (pos, f)
        ++n;
    return n;
}

// slow O(N) walk fifo and validate tail pointer is consistent
inline static int
fifo_is_valid(const fifo_t *f)
{
    const slink_t *pos = FIFO_FRONT(f);
    if (pos) {
        const slink_t *prev;
        do {
            prev = pos;
            pos = pos->next;
        } while (pos);
        return prev == f->tail;
    }
    else
        return f->tail == 0;
}

#define _FIFO_DELETE(f,elt,found) do { \
    fifo_t *const _f = (f); \
    slink_t *_prev = _f->head; \
    if (_prev) \
    { \
        slink_t *const _e = (elt); \
        if (_prev == _e) \
        { \
            _f->head = _prev->next; \
            if (_f->head == 0) \
            { \
                _f->tail = 0; \
            } \
            *(found) = 1; \
        } \
        else \
        { \
            int _found = 0; \
            slink_t *_t = _prev->next; \
            while (_t) \
            { \
                if (_t == _e) \
                { \
                    _prev->next = _e->next; \
                    if (_e == _f->tail) \
                    { \
                        _f->tail = _prev; \
                    } \
                    _found = 1; \
                    break; \
                } \
                _prev = _t; \
                _t = _t->next; \
            } \
            *(found) = _found; \
        } \
    } \
    else \
    { \
        /* empty list */ \
        *(found) = 0; \
    } \
} while (0)

#define FIFO_DELETE(f,elt,found) do { \
    _FIFO_DELETE((f),(elt),(found)); \
    if (*(found)) \
        (elt)->next = 0; \
} while (0)

#ifdef __cplusplus
}
#endif

#endif
