#ifndef SPUJOB_INCLUDED
#define SPUJOB_INCLUDED

// (C) Copyright IBM Corp. 2009, 2014  All Rights Reserved.
#include "jobbase.h"
#include "rcAtomic.h"
#include "sdlist.h"
#include "nztimestamp.h"
#include "storagedefs.h"
#include "nzstd.h"
#include "dpm_common.h" // for TCHand
#include "groomspustate.h"
#include "nzpthread.h"

struct create_job_msg_t;
struct incr_resacct_msg_t;
struct final_resacct_msg_t;
extern "C" { typedef struct DiskJob TDiskJob; }
class CSPUInteraction;
struct ChildRsUsage;

typedef void (prio_notify_func_t)(void *param, int newprio, bool sqb_active);
typedef void (idle_func_t)(void *param);

/// INZA child resource counters
struct udx_child_resacct_t
{
    pid_t   cr_pid;             // Child PID
    uint32  cr_type;            // Child SQL function type
    uint32  cr_index;           // Index in plan m_udxObject array
    bool    cr_terminated;      // 0: child is running, 1: child terminated
    uint64  cr_cur_cpu_time;    // Current child CPU usage
    uint64  cr_last_cpu_time;   // Previous child CPU usage
                                // CONSIDER: Modify/use RusageTracker?
    unsigned int cr_mem_hwm;    // Child mem high water mark
};

struct process_cputime_accum
{
    pthread_mutex_t cp_lock;       // lock
    unsigned long cp_accum_ticks;  // accumulated cpu ticks
    unsigned long cp_last_ticks;   // cpu ticks at last update
    bool cp_running;               // true when process is running
};

struct job_udx_tracker
{
    /// set to true if any fenced udx ever registered
    bool ju_fenced_udx;

    /// Lock to protect child stats
    pthread_mutex_t ju_child_lock;

    /// Max number of child processes tracked
    int ju_child_max;

    /// Number of child processes actually registered
    int ju_child_count;

    /// Child resource counters (jr_child_max allocated)
    udx_child_resacct_t *ju_child_resacct_ary;
};

/// job dataslice resource counters
/// The "_reported" variables are used when reporting incremental data
struct jds_resacct_t
{
    /// Data disk (nzds) sectors read. Includes ZM.
    uint64 jr_data_disk_rd;
    uint64 jr_data_disk_rd_reported;

    /// Data disk (nzds) sectors written. Includes ZM.
    uint64 jr_data_disk_wr;
    uint64 jr_data_disk_wr_reported;

    // Data disk (nzds) sectors found in SPU table cache. Does not include ZM.
    uint64 jr_data_disk_rd_cache;

    // FPGA ticks used.
    uint64 jr_fpga_ticks;
    uint64 jr_fpga_ticks_reported;

    // Zone map sectors (nzds I/O w/o DISK_JOB_FLAG_FOR_FPGA)
    uint64 jr_zm_rd;
    uint64 jr_zm_wr;
    uint64 jr_zm_cache;

    /// Temp disk (nzlocal) sectors read.
    uint64 jr_temp_disk_rd;
    uint64 jr_temp_disk_rd_reported;

    /// Temp disk (nzlocal) sectors written.
    uint64 jr_temp_disk_wr;
    uint64 jr_temp_disk_wr_reported;

    /// Comm bytes sent.
    uint64 jr_comm;
    uint64 jr_comm_reported;

    /// Plan memory usage high water mark.
    unsigned int jr_mem_hwm;

    /// Snippet temp disk page allocation high water mark
    unsigned int jr_temp_disk_hwm;

    /// cpu time accumulator
    process_cputime_accum jr_cpu;
    unsigned long jr_cpu_reported;

    /// Timestamp of when the last resource account message
    /// went out.
    time64_t jr_resacct_next_send_time;

    /// The following tracks resource usage of child processes (INZA udx/AE)
    job_udx_tracker jr_udx_tracker;
};

// per-dataslice job data
struct job_dslice_t
{
    pid_t jds_jobtask;
    nzmutex_t jds_jobtask_lock;
    pthread_cond_t jds_jobtask_cond;
    unsigned int jds_jobtask_waiters;

    jds_resacct_t jds_resacct;

    // Linux page cache page count estimates
    atomic_t jds_residentPagesUcast; // pages from unicast records
    atomic_t jds_dirtyPagesUcast; // dirty pages

    // list of recordstores (excluding bcast stores)
    dlink_t jds_recstores;
    nzmutex_t jds_recstores_lock;
    atomic_t jds_recstores_cnt;

    // For mview index scans
    TCHand jds_prsrvdTch;

    // Groom state for reorg & reclaim
    GroomSpuState jds_groomState;
};

struct idle_cb
{
    slink_t link;
    idle_func_t *func;
    void *arg;
};

struct idle_cb_list
{
    unsigned int nr_jobs;
    fifo_t cblist;
    pthread_mutex_t lock;
};

class SpuJob : public JobBase
{
    int sj_prio_bias; // current priority bias
    bool sj_sqb_active; // Short Query Bias. When true, give preference to
                       // this job.

    // not implemented
    SpuJob(const SpuJob&);
    SpuJob& operator=(const SpuJob&);

protected:

    // not implemented (protected rather than private to suppress compiler
    // warning)
    SpuJob();
    ~SpuJob();

public:

    unsigned int sj_magic;
    atomic_t sj_refcnt;         // internal use only

    SpuJob *sj_next;            // job table use only

    bool sj_udx_stats_enabled;          // collect and use stats
    bool sj_udx_stats_detail_enabled;   // report detailed stats

    // Linux page cache page count estimates
    atomic_t sj_residentPagesBcast; // pages from broadcast records
    atomic_t sj_dirtyPagesBcast; // dirty pages

    dlink_t sj_recstores; // list of broadcast record stores
    nzmutex_t sj_recstores_lock;

    job_dslice_t sj_job_dslice[0]; // one element per dataslice

    static void init_prefork();
    static void start_jobmon_task();
    static void init_preonline();

    static void callWhenIdle(idle_func_t *func, void *arg);

    static SpuJob *createSpuJob(const create_job_msg_t *msg);
    static void releaseSpuJob(int jobid);

    static void cleanupSpuJobForTxAbort( int jobid );

    static unsigned int spuJobCnt(); // current number of spu jobs in existence

    void addReference();
    void removeReference();

    int prio_bias() const;
    void set_prio_bias(int bias);

    bool sqb_active() const;
    void set_sqb_active(bool val);

    void attach_dslice_process(int dsidx, pid_t pid);
    void detach_dslice_process(int dsidx);
    void join_dslice_process(int dsidx);

    int priority() const;

    void notify_priority() const;

    static void updateJobPrio(int jobid, int newprio, bool sqb_active);

    static void update_resacct_data_disk(const TDiskJob *dj);
    static void update_resacct_temp_disk(unsigned int sectors_wr, unsigned int sectors_rd);
    static void update_resacct_comm(size_t bytes);
    static void update_resacct_mem_hwm(unsigned int pageCount);
    static void update_resacct_temp_disk_hwm(unsigned int tempPageCount);

    void start_cpu_tracker();
    void stop_cpu_tracker();

    /////////////////////////////////
    // Indicates whether or not it's time to send a Snippet Resource
    // Accounting message. Based on the registry setting and when
    // fetch_incr_resacct or reset_ressacct was last called.
    // @returns: true if it's time to send the message. False otherwise.
    /////////////////////////////////
    bool isResacctDue() const;

    /////////////////////////////////
    // Retrieves incremental resource usage for this job along with some
    // stats of interest for tracking progress.
    // Resets the Resource Accounting timer.
    // @param dst: the buffer to write to.
    /////////////////////////////////
    void fetch_incr_resacct(incr_resacct_msg_t *dst);

    /////////////////////////////////
    // Retrieves incremental resource usage for this job along with final
    // stats of interest.
    // Resets the Resource Accounting timer.
    // @param dst: the buffer to write to.
    /////////////////////////////////
    void fetch_final_resacct(final_resacct_msg_t *dst, int num_children);

    /////////////////////////////////
    // Resets the Snippet Resource Accounting timer and the CPU
    // counter.
    /////////////////////////////////
    void reset_resacct();

    int register_prio_notify(prio_notify_func_t *func, void *param);
    void unregister_prio_notify(int cookie) throw ();

    // Used only by groom (reorg & reclaim)
    GroomSpuState *groomState(int dsIx);

    // Used by mview index scans to share scanlist between snippets
    TCHand preservedTCHand(int dsIx);
    void setPreservedTCHand(int dsIx, TCHand handle);

    // UDX child process tracking
    void register_udx_child(pid_t pid, int udxType, void *udxObject);
    void unregister_udx_child(pid_t pid, ChildRsUsage *cru);

    // Active children in resacct
    unsigned int udx_child_count();

    // Number of children to be reported in detailed statistics
    unsigned int udx_child_count_reported();

    // Return memory usage of all children of the indicated dataslice in
    // NPS pages. Safe to call from report process.
    unsigned int udx_children_mem_page_count(int dslice);

    // Return the private memory usage (as reported by OS as "data") in
    // units of 128K pages. Safe to call from report process.
    unsigned int getUdxPrivatePageCnt(int dslice);

    // Active children in resacct. Safe to call from report process.
    unsigned int getUdxChildCount(int dslice);
};

inline int
SpuJob::prio_bias() const
{
    return sj_prio_bias;
}

inline void
SpuJob::set_prio_bias(int bias)
{
    sj_prio_bias = bias;
}

inline bool
SpuJob::sqb_active() const
{
    return sj_sqb_active;
}

inline void
SpuJob::set_sqb_active(bool val)
{
    sj_sqb_active = val;
}

inline int
SpuJob::priority() const
{
    return base_prio() + prio_bias();
}

inline GroomSpuState *
SpuJob::groomState(int dsIx)
{
    return &sj_job_dslice[dsIx].jds_groomState;
}

inline TCHand
SpuJob::preservedTCHand(int dsIx)
{
    return sj_job_dslice[dsIx].jds_prsrvdTch;
}

inline void
SpuJob::setPreservedTCHand(int dsIx, TCHand handle)
{
    sj_job_dslice[dsIx].jds_prsrvdTch = handle;
}

#ifndef MYTHREAD_INCLUDED
# include "mythread.h"
#endif

struct auto_mythread_job
{
    SpuJob *m_job;

    auto_mythread_job() : m_job(0) { }

    auto_mythread_job(SpuJob *job, int dsidx)
        : m_job(job)
    {
        MyThread::tls_set_job(job, dsidx);
        job->addReference();
    }

    void set_job(SpuJob *job, int dsidx)
    {
        m_job = job;
        MyThread::tls_set_job(job, dsidx);
        job->addReference();
    }

    ~auto_mythread_job()
    {
        if (m_job)
        {
            MyThread::tls_set_job(0, -1);
            m_job->removeReference();
        }
    }
};

// helper class for updating a job-dslice's cpu resource utilization
class auto_job_cpu_tracker
{
    unsigned long m_tick0;
public:
    // start cpu resource counter
    auto_job_cpu_tracker();
    // end cpu resource counter and accumulate cpu time into current job
    // dslice
    ~auto_job_cpu_tracker();
};

#endif
