#ifndef STM_LABEL_INCLUDED
#define STM_LABEL_INCLUDED

#include "nzstd.h"
#include "locktracker.h"

LKTRKR_DECLARE_NAME(MTX_LABEL_DEVICE);

//----------------------------------------------------------------
// For SSV1/SSV2 upgrade/downgrade 
//
// TScatterExtents tracks scattered extents which contain backup
// copies of CORE partition extents.
//
enum {
    // sized to keep 2 structures in 'pad2'
    // limits acceptable disk size to approx 2.75 TB
    // exceeding this requires a new SSV1 upgrade scheme
    MAX_UPGRADE_SCATTER_EXTENTS = 11         
};

typedef enum {
    UPGRADE_SET_SSV1,
    UPGRADE_SET_SSV2,
    UPGRADE_SET_MAX
} UpgradeSet;

struct TScatterExtents { 
    int extentCount;                             // count of lba's which follow.
    int extentLba[MAX_UPGRADE_SCATTER_EXTENTS];  // variable sized list of lbas  (only used in systems with 32-bit lbas)
};

struct TUpgradeInfo {
    // upgrade storage tracking copies of CORE partition data for each version
    int state;
    typedef enum {
        NONE,
        DOWNGRADE_PREPARED,
        DOWNGRADE_COMMITTED,
        UPGRADE_PREPARED,
        UPGRADE_COMMITTED,
    } UpgradeState;

    TScatterExtents copyOfCore[UPGRADE_SET_MAX];
};

//----------------------------------------------------------------

enum {
    REVISION_ORIGINAL          = 1,
    REVISION_64BIT_PARTITIONS  = 2
};

//
// disk label structure  (Historically located at 2049)
// WARNING:  This structure and constant definitions are also defined in nzgilocal.h
// 
// Note: revision==2 changed the following:
//     * expanded core and partA sizes from 32 bits to 64 bits.
//     * obsoleted partB, swap, diaglo, diaghi regions.
//     * compatible with existing revision==1, as the obsoleted fields
//       have not been used (only printed with 'nzsqa disk -label')
//
// Q: why create hi and lo fields, instead of simply creating int64 fields?
// A: to keep the 32-bit label common between revision 1 and revision 2.
//    avoiding the need for heavy-handed upgrade/downgrades.
//
struct TDiskLabel
{
    int sig;                   //signature, int32 123456789
    int extentsize;            //global extent size for all partitions but core
    uint32 corestartlo;        //start sector of core partition (low order 32 bits)
    uint32 coresizelo;         //size in sectors of core partition (low order 32 bits)
    uint32 partAstartlo;       //start sector of partition A (low order 32 bits)
    uint32 partAsizelo;        //size in sectors of partition A (low order 32 bits)
    uint32 corestarthi;        //start sector of core partition (high order 32 bits)
    uint32 coresizehi;         //size in sectors of core partition (high order 32 bits)
    uint32 partAstarthi;       //start sector of partition A (high order 32 bits)
    uint32 partAsizehi;        //size in sectors of partition A (high order 32 bits)
    int revision;              //disk format version (*-see design note above)
    int obsolete[5];           //obsolete (corebackup and diag regions obsoleted with revision==2)
    int coreextentsize;        //extent size of core partition
    int maxread;               //max number of sectors in one read
    uint32 disksizelo;         //size of disk in sectors (low order 32 bits)
    int labelstart;            //start sector of label
    int labelsize;             //size in sectors of label
    uint32 disksizehi;         //size of disk in sectors (high order 32 bits)
    uint32 reserved1;          //unused
    int crc_upgrade_partitions;//used to auto-(re)start crc upgrade process
    int crc_upgrade_lba;       //used to auto-(re)start crc upgrade process
    int crc_upgrade_error_count; // used to know that errors have occurred prior to a restart
    uint64 crc_upgrade_start_ts; // when full auto-upgrade last started
    uint64 crc_upgrade_end_ts;   // when auto-upgrade last completed
    int metadataVersion;        // Version of Storage mgr meta data
    int metadataState;          // State of Meta data (0-stable, +n upgrading, -n downgrading)
    int flags;
    enum {
        CompressedTablesPresent  = 0x1,  // Compressed tables may be present
        LargeTablesPresent       = 0x2,  // Large tables may be present
        LargeCompressionRatio    = 0x4   // Tables with comp ratio > Mistral default may be present
    };
     // 1 => compiled tables are or were present on this SPU
    int pad1[30];              //not used
    int othersig;              //signature int32 987654321

    // sysmgr data (for host restore)
    int  idmagic;
    int  idpad;
    bool regenactive;
    bool regenactiveprecommit;
    bool unused[2];
    char dbuuid[40];
    int  numslices;
    int  topover;
    int  dsid;
    int  spuLoc;
    int  topoverprecommit;
    int  dsidprecommit;
    int  spuLocprecommit;
    int  devmapver;
    int  devmapverprecommit;
    char serialnum[32];
    char serialnumprecommit[32];

    union {
        TUpgradeInfo upgrade;
        int pad2[25];              // make sizeof(TDiskLabel) be SECTOR_SIZE
    };

    uint32  crc;               // label checksum (crc32)

#define ACCESSOR_READ_64BIT_FIELD(field)                \
    inline uint64 field() const {                       \
        return (revision == REVISION_64BIT_PARTITIONS)? \
            ((uint64)field##hi << 32) | field##lo:       \
            field##lo;                                  \
    }

    ACCESSOR_READ_64BIT_FIELD(partAstart);
    ACCESSOR_READ_64BIT_FIELD(partAsize);
    ACCESSOR_READ_64BIT_FIELD(corestart);
    ACCESSOR_READ_64BIT_FIELD(coresize);
    ACCESSOR_READ_64BIT_FIELD(disksize);

#define LABEL_WRITE_64BIT_FIELD(field, val)  \
    field##lo = val & 0xFFFFFFFF;            \
    field##hi = val >> 32;                   \

    inline void setPartAStart( uint64 val ) {
        LABEL_WRITE_64BIT_FIELD(partAstart, val);
    };
    inline void setPartASize( uint64 val ) {
        LABEL_WRITE_64BIT_FIELD(partAsize, val);
    };
    inline void setCoreStart( uint64 val ) {
        LABEL_WRITE_64BIT_FIELD(corestart, val);
    };
    inline void setCoreSize( uint64 val ) {
        LABEL_WRITE_64BIT_FIELD(coresize, val);
    };
    inline void setDiskSize( uint64 val ) {
        LABEL_WRITE_64BIT_FIELD(disksize, val);
    };
};

enum {
    METADATA_VER_PREHISTORY    = 0,
    METADATA_VER_NO_DICTIONARY = 1,
    METADATA_VER_LOG_BASED     = 2
};

enum {
    METADATA_STATE_DOWNGRADE_STAGE_2 = -2,
    METADATA_STATE_DOWNGRADE_STAGE_1 = -1,
    METADATA_STATE_STABLE            = 0,
    METADATA_STATE_UPGRADE_STAGE_1   = 1,
    METADATA_STATE_UPGRADE_STAGE_2   = 2
};

#endif // STM_LABEL_INCLUDED
