#ifndef STORAGEDEFS_INCLUDED
#define STORAGEDEFS_INCLUDED

#include <limits.h>

#include "nzsector.h"

// 13-Aug-2003: common constants/definitions


// Disk sizes for Mustang systems ONLY (150 and 400.  320 for early dev systems).
//     150GB - First generation production mustang
//     320GB - Early dev systems.  Not for production.
//     400GB - Large mustang production systems (same partitioning
//             as 320GBs with leftover allocated to SWAP).
// Disk sizes for Finch and Sparrow
//     120GB - systems ship with 120GB disk drives ONLY.
// Disk sizes for Mercury
//      40GB - systems ship with 40GB disk drives ONLY.
// NOTE: GB in terms of vendor specified disk sizes is
//       in units of 1000000000 (billion decimal)
#define DISK_SIZE_150GB   (292968750)  // sectors
#define DISK_SIZE_250GB   (488281250)  // sectors
#define DISK_SIZE_320GB   (625000000)  // sectors
#define DISK_SIZE_400GB   (781250000)  // sectors

// extracted from storagemgr.h:

enum {
    SECTOR_SIZE         = 512,  // sector size in bytes
    SECTORS_PER_KB      = 1024 / SECTOR_SIZE,
    SECTORS_PER_MB      = SECTORS_PER_KB * 1024,
    SECTORS_PER_GB      = SECTORS_PER_MB * 1024,
    SECTORS_PER_PAGE    = 256,  // source of the Netezza 128KB data page
    BYTES_PER_PAGE      = SECTOR_SIZE * SECTORS_PER_PAGE,
    PAGES_PER_EXTENT    = 24,
    SECTORS_PER_EXTENT  = SECTORS_PER_PAGE * PAGES_PER_EXTENT, // 6144
    BYTES_PER_EXTENT    = BYTES_PER_PAGE   * PAGES_PER_EXTENT, // 3145728
    MEGABYTES_PER_EXTENT= SECTORS_PER_EXTENT / SECTORS_PER_MB,

    // archaic aliases
    SECTORS_PER_BLOCK   = SECTORS_PER_PAGE,
    BLOCKS_PER_EXTENT   = PAGES_PER_EXTENT,

    NZ_TABLE_DEFAULT_BLOCK_SIZE = BYTES_PER_PAGE // extracted from table.h
};

enum {
    MAX_SPU_DATASLICES = 128
};

#define DEVICE0 0

// 22-Sep-2003: try to use EPartition instead of int for parameter types !
// Do not change the values for the partitions without searching the
// sources for the usage of these
enum EPartition {
    CORE = 0,           //partition number of core
    PART_A = 1,
    PART_B = 2,
    SWAP = 3,
    // XS2 has only PART_A and CORE partitions under storagemgr management
    NUM_STORAGEMGR_PARTITIONS = 2,
    UNKNOWN_PART = 4,   // must be >= NUM_STORAGEMGR_PARTITIONS
    CORE_BACKUP = 4,
    DEFAULT_PART = 1    // default is PART_A
};

// This is the ID for the one and only Netezza-style table that used
// to live in the SWAP partition. Catalog upgrade/downgrade still has
// need to know
// #define NZ_SWAP_TABLE_ID 2      // Any number that PG won't use is ok

// convert EPartition to EDiskPartition someday...
enum EDiskPartition {
    DISK_PART_CORE = 0,        // partition number of core
    DISK_PART_A = 1,
    DISK_PART_B = 2,
    DISK_PART_SWAP = 3,
    DISK_PART_CORE_BACKUP = 4,
    // XS2 has only PART_A and CORE partitions under storagemgr management
    NUM_STORAGEMGR_DISK_PARTITIONS = 2
};

enum EDataPartition {
    DATA_PART_A = 0,
    DATA_PART_B = 1,
    NUM_DATA_PARTITIONS = 1
};

// Magic words

enum {
    NZ_PAGE_HEADER_MAGIC    = 0xFEEDFACE, // first word of page header
    NZ_PAGE_HEADER_EMULATED = 0x10012002  // first word of a buffer processed by FPGA emulator
};

// Layout IDs in block headers
//  NOTE: must be odd parity (high bit reserved for parity)
//        0x40000000 bit indicates 'compressed'
// Think of layout ID as being structured as:
// +-+---------------+--------+--------+
// |P|000000000000000|CCCCCCCC|TTTTTTTT|
// +-+---------------+-----------------+
//  where P = odd-parity bit
//        C = compression type
//    T...T = CRC type

enum ELayoutID {
    LAYOUT_BIT_PARITY                     = 0x80000000,

    COMPRESSED_TYPE_UNCOMPRESSED          = 0x00000000,
    COMPRESSED_TYPE_NUMBERS_ONLY          = 0x00000100,
    COMPRESSED_TYPE_ALL_COLUMNS           = 0x00000200,
    COMPRESSED_TYPE_ALL_COLUMNS_STABLE    = 0x00008200,
    COMPRESSED_TYPE_COLUMNAR              = 0x00000300,
    COMPRESSED_TYPE_COLUMNAR_STABLE       = 0x00008300,
    COMPRESSED_TYPE_MASK                  = 0x0000FF00,

    COMPRESSED_TYPE_STABLE_MASK           = 0x00008000,
    COMPRESSED_TYPE_NONSTABLE_MASK        = 0x00007F00,
    
    LAYOUT_FACT_DTID_CLEAN                = 0x00010000, // set if we *know* that all dtids are zero in the page

    CRC_TYPE_NONE                         = 0x00000001,
    CRC_TYPE_FULL_BLOCK                   = 0x00000002,
    CRC_TYPE_MASK                         = 0x00000003,

    LAYOUT_FULL_BLOCK_CRC                 = CRC_TYPE_FULL_BLOCK,

    // let us not care about STABLE bit and PARITY and DTID_CLEAN bits when checking
    // for CTA2 pages during downgrade
    LAYOUT_CHECKTABLE_MASK                = 0x7FFE7FFF
};
#define COMPRESSED_TYPE(layout) ( (layout) & COMPRESSED_TYPE_MASK )
#define IS_COMPRESSED(layout)   ( COMPRESSED_TYPE(layout) != 0 )

#define VALID_LAYOUT(layout) ( ( layout == LAYOUT_FULL_BLOCK_CRC )        \
                            || ( layout == LAYOUT_FULL_BLOCK_CRC_COMP_NUMB_ONLY ) \
                            || ( layout == LAYOUT_FULL_BLOCK_CRC_COMP_ALL_COLUMNS ) \
                            || ( layout == LAYOUT_FULL_BLOCK_CRC_COMP_ALL_COLUMNS_STABLE ) \
                            )

// (in)convenience macros
#define DISK_TO_DATA_PART(diskPart) ( (diskPart) == DISK_PART_A ? DATA_PART_A : \
                                    ( (diskPart) == DISK_PART_B ? DATA_PART_B : -1 ) )
#define DATA_TO_DISK_PART(dataPart) ( (dataPart) == DATA_PART_A ? PART_A : \
                                      ( (dataPart) == DATA_PART_B ? PART_B : \
                                        (EPartition) -1 ) )

#define IS_DATA_PARTITION(diskPart) ((diskPart) == PART_A || (diskPart) == PART_B)

inline static const char *
diskPartitionName (EDiskPartition diskPart)
{
    switch (diskPart) {
    case DISK_PART_CORE:
        return "CORE";
    case DISK_PART_A:
        return "PART_A";
    case DISK_PART_B:
        return "PART_B";
    case DISK_PART_SWAP:
        return "SWAP";
    case DISK_PART_CORE_BACKUP:
        return "CORE_BACKUP";
    default:
        return "Unknown diskPartition";
    }
}

/*
 * FIXME LBA64: This struct is defined in 2 places!  The other is
 * spucbs.h as sector_range_t.
 */
struct TSectorRange
{
    nzsector_t   sr_sector __attribute__ ((aligned(8)));
    unsigned int sr_count;
};

struct TScanList
{
    unsigned int sl_ranges;   // number of sector ranges
    TSectorRange sl_range[1]; // array of sector ranges
};

#define SL_SIZE(num_ranges) (sizeof(TScanList) + (num_ranges-1) * sizeof(TSectorRange))


// !!! in dev-osr2 the EWriteOrderType definition has moved to fpgaDiskApi.h

#define NZ_ZONE_MAP_TABLE_ID  (5665)  
#define ZONE_MAP_MAX_COLUMNS  (200)
// Number of specials supported in zone map
#define ZONE_MAP_SPECIALS_COUNT (2)


namespace hc {
struct DiskPartitionInfo;
};

enum ECrcError {
    eReadError,     // error on upgrade read
    eWriteError,    // error on upgrade write
    eInvalidLayout, // read successful but block has invalid layout ID
    eCheckpoint,    // failure to write checkpoint
    eMemory         // problem freeing IoStream at end (?!)
};

enum {
    eAddStartSector = 1,
    eSubStartSector = 2
};

// Disk serial numbers of failed disks in an MD Raid array
#define MAX_SN_LEN (128)  // Also in hcMsgs.h *must be kept in sync*!!!

struct DevMap {
    int          lunGid;
    int          partId;     // 0-15, 100, 101, 110-113
    int          partSize;   // in MiB
    char         primarySn[MAX_SN_LEN];
    int          primaryPhysPartId;
    char         regenSn[MAX_SN_LEN];
    int          regenPhysPartId;
};

#endif
