/* (C) Copyright IBM Corp. 2001, 2011  All Rights Reserved. */

#ifndef STORAGEMGR_INCLUDED
#define STORAGEMGR_INCLUDED

/* Copyright (c) 2001-2007 Netezza Corporation.  All rights reserved. */

#include "dpm_common.h"
#include "error.h"
#include "memNewDelete.h"
#include "nzlimits.h"
#include "nzmutex.h"
#include "nzsector.h"
#include "spucbs.h"
#include "stmlabel.h"
#include "storagedefs.h"
#include "TXID.h"

#include <stdio.h>
#include <string>

namespace rc { class TxtTbl; };

class CPlan;
class CRegenTarget;
class CStorageMgr;
class CTable;
class ClearHeadersJob;
class CoreTable;
class DSConfig;
class TInsertNode;
class TIoStream;

struct TExtentList;
struct TKeyValue;
struct TMemMgmt;
struct TScanList;
struct blockheader_t;
struct scan_state_t;

#define MAX_DISK_SN_LEN 32

// utility function to decompress a compressed disk block
extern bool _decompress_block(
    CTable* tableP,
    blockheader_t* decompressBufP,
    int bufferSize,
    blockheader_t* compressedBufP );

extern TScanList *makeEmptyScanList();

/*
 CStorageMgr Object

The storage manager provides disk space management functions.


Disk layout:
  The disk is logically divided into four partitions - core, swap, part A, and
part B.  Parts A and B are for table storage. Swap is for workspace and
some sort of virtual store for large tables. Core contains
directories, catalogs, the dbos dictionary and backup dictionary.  Space is
accounted for and managed using directories.  Tables are managed using
catalogs.
  A directory contains extent descriptors.  An extent is a fixed-size area of
disk.  Each partition is divided into a number of extents (set when the
disk is "formatted").  A table is allocated as many extents as it needs.
A directory is loaded into memory at spu startup and periodically rewritten
to the disk.
  A catalog contains data about tables.  There is one catalog entry for each
table.  Catalogs are loaded at startup and periodically rewritten.
  Tables are stored in either part A or Part B partitions.  Extents are
preallocated if requested at table creation time.  When end of file is reached
during writes, new extents are allocated as needed.  A table is created with
a fixed blocksize at creation time.  Records are packed into these blocks as
densely as possible.  Each block has a blockheader and each record has a
recordheader.
*/


// data slice location
struct dslice_loc_t
{
    int16 dl_dsid;
    uint8 dl_device;  // zero-based device index
};

struct dslice_loc_list_t
{
    unsigned int nr;
    dslice_loc_t ary[0];
};

struct sm_instance_list_t
{
    unsigned int nr;      // number of instances (one per device)
    CStorageMgr *ary[0];  // array of instances, indexed by device index
};

#define SM_NR_DEVICES() (CStorageMgr::s_deviceStorageMgr->nr)
#define SM_FOR_EACH_DEVICE(sm) for (CStorageMgr **__pp = &CStorageMgr::s_deviceStorageMgr->ary[0], *sm = *__pp; \
                                    __pp < &CStorageMgr::s_deviceStorageMgr->ary[CStorageMgr::s_deviceStorageMgr->nr]; \
                                    sm = *++__pp)

// mount states
enum {
    SMMNT_NOT_MOUNTED,
    SMMNT_MOUNTED_FOR_REGEN,  // mounted but dirs and cats not valid
    SMMNT_MOUNTED
};

//In memory PartACat, CoreCat, SwapCat and PartBCat point to structures
//containing a catalog header followed by catalog entries.

struct SMConfigVars
{
    int fpgaBools;
    int emulatorThreads;
    int appendDataCap;
    int regenGenericCtrl;
    int tableExtentLimit;
    int fpgaRecSizeIncrPct;
    bool fpgaDump;
    bool insertToNull;
    bool regenSkipBadHeaderCheck;
    bool enableLargeTables;
    bool avoidSilentMode;
    bool reclaimRecoveryEnable;
    cbs_cfg_t cbs_cfg;
};
extern const SMConfigVars *SMConfig;

static inline bool SMCompressionEnabled()
{
    return true;
}

////////////////////
/// CStorageMgr - allocate from global shared memory
/// true means IO aligned.  That is required because we do a read IO directly
/// into this structure; the label is the first data member, so aligning the
/// structure should make it aligned properly for the IO.
////////////////////
class CStorageMgr : public mem::newDeleteClear<mt_storage,
                                               mem::REGION_GLOBAL,
                                               mem::MEM_ALIGN,
                                               mem::MEM_SCOPE_PERMANENT>
{
    friend class CPlan;
    friend class CIndexMgr;
    friend class CRecordMgr;
    friend class CMirrorMgr;
    friend class CDictionary;
    friend class CRegenTarget;
    friend class CRegen;
    friend bool VtStoragePartitionInfo(CTable *table, int32 hwId, CPlan *plan);
    friend bool VtStorageTableInfo(CTable *table, int32 hwId, CPlan *plan);
    friend void downgradeCommit(CStorageMgr &stm);


protected:

    TDiskLabel m_Label;        //disk label as read from sector 0
    bool       m_goodLabel;

    int m_mounted;             // SMMNT_*

    char m_primarySn[MAX_DISK_SN_LEN];
    char m_mirrorSn[MAX_DISK_SN_LEN];

    CMutex *m_labelMutex;   // internal concurrent protection

    // cached values from registry
    SMConfigVars m_smconfig;

    ClearHeadersJob *m_clearheaderjob;

public:

    // writeLabel option bits
    enum {
        WL_DEFAULT = 0, // verify m_goodLabel
        WL_FORCE = 1,   // write regardless of m_goodLabel
        WL_CHECK = 2    // write if m_goodLabel, ignore otherwise
    };

    // array of StorageMgr instances, indexed by device number
    static sm_instance_list_t *s_deviceStorageMgr;

    // array of data slice locations
    static dslice_loc_list_t *s_dsliceLoc;

    unsigned int m_device;   // storage device index

    int m_dsidA;             // dataslice id of PART_A
    int m_topover_precommitted;   // for use by writeTopology() - see comment there
    int m_devmapver_precommitted; // for use by writeTopology() - see comment there

public:

    CStorageMgr(unsigned int device);

    static void InitDsliceLocMap();

    void InitStorageMgrConfig();

    bool StorageMgrMount(
        bool reformat,
        bool forRegen);

    void StorageMgrUnmount();

    void dpmDefine();
    void dpmDefineOneDataSlice( unsigned systemDsId
                              , unsigned device
                              , nzsector_t dataLba
                              , nzsector_t logLba
                              , DSConfig const& dsconfig
                              , unsigned whichPartition);
    void dpmActivate(bool reformat);
    void dpmActivateDsId(bool reformat, DsId dsid);
    void dpmDeactivate();

    bool TruncTable(           //truncates a table
        TXid xId,              //transaction id
        int tblId,             //table number to delete
        EPartition which);      //partition

    int GetTableSizeInBlocks(  //returns table size in blocks
        int tablenum,          //table number
        EPartition which);     //partition

    nzsector_t i_disksize() const { return m_Label.disksize(); }
    int i_extentsize() const { return m_Label.extentsize; }

    const char *primarySn() const { return m_primarySn; }
    const char *mirrorSn() const { return m_mirrorSn; }

    unsigned int blocksPerExtent() const { return m_Label.extentsize / SECTORS_PER_BLOCK; }
    unsigned int mbPerExtent() const { return m_Label.extentsize / SECTORS_PER_MB; }

    bool i_mounted() const { return m_mounted != SMMNT_NOT_MOUNTED; }

    int DsidForPart(EPartition part) const;

    TDiskLabel *i_readlabel() { return &m_Label; }
    TDiskLabel const *i_readlabel() const { return &m_Label; }
    void writeLabel(unsigned int options);

    void updateLabelFlags(bool hasCompressedTables, bool hasLargeTables, bool hasHighlyCompressedTables);

    void ResetCta();

    unsigned int GetTableCount(int dsId) const;

    void Dump (
        char *title,
        int detail,
        FILE *fp,
        int in);

    static unsigned int numStorageDevices();
    static unsigned int numPartitionsPerStorageDevice() { return 2; } // data, meta-data
    static unsigned int numDataPartsPerStorageDevice() { return 1; } // RIP Mustang

    uint64 totalDiskSize() const;
    const hc::DiskPartitionInfo& partitionInfo(int which) const;

    int labelHwIdVersion() const;
    const char *labelDbUuid() const;
    void writeTopology(bool commit);
    void writeTopology(bool commit,
                       int topover,
                       int devmapver,
                       int numdataslices,
                       int dsidpri,
                       int spuLoc,
                       const char *pridisksn=NULL,
                       bool activeregen=false);

    void        dumpLabel();
    void        dumpDisk();

    void        dumpPageHdrs( int tblId         // table object
                            , int startBlock
                            , int count );

    void        dumpPageData( int absBlockOffset
                            , int tableId
                            , int byteCount
                            , uint64 pageId
                            , int byteOffset = 0
                            , nzsector_t fromSectorNum = 0
                            , bool ignoreHeader = false
                            , bool decompress = false);

    void        dumpRecordHdrs( int TableNum
                              , int relBlockOffset
                              , uint64 pageId
                              , int recCount
                              , int recOffset);

    nzsector_t i_lastSector() const
    {
        return m_Label.disksize() - 1;
    }

    unsigned int device () const
    {
        return m_device;
    }

    CMutex* mutex()
    {
        return m_labelMutex;
    }

    void setPrimarySn(const std::string& primarySn);
    void setMirrorSn(const std::string& mirrorSn);
    bool emptyPage (DevNo dev, nzsector_t lba, table_id_t tblId);

    // virtual table materializers need these:
    void getPartitionStartAndSizes(EPartition which,
                                   nzsector_t *partitionStartP,
                                   nzsector_t *partitionSizeP,
                                   unsigned int *partitionExtentSizeP);

    void setupStorageMgr();

    // Associate a running clear header job with this storagemgr.
    // (This simplifies finding the job based on DSID)
    void setclearjob(ClearHeadersJob *job) { m_clearheaderjob = job; }
    ClearHeadersJob *getclearjob() const { return m_clearheaderjob; }

 protected:

    // utilities
    void    validatePartitionSizes();
    void    initLabel(int partitionSectorCount);
    bool    readLabel();

    // disk dump utilities
    void    setupBlockHdrTbl(rc::TxtTbl* tbl, int rowNum);
    void    addBlockHdrTblRow(rc::TxtTbl* tbl, void* block, int rowNum, int index);

    void formatStorageDevice(); // partition disk, writes skelton file system

    nzsector_t getStart(EPartition which);
};

inline int
CStorageMgr::DsidForPart(EPartition part) const
{
//    NZ_VERIFY(part == PART_A);
    return part == PART_A ? m_dsidA : 0;
}

inline static CStorageMgr *
SM_FOR_DEVICE(unsigned int device)
{
    NZ_CHECK(device < CStorageMgr::s_deviceStorageMgr->nr,
             NZ_ERROR_ILLVAL, "Storage Manager doesn't exist for device %u\n",
             device);

    return CStorageMgr::s_deviceStorageMgr->ary[device];
}

#endif
