#ifndef STORAGEREFS_H
#define STORAGEREFS_H

//======================================================================
///\file: storagerefs.h
//
// (C) Copyright IBM Corp. 2008, 2011  All Rights Reserved.
//======================================================================

#include "error.h"

#include "pmScanListDefs.h"     // for PageNo
typedef pm::PageNo PageNo;      ///< PState24 index (+1 bias; non-zero)
// end of things from dpm_pvt needed ^^^^^

class StorageId;
class ExtentId;
class PageId;

#include "storagedefs.h"

#define NULL_SID  StorageId::null()
#define NULL_EID   ExtentId::null()
#define NULL_PID     PageId::null()

// Well-formed but guaranteed greater than any generatable SID, EID & PID
#define NEVER_SID StorageId::never()
#define NEVER_EID  ExtentId::never()
#define NEVER_PID    PageId::never()

#define RAW_NEVER_SID  (((1ULL << StorageId::EXTENT_SERIAL_BITS) - 1) << StorageId::NZPAGENO_BITS)
#define RAW_NEVER_EID  (((1ULL << StorageId::EXTENT_SERIAL_BITS) - 1) << StorageId::NZPAGENO_BITS)
#define RAW_NEVER_PID ((((1ULL << StorageId::EXTENT_SERIAL_BITS) - 1) << StorageId::NZPAGENO_BITS) | 1)


class StorageId
    /// A storage denotation base class composed of a persistent symbolic,
    /// monotonic extent ID and a page tag.  Existence of the extent ID (or
    /// serial number) can solve various dangling reference issues but
    /// requires translation through some form of map (i.e. an ExtIdHT
    /// hash table) to locate the actual extent with a data partition.
{
public:
    enum
    // Disk geometry definitions
    { NZPAGENO_BITS     = 5     // i.e. log2(PAGES_PER_EXTENT)
    , NZPAGENO_MASK     = ((size_t)1 << NZPAGENO_BITS) - 1
    , NZPAGENO_FIRST    = 1
    , NZPAGENO_LAST     = PAGES_PER_EXTENT
    , ALL_NZPAGES_MASK  = (1 << NZPAGENO_LAST) - 1

    // Subfields
    , NZPAGE_BITS       = NZPAGENO_BITS
    , EXTENT_SERIAL_BITS= 39
    , EXTENTNO_FIRST    = 1

    // 2^20 extents => 'extent size in MBs' equals 'data slice size in TBs'
    , ID_BASE           = (1 << 20)
    };

    static StorageId null();
    static StorageId never();
    uint64 const& rep() const;
    int64 const& sqlRep() const;
    uint64 extentSerial() const;
    unsigned pageBits() const;
    bool isPage() const;
    bool isFirstPage() const;
    bool isLastPage() const;
    unsigned mask() const;
    ExtentId extentId() const;
    char const* pageTag() const;

    // Default constructor for use in TScanNode c'tor
    StorageId() { m_rep = 0LL; }

    // Support for textual display of 2-part StorageId values
    #define SIDFMT "[SID:%010llx|%s]"
    #define SIDARG(SID) (SID).extentSerial(), (SID).pageTag()

protected:

    // Protected allows access from ExtentId and PageId classes.
    explicit StorageId(uint64 const&);

private:

    uint64      m_rep;

    static char const* pageTags[];

friend StorageId toStorageId(uint64 const&);

} __attribute__(( packed, aligned(2) ));


//----------------------------------------------------------------------

class ExtentId : public  StorageId
    /// A specialization of StorageId denoting only entire extents.
    /// It consists of a symbolic extent ID and a null PageNo.
{
public:

    ExtentId();
    static ExtentId null();
    static ExtentId never();
    PageId makePageId(PageNo pageNo) const;

    // Support for textual display of 2-part ExtentId values
    #define EIDFMT "[EID:%010llx|%s]"
    #define EIDARG(EID) (EID).extentSerial(), (EID).pageTag()

private:

    explicit ExtentId(uint64 const&);

friend ExtentId toExtentId(uint64 const&);
friend ExtentId makeExtentId(uint64);
friend ExtentId StorageId::extentId() const;

} __attribute__(( packed, aligned(2) ));


//----------------------------------------------------------------------

class PageId : public StorageId
    /// A specialization of StorageId denoting only single pages.
    /// It consists of a symbolic extent ID and a non-null PageNo.
{
public:

    PageId();
    static PageId null();
    static PageId never();
    size_t extentDiffers(PageId const& pid) const;

    // Support for textual display of 2-part PageId values
    #define PIDFMT "[PID:%010" PRIx64 "|%s]"
    #define PIDARG(PID) (PID).extentSerial(), (PID).pageTag()

private:

    explicit PageId(uint64 const&);

friend PageId toPageId(uint64 const&);
friend PageId toPageIdUnchecked(uint64 const&);
friend PageId toPageIdNonFatal(uint64 const&);
friend PageId makePageId(uint64, unsigned);
friend PageId ExtentId::makePageId(PageNo) const;

} __attribute__(( packed, aligned(2) ));


//----------------------------------------------------------------------
// Related non-members functions

StorageId toStorageId(uint64 const&);
ExtentId toExtentId(uint64 const&);
ExtentId makeExtentId(uint64 extentSerial);
PageId toPageId(uint64 const&);
PageId toPageIdUnchecked(uint64 const&);
PageId toPageIdNonFatal(uint64 const&);
PageId makePageId(uint64 extentSerial, unsigned pageNo);


#define DEFINE_COMPARISON(CLASS, OP) \
inline bool operator OP (CLASS const& left, CLASS const& right) \
{ \
    return left.rep() OP right.rep(); \
} \

#define DEFINE_COMPARISONS(CLASS) \
    DEFINE_COMPARISON(CLASS, ==) \
    DEFINE_COMPARISON(CLASS, !=) \

#if 0

    DEFINE_COMPARISON(CLASS, <=) \
    DEFINE_COMPARISON(CLASS,  <) \
    DEFINE_COMPARISON(CLASS, >=) \
    DEFINE_COMPARISON(CLASS,  >) \

#endif

DEFINE_COMPARISONS(StorageId)
DEFINE_COMPARISONS(ExtentId)
DEFINE_COMPARISONS(PageId)

#undef DEFINE_COMPARISON
#undef DEFINE_COMPARISONS



////////////////////////////////////////////////////////////////////////
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

//= Inline implementations: StorageId ==================================

inline
StorageId StorageId::null
    /// Return a NULL StorageId value.
    ()
{
    return StorageId(0);
}

inline
StorageId StorageId::never
    /// Return an alternative NULL StorageId with a non-zero representation.
    ()
{
    return StorageId(RAW_NEVER_SID);
}

inline
uint64 const& StorageId::rep
    /// Return the underlying representation.  Needed because C++
    /// forbids passing a non-POD type to a ... prototype (e.g. printf).
    () const
{
    return *(uint64*)&m_rep;
}

inline
int64 const& StorageId::sqlRep
    /// Return an SQL-friendly rendering of the underlying representation
    () const
{
    return *(int64*)&m_rep;
}

inline
uint64 StorageId::extentSerial
    /// Return the right-justified extent serial number.
    () const
{
    return m_rep >> NZPAGE_BITS;
}

inline
unsigned StorageId::pageBits
    /// Return the page tag value.
    () const
{
    return ((size_t)m_rep) & NZPAGENO_MASK;
}

inline
bool StorageId::isPage
    /// Return true IFF the value denotes a single page.
    // review-dle: why 0 m_rep is a page?
    () const
{
    return m_rep == 0 || pageBits();
}

inline
bool StorageId::isFirstPage
    /// Return true IFF value references the first page within an extent.
    () const
{
    return pageBits() == NZPAGENO_FIRST;
}

inline
bool StorageId::isLastPage
    /// Return true IFF value references the last page within an extent.
    () const
{
    return pageBits() == NZPAGENO_LAST;
}

inline
unsigned StorageId::mask
    /// Return either a mask for a single page or for an entire extent.
    () const
{
    size_t pb = pageBits();
    return pb ? 1 << (pb-1) : ALL_NZPAGES_MASK;
}

inline
ExtentId StorageId::extentId
    /// Return a StorageId's underlying extent as an ExtentId.
    () const
{
    return ExtentId(m_rep & ~((uint64)NZPAGENO_MASK));
}

inline
char const * StorageId::pageTag
    () const
{
    return pageTags[ pageBits() ];
}

inline
StorageId::StorageId
    /// A protected unchecked constructor.
    ( uint64 const& rep
    )
    : m_rep(rep)
{}


//= Inline implementations: ExtentId ===================================

inline
ExtentId::ExtentId
    /// Default constructor.
    ()
    : StorageId(0)
{}

inline
ExtentId ExtentId::null
    /// Return a NULL ExtentId value.
    ()
{
    return ExtentId(0);
}

inline
ExtentId ExtentId::never
    /// Return an alternative NULL ExtentId with a non-zero representation.
    ()
{
    return ExtentId(RAW_NEVER_EID);
}

inline
PageId ExtentId::makePageId
    /// Specialize an ExtentId to reference one of its pages.
    ( PageNo pageNo
    ) const
{
    NZ_VERIFY(NZPAGENO_FIRST <= pageNo && pageNo <= NZPAGENO_LAST); // bad pageNo => fatal
    return PageId(rep() | pageNo);
}

inline
ExtentId::ExtentId
    /// A private unchecked constructor.
    ( uint64 const& rep
    )
    : StorageId(rep)
{}


//= Inline implementations: PageId =====================================

inline
PageId::PageId
    /// Default constructor.
    ()
    : StorageId(0)
{}

inline
PageId PageId::null
    /// Return a NULL PageId value.
    ()
{
    return PageId(0);
}

inline
PageId PageId::never
    /// Return an alternative NULL PageId with a non-zero representation.
    ()
{
    return PageId(RAW_NEVER_PID);
}

inline
size_t PageId::extentDiffers
    /// Return non-zero IFF presented PageId falls within a distinct extent.
    ( PageId const& pid
    ) const
{
    // Implementation optimized for a 32-bit architecture
    return ((((size_t)(rep() >>  0) ^ (size_t)(pid.rep() >>  0)) & ~NZPAGENO_MASK)
           | ((size_t)(rep() >> 32) ^ (size_t)(pid.rep() >> 32))
           );
}

inline
PageId::PageId
    /// A private unchecked constructor.
    ( uint64 const& rep
    )
    : StorageId(rep)
{}


//= Inline implementations: non-member functions =======================

inline
StorageId toStorageId
    /// Checked conversion of a raw uint64 into a StorageId.
    ( uint64 const& rep
    )
{
    NZ_VERIFY((rep & StorageId::NZPAGENO_MASK) <= StorageId::NZPAGENO_LAST);
    return StorageId(rep);
}

inline
ExtentId toExtentId
    /// Checked conversion of a raw uint64 into an ExtentId.
    ( uint64 const& rep
    )
{
    NZ_VERIFY((rep & StorageId::NZPAGENO_MASK) == 0); // bad pageNo => fatal
    return ExtentId(rep);
}

inline
ExtentId makeExtentId
    /// Construct an ExtentId from an extent serial.
    ( uint64 extentSerial
    )
{
    return ExtentId(extentSerial << StorageId::NZPAGENO_BITS);
}

inline
PageId toPageId
    /// Checked conversion of a raw uint64 into a PageId.
    ( uint64 const& rep
    )
{
    unsigned pageNo = rep & StorageId::NZPAGENO_MASK;
    NZ_VERIFY((StorageId::NZPAGENO_FIRST <= pageNo && pageNo <= StorageId::NZPAGENO_LAST) || rep==0); // bad pageNo => fatal
    return PageId(rep);
}

inline
PageId toPageIdUnchecked
    /// Unchecked conversion of a raw uint64 into a PageId.
    ( uint64 const& rep
    )
{
    return PageId(rep); // Caller beware
}

inline
PageId toPageIdNonFatal
    /// Checked conversion of a raw uint64 into a PageId.
    ( uint64 const& rep
    )
{
    unsigned pageNo = rep & StorageId::NZPAGENO_MASK;
    if (StorageId::NZPAGENO_FIRST <= pageNo && pageNo <= StorageId::NZPAGENO_LAST)
        return PageId(rep);
    else
        return NULL_PID;
}

inline
PageId makePageId
    /// Construct a PageId from an extent serial and page bits (checked).
    ( uint64 extentSerial
    , unsigned pageNo
    )
{
    NZ_VERIFY(StorageId::NZPAGENO_FIRST <= pageNo && pageNo <= StorageId::NZPAGENO_LAST); // bad pageNo => fatal
    return PageId((extentSerial << StorageId::NZPAGENO_BITS) | pageNo);
}

#endif // STORAGEREFS_H
