/** \file table.h
 *
 * (C) Copyright IBM Corp. 2001, 2016  All Rights Reserved.
 */

#ifndef TABLE_INCLUDED
#define TABLE_INCLUDED

#include "error.h"
#if (AIX ||SUN4S || HP800 || HPITANIUM)
#else
#include "mirrortopo.h"
#endif
#include "nzlog.h"     // for Dump( ) defaults
#include "memInterface.h"
#include "memStrdup.h"
#include "nzmutex.h"
#include "rcAtomic.h"
#include "sdlist.h"
#include "tableshape.h"

#if FOR_SPU
#include "TxIdList.h"
#include "autolock.h"

# ifndef TNS
// DPM's captureWaterMark needs a TCHand
#  include "dpm_common.h"
#  include "storagerefs.h"
# endif // TNS
#endif // FOR_SPU

#include "BitVector.h"
#include "rcAtomicSmp.h"

struct TIoStream;           // so iomgr.h can go limited distribution
struct TMemblockInfo;
class CSpuSwapMgr;
class TProjectNode;
class CRecordStore;
struct TDistRowsRqst;
struct TCreateTempTableMsg;
class CTable;
class CPlan;
struct dbos_create_table_t;
struct dbos_create_table_entry_t;
struct Var;
struct ExprContext;
struct Expr;
struct Node;
struct TPgVarGroup;
struct spu_msg_header_t;
struct TranInfo;
struct bcast_rs_pgacct_t;
struct TAlterColumnInfo;
#if !FOR_SPU
class TruncateStats;
#endif

#if 0 // FOR_SPU   // jpb- handy for finding leaking ctable objects
#define PROJTBLLOG(addr)  Log(0,0,"CTable projected 0x%08x id=%d m=%d t=%d X=%" PRId64 ": %s#%d\n", \
                          addr,(addr)->i_id(),(addr)->i_Ttype(),(addr)->i_isTransient(), \
                          (addr)->i_xId().val(), __FILE__,__LINE__)
#define CLONETBLLOG(addr)  Log(0,0,"CTable Clone 0x%08x id=%d m=%d t=%d X=%" PRId64 ": %s#%d\n", \
                          addr,(addr)->i_id(),(addr)->i_Ttype(),(addr)->i_isTransient(), \
                          (addr)->i_xId().val(), __FILE__,__LINE__)
#define TBLLOG(addr)  Log(0,0,"CTable 0x%08x id=%d m=%d t=%d X=%" PRId64 ": %s#%d\n", \
                          addr,(addr)->i_id(),(addr)->i_Ttype(),(addr)->i_isTransient(), \
                          (addr)->i_xId().val(), __FILE__,__LINE__)
#define TBLDESTRUCTORLOG() Log(0,0,"0x%08x id=%d m=%d t=%d m= X=%" PRId64 ": destructor\n", \
                               this, m_id, m_Ttype, i_isTransient(),m_xId.val())
#else
#define PROJTBLLOG(addr)
#define CLONETBLLOG(addr)
#define TBLLOG(addr)
#define TBLDESTRUCTORLOG()
#endif // FOR_SPU

#ifndef NZGLOBALS_INCLUDED
extern int NzMyProcessId;
#endif // NZGLOBALS_INCLUDED
#if !FOR_SPU
#define TABLE_EXCLUSIVE_NOT_SELF(table) \
    ({ int const _pid = (table)->exclusivePid(); _pid != 0 && _pid != NzMyProcessId; })
#endif

/** \class CTable

The CTable class is currently 256 bytes on disk.  There are a number
of available "reserved" members.  These members are guaranteed to be zero
filled in 'previous' versions.
When we run out of those, there are a number of members that are
'oversized' and can be reduced for 32 to 16 bit values.  There are also
member groups that are rarely used and can be allocated separately for
those few tables that use them.

There are also many members that have meaning only when the CTable is
memory resident (e.g., all pointers) and it would be simple to have
a "on-disk" table image and today's CTable with two translation
methods.

Tables are stored in memory using a table descriptor object, CTable and
a record storage object CRecordStore. NEXT S WILL CHANGE -- DK:
In the host, they are always
allocated in shared memory and thus visible to Postgres, and the DBOS
processes.

If a CTable references a clustered index, the m_id member is the
clustered index id and the m_parentId member is the base table id.

THIS NEXT SECTION IS OBSOLETE AND NEEDS A REWRITE --DK
A CTable object references its records via an array of pointers,
m_records.  Thus, records in a table are sorted by sorting the pointer
array, not the records themselves.  Likewise, a 'restrict' can be
performed by 'orphaning' records in the record store.

The records themselves are stored packed in memory, with sufficient
padding to allow field access according to the machine's alignment
rules.

Records in memory may be preceeded by a length word.  If the record is
composed solely of fixed size fields, there is no size word before the
record because it is simply the m_fixedFieldsSize member of the table
object.  If the record has varying size fields, the size word is still
optional.  In this latter case, the record is scanned to determine its
length.  See the CTable.RecordSize().

If records allow null for any fields, there is a null-mask whose bitcount
includes all fields, not just the nullable fields.  If present, this
null-mask preceeds the first field value.

Also in the records header may be some optional special fields.  The special
fields do not use the m_fields descriptor array.  Instead, their presence is
defined by the m_enableSpecials member and the m_stickySpecials member. These
are bitmasks that specify whether to include the (1) physical row address,
(2) logical row number, (3) transaction id it was created with, and/or (4) the
transaction id it was created with. In general, the row address and
row number special fields are enabled for locking on delete and update queries.
They are also used for two-step joining where the dbos code joins on a
minimal set of colums and then fetches the rest of the columns by row number
or row address. The transaction id information is generally just needed
for transaction rollback.

The table also uses a m_stickySpecials member to define whether these special
fields are included in the Clone() and CreateProjected() methods.

*/

#define TABLE_DISK_SIZE 256           //256 bytes
#define PAGE_SIZE_BYTES_SHIFT 17      //128KB
#define PAGE_SIZE_BYTES (1 << PAGE_SIZE_BYTES_SHIFT)
#define DEFAULT_RECSTORE_SIZE PAGE_SIZE_BYTES
#define ROUND_BYTES_UP_TO_PAGES(BYTES) (((BYTES)+PAGE_SIZE_BYTES-1)>>PAGE_SIZE_BYTES_SHIFT)

#define DEFAULT_VTSMALL_RECSTORE_SIZE 0x1000 // 4K

#ifndef MAX_DIST_KEY_FIELDS
#define MAX_DIST_KEY_FIELDS 4                  //!FIX-FR-dw - move this higher?
#endif

// Distribution algorithms.  User written algorithms will use dynamically
// provided high numbers.
//
// DIST_ALGOR numbers are persisted in the compressed external table header.
// If you change an existing number (extensions are fine), we need to write
// backward compatibility code.
// See Scott or TDownloadTableNode::readExternalTableHeader.
#define DIST_ALGOR_CRC_HASH 0    // self-hash ints, crc other types
#define DIST_ALGOR_RND_ROBIN 1   // round robin
#define DIST_ALGOR_ALWAYS_CRC 2     // (todo) crc all types
#define DIST_ALGOR_ALWAYS_KNUTH 3   // (todo) knuth all types
#define DIST_ALGOR_KNUTH_HASH 4     // (todo) self-hash ints, knuth other types

#define SPRING_SPACE (ALIGN_VARLENS(512))

#define NZ_TABLE_DEFAULT_BLOCK_SIZE 131072

#define NZ_TABLE_MAX_BLOCK_SIZE 131072  // Apr2007(dk)ndelink.cpp DEPRECATED

// See also NZ_TABLE_MAX_COLUMN_SIZE defined in nzlimits.h
// see also NZ_MAX_RECORD_SIZE defined in nzlimits.h
// see also MAX_TABLE_COLUMNS  defined in nzlimits.h


// "Identity" tables used for place holders, EXISTS(), etc. a/k/a "degenerate"
#define TABLE_1_COL_EMPTY     17 // used by pg for non-scanning plans
#define TABLE_1_COL_1_ROW     18 // used by pg for non-scanning plans
#define TABLE_TRUE_COL_1_ROW  19 // used OpenParamTable for exists subplans
#define TABLE_FALSE_COL_1_ROW 20 // used OpenParamTable for exists subplans
#define TABLE_NULL_COL_1_ROW  21 // used OpenParamTable for subplan empty result


const int MAX_SUBXIDS=1000;

// Association of transient table CRecordStores and physical partitions on SPU.
enum EDutyTypes {
    DD_UNKNOWN = 0,    // by "construction"
    DD_SLICES  = 1,
    DD_BROADCAST = 2   // one record store used by all dataslices' plans on blade
};


CEXTERN Datum NzFetchVar(   // entry point for postgres to get colval
        Var *variable,  // pg var
        ExprContext *econtext,  // pg expression context
        bool *isNull);          // returned

struct TJoinTableSlot {    // Used to control pg expression evaluator's record fetch
    CTable *table;
    record_t *record;
};

struct CTransTableCtx {

    // WARNING!!!!!!!!!
    // This enum must track NzTransientTblNames in recordstore.cpp.
    enum ETransTableType {
        eTransTableNone = 0,
        eDHJNodeStreamBitVector,
        eDHJNodeStreamTable,
        eDHJNodeTempTable,
        eVtScanNodeTable,
        eScanBlockCountNodeTable,
        eSaveTempNodeTable,
        eDownloadTableNodeTable,
        eDataRcvDownloadTable,
        eDiskBasedSortSwapTable,
        eInsertNodeTable,
        eDeleteNodeTable,
        eAggrNodeTable,
        eAggrNodeHashDistinctTable,
        eAggrNodeAggrInfoFinalTable,
        eMtlzNodeTable,
        eTableSizeNodeTable,
        eSubPlanParamTable,
        ePlanExecDoneDummyTable,
        eGroomReorgPrepTable,
        eWindowAggrFrameTable,
        eSortRecordsTable,
        eTableLast
    };

    ETransTableType m_transTblType;
    int16   m_index;

    explicit CTransTableCtx(ETransTableType type) :
        m_transTblType(type), m_index(-1) { }
    explicit CTransTableCtx(ETransTableType type, int idx) :
        m_transTblType(type), m_index(idx) { }
};

extern CTransTableCtx s_transTblCtxUnknown;
#define TTCTX(e) CTransTableCtx((CTransTableCtx::e))
#define TTCTXI(e,i) CTransTableCtx((CTransTableCtx::e), (i))
#define TTCTX_UNKNOWN s_transTblCtxUnknown

LKTRKR_DECLARE_NAME(MTX_VT);

#if FOR_SPU
# ifndef TNS
// Obsolete under TNS

struct TableInserter {
    TXid insertXid;
    bool insertsBuffered; // istream has buffered inserts
    PageId firstPageId;   // lowest inclusive insert pageId
    TRecLoc waterMark;
};

typedef std::vector<TableInserter*, NzShmAlloc<char, mt_inserters>::stl_alloc> InsertersVector;
typedef InsertersVector::iterator                                              InsertersIter;
typedef InsertersVector::const_iterator                                        InsertersIterConst;

struct TDSInsertInfo {
    TIoStream *iStream;
    InsertersVector inserters;
};

struct TDSGroomInfo {
    TIoStream *groomStream;
    TXid groomXid;             // only one concurrent groom transaction is allowed
    PageId groomEotPageId;     // highest fence (exclusive) pageId of groom scan list
};

struct TableDslice {
    pthread_mutex_t tds_lock;
    TDSInsertInfo *tds_ins;
    TDSGroomInfo *tds_groom;
    TableDslice();
    ~TableDslice();

    void AllocInsertInfo();
    void TryFreeInsertInfo() throw ();
    void FreeInsertInfo() throw ();
    void AllocGroomInfo();
    void TryFreeGroomInfo() throw ();
    void FreeGroomInfo() throw ();

    unsigned int getNumActiveInserters(const CTable *table) const;

    // This should be given an empty vector of type TxIdsVector
    void getActiveInserters(const CTable *table, TxIdList& activeInserters) const;

    void clearInsertsBuffered(const CTable *table, const TxIdList& inserters);

    unsigned int addInserter(
        const CTable *table,
        const TXid& xId,
        bool insertsBuffered);

    int FindInserterDex(const TXid& xId) const;

    void addGroomer(const TXid& xId);

};


# endif // TNS
#endif // FOR_SPU

struct TableCtx {

    enum ETableType {
        eCTonDisk=1,      // On disk, SQL-referenceable tables (incl session temps)
        eCTvirtual,       // "ring" or "dynamic", all the same now
//        eCTstoredVirtual, // virtual table with globally shared storage
//        eCTdynamicVirtual,// virtual table with dynamic materializer
        eCTtransient,     // Intra-job transient record sets (multi-snippet)
        eCTslice,         // Single slice table (unique to single slice)
    };

    ETableType ctx_type;

    union {
        // for eCTslice
        CPlan *ctx_plan;

        // for eCTonDisk, eCTvirtual, eCTtransient
        struct {
            int ctx_tblid;
            const TXid *ctx_xid;
        };
    };

    static TableCtx
    ForPlan(CPlan *plan) { return TableCtx(plan); }

    static TableCtx
    DiskTable(int tblid) { return TableCtx(eCTonDisk, tblid, 0); }

    static TableCtx
    VirtualTable(int tblid) { return TableCtx(eCTvirtual, tblid, 0); }

    static TableCtx
    TransTable(int tblid, const TXid *xid) { return TableCtx(eCTtransient, tblid, xid); }

private:
    TableCtx(CPlan *plan) : ctx_type(eCTslice), ctx_plan(plan) { }
    TableCtx(ETableType type, int tblid, const TXid *xid) : ctx_type(type), ctx_tblid(tblid), ctx_xid(xid) { }
};

class CTable {

public:

    enum ESizeSource {
        eLSSnone=0,
        eLSSpostgres=1,
        eLSSsnipSched=2,
        eLSStblNode=3,
        eLSSsniprescache=4
    };

    static CTable *CreateTable( // primary public interface for new CTable objects
        const TableCtx& ctx,
        int numFields,              // number of columns, see DefineField() method
        bool permanent = false);    // mark allocated memory as permanent (degenerate tables)

    static CTable *CreateTableFromShape( // primary public interface for new CTable objects
        const TableShape *shape,
        const TableCtx& ctx);

    static void DestroyTable(CTable *table) throw (); // Antonym of CreateTable, vice destructor

    void addTableReference() {
        atomic_inc(&m_tableRefCnt);
    }

    void removeTableReference() {
        int const val = atomic_dec_return(&m_tableRefCnt);
        NZ_VERIFY(val >= 0);
        if (val == 0)
            delete this;
    }

#if !FOR_SPU
    void beginExclusive(int pid) {
        int prior = atomic_cmpxchg(&m_exclusivePid, 0, pid);
        NZ_VERIFY(prior == 0);
    }

    void endExclusive() {
        atomic_set(&m_exclusivePid, 0);
    }

    int exclusivePid() const {
        return atomic_read(&m_exclusivePid);
    }
#endif

    void lockVtRead() {
        nzpthread_rwlock_rdlock_trk(&m_vtRwLock, MTX_VT, m_id);
    }

    void lockVtWrite() {
        nzpthread_rwlock_wrlock_trk(&m_vtRwLock, MTX_VT, m_id);
    }

    void unlockVtRw() {
        nzpthread_rwlock_unlock_trk(&m_vtRwLock, MTX_VT, m_id);
    }

    void lockTable() {
        NzSetMutex(&m_tableLock);
    }

    void unlockTable() {
        NzClearMutex(&m_tableLock);
    }

 private:

    struct TableInitArgs {
        TableCtx::ETableType ttype;
        int tblid;
        CPlan *plan;
        const TXid *xid;
        int numFields;
        TableShape *shapeAlloc;
        TranInfo *tranInfoAlloc;
#if FOR_SPU
# ifndef TNS
        TableDslice *tableDsliceAlloc;
# endif
#endif
    };

    // Constructor is private, use CTable::CreateTable()
    CTable(const TableInitArgs& initargs);
    // Destructor is private, use CTable::DestroyTable()
    ~CTable();

    CTable *CloneTable( // Create a new table with same field types but new specialFields
        TUseSpecialFields specials, // special fields enabling bitmask
        TUseSpecialFields stickySpecials, // sticky special fields enabling bitmask
        bool bForceNulls, // true to force nullable fields in the clone
        const TableCtx& ctx) const;

 public:
    CTable *CloneTable( // Create a new table with same field types
        const TableCtx& ctx,
        CloneType how) const;

    CTable *CloneTable(
        CPlan* plan,
        TUseSpecialFields specials) const
    {
        return CloneTable(specials, specials, false, TableCtx::ForPlan(plan));
    }

    void AlterField( // Set truesize of a altered field
        int fdex,   // ordinal 0 - (m_numFields-1)
        int type,   // NZ_FIELD_TYPE_XXXX...
        unsigned int size,// for varchar, nvarchar, nchar
        const TXid& xId); // in case of rollback

    void AlterFieldZmap(  // Start/stop zonemapping of the specified field
        int    fdex,      // 0-base field index
        bool   zmapped,   // true: zonemap; false: do not zonemap
        const TXid& xId); // for rollback

#if !FOR_SPU
// used by HostTXMgr (and CTable) methods
    void restorePrevSizesForAlter (const TXid& xId);
    void recordPrevSizeForAlter (const TXid& xId, int coldex, int prevsize, uint8 prevflags);
    void freePrevSizesForAlter (const TXid& xId);

    TXid truncateTXid() { return m_truncTXid; }
    void setTruncateTXid(const TXid& txid) { m_truncTXid = txid; }
    TruncateStats* truncateStats() { return m_truncateStats; }
    void setTruncateStats(TruncateStats* p) { m_truncateStats = p; }

    void setNotForSpu()     { m_notForSpu = true; }
    bool notForSpu()  const { return m_notForSpu; }
#endif // FOR_SPU

    void ComputeCompilability(); // Set FieldIsCompilable in m_fields members where appropriate

    void RemoveRecordStore(CPlan *pPlan); // Detach record store (store deletes itself)
    void CloseBridgeStore(CPlan *pPlan); // Detach record store
                                         // (special for Bridge Table)
    void ResetRecordStore(); // Make record store empty, without deleting it

    int64 TotalSize() const; // Used Size of the record store

    uint32 TotalAllocatedSize() const; // Allocated Size of the record store

    const TableShape *Shape() const { return m_shape; }
    const TableShape *ValidateShape() {
        m_shape->ValidateShape();
        return m_shape;
    }

    TableShape *writableShape() const { return m_shape; }

    void FetchSpecials( // Get special fields (rowAddr,rowNum,txCreated,txDeleted)
        record_t const* recP,   // record to fetch from
        TRecLoc* ordLocP,       // returned recOrdLoc    if m_enableSpecials.recOrd
        TRowId* rowNumP,        // returned rowId        if m_enableSpecials.rowId
        TXid* txCreatedP,       // returned txCreated    if m_enableSpecials.txCreated
        TXid* txDeletedP,       // returned txDeleted    if m_enableSpecials.txDeleted
        TRecLoc* dtidLocP) const // returned dtidOffLoc if m_enableSpecials.dtidOff
    {
        m_shape->FetchSpecials(recP, ordLocP, rowNumP, txCreatedP, txDeletedP, dtidLocP);
    }

    void SetSpecials( // Set special fields (rowAddr,rowNum,txCreated,txDeleted)
        record_t *recP,         // record to set in
        TRecLoc ordLoc,         // recOrd    if m_enableSpecials.recOrd
        TRowId rowNum,          // rowId     if m_enableSpecials.rowId
        TXid const& txCreated,  // txCreated if m_enableSpecials.txCreated
        TXid const& txDeleted,  // txDeleted if m_enableSpecials.txDeleted
        TRecLoc dtidLoc) const  // dtidOff   if m_enableSpecials.dtidOff
    {
        m_shape->SetSpecials(recP, ordLoc, rowNum, txCreated, txDeleted, dtidLoc);
    }

    void FetchDiskSpecials( // Get special fields (rowNum,txCreated,txDeleted)
        const record_t *recP,   // record to fetch from
        TRowId *rowNumP,        // returned rowId if m_diskSpecials.rowId
        TXid *txCreatedP,       // returned txCreated if m_diskSpecials.txCreated
        TXid *txDeletedP) const // returned txDeleted if m_diskSpecials.txDeleted
    {
        m_shape->FetchDiskSpecials(recP, rowNumP, txCreatedP, txDeletedP);
    }

    void SetDiskSpecials( // Set special fields (rowNum,txCreated,txDeleted)
        record_t *recP,         // record to set in
        TRowId rowNum,          // rowId if m_diskSpecials.rowId
        const TXid& txCreated,  // txCreated if m_diskSpecials.txCreated
        const TXid& txDeleted) const // txDeleted if m_diskSpecials.txDeleted
    {
        m_shape->SetDiskSpecials(recP, rowNum, txCreated, txDeleted);
    }

    void AddRecordToStore(    // Create record from passed array of field vals
        const void **valsP, CPlan *plan ); // array of ptrs to field values and/or null ptrs

    int PackFromFieldArray( // Create record from passed array of field vals
        unsigned char sbcspc,
        void **valsP,   // array of ptrs to field values (and/or null ptrs)
        const CTransTableCtx& transTblCtx = TTCTX_UNKNOWN,
        bool charsAreVarlenas=true,
        bool numericsAreRecordReady=false,
        char *outBuf=NULL  // if null, add to table, else just pack
        );

    int CreateBlankRecordPlusSpringSpace( CPlan* plan); // Create rec w/ empty v-fields except for spring

    const char *fullyQualifiedName() const; // return the fully qualified name of table

    void Dump(const char *title=NULL,
              int detail=NzDumpDetail,
              FILE *fp=NzDumpFile,
              int indent=NzDumpIndent,
              int dumpTblFields=false) const;
    void dump () const;   // For DBG_DUMP_OBJ ()
    void Dumper() const;  // for debugger access
    void dumpFields(const int indent) const;
    void RecordHexDumper(  // for debugger access
        const void *recP) const;

    void RecordHexDump( // Printf record in hex
            const char *recordP,int dex,FILE *fp, int in) const;

    void RecordDump( // Printf record (called by Dump())
        const char *recordP, int dex,
        int detail, FILE *fp, int indent) const;

    void FieldDump( // Printf field (called by RecordDump/FieldsDump)
        const field_t *fldP, const void *fieldP, FILE *fp) const;

    void noteRecordSize( // Copy record size to size word, if present
        record_t *recordP, int size);

    void i_noteRecordSize( // Copy record size to size word, if present
        record_t *recordP, int size) {
        if ( NZ_LIKELY(m_shape->m_sizeWordSize == 2) )
            *((uint16*)recordP) = size;
        else if ( m_shape->m_sizeWordSize == 4 )
            *((uint32*)recordP) = size;
    }

    //!NOTE-dk too minimal a cliche?
    CRecordStore *GrabSingleRecordSpace(CPlan *plan, const CTransTableCtx& transTblCtx); // makes fixed size record buffer

    static TableShape *CreateProjected( // Create a table shape that is a field subset
        const TableShape *src,
        TUseSpecialFields specials, // special fields enabling bitmask
        TUseSpecialFields stickySpecials, // sticky special fields enabling bitmask
        int numFields,  // number of fields in new table (or 0 for 'all')
        int *fieldNumsP,    // array of fieldnums to clone (or <0 for computed)
        Node **exprList,   // expr if fieldNumsP[ix] == -1
        int *physIDs = 0, // if non-zero, physical IDs in logical-ID order
        bool suppressMaxCheck=false,
        bool unloadAlterTblCompressedNoNull=false);

    static TableShape *CreateProjected( // From plan's project node, create field subset
        const TableShape *src,
        TProjectNode *proj,
        int *physIDs = 0 ); // if non-zero, physical IDs in logical-ID order

    static TableShape *CreateJoined ( // Create a table that is a field subset of two tables
        const TableShape *src,
        const TableShape *tbl2P,     // the 'right' table
        TProjectNode *proj, // project node that follows join nodes
        int joinType);      // inner/outer-left.right.full

    static TableShape *CreateAggrTbl(  //create a table that matches proj after aggr
        const TableShape *src,
        TProjectNode *proj );

    // DEPRECATED "High level" record store interface
    CRecordStore *RecordStore(void) const; // Return "my query" record store (compatibility with the past)

    CRecordStore *NeedRecordStore(const CTransTableCtx& transTblCtx = TTCTX_UNKNOWN); // Create record store if there isn't one, return it
    CRecordStore *NeedRecordStore(CPlan *plan, const CTransTableCtx& transTblCtx); // Create record store (for named plan) if there isn't one, return it
    CRecordStore *NeedRecordStore(CPlan *plan, int size, int capacity, const CTransTableCtx& transTblCtx);

    CRecordStore *MakeOutputOnlyRecordStore(CPlan *plan, // Create sequentially
                                            bool forceImplicitClose,
                                            const CTransTableCtx& transTblCtx);
    CRecordStore *MakeClientOutputRecordStore(const CTransTableCtx& transTblCtx);  // Create sequentially
    CRecordStore *MakeReturnNodeRecordStore(CPlan *plan);     // Special RS for ReturnNode
    CRecordStore *MakeWorkTempRecordStore(CPlan *plan, const CTransTableCtx& transTblCtx);   // snippet-local RS
    // written record store (assert if another one beat us to it!)
    CRecordStore *MakeDownloadRecordStore(int dsId);  // SPU datarcv special

    CRecordStore *ReopenStore(CPlan *plan, bool forceImplicitClose=true, bool readOnly = true);   // Later snippets' uses of store
    CRecordStore *ReopenStoreForCaching(int dsid );

    TPgVarGroup **PgVarGroupSlot(int dsid); // for postgres fetchVar interface

    bcast_rs_pgacct_t *BcastRsPgacct() const; // broadcast recordstore page accounting info

    // Parallel execution record store interfaces.
#define HOST_DSID (0)
    CRecordStore *PlanRecordStore(const CPlan *planP) const;
    CRecordStore *DsidToRecordStore(int dsId) const;
    CRecordStore *DownloadRecordStore(int dsId) const; // SPU datarcv special

    void DestroyDownloadRecordStore(int dsId) throw();
    void CloseDownloadRecordStore(int dsId); // SPU datarcv special

    int CopyRecordSubset( // Copy some of the fields of a record to new rec
        ExprContext *exontextP, // pg expr context for pg eval mem allocs
        Timestamp stmtTime,     // for CURRENT_TIME, etc, evaluation
        CTable *srcP,   // src table
        record_t *srcRecP, // src record
        TProjectNode *pNode);    // from the plan

    char *packOneField ( // Load a single field into a record at specified offset
        char *buf,
        const void *fp,
        int type,
        int declaredSize,
        int dataSize, unsigned char sbcspc);

    int CopyRecord( // Make a copy of the record to this table
        const TableShape *srcP, const record_t *srcRecP);

    int CopyRecordSansSpecials( // Make a copy of the record to this table
        const TableShape *srcP, const record_t *srcRecP);

    // CopyInVarField is called by generated code (genc.cpp,spubindsymbols.h)
    void CopyInVarField( // Update varying field, adjusting spring to keep same recsize
         char *rec, int fdex, char *bodyP, int bodyLen);

    void AdjustVarField( // Update varying field, adjusting spring to keep same recsize
         char *rec, int fdex, int bodyLen);

    void BecomeIndex(int parent, int nIxFlds) { m_parentId = parent; m_numIndexFields = nIxFlds; }

    uint32 EqualKeys( // Given 2 records, return TRUE iff equal keys (nulls compare equal)
        record_t *srcRec1, record_t *srcRec2, int numCols, const int *colNums, unsigned char space) const;

    int64 NeedIntVal( // Given a record/column, return int value there
        record_t *rec,
        int field);

    void fillCreateTableEntry(dbos_create_table_entry_t *e) const;

    dbos_create_table_t *PackCreateTableMsg( // Allocate and load create table msg for this
        size_t *sizeP,   // returned size of created message
        const TXid& xId,
        memType_t mt) const;

    void SendCreateTable();   // called only by client (maybe dbos someday)
    void SendTruncateTable(); // called only by client (maybe dbos someday)
    void SendDropTable();     // called only by client (maybe dbos someday)
    static void SendDropTable (
        table_id_t tblId,
        const TXid& prevXid = TXid_NULL);

    void StartRecordInsertTransaction( // Let dbos lock table...
        const TXid& xId);

    TDistRowsRqst *CreateDistRowsRqst( // Produce message packet for later sending
        const TXid& xId,   // transaction id of the inserting transaction
        memType_t mt);

    TCreateTempTableMsg *CreateTempTableMsg( // Produce message packet for later sending
        CPlan *plan,       // alloc pool, TXId
        int *fullMsgSize,  // OUT: length of this created message
        EDutyTypes partDest);

    static int ExprDbosType( // return the Dbos data type of an expression's return value
        const Node* const expr, const int32 col = -1);

    static uint16 typmodToLen(int type, int32 typmod);
    static uint16 exprLen(const Expr* const expr, const int32 col = -1);

    //bvjoin table
    CRecordStore *CreateBVJoinStore(CPlan *plan, uint numBits);

    void assignRowNum( // Set row number from passed arg
        record_t *recP, TRowId rowNum);

    static ESchemaType schemaType(
        int nzType );  // NzType of field

    static int GetNextTempTableId();    //Get next unused temp table id

    // Distribution key support

    void    SetDistKeyFields(int numFields, int16* fieldIds);
    void    SetDistKeyFields(int numFields, int* fieldIds);
    int     GetNumDistKeyFields() const;
    int16 *GetDistKeyFields() { return m_distKeyFields; }
    int GetNumDistKeyFields() { return m_numDistKeyFields; }
    const int16 *GetDistKeyFields() const { return m_distKeyFields; }
    unsigned int i_distAlgorithm() const { return m_distAlgorithm; }
    void  i_setDistAlgorithm(unsigned int algor) {
        m_distAlgorithm = algor;
    }
    void  i_setDistKeyFields(int cnt, uint32 *flds) {
        NZ_VERIFY(cnt <= MAX_DIST_KEY_FIELDS);
        for ( int ix = 0; ix < cnt; ix++ )
            m_distKeyFields[ix] = flds[ix];
        m_numDistKeyFields = cnt;
    }

    uint16* GetDistKeyMap();
    int16 i_getDistKeyField (uint32 colNum);

    // virtual table / concurrency support
    bool IsStoredVt () const;

#if FOR_SPU
# ifndef TNS
    // support for AppendData/FlushAppendBuffer
    void RemoveInserter(const TXid& xId);

    // Insert/groom coordination.
    void CaptureGroomEotPages(TXid groomXid, bool concurrentInsertAllowed);
    void CaptureInsertEotPages(TXid insertXid, bool concurrentGroomAllowed);
    bool hasInserters();

    // Support for insertions into table being scanned and mview
    void captureWaterMark(int dsid, const CPlan* plan, TCHand tch );
    TRecLoc getWaterMark(int dsid, const TXid& xId);
# endif // TNS
#endif // FOR_SPU


/****************** ACCESSOR FUNCTIONS ********************************/

    static bool idIsTransient (table_id_t id) {
        return (unsigned int) id >= NZ_BASE_DBOS_TEMP_TABLE_ID;
    }

    static bool idIsSPUTransient (table_id_t id) {
        return (unsigned int) id >= NZ_BASE_SPU_TEMP_TABLE_ID;
    }

    static bool idIsPostgres (table_id_t id) {
        return (unsigned int) id <= MAX_SYSOID;
    }

#if !FOR_SPU
    int i_version () const {
        return 0;
    }
#endif

    int i_id() const {
        return m_id;
    }
    int i_Ttype() const {
        return (int)m_Ttype;
    }

    int i_fieldsVersion() const {
        return atomic_read(&m_fieldsVersion);
    }
    void i_incFieldsVersion() {
        atomic_inc(&m_fieldsVersion);
        NZ_VERIFY (i_fieldsVersion() > 0);
    }
    void i_setFieldsVersion(int version) {
        atomic_set(&m_fieldsVersion, version);
    }

    int i_parentId () const {
        return m_parentId;
    }
    void i_setParentId (int parentId) {
        m_parentId = parentId;
    }
    int i_parentOfAlt () const {
        return m_parentOfAlt;
    }
    void i_setParentOfAlt (int parentOfAlt) {
        m_parentOfAlt = parentOfAlt;
    }
    const TXid& i_xId () const {
        return m_xId;
    }
    void i_setXid (const TXid& xId ) {
        m_xId = xId;
    }
    bool i_isTransient () const {
       return idIsTransient(m_id);
    }

    static unsigned int dsliceIndex(int dsid);

    void DetachInsertStream(int dsid); // SPU datarcv special

    EDutyTypes i_fillType() const {
        return m_fillType;
    }
    void setFillType(int fillType);

    int i_numIndexFields () const {
        return m_numIndexFields;
    }
    unsigned int tblBlockSize() const {
        return NZ_TABLE_DEFAULT_BLOCK_SIZE;
    }

    void setTransTblType( uint8 type ) {
        m_transTblType = type;
    }

    uint8 i_virtualLocus () const {
        return m_virtualLocus;
    }
    void setVirtualLocus(uint8 place);

    int16 i_virtualMaterializer () const {
        return m_virtualMaterializer;
    }
    void i_setVirtualMaterializer (int16 val) {
        m_virtualMaterializer = val;
    }

    unsigned int i_virtualRingSize() const {
        return m_virtualRingSize;
    }
    void setVirtualRingSize(unsigned int val);

    bool i_virtualEnabled () const {
        return m_virtualEnabled;
    }
    void setVirtualEnabled(bool val);

    int i_defaultRecstoreSize() const {
        if ( m_virtualLocus && m_virtualRingSize
             && (DEFAULT_VTSMALL_RECSTORE_SIZE > 4 * m_shape->i_maxRecordLen()) )
            return(DEFAULT_VTSMALL_RECSTORE_SIZE);     // 4K
        else
            return(DEFAULT_RECSTORE_SIZE);                 // 128K
    }

    bool i_isMemoryTable() const {
        return m_virtualLocus;  // degenerates have virtual markings
    }

#if FOR_SPU
    const char *i_nameHack() const {
        return 0;
    } 
#else // FOR_SPU
    const char *i_nameHack() const {
        return m_nameHack;
    }

    const char *i_schemaName() const {
        return m_schemaName;
    }
    const char *i_catalogName() const {
        return m_catalogName;
    }

    void setNameHack(const char *name);
    void setSchemaName(const char* name);
    void setCatalogName(const char* name);
#endif // FOR_SPU

    // size of varying fields, just the actual payload
    vlen_t i_varFieldSizeContent(const void *fp) const {
        return *((const vlen_t *)fp) - sizeof(vlen_t);
    }

    void MarkAsTempTable() {
        m_isTempTable = true;
    }

    bool i_isTempTable() const {
        return m_isTempTable;
    }

#if !FOR_SPU
    void MarkAsBridgeTable();    // special for DbosTableCreate()
    bool isBridgeTable() const { return m_isBridgeTable; }

    void setReplicated(bool replicated) { m_isReplicated = replicated; }
    bool isReplicated() const { return m_isReplicated; }
#endif

    bool i_isThinTable () const {
        return m_parentId != m_id;
    }

    void i_setSourceLargestSlice (ESizeSource ss) {
        m_sourceLargestSlice = ss;
    }

    bool i_usedInMultipleJoins() const { //true if used in multiple joins
        return m_bUsedInMultipleJoins;
    }
    void setUsedInMultipleJoins(bool usedInMultipleJoins) {
        m_bUsedInMultipleJoins = usedInMultipleJoins;
    }

    bool i_isPrebroadcastTbl() const {
        return m_isPrebroadcastTbl;
    }
    void setIsPrebroadcastTbl(bool isPrebroadcastTbl) {
        m_isPrebroadcastTbl = isPrebroadcastTbl;
    }

    // Identifies broadcasted tables on both host and SPU
    bool isBroadcastTbl() const {
        return m_isBroadcastTbl;
    }
    void setIsBroadcastTbl(bool value) {
        m_isBroadcastTbl = value;
    }

    static void CreateDegenerateTables();

    bool i_distincter() const { //true if this table is for a distinct aggr
        return m_bDistincterTable;
    }

    void i_setDistincter(bool isDistincter) {
        m_bDistincterTable = isDistincter;
    }

    void i_setDataOnDisk(bool value) {
        m_dataOnDisk = value;
    }
    
    uint64 i_nTransPages () const {
        return m_nTransPages;
    }
    
    uint64 i_nTransRecords () const {
        return m_nTransRecords;
    }
    
#if !FOR_SPU
    uint64 i_nTransBytes () const {
        return m_nTransBytes;
    }
#endif

    uint64 i_nPagesInLargestSlice () const {
        return m_nPagesInLargestSlice;
    }
    
    uint64 i_nRecordsInLargestSlice () const {
        return m_nRecordsInLargestSlice;
    }
    
#if !FOR_SPU
    uint64 i_nBytesInLargestSlice () const {
        return m_nBytesInLargestSlice;
    }
#endif
    
    bool HasNullsInJoinCols()      const { return m_hasNullsInJoinCols; }
    void SetHasNullsInJoinCols(bool val) { m_hasNullsInJoinCols = val; }

    bool SqbHasRows()              const { return m_sqbHasRows; }
    void SetSqbHasRows(bool val)         { m_sqbHasRows = val; }

    uint8 i_transTblType() {
        return m_transTblType;
    }

    int i_numFields() {
        return m_shape->i_numFields();
    }

#if FOR_SPU
# ifndef TNS
    TableDslice *tableDslice(int dsid) {
        NZ_ASSERT(m_Ttype == TableCtx::eCTonDisk);
        NZ_ASSERT(m_tableDslice);
        unsigned int const dsdex = dsliceIndex(dsid);
        return &m_tableDslice[dsdex];
    }
    unsigned int tableDsliceIdx(const TableDslice *tds) const {
        unsigned int idx = tds - &m_tableDslice[0];
        NZ_ASSERT(idx < mt::mt_dsids->nr);
        return idx;
    }

    unsigned int tableDsliceDsid(const TableDslice *tds) const { 
        unsigned int idx = tableDsliceIdx(tds);
        return mt::mt_dsids->dsid[idx];
    }  
# endif // TNS
#endif

#if FOR_SPU
    int i_planId() const {
        return m_planId;
    }
    int i_snippetNum() const {
        return m_snippetNum;
    }
    void i_setPlanId(int planId) {
        m_planId = planId;
    }
    void i_setSnippetNum(int snipNum) {
        m_snippetNum = snipNum;
    }

    PageId groomEot(int dsid);

#endif // FOR_SPU

    void IncrRecStoreBcastAliasCnt() {
        atomic_inc(&m_recStoreBcastAliasCnt);
    }
    void SetSliceRecStore( int ix, CRecordStore* rsP );

#if !FOR_SPU
    uint64 lastModTime() {
        return __sync_fetch_and_add(&m_lastModTime,0);
    }
    void setLastModTime(uint64 lastModTime) {
        uint64 cachedTime = this->lastModTime();
        while (!__sync_bool_compare_and_swap(&m_lastModTime,cachedTime,lastModTime) && (cachedTime < lastModTime))
        {
            cachedTime = this->lastModTime();
        }
    }
#endif //!FOR_SPU

    bool hasPlan() {
        return m_plan != 0;
    }

    void *operator new (size_t);   // Shared memory constructor helper
    void operator delete (void *); // Shared memory constructor helper

    static void computeFieldSizes( // get logical and physical sizes
        int type,   // NZ_FIELD_XXX
        uint16 *size,  // IN/OUT logical size
        uint16 *trueSize); // OUT physical size (same as logical for nonvarying

/********************* PRIVATE ****************************************/

PROTECTED:
    int m_id;          // table instance id
    int m_parentId;    // master table id
    TXid m_xId;        // nonzero for transaction temp tables
    bool m_dataOnDisk; // true if table not in memory
    uint8 m_virtualLocus; // spu/host locus (e.g. VIRTUAL_LOCUS_SPU...)
    int16 m_virtualMaterializer; // record loading function in virtab.cpp
    TableShape *m_shape;
    bool m_isTempTable;     // true if temp table (SQL CREATE TEMP TABLE)
    int m_numIndexFields;   // number of columns specified in "order by" clause of MView
#if !FOR_SPU
    char *m_nameHack;  // !NOTE-bmz temp for dev
#endif
    int     m_numDistKeyFields; // <= MAX_DIST_KEY_FIELDS
    int16   m_distKeyFields[MAX_DIST_KEY_FIELDS];
    int m_parentOfAlt; // parent table if this is an alternate base table
    unsigned int m_distAlgorithm; // DIST_ALGOR_CRC_HASH=0, DIST_ALGOR_RND_ROBIN...

    TableCtx::ETableType m_Ttype;    // table sub-type
    CPlan *m_plan;                   // owning plan (for eCTslice type tables)
    bool m_bUsedInMultipleJoins;     //true if table used in multipe joins
    bool m_isPrebroadcastTbl;        //true if this is a pre-broadcasted or pre-collocated table
    bool m_isBroadcastTbl;           //true if this is a broadcasted table
    bool m_bDistincterTable;         //true if this is a distinct aggr table
    bool m_hasOutOfBandValsInBV;     //true if this is a BV table and has out-of-band values
    uint8  m_transTblType;           // transient table type.

    bool m_hasNullsInJoinCols;      // For NOT IN/relOp ALL
                                    // true  - join cols have nulls
                                    // false - join cols have no nulls
    bool m_sqbHasRows;              // For NOT IN/relOp ALL
                                    // true  - subquery has rows
                                    // false - subquery has no rows

    uint16* m_distKeyMap;

    uint m_virtualRingSize;  // if ring-buffer virtab, its size else 0
    bool m_virtualEnabled;   // whether loading of this table is enabled
    pthread_rwlock_t m_vtRwLock; // protect virtabs against simultaneous scans & writes

    //for long-term DHJ
    uint64 m_nTransPages;  //number pages in transient table
    uint64 m_nTransRecords;//number of records in transient table
#if !FOR_SPU
    uint64 m_nTransBytes; //number of bytes in transient table
    uint64 m_nBytesInLargestSlice; //number of bytes in largest table slice - transient only
#endif
    
    uint64 m_nPagesInLargestSlice;  //number pages in largest table slice "DHJ"
    uint64 m_nRecordsInLargestSlice;//number of records in largest table slice "DHJ"
    //END of members required by GENC
    bool m_bTableSizesInited; //table sizes have been initialized for DHJ, temp for debugging

    EDutyTypes m_fillType;       // Broadcast recordstores are special (weird)

    atomic_t m_tableRefCnt;      // table reference count

#if !FOR_SPU
    atomic_t m_exclusivePid;     // for validating exclusive access assumptions
#endif

    nzmutex_t m_tableLock;

#if FOR_SPU
    atomic_t m_tranPlansRefs; // for transient table: num plans attached
#endif // FOR_SPU

    atomic_t m_recStoreBcastAliasCnt; // broadcast table number of cloned CRecordStore instances

#if !FOR_SPU
    char *m_schemaName;             // cross-db
    char *m_catalogName;
    char *m_fullyQualifiedName;
#endif

    ESizeSource m_sourceLargestSlice;  // Who set m_nPagesInLargestSlice.

#if !FOR_SPU
    bool m_isBridgeTable; // [host only] true iff transient table created by pg
    bool m_isReplicated;  // [host only] is table replicated on subordinate node
#endif

    atomic_t m_fieldsVersion; // [host and spu] "schema version" (for alter table)
#if !FOR_SPU
    TXid     m_alterTXid;           // transaction altering the table
    TAlterColumnInfo *m_prevSizes;  // previous column sizes (in case of rollback of m_alterTXid)
    TXid     m_truncTXid;           // transaction truncating the table
    TruncateStats* m_truncateStats; // cached pre-truncate statistics
    bool m_notForSpu;         // true == do not instantiate this table on SPU (e.g. external tables)
    uint64 m_lastModTime;     // modification time for snippet result cache
#endif // FOR_SPU

    TranInfo *m_tranInfo;     // transient table extension info

public:

#if FOR_SPU
# ifndef TNS
    uint8 m_droppedOnSlice[MAX_SPU_DATASLICES / 8];
    TableDslice *m_tableDslice;  // non-transient table (eCTonDisk) per-dataslice info
# endif
    int m_planId;                   // plan id associated with this transient table.
    int m_snippetNum;               // snippet num associated with this transient table.
#endif // FOR_SPU

PROTECTED:

    void addRecordNoLock( // called from AddRecordToStore, who sets lock if necessary
        const void **valsP , unsigned char spc); // array of ptrs to field values and/or null ptrs

    void adjustMap(
        TPhysIDMap* pMap,
        int pNextPid );

    void adjustVtRing(unsigned int ringSize);

    void *evalExpression(       // Evaluate Postgres record based expression
        Node *expr,             // Postgress style expression from execution plan
        ExprContext *econtextP, // pg expr context for pg eval code mem allocs
        CPlan *pPlan,           // for expr contect
        record_t *recP,         // current record to use for field refs
        long int aggrBias=0);   // from ProjectNode

    void reclaimVtRing();

    static CTable *instantiateDegenerateTable(int protoTblId, bool isProto);
    void createDegenerateRecordStore();

    friend Datum NzFetchVar(
        Var *variable,  // pg var
        ExprContext *econtext,  // pg expression context
        bool *isNull);           // returned

public:

    dlink_t m_link;        // table dictionary link
    dlink_t m_hash_link;   // table dictionary link
    bool m_is_linked;

#if !FOR_SPU
    CBitVector m_dataSlicesWithData;
#endif
    bool filterDataSlices;

    atomic_t m_randomChunk;
    int getRandomDistChunkSize(int chunkSize = 0);

    // leakage to CDictionary
    // and... hcSpuMgr.cpp, _dump_table_dictionary
    // a/k/a SpuMgr::handleHwOpCb(), case: HWDUMP_TBL_DICTIONARY
    CRecordStore *diagRecordStore(void) const;

    // leakage to CRecordStore
    void dissociateRecordStore(CRecordStore *pRS);

private:

    void setVTRecordStore(CRecordStore *pRS);
    CRecordStore *setPlanRecordStore(CPlan *plan, CRecordStore *pRS);
    void cleanupTranInfo(void);

    // not implemented
    CTable(const CTable&);
    CTable& operator=(const CTable&);

};

#if FOR_SPU

LKTRKR_DECLARE_NAME(MTX_TABLEDS);

struct TDSLock : public auto_lock_trk {
    TDSLock(const CTable *table, TableDslice *tds) : auto_lock_trk(&tds->tds_lock, MTX_TABLEDS, table->tableDsliceDsid(tds)) { }
};

#endif // FOR_SPU


extern void *adjustedComputedNumericScale(
        void *fp, CTable *dst, field_t *fldP, int destFldSize = 0);

/* i_SimpleHasher() and i_SimpleHasherBe() now in simplehasherinline.h */

#if FOR_CLIENT || FOR_POSTGRES
// --------------------- Table creation interface (client.cpp)

extern CTable *DbosTableCreate(    // Create table descriptor
    unsigned int tblId,
    TableCtx::ETableType tTyp,
    bool isTempTable,
    bool isBridgeTable,
    bool isReplicated,
    int numFields)
        throw (CError);

extern void DbosTableColumnDefine(  // define basic column attributes
    CTable *table,
    unsigned int colId,
    int type,
    bool nullAllowed,
    int size)
        throw (CError);

extern void DbosTableCompleteCreation(  // Completes column definitions, writes table to disk
    unsigned int tblId)
        throw (CError);

extern void DbosTableAlterColSize(
    int tblid,
    int coldex,
    int newsize)
        throw (CError);

extern void DbosIndexCreate(
    unsigned int indexId,
    unsigned int parentId,
    int numFields,
    int *fieldList)
        throw (CError);

extern void DbosThinTableCreate (
    unsigned int indexId,
    unsigned int parentId,
    int numFields,
    int *fieldList,
    int numProject,
    int *projectList)
        throw (CError);

#endif // FOR_CLIENT

struct VtReadLock
{
    CTable *m_tbl;
    VtReadLock(CTable *tbl) : m_tbl(tbl) { m_tbl->lockVtRead(); }
    ~VtReadLock() { if (m_tbl) m_tbl->unlockVtRw(); }
    void Unlock() { if (m_tbl) { m_tbl->unlockVtRw(); m_tbl = 0; } }
};

struct VtWriteLock
{
    CTable *m_tbl;
    VtWriteLock(CTable *tbl) : m_tbl(tbl) { m_tbl->lockVtWrite(); }
    ~VtWriteLock() { if (m_tbl) m_tbl->unlockVtRw(); }
    void Unlock() { if (m_tbl) { m_tbl->unlockVtRw(); m_tbl = 0; } }
};

struct TableGuard
{
    CTable *m_tbl;
    TableGuard(CTable *tbl) : m_tbl(tbl) { m_tbl->lockTable(); }
    ~TableGuard() { if (m_tbl) m_tbl->unlockTable(); }
    void Unlock() { if (m_tbl) { m_tbl->unlockTable(); m_tbl = 0; } }
};

#if !FOR_SPU
struct TableExclusive
{
    CTable *m_table;
    TableExclusive(CTable *table, int pid) : m_table(table) { table->beginExclusive(pid); }
    ~TableExclusive() { m_table->endExclusive(); }
};

#define DECLARE_TABLE_EXCLUSIVE_SCOPE(table) TableExclusive _exclusive((table), NzMyProcessId)

#endif //!FOR_SPU

#endif // TABLE_INCLUDED
