/*
 * Copyright 2009, 2014 IBM Corporation.
 */
 /* tableshape.h
 * Copyright (c) 2011 IBM/Netezza
 */

#ifndef TABLESHAPE_INCLUDED
#define TABLESHAPE_INCLUDED

#include "error.h"
#include "fpgaprepapi.h"
#include "hashenum.h"
#include "nzlimits.h"
#include "nzlog.h"
#include "nzrecord.h"
#include "projectspecials.h"
#include "TXID.h"
#include "utils.h"


void computeFieldSizes(         // get logical and physical sizes
          int type,             // NZ_FIELD_XXX
          uint16 *size,         // IN/OUT logical size
          uint16 *trueSize);    // OUT physical size (same as logical for nonvarying

#define COLUMN_CANT_NULL (-1)

#define DEFER_SCALE 0xffff // signals scale to be determined from data values
#define DEFER_FIXED_CHAR_SIZE 0xffff // signals size to be determined from data values

#define PRECISION_SCALE(precision, scale) (((precision)<<8) + (scale))
#define PRECISION_FROM_SIZE(size) (((size) >> 8) & 0x7F)
#define BYTECOUNT_FROM_PRECISION(precision) \
    ((precision)<=9  ? sizeof(uint32) : \
     (precision)<=18 ? 2*sizeof(uint32) : \
     4*sizeof(uint32))

#define SCALE_FROM_SIZE(size) ((size) & 0x00FF)

#define TYPE_IS_VARYING(type) (type==NzTypeVarChar || type==NzTypeVarFixedChar || \
                                  type==NzTypeNChar || type==NzTypeNVarChar || \
                                  type==NzTypeGeometry || type==NzTypeVarBinary)
#define TYPE_IS_STRING_OR_BINARY(type) (type == NzTypeChar || (TYPE_IS_VARYING(type)))
#define TYPE_IS_NSTRING(type) (type==NzTypeNChar || type==NzTypeNVarChar)
#define TYPE_IS_INT(type) (type==NzTypeInt || type==NzTypeInt1 || \
                           type==NzTypeInt2 || type==NzTypeInt8)

// We keep TYPE_IS_DEPRECATED as a last ditch defense against creating a
// CTable with these never-perfected/never-implemented types. See table.cpp
#define TYPE_IS_DEPRECATED(type) (type==NzDEPR_Text || type==NzDEPR_Blob || \
                                  type==NzDEPR_NText)

extern ESchemaType NzSchemaTypeTable[];

inline
ESchemaType schemaType(
    int nzType )
{
    return NzSchemaTypeTable[nzType];
}

#define HASH_AS_CHAR(type)    ((NzSchemaTypeTable[type]==SCHEMA_TYPE_FIXEDCHAR) || \
                               (NzSchemaTypeTable[type]==SCHEMA_TYPE_VARCHAR))
#define HASH_AS_INT(type)     (NzSchemaTypeTable[type]==SCHEMA_TYPE_INT)
#define HASH_AS_NUMERIC(type) (NzSchemaTypeTable[type]==SCHEMA_TYPE_NUMERIC)
#define HASH_AS_TIME(type)    (NzSchemaTypeTable[type]==SCHEMA_TYPE_TIME)
#define HASH_AS_FLOAT(type)   (NzSchemaTypeTable[type]==SCHEMA_TYPE_FLOAT)



struct TPhysIDMap { // Used in preserving physical field IDs across projects.
    int type;       // Nz data type
    int pid;        // physical ID (or -1)
};

// Alignment boundaries of interest
enum EAlignSize {
    SIZE_128,
    SIZE_64,
    SIZE_32,
    SIZE_16,
    SIZE_8,
    SIZE_FIXED_CHAR1, // special cases for fixed-char fields of possible sizes
    SIZE_FIXED_CHAR2,
    SIZE_FIXED_CHAR3,
    SIZE_FIXED_CHAR4,
    SIZE_FIXED_CHAR5,
    SIZE_FIXED_CHAR6,
    SIZE_FIXED_CHAR7,
    SIZE_FIXED_CHAR8,
    SIZE_FIXED_CHAR9,
    SIZE_FIXED_CHAR10,
    SIZE_FIXED_CHAR11,
    SIZE_FIXED_CHAR12,
    SIZE_FIXED_CHAR13,
    SIZE_FIXED_CHAR14,
    SIZE_FIXED_CHAR15,
    SIZE_FIXED_CHAR16=MAX_FIXED_CHAR_LENGTH+SIZE_8,
    SIZE_VARCHAR
};

enum CloneType {
    CLONE_DEFAULT,
    CLONE_WITHOUT_SPECIALS,
    CLONE_WITHOUT_SPECIALS_ADD_NULLS,
    CLONE_WITH_SPECIALS,
};


// struct TableShape: describes the shape of a table, including the number
// of fields and the field types.
//
// Note: This structure often resides in shared memory and may be accessed
// by different executables, so virtual methods are not permitted.

struct TableShape {

    int m_leaderSize;   // bytecount of recSize & nulls flags bytes (may be 0)
    int m_specialsOffset;    // where special fields begin
    int m_numFixedFields;
    int m_fixedFieldsSize;   // bytecount of fixed fields
    int m_numVaryingFields; // how many fields are variant (after fixed)
    int m_minRecordSize;    // space if all varyings are empty
    int m_maxRecordSize;    // space if all varyings are fully expanded
    bool m_suppressMaxCheck;
    bool m_nullsAllowed;    // if any fields allow null values
    bool m_sizeWord;     // if TRUE, a size word preceeds record
    TUseSpecialFields m_enableSpecials; // what special fields in record
    TUseSpecialFields m_stickySpecials; // flds that carry across projections
    TUseSpecialFields m_diskSpecials;   // record image on disk has which specials?
    int m_sizeWordSize;   //!< 0/2/4 bytecount of length word
    int m_numFields;
    field_t m_fields[0];    // actual length specified by m_numFields

private:

    TableShape(const TableShape&); // not implemented
    TableShape& operator=(const TableShape&); // not implemented


    // Private here because it probably belongs somewhere else
    static
    unsigned _beFetch_vlen_t(char const* p) { return ((*(unsigned char const*)p) << 8) | *(unsigned char const*)(p+1); }

public:

    // Default constructor returns a 0-field zeroed TableShape,
    // which caller can then adjust as long as caller is OK with the 0-field size tableshape.
    // This is not entirely useless: TNS/Wahoo "delete" uses this to create
    // 0-field tableshapes and then sets the dtid special and then uses these
    // to create dtid CCC's...
    // However, most users of TableShape should use the factory methods in CTable
    // instead of this constructor.
    TableShape() { memset(this, 0, sizeof(*this)); }

    static TableShape *MakeTransientShape(
        unsigned int numFields);

    void init(int numFields, bool isTransient);

    // Note: Do not make this virtual; see "Note to Wise" above.
    ~TableShape() { }

    void decodeCloneType(CloneType how,
                         TUseSpecialFields *pSpecials,
                         TUseSpecialFields *pStickySpecials,
                         bool *addnulls) const;

    TableShape* Clone(
        TUseSpecialFields specials,
        TUseSpecialFields stickySpecials,
        bool forceNulls = false) const;

    TableShape *Clone(CloneType how) const;

    TableShape& CopyFrom(
        const TableShape& src,
        TUseSpecialFields specials, // special fields enabling bitmask
        TUseSpecialFields stickySpecials, // sticky special fields enabling bitmask
        bool bForceNulls);

    void setFieldFlags(int coldex, uint8 flags) {
        NZ_ASSERT((coldex < m_numFields) && (coldex >= 0));
        m_fields[coldex].flags = flags;
    }

    const field_t *i_field(int coldex) const {
        NZ_ASSERT((coldex < m_numFields) && (coldex >= 0));
        //NZ_ASSERT(!TABLE_EXCLUSIVE_NOT_SELF(this));
        return &m_fields[coldex];
    }

    int i_pid (int lid) const {
        NZ_ASSERT((lid < m_numFields) && (lid >= 0));
        return m_fields[lid].physField;
    }

    // Map a logical column to its physical index.
    // Can deal with stored specials and the user columns.
    int i_physicalId(ColumnId) const;

    int i_lid (int pid) const {
        NZ_ASSERT((pid < m_numFields) && (pid >= 0));
        return m_fields[pid].logField;
    }

    EFieldType i_fieldType (int coldex) const;
    int i_fieldSize(int coldex) const;

    bool i_fieldPadEbcdic (int lid) const {
        if(lid < 0)
            // No SpecialField is of char/varchar type
            return false;
        NZ_ASSERT(lid < m_numFields);
        return m_fields[lid].flags & FieldPadEbcdic;
    }

    void setFieldPadEbcdic(int lid, bool value) {
        NZ_ASSERT((lid < m_numFields) && (lid >= 0));
        if (value)
            m_fields[lid].flags |= FieldPadEbcdic;
        else
            m_fields[lid].flags  &= ~FieldPadEbcdic;
    }

    void i_setFieldTrueSize (int coldex, int size) {
        NZ_ASSERT((coldex < m_numFields) && (coldex >= 0));
        m_fields[coldex].trueSize = size;
    }

    int i_fieldTrueSize (int coldex) const {
        NZ_ASSERT((coldex < m_numFields) && (coldex >= 0));
        return m_fields[coldex].trueSize;
    }

    int i_fieldInstanceSize(void *fieldP,int coldex) const {
        return TYPE_IS_VARYING(m_fields[coldex].type) ?
                   *((vlen_t *)fieldP) : m_fields[coldex].trueSize;
    }

    // i_fieldInstanceSize for a big-endian (spu) record
    int i_fieldInstanceSizeBe(void *fieldP,int coldex) const
    {
        return TYPE_IS_VARYING(m_fields[coldex].type) ?
            _beFetch_vlen_t((char const*)fieldP) : m_fields[coldex].trueSize;
    }

    int i_singleRepresentation (int coldex) const {
        NZ_ASSERT((coldex < m_numFields) && (coldex >= 0));
        return m_fields[coldex].type != NzTypeInterval &&
               m_fields[coldex].type != NzTypeTimeTz;
    }

    int i_fieldScale (int coldex) const {
        NZ_ASSERT((coldex < m_numFields) && (coldex >= 0));
        NZ_VERIFY(m_fields[coldex].size != DEFER_SCALE);
        return SCALE_FROM_SIZE(m_fields[coldex].size);
    }

    int i_fieldPrecision (int coldex) const {
        NZ_ASSERT((coldex < m_numFields) && (coldex >= 0));
        return PRECISION_FROM_SIZE(m_fields[coldex].size);
    }

    bool i_fieldSizeIsDeferred(int coldex) const {
        NZ_ASSERT((coldex < m_numFields) && (coldex >= 0));
        return (m_fields[coldex].type==NzTypeNumeric && m_fields[coldex].size==DEFER_SCALE) ||
            ((m_fields[coldex].type==NzTypeChar || m_fields[coldex].type==NzTypeVarFixedChar) &&
             m_fields[coldex].size==DEFER_FIXED_CHAR_SIZE);
    }

    void i_setFieldSize(int coldex, int size) {
         NZ_ASSERT((coldex < m_numFields) && (coldex >= 0));
         m_fields[coldex].size = size;
    }

    int i_fieldNumericDigit32Count (int coldex) const {
        NZ_ASSERT((coldex < m_numFields) && (coldex >= 0) && (m_fields[coldex].type==NzTypeNumeric));
        return (m_fields[coldex].trueSize / sizeof(TNumericDigit));
    }

    static int i_fieldNumericDigit32Count (const field_t *fldP) {
        NZ_ASSERT(fldP->type == NzTypeNumeric);
        return fldP->trueSize / sizeof(TNumericDigit);
    }

    static int i_fieldNumericByteCountFromSize (int size) {
        int precision = PRECISION_FROM_SIZE(size);
        return BYTECOUNT_FROM_PRECISION(precision);
    }

    const field_t *i_fields() const {
        return m_fields;
    }

    int i_numFields() const {
        return m_numFields;
    }

    int i_maxRecordLen () const {
        return m_maxRecordSize;
    }

    int i_maxDiskRecordLen () const {
        return m_maxRecordSize - m_leaderSize + i_diskLeaderSize();
    }

    bool IsFieldNullable( int lid ) const {
        return m_fields[lid].nullAllowed;
    }

    int i_fixedFieldsSize () const {
        return m_fixedFieldsSize;
    }

    int i_numFixedFields () const {
        return m_numFixedFields;
    }

    bool i_isFixedField(int col) const {
//         return(col < m_numFixedFields);  WRONG, would be physId (12/15/01)
        NZ_ASSERT((col < m_numFields) && (col >= 0));
        return m_fields[col].fixedSize;
    }

    int i_numVaryingFields () const {
        return m_numVaryingFields;
    }

    int i_leaderSize () const {
        return(m_leaderSize);
    }

    int i_diskLeaderSize () const {
        return i_diskLengthOfSpecials() + i_sizeAndNullsLength();
    }

    int i_sizeWordSize () const {
        return m_sizeWordSize;
    }

    bool i_sizeWord () const {
        return m_sizeWord;
    }

    int i_specialsOffset () const {
        return m_specialsOffset;
    }

    TUseSpecialFields i_enableSpecials () const {
        return m_enableSpecials;
    }

    TUseSpecialFields i_stickySpecials () const {
        return m_stickySpecials;
    }

    int i_nullsSize () const {
        //return((m_numFields+7)/8);
        return m_nullsAllowed ? ((m_numFields+7)>>3) : 0;  // faster?
    }

    int i_sizeAndNullsLength () const {
        return i_leaderSize() - i_lengthOfSpecials();
    }

    int i_diskOffsetOfDeletedTXid () const {
        NZ_ASSERT(m_diskSpecials.txDeleted);
        return ( m_diskSpecials.rowId ? sizeof(TRowId) : 0 )
             + ( m_diskSpecials.txCreated ? sizeof(TXid) : 0 );
        //NZ_ASSERT(m_diskSpecials.rowId && m_diskSpecials.txCreated && m_diskSpecials.txDeleted);
        //return(sizeof(TRowId) + sizeof(TXid));
    }

    int i_diskOffsetOfCreatedTXid () const {
        NZ_ASSERT(m_diskSpecials.txCreated);
        return ( m_diskSpecials.rowId ? sizeof(TRowId) : 0 );
        //NZ_ASSERT(m_diskSpecials.rowId && m_diskSpecials.txCreated && m_diskSpecials.txDeleted);
        //return(sizeof(TRowId));
    }

    int i_diskLengthOfSpecials () const {
        return( m_diskSpecials.rowId ? sizeof(TRowId) : 0 )
            + ( m_diskSpecials.txCreated ? sizeof(TXid) : 0 )
            + ( m_diskSpecials.txDeleted ? sizeof(TXid) : 0 );
        //NZ_ASSERT(m_diskSpecials.rowId && m_diskSpecials.txCreated && m_diskSpecials.txDeleted);
        //return(sizeof(TRowId) + 2*sizeof(TXid));
    }

    int i_lengthOfSpecials () const {
        return( m_leaderSize - m_specialsOffset );
    }

    int i_nullsAllowed () const {
        return m_nullsAllowed;
    }

    TRowId i_rowNum(const record_t *recP) const {
        NZ_ASSERT(i_enableSpecials().rowId);
        return *((TRowId*)(recP + m_specialsOffset));
    }

    TRowId *i_rowNumP(const record_t *recP) const {
        NZ_ASSERT(m_enableSpecials.rowId);
        return (TRowId*)(recP + m_specialsOffset);
    }

    bool i_enableRowAddr () const {
        return m_enableSpecials.recOrd;
    }

    bool i_enableDTIDBitAddr () const {
        return m_enableSpecials.dtidOff;
    }

    bool i_enableRowNum () const {
        return m_enableSpecials.rowId;
    }

    bool i_enableTxCreated () const {
        return m_enableSpecials.txCreated;
    }

    bool i_enableTxDeleted () const {
        return m_enableSpecials.txDeleted;
    }

    void i_setEnableRowNum (bool val) {
        m_enableSpecials.rowId = val;
    }

    bool i_enableDiskRowNum () const {
        return m_diskSpecials.rowId;
    }

    bool i_enableDiskTxCreated () const {
        return m_diskSpecials.txCreated;
    }

    bool i_enableDiskTxDeleted () const {
        return m_diskSpecials.txDeleted;
    }

    bool isFieldSoleDistCol(int lid) const {
        return lid >= 0 && i_field(lid)->flags & FieldIsSoleDistCol;
    }
    void setFieldSoleDistCol(int lid) {
       const field_t *fp = i_field(lid);
       // Zone map logic expects this bit to be set _ONLY_ if data slice
       // determination is viable.  That is the column can be both zone
       // mapped and a hashed distribution column.  To be even more
       // conservative we currently exclude floating point plus those
       // types whose representation in memory and in zone maps differ.
       if (fp->type != NzTypeFloat && fp->type != NzTypeDouble && (fp->type == NzTypeChar || fp->trueSize <= 8))
           setFieldFlags(lid, fp->flags | FieldIsSoleDistCol);
    }

    bool isFieldZoneMapped(int lid) const {
        return lid < 0 ? TableShape::IsZMapSpecial(lid) : i_field(lid)->flags & FieldIsZoneMapped;
    }

    void setFieldZoneMapped(int lid, bool value) {
        uint8 flags = i_field(lid)->flags;
        if (value)
            flags |= FieldIsZoneMapped;
        else
            flags &= ~FieldIsZoneMapped;
        setFieldFlags(lid, flags);
    }

    // Is this type zone mappable for mview ORDER BY columns and DOT
    // ORGANIZE ON columns?
    bool isFieldZoneMappable(int lid) const {
        // All types except 128 bit numerics are zonemapped.
        return lid < 0 ?
                 IsZMapSpecial(lid)
               // All user types except 128 bit numerics are zonemappable.
               : ! (i_fieldType(lid) == NzTypeNumeric && i_fieldNumericDigit32Count(lid) > 2);
    }


    // Is this type zone mappable by default on all tables?
    static bool isFieldZoneMapDataTypeDefault (int fieldType)
    {
        switch ( fieldType ) {    // current default zone-mappable types
        case NzTypeDate:
        case NzTypeTimestamp:
        case NzTypeInt:
        case NzTypeInt1:
        case NzTypeInt2:
        case NzTypeInt8:
            return true;
        default:
            return false;
        }
    }


    void *specialFieldAt(char *recordP, int field) const;

    void *i_fieldAt(char *recordP, int field) const;

    void *diskSpecialFieldAt(char *recordP, int field) const;

    void *i_diskFieldAt(char *recordP, int field) const;

    void *FieldAt( // Return address of field given record & fdex
        char *recordP,         // ptr to record
        int field) const;      // field dex

    //!NOTE-dk This is here to catch callers of the original CTable method
    //that became a CRecordStore method (for the array of record pointers)
    //Unfortunately, it is a link-time catch, not a compile time catch
    //for table->FieldAt(0,k); [compiler happy to call 0 a "(char *) 0"]
    void *FieldAt( // Return address of field given recDex & fdex
        int recNum,            // record pos in m_records array (0-N)
        int field) const;      // field dex

    void *FieldAtBe( // FieldAt for a big-endian record (from spu)
        char *recordP,
        int field) const;

    void *DiskFieldAt( // Return address of field given ptr to rec in disk buffer
        char *recordP,  // ptr to start of raw disk record
        int field) const; // field dex

    void FetchSpecials( // Get special fields (rowAddr,rowNum,txCreated,txDeleted)
        record_t const* recP,   // record to fetch from
        TRecLoc* ordLocP,       // returned recOrdLoc    if m_enableSpecials.recOrd
        TRowId* rowNumP,        // returned rowId        if m_enableSpecials.rowId
        TXid* txCreatedP,       // returned txCreated    if m_enableSpecials.txCreated
        TXid* txDeletedP,       // returned txDeleted    if m_enableSpecials.txDeleted
        TRecLoc* dtidLocP) const; // returned dtidOffLoc if m_enableSpecials.dtidOff

    void SetSpecials( // Set special fields (rowAddr,rowNum,txCreated,txDeleted)
        record_t *recP,         // record to set in
        TRecLoc ordLoc,         // recOrd    if m_enableSpecials.recOrd
        TRowId rowNum,          // rowId     if m_enableSpecials.rowId
        TXid const& txCreated,  // txCreated if m_enableSpecials.txCreated
        TXid const& txDeleted,  // txDeleted if m_enableSpecials.txDeleted
        TRecLoc dtidLoc) const; // dtidOff   if m_enableSpecials.dtidOff

    void FetchDiskSpecials( // Get special fields (rowNum,txCreated,txDeleted)
        const record_t *recP,    // record to fetch from
        TRowId *rowNumP,         // returned rowId if m_diskSpecials.rowId
        TXid *txCreatedP,        // returned txCreated if m_diskSpecials.txCreated
        TXid *txDeletedP) const; // returned txDeleted if m_diskSpecials.txDeleted

    void SetDiskSpecials( // Set special fields (rowNum,txCreated,txDeleted)
        record_t *recP,         // record to set in
        TRowId rowNum,          // rowId if m_diskSpecials.rowId
        const TXid& txCreated,  // txCreated if m_diskSpecials.txCreated
        const TXid& txDeleted) const; // txDeleted if m_diskSpecials.txDeleted

    size_t emptyRecordSize() const; // size of empty record

    void MakeEmptyRecord(void *buf) const;

    const void ValidateShape() const; 
    
    void VerifyRecordSize (const char *recP) const { // Validate actual size of record in memory
        if ( m_numVaryingFields )
        {
            // Size includes itself, so can't correctly be zero
            // This 'verify' will prevent some infinite loops
            //  where *recP is used to advance from one rec to next.
            const uint16 size = *((const uint16*)recP);
            NZ_CHECK_DUMP( (size >= m_minRecordSize) && ((size%4) == 0) && (size <= m_maxRecordSize), NZ_ERROR_INCORRECT_DATA_SIZE, "Bad record length", size);
        }
    }

    int RecordSize (const char *recP) const { // Return actual size of record in memory
        if ( !m_numVaryingFields )
            return m_maxRecordSize;
        else {
            VerifyRecordSize (recP);
            return *((const uint16*)recP);
        }
    }

    void Dump(const char *title=0,
              int detail=NzDumpDetail) const;

    int BuildTableDescription(
        TableDescriptionEntry* pTdeArray ) const;

    int AdjustTableDescriptionForCompression(
        TableDescriptionEntry* tableDescP,
        int oldCount,
        bool returnRowAddr,
        bool returnDTIDBitAddr,
        TableDescriptionEntry* compTableDescP )  // OUT: must be large enough for
            const;                               //  worst case (32 entries

    void ComputeNullVecSizes(
        int* compiledPageNullVecBitsP,
        int* uncompiledPageNullVecBytesP ) const;

    void DefineField( // Set field attributes
        int fdex,   // ordinal 0 - (m_numFields-1)
        int type,   // NZ_FIELD_TYPE_XXXX...
        bool nullAllowed=false, // true if nulls value is allowed for this field
        unsigned int size=0xffff,   // for char(N), varchar(N)...
        int* physIDs=0, // if non-zero, physical IDs in logical-ID order
        bool springField=false, // Compensates for changes to preceding varyings
        uint8 flags = FieldNone);   // FieldNone, FieldOrderBy, etc

    void DefineField( // Set field attributes using passed model field
        int fdex,       // ordinal 0 - (m_numFields-1)
        const TableShape *src,    // table model field is from
        int srcDex,      // position in src table to get model field from
        bool forceAllowNull=false,    // overrides field's null allowance
        int* physIDs=0); // if non-zero, physical IDs in logical-ID order

    void finishCreate( // called after last field defined
        int* physIDs);    // if non-zero, physical IDs in logical-ID order

    void updateField(   // Set/modify field attributes
        bool firstTime, // false=DefineField, true=AlterField
        int fdex,   // ordinal 0 - (m_numFields-1)
        int type,   // NZ_FIELD_TYPE_XXXX...
        unsigned int size,   // for char(N), varchar(N)...
        bool nullAllowed=false, // true if nulls value is allowed for this field
        int* physIDs=0, // if non-zero, physical IDs in logical-ID order
        bool springField=false, // Compensates for changes to preceding varyings
        uint8 flags = FieldNone);   // FieldNone, FieldOrderBy, etc

    void assignPhysIds(
        EAlignSize size,        // field size of interest this time
        ESchemaType comp,       // schema type of interest this time
        TPhysIDMap* pMap,       // adjustable map
        int* pNextPid );        // pointer to next PID to be assigned

    void assignPhysIds(
        EAlignSize size,        // field size of interest this time
        ESchemaType comp,       // schema type of interest this time
        int *pPhysFieldDex,     // running physical-field ID
        int *pOffset,           // running record offset
        field_t **fPP);         // address of field_t array

    int ColumnByteOffset(int fdex) const
    {
        NZ_ASSERT(fdex < m_numFields);
        return m_fields[fdex].offset;
    }

    bool FieldIsSpring(int field) const
    {
        NZ_ASSERT(field < m_numFields);
        return m_fields[field].springField;
    }

    int NullBitOffset(int fdex) const;

    bool IsVarField( // Return whether passed field dex is a var field
        int dex) const;

    int DiskRecordSize( // Return size of record in Disk Buffer
        const char *recordP) const;

    void VfyRecordSize( // Verify varlens match size word & other sanity checks
        const record_t *recordP,     // points to beginning of record
        bool diskFormat) const;      // false if memory format

    void SetNullFlag( // Set the null flag for the given field
        record_t *recP,  // the record
        int fdex) const; // field index

    void i_SetNullFlag( // Set the null flag for the given field
        record_t *recP,  // the record
        int fdex) const // field index
    {
        NZ_VERIFY(m_nullsAllowed);
        fdex = m_fields[fdex].physField;
        NzSetBit((uint8*)recP+m_sizeWordSize,fdex);
    }

    void ClearNullFlag( // Clear the null flag for the given field
        record_t *recP,  // the record
        int fdex) const; // field index

    bool IsNull( // return true if null flag is set
        const record_t *recP, // the record
        int fdex) const;      // field index

    void AdjustVarField( // Update varying field, adjusting spring to keep same recsize
         char *rec, int fdex, int bodyLen);

    static ESimpleHashType FindHashType(int type,int size);

    ESimpleHashType CheckForSimpleHashing( // Return whether can use degenerate hash scheme
        int numCols,    // how many columns we are hashing on
        const int *colNums) const;   // array of the columns

    static uint32 hashOneColumn(int fieldType, void *srcRec,int len,uint32 hash, bool hashTimeAsInt,uint8 c_space, bool khash);

    uint32 GetHashForRec( // For a given record, return hash value
        record_t *srcRec, int numCols, const int *colNums, bool khash=false) const;

    uint32 GetHashForRecBe( // GetHashForRec for a big-endian record
        record_t *srcRec, int numCols, const int *colNums) const;

    static char *i_fixedFieldPtr(record_t *recP, int offset) {
        return (char *)recP + offset;
    }

    static char *i_varFieldPtr(record_t *recP, int fixedOffset, int varDex) {
        char *lenP;
        int  ctr;
        lenP = (char *)recP + fixedOffset;
        for (ctr = 0; ctr < varDex; ctr++) {
            lenP += ALIGN_VARLENS(*(vlen_t *)lenP);
        }
        return lenP;
    }

    // i_varFieldPtr for a big-endian (spu) record
    static char *i_varFieldPtrBe(record_t *recP, int fixedOffset, int varDex)
    {
        char *lenP;
        int  ctr;
        lenP = (char *)recP + fixedOffset;
        for (ctr = 0; ctr < varDex; ctr++) {
            lenP += ALIGN_VARLENS(_beFetch_vlen_t(lenP));
        }
        return lenP;
    }

    static bool IsZMapSpecial( // Passed field dex is a zmap-supported special
        int dex)
    {
        return (dex == PROJECT_ROWNUM) || (dex == PROJECT_CREATETXN);
    }

    static bool IsSpecialField( // Return whether passed field dex is a special
        int dex)
    {
        return (dex == PROJECT_ROWNUM) || (dex == PROJECT_CREATETXN) || (dex == PROJECT_DELETETXN);
    }

    static EFieldType specialFieldType (int coldex);

    static int specialFieldSize (int coldex);


    // Count the number of non generated columns in the table (including the
    // specials).  The compressor is going to compress exactly these many
    // number of columns.  And the SL needs to store only these many
    // columns.
    unsigned int nonGeneratedColumnCount() const {
        unsigned int numSpecials = 0;
        if (i_enableDiskRowNum()) numSpecials ++;
        if (i_enableDiskTxCreated()) numSpecials ++;
        if (i_enableDiskTxDeleted()) numSpecials ++;
        return i_numFields() + numSpecials;
    }

    static void computeFieldSizes( // get logical and physical sizes
        int type,    // NZ_FIELD_XXX
        uint16 *size,  // IN/OUT logical size
        uint16 *trueSize); // OUT physical size (same as logical for nonvarying)
};

inline
void TableShape::init(int numFields, bool isTransient)
{
    m_numFields = numFields;

    m_diskSpecials.txCreated = true;
    m_diskSpecials.txDeleted = true;
    m_diskSpecials.rowId = true;

    if (!isTransient)   // when 'real' tables have recordsets, they always have rowNums
        m_enableSpecials.rowId = true;
}

inline
EFieldType TableShape::specialFieldType (int coldex) {
    switch( coldex ) {

    case PROJECT_ROWADDR:     //  -2 used to place rowAddrs in projected field
        return NzTypeRecAddr;

    case PROJECT_SPUID:       //  -7 place spu id in projected field
        return NzTypeInt2;

    case PROJECT_ROWNUM:      //  -3 used to place rowNums in projected field
    case PROJECT_CREATETXN:   //  -8 place created txn id in projected field
    case PROJECT_DELETETXN:   //  -9 place deleted txn id in projected field
    case PROJECT_PAGEID:      // -12
    case PROJECT_EXTENTID:    // -13
        return NzTypeInt8;

// PROJECT_BLOCKNUM has been superseded by PROJECT_PAGEID and is no longer used,
// should be removed from all code that checks for it.
//    case PROJECT_BLOCKNUM:    // -10 used to project rowAddr's blocknum
//        return NzTypeInt;

    case PROJECT_DTIDBITADDR: // -11 put deleted txn id's bit address in proj fld
        return NzTypeDTIDBitAddr;
    case PROJECT_ENDOFPARTITION:
	return NzTypeBool;
    case PROJECT_EXPR:        //  -1 PG expr evaluator computed expression
    case PROJECT_HASH:        //  -4 used to project in a hash column
    case PROJECT_BYTE:        //  -5 !NOTE-bmz need more consisteny scheme!
    case PROJECT_COMPUTE:     //  -6 fsm based computed expression
    default:
        NZ_VERIFY(false);
    }
}

inline
EFieldType TableShape::i_fieldType (int coldex) const
{
    if (coldex < 0)
        return specialFieldType(coldex);

    NZ_ASSERT(coldex < m_numFields);
    return (EFieldType) m_fields[coldex].type;
}

inline
int TableShape::specialFieldSize (int coldex) {
    switch( coldex ) {

    case PROJECT_ROWADDR:     //  -2 used to place rowAddrs in projected field
    case PROJECT_ROWNUM:      //  -3 used to place rowNums in projected field
    case PROJECT_DTIDBITADDR: // -11 put deleted txn id's bit address in proj fld
        return sizeof( TRowId );

    case PROJECT_SPUID:       //  -7 place spu id in projected field
        return sizeof(int16);

    case PROJECT_CREATETXN:   //  -8 place created txn id in projected field
    case PROJECT_DELETETXN:   //  -9 place deleted txn id in projected field
        return sizeof( TXid );

// PROJECT_BLOCKNUM has been superseded by PROJECT_PAGEID and is no longer used,
// should be removed from all code that checks for it.
//    case PROJECT_BLOCKNUM:    // -10 used to project rowAddr's blocknum
//        return sizeof(int32);

    case PROJECT_PAGEID:      // -12
    case PROJECT_EXTENTID:    // -13
        return sizeof(int64);
    case PROJECT_ENDOFPARTITION:
	return sizeof(bool);
    case PROJECT_EXPR:        //  -1 PG expr evaluator computed expression
    case PROJECT_HASH:        //  -4 used to project in a hash column
    case PROJECT_BYTE:        //  -5 !NOTE-bmz need more consisteny scheme!
    case PROJECT_COMPUTE:     //  -6 fsm based computed expression
    default:
        NZ_VERIFY(false);
    }
}

inline
int TableShape::i_fieldSize(int coldex) const
{
    if (coldex < 0)
        return specialFieldSize(coldex);

    NZ_ASSERT(coldex < m_numFields);
    return m_fields[coldex].size;
}

inline
void* TableShape::i_fieldAt(char *recordP, int field) const
{
    if (IsSpecialField(field)) {
        return specialFieldAt(recordP, field);
    }
    NZ_ASSERT((field < m_numFields) && (field >= 0));
    if (m_fields[field].fixedSize)
        return (char *)recordP + m_fields[field].offset;
    return i_varFieldPtr(recordP, m_fixedFieldsSize, m_fields[field].offset);
}

inline
void* TableShape::i_diskFieldAt(char *recordP, int field) const
{
    if( IsSpecialField(field)) {
        return diskSpecialFieldAt(recordP, field);
    }
    recordP += i_diskLengthOfSpecials();    // get beyond disk specials
    int skipMe = i_lengthOfSpecials();  // see note above!
    NZ_ASSERT((field < m_numFields) && (field >= 0));
    if ( m_fields[field].fixedSize )
        return (char *)recordP + m_fields[field].offset - skipMe;
    return i_varFieldPtr(recordP, m_fixedFieldsSize-skipMe, m_fields[field].offset);
}

inline void
TableShape::decodeCloneType(CloneType how,
                            TUseSpecialFields *pSpecials,
                            TUseSpecialFields *pStickySpecials,
                            bool *addnulls) const
{
    static TUseSpecialFields const nospecials = { 0 };

    switch (how) {
    case CLONE_DEFAULT:
        *pSpecials = m_stickySpecials;
        *pStickySpecials = m_stickySpecials;
        *addnulls = false;
        break;
    case CLONE_WITHOUT_SPECIALS:
        *pSpecials = nospecials;
        *pStickySpecials = nospecials;
        *addnulls = false;
        break;
    case CLONE_WITHOUT_SPECIALS_ADD_NULLS:
        *pSpecials = nospecials;
        *pStickySpecials = nospecials;
        *addnulls = true;
        break;
    case CLONE_WITH_SPECIALS:
        *pSpecials = m_enableSpecials;
        *pStickySpecials = m_stickySpecials;
        *addnulls = false;
        break;
    default:
        NZ_VERIFY(false);
    }
}

#endif // TABLESHAPE_INCLUDED
