// udxbase.h
// Base class for User Defined Functions and Aggregates
//
// (C) Copyright IBM Corp. 2007, 2011  All Rights Reserved.

#ifndef UDX_BASE_INCLUDED
#define UDX_BASE_INCLUDED

#ifdef LOG_DEBUG
#undef LOG_DEBUG
#endif

// These macros are what should be used to return from evaluate or finalResult
#define NZ_UDX_RETURN_NULL() do { setReturnNull(true); return (void*)NULL; } while (0)

#define NZ_UDX_RETURN_STRING(x) return stringReturn(x)
#define NZ_UDX_RETURN_BOOL(x) return boolReturn(x)
#define NZ_UDX_RETURN_DATE(x) return dateReturn(x)
#define NZ_UDX_RETURN_TIME(x) return timeReturn(x)
#define NZ_UDX_RETURN_TIMETZ(x) return timetzReturn(x)
#define NZ_UDX_RETURN_NUMERIC32(x) return numeric32Return(x)
#define NZ_UDX_RETURN_NUMERIC64(x) return numeric64Return(x)
#define NZ_UDX_RETURN_NUMERIC128(x) return numeric128Return(x)
#define NZ_UDX_RETURN_FLOAT(x) return floatReturn(x)
#define NZ_UDX_RETURN_DOUBLE(x)  return doubleReturn(x)
#define NZ_UDX_RETURN_INTERVAL(x) return intervalReturn(x)
#define NZ_UDX_RETURN_INT64(x) return int64Return(x)
#define NZ_UDX_RETURN_INT32(x) return int32Return(x)
#define NZ_UDX_RETURN_INT16(x) return int16Return(x)
#define NZ_UDX_RETURN_INT8(x) return int8Return(x)
#define NZ_UDX_RETURN_TIMESTAMP(x) return timestampReturn(x)

#define CHECK_INFO_AVAILABLE() if (!m_pInfo) \
          throwError(NZ_ERROR_ILLVAL, "UDF/UDA not yet initialized", 0)

#define CHECK_FUNC_RANGE(n) if (n < 0 || n >= m_pInfo->numArgs) \
		  throwError(NZ_ERROR_ILLVAL, "Argument number out of range", n)



namespace nz {
namespace udx {


class FuncCallInfo
{
 public:
  FuncDatum	pReturnInfo;
  bool*   	pReturnNull;
  int		returnType;
  int     	numArgs;
  int		argTypes[NZ_FUNC_MAX_ARGS];
  FuncDatum	args[NZ_FUNC_MAX_ARGS];
  bool		argNulls[NZ_FUNC_MAX_ARGS];
  bool      argConsts[NZ_FUNC_MAX_ARGS];
};


class AggCallInfo : public FuncCallInfo
{
 public:
  int       numStateVars;
  int		stateVarTypes[NZ_FUNC_MAX_ARGS];
  FuncDatum	stateVars[NZ_FUNC_MAX_ARGS];
  bool*		pStateVarNulls[NZ_FUNC_MAX_ARGS];
};


class UdxBase
{
 public:
  UdxBase()
    {
      m_pInfo = NULL;
      m_nLogMask = LOG_NONE;
      m_memory = 0;
    }
  virtual ~UdxBase() {} 
  enum LogValue
    {
      LOG_NONE=0,
      LOG_TRACE=1,
      LOG_DEBUG=2
    };
  
  
  enum DataType
    {
      UDX_UNKNOWN=-1,
      UDX_FIXED,
      UDX_VARIABLE,
      UDX_NATIONAL_FIXED,
      UDX_NATIONAL_VARIABLE,
      UDX_BOOL,
      UDX_DATE,
      UDX_TIME,
      UDX_TIMETZ,
      UDX_NUMERIC32,
      UDX_NUMERIC64,
      UDX_NUMERIC128,
      UDX_FLOAT,
      UDX_DOUBLE,
      UDX_INTERVAL,
      UDX_INT8,
      UDX_INT16,
      UDX_INT32,
      UDX_INT64,
      UDX_TIMESTAMP,
      UDX_GEOMETRY,
      UDX_VARBINARY,
      UDX_NUMERIC=UDX_NUMERIC128 // Used for calculateSize only
    };
  
 private:
  int8 m_nLogMask;
  int m_memory;
 public:
  FuncCallInfo *m_pInfo;

  
  // Return methods
  ReturnValue numeric32Return(Numeric32Val* ret) const;
  ReturnValue numeric64Return(Numeric64Val* ret) const;
  ReturnValue numeric128Return(Numeric128Val* ret) const;
  ReturnValue stringReturn(StringReturn* ret) const;
  ReturnValue timetzReturn(TimeTzReturn* ret) const;
  ReturnValue intervalReturn(IntervalReturn* ret) const;
  ReturnValue boolReturn(bool ret) const;
  ReturnValue dateReturn(int32 ret) const;
  ReturnValue timeReturn(int64 ret) const;
  ReturnValue timestampReturn(int64 ret) const;
  ReturnValue int8Return(int8 ret) const;
  ReturnValue int16Return(int16 ret) const;
  ReturnValue int32Return(int32 ret) const;
  ReturnValue int64Return(int64 ret) const;
  ReturnValue floatReturn(float ret) const;
  ReturnValue doubleReturn(double ret) const;

  void setReturnNull(bool val) const;
  int returnType() const;

  Numeric32Val* numeric32ReturnInfo() const;
  Numeric64Val* numeric64ReturnInfo() const;
  Numeric128Val* numeric128ReturnInfo() const;
  StringReturn* stringReturnInfo() const;
  TimeTzReturn* timetzReturnInfo() const;
  IntervalReturn* intervalReturnInfo() const;

  // Arguments methods
  int numArgs() const;

  bool isArgConst(int n) const;
  int argType(int n) const;
  bool isArgNull(int n) const;

  Numeric32Val* numeric32Arg(int n) const;
  Numeric64Val* numeric64Arg(int n) const;
  Numeric128Val* numeric128Arg(int n) const;
  StringArg* stringArg(int n) const;
  TimeTzADT* timetzArg(int n) const;
  struct Interval* intervalArg(int n) const;
  bool boolArg(int n) const;
  int32 dateArg(int n) const;
  int64 timeArg(int n) const;
  int64 timestampArg(int n) const;
  int8 int8Arg(int n) const;
  int16 int16Arg(int n) const;
  int32 int32Arg(int n) const;
  int64 int64Arg(int n) const;
  float floatArg(int n) const;
  double doubleArg(int n) const;

  

  // Log methods
  void setLogMask(int8 val);

  void logMsg(int8 flags, const char* fmt, ...) const ;

  // memory methods
  void setMemory(int memory);

  int getMemory() const;

};

inline void UdxBase::setMemory(int memory)
{
  m_memory = memory;
}

inline int UdxBase::getMemory() const
{
  return m_memory;
}

inline bool checkInterrupt() 
{
  return checkPlanInterrupt();
}

inline uint32 checkStack() 
{
#if FOR_SPU && __DCC__
  // This will have the whole SPU 
  // error out if we are within ~ 1K - 1.2K of stack
  // limit
  uint32 nRemaining = TCT_Check_Stack();

  // if we get here, we are fine, now we try a softer
  // limit:
  if (nRemaining < 4096)
    {
      // Make sure we use the throwError which does not allocate a 1k
      // string on the stack.
      throwError(NZ_ERROR_ILLVAL, "Stack overflow", (const char*)NULL);
    }
  return nRemaining;
#else
  // We are on the host, return dummy value
  return 16*1024;
#endif
}

inline void UdxBase::setLogMask(int8 val)
{
  m_nLogMask = val;
}

inline void UdxBase::logMsg(int8 flags, const char* fmt, ...) const
{
#ifdef GENCODE
  
  if ((m_nLogMask & flags) != 0)
    {
      va_list ap;
      va_start(ap, fmt);
      logvprint(fmt, ap);
      va_end(ap);
    }
#endif
}

  

inline int UdxBase::numArgs() const
{
    CHECK_INFO_AVAILABLE();
  return m_pInfo->numArgs;
}
inline int UdxBase::returnType() const
{
    CHECK_INFO_AVAILABLE();
  return m_pInfo->returnType;
}
inline bool UdxBase::isArgConst(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  
  return m_pInfo->argConsts[n];
}
inline int UdxBase::argType(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  return m_pInfo->argTypes[n];
}
inline bool UdxBase::isArgNull(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  return m_pInfo->argNulls[n];
}
inline ReturnValue UdxBase::numeric32Return(Numeric32Val* ret) const
{
  if (m_pInfo->returnType != UDX_NUMERIC32)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric32 return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
	
inline ReturnValue UdxBase::numeric64Return(Numeric64Val* ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_NUMERIC64)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric64 return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxBase::numeric128Return(Numeric128Val* ret) const
{
  if (m_pInfo->returnType != UDX_NUMERIC128)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric128 return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxBase::stringReturn(StringReturn* ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_FIXED && m_pInfo->returnType != UDX_VARIABLE &&
      m_pInfo->returnType != UDX_NATIONAL_FIXED && m_pInfo->returnType != UDX_NATIONAL_VARIABLE &&
      m_pInfo->returnType != UDX_GEOMETRY && m_pInfo->returnType != UDX_VARBINARY)
    throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxBase::timetzReturn(TimeTzReturn* ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_TIMETZ)
    throwError(NZ_ERROR_ILLVAL, "Expected timetz return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxBase::intervalReturn(IntervalReturn* ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_INTERVAL)
    throwError(NZ_ERROR_ILLVAL, "Expected interval return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxBase::boolReturn(bool ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_BOOL)
    throwError(NZ_ERROR_ILLVAL, "Expected bool return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)(int8)ret;
}
inline ReturnValue UdxBase::dateReturn(int32 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_DATE)
    throwError(NZ_ERROR_ILLVAL, "Expected date return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxBase::timeReturn(int64 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_TIME)
    throwError(NZ_ERROR_ILLVAL, "Expected time return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxBase::timestampReturn(int64 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_TIMESTAMP)
    throwError(NZ_ERROR_ILLVAL, "Expected timestamp return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxBase::int8Return(int8 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_INT8)
    throwError(NZ_ERROR_ILLVAL, "Expected int8 return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxBase::int16Return(int16 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_INT16)
    throwError(NZ_ERROR_ILLVAL, "Expected int16 return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxBase::int32Return(int32 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_INT32)
    throwError(NZ_ERROR_ILLVAL, "Expected int32 return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxBase::int64Return(int64 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_INT64)
    throwError(NZ_ERROR_ILLVAL, "Expected int64 return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxBase::floatReturn(float ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_FLOAT)
    throwError(NZ_ERROR_ILLVAL, "Expected float return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxBase::doubleReturn(double ret) const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_DOUBLE)
    throwError(NZ_ERROR_ILLVAL, "Expected double return type", m_pInfo->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline void UdxBase::setReturnNull(bool val) const
{
    CHECK_INFO_AVAILABLE();
  *m_pInfo->pReturnNull = val;
}
inline Numeric32Val* UdxBase::numeric32ReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_NUMERIC32)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric32 return type", m_pInfo->returnType);
  if (!m_pInfo->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric32 return structure", 0);
  return (Numeric32Val*)m_pInfo->pReturnInfo;
}
inline Numeric64Val* UdxBase::numeric64ReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_NUMERIC64)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric64 return type", m_pInfo->returnType);
  if (!m_pInfo->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric64 return structure", 0);
  return (Numeric64Val*)m_pInfo->pReturnInfo;
}
inline Numeric128Val* UdxBase::numeric128ReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_NUMERIC128)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric128 return type", m_pInfo->returnType);
  if (!m_pInfo->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric128 return structure", 0);
  return (Numeric128Val*)m_pInfo->pReturnInfo;
}
inline StringReturn* UdxBase::stringReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_FIXED && m_pInfo->returnType != UDX_VARIABLE && 
      m_pInfo->returnType != UDX_NATIONAL_FIXED && m_pInfo->returnType != UDX_NATIONAL_VARIABLE &&
      m_pInfo->returnType != UDX_GEOMETRY && m_pInfo->returnType != UDX_VARBINARY)
    throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary return type", m_pInfo->returnType);
  if (!m_pInfo->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary return structure", 0);
  return (StringReturn*)m_pInfo->pReturnInfo;
}
inline TimeTzReturn* UdxBase::timetzReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_TIMETZ)
    throwError(NZ_ERROR_ILLVAL, "Expected timetz return type", m_pInfo->returnType);
  if (!m_pInfo->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected timetz return structure", 0);
  return (TimeTzReturn*)m_pInfo->pReturnInfo;
}
inline IntervalReturn* UdxBase::intervalReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (m_pInfo->returnType != UDX_INTERVAL)
    throwError(NZ_ERROR_ILLVAL, "Expected interval return type", m_pInfo->returnType);
  if (!m_pInfo->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected interval return structure", 0);
  return (IntervalReturn*)m_pInfo->pReturnInfo;
}
inline StringArg* UdxBase::stringArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_FIXED && nType != UDX_VARIABLE && nType != UDX_NATIONAL_FIXED &&
      nType != UDX_NATIONAL_VARIABLE && nType != UDX_GEOMETRY && nType != UDX_VARBINARY)
    throwError(NZ_ERROR_ILLVAL, "Expected string,  st_geometry, or varbinary argument", nType);
  return ((StringArg*)m_pInfo->args[n]);
}
inline Numeric32Val* UdxBase::numeric32Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_NUMERIC32)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric32 argument", nType);
  return ((Numeric32Val*)m_pInfo->args[n]);
}
inline Numeric64Val* UdxBase::numeric64Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_NUMERIC64)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric64 argument", nType);
  return ((Numeric64Val*)m_pInfo->args[n]);
}
inline Numeric128Val* UdxBase::numeric128Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_NUMERIC128)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric128 argument", nType);
  return ((Numeric128Val*)m_pInfo->args[n]);
}

inline TimeTzADT* UdxBase::timetzArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_TIMETZ)
    throwError(NZ_ERROR_ILLVAL, "Expected timetz argument", nType);
  return ((TimeTzADT*)m_pInfo->args[n]);
}

inline struct Interval* UdxBase::intervalArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INTERVAL)
    throwError(NZ_ERROR_ILLVAL, "Expected interval argument", nType);
  return ((struct Interval*)m_pInfo->args[n]);
}
inline float UdxBase::floatArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_FLOAT)
    throwError(NZ_ERROR_ILLVAL, "Expected float argument", nType);
  return *((float*)m_pInfo->args[n]);
}

inline double UdxBase::doubleArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_DOUBLE)
    throwError(NZ_ERROR_ILLVAL, "Expected double argument", nType);
  return *((double*)m_pInfo->args[n]);
}
inline int8 UdxBase::int8Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INT8)
    throwError(NZ_ERROR_ILLVAL, "Expected int8 argument", nType);
  return *((int8*)m_pInfo->args[n]);
}
inline int16 UdxBase::int16Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INT16)
    throwError(NZ_ERROR_ILLVAL, "Expected int16 argument", nType);
  return *((int16*)m_pInfo->args[n]);
}
inline int32 UdxBase::int32Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INT32)
    throwError(NZ_ERROR_ILLVAL, "Expected int32 argument", nType);
  return *((int32*)m_pInfo->args[n]);
} 
inline int64 UdxBase::int64Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INT64)
    throwError(NZ_ERROR_ILLVAL, "Expected int64 argument", nType);
  return *((int64*)m_pInfo->args[n]);
}
inline bool UdxBase::boolArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_BOOL)
    throwError(NZ_ERROR_ILLVAL, "Expected boolean argument", nType);
  return *((int8*)m_pInfo->args[n]);
}
inline int32 UdxBase::dateArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_DATE)
    throwError(NZ_ERROR_ILLVAL, "Expected date argument", nType);
  return *((int32*)m_pInfo->args[n]);
}
inline int64 UdxBase::timeArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_TIME)
    throwError(NZ_ERROR_ILLVAL, "Expected time argument", nType);
  return *((int64*)m_pInfo->args[n]);
}
inline int64 UdxBase::timestampArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_TIMESTAMP)
    throwError(NZ_ERROR_ILLVAL, "Expected timestamp argument", nType);
  return *((int64*)m_pInfo->args[n]);
}

} // end namespace udx
} // end namespace nz

#endif //  #ifndef UDX_BASE_INCLUDED
