// udxbasev2.h
// Base class for Ver 2 User Defined Functions and Aggregates
//
// (C) Copyright IBM Corp. 2009, 2011  All Rights Reserved.

#ifndef UDX_BASEV2_INCLUDED
#define UDX_BASEV2_INCLUDED

// These macros are what should be used to return from evaluate or finalResult
#define NZ_UDX_RETURN_NULL() do { setReturnNull(true); return (void*)NULL; } while (0)

#define NZ_UDX_RETURN_STRING(x) return stringReturn(x)
#define NZ_UDX_RETURN_BOOL(x) return boolReturn(x)
#define NZ_UDX_RETURN_DATE(x) return dateReturn(x)
#define NZ_UDX_RETURN_TIME(x) return timeReturn(x)
#define NZ_UDX_RETURN_TIMETZ(x) return timetzReturn(x)
#define NZ_UDX_RETURN_NUMERIC32(x) return numeric32Return(x)
#define NZ_UDX_RETURN_NUMERIC64(x) return numeric64Return(x)
#define NZ_UDX_RETURN_NUMERIC128(x) return numeric128Return(x)
#define NZ_UDX_RETURN_FLOAT(x) return floatReturn(x)
#define NZ_UDX_RETURN_DOUBLE(x)  return doubleReturn(x)
#define NZ_UDX_RETURN_INTERVAL(x) return intervalReturn(x)
#define NZ_UDX_RETURN_INT64(x) return int64Return(x)
#define NZ_UDX_RETURN_INT32(x) return int32Return(x)
#define NZ_UDX_RETURN_INT16(x) return int16Return(x)
#define NZ_UDX_RETURN_INT8(x) return int8Return(x)
#define NZ_UDX_RETURN_TIMESTAMP(x) return timestampReturn(x)

#define CHECK_FUNC_RANGE(n) if (n < 0 || n >= m_pInfo->numArgs) \
		  throwError(NZ_ERROR_ILLVAL, "Argument number out of range", n)
#define CHECK_INFO_AVAILABLE() if (!m_pInfo) \
          throwError(NZ_ERROR_ILLVAL, "UDF/UDA not yet initialized", 0)

#define CHECK_SIZER_TYPMOD(n, typmod) if (typmod == 0xFFFFFFFFFFFFFFFFLL) \
                  throwError(NZ_ERROR_ILLVAL, "Sizer argument invalid", n)

#ifndef MAX_TABLE_COLUMNS
#define MAX_TABLE_COLUMNS 4000
#endif

#define NZ_TABLE_FUNC_MAX_RETURN_COLUMNS MAX_TABLE_COLUMNS 


namespace nz {
namespace udx_ver2 {


using nz::udx::FuncDatum ;
using nz::udx::ReturnValue ;
using nz::udx::Numeric32Val ;
using nz::udx::Numeric64Val ;
using nz::udx::Numeric128Val ;
using nz::udx::StringReturn ;
using nz::udx::StringArg ;
using nz::udx::IntervalReturn ;
using nz::udx::TimeTzReturn ;
using nz::udx::CPad ;

namespace utf8helpers = nz::udx::utf8helpers;
namespace dthelpers = nz::udx::dthelpers;

using nz::udx::getPad;

class CallInfo
{
public:
    int     	numArgs;
    int		argTypes[NZ_FUNC_MAX_ARGS];
    FuncDatum	args[NZ_FUNC_MAX_ARGS];
    bool		argNulls[NZ_FUNC_MAX_ARGS];
    bool      argConsts[NZ_FUNC_MAX_ARGS];

};

class FuncCallInfo : public CallInfo
{
 public:
  FuncDatum	pReturnInfo;
  bool*   	pReturnNull;
  int		returnType;
  
};


class AggCallInfo : public FuncCallInfo
{
 public:
  int       numStateVars;
  int		stateVarTypes[NZ_FUNC_MAX_ARGS];
  FuncDatum	stateVars[NZ_FUNC_MAX_ARGS];
  bool*		pStateVarNulls[NZ_FUNC_MAX_ARGS];
};


class RowFuncCallInfo : public CallInfo
{
public:

    int        numReturnColumns;
    int        returnColumnTypes[NZ_TABLE_FUNC_MAX_RETURN_COLUMNS];
    FuncDatum  returnColumns[NZ_TABLE_FUNC_MAX_RETURN_COLUMNS];
    bool       returnColumnNulls[NZ_TABLE_FUNC_MAX_RETURN_COLUMNS];

};


class RowAggCallInfo : public RowFuncCallInfo
{
 public:
  int           numStateVars;
  int           stateVarTypes[NZ_FUNC_MAX_ARGS];
  FuncDatum     stateVars[NZ_FUNC_MAX_ARGS];
  bool*         pStateVarNulls[NZ_FUNC_MAX_ARGS];
};

class TableFuncCallInfo : public RowFuncCallInfo
{
public:
    

};

class UdxInit
{
public:
    UdxInit(UdxEnvironment *pEnviron, UdxParameter *pParam,
            int nLogMask, int nMemory) {
        m_pEnvironment = pEnviron;
        m_nLogMask = nLogMask;
        m_nMemory = nMemory;
        m_pParameter = pParam;
    }
    ~UdxInit() {
        if (m_pEnvironment)
            delete m_pEnvironment;
        if (m_pParameter)
            delete m_pParameter;
    }

    int getLogMask() {
        return m_nLogMask;
    }
    int getMemory() {
        return m_nMemory;
    }
    UdxEnvironment* getEnvironment() {
        return m_pEnvironment;
    }
    UdxParameter* getParameter() {
        return m_pParameter;
    }

protected:
    UdxParameter   *m_pParameter;
    UdxEnvironment *m_pEnvironment;
    int m_nLogMask;
    int m_nMemory;
};

class UdxBase
{
 public:
    UdxBase(UdxInit *pInit)
    {
      m_pInfo = NULL;
      m_nLogMask = pInit->getLogMask();
      m_memory = pInit->getMemory();
      m_pEnviron = pInit->getEnvironment();
      m_pParameter = pInit->getParameter();
      m_pInit = pInit;
    }
  virtual ~UdxBase() {
      if (m_pInit)
          delete m_pInit;
  } 
  enum LogValue
    {
      LOG_NONE=0,
      LOG_TRACE=1,
      LOG_DEBUG=2
    };
  
  
  enum DataType
    {
      UDX_UNKNOWN=-1,
      UDX_FIXED,
      UDX_VARIABLE,
      UDX_NATIONAL_FIXED,
      UDX_NATIONAL_VARIABLE,
      UDX_BOOL,
      UDX_DATE,
      UDX_TIME,
      UDX_TIMETZ,
      UDX_NUMERIC32,
      UDX_NUMERIC64,
      UDX_NUMERIC128,
      UDX_FLOAT,
      UDX_DOUBLE,
      UDX_INTERVAL,
      UDX_INT8,
      UDX_INT16,
      UDX_INT32,
      UDX_INT64,
      UDX_TIMESTAMP,
      UDX_GEOMETRY,
      UDX_VARBINARY,
      UDX_NUMERIC=UDX_NUMERIC128 // Used for calculateSize only
    };
  
  
  // Log methods
  void setLogMask(int8 val);
  bool isLoggingEnabled(int8 val);
  int8 getLogMask();
	
  void logMsg(int8 flags, const char* fmt, ...) const ;

  // memory methods
  void setMemory(int memory);

  int getMemory() const;

  // Arguments methods
  int numArgs() const;

  bool isArgConst(int n) const;
  int argType(int n) const;
  bool isArgNull(int n) const;

  Numeric32Val* numeric32Arg(int n) const;
  Numeric64Val* numeric64Arg(int n) const;
  Numeric128Val* numeric128Arg(int n) const;
  StringArg* stringArg(int n) const;
  TimeTzADT* timetzArg(int n) const;
  struct Interval* intervalArg(int n) const;
  bool boolArg(int n) const;
  int32 dateArg(int n) const;
  int64 timeArg(int n) const;
  int64 timestampArg(int n) const;
  int8 int8Arg(int n) const;
  int16 int16Arg(int n) const;
  int32 int32Arg(int n) const;
  int64 int64Arg(int n) const;
  float floatArg(int n) const;
  double doubleArg(int n) const;

  UdxEnvironment* getEnvironment() {
      return m_pEnviron;
  }
  UdxParameter* getParameter() {
        return m_pParameter;
  }

  // This will be used to provide information
  // about the input from the optimizer and to
  // give the UDX writer a chance to give hints
  // to the optimizer about how to plan the query.
  // You will need to consult the release documentation 
  // in order to see what information is provided and
  // can be set for a particular UDX type.
  // Do nothing by default. Return true if 
  // information is set in info for the optimizer.
  // For aggregates the argument will actually be
  // a UdxAggregateOptimizerInfo object.
  virtual bool provideOptimizerInfo(UdxOptimizerInfo *info){

      return false;
  }

  // This is relevant for UDTF, UDRA and UDRF.
  // For UDRA it can be used for either return, state
  // or either and will actually be a UdxAggregateShaper
  // object.
  // This is present in the base class in case we ever
  // want to support shapers for scalar UDFs and UDAs
  // although there are currently no plans to do so
  virtual void calculateShape(UdxOutputShaper *shaper) {
      throwUdxException("Must inplement calculateShape for a UDX returning any or using state any");
  }

 private:
  int8 m_nLogMask;
  int m_memory;
 public:
  CallInfo *m_pInfo;
  UdxEnvironment *m_pEnviron;
  UdxParameter *m_pParameter;
  UdxInit *m_pInit;
};



class UdxScalarBase : public UdxBase
{
public:
    UdxScalarBase(UdxInit *pInit) : UdxBase(pInit) {}
  
  // Return methods
  ReturnValue numeric32Return(Numeric32Val* ret) const;
  ReturnValue numeric64Return(Numeric64Val* ret) const;
  ReturnValue numeric128Return(Numeric128Val* ret) const;
  ReturnValue stringReturn(StringReturn* ret) const;
  ReturnValue timetzReturn(TimeTzReturn* ret) const;
  ReturnValue intervalReturn(IntervalReturn* ret) const;
  ReturnValue boolReturn(bool ret) const;
  ReturnValue dateReturn(int32 ret) const;
  ReturnValue timeReturn(int64 ret) const;
  ReturnValue timestampReturn(int64 ret) const;
  ReturnValue int8Return(int8 ret) const;
  ReturnValue int16Return(int16 ret) const;
  ReturnValue int32Return(int32 ret) const;
  ReturnValue int64Return(int64 ret) const;
  ReturnValue floatReturn(float ret) const;
  ReturnValue doubleReturn(double ret) const;

  void setReturnNull(bool val) const;
  int returnType() const;

  Numeric32Val* numeric32ReturnInfo() const;
  Numeric64Val* numeric64ReturnInfo() const;
  Numeric128Val* numeric128ReturnInfo() const;
  StringReturn* stringReturnInfo() const;
  TimeTzReturn* timetzReturnInfo() const;
  IntervalReturn* intervalReturnInfo() const;

  // The following will be used to support generic sized
  // string and numeric return values.

  // Can return UDX_NUMERIC, UDX_FIXED, UDX_VARIABLE,
  // UDX_NATIONAL_FIXED or UDX_NATIONAL_VARIABLE,
  int sizerReturnType() const;

  // Number of arguments the function is called with
  int numSizerArgs() const;

  // The type of the specified argument. (Can be any
  // of the DataType enums (Except UDX_NUMERIC). Will throw
  // an exception if n is out of range. 
  int sizerArgType(int n) const;

  // Returns the size of the specified argument. Will throw
  // exceptions if the n is out of range, if the specified
  // argument is not a string type or the specified argument
  // does not have a size. (shouldn't happen) This size is in characters,
  // not bytes. (this matters for the national strings)
  int sizerStringArgSize(int n) const;

  // Returns the precision of the specified argument. Will throw
  // exceptions if the n is out of range, if the specified
  // argument is not a numeric type or the specified argument
  // does not have a size. (shouldn't happen)
  int sizerNumericArgPrecision(int n) const;

  // Returns the scale of the specified argument. Will throw
  // exceptions if the n is out of range, if the specified
  // argument is not a numeric type or the specified argument
  // does not have a size. (shouldn't happen)
  int sizerNumericArgScale(int n) const;

  // Builds a return value for calculateSize given the
  // string length desired. Will throw an exception if
  // the return type is not a string. This length is in characters,
  // not bytes. (this matters for the national strings)
  uint64 sizerStringSizeValue(int len) const;

  // Builds a return value for calculateSize given the
  // precision and scale desired. Will throw an exception
  // if the return type is not a numeric.
  uint64 sizerNumericSizeValue(int prec, int scale) const;

  // In order to support certain methods like round(val, scale)
  // we have a mechanism for passing constant arguments to the
  // sizer so long as they are of int32 type.
  // This method will return true if the specified argument is
  // a constant integer. Will throw and exception if the n is
  // out of range or if the specified argument is not an int32
  // If the constant specified is -1, isSizerArgConstants
  // will return false.
  bool isSizerArgConstant(int n) const;

  // This method will return the specified constant int32 argument
  // Will throw and exception if the n is out of range, the specified
  // argument is not an int32 or the specified argument is
  // not constant.
  int32 sizerGetConstantArg(int n) const;
    
  // This method must be overriden to provide sizing
  // capabilities for numerics and strings. This must
  // be done in conjunction with the DDL.
  virtual uint64 calculateSize() const
  {
      return 0xFFFFFFFFFFFFFFFFLL;    
  }
};


inline void UdxBase::setMemory(int memory)
{
  m_memory = memory;
}

inline int UdxBase::getMemory() const
{
  return m_memory;
}

inline bool checkInterrupt() 
{
  return checkPlanInterrupt();
}

inline void UdxBase::setLogMask(int8 val)
{
  m_nLogMask = val;
}

inline int8 UdxBase::getLogMask()
{
  return m_nLogMask;
}

inline void UdxBase::logMsg(int8 flags, const char* fmt, ...) const
{
#ifdef GENCODE
  
  if ((m_nLogMask & flags) != 0)
    {
      va_list ap;
      va_start(ap, fmt);
      logvprint(fmt, ap);
      va_end(ap);
    }
#endif
}

inline bool UdxBase::isLoggingEnabled(int8 val)
{
    return ((m_nLogMask & val) != 0);
}

inline int UdxBase::numArgs() const
{
    CHECK_INFO_AVAILABLE();
  return m_pInfo->numArgs;
}

inline bool UdxBase::isArgConst(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  
  return m_pInfo->argConsts[n];
}
inline int UdxBase::argType(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  return m_pInfo->argTypes[n];
}
inline bool UdxBase::isArgNull(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  return m_pInfo->argNulls[n];
}

inline StringArg* UdxBase::stringArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_FIXED && nType != UDX_VARIABLE && nType != UDX_NATIONAL_FIXED &&
      nType != UDX_NATIONAL_VARIABLE && nType != UDX_GEOMETRY && nType != UDX_VARBINARY)
    throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary argument", nType);
  return ((StringArg*)m_pInfo->args[n]);
}
inline Numeric32Val* UdxBase::numeric32Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_NUMERIC32)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric32 argument", nType);
  return ((Numeric32Val*)m_pInfo->args[n]);
}
inline Numeric64Val* UdxBase::numeric64Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_NUMERIC64)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric64 argument", nType);
  return ((Numeric64Val*)m_pInfo->args[n]);
}
inline Numeric128Val* UdxBase::numeric128Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_NUMERIC128)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric128 argument", nType);
  return ((Numeric128Val*)m_pInfo->args[n]);
}

inline TimeTzADT* UdxBase::timetzArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_TIMETZ)
    throwError(NZ_ERROR_ILLVAL, "Expected timetz argument", nType);
  return ((TimeTzADT*)m_pInfo->args[n]);
}

inline struct Interval* UdxBase::intervalArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INTERVAL)
    throwError(NZ_ERROR_ILLVAL, "Expected interval argument", nType);
  return ((struct Interval*)m_pInfo->args[n]);
}
inline float UdxBase::floatArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_FLOAT)
    throwError(NZ_ERROR_ILLVAL, "Expected float argument", nType);
  return *((float*)m_pInfo->args[n]);
}

inline double UdxBase::doubleArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_DOUBLE)
    throwError(NZ_ERROR_ILLVAL, "Expected double argument", nType);
  return *((double*)m_pInfo->args[n]);
}
inline int8 UdxBase::int8Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INT8)
    throwError(NZ_ERROR_ILLVAL, "Expected int8 argument", nType);
  return *((int8*)m_pInfo->args[n]);
}
inline int16 UdxBase::int16Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INT16)
    throwError(NZ_ERROR_ILLVAL, "Expected int16 argument", nType);
  return *((int16*)m_pInfo->args[n]);
}
inline int32 UdxBase::int32Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INT32)
    throwError(NZ_ERROR_ILLVAL, "Expected int32 argument", nType);
  return *((int32*)m_pInfo->args[n]);
} 
inline int64 UdxBase::int64Arg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_INT64)
    throwError(NZ_ERROR_ILLVAL, "Expected int64 argument", nType);
  return *((int64*)m_pInfo->args[n]);
}
inline bool UdxBase::boolArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_BOOL)
    throwError(NZ_ERROR_ILLVAL, "Expected boolean argument", nType);
  return *((int8*)m_pInfo->args[n]);
}
inline int32 UdxBase::dateArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_DATE)
    throwError(NZ_ERROR_ILLVAL, "Expected date argument", nType);
  return *((int32*)m_pInfo->args[n]);
}
inline int64 UdxBase::timeArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_TIME)
    throwError(NZ_ERROR_ILLVAL, "Expected time argument", nType);
  return *((int64*)m_pInfo->args[n]);
}
inline int64 UdxBase::timestampArg(int n) const
{
    CHECK_INFO_AVAILABLE();
  CHECK_FUNC_RANGE(n);
  int nType = argType(n);
  if (nType != UDX_TIMESTAMP)
    throwError(NZ_ERROR_ILLVAL, "Expected timestamp argument", nType);
  return *((int64*)m_pInfo->args[n]);
}

inline int UdxScalarBase::returnType() const
{
    CHECK_INFO_AVAILABLE();
  return ((FuncCallInfo*)m_pInfo)->returnType;
}

inline ReturnValue UdxScalarBase::numeric32Return(Numeric32Val* ret) const
{
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_NUMERIC32)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric32 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
	
inline ReturnValue UdxScalarBase::numeric64Return(Numeric64Val* ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_NUMERIC64)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric64 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxScalarBase::numeric128Return(Numeric128Val* ret) const
{
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_NUMERIC128)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric128 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxScalarBase::stringReturn(StringReturn* ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_FIXED && ((FuncCallInfo*)m_pInfo)->returnType != UDX_VARIABLE &&
      ((FuncCallInfo*)m_pInfo)->returnType != UDX_NATIONAL_FIXED && ((FuncCallInfo*)m_pInfo)->returnType != UDX_NATIONAL_VARIABLE &&
      ((FuncCallInfo*)m_pInfo)->returnType != UDX_GEOMETRY && ((FuncCallInfo*)m_pInfo)->returnType != UDX_VARBINARY)
    throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxScalarBase::timetzReturn(TimeTzReturn* ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_TIMETZ)
    throwError(NZ_ERROR_ILLVAL, "Expected timetz return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxScalarBase::intervalReturn(IntervalReturn* ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_INTERVAL)
    throwError(NZ_ERROR_ILLVAL, "Expected interval return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)(void*)ret;
}
inline ReturnValue UdxScalarBase::boolReturn(bool ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_BOOL)
    throwError(NZ_ERROR_ILLVAL, "Expected bool return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)(int8)ret;
}
inline ReturnValue UdxScalarBase::dateReturn(int32 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_DATE)
    throwError(NZ_ERROR_ILLVAL, "Expected date return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxScalarBase::timeReturn(int64 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_TIME)
    throwError(NZ_ERROR_ILLVAL, "Expected time return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxScalarBase::timestampReturn(int64 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_TIMESTAMP)
    throwError(NZ_ERROR_ILLVAL, "Expected timestamp return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxScalarBase::int8Return(int8 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_INT8)
    throwError(NZ_ERROR_ILLVAL, "Expected int8 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxScalarBase::int16Return(int16 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_INT16)
    throwError(NZ_ERROR_ILLVAL, "Expected int16 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxScalarBase::int32Return(int32 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_INT32)
    throwError(NZ_ERROR_ILLVAL, "Expected int32 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxScalarBase::int64Return(int64 ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_INT64)
    throwError(NZ_ERROR_ILLVAL, "Expected int64 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxScalarBase::floatReturn(float ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_FLOAT)
    throwError(NZ_ERROR_ILLVAL, "Expected float return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline ReturnValue UdxScalarBase::doubleReturn(double ret) const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_DOUBLE)
    throwError(NZ_ERROR_ILLVAL, "Expected double return type", ((FuncCallInfo*)m_pInfo)->returnType);
  setReturnNull(false);
  return (ReturnValue)ret;
}
inline void UdxScalarBase::setReturnNull(bool val) const
{
    CHECK_INFO_AVAILABLE();
  *((FuncCallInfo*)m_pInfo)->pReturnNull = val;
}
inline Numeric32Val* UdxScalarBase::numeric32ReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_NUMERIC32)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric32 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  if (!((FuncCallInfo*)m_pInfo)->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric32 return structure", 0);
  return (Numeric32Val*)((FuncCallInfo*)m_pInfo)->pReturnInfo;
}
inline Numeric64Val* UdxScalarBase::numeric64ReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_NUMERIC64)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric64 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  if (!((FuncCallInfo*)m_pInfo)->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric64 return structure", 0);
  return (Numeric64Val*)((FuncCallInfo*)m_pInfo)->pReturnInfo;
}
inline Numeric128Val* UdxScalarBase::numeric128ReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_NUMERIC128)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric128 return type", ((FuncCallInfo*)m_pInfo)->returnType);
  if (!((FuncCallInfo*)m_pInfo)->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected numeric128 return structure", 0);
  return (Numeric128Val*)((FuncCallInfo*)m_pInfo)->pReturnInfo;
}
inline StringReturn* UdxScalarBase::stringReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_FIXED && ((FuncCallInfo*)m_pInfo)->returnType != UDX_VARIABLE && 
      ((FuncCallInfo*)m_pInfo)->returnType != UDX_NATIONAL_FIXED && ((FuncCallInfo*)m_pInfo)->returnType != UDX_NATIONAL_VARIABLE &&
      ((FuncCallInfo*)m_pInfo)->returnType != UDX_GEOMETRY && ((FuncCallInfo*)m_pInfo)->returnType != UDX_VARBINARY)
    throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary return type.", ((FuncCallInfo*)m_pInfo)->returnType);
  if (!((FuncCallInfo*)m_pInfo)->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected string return, geometry structure, or varbinary structure.", 0);
  return (StringReturn*)((FuncCallInfo*)m_pInfo)->pReturnInfo;
}
inline TimeTzReturn* UdxScalarBase::timetzReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_TIMETZ)
    throwError(NZ_ERROR_ILLVAL, "Expected timetz return type", ((FuncCallInfo*)m_pInfo)->returnType);
  if (!((FuncCallInfo*)m_pInfo)->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected timetz return structure", 0);
  return (TimeTzReturn*)((FuncCallInfo*)m_pInfo)->pReturnInfo;
}
inline IntervalReturn* UdxScalarBase::intervalReturnInfo() const
{
    CHECK_INFO_AVAILABLE();
  if (((FuncCallInfo*)m_pInfo)->returnType != UDX_INTERVAL)
    throwError(NZ_ERROR_ILLVAL, "Expected interval return type", ((FuncCallInfo*)m_pInfo)->returnType);
  if (!((FuncCallInfo*)m_pInfo)->pReturnInfo)
    throwError(NZ_ERROR_ILLVAL, "Expected interval return structure", 0);
  return (IntervalReturn*)((FuncCallInfo*)m_pInfo)->pReturnInfo;
}


inline bool UdxScalarBase::isSizerArgConstant(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_SIZER_RANGE(n);
    int nType = sizerArgType(n);
    if (nType != UDX_INT32 )
        throwError(NZ_ERROR_ILLVAL, "Expected int32 argument", nType);
    uint64 arg = (uint64)*(int64*)((FuncCallInfo*)m_pInfo)->args[n];
    int lower = arg >> 32;
    if (lower == -1)
        return false;
    return true;
}
inline int32 UdxScalarBase::sizerGetConstantArg(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_SIZER_RANGE(n);
    int nType = sizerArgType(n);
    if (nType != UDX_INT32)
        throwError(NZ_ERROR_ILLVAL, "Expected int32 argument", nType);
    uint64 arg = (uint64)*(int64*)((FuncCallInfo*)m_pInfo)->args[n];
    int lower = arg >> 32;
    if (lower == -1)
        throwError(NZ_ERROR_ILLVAL, "Not a constant integer argument", 0);
        
    return lower;

}

 
inline int UdxScalarBase::sizerReturnType() const
{
    CHECK_INFO_AVAILABLE();
    return ((FuncCallInfo*)m_pInfo)->returnType;
}
    
inline int UdxScalarBase::numSizerArgs() const
{
    CHECK_INFO_AVAILABLE();
    return m_pInfo->numArgs;
}

inline int UdxScalarBase::sizerArgType(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_SIZER_RANGE(n);
    return m_pInfo->argTypes[n];
}

inline int UdxScalarBase::sizerStringArgSize(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_SIZER_RANGE(n);
         
    int nType = sizerArgType(n);
    if (nType != UDX_FIXED && nType != UDX_VARIABLE && nType != UDX_NATIONAL_FIXED &&
        nType != UDX_NATIONAL_VARIABLE && nType != UDX_GEOMETRY && nType != UDX_VARBINARY)
        throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary argument", nType);
    uint64 typmod = ((uint64)*(int64*)m_pInfo->args[n]) & 0xFFFFFFFF;
    CHECK_SIZER_TYPMOD(n, typmod);
    typmod -= VARHDRSZ;
    return (int32)typmod;
} 

inline int UdxScalarBase::sizerNumericArgPrecision(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_SIZER_RANGE(n);
         
    int nType = sizerArgType(n);
    if (nType != UDX_NUMERIC32 && nType != UDX_NUMERIC64 && nType != UDX_NUMERIC128)
        throwError(NZ_ERROR_ILLVAL, "Expected numeric argument", nType);
    uint64 typmod = ((uint64)*(int64*)m_pInfo->args[n]) & 0xFFFFFFFF;
    CHECK_SIZER_TYPMOD(n, typmod);
    typmod -= VARHDRSZ;
    return (typmod >> 16) & 0x7fff;
}
inline int UdxScalarBase::sizerNumericArgScale(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_SIZER_RANGE(n);
         
    int nType = sizerArgType(n);
    if (nType != UDX_NUMERIC32 && nType != UDX_NUMERIC64 && nType != UDX_NUMERIC128)
        throwError(NZ_ERROR_ILLVAL, "Expected numeric argument", nType);
    uint64 typmod = ((uint64)*(int64*)m_pInfo->args[n]) & 0xFFFFFFFF;
    CHECK_SIZER_TYPMOD(n, typmod);
    typmod -= VARHDRSZ;
    return typmod & 0xffff;
}

inline uint64 UdxScalarBase::sizerStringSizeValue(int len) const
{
    CHECK_INFO_AVAILABLE();
    int nType = ((FuncCallInfo*)m_pInfo)->returnType;
    if (nType != UDX_FIXED && nType != UDX_VARIABLE && nType != UDX_NATIONAL_FIXED &&
        nType != UDX_NATIONAL_VARIABLE && nType != UDX_GEOMETRY && nType != UDX_VARBINARY)
        throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary return type", nType);
         
    return ((uint64)0xFFFFFFFF00000000LL + len + VARHDRSZ);
}

inline uint64 UdxScalarBase::sizerNumericSizeValue(int prec, int scale) const
{
    CHECK_INFO_AVAILABLE();
    int nType = ((FuncCallInfo*)m_pInfo)->returnType;
    if (nType != UDX_NUMERIC)
        throwError(NZ_ERROR_ILLVAL, "Expected numeric return type", nType);
    return ((uint64)0xFFFFFFFF00000000LL + (((prec)&0x7fff)<<16) + (scale) + VARHDRSZ);
}
 

} // end namespace udx_ver2
} // end namespace nz

#endif //  #ifndef UDX_BASEV2_INCLUDED
