// udxenvironment.h
// Environment class for User Defined Functions and Aggregates
//
// (C) Copyright IBM Corp. 2009, 2014  All Rights Reserved.

#ifndef UDX_ENVIRONMENT_INCLUDED
#define UDX_ENVIRONMENT_INCLUDED

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <string>
#include <map>



namespace nz {  
namespace udx_ver2 {

using nz::udx::FuncDatum ;
using nz::udx::ReturnValue ;
using nz::udx::Numeric32Val ;
using nz::udx::Numeric64Val ;
using nz::udx::Numeric128Val ;
using nz::udx::StringReturn ;
using nz::udx::StringArg ;
using nz::udx::IntervalReturn ;
using nz::udx::TimeTzReturn ;


} // end namespace udx_ver2
} // end namespace nz

#include <udxnamevalue.h>
#include <udxcolumninfo.h>
#include <udxshaper.h>
#include <udxoptimizer.h>
#include <udxparameter.h>

namespace nz {  
namespace udx_ver2 {

class UdxEnvironmentEntry : public UdxNameValuePair
{
public:
    UdxEnvironmentEntry(const char* strKey, const char* strValue) :
        UdxNameValuePair(strKey, strValue) {}
    virtual void setValue(const char* strValue) {
        throwUdxException("Can't modify a UdxEnvironmentEntry");
    }
    
};



class UdxEnvironment : public UdxNameValuePairs 
{
public:
    
    virtual void deleteEntry(int idx) {
        throwUdxException("Can't delete entries in UdxEnvironment");
    }
    virtual void modifyEntry(int idx, const char* strNewValue) {
        throwUdxException("Can't modify entries in UdxEnvironment");
    }
    
    virtual const UdxEnvironmentEntry* getEntry(int idx) const {
        return (UdxEnvironmentEntry*)UdxNameValuePairs::getEntry(idx);
    }
    virtual const UdxEnvironmentEntry* getEntry(const char* strKey) const {
        return (UdxEnvironmentEntry*)UdxNameValuePairs::getEntry(strKey);
    }

protected:
    virtual UdxNameValuePair* allocateEntry(const char* strKey, const char* strValue) {
        return new UdxEnvironmentEntry(strKey, strValue);
    }

};


class UdxParameterToEnvironment
{
public:
    UdxParameterToEnvironment(UdxParameter *pParam) {

        m_pParam = pParam;
        m_pEnv = new UdxEnvironment;
    }

    virtual ~UdxParameterToEnvironment() {
        delete m_pEnv;
    }

    int processString(const char *entry, int idx, std::string &accum, const char* context) {
        int len = strlen(entry); 
        bool inquote=false;
        char buf[1024];

        // skip space
        while (idx < len && isspace(entry[idx])) 
            idx++;

        while (idx < len) {
            char c = entry[idx];

            // handle escaped characters
            if (c == '\\') 
            {
                if (idx+1 < len) {
                    char c2 = entry[idx+1];
                    if (c == '"' || c == '\\') {
                        idx++;
                        accum += c2;
                    }
                    else {
                        accum += c;
                    }
                    idx++;
                }
                else {
                    sprintf(buf, "%s: Escape character at end of string", context);
                    throwUdxException(buf);
                }
            }
            else if (c == '"') {
                inquote = !inquote;
                idx++;
            }
            else {
                if (inquote) 
                {
                    accum += c;
                    idx++;
                }
                else {
                    if (isspace(c)) {
                        idx++;
                    }
                    else if (c == '=') {
                        idx++;
                        return idx;
                    }
                    else {
                        accum += c;
                        idx++;
                    }
                }
            }
            
        }
        if (inquote) {
            sprintf(buf, "%s: Unterminated quote at end of string", context);
            throwUdxException(buf);
        }
        if (idx == len) 
            return len;
        
    }
    void parseEntry(const char* entry, const char **key, const char **value) {

        m_key = "";
        m_value = "";
        char buf[1024];
        int idx = 0;
        idx = processString(entry, idx, m_key, "Processing name");
        idx = processString(entry, idx,  m_value, "Processing value");
        int len = strlen(entry);
        // skip space
        while (idx < len && isspace(entry[idx])) 
            idx++;
        if (idx != len) {
            sprintf(buf, "Unexpected character at end of entry at position %d", idx-1);
            throwUdxException(buf);
        }
        *key = m_key.c_str();
        *value = m_value.c_str();
    }
    void process() {
        int i;
        for (i=0; i < m_pParam->getNumEntries(); i++) {
            const char* entry = m_pParam->getEntry(i);
            const char*key;
            const char*value;

            parseEntry(entry,&key, &value);
            m_output.insert(std::pair<std::string, std::string>(std::string(key), std::string(value)));

        }
        std::map<std::string, std::string>::iterator p;
        for (p = m_output.begin(); p != m_output.end(); p++)
        {
            std::string key = p->first;
            std::string value = p->second;
            m_pEnv->addEntry(key.c_str(), value.c_str());
        }
        m_pEnv->setReadOnly();

    }
    UdxEnvironment * getEnvironment() {
        return m_pEnv;
    }
private:
    std::string m_key;
    std::string m_value;
    std::map<std::string, std::string> m_output;
    UdxParameter *m_pParam;
    UdxEnvironment *m_pEnv;
};


} // end namespace udx_ver2
} // end namespace nz

#endif //  #ifndef UDX_ENVIRONMENT_INCLUDED
