// udxfunc.h
// Function class for User Defined Functions and Aggregates
//
// (C) Copyright IBM Corp. 2007, 2011  All Rights Reserved.

#ifndef UDX_FUNC_INCLUDED
#define UDX_FUNC_INCLUDED

#include "bitarch.h"

#define CHECK_SIZER_RANGE(n) if (n < 0 || n >= m_pInfo->numArgs) \
                  throwError(NZ_ERROR_ILLVAL, "Sizer argument number out of range", n)

#define CHECK_SIZER_TYPMOD(n, typmod) if (typmod == 0xFFFFFFFFFFFFFFFFLL) \
                  throwError(NZ_ERROR_ILLVAL, "Sizer argument invalid", n)
                          
namespace nz {  
namespace udx {


class Udf : public UdxBase
{
public:
	Udf() {}	
	virtual ~Udf() {}

	// Interface methods
	virtual ReturnValue evaluate() = 0;
	
	// The derived class must have a static method like such:
	// static Udf* instantiate();

	// Called from engine
	ReturnValue performEvaluate(FuncCallInfo *pFcInfo)
	  {
	    checkStack();
	    m_pInfo = pFcInfo;
	    ReturnValue oRet = evaluate();
	    m_pInfo = NULL;
	    return oRet;
	  }

    // The following will be used to support generic sized
    // string and numeric return values.

    // Can return UDX_NUMERIC, UDX_FIXED, UDX_VARIABLE,
    // UDX_NATIONAL_FIXED or UDX_NATIONAL_VARIABLE,
    int sizerReturnType() const;

    // Number of arguments the function is called with
    int numSizerArgs() const;

    // The type of the specified argument. (Can be any
    // of the DataType enums (Except UDX_NUMERIC). Will throw
    // an exception if n is out of range. 
    int sizerArgType(int n) const;

    // Returns the size of the specified argument. Will throw
    // exceptions if the n is out of range, if the specified
    // argument is not a string type or the specified argument
    // does not have a size. (shouldn't happen) This size is in characters,
    // not bytes. (this matters for the national strings)
    int sizerStringArgSize(int n) const;

    // Returns the precision of the specified argument. Will throw
    // exceptions if the n is out of range, if the specified
    // argument is not a numeric type or the specified argument
    // does not have a size. (shouldn't happen)
    int sizerNumericArgPrecision(int n) const;

    // Returns the scale of the specified argument. Will throw
    // exceptions if the n is out of range, if the specified
    // argument is not a numeric type or the specified argument
    // does not have a size. (shouldn't happen)
    int sizerNumericArgScale(int n) const;

    // Builds a return value for calculateSize given the
    // string length desired. Will throw an exception if
    // the return type is not a string. This length is in characters,
    // not bytes. (this matters for the national strings)
    uint64 sizerStringSizeValue(int len) const;

    // Builds a return value for calculateSize given the
    // precision and scale desired. Will throw an exception
    // if the return type is not a numeric.
    uint64 sizerNumericSizeValue(int prec, int scale) const;

    // In order to support certain methods like round(val, scale)
    // we have a mechanism for passing constant arguments to the
    // sizer so long as they are of int32 type.
    // This method will return true if the specified argument is
    // a constant integer. Will throw and exception if the n is
    // out of range or if the specified argument is not an int32
    // If the constant specified is -1, isSizerArgConstants
    // will return false.
    bool isSizerArgConstant(int n) const;

    // This method will return the specified constant int32 argument
    // Will throw and exception if the n is out of range, the specified
    // argument is not an int32 or the specified argument is
    // not constant.
    int32 sizerGetConstantArg(int n) const;
    
    // This method must be overriden to provide sizing
    // capabilities for numerics and strings. This must
    // be done in conjunction with the DDL.
    virtual uint64 calculateSize() const
    {
        return 0xFFFFFFFFFFFFFFFFLL;    
    }


};

inline bool Udf::isSizerArgConstant(int n) const
    {
        CHECK_INFO_AVAILABLE();
        CHECK_SIZER_RANGE(n);
        int nType = sizerArgType(n);
        if (nType != UDX_INT32 )
            throwError(NZ_ERROR_ILLVAL, "Expected int32 argument", nType);
        uint64 arg = (uint64)*(int64*)m_pInfo->args[n];
        int lower = arg >> 32;
        if (lower == -1)
            return false;
        return true;
    }
inline int32 Udf::sizerGetConstantArg(int n) const
    {
        CHECK_INFO_AVAILABLE();
        CHECK_SIZER_RANGE(n);
        int nType = sizerArgType(n);
        if (nType != UDX_INT32)
            throwError(NZ_ERROR_ILLVAL, "Expected int32 argument", nType);
        uint64 arg = (uint64)*(int64*)m_pInfo->args[n];
        int lower = arg >> 32;
        if (lower == -1)
            throwError(NZ_ERROR_ILLVAL, "Not a constant integer argument", 0);
        
        return lower;

    }

 
inline int Udf::sizerReturnType() const
    {
        CHECK_INFO_AVAILABLE();
        return m_pInfo->returnType;
    }
    
 inline int Udf::numSizerArgs() const
    {
        CHECK_INFO_AVAILABLE();
        return m_pInfo->numArgs;
    }

 inline int Udf::sizerArgType(int n) const
     {
         CHECK_INFO_AVAILABLE();
         CHECK_SIZER_RANGE(n);
         return m_pInfo->argTypes[n];
     }

 inline int Udf::sizerStringArgSize(int n) const
     {
         CHECK_INFO_AVAILABLE();
         CHECK_SIZER_RANGE(n);
         
         int nType = sizerArgType(n);
         if (nType != UDX_FIXED && nType != UDX_VARIABLE && nType != UDX_NATIONAL_FIXED &&
             nType != UDX_NATIONAL_VARIABLE && nType != UDX_GEOMETRY && nType != UDX_VARBINARY)
             throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary argument", nType);
         uint64 typmod = ((uint64)*(int64*)m_pInfo->args[n]) & 0xFFFFFFFF;
         CHECK_SIZER_TYPMOD(n, typmod);
         typmod -= VARHDRSZ;
         return (int32)typmod;
     } 

 inline int Udf::sizerNumericArgPrecision(int n) const
     {
         CHECK_INFO_AVAILABLE();
         CHECK_SIZER_RANGE(n);
         
         int nType = sizerArgType(n);
         if (nType != UDX_NUMERIC32 && nType != UDX_NUMERIC64 && nType != UDX_NUMERIC128)
             throwError(NZ_ERROR_ILLVAL, "Expected numeric argument", nType);
         uint64 typmod = ((uint64)*(int64*)m_pInfo->args[n]) & 0xFFFFFFFF;
         CHECK_SIZER_TYPMOD(n, typmod);
         typmod -= VARHDRSZ;
         return (typmod >> 16) & 0x7fff;
     }
 inline int Udf::sizerNumericArgScale(int n) const
     {
         CHECK_INFO_AVAILABLE();
         CHECK_SIZER_RANGE(n);
         
         int nType = sizerArgType(n);
         if (nType != UDX_NUMERIC32 && nType != UDX_NUMERIC64 && nType != UDX_NUMERIC128)
             throwError(NZ_ERROR_ILLVAL, "Expected numeric argument", nType);
         uint64 typmod = ((uint64)*(int64*)m_pInfo->args[n]) & 0xFFFFFFFF;
         CHECK_SIZER_TYPMOD(n, typmod);
         typmod -= VARHDRSZ;
         return typmod & 0xffff;
     }

 inline uint64 Udf::sizerStringSizeValue(int len) const
     {
         CHECK_INFO_AVAILABLE();
         int nType = m_pInfo->returnType;
         if (nType != UDX_FIXED && nType != UDX_VARIABLE && nType != UDX_NATIONAL_FIXED &&
             nType != UDX_NATIONAL_VARIABLE && nType != UDX_GEOMETRY && nType != UDX_VARBINARY)
             throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary return type", nType);
         
         return ((uint64)0xFFFFFFFF00000000LL + len + VARHDRSZ);
     }

 inline uint64 Udf::sizerNumericSizeValue(int prec, int scale) const
     {
         CHECK_INFO_AVAILABLE();
         int nType = m_pInfo->returnType;
         if (nType != UDX_NUMERIC)
             throwError(NZ_ERROR_ILLVAL, "Expected numeric return type", nType);
         return ((uint64)0xFFFFFFFF00000000LL + (((prec)&0x7fff)<<16) + (scale) + VARHDRSZ);
     }
    
} // end namespace udx
} // end namespace nz

#endif //  #ifndef UDX_FUNC_INCLUDED
