// udxnamevalue.h
// NameValue classes for User Defined Functions and Aggregates
//
// (C) Copyright IBM Corp. 2009, 2014  All Rights Reserved.

#ifndef UDX_NAMEVALUE_INCLUDED
#define UDX_NAMEVALUE_INCLUDED


namespace nz {
namespace udx_ver2 {



class UdxNameValuePair
{
public:
    UdxNameValuePair(const char* strKey, const char* strValue) {
        m_strKey = strdup(strKey);
        if (m_strKey == NULL)
            throwUdxException("Couldn't duplicate key");
        m_strValue = strdup(strValue);
        if (m_strValue == NULL)
            throwUdxException("Couldn't duplicate value");
    }
    virtual ~UdxNameValuePair() {
        free(m_strKey);
        free(m_strValue);
    }

    // Key should be immutable.
    // Only the container class should use
    // this method since all users
    // will get a const pointer to this object
    virtual void setValue(const char* strValue) {
        if (m_strValue)
            free(m_strValue);
        m_strValue = strdup(strValue);
        if (m_strValue == NULL)
            throwUdxException("Couldn't duplicate value");
    }
    const char* getKey() const { return m_strKey; }
    const char* getValue() const { return m_strValue; }

private:
    char *m_strKey;
    char *m_strValue;
};


class UdxNameValuePairs
{
public:
    UdxNameValuePairs() {
        m_nSize = 0;
        m_nAllocated = 8;
        m_pEntries = new UdxNameValuePair*[m_nAllocated];
        if (m_pEntries == NULL)
            throwUdxException("Couldn't allocate entries array");
        m_bReadOnly = false;
    }
    virtual ~UdxNameValuePairs() {
        if (m_pEntries)
        {
            for (int i=0; i < m_nSize; i++)
                delete m_pEntries[i];
            delete [] m_pEntries;
        }
    }
    void setReadOnly() { m_bReadOnly = true; }
    int getNumEntries() { return m_nSize; };

    virtual const UdxNameValuePair* getEntry(int idx) const;
    virtual const UdxNameValuePair* getEntry(const char* strKey) const;
    int findEntry(const char* strKey) const;

    void addEntry(const char* strKey, const char *strValue);

    void deleteEntry(const char* strKey);
    virtual void deleteEntry(int idx);
    void modifyEntry(const char* strKey, const char* strNewValue);
    virtual void modifyEntry(int idx, const char* strNewValue);


protected:
    virtual UdxNameValuePair* allocateEntry(const char* strKey, const char* strValue) {
        return new UdxNameValuePair(strKey, strValue);
    }

private:
    UdxNameValuePair **m_pEntries;
    int m_nSize;
    int m_nAllocated;
    bool m_bReadOnly;
};

inline void  UdxNameValuePairs::deleteEntry(int idx)
{
    if (idx < 0 || idx >= m_nSize)
        throwUdxException("Index is out of bounds");
    if (m_bReadOnly)
        throwUdxException("Object is read only now");
    delete m_pEntries[idx];
    for (int i=idx+1; i < m_nSize; i++)
        m_pEntries[i-1] = m_pEntries[i];
    m_nSize--;

}

inline void UdxNameValuePairs::deleteEntry(const char *strKey)
{
    int idx = findEntry(strKey);
    deleteEntry(idx);

}

inline void UdxNameValuePairs::modifyEntry(int idx, const char* strNewValue)
{
    if (idx < 0 || idx >= m_nSize)
        throwUdxException("Index is out of bounds");
    if (m_bReadOnly)
        throwUdxException("Object is read only now");
    m_pEntries[idx]->setValue(strNewValue);

}

inline void  UdxNameValuePairs::modifyEntry(const char* strKey, const char* strNewValue)
{
    int idx = findEntry(strKey);
    modifyEntry(idx, strNewValue);

}

inline const UdxNameValuePair* UdxNameValuePairs::getEntry(int idx) const
{
    if (idx < 0 || idx >= m_nSize)
        throwUdxException("Index is out of bounds");
    return m_pEntries[idx];

}
inline int UdxNameValuePairs::findEntry(const char* strKey) const
{
    for (int i=0; i < m_nSize;i++)
    {
        if (strcmp(strKey, m_pEntries[i]->getKey()) == 0)
            return i;
    }
    return -1;
}
inline const UdxNameValuePair* UdxNameValuePairs::getEntry(const char* strKey) const
{
    int idx = findEntry(strKey);
    if (idx == -1)
        return NULL;
    return m_pEntries[idx];

}

inline void UdxNameValuePairs::addEntry(const char* strKey, const char *strValue)
{
    if (strKey == NULL)
        throwUdxException("NULL key");
    if (strValue == NULL)
        throwUdxException("Null value");
    if (m_bReadOnly)
        throwUdxException("Object is read only now");
    if (findEntry(strKey) != -1)
    {
        const char *msgText = "Key already exists: %s";
        char initialBuf[100];
        char *msg = initialBuf;
        int required = snprintf(msg, sizeof(initialBuf), msgText, strKey);
        if (required >= sizeof(initialBuf)) {
            msg = (char *)malloc((required + 1) * sizeof(char));
            snprintf(msg, required + 1, msgText, strKey);
        }
        throwUdxException(msg);
    }

    if (m_nSize == m_nAllocated-1)
    {
        m_nAllocated *= 2;
        UdxNameValuePair ** temp = new UdxNameValuePair*[m_nAllocated];
        if (temp == NULL)
            throwUdxException("Couldn't allocate new entries array");
        for (int i=0; i < m_nSize; i++)
            temp[i] = m_pEntries[i];
        delete [] m_pEntries;
        m_pEntries = temp;
    }
    m_pEntries[m_nSize] = allocateEntry(strKey, strValue);
    if (m_pEntries[m_nSize] == NULL)
        throwUdxException("Couldn't allocate new NameValue entry");
    m_nSize++;

}


} // end namespace udx_ver2
} // end namespace nz

#endif //  #ifndef UDX_NAMEVALUE_INCLUDED
