// udxoptimizer.h
// Optimizer classes for User Defined Functions and Aggregates
//
// (C) Copyright IBM Corp. 2009, 2011  All Rights Reserved.

#ifndef UDX_OPTIMIZER_INCLUDED
#define UDX_OPTIMIZER_INCLUDED


namespace nz {  
namespace udx_ver2 {


class UdxOptimizerInfo : public UdxOutputInfo
{
public:   
    UdxOptimizerInfo(ArgInfo *pInfo, int numInputs, int numOutputs) : UdxOutputInfo(pInfo) {
        m_numInputs = numInputs;
        m_numOutputs = numOutputs;
        if (numInputs) {
            m_inputInfo =  new UdxNameValuePairs[numInputs];
        }
        else
            m_inputInfo = NULL;
        if (numOutputs) {
            m_outputInfo =  new UdxNameValuePairs[numOutputs];
        }
        else
            m_outputInfo = NULL;
    }
    ~UdxOptimizerInfo() {
        if (m_inputInfo)
            delete[] m_inputInfo;
        if (m_outputInfo)
            delete[] m_outputInfo;
    }
    int getNumInputs() { return m_numInputs; };
    int getNumOutputs() { return m_numOutputs; };

    // optimizer input information
    // should only be called by system
    // Same with addOutputColumn
    void setInputReadOnly() {
        m_globalInput.setReadOnly();
        for (int i=0; i < m_numInputs; i++)
            m_inputInfo[i].setReadOnly();
    }
    UdxNameValuePairs* getGlobalInputInfo() { return &m_globalInput; }
    UdxNameValuePairs* getArgumentInputInfo(int idx) {
        if (idx <0 || idx >= m_numInputs)
            throwUdxException("Input index is out of bounds");
        return &m_inputInfo[idx];
    }

    // optimizer output information
    UdxNameValuePairs* getGlobalOutputInfo() { return &m_globalOutput; }
    UdxNameValuePairs* getOutputInfo(int idx) {
        if (idx <0 || idx >= m_numOutputs)
            throwUdxException("Output index is out of bounds");
        return &m_outputInfo[idx];
    }
    
    // Output information 
    // These should only be called by system
    void addOutputColumn(int nType, const char* strName, int nSize) {
        addNamedColumn(nType, strName, nSize);
    }
    void addOutputColumn(int nType, const char* strName, int precision, int scale) {
        addNamedColumn(nType, strName, precision, scale);
    }
    void addOutputColumn(int nType, const char* strName) {
        addNamedColumn(nType, strName);
    }
private:
    int m_numInputs;
    UdxNameValuePairs m_globalInput;
    UdxNameValuePairs* m_inputInfo;

    int m_numOutputs;
    UdxNameValuePairs m_globalOutput;
    UdxNameValuePairs* m_outputInfo;
};


class UdxAggregateOptimizerInfo : public UdxOptimizerInfo
{
public:
    UdxAggregateOptimizerInfo(ArgInfo *pInfo, int numInputs, int numOutputs) : 
        UdxOptimizerInfo(pInfo, numInputs, numOutputs) {}

    // output information
    int numStateColumns() {
        return numUnNamedColumns();
    }
    const UdxColumnInfo* getStateColumn(int n) {
        return getUnNamedColumn(n);
    }

    // These should only be called by system
    void addStateColumn(int nType, int nSize) {
        addUnNamedColumn(nType, nSize);
    }
    void addStateColumn(int nType, int precision, int scale) {
        addUnNamedColumn(nType, precision, scale);
    }
    void addStateColumn(int nType) {
        addUnNamedColumn(nType);
    }

};

} // end namespace udx_ver2
} // end namespace nz

#endif //  #ifndef UDX_OPTIMIZER_INCLUDED
