// udxragg.h
// Row aggregate class for User Defined Functions and Aggregates
//
// (C) Copyright IBM Corp. 2009, 2011  All Rights Reserved.

#ifndef UDX_RAGG_INCLUDED
#define UDX_RAGG_INCLUDED


namespace nz {  
namespace udx_ver2 {


#define CHECK_ROW_AGG_RANGE(n) if (n < 0 || n >= ((RowAggCallInfo*)m_pInfo)->numStateVars) \
		  throwError(NZ_ERROR_ILLVAL, "State number out of range", n)

class Udra : public UdxNonScalarBase
{
public:
    

	Udra(UdxInit *pInit) : UdxNonScalarBase(pInit) 
	  {
	    m_bInit = false;
        m_aggregateType = UnknownAggregateType;
	  }
	virtual ~Udra() {}
	
    enum AggregateType
    {
        UnknownAggregateType,
        Grouped,
        Analytic    
    };
    AggregateType aggregateType() const {
        return m_aggregateType;
    }
    void setAggregateType(AggregateType type) {
        m_aggregateType = type;
    }
	// Interface methods
				
	// Must initialize all state variables and state null indicators in this method
	virtual void initializeState() = 0;

	// Called once per each row (called by the  engine)
	// Called with state values. which will be updated when method returns
	virtual void accumulate() = 0;

	// Merge serialized state of another aggregate with this one.
	virtual void merge() = 0;

	// Finish up and return result.
	// Will not cleanup memory, destructor should do this.
	virtual void finalResult() = 0;

	// The derived class must have a static method like such:
	// static Udra* instantiate(UdxInit *pInit);

	// State methods
	int numStateVars() const;
	int stateType(int n) const;
	void setStateNull(int n, bool val) const;
	bool isStateNull(int n) const;	
	StringArg* stringState(int n) const;
	Numeric32Val* numeric32State(int n) const;
	Numeric64Val* numeric64State(int n) const;
	Numeric128Val* numeric128State(int n) const;
	TimeTzADT* timetzState(int n) const;	
	struct Interval* intervalState(int n) const;
	float* floatState(int n) const;	
	double* doubleState(int n) const;
	int8* int8State(int n) const;
	int16* int16State(int n) const;
	int32* int32State(int n) const;
	int64* int64State(int n) const;
	bool* boolState(int n) const;
	int32* dateState(int n) const;
	int64* timeState(int n) const;
	int64* timestampState(int n) const;

	
	// Called from Engine
	void init();
	void performAccumulate(RowAggCallInfo *pAcInfo);	
	void performMerge(RowAggCallInfo *pAcInfo);
	void performFinalResult(RowAggCallInfo *pAcInfo);

	

 private:
	bool m_bInit;
    AggregateType m_aggregateType;
};


inline int Udra::numStateVars() const
{
    CHECK_INFO_AVAILABLE();
    return ((RowAggCallInfo *)m_pInfo)->numStateVars;
}
inline int Udra::stateType(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    return ((RowAggCallInfo *)m_pInfo)->stateVarTypes[n];
}
inline void Udra::setStateNull(int n, bool val) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    *((RowAggCallInfo *)m_pInfo)->pStateVarNulls[n] = val;
}
inline bool Udra::isStateNull(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    return *((RowAggCallInfo *)m_pInfo)->pStateVarNulls[n];
}

inline StringArg* Udra::stringState(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_FIXED && nType != UDX_VARIABLE && nType != UDX_NATIONAL_FIXED &&
        nType != UDX_NATIONAL_VARIABLE && nType != UDX_GEOMETRY && nType != UDX_VARBINARY)
        throwError(NZ_ERROR_ILLVAL, "Expected string, st_geometry, or varbinary state value", nType);
    return ((StringArg*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline Numeric32Val* Udra::numeric32State(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_NUMERIC32)
        throwError(NZ_ERROR_ILLVAL, "Expected numeric32 state value", nType);
    return ((Numeric32Val*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline Numeric64Val* Udra::numeric64State(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_NUMERIC64)
        throwError(NZ_ERROR_ILLVAL, "Expected numeric64 state value", nType);
    return ((Numeric64Val*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline Numeric128Val* Udra::numeric128State(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_NUMERIC128)
        throwError(NZ_ERROR_ILLVAL, "Expected numeric128 state value", nType);
    return ((Numeric128Val*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}

inline TimeTzADT* Udra::timetzState(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_TIMETZ)
        throwError(NZ_ERROR_ILLVAL, "Expected timetz state value", nType);
    return ((TimeTzADT*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
	
inline struct Interval* Udra::intervalState(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_INTERVAL)
        throwError(NZ_ERROR_ILLVAL, "Expected interval state value", nType);
    return ((struct Interval*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline float* Udra::floatState(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_FLOAT)
        throwError(NZ_ERROR_ILLVAL, "Expected float state value", nType);
    return ((float*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
	
inline double* Udra::doubleState(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_DOUBLE)
        throwError(NZ_ERROR_ILLVAL, "Expected double state value", nType);
    return ((double*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline int8* Udra::int8State(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_INT8)
        throwError(NZ_ERROR_ILLVAL, "Expected int8 state value", nType);
    return ((int8*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline int16* Udra::int16State(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_INT16)
        throwError(NZ_ERROR_ILLVAL, "Expected int16 state value", nType);
    return ((int16*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline int32* Udra::int32State(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_INT32)
        throwError(NZ_ERROR_ILLVAL, "Expected int32 state value", nType);
    return ((int32*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline int64* Udra::int64State(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_INT64)
        throwError(NZ_ERROR_ILLVAL, "Expected int64 state value", nType);
    return ((int64*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline bool* Udra::boolState(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_BOOL)
        throwError(NZ_ERROR_ILLVAL, "Expected boolean state value", nType);
    return (bool*)((int8*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline int32* Udra::dateState(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_DATE)
        throwError(NZ_ERROR_ILLVAL, "Expected date state value", nType);
    return ((int32*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline int64* Udra::timeState(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_TIME)
        throwError(NZ_ERROR_ILLVAL, "Expected time state value", nType);
    return ((int64*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline int64* Udra::timestampState(int n) const
{
    CHECK_INFO_AVAILABLE();
    CHECK_ROW_AGG_RANGE(n);
    int nType = stateType(n);
    if (nType != UDX_TIMESTAMP)
        throwError(NZ_ERROR_ILLVAL, "Expected timestamp state value", nType);
    return ((int64*)((RowAggCallInfo *)m_pInfo)->stateVars[n]);
}
inline void Udra::init()
{
    m_bInit = false;
}
	
inline void Udra::performAccumulate(RowAggCallInfo *pAcInfo)
{

    m_pInfo = pAcInfo;
    if (!m_bInit)
        {
            initializeState();
            m_bInit = true;
        }
    accumulate();
    m_pInfo = NULL;
}
	
	
inline void Udra::performMerge(RowAggCallInfo *pAcInfo)
{

    m_pInfo = pAcInfo;
    if (!m_bInit)
        {
            initializeState();
            m_bInit = true;
        }
    merge();
    m_pInfo = NULL;
}

inline void Udra::performFinalResult(RowAggCallInfo *pAcInfo)
{
    
    m_pInfo = pAcInfo;
    if (!m_bInit)
        {
            int i;
            // This works since the arg information
            // in this case is actually state info.
            // This should only happen in odd window
            // aggregate cases where accumulate/merge
            // are never called, and thus initializeState
            // is never called.
             
            // Set num state vars from num args. 
            pAcInfo->numStateVars = pAcInfo->numArgs;

            // Set state var types. These should be integer values
            for (i=0; i < pAcInfo->numArgs; i++)
                pAcInfo->stateVarTypes[i] = pAcInfo->argTypes[i];
             
            // Set state vars. These should all be pointers.
            for (i=0; i < pAcInfo->numArgs; i++)
                pAcInfo->stateVars[i] = pAcInfo->args[i];

            // Set state null indicators. These are pointers to bools.
            // The arg indicators are bool values, so the changes made
            // to them won't persist beyond the lifetime of performFinalResult.
            // Nevertheless, this should not be an issue as the state indicators
            // should not be used by the genereated code after this point.
            for (i=0; i < pAcInfo->numArgs; i++)
                pAcInfo->pStateVarNulls[i] = &pAcInfo->argNulls[i];

            initializeState();
            m_bInit = true;
        }
    finalResult();
    m_pInfo = NULL;
    
}


} // end namespace udx_ver2
} // end namespace nz

#endif //  #ifndef UDX_RAGG_INCLUDED
