#ifndef XPDELETE_INCLUDED
#define XPDELETE_INCLUDED

#include "nodecommon.h"
#include "nzlimits.h"
#include "storagerefs.h"

struct DeletePageInfo {
    PageId id;    ///< id of the page
    int startRec; ///< index of first record for that page
    int endRec;   ///< one past index of last record for that page, or -1
};

// helper function for TDeleteNode::GetPagesForRecords
// for a simple solution, we compute this information on the fly with a full pass
// over the reclocs that should be deleted.
// for slightly improved performance, we can avoid this computation by storing the
// recloc offsets at which the page ID changes whenever TDeleteNode::SetPage2Modify
// is called. This would imply gencode modifications that we want to avoid for the
// moment.
int LocatePagesForRecords (
    TRecLoc *recLocs, // records to process
    int numrecs,  // number of records
    DeletePageInfo *pageInfo  // output information
);


class TDeleteNodeDef {
protected:
    table_id_t  m_tblId;        // id of the table to be updated
    bool        m_forDelete;    // false if used in update.
    // The index of the column that contains the record
    // locators (addresses) to delete. Usually PROJECT_DTIDBITADDR.
    int         m_recLocCol;
    bool        m_mustSortRecLocs;
};

class TDeleteNodeRuntimeInfo {
protected:
    uint64      m_delCount;     // total cnt deleted
    TRecLoc*    m_recLocs;      // records to delete on this pass
    unsigned    m_numToDelete;  // cnt this pass (size of m_recAddrs)
    unsigned    m_nPages;       // span in pages of deleted records
    PageId      m_prevPage;     // PageId of most recently recorded delete
    PageId      m_pages2Modify[MAX_GENC_DELETE_PAGES];
    unsigned    m_nMaxRecLocs;  // size of RecLoc array
    CTable*     m_delTable;     // table being deleted from
};

class TDeleteNode : public SnippetNode,
    private TDeleteNodeDef, private TDeleteNodeRuntimeInfo {
public:
    TableShape const * Init(TableShape const * src, EPlanNodeLocus place);
    virtual ~TDeleteNode() { /* nothing to do here */ }

    // Address-of accessors
    // Used for things like virtual-table materialization.
    inline uint64*    AddrOfDelCount()        { return &m_delCount; } // yuck!

    // Accessors
    inline uint64     DelCount()        const { return m_delCount; }
    inline bool       ForDelete()       const { return m_forDelete; }
    inline unsigned   NMaxRecLocs()     const { return m_nMaxRecLocs; }
    inline unsigned   NumToDelete()     const { return m_numToDelete; }
    inline unsigned   Pages()           const { return m_nPages; }
    inline PageId     PrevPage()        const { return m_prevPage; }
    inline TRecLoc*   RecLocs()         const { return m_recLocs; }
    inline table_id_t TblId()           const { return m_tblId; }
    inline int        RecLocCol()       const { return m_recLocCol; }
    inline CTable*    DelTable()        const { return m_delTable; }
    inline bool       MustSortRecLocs() const { return m_mustSortRecLocs; }

    // Special accessors
    inline uint64     Page2ModifyRep( unsigned ix ) const { return m_pages2Modify[ix].rep(); }

    // Non-void mutators
    inline unsigned PostIncNPages()                    { return m_nPages++; }
    inline unsigned PostIncNumToDelete()               { return m_numToDelete++; }

    // Mutators
    inline void     DecNumToDelete()               { m_numToDelete--; }
    inline void     IncDelCount( uint64 ct )       { m_delCount += ct; }

    // Setters
    inline void SetForDelete( bool bb )                   { m_forDelete = bb; }
    inline void SetMaxRecLocs( unsigned num )             { m_nMaxRecLocs = num; }
    inline void SetNumToDelete( unsigned num )            { m_numToDelete = num; }
    inline void SetNPages( unsigned num )                 { m_nPages = num; }
    inline void SetPage2Modify( unsigned ix, PageId val ) { m_pages2Modify[ix] = val; }
    inline void AddPage2Modify( PageId val )              { m_pages2Modify[m_nPages++] = val; }
    inline void SetPrevPage( PageId val )                 { m_prevPage = val; }
    inline void SetRecLoc( unsigned ix, TRecLoc val )     { m_recLocs[ix] = val; }
    inline void AddRecLoc( TRecLoc val )                  { m_recLocs[m_numToDelete++] = val; }
    inline void SetTblId( table_id_t tbl )                { m_tblId = tbl; }
    inline void SetRecLocCol( int recLocColIdx )          { m_recLocCol = recLocColIdx; }
    inline void SetMustSortRecLocs( bool bb )             { m_mustSortRecLocs = bb; }

    inline bool IsPageNew(const PageId& p) const {
        for (int i = m_nPages; --i >= 0;) {
            if (m_pages2Modify[i] == p) {
                return false;
            }
        }
        return true;
    }
    
    inline int GetPagesForRecords( DeletePageInfo *pageInfo ) {
        return LocatePagesForRecords(m_recLocs, m_numToDelete, pageInfo);
    }
};

void ExecDeleteDone( // Perform the actual storage level deletion (called by gen'd code)
    TDeleteNode *pNode, bool lastCall, bool forDelete=false);

#endif // !XPDELETE_INCLUDED
