#ifndef XPHASHJOIN_INCLUDED
#define XPHASHJOIN_INCLUDED

#include "xpjoin.h"

class CPlan;
class CRecordStore;
struct THashInfo;
struct TMemblockInfo;

/* Support for execution plan joining */

class TDHJNode {   //container class for DHJ nodes
public:
    CRecordStore **   m_hashStoreArr;     //hash value array of CRecordStore ptrs
    nzotuid_t         m_hashHeapOwner;    // process that allocated arrays from heap
    uint32            m_nHashVals; //number of hash vals in the histogram
    void             *m_pHashSizesArr; //sum of sizes of records for each hash value
    bool              m_isLongInt;   // true if m_pHashSizesArr is of type 8-byte ints 
                                     // else 4-byte ints
    nzotuid_t         m_hashSizesArrOwner;      // process that allocated m_pHashSizesArr
    
    uint              m_nHashParts; //number of hash partitions for the temp table
    uint              m_hashMask;     //mask for hash values to index into hash table
    struct THashPart *m_pHashParts; //hash partitions
    uint              m_nValsForPartitioning; //vals for partitioning
    CRecordStore **   m_streamStores;     //stream partition record store
    CRecordStore **   m_tempStores;       //temp partition record store
    CRecordStore *    m_streamBVStore; //bit vector for stream outer joins
    uint16 *          m_pTempHashTableArr; //hash table containing partition identifiers
                                 //used for temp table partitioning
    uint              m_dhjPartitionPages;   // working memory for partitioning 


    void allocateHashHist(CRecordStore* pTmpRS);
    void computeHashHist(TJoinNode *pNode);
    void computeHashPartIntervals(TJoinNode *pNode);
    void dhjWorkingMemQuota(TJoinNode *pNode);
    void initStreamPartitions(const TableShape *src, TJoinNode *pNode);
    void produceTempPartitions(TJoinNode *pNode);
    uint getMaxNumStreamRecords( //get num records in largest stream partition
         TJoinNode *pNode );
    void FreeHashSizesArr();
    void FreeHashStoreArr();
    void FreeHashPart();
    void primeHashTableForStreamPartitioning(CPlan *plan, CRecordStore *storeArr[] );
    inline void SetHashMask( uint val ) { m_hashMask = val; }
};

class THashJoinNodeDef {
protected:
    int  m_numHashCols;
    int *m_tmpHashCols;
    int *m_streamHashCols;
    // true => nulls considered equal for this join field  (/=/ operator)
    // may be NULL => all false
    bool *m_nullsCompareCols;
};

class THashJoinNodeRuntimeInfo {
protected:
    TDHJNode       m_dhj;

    CTable *       m_pJoinTbl;       //temp table partition for DHJ
    CRecordStore*  m_pStreamBVStore; //stream bit vector recordstore
    TMemblockInfo* m_pBVBlock; //transient for use in gencode
    uint32         m_iStreamRec;   //sequence number of stream record being processed 
    bool           m_bSetStreamBits; //if true set bits in the stream bit vector
    bool           m_bWriteStreamOuter;  //write non-matching outer records for outerjoins
    uint32         m_HJnumBuckets;        // staged hash parameter for gencode
    THashInfo**    m_HJhashBuckets;  // staged hash bucket array for gencode

    // for node stats gathering
    uint64         m_numRecsInInner;
    uint64         m_numIncomingRecs; // The usual "recs" field in node stats is the *outgoing* count.
    uint64         m_numExtraRecs;    // Number of additional output recs beyond the one-output-per-input count.
    // Note: We can get the number of unmatched streaming records as incoming+extra-outgoing
};

class THashJoinNode : public TJoinNode,
    private THashJoinNodeDef, private THashJoinNodeRuntimeInfo  {
public:
    TableShape const * Init(TableShape const * src, EPlanNodeLocus calledFrom);
    virtual ~THashJoinNode() { /* Nothing to do here */ }

    // Address-of accessors
    // Used for things like virtual-table materialization.
    inline TDHJNode*      AddrOfDhj()              { return &m_dhj; }
    inline CTable**       AddrOfJoinTbl()          { return &m_pJoinTbl; }
    inline uint64*        AddrOfNumExtraRecs()     { return &m_numExtraRecs; }
    inline uint64*        AddrOfNumIncomingRecs()  { return &m_numIncomingRecs; }
    inline uint64*        AddrOfNumRecsInInner()   { return &m_numRecsInInner; }

    // Accessors
    inline TMemblockInfo* BVBlock()                { return m_pBVBlock; }
    inline TDHJNode &     Dhj()                    { return m_dhj; }
    inline THashInfo**    HJhashBuckets()    const { return m_HJhashBuckets; }
    inline uint32         HJnumBuckets()     const { return m_HJnumBuckets; }
    inline uint32         IStreamRec()       const { return m_iStreamRec; }
    inline CTable*        JoinTbl()          const { return m_pJoinTbl; }
    inline uint64         NumExtraRecs()     const { return m_numExtraRecs; }
    inline int            NumHashCols()      const { return m_numHashCols; }
    inline uint64         NumIncomingRecs()  const { return m_numIncomingRecs; }
    inline uint64         NumRecsInInner()   const { return m_numRecsInInner; }
    inline bool           SetStreamBits()    const { return m_bSetStreamBits; }
    inline CRecordStore*  StreamBVStore()    const { return m_pStreamBVStore; }
    inline int*           StreamHashCols()   const { return m_streamHashCols; }
    inline int*           TmpHashCols()      const { return m_tmpHashCols; }
    inline bool*          NullsCompareCols() const { return m_nullsCompareCols; }
    inline bool           WriteStreamOuter() const { return m_bWriteStreamOuter; }

    // Testers: Overrides of JoinNode defaults:
    bool IsHashed() const { return true; }

    // Mutators
    inline void IncIStreamRec()                         { m_iStreamRec++; }

    // Non-void setters
    inline TMemblockInfo*  SetBVBlock( TMemblockInfo* block ) { return m_pBVBlock = block; }

    // Setters
    inline void SetIStreamRec( uint32 rec )             { m_iStreamRec = rec; }
    inline void SetJoinTbl( CTable* tbl )               { m_pJoinTbl = tbl; }
    inline void SetHJhashBuckets( THashInfo** bckts )   { m_HJhashBuckets = bckts; }
    inline void SetHJnumBuckets( uint32 num )           { m_HJnumBuckets = num; }
    inline void SetNumExtraRecs( uint64 val )           { m_numExtraRecs = val; }
    inline void SetNumHashCols( int num )               { m_numHashCols = num; }
    inline void SetNumIncomingRecs( uint64 val )        { m_numIncomingRecs = val; }
    inline void SetNumRecsInInner( uint64 val )         { m_numRecsInInner = val; }
    inline void SetSetStreamBits( bool bb )             { m_bSetStreamBits = bb; }
    inline void SetStreamBVStore( CRecordStore* store ) { m_pStreamBVStore = store; }
    inline void SetStreamHashCols( int* cols )          { m_streamHashCols = cols; }
    inline void SetTmpHashCols( int* cols )             { m_tmpHashCols = cols; }
    inline void SetNullsCompareCols( bool* val )        { m_nullsCompareCols = val; }
    inline void SetWriteStreamOuter( bool bb )          { m_bWriteStreamOuter = bb; }

protected:
    // Non-inlines
    void initDHJ(const TableShape *src);
};

//data structure to maintain linked list of hash intervals in each partition
struct  THashPartPair   {
    struct THashPartPair *next; //next interval
    uint16 start;   //start of interval
    uint16 end;     //end of interval
};

struct  THashPart   {
    uint64 partSize;
    THashPartPair *head, 
                  *tail; //just for debug
    uint8  type;    //type of partition, see below for definitions
};


/*===========================================================================
* Hash partition types
* 1) Normal paritition - contains one or more hash intervals
*    (HASH_PART_NORMAL | HASH_PART_FIRST | HASH_PART_LAST )
* 2) First of a multi-part sequence - first of a sequence of hash parts
*    that contain a single hash value
*    (HASH_PART_MULTIPLE | HASH_PART_FIRST)
* 3) Continuation partition - a continuation part of a sequence of hash parts
*    that contain a single hash value
*    (HASH_PART_MULTIPLE)
* 4) Last of a multi-part sequence - last of a sequence of hash parts
*    that contain a single hash value
*    (HASH_PART_MULTIPLE | HASH_PART_LAST)
* 5) Hybrid partition - last of a multi-part sequence that also contains
*    Normal hash intervals
*    (HASH_PART_MULTIPLE | HASH_PART_LAST | HASH_PART_NORMAL)
*
*===========================================================================
*/

#define HASH_PART_NORMAL    0x1 //a partition contains one or more hash 
                                //intervals 
#define HASH_PART_FIRST     0x2 //first of a multi-partition sequence or a normal                               //partition
#define HASH_PART_LAST      0x4 //last of a multi-partition sequence or a normal
                                //partition
#define HASH_PART_MULTIPLE  0x8 //partition that shares a single hash value with
                                //other partitions
#endif // !XPHASHJOIN_INCLUDED
