#ifndef XPJOIN_INCLUDED
#define XPJOIN_INCLUDED

#include "nodecommon.h"

/* Support for execution plan joining */

class THashNode;
class CRecordStore;
struct List;

// distinct from postgres JoinType -- for m_joinType, below
enum NzJoinType {
    NZ_JOIN_INNER = 0,
    NZ_JOIN_OUTER_STREAM = 1,
    NZ_JOIN_OUTER_TEMPTBL = 2,
    NZ_JOIN_OUTER_BOTH = 3
};

// For m_joinFlags below
enum NzJoinFlags {
    PQB_COL_NULLABLE      = (1<<0), // For "not exists" joins (NOT IN and relOp ALL)
                                    //     - Parent query join col is nullable
    SQB_COL_NULLABLE      = (1<<1), // For "not exists" joins (NOT IN and relOp ALL)
                                    //     - Subquery join col is nullable
    NULLS_IN_JOIN_COLS    = (1<<2), // For "not exists" joins (NOT IN and relOp ALL)
                                    //     - Nulls seen in join cols
    SQB_HAS_ROWS          = (1<<3)  // For "not exists" joins (NOT IN and relOp ALL)
                                    //     - Subquery has rows (or is not empty)
};


#define NODE_IS_HASHJOIN(p) \
((p)->IsJoinType() && (((TJoinNode *)(p))->IsHashed()))

class TJoinNodeDef {
protected:
    int    m_tblId;        // already existant temp table
    int    m_tblField;     // field in temp table
    int    m_joinField;    // field in pipeline table
    int    m_joinType;     // join operation
    int    m_joinFlags;    // join flags (see NzJoinFlags)
    List * m_joinQual;     // PG expression for matching
    List * m_otherQual;    // PG qual for outerjoins
    List * m_nullTestQualInner;   // null test qual for "not exists" joins only
    List * m_nullTestQualOuter;   // null test qual for "not exists" joins only
    bool   m_bNullsCompare;       // true => nulls considered equal for join (intersect/minus)
    bool   m_collectRC;           // true id count is needed (must mark the node in postgres world)
    bool   m_bDoOuterJoins;       // for generated code, process outerjoins if true
};

class TJoinNodeRuntimeInfo {
protected:
    bool   m_bDiskBased;           // true if disk based join  (host computes this and spu uses)
    bool   m_bLoaded;              // true if temp table loaded false if not
    int    m_nDataPagesForDHJ;     // pages available for disk based join  (host computes this and spu uses)
    int    m_nTotalPagesForDHJ;    // total pages available for disk based join (includes hashInfo and hashBucket)
    int    m_nPartsForDHJ;         // number of partitions to use for DHJ (host computes this and spu uses)

    int    m_pcodeStart;          // number of first pcode instruction
    int    m_pcodeCount;          // number of pcode instructions
    int    m_pcodeStartOtherQual; // first pcode instr for otherQual
    int    m_pcodeCountOtherQual; // number of pcode instrs for otherQual
    int    m_pcodeStartNullTestQualInner; // first pocde instr for nullTestQual
    int    m_pcodeCountNullTestQualInner; // number of pcode instr for nullTestQual
    int    m_pcodeStartNullTestQualOuter; // first pocde instr for nullTestQual
    int    m_pcodeCountNullTestQualOuter; // number of pcode instr for nullTestQual

    CTable       * m_tempTbl;            //inner table
    CRecordStore * m_outerTempTracking;  // bit vector for outerjoins
    void         * m_nullTempRecord;     // for expressions on nulled oj temp rec
    int            m_releaseTbl;         // ID of inner table to release / clean-up
    int64          m_tempTblSize;
    int64          m_tempTblNumRecords;
    int64          m_rCStreamingInto;    // row count streaming into the join.

    int            m_estDataSliceCount;  // Estimated number of data slices targeted. Zero if it can't be easily determined.
    int            m_pagesFiltered; // Total number of pages filtered by join above pages already filtered.
};

class TJoinNode : public SnippetNode,
     private TJoinNodeDef, private TJoinNodeRuntimeInfo { //base class for JoinNodes

public:                      // join (always followed by a project)
    virtual TableShape const * Init(TableShape const * src, EPlanNodeLocus place) = 0; // Abstract base class
    virtual ~TJoinNode() { /* nothing to do here */ }

    // Address-of accessors
    // Used for things like virtual-table materialization.
    inline const int*    AddrOfJoinField()   const { return &m_joinField; }
    inline const int*    AddrOfTblField()    const { return &m_tblField; }
    inline const int*    AddrOfTblId()       const { return &m_tblId; }

    // Accessors
    inline bool          CollectRC()         const { return m_collectRC; }
    inline bool          DiskBased()         const { return m_bDiskBased; }
    inline bool          DoOuterJoins()      const { return m_bDoOuterJoins; }
    inline int           JoinField()         const { return m_joinField; }
    inline int           JoinFlags()         const { return m_joinFlags; }
    inline List*         JoinQual()          const { return m_joinQual; }
    inline int           JoinType()          const { return m_joinType; }
    inline bool          Loaded()            const { return m_bLoaded; }
    inline int           NDataPagesForDHJ()  const { return m_nDataPagesForDHJ; }
    inline int           NTotalPagesForDHJ() const { return m_nTotalPagesForDHJ; }
    inline int           NPartsForDHJ()      const { return m_nPartsForDHJ; }
    inline bool          NullsCompare()      const { return m_bNullsCompare; }
    inline void*         NullTempRecord()    const { return m_nullTempRecord; }
    inline List*         NullTestQualInner() const { return m_nullTestQualInner; }
    inline List*         NullTestQualOuter() const { return m_nullTestQualOuter; }
    inline List*         OtherQual()         const { return m_otherQual; }
    inline CRecordStore* OuterTempTracking() const { return m_outerTempTracking; }

    inline int PcodeCount()                  const { return m_pcodeCount; }
    inline int PcodeCountNullTestQualInner() const { return m_pcodeCountNullTestQualInner; }
    inline int PcodeCountNullTestQualOuter() const { return m_pcodeCountNullTestQualOuter; }
    inline int PcodeCountOtherQual()         const { return m_pcodeCountOtherQual; }

    inline int PcodeStart()                  const { return m_pcodeStart; }
    inline int PcodeStartNullTestQualInner() const { return m_pcodeStartNullTestQualInner; }
    inline int PcodeStartNullTestQualOuter() const { return m_pcodeStartNullTestQualOuter; }
    inline int PcodeStartOtherQual()         const { return m_pcodeStartOtherQual; }

    inline int64         RCStreamingInto()   const { return m_rCStreamingInto; }
    inline int           TblField()          const { return m_tblField; }
    inline int           TblId()             const { return m_tblId; }
    inline CTable*       TempTbl()           const { return m_tempTbl; }
    inline int64         TempTblNumRecords() const { return m_tempTblNumRecords; }
    inline int64         TempTblSize()       const { return m_tempTblSize; }

    inline int           EstDataSliceCount() const { return m_estDataSliceCount; }
    inline int           PagesFiltered() const { return m_pagesFiltered; }
   // Testers
    inline bool IsDHJ()                      const { return (IsHashed() && m_bDiskBased); }

    // Virtual testers: JoinNode defaults
    virtual inline bool IsExprStyle()             const { return false; }                              // default for JoinNode types
    virtual inline bool IsHashed()                const { return false; }                              // default for JoinNode types
    virtual inline bool IsMergeStyle()            const { return false; }                              // default for JoinNode types
    virtual inline bool IsStreamOuterStyle()      const { return JoinType() & NZ_JOIN_OUTER_STREAM; }  // default for JoinNode types
    virtual inline bool NeedsRightOuterJoinPass() const { return JoinType() & NZ_JOIN_OUTER_TEMPTBL; } // default for JoinNode types

    // Testers: overrides of SnippetNode methods
    virtual inline bool IsJoinType()              const { return true; }                               // JoinNode override SnippetNode

    // Mutators
    inline void IncRCStreamingInto()                     { m_rCStreamingInto++; }
    inline void AddJoinFlags( int flags )                { m_joinFlags |= flags; }

    // Setters
    inline void SetCollectRC( bool bb )                  { m_collectRC = bb; }
    inline void SetDiskBased( bool bb )                  { m_bDiskBased = bb; }
    inline void SetDoOuterJoins( bool bb )               { m_bDoOuterJoins = bb; }
    inline void SetJoinField( int fld )                  { m_joinField = fld; }
    inline void SetJoinFlags( int flags )                { m_joinFlags = flags; }
    inline void SetJoinQual( List* qual )                { m_joinQual = qual; }
    inline void SetJoinType( int type )                  { m_joinType = type; }
    inline void SetNullsCompare(bool flag)               { m_bNullsCompare = flag; }
    inline void SetNullTempRecord( void* ptr )           { m_nullTempRecord = ptr; }
    inline void SetNullTestQualInner( List* qual )       { m_nullTestQualInner = qual; }
    inline void SetNullTestQualOuter( List* qual )       { m_nullTestQualOuter = qual; }
    inline void SetNDataPagesForDHJ( int pages )         { m_nDataPagesForDHJ = pages; }
    inline void SetNTotalPagesForDHJ( int pages )        { m_nTotalPagesForDHJ = pages; }
    inline void SetNPartsForDHJ( int parts )             { m_nPartsForDHJ = parts; }
    inline void SetOtherQual( List* qual )               { m_otherQual = qual; }
    inline void SetOuterTempTracking( CRecordStore* rs ) { m_outerTempTracking = rs; }

    inline void SetPcodeCount( int count )                  { m_pcodeCount = count; }
    inline void SetPcodeCountNullTestQualInner( int count ) { m_pcodeCountNullTestQualInner = count; }
    inline void SetPcodeCountNullTestQualOuter( int count ) { m_pcodeCountNullTestQualOuter = count; }
    inline void SetPcodeCountOtherQual( int count )         { m_pcodeCountOtherQual = count; }

    inline void SetPcodeStart( int start )                  { m_pcodeStart = start; }
    inline void SetPcodeStartNullTestQualInner( int start ) { m_pcodeStartNullTestQualInner = start; }
    inline void SetPcodeStartNullTestQualOuter( int start ) { m_pcodeStartNullTestQualOuter = start; }
    inline void SetPcodeStartOtherQual( int start )         { m_pcodeStartOtherQual = start; }

    inline void SetTblField( int fld )            { m_tblField = fld; }
    inline void SetTblId( int tid )               { m_tblId = tid; }
    inline void SetTempTbl( CTable* tbl )         { m_tempTbl = tbl; }
    inline void SetTempTblNumRecords( int64 num ) { m_tempTblNumRecords = num; }
    inline void SetTempTblSize( int64 size )      { m_tempTblSize = size; }

    inline void SetEstDataSliceCount ( int num )  { m_estDataSliceCount = num; }
    inline void SetPagesFiltered ( int num )      { m_pagesFiltered = num; }
    // Non-inlines
    CTable *GetTempTblPtr();
    void NeedTempTblRelease();
    void ReleaseTempTbl();

    void LoadTempTable(); //bring temp table into memory

    bool colInProjection( // Return whether passed fieldId part of join projection
        int fdex,         // logical field number in the table
        bool fromInner);  // if true, adjust field for numFields of outer
};

// *** Move following to new xpexprjoin.h ***

class TExprJoinNodeDef {
protected:
    // empty at the moment
};

class TExprJoinNodeRuntimeInfo {
protected:
    // empty at the moment
};

class TExprJoinNode : public TJoinNode,
    private TExprJoinNodeDef, private TExprJoinNodeRuntimeInfo {
public:
    const TableShape *Init(const TableShape *src, EPlanNodeLocus calledFrom);
    virtual ~TExprJoinNode() { /* nothing to do here */ }

    bool IsExprStyle() const { return true; }  // ExprJoin override JoinNode
};

// *** Move following to new xpcrossproduct.h ***

class TCrossProductNodeDef {
protected:
    // empty at the moment
};

class TCrossProductNodeRuntimeInfo {
protected:
    // empty at the moment
};

class TCrossProductNode : public TJoinNode,
    private TCrossProductNodeDef, private TCrossProductNodeRuntimeInfo  {
public:
    TableShape const * Init(TableShape const * src, EPlanNodeLocus calledFrom);
    virtual ~TCrossProductNode() { /* nothing to do here */ }

    bool IsExprStyle() const { return true; }  // CrossProduct override JoinNode
};

THashNode *GetHashNode( //get hash node associated with specified temp table
    int tblId, SnippetNode *pNode);
bool HashNodeIsForDHJ( //returns true if this hash node is for DHJ
    THashNode *pNode );

#endif // !XPJOIN_INCLUDED
