#ifndef XPRECLAIM_INCLUDED
#define XPRECLAIM_INCLUDED

/****************************************************************/
/* FOR CLIENT+DBOS+SPU */
/****************************************************************/

#include "nodecommon.h"
#include "groomcgen.h"
#include "TXIDCommon.h"
#include "groomstmt.h"

#include "error.h"  // for NZ_VERIFY

class CCommBuf;
class CPlan;

enum ReclaimNodeFlags {
    ReclfNone             =  0,      // no reclaim flags
    ReclfTrackNoDeletes   = (1<<0),  // mark extents that have no deleted records
                                     // (groom pages only)
};

class TReclaimClntNode: public GroomNodeBase {
    const TableShape* Init( const TableShape* src, EPlanNodeLocus place ) { NZ_VERIFY( false ); } // dummy to make this class concrete
public:
    TReclaimClntNode(TXidValue stableXid, GroomMode groomMode, uint32 reclaimFlags);

    void Dump(int detail, int indent);
    
    TXidValue stableXid() const { return m_stableXid; }
    GroomMode groomMode() const {return m_groomMode; }
    uint32 reclaimFlags() const { return m_reclaimFlags; }
    bool isPageReclaim() const;

    virtual ~TReclaimClntNode();
    virtual void Serialize(CCommBuf* cbuf);
    virtual void Deserialize(CCommBuf* cbuf);

protected:
    TReclaimClntNode(); // For derived classes constructors from CCommBuf
    
protected:
    TXidValue m_stableXid;
    GroomMode m_groomMode;
    uint32     m_reclaimFlags;
};

inline bool TReclaimClntNode::isPageReclaim() const
{
    switch (groomMode()) {
    case GROOM_PAGES_ALL:
    case GROOM_PAGES_START:
        return true;
    }
    return false;
}

class TReclaimDbosNode: public TReclaimClntNode, public IGroomCGen {
public:

    //! Creates node from communication buffer
    TReclaimDbosNode(CCommBuf* buf, CPlan* plan);
    
    virtual const TableShape* Init(const TableShape *src, EPlanNodeLocus place);

/* IGroomCGen interface ***********************************/

    virtual const TableShape* DeriveSnippetProjection(const TableShape* src);
    virtual bool IsNextLoopStartingNode();
    virtual void EmitNeedNode(CGenCode*  genc);
    virtual TStreamInfo* EmitStreamInfo(CGenCode* genc, int nodeIx, TStreamInfo *flowInfo);
    virtual void EmitLoopHdrStuff(
        CGenCode* genc, TStreamInfo* sP, TStreamInfo* flowP);
    virtual void EmitLoopHdr(CGenCode* genc, TStreamInfo* sP);
    virtual void EmitLoopBodyStuff(
        CGenCode* genc, TStreamInfo* sP, TStreamInfo* flowP);
    virtual void EmitLoopFtr(CGenCode* genc, TStreamInfo* sP, TStreamInfo* flowP);

    virtual void InitOnNewPlan();

    virtual void SpuSnippetSchedCost(
        unsigned& numChannels,  // incremental
        bool& sorter,  // default: false; optionally assigned true; never assign false
        bool& spuDist, // default: false; optionally assigned true; never assign false
        int& readPasses,     // incremental
        int& writePasses,    // incremental
        int& outBlocks,      // default: set in ScanNode; optionally assigned
        int& dest,           // default: outToFabric; optionally assigned
        int& scanBlocks,     // assigned in ScanNode, used by some other nodes
        int& newMem,         // default 0, reset per node; optionally assigned
        int& newInnerRead,   // default 0, reset per node; optionally assigned
        int& newInnerWrite,  // default 0, reset per node; optionally assigned
        bool& looper         // default false; optionally assigned true
        );
/* IGroomCGen interface end ***/
};

/****************************************************************/
#if FOR_SPU
/****************************************************************/

#include "groomspustate.h"
#include "jobintr.h"
#include "plan.h"
#include "storagerefs.h"

class TableShape;

class TReclaimSpuNode: public TReclaimClntNode {
public:

    //! Creates node from communication buffer
    TReclaimSpuNode(CCommBuf* buf, CPlan* plan);

    //! Initilalization
    virtual const TableShape* Init(const TableShape *src, EPlanNodeLocus place);

    GroomSpuState* groomState() const;

    /**
     * Generated code for page-level reclaim calls it
     * before each loop over records.
     */
    void OnLoopStartPR();

    /** 
     * Generated code for record-level reclaim calls it
     * before each loop over records.
     */
    void OnLoopStartRR();

    /** 
     * Generated code for page-level reclaim calls it on last call
     * after looping over all records
     */
    void OnLastCallPR();

    /** 
     * Generated code for record-level reclaim calls it on last call
     * after looping over all records
     */
    void OnLastCallRR();

    /** 
     * Returns true if the record can be reclaimed.
     *
     * Generated code for both record and page level reclaim 
     * calls it for each record inside the loop.
     */
    bool IsRecordReclaimable(TXidValue delXid);

    /** 
     * Record current page as having / not having deleted records.
     * Generated code for groom pages calls it on encountering a non-zero deletexid.
     * Also called from emptyPage().
     */
    void setHasDeletes(bool hasDeletes);

private:

    /**
     * Helper function that unconditionally marks current page empty
     */
    void emptyPage();

    /**
     * Helper function to mark extent as having / not having deleted records
     */
    void logExtentHasDeletes(StorageId sid, bool hasDeletes);

    /** Called when end-of-page is detected for Page-level Reclaim*/
    void handleEopPR();

    /** Called when end-of-extent is detected for Pages All Reclaim*/
    void handleEoxPR();

    /** Called when end-of-page is detected for Record-level Reclaim*/
    void handleEopRR();

    /** Used both by record and page level reclaim on handling end-of-page */
    void updateCounters();

    /** */
    void updateGroomState();

private:

    // Current page; on crossing page border means previous page
    PageId m_pageId;

    // First page seen by the grooming step
    PageId m_firstPageId; 

    // Counter of input (read) records on the current page
    int m_pageRecsIn; 

    // Counter of output (non-reclaimable) records on the current page
    int m_pageRecsOut; 

    // Any deleted records in the current extent
    // (for setting 'noDeletes') - one bit per page.
    uint32 m_hasDeletedRecs;

    // Cumulative number of input (read) records
    int m_recsIn;

    // Cumulative number of output (non-reclaimable) records
    int m_recsOut;

    // Counter of scanned pages
    int m_scannedPages;
    
    // max reclaimed delete xid on page
    TXidValue m_pageMaxReclaimedXid;
    // max reclaimed delete xid overall
    TXidValue m_maxReclaimedXid;
};


inline void TReclaimSpuNode::updateCounters() 
{
    // Update Step cumulative stats
    m_recsIn += m_pageRecsIn; 
    m_recsOut += m_pageRecsOut;
    m_scannedPages++;
    if (m_maxReclaimedXid < m_pageMaxReclaimedXid)
        m_maxReclaimedXid = m_pageMaxReclaimedXid;
    
    // Reset page stats for the next page
    m_pageRecsIn = 0;
    m_pageRecsOut = 0;
    m_pageMaxReclaimedXid = 0;
}

inline void TReclaimSpuNode::updateGroomState() 
{
    GroomSpuState* gs = groomState();
    gs->scannedPages  = m_scannedPages;
    gs->inputRecs     = m_recsIn;
    gs->outputRecs    = m_recsOut;
    gs->firstPageId   = m_firstPageId;
    gs->lastPageId    = m_pageId;
    gs->maxReclaimedXid = m_maxReclaimedXid;

    if (Plan()->i_scanNode()->SlCapturedAll() || Plan()->m_limitReached) 
        gs->completed = true;

    if (groomMode() ==  GROOM_PAGES_ALL) {
        // Remember last processed page to be used as scan start pageid for the next grooming step
        gs->lastReclaimedPage = m_pageId;
    }
}

inline void TReclaimSpuNode::handleEopRR() 
{
    emptyPage();
    updateCounters();
}

inline void TReclaimSpuNode::handleEopPR()
{
    if (m_pageRecsOut == 0) {
        emptyPage();           // All the records on fully scanned page are reclaimed
    }
    else {
        m_pageRecsOut = m_pageRecsIn;  // None of the records on the page are reclaimed
        m_pageMaxReclaimedXid = 0;
        // Is this the first non-empty page of the extent with deleted records?
        int bitPos = m_pageId.pageBits() - 1;
        NZ_ASSERT(bitPos >= 0 && bitPos < PAGES_PER_EXTENT);
        if ((m_hasDeletedRecs & 1<<bitPos)
            && !(m_hasDeletedRecs & ~(1<<bitPos))) {
            logExtentHasDeletes(m_pageId, true);
        }
    }
    updateCounters();
}

inline void TReclaimSpuNode::handleEoxPR()
{
    // Are we tracking 'noDeletes' on extents?
    if ((m_reclaimFlags & ReclfTrackNoDeletes)
        // If m_hasDeletedRecs is still zero, all non-empty pages in the extent
        // have been scanned and no deleted records found.
        // Note that if there are deleted records, logExtentHasDeletes(true) has already
        // been called.
        && m_hasDeletedRecs == 0) {
        logExtentHasDeletes(m_pageId, false);
    }
    m_hasDeletedRecs = 0;
}

inline void TReclaimSpuNode::setHasDeletes(bool hasDeletes)
{
    int bitPos = m_pageId.pageBits() - 1;
    NZ_ASSERT(bitPos >= 0 && bitPos < PAGES_PER_EXTENT);
    if (hasDeletes) {
        m_hasDeletedRecs |= 1<<bitPos;
    }
    else {
        m_hasDeletedRecs &= ~(1<<bitPos);
    }
}

inline void TReclaimSpuNode::OnLoopStartRR() 
{
    // Initialization on first input buffer
    if (m_pageId == NULL_PID) {
        m_pageId = m_firstPageId = Plan()->m_curTablePageId;
    }
    // Check for and handle page border condition
    if (m_pageId != Plan()->m_curTablePageId) {
        handleEopRR();
        m_pageId = Plan()->m_curTablePageId;
    }
}

inline void TReclaimSpuNode::OnLoopStartPR() 
{
    // Initialization on first input buffer
    if (m_pageId == NULL_PID) {
        m_pageId = m_firstPageId = Plan()->m_curTablePageId;
    }
    // Check for and handle page border condition
    // Even if we set limitReached [used for PAGES START] on the previous gen'd code call, 
    // we still can be called again [with lastCall == 0]
    // with a new page/buffer until CBS drains what it already scanned into buffers
    if (m_pageId != Plan()->m_curTablePageId && !Plan()->m_limitReached) {
        handleEopPR();
        if (m_pageId.extentDiffers(Plan()->m_curTablePageId)) {
            handleEoxPR();
        }
        m_pageId = Plan()->m_curTablePageId;
    }
}

inline void TReclaimSpuNode::OnLastCallRR() 
{
    if (m_pageId != NULL_PID) {
        handleEopRR();
    }
    updateGroomState();
}

inline void TReclaimSpuNode::OnLastCallPR() 
{
    if (m_pageId != NULL_PID) {
        handleEopPR();
        // For GROOM_PAGES_START we don't know if we've seen all records in the extent.
        if (groomMode() == GROOM_PAGES_ALL) {
            handleEoxPR();
        }
    }
    updateGroomState();
}

inline bool TReclaimSpuNode::IsRecordReclaimable(TXidValue delXid) 
{
    // Gen'd code calls this method at most once for each record.
    // For groom pages modes we break out of loop over buffered records 
    // as soon as we see the first non-reclaimable record; 
    // So for those modes m_pageRecsIn and m_pageRecsOut can be less than the actuall number 
    // of input and output records. Groom step aggregates of these counters go into
    // _vt_groom_step; so care must be taken when interpreting them for groom_pages modes.
    // For statistics we are only interested in the number of purged
    // records (purged_records == records_in - records_out), which is still
    // valid for all groom modes
    m_pageRecsIn++; 

    if (delXid && delXid < stableXid()) {
        if (delXid > m_pageMaxReclaimedXid)
            m_pageMaxReclaimedXid = delXid;
        return true;
    }
    else {
        m_pageRecsOut++;  
        return false;
    }
}

/****************************************************************/
#endif // FOR_SPU
/****************************************************************/


#endif // XPRECLAIM_INCLUDED
