#ifndef XPREORGPREP_INCLUDED
#define XPREORGPREP_INCLUDED

/****************************************************************/
/* FOR CLIENT+DBOS+SPU */
/****************************************************************/

#include "nodecommon.h"
#include "groomcgen.h"
#include "TXIDCommon.h"
#include "storagerefs.h"
#include "plan.h" // for groomState()

#include "error.h"  // for NZ_VERIFY

class CCommBuf;
class CPlan;
class TableShape;

//! Enumerates available key normalization algorithms
enum KeyNormAlg {
    ProportionalNormAlg,
    SortNormAlg,
    SortRankNormAlg,
    EquiBandsNormAlg
};

enum {
    //! Max supported number of reorg keys
    MAX_REORG_KEYS_CNT = 4
};


class TReorgPrepClntNode: public GroomNodeBase {
    const TableShape* Init( const TableShape* src, EPlanNodeLocus place ) { NZ_VERIFY( false ); } // dummy to make this class concrete
public:

    TReorgPrepClntNode(
        int nKeys,
        int tempTableId,
        KeyNormAlg* normAlgs,
        int* normParams,
        int64 maxSpuMem);

    void Dump(int detail, int indent);

    // Gets/sets stable txid
    void StableTxid(TXidValue txid);
    TXidValue StableTxid();

    /** 
     * Returns max number of records that will be processed per grooming step.
     * keyCols is a pointer to array of at least nKeys elements where
     * each element is a 0-based logical column index. 
     * If keyCols is NULL, use the table's first nKeys columns as key columns.
     */
    static int calcMaxRecs(int64 maxMem, const TableShape* table, int nKeys, const int* keyCols);

    virtual ~TReorgPrepClntNode();
    virtual void Serialize(CCommBuf* cbuf);
    virtual void Deserialize(CCommBuf* cbuf);

protected:
    TReorgPrepClntNode(); // For derived classes constructors from CCommBuf

protected:
    //! Number of reorg keys
    int m_nKeys;
    //! Key normalization algorithms to use for each key
    KeyNormAlg m_normAlgs[MAX_REORG_KEYS_CNT];
    //! Parameters used by key normalization algorithms
    unsigned m_normParams[MAX_REORG_KEYS_CNT];
    //! Transient table id (to pass data from DotPrep to DotReorg node)
    int m_tempTableId;
    //! Max amount of memory (in bytes) we can use on SPU for reorg process
    int64 m_maxMem;
    //! Stable transaction id
    TXidValue m_stableTxid;
};


inline void TReorgPrepClntNode::StableTxid(TXidValue txid) 
{
    m_stableTxid = txid;
}

inline TXidValue TReorgPrepClntNode::StableTxid() 
{
    return m_stableTxid;
}


class TReorgPrepCommonNode: public TReorgPrepClntNode {
public:
//    TReorgPrepCommonNode(CCommBuf* buf, CPlan* plan) {};
    /** Node initialization before node excecution. */
    virtual const TableShape* Init(const TableShape *src, EPlanNodeLocus place) = 0;
    
};



/****************************************************************/
/* end FOR CLIENT+DBOS+SPU */
/****************************************************************/


/****************************************************************/
#if FOR_DBOS
/****************************************************************/

class TReorgPrepDbosNode: public TReorgPrepCommonNode, public IGroomCGen {
public:

    //! Creates node from communication buffer
    TReorgPrepDbosNode(CCommBuf* buf, CPlan* plan);
    
    virtual const TableShape* Init(const TableShape *src, EPlanNodeLocus place);

/* IGroomCGen interface ***********************************/

    virtual const TableShape* DeriveSnippetProjection(const TableShape* src);
    virtual bool IsNextLoopStartingNode();
    virtual void EmitNeedNode(CGenCode*  genc);
    virtual TStreamInfo* EmitStreamInfo(CGenCode* genc, int nodeIx, TStreamInfo *flowInfo);
    virtual void EmitLoopHdrStuff(
        CGenCode* genc, TStreamInfo* sP, TStreamInfo* flowP);
    virtual void EmitLoopHdr(CGenCode* genc, TStreamInfo* sP);
    virtual void EmitLoopBodyStuff(
        CGenCode* genc, TStreamInfo* sP, TStreamInfo* flowP);
    virtual void EmitLoopFtr(CGenCode* genc, TStreamInfo* sP, TStreamInfo* flowP);

    virtual void InitOnNewPlan();

    virtual void SpuSnippetSchedCost(
        unsigned& numChannels,  // incremental
        bool& sorter,  // default: false; optionally assigned true; never assign false
        bool& spuDist, // default: false; optionally assigned true; never assign false
        int& readPasses,     // incremental
        int& writePasses,    // incremental
        int& outBlocks,      // default: set in ScanNode; optionally assigned
        int& dest,           // default: outToFabric; optionally assigned
        int& scanBlocks,     // assigned in ScanNode, used by some other nodes
        int& newMem,         // default 0, reset per node; optionally assigned
        int& newInnerRead,   // default 0, reset per node; optionally assigned
        int& newInnerWrite,  // default 0, reset per node; optionally assigned
        bool& looper         // default false; optionally assinged true
        );
/* IGroomCGen interface end ***/
};

/****************************************************************/
#endif // FOR_DBOS
/****************************************************************/



/****************************************************************/
#if FOR_SPU
/****************************************************************/

struct GroomSpuState;

class TableShape;

/** Key normalization functions are responsible for convertion/normalization of
    array of keys of some data type into array of ints.
    This function frees keys memory and sets keys pointer to 0 and 
    allocates normKeys memory. This is why they are passed by reference.
    Record reorganization algorithms work with these normalized keys.
*/
typedef void KeyNormFunc(void*& keys, unsigned*& normKeys, 
                         int elemCnt, unsigned maxNormKeyValue);

class TReorgPrepSpuNode: public TReorgPrepCommonNode {
public:

    //! Creates node from communication buffer
    TReorgPrepSpuNode(CCommBuf* buf, CPlan* plan);
    virtual ~TReorgPrepSpuNode();

    //! Initilalization
    virtual const TableShape* Init(const TableShape *src, EPlanNodeLocus place);

    /** Stores key value for current record.
     *  Gen'd code must call this method for each key before calling AddRecord()
     *  Make sure this method is inlined! 
     */
    template <typename T>
    void AddKey(int keyIdx, bool isNull, T keyValue);

    /** Returns false when we do not have enough space.
     *  Gened code must call it after it added all keys for the record.
     *  When this method returns false, gened code must exit the scan.
    */
    bool AddRecord();

    /** Checks if we have started new extent.
     *  Called by gened code before scan loop
     */
    void CheckExtentBoundary();

    /**
     *  Gened code calls it when it encounters a reclaimable record
     */
    void OnReclaimedRec(TXidValue deleteXid);

    /**
     *  Gened code calls it on last call of first snippet
     */
    void OnLastCall();

private:

    /** Prepares and stores organized indexes.
     *  Called from OnLastCall()
     */
    void PrepareOrgIxs();

    //! Applies appropriate key normalization function to buffered key values
    void normalizeKeys();

    //! Initializes for individual reorg key
    template <typename T>  void initKeyField(int keyIx);

    GroomSpuState* groomState() const {
        return Plan()->groomState();
    }

private:
    //! Max number of records we can buffer. Calculated from m_maxMem.
    int m_maxRecs;
    //! Owner Id for deleting key and norm values
    nzotuid_t m_owner;
    //! NzTypes of the keys
    int m_keyTypes[MAX_REORG_KEYS_CNT];
    /** Buffers for each key. Actual data  will be consistent with m_keyTypes. */
    void* m_keyValues[MAX_REORG_KEYS_CNT];
    //! Normalized key values
    unsigned int* m_normKeys[MAX_REORG_KEYS_CNT];
    //! Key normalization functions to use for each key normalization
    KeyNormFunc* m_keyNormFuncs[MAX_REORG_KEYS_CNT];
    //! Current number of records in the buffers
    int m_recCnt; 
    //! Number of records in fullly scanned extents
    int m_extRecCnt;
    //! Current page id: used to track extent boundaries
    PageId m_curPageId;
    //! Number of full input extents to reorganize
    unsigned m_extCount;
    /** Transient table to store and pass data between DotPrep and DotReorg
     *  nodes across snippets. 
     *  Created and filled by ReorgPrepNode and read and destroyed by ReorgMergeNode.
     */
    CTable* m_tempTable;

    /** Used to determine if we have accumulated enough records to convert 
     *  full written extents to "done"
     */
    double m_groomAutoDoneRatio;

    //////////////// FOR STATS
    //! Current number of reclaimable records
    int m_reclaimedRecs;
    //! Number of reclaimable records in fully scanned extents
    int m_extReclaimedRecs;
    //! First page id of the grooming step
    PageId m_firstPageId;
    //! Last page id of the grooming step
    PageId m_lastPageId;
    //! Counter of scanned pages
    int m_scannedPages;
    //! Snippet 1 start time
    int64 m_startTime;
    // Incremental backup/restore needs it to operate correctly
    TXidValue m_maxReclaimedXid;
    //////////////// END FOR STATS
};


inline void TReorgPrepSpuNode::OnReclaimedRec(TXidValue deleteXid) 
{
    m_reclaimedRecs++;
    if (deleteXid > m_maxReclaimedXid)
        m_maxReclaimedXid = deleteXid;
}


template <typename T>
inline void TReorgPrepSpuNode::AddKey(int keyIdx, bool isNull, T keyValue) 
{
    NZ_VERIFY(m_recCnt < m_maxRecs);

    T *p = (T*)m_keyValues[keyIdx];
    if (isNull)
        p[m_recCnt] = T();
    else
        p[m_recCnt] = keyValue;
}

inline bool TReorgPrepSpuNode::AddRecord() 
{
    NZ_VERIFY(m_recCnt < m_maxRecs);

    // Gen'd code must exit scan when we return false
    return (++m_recCnt < m_maxRecs);
}

/****************************************************************/
#endif // FOR_SPU
/****************************************************************/


#endif // XPREORGPREP_INCLUDED
