#ifndef XPSCAN_INCLUDED
#define XPSCAN_INCLUDED


#include "nodecommon.h"
#include "nzrecord.h"
#include "storagedefs.h"
#include "extdropxidmap.h"

#include "storagerefs.h"
namespace zm { class Column; };

struct TZoneMapUnion;
class IvList;
struct fpgaPortablePacket;
class TKeyValue;
class TXid;

// for decoding m_scanFlags
#define SfNone          (0)     // no scan flag
#define SfJoinStream    (1<<0)  // this is last stream into joinlist
#define SfOrdered       (1<<1)  // scan of an ordered table on disk
#define SfFabric        (1<<2)  // fabric scan
#define SfForModify     (1<<3)  // scanned table being inserted/updated into
#define SfFabricImmed   (1<<4)  // process fabric blocks as they arrive
#define SfAllRecords    (1<<5)  // show deleted and undeleted records (reclaim, stats)
#define SfGenStats      (1<<6)  // Generate Statistics scan
#define SfUNUSED7       (1<<7)  // Available
#define SfIndirect      (1<<8)  // block content points to recordstore (loader)
#define SfRawScan       (1<<9)  // use raw reads (because table has blobs, for now)
#define SfPrepScan      (1<<10) // Scan node hijacked for prep snippet
#define SfForNewStuff   (1<<11) // Table being scanned beginning at watermark
#define SfMrUnsafe      (1<<12) // Micro-regen is unsafe on errors in this scan
#define SfMrWithBias    (1<<13) // Micro-regen is safe but needs bias adjustment
#define SfNeedSkipSpace (1<<14) // gened code requires skip space before first record
#define SfNeedsRescan   (1<<15) // Param resulted in need to reevaluate zone-map
#define SfRedoNew       (1<<16) // Requests scan siRedoNew (used by groom)
#define SfUNUSED17      (1<<17) // Available
#define SfDeletedRecsOnly    (1<<18) // Show only deleted records (incremental backup, replication).
                                     // Use in conjunction with SfAllRecords.
#define SfForceSwTxFilter    (1<<19) // Apply invisibility list in software even when FPGAs ignore it
#define SfSkipDroppedExtents (1<<20) // Don't scan truncated extents even if "visible"

// Buffer for Scan operation
// Improve system performance by usually scanning an extent at a time.
#define FPGA_BUFFER_SIZE 1024*1024*3

enum EScanType {
    eScanTypeOrderedOnly,
    eScanTypeNewStuffOnly,
    eScanTypeBoth
};

enum {
    eBlockCostFromPgPlanner = 'o',  // postgres planner/optimizer
    eBlockCostFromZoneMap = 'z',    // zone map via prep scan
    eBlockCostFromSampleScan = 'j',  // JIT-Stats sample scan,
    eBlockCostFromDROCorrection = 'r'  // corrected estimate based on DRO
};

enum EVtMtlzScope {
    eVtMtlzBlade,
    eVtMtlzDataslice
};

class TScanNodeDef {
protected:
    EPlanNodeTableType tableType; 
    table_id_t         tblId;        // table 'name'
    int                m_scanFlags;  // hints - SfJoinStream, etc.
    int                numKeyValues; // number of elements in pKeyValues
    TKeyValue        * pKeyValues;   // pointer to 1st of key values for index scan
    int                spuId;        // ALL_SPUS_LID or specific spu id
    EScanType          scanType;     // scan ordered stuff ,new stuff or both
    int                m_totalXIDsCount;
    int                m_invisibleXIDsCount;
    int                m_XIDMode;    // actually fpga::EXIDMode, but avoids an include
    TXid             * m_pFpgaXidList;
    table_id_t         m_relOrigOid; // original relation OID; still referred to by later versions
};

class TScanNodeRuntimeInfo {
protected:
    fpgaPortablePacket * fpgaMicrocode;
    TZoneMapUnion      * pZoneMapInfo;         // pointer to start of zone-map info tree
    int                  m_blockCost;          // cost in blocks (via stats? zone map?)
    int                  m_blockCostSource;    // eBlockCostXXX
    int                  m_estDataSliceCount; // Estimated number of data slices targeted. Zero if it can't be easily determined.

    table_id_t           m_indexTable;         // table id holding blockList (or 0)

    // for _vt_plan_scan
    int totalSectors;
    int tblSectors;

    // Watermarks can be used as a starting point (low) or an ending point (high).
    // In either case, the watermark is the end of the table at the time the plan
    // starts to run.  While the watermark members are essentially the same, they
    // have been created for clarity.
    TRecLoc m_highWater;        // highest rowAddr present before this plan
    TRecLoc m_lowWater;         // highest rowAddr present before this plan

    CTable          * scanTable;  // table being scanned
    TableShape      * fpgaResultTable;  // projected table out of fpga
    size_t            m_skipSpace;    // amount of buffer (at head) to skip before IO
    TUseSpecialFields m_extraSpecials;  // specials needed due to restricts not handled in FPGA
    void            * m_vt_materializer; // dynamic virtual materializer function pointer
    EVtMtlzScope      m_vt_scope;

    // Sample scanning
    double m_sampleGrowthPower;  // growth power for sample scan; default is zero (no sampling)
    int    m_sampleSize;         // total number of extents to sample scan.
    int    m_scanSize;           // total number of extents in scan before sampling
    int    m_scanSizeBJ;         // total number of extents in scan before sampling and join time zmaps

    int    m_scanTableBias;      // table block number bias

    // Parameters controlling scan list setup (see SpuSetupSectorList)
    StorageId m_slStartExtent; // Scan list will include only _later_ (greater extentid) extents
    StorageId m_slLastExtent;  // Scan list will include only _prior_ (lesser extentid)  extent
    unsigned  m_slMaxExtents;  // Limit scan list to this many extents
    bool      m_slCapturedAll; // Indicates if scan list reached end of table or was truncated

    ExtRangeDropXidMap m_extentsWithDropXid;
};

class TScanNode : public SnippetNode,
    private TScanNodeDef, private TScanNodeRuntimeInfo { // scan (the start of any select)

public:
    TScanNode();
    virtual ~TScanNode();
    
    TableShape const * Init( TableShape const * src, EPlanNodeLocus place );
// dlm-check    virtual ~TScanNode() { /* nothing to do here */ }
    bool IsValidForSrc() { return true; }
    bool Serialize( CCommBuf * cbuf, ESerializeMode mode );
    void Deserialize( CCommBuf const * cbuf, ESerializeMode mode );

    // Address-of accessors
    // Used for things like virtual-table materialization.
    inline double*     AddrOfSampleGrowthPower() { return &m_sampleGrowthPower; }
    inline table_id_t* AddrOfTblId()             { return &tblId; }
    inline int*        AddrOfTblSectors()        { return &tblSectors; }
    inline int*        AddrOfTotalSectors()      { return &totalSectors; }

    // Accessors
    inline int         BlockCost()              const { return m_blockCost; }
    inline int         BlockCostSource()        const { return m_blockCostSource; }
    inline int         EstDataSliceCount()      const { return m_estDataSliceCount; }
    inline TUseSpecialFields ExtraSpecials()    const { return m_extraSpecials; }
    inline TableShape* FpgaResultTable()        const { return fpgaResultTable; }
    inline TXid*       FpgaXidList()            const { return m_pFpgaXidList; }
    inline TRecLoc     HighWater()              const { return m_highWater; }
    inline table_id_t  IndexTable()             const { return m_indexTable; }
    inline int         InvisibleXIDsCount()     const { return m_invisibleXIDsCount; }
    inline bool        IsFabricScan()           const { return ( m_scanFlags & SfFabric ) != 0; }
    inline bool        IsScanNewStuffOnly()     const { return scanType == eScanTypeNewStuffOnly; }
    inline bool        IsScanOrderedOnly()      const { return scanType == eScanTypeOrderedOnly; }
    inline TKeyValue*  KeyValues()              const { return pKeyValues; }
    inline TRecLoc     LowWater()               const { return m_lowWater; }
    inline fpgaPortablePacket * FpgaMicrocode() const { return fpgaMicrocode; }
    inline int         NumKeyValues()           const { return numKeyValues; }
    inline table_id_t  RelOrigOid()             const { return m_relOrigOid; }
    inline double      SampleGrowthPower()      const { return m_sampleGrowthPower; }
    inline int         SampleSize()             const { return m_sampleSize; }
    inline int         ScanFlags()              const { return m_scanFlags; }
    inline int         ScanSize()               const { return m_scanSize; }
    inline int         ScanSizeBJ()             const { return m_scanSizeBJ; }
    inline CTable*     ScanTable()              const { return scanTable; }
    inline EScanType   ScanType()               const { return scanType; }
    inline size_t      SkipSpace()              const { return m_skipSpace; }
    inline bool        SlCapturedAll()          const { return m_slCapturedAll; }
    inline unsigned    SlMaxExtents()           const { return m_slMaxExtents; }
    inline StorageId   SlStartExtent()          const { return m_slStartExtent; }
    inline StorageId   SlLastExtent()           const { return m_slLastExtent; }
    inline int                SpuId()           const { return spuId; }
    inline EPlanNodeTableType TableType()       const { return tableType; }
    inline table_id_t         TblId()           const { return tblId; }
    inline int                TblSectors()      const { return tblSectors; }
    inline int                TotalSectors()    const { return totalSectors; }
    inline int                TotalXIDsCount()  const { return m_totalXIDsCount; }
    inline int                XIDMode()         const { return m_XIDMode; }
    inline void*              Vt_materializer() const { return m_vt_materializer; }
    inline EVtMtlzScope       Vt_scope()        const { return m_vt_scope; }
    inline TZoneMapUnion    * ZoneMapInfo()     const { return pZoneMapInfo; }
    inline bool         ForceSwTxFilter()       const { return ( m_scanFlags & SfForceSwTxFilter ) != 0; }

    // Mutators
    inline void AddScanFlag( int flag )           { m_scanFlags |= flag; }
    inline void RemoveScanFlag( int flag )        { m_scanFlags &= ~flag; }
    inline void IncTotalSectors( int val )        { totalSectors += val;}

    // Setters
    inline void SetBlockCost( int cost)                     { m_blockCost = cost; }
    inline void SetBlockCostSource( int src )               { m_blockCostSource = src; }
    inline void SetEstDataSliceCount( int num )             { m_estDataSliceCount = num; }
    inline void SetExtraSpecials( TUseSpecialFields ss )    { m_extraSpecials = ss; }
    inline void SetFpgaMicrocode( fpgaPortablePacket* pkt ) { fpgaMicrocode = pkt; }
    inline void SetFpgaResultTable( TableShape* tbl )       { fpgaResultTable = tbl; }
    inline void SetFpgaXidList( TXid* list )         { m_pFpgaXidList = list; }
    inline void SetHighWater( TRecLoc mark )         { m_highWater = mark; }
    inline void SetIndexTable( table_id_t tbl )      { m_indexTable = tbl; }
    inline void SetInvisibleXIDsCount( int count)    { m_invisibleXIDsCount = count; }
    inline void SetKeyValues( TKeyValue* keys )      { pKeyValues = keys; }
    inline void SetLowWater( TRecLoc mark )          { m_lowWater = mark; }
    inline void SetNumKeyValues( int num )           { numKeyValues = num; }
    inline void SetRelOrigOid( table_id_t tbl )      { m_relOrigOid = tbl; }
    inline void SetSampleGrowthPower( double pwr )   { m_sampleGrowthPower = pwr; }
    inline int  SetSampleSize( int val )             { m_sampleSize = val; return val;}
    inline void SetScanFlags( int flags )            { m_scanFlags = flags; }
    inline int  SetScanSize( int val )               { m_scanSize = val; return val; }
    inline void SetScanSizeBJ( int val )             { m_scanSizeBJ = val; }
    inline void SetScanType( EScanType type )        { scanType = type; }
    inline void SetScanTable( CTable* tbl )          { scanTable = tbl; }
    inline void SetSlCapturedAll( bool bb )          { m_slCapturedAll = bb; }
    inline void SetSlStartExtent( StorageId sid )    { m_slStartExtent = sid; }
    inline void SetSlLastExtent( StorageId sid )     { m_slLastExtent = sid; }
    inline void SetSlMaxExtents( int maxExt )        { m_slMaxExtents = maxExt; }
    inline void SetSpuId( int id )                      { spuId = id; }
    inline void SetTableType( EPlanNodeTableType type ) { tableType = type; }
    inline void SetTblId( table_id_t tbl )              { tblId = tbl; }
    inline void SetTblSectors( int val )                { tblSectors = val; }
    inline void SetTotalSectors( int val )              { totalSectors = val; }
    inline void SetTotalXIDsCount( int count )          { m_totalXIDsCount = count; }
    inline void SetXIDMode( int mode )                  { m_XIDMode = mode; }
    inline void SetZoneMapInfo( TZoneMapUnion* zmu )    { pZoneMapInfo = zmu; }
    inline void Set_vt_materializer( void* vtm )        { m_vt_materializer = vtm; }
    inline void Set_vt_scope( EVtMtlzScope vts )        { m_vt_scope = vts; }
    inline void clearExtentsWithDropXid()               { m_extentsWithDropXid.clear(); }

    // Non-inlines
    void SetVisibility( const TXid& selfTxId, const IvList& list); // set m_XIDMode, and then visibility list if applicable
    int ScanReadBlocks() const;
    bool hasDroppedExtents() const;
    void InsertExtRangeDropXidPair( RawSid minExtentId, RawSid maxExtentId, uint64 txid );
    uint64 LookupDropXid( RawSid eid ) const;
};


class TScanPrepNodeRuntimeInfo {
protected:
    TScanNode* m_pScanNode;      // Scan node this is prep for
};

class TScanPrepNode : public SnippetNode,
                      private TScanPrepNodeRuntimeInfo { // scan (the start of any select)
 public:
     virtual ~TScanPrepNode() { /* nothing to do here */ }

     // Accessors:
     inline TScanNode* ScanNode()  const { return m_pScanNode; }

     inline void SetScanNode( TScanNode* val ) { m_pScanNode = val; }
};

#endif // !XPSCAN_INCLUDED
