#ifndef XPSORT_INCLUDED
#define XPSORT_INCLUDED

#include "nodecommon.h"

typedef int (*TCmpRecFn)(void *, record_t**, record_t**); //from recordcompare.h
class CTable;
class CRecordStore;
class CMergePlan;
class CComparison;
struct TMemblockInfo;
struct CMemForSorting;

enum EMergePhase {        // Markers in TMergeSeq nodes, "Node contains...
    PHASE_PRELIM = 0,     // ... an initialization-time place-holder (dummy)
    PHASE_RECORD,         //     a real record 
    PHASE_EXHAUSTED       //     an end-of-data dummy"
};

struct TMergeSeq {
    CRecordStore * seqStore; // accessor to stream of blocks
    int seqLimit;         // limit block just after sequence
    int seqCurrent;       // block being worked on from sequence
    TMemblockInfo *curBlockP; // pointer to current record's block
    record_t* curRecP;    // current record pointer in current block
    int loserSeq;         // (tree internal) sequence of loser record
    int loserPhase;       // (tree internal) phase of loser
};

// Empirical parameters for sort setup.
enum {HOST_MAX_SORT_BLOCKS = (512*8) }; // 512MB * 108 is a LOT of data

// Specify output target of a merge step
enum EMergeOutType { OUT_TO_RECORDSTORE=0, OUT_TO_STREAM };

// Management of sorting one table: can have sorting of transient tables
// involved in GROUP BY and DISTINCT processing as well as plain old ORDER BY
class CSorter {
    friend class CHash;
PRIVATE:
    CTable *m_table;      // materializing work area (alloced by caller)
    CRecordStore *m_RSt;  // RecordStore of m_table
    CPlan *m_plan;        // plan containing this sorter
    TCmpRecFn m_cmpRecFn; // record compare function for merge phase
    void *m_cmpRecFnData; // 'control data' arg for cmpRecFn
    bool m_diskBased;     // true if didn't fit in memory
    CComparison *m_sortCompare; // record compare for sort phase

    // state for disk-based sorting & merging
    int m_numDataPages;   // count of 128KB data pages in materialization phase (incl hashInfo)
    int m_numPRecPages;   // number of 128KB pages used in m_pRecords (estimate)
    int m_numBucketPages; // size of hash bucket pointers (in 128KB pages) (zero?)
    int m_memReleaseMinSort;

    int m_lastBlocksRecords; // worst case num records in one data page
    int m_numSections;    // count of sorted sections in swap
    int m_allocSections;  // allocated pointers for sections
    CRecordStore **m_sections; // list of sorted sections
    CRecordStore *m_outputSection;  // collects sorted output for re-use
    int m_numSeqBlocks;   // for _vt_plan_<node> Merge phase read prefetch amount
    int m_bestSeq;        // continuing merge tree state scalars
    int m_bestPhase;

    int m_maxNumSeq;      // max number of sorted sequences (alloc size of m_seq array)
    int m_mergeDimension; // number of sequences being merged by this setup
    TMergeSeq *m_seq;     // array of structs describing sorted sections to be merged
    TMergeSeq *m_seqPrevRec; // seq containing returned record's block to release
                          // ("delay line" so Unique or merge-style Aggr
                          // gets one last peek at "previously selected record")
    CMemForSorting* m_memForP; // Memory Source for this sort
    CSorter* m_nextSI;    // Link to next managed table sort
    int m_sortNumSections;// numSections before intermediate merging

    int m_progressLevels;
    int m_progressPages;
    int m_progressCount;  // How many pages we have read from sort sections
    int m_progressTotal;  // Total number of pages in sort sections

public:

    void SortInit(
        CMemForSorting& memFor,
        int finalID,
        CPlan *plan,
        CTable *tbl,
        CComparison *comparer,
        int nHashFrameworkBlocks);

    void becomeDiskBased(); // Set up swap table and
    void mergeCapacity(int finalPages,
                       int limitPages,
                       int mergeLevel,
                       int *pMergeWays,
                       int *pNextLevelCapacity,
                       int *pMoreLevels);

    void addSection(); // Put current sorted table into swaptable as section

    void setWideMerge(CComparison *comparer, CTable *table); // shape-changer for DISTINCTer use
    void setupMergeArrays(int limitPages,     // Set up merger trains for disk-based joins (bz)
                          int firstSection,
                          int numSections);

    void IntermediateMerging(int limitPages, EMergeOutType final);

    void doPartialMerge(int outPart,
                        int firstSection,
                        int numSections,
                        int memLimit,
                        int mergeLevel);

    record_t *mergeStep(record_t **priorP);   // Run the merger to get the next rec

    record_t *getNextFromSection( // Fetch the next record of this section (for merge engine)
            TMergeSeq &section,record_t **priorP);

    void setCmpRecFnForSort( // install the generated compare records functions
            TCmpRecFn cmpRecFn, CTable *tbl);

    void qsortTable(); // sort the current state of the table

    void sendNodeMemUpdate(int nid);

    void sortSection(  // Notifcation of full sort store
            bool lastCall);

    bool overThreshold(int nPages, int nHashTablePages);

    void cleanUp(); // Free storage

    bool DiskBased () const {
        return m_diskBased;
    }

    CRecordStore * createOutputSection(const TableShape *pattern);
    void newInitSection(CRecordStore *newRS, int sourceID);
    void deleteSectionStores(int firstStore, int limitStore);
    enum {SECTIONS_GROWTH = 250 };

    // manipulate m_sections
    void packSections(int outCount, int levelDone);
    void pivotTail(int tail);
}; // CSorter

// Segregates memory management data for the collection of CSorters
// used by a TAggregateNode (a modest number of sorts) or TSortNode (one sort)
class CMemForSorting {
public:
    CSorter* m_firstSI;          // Singly linked list of sorts to get memory
    int m_nAvailPagesForMerge;   //blocks avail for merge phase (by DBOS)
    int m_nAvailPagesForCollect; // blocks avail to collect records (by DBOS)
    int m_nSpuPagesForMerge;     // SPU's guess at pages for merge
    int m_nSpuPagesForCollect;   // SPU's guess at pages for collecting records
    int m_finalPagesForMerge;    // merge pages after testing-point reduction
    int m_finalPagesForCollect;  // collecting pages after testing-point reduction
    int m_nodePagesLimit;        // for _vt_plan_<node>
    int m_maxNumPages;           // for _vt_plan_<node>
    int m_allSections;           // for _vt_plan_<node>
    bool m_anyDiskBased;         // for _vt_plan_<node>
    EMergeOutType m_outputTo;

    void InitWithTestLimit(const char *cfgParam); // and in-lieu of real _ctor_
    void ValidateTestLimit(const char *whichNode);
    int  CountPagesUsed(EMergeOutType outTo);   // pages subscribed to in CSorter list
    int  CountPagesUsed(void) { return CountPagesUsed(m_outputTo); }
    void SetOutType(EMergeOutType newTo) { m_outputTo = newTo; }
};

class TSortNodeDef {
protected:
    int   m_numFields;
    int   m_topN;        // if > 0 then a tournament sort   
    int   m_mergeType;   // FALSE complete table sort (inSPU sort must be this)
                         // TRUE merge SPU return sets (inDBOS in tandem with inSPU)
    int * m_fieldsList;
    int * m_ascendingList;
    int * m_nullsLoList;
};

class TSortNode;

class TSortNodeRuntimeInfo {
protected:
    // Support for generated code's compare logic
//public:
//    int            m_numMergeReturnRec;
//    record_t    ** m_mergeReturnArray;
//protected:
    nzbool         m_isTie;    // if tie on last node for topN
    CComparison  * m_comparer;
    CMergePlan   * m_mergePlan;  // for m_mergeType TRUE operation FOR_DBOS
    bool           m_bInitialized; //true if sortInfo object is initialized
    CMemForSorting memFor;
    CSorter        m_sortInfo;  // main engine of the hybrid quick/merge sorter
public:
    nzbool isTie();
    void   SetIsTie( nzbool isTie );
    void   SetTopN( int topN );
};

// Need these to be public for MergePlan->Step call in xpsort.cpp.
// Step takes its last two args as references.
struct TSortNodeRuntimeInfoWart {
    int            m_numMergeReturnRec;
    record_t    ** m_mergeReturnArray;
};

class  TSortNode : public SnippetNode,
    private TSortNodeDef, private TSortNodeRuntimeInfo,
    /*private CMemForSorting,*/ public TSortNodeRuntimeInfoWart {
public:
    TSortNode();
    virtual ~TSortNode();

    TableShape const * Init(TableShape const * src, EPlanNodeLocus calledFrom);
// dlm-check    virtual ~TSortNode() { /* nothing to do here */ }

    bool IsValidForSrc() { return true; }
    bool Serialize( CCommBuf * cbuf, ESerializeMode mode );
    void Deserialize( CCommBuf const * cbuf, ESerializeMode mode );

    // Accessors
    inline CMemForSorting*  AddrOfMemFor()            { return &memFor; }
    inline CSorter*         AddrOfSortInfo()          { return &m_sortInfo; }

    inline int*             AscendingList()     const { return m_ascendingList; }
    inline CComparison*     Comparer()          const { return m_comparer; }
    inline bool             DiskBased()         const { return m_sortInfo.DiskBased(); }
    inline int*             FieldsList()        const { return m_fieldsList; }
    inline bool             Initialized()       const { return m_bInitialized; }
    inline nzbool           IsTie()             const { return m_isTie; }
    inline CMemForSorting & MemFor()                  { return memFor; }
    inline CMergePlan*      MergePlan()         const { return m_mergePlan; }
    inline record_t**       MergeReturnArray()        { return m_mergeReturnArray; }
    inline int              MergeType()         const { return m_mergeType; }
    inline int*             NullsLoList()       const { return m_nullsLoList; }
    inline int              NumFields()         const { return m_numFields; }
    inline int              NumMergeReturnRec() const { return m_numMergeReturnRec; }
    inline CSorter &        SortInfo()                { return m_sortInfo; }
    inline int              TopN()              const { return m_topN; }
//    inline record_t **& MergeReturnArray()     { return m_mergeReturnArray; }

    // Setters
    inline void SetAscendingList( int* list )         { m_ascendingList = list; }
    inline void SetFieldsList( int* list )            { m_fieldsList = list; }
    inline void SetIsTie( nzbool tie )                { m_isTie = tie; }
    inline void SetMergePlan( CMergePlan* plan )      { m_mergePlan = plan; }
    inline void SetMergeType( int type )              { m_mergeType = type; }
    inline void SetMergeReturnArray( record_t** ary ) { m_mergeReturnArray = ary; }
    inline void SetNullsLoList( int* list )           { m_nullsLoList = list; }
    inline void SetNumFields( int num )               { m_numFields = num; }
    inline void SetNumMergeReturnRec( int num )       { m_numMergeReturnRec = num; }
    inline void SetTopN( int num )                    { m_topN = num; }

//    void    SetLimitAttr(int in_topN, bool in_isTie)  {
//        SetIsTie( in_isTie ); // set tournament sort attrs
//        SetTopN( in_topN );
//    }

    // Non-inlines
    void InitSorter(); //initialize the sort info object
    void sortNewBlock( // Notifcation of growing the sort store
            bool lastCall);
    void CleanMerge();
    void sortSwapRestore(void);
};

// Glue between genc and class routines for DBOS' SPU-return merge
void execMergeSort(
    CPlan *planP,
    CTable *dst,
    TSortNode *pNode,
    void *firstRec,         // &memblock.bytes[0]
    bool lastCall);

#endif // !XPSORT_INCLUDED
