#ifndef XPUNLOAD_INCLUDED
#define XPUNLOAD_INCLUDED

#define UNL_IO_SIZE 65536 // IO buffering value

#include "nodecommon.h"
#include "eretstrmode.h"
#include "xpunload_enums.h"

struct TMemblockInfo;

class CUnloadCtrl {
public:
    char m_tokDelim;      // character that seperates fields
    bool m_trailingDelim; // true if there is a delim before the newline
    char m_dateDelim;     // char between mon, day, year
	char m_dateTimeDelim; // char between date and time
    char m_timeDelim;     // char between hour, min, sec
	char m_meridianDelim; // char between sec and meridian
    char m_tokEscape;     // if non-zero, char to use as escape char (usually \)
    enum ParserBools m_boolStyle; // ParserBool_1_0, ParserBool_T_F, ...
    enum ParserDates m_dateStyle; // ParserDate_YMD, ParserDate_MDY, ...
    enum ParserTimes m_timeStyle; // ParserTime_12, ParserTime_24
    enum EUnloaderFormat m_dataFormat;  // UnloadFormatText...
    char m_nullValue[5];  // null string
    bool m_crInString;    // if true allow unescaped CR in strings
    int m_relId;          // relId of external table
    bool m_includezeroseconds; // true if 00 is to be included for seconds
    char m_decimalDelim;   // decimal delimiter
    uint64 m_recordDelimiter;
    bool m_includeheader;  // include header in external file

    CUnloadCtrl ( // (in plan.cpp for convenience of clients)
        char delim,        // character that seperates fields
        bool trailingDelim,// true if there is a delim before the newline
        char dateDelim,    // char between mon, day, year
		char dateTimeDelim, //char between date and time
        char timeDelim,    // char between hour, min, sec
		char meridianDelim, // char between sec and meridian
        char tokEscape,    // if non-zero, char to use as escape char (usually \)
        enum ParserBools boolStyle, // ParserBool_1_0, ParserBool_T_F, ...
        enum ParserDates dateStyle, // ParserDate_YMD, ParserDate_MDY, ...
        enum ParserTimes timeStyle, // ParserTime_12, ParserTime_24
        enum EUnloaderFormat dataFormat, // UnloadFormatText...
        const char *nullValue,   // null string
        bool crInString,   // if true allow unescaped CR in strings
        int relId,         // relId of external table
        bool includezeroseconds, // true if 00 is to be included for seconds
        char decimalDelim,  // decimal delimiter
        uint64 recordDelim,  // record delimiter
        bool includeheader); // include header in external file
                
};

struct TUnloadNodeDef {
protected:
    EUnloaderMode m_mode;            // file/socket...
};

struct TUnloadNodeRuntimeInfo {
protected:
    char          *m_filenameP;      // name of file/socket/program

    char          *m_outBuf;         // write buffer (in front of disk writes)
    int           m_maxSize;         // max ascii record size
    char          *m_dst;            // write ptr into m_outBuf
    uint64        m_written;         // how many rows written
    uint64        m_bytesUnflushed;  // how many bytes written
    uint          m_flushThreshold;  // flush after this threshold (in MB)
    ERetStrMode   m_retStrMode;      // non-zero for formatted strings method

    // If a table in the source query for the unload is a modeled as a view, 
    // like MLS or altered tables  then we have  to handle the compressed 
    // format differently.  In particular, when we make the external 
    // table header, we cannot use the current scan CTable to get the shape 
    // of the rows  in the external table. We do not store distribution keys, 
    // since they are not useful with a join result.  As a consequence, 
    // if the external table is reloaded, we will get uncompress-on-the-host 
    // behavior.  And this OID does it all...
    // 
    // It is set by scanning the range table entries of the Plan node to 
    // see if any  of them have the object class SECURE_TABLE_OID, 
    // TABLE_VERSION_OID, SECURE_TABLE_VERSION_OID.
    // 
    int64          m_tableSource;
    bool           m_mlsTable;
    const TableShape *m_srcTableShape;

    int            m_timeout;	     // timeout for load/unload
    bool           m_headerIncluded; // chk whether the header already included or not 

};

class TUnloadNode : public SnippetNode,
    private TUnloadNodeDef, private TUnloadNodeRuntimeInfo {
public:
    TUnloadNode();
    virtual ~TUnloadNode();
    
    TableShape const * Init(TableShape const * src, EPlanNodeLocus place);

    // Accessors
    inline char*         Dst()         const { return m_dst; }
    inline int           Fd()          const { return m_fd; }
    inline char*         Filename()    const { return m_filenameP; }
    inline bool          MlsTable()    const { return m_mlsTable; }
    inline EUnloaderMode Mode()        const { return m_mode; }
    inline char*         OutBuf()      const { return m_outBuf; }
    inline ERetStrMode   RetStrMode()  const { return m_retStrMode; }
    inline int64         TableSource() const { return m_tableSource; }
    inline uint64        Written()     const { return m_written; }
    inline int		 TimeOut()     const { return m_timeout; }

    // Testers
    inline bool usingNamedPipe () { return (m_mode == UnloadToNamedPipe); }
    inline bool usingFile ()      { return (m_mode == UnloadToFile); }
    inline bool usingSocket ()    { return (m_mode == UnloadToSocket); }

    // Setters
    inline void SetDst( char* val )               { m_dst = val; }
    inline void SetFd( int fd )                   { m_fd = fd; }
    inline void SetFilename( char* val )          { m_filenameP = val; }
    inline void SetMlsTable( bool val )           { m_mlsTable = val; }
    inline void SetMode( EUnloaderMode val )      { m_mode = val; }
    inline void SetRetStrMode( ERetStrMode mode ) { m_retStrMode = mode; }
    inline void SetTableSource( int64 val )       { m_tableSource = val; }
    inline void SetWritten( uint64 val )          { m_written = val; }
    inline void SetTimeOut( int val )             { m_timeout = val; }

    // Non-inlines
    void InitUnloader(const CUnloadCtrl *unload); // initializer called on plan-init
    void vfyState(CPlan *plan);      // Make sure file is open...
    void handleFileError();          // Close and delete file and throw error

    int writeBuf(char *fence);       // Write current m_outBuf to file
                                     // size is fence-m_outBuf
    int writeBlock(TMemblockInfo *block, bool includeMemInfo = true); // Write passed recordstore block to file. if includeMemInfo is true write meminfo header
                                                                      // the block to write.


    void clearFD();                  // clear file descriptor associated w/ unload file
    void closeFile();                // close file and clear associated vars
    bool fileIsOpen() const;         // unload file represented by m_fd is open
    void removeFile();               // for remote unload, remove the fifo file.

    int getBufferSize (const TableShape *src, const CUnloadCtrl *unload, bool bPacked);

 private:
    int           m_fd;              // file descriptor
    nzotuid_t     m_fd_Owner;        // process that allocated m_fd

    void initWriteFlush();        // init write flushing
    void flushWrite();            // Flush write (if have reached threshold)
    void setFD(int fd);           // set file descriptor associated w/ unload file
    void writeExternalFileHeader(CPlan * plan); // Compressed external files have a header
    bool writeFlushEnabled() const;
};

// Unpack nibbles until first stop nibble (value == 0).
// [See BCDPackNumber block comment for the encoding.]
//
// These are ASMs so effectively statements with multiple
// side-effects.  (The point here is that they are NOT
// value yielding functions.)  There are four variants:
//
//  [ START, CHAIN ] X [ FIXED, FLOAT ]
//
// START implies that this is the first nibble-packed
// field in a run; hence that there is no previous nibble
// and that the STATE argument (an int) is  OUT-only.
//
// CHAIN implies that this is a subsequent nibble-packed
// field in a run; hence that there may be a previous
// nibble and that the STATE argument (an int) is IN/OUT.
//
// FIXED implies that nibble decoding does not need to
// account for the letter 'e'; FIXED is suitable for
// decoding all integer and fixed point numeric fields.
//
// FLOAT implies that nibble decoding must be prepared
// the letter 'e', necessary for floating point values
// expressed in exponential notation.
//
// The macros share a common set of arguments:
// - ASCII points to the next output position in the
//   buffer of ascii text being emitted.
// - BYTES points to the next nibble pair in stream
//   being decoded.
// - STATE is an integer width truth value (zero, non-
//   zero); it conveys from one macro invocation to the
//   next whether or not the previous macro invocation
//   left an unconsumed nibble in the most recent byte.
//
// All macros are structured as a loop that decodes two
// nibbles and emits two ascii characters every full loop
// iteration.  The low order nibble (0x0F) is the first
// to be decoded.  A null nibble terminates decoding but
// does not imply end of the entire stream of encoded
// values.
//
// The CHAIN macro variants contain the exact same loop
// as the START variants.  They differ by including logic
// to test STATE to select a loop entrypoint.
//
// The FLOAT macro variants includes logic to look for a
// nibble value of 5 and to map it to 'e'.

#if __i386__ || __x86_64__
#define UNPACK_START_FIXED(ASCII, STATE, BYTES)\
asm(   "jmp	3f\n"\
    "1:	add	$42,%%eax\n"\
    "	stosb\n"\
    "2:	shr	$4,%[state]\n"\
    "	jz	4f\n"\
    "	lea	42(%[state]),%%eax\n"\
    "	stosb\n"\
    "3:	lodsb\n"\
    "	movzb	%%al,%[state]\n"\
    "	and	$0xF,%%eax\n"\
    "	jnz	1b\n"\
    "	not	%[state]\n"\
    "4:"\
    : [ascii]  "+D" (ASCII)\
    , [state] "=&r" (STATE)\
    , [bytes]  "+S" (BYTES)\
    :\
    : "eax", "cc"\
    )

#define UNPACK_CHAIN_FIXED(ASCII, STATE, BYTES)\
asm(   "test	%[state],%[state]\n"\
    "	jz	3f\n"\
    "	movzb	-1(%[bytes]),%[state]\n"\
    "	jmp	2f\n"\
    "1:	add	$42,%%eax\n"\
    "	stosb\n"\
    "2:	shr	$4,%[state]\n"\
    "	jz	4f\n"\
    "	lea	42(%[state]),%%eax\n"\
    "	stosb\n"\
    "3:	lodsb\n"\
    "	movzb	%%al,%[state]\n"\
    "	and	$0xF,%%eax\n"\
    "	jnz	1b\n"\
    "	not	%[state]\n"\
    "4:"\
    : [ascii]  "+D" (ASCII)\
    , [state]  "+r" (STATE)\
    , [bytes]  "+S" (BYTES)\
    :\
    : "eax", "cc"\
    )

#define UNPACK_START_FLOAT(ASCII, STATE, BYTES)\
asm(   "jmp	3f\n"\
    "1:	cmp	$5,%%eax\n"\
    "	lea	42(%%eax),%%eax\n"\
    "	cmovz	%[e],%%eax\n"\
    "	stosb\n"\
    "2:	shr	$4,%[state]\n"\
    "	jz	4f\n"\
    "	cmp	$5,%[state]\n"\
    "	lea	42(%[state]),%%eax\n"\
    "	cmovz	%[e],%%eax\n"\
    "	stosb\n"\
    "3:	lodsb\n"\
    "	movzb	%%al,%[state]\n"\
    "	and	$0xF,%%eax\n"\
    "	jnz	1b\n"\
    "	not	%[state]\n"\
    "4:"\
    : [ascii]  "+D" (ASCII)\
    , [state] "=&r" (STATE)\
    , [bytes]  "+S" (BYTES)\
    : [e]       "r" ((int)'e')\
    : "eax", "cc"\
    )

#define UNPACK_CHAIN_FLOAT(ASCII, STATE, BYTES)\
asm(   "test	%[state],%[state]\n"\
    "	jz	3f\n"\
    "	movzb	-1(%[bytes]),%[state]\n"\
    "	jmp	2f\n"\
    "1:	cmp	$5,%%eax\n"\
    "	lea	42(%%eax),%%eax\n"\
    "	cmovz	%[e],%%eax\n"\
    "	stosb\n"\
    "2:	shr	$4,%[state]\n"\
    "	jz	4f\n"\
    "	cmp	$5,%[state]\n"\
    "	lea	42(%[state]),%%eax\n"\
    "	cmovz	%[e],%%eax\n"\
    "	stosb\n"\
    "3:	lodsb\n"\
    "	movzb	%%al,%[state]\n"\
    "	and	$0xF,%%eax\n"\
    "	jnz	1b\n"\
    "	not	%[state]\n"\
    "4:"\
    : [ascii]  "+D" (ASCII)\
    , [state]  "+r" (STATE)\
    , [bytes]  "+S" (BYTES)\
    : [e]       "r" ((int)'e')\
    : "eax", "cc"\
    )
#endif // __i386__  || __x86_64__

#ifndef CEXTERN
#define CEXTERN extern "C"
#endif

CEXTERN int throwError(int code, const char *buf, const char *obj);

CEXTERN char *HackTypeStr( // Temp function for unloader dev. 
    char *buf, int type, void *fieldP, int size, int trueSize);

char *BCDPackNumber(  // Convert numeric string to a digit per nibble (bcd)
    char* bytes,      // dst buffer
    char& halfFull,   // flag whether was space left for a nibble in prev
    char decimalDelim,// decimal delimiter 
    char* ascii);     // src buffer (null-termed)

char *BCDUnpackNumber( // Convert bcd to ascii
    char* asc,  // dst buffer, must be big enough to hold val
    char* &bcd); // src buffer

union UAllNumbers {
    long long int full64;
    struct {
#if __PPC__
        int high32;
        unsigned int low32;
#else
        unsigned int low32;
        int high32;
#endif
    } i2;
    UAllNumbers(int i) : full64(i) {}
    UAllNumbers(long l) : full64(l) {}  // for int64 in 64bit mode
    UAllNumbers(long long int ll) : full64(ll) {}
};

char* NumbersToBCD( // Convert 32 & 64 bit numbers to BCD
    char* bytes,            // dst buffer
    char& state,            // in/out flag for nybble packing
    UAllNumbers value,      // input value - 32 or 64 bits
    char decimalDelim,      // decimal delimiter
    unsigned dpPos = 0);    // digits after decimal (i.e. scale)

CEXTERN int RecDelimLength (uint64 recordDelim);

CEXTERN char *XpReturnAddChar( // Add fixed char string to packed unloader buffer
    char *dst, char *src, int len, char tokDelim, char tokEscape, char *nullValue, int nullValueLen, bool crInString, unsigned char chspc, uint64 recordDelim);

CEXTERN char *XpReturnAddCharAndTerm( // Add null-terminated fixed char string to packed unloader buffer
    char *dst, char *src, int len, char tokDelim, char tokEscape, char *nullValue, bool crInString, unsigned char chspc, uint64 recordDelim);

CEXTERN char *XpReturnAddVarChar (// Add varchar string to packed unloader buffer
    char *dst, char *src, int len, char tokDelim, char tokEscape, char *nullValue, bool crInString, uint64 recordDelim);

// This macro appends an empty field (the external rendering of a NULL value)
// to the packed nibble stream represented by the vaiables char* BYTES and
// char STATE.
#define PACK_NULL(BYTES, STATE)\
do {\
    if (STATE)\
        STATE = 0;\
    else {\
        *BYTES = STATE;\
        ++BYTES;\
        STATE = 1;\
        }\
}\
while (0)

#define COMPARE_RECDELIM(STR, LEN, RECORDDELIM, COMPARE_LEN, EQUAL)\
do {\
    if ( *STR++ != ((uint8*)&RECORDDELIM) [COMPARE_LEN++]) {\
        EQUAL=0;\
        break;\
    }\
}\
while (COMPARE_LEN < LEN)

// This macro checks single tokens and nullValue. Add escape token in the front of the string 
// if they are matched and escape token is specified.
#define XP_TOK_CHECK(DST, CH, TOK_DELIM, TOK_ESCAPE, NULL_VALUE, NULL_VAL_FOUND, CRINSTRING, RECORDDELIM)\
    if (CH == 13 && CRINSTRING);\
    else if (NULL_VAL_FOUND || CH == TOK_ESCAPE || CH == TOK_DELIM || CH == 10 || CH == 13 || CH == 0)  {\
      if (TOK_ESCAPE)\
        *DST++ = TOK_ESCAPE;\
      else {\
        if (RECORDDELIM == 10)\
            XpHandleBadChar(TOK_DELIM, CH, NULL_VALUE, NULL_VAL_FOUND);\
        }\
    }     

extern const char *UnloadByteVals[256];

// Version stamp.  Bump when anything affecting existing files changes,
// and write some backwards compatibility code.  Candidates are this
// struct, TMemblockInfo, dbos_msg_header_t, spu_msg_header_t, and
// anything that might change the distribution scheme and invalidate
// the data slices saved with the record blocks.
//
// Temporarily bumped to 3 for 5.0. Disabled compatibility to conserve
// QA resources. In 5.0.2, going back down to 2, and will explicity allow
// reads of version 3. When need to increase next time, go to 4. See SWS-47064.
#define EXTTAB_HEADER_V2      2
#define EXTTAB_HEADER_V3      3
#define EXTTAB_HEADER_VERSION 2
// Column definition. Members from field_t, not use directly b/c didn't
// all of it and so unaffected by changes to that structure.
struct TColumnDef {
    uint16 type;        // datatype
    uint16 size;        // column size (& precision)
    uint16 physField;   // physical field dex for this logical field
    nzbool nullAllowed;
};
struct TExttabHeader {
    uint32     version;    // Version stamp. 
    uint32     numSPUs;    // Number of dataslices on source (unload) machine
    TRowId     rowIDFence; // Current max rowid on source (unload) machine
    int        numDistKeys;// # of distribution keys
    int16      distKeys[4];// array of distKeys MAX_DIST_KEY_FIELDS
    int        magic;      //0xFEEDFACE
    uint8      spuLittleEndian;// true when spu is in little endian mode
    uint32     distAlgorithm;// random vs. hash
    uint       dbCharset;  // 8-bit charset oid
    int        numFields;  // number of fields in the table
    TColumnDef *tableColumnDefs; // table column info
};

// Only implementations are x86
#  if __i386__  || __x86_64__

// These are the actual out-of-line implementations.  They
// appear first to satisfy the C/C++ declaration before use
// requirement.  See below for more user friendly wrappers.

char* int32_to_ascii(char* ascii
                   , unsigned int value_31_0
                   );

char* int64_to_ascii(char* ascii
                   , unsigned int value_31_0
                   , unsigned int value_63_32
                   );

char* cnum32_to_ascii(char* ascii
                   , unsigned int value_31_0
                   , char decimalDelim
                   , unsigned int scale = 0
                   );

char* cnum64_to_ascii(char* ascii
                   , unsigned int value_31_0
                   , unsigned int value_63_32
                   , char decimalDelim
                   , unsigned int scale = 0
                   );

char* cnum128_to_ascii(char* ascii
                   , unsigned int value_31_0
                   , unsigned int value_63_32
                   , unsigned int value_95_64
                   , unsigned int value_127_96
                   , char decimalDelim
                   , unsigned int scale = 0
                   );


// Wrapper for mere mortals
inline
char* int32_to_ascii(char* ascii, int i) {
//    return int32_to_ascii(ascii, i);
    return cnum32_to_ascii(ascii, i, '.', 0);
}

inline
char* int64_to_ascii(char* ascii, long long int ll) {
//    return int64_to_ascii(ascii, ll, ll>>32);
    return cnum64_to_ascii(ascii, ll, ll>>32, '.', 0);
}

inline
char* cnum32_to_ascii(char* ascii, int i, char decimalDelim, unsigned int scale = 0)  {
    return cnum32_to_ascii(ascii, i, decimalDelim, scale);
}

inline
char* cnum64_to_ascii(char* ascii, long long int ll, char decimalDelim, unsigned int scale = 0) {
    return cnum64_to_ascii(ascii, ll, ll>>32, decimalDelim, scale);
}


// The following functions should use the definitions of CNumeric64 and CNumeric128
// from cnumeric.h.  Unfortunately that triggers a series of cascaded dependencies
// ultimately involving placement operator new().  I have been involved with that
// particular tar baby on multiple occassions and am unwilling to solve it right now.
//
// For now I cheat and exploit my knowledge of the layout of the cnumeric types.

class CNumeric32;
inline
char* cnum32_to_ascii(char* ascii, CNumeric32 const & cn, char decimalDelim, unsigned int scale = 0) {
    return cnum32_to_ascii(ascii, ((unsigned*)&cn)[0], decimalDelim, scale);
}

class CNumeric64;
inline
char* cnum64_to_ascii(char* ascii, CNumeric64 const & cn, char decimalDelim, unsigned int scale = 0) {
    return cnum64_to_ascii(ascii, ((unsigned*)&cn)[1], ((unsigned*)&cn)[0], decimalDelim, scale);
}

class CNumeric128;
inline
char* cnum128_to_ascii(char* ascii, CNumeric128 const & cn, char decimalDelim, unsigned int scale = 0) {
    return cnum128_to_ascii(ascii, ((unsigned*)&cn)[3], ((unsigned*)&cn)[2],
        ((unsigned*)&cn)[1], ((unsigned*)&cn)[0], decimalDelim, scale);
}

#  endif    // i386

#endif // !XPUNLOAD_INCLUDED
