/*
 * Copyright IBM Corp. 2001, 2012.
 */

#ifndef INCLUDED_WINDOWAGGR
#define INCLUDED_WINDOWAGGR
/* Header for sql/olap window aggregation node */

#include "nodecommon.h"
#include "xpaggregate.h"

class TWindowAggrNode;
struct TStreamInfo;
struct UdxLibraryInfo;
struct Node; // postgres
class UdxEnvInfo;

struct TWindowAggrField {
    EAggrTypes m_aggrType;
    int m_frameDex;       // frame entry in parent WindowAggr node
    int m_srcColDex;      // column to aggregate in input stream
    int m_arg2Val;        // -1 if default, else lead/lag arg2 literal (TEMP - will be colDex one day)
    int m_arg3Val;        // -1 if default, else coldex of "default" lead/lag arg3
    bool m_distinct;
    int m_dstType;        // PG field type
    int32 m_dstTypmod;    // PG aggr field shape
    int m_srcType;        // type of src material (differs from fType when COUNT(src) )
    int m_filterDex;      // 0 =>no FILTER expr, else 1+(offset of WindowFrame->m_filterList entry)

    char *name;           // !REWORK - need comments on this and following UDA stuff
    char* externalName;
    oidvector argTypmods;
    oidvector intTypmods;
    oidvector intTypes;
    oidvector argCols;
    int nargs;
    int nints;
  		
    bool        fenced;
    int16       log;
    char*       memory;
    char*       hostcksum;
    char*       spucksum;
    char*       datadir;
    Oid         aggid;
    bool        stub;
    int         numdependencies;
    UdxLibraryInfo  *dependencies;
    int         numenvironment;
    UdxEnvInfo *environment;

    int numparameter;
    char** parameters;

    int version;

    // code gen runtime state:
    int m_colDex;  //position of this aggregate in aggr result rec
    bool m_bMinMaxByPointer;  // true if tracking min/max by record pointer
    bool m_bIterateOverFrame; // false for LEAD/LAG/FIRST_VALUE/LAST_VALUE
    bool m_bRunner;   // simple running (unbounded prec to current) sum or count
    bool m_bIgnoreNulls;   // true if ignore nulls set . poisble only for FIRST_VALUE , LAST_VALUE
};

enum EWindowFrameUnits {
    frameRows = 1,
    frameRange
};

enum EWindowFrameBound {
    boundCurrentRow = 1,
    boundUnboundedPreceding,
    boundConstPreceding,
    boundUnboundedFollowing,
    boundConstFollowing,
    boundNonNullFollowing
};

enum EWindowFrameExclusion {
    excludeNone = 1,
    excludeCurrentRow,
    excludeGroup,
    excludeTies
};

struct TFrameDelta {
    int m_type;        // -2 => unbounded till non null , -1 => unbounded; 0 => current row; else datatype of delta value
    union {
        int32 m_i32;
        int64 m_i64;
        double m_double;
        struct {
            signed int hi;
            unsigned int hl;
            unsigned int lh;
            unsigned int lo;
        } m_numeric;
    } m_val;
    // methods to load appropriate delta values from a const node
    void int32Delta(void *pNode);
    void int64Delta(void *pNode);
    void dateDelta(void *pNode);
    void timeDelta(void *pNode);
    void timestampDelta(void *pNode);
    void intervalDelta(void *pNode);
    void doubleDelta(void *pNode);
    void numericDelta(void *pNode, int scale, int digitCount);
};

struct TWindowFrame {
    EWindowFrameUnits m_units;
    EWindowFrameBound m_startBound;
    Node *m_startConstNode;
    EWindowFrameBound m_endBound;
    Node *m_endConstNode;
    EWindowFrameExclusion m_exclusion;
    int  m_numFilters;         // number of entries in m_filterExpr array
    Node **m_filterList;       // array of FILTER expressions for aggrs in this node.
    int m_pcodeStart;          // number of first pcode instr
    int m_pcodeCount;          // count of pcode instrs

    // runtime state
    int m_frameStart;          // m_row entry number of first rec of current frame
    int m_frameEnd;            // m_row entry number of last rec of current frame
    int64 m_rowNumber;         // row_number of current row
    int64 m_sparseRank;        // rank of current row
    int64 m_denseRank;         // dense rank of current row

    // code gen runtime state
    TFrameDelta m_precedeDelta;// frame precede delta const
    TFrameDelta m_followDelta; // frame follow delta const
    bool m_frameIteration;     // iteration over this frame needed
    int m_rankFlags;           // inidicate flavors of rank/rownum needed for this frame
    bool m_uncentered;         // an uncentered frame
    bool m_bRunner;            // rows or range between unbounded preceding and current row
    bool m_fullPartition;      // true iff frame is full partition (unbounded preceding and following)
    bool m_needOldFrame;         //do we have a first value ?
    bool isFrameRunner() const;
    bool isFrameRowBasedRunner() const;
    bool isFrameOnlyPriorAndCurrenRows() const;
    void setNeedOldFrame(); 
    bool needOldFrameComputation() const;

    bool isFullPartitionFrame() const;
    bool isUncenteredFrame() const;
    void initCodeGenRunTimeState( int rangeType, int scale, int digitCount);
    TFrameDelta getPrecedeDelta() const;
    TFrameDelta getFollowDelta() const;
    TFrameDelta computeFrameDelta(
            bool startBound,  // true if called for start bound, else for end bound
            EWindowFrameBound bound,
            void *constNode,
            int rangeType,
            int scale,        // relevant to NzTypeNumeric ranges only
            int digitCount
    ); 
};

// m_rankFlags flag bits:
#define RANK_ROWNUM 0x0001
#define RANK_DENSE  0x0002
#define RANK_SPARSE 0x0004

struct RowSpec {
    void *m_recP;              // ptr to row, null if freed
    uint32 m_blockNum;         // of this row's block
    int8 m_tieVal;             // toggled value to indicate run of peers
};

enum EPartitionMode {
    GotFrameStart       = -1,
    GotFrame            =  0,
    FirstPartition      =  1,
    SubsequentPartition =  2
};

class TWindowAggrNodeHelper;

class TWindowAggrNodeDef {
protected:
    int m_numAggrs;            // number of aggregates
    TWindowAggrField *m_aggrs; // list of agrregates
    int m_numPartitionCols;    // number of partition columns
    int *m_partitionColList;   // list of partition columns
    int m_numOrderCols;        // number of order columns
    int *m_orderColList;       // list of order columns
    int *m_ascendingList;      // list of order direction flags
    int *m_nullsLoList;        // list of nulls collation booleans
};

class TWindowAggrNodeRuntimeInfo {
protected:
    int m_numFrames;           // number of frames in partition/order window
    TWindowFrame *m_frames;    // list of window frames
    int m_srcMaxRecLen;        // max rec len of source table - needed for pagesNeeded()
    bool m_minimalPinningRequired; // this node requires pinning at most one prior row
    bool m_prevWindowAggrDeepCopy; //how to use prevWindowAggr buffer
    TableShape *m_pinnedFrameTblShape;
    CTable *m_pinnedFrameTbl;  // partially-materialized pinned frame rows table
    TStreamInfo *m_pinnedFrameS; // stream info for pinned frame rows table
    void *m_resultRow[2];      // materialized result row (will need current and prev for "running" optimization)
    void *m_prevWindowAggrCopy; //deep copy contenets of prevWindowAggr here (required for
                                // range frames with exclusions that have min/max
                                // on varchar datatypes)
    int m_curBuf;              // 0/1 toggle for cur/prev m_resultRow  swapping
    int m_availPages;          // our claim on snippetCap
    int m_maxPinnedRecBlocks;  // max blocks we'll pin
    int m_maxRowArrayCapacity; // max ents in m_row array
    int m_loFrameStart;        // m_row entry number of first rec of any frame
    int m_hiFrameEnd;          // m_row entry number of last rec of any frame
    int m_curRec;              // m_row entry number of current record for which aggregate is being computed
    int m_nextPartition;       // m_row entry number of first rec of next partition (0 if not known)
    EPartitionMode m_mode;     // controls initialization for new partition
    int m_loBlock;             // logical block number of first pinned frame table block
    int m_logicalFrameTblBlocks; // logical number of blocks in pinned frame table
    RowSpec *m_row;            // array of partition rowspecs (structure RowSpec)
    int m_rowArrayRow;         // entry number of most recent row in m_row array
    int m_rowArrayCapacity;    // number of slots currently in m_row array
    int m_curTieVal;           // current value of m_tieVal as of last row seen
    int *m_fvRecNum;           // current firstValue row index in m_row
                               // one per each first_value ignore nulls field in aggrs
    bool *m_fvRecNull;         // corresponding to m_fvRecNum,
                               // is the first value null or not

    TFrameDelta m_maxPrecedeDelta;   // max frame precede delta
    TFrameDelta m_maxFollowDelta;    // max frame follow delta
    bool m_groupExclusions;          // group/tie exclusion code needed
    bool m_frameIterations;          // iteration over frames needed
    bool m_ranks;                    // true iff any rank/dense_rank funcs
    bool m_distincts;                // true iff any distinct aggrs
    bool m_runners;                  // true iff any row-based running aggrs
    bool m_nonRunners;               // true iff anything other than row-based running aggrs
    bool m_fullPartitionFrames;      // true iff any frames are full partition (unbounded preceding and following)
    int m_rowFrameCnt;               // number of row-based frames this node
    int m_rangeFrameCnt;             // number of range-based frames this node
    TStreamInfo *m_prevResultRecS;   // helper stream aid to code gen reference to previous m_result row
    TStreamInfo *m_curRecS;          // helper stream aid to code gen reference to current record
    TStreamInfo *m_partRecS;         // helper stream aid to code gen reference to first rec of partition
    TStreamInfo *m_frameRecS;        // helper stream aid to code gen reference to frame looping record
    TStreamInfo *m_tmpRecS;        // helper stream aid to code gen reference to tmp record
    TStreamInfo **m_extremeRecSArray;// helpers for pointer type min/max
    TWindowAggrNodeHelper *m_helper;
};

class TWindowAggrNode : public SnippetNode,
    private TWindowAggrNodeDef, private TWindowAggrNodeRuntimeInfo {
public:
    TWindowAggrNode();
    virtual ~TWindowAggrNode();
    
    TableShape const * Init(TableShape const * src, EPlanNodeLocus place);

    // Address-of accessors
    // Used for things like virtual-table materialization.
    inline TFrameDelta*  AddrOfMaxFollowDelta()   { return &m_maxFollowDelta; }
    inline TFrameDelta*  AddrOfMaxPrecedeDelta()  { return &m_maxPrecedeDelta; }

    // Accessors
    inline TWindowAggrField* Aggrs()              const { return m_aggrs; }
    inline int*          AscendingList()          const { return m_ascendingList; }
    inline int           CurBuf()                 const { return m_curBuf; }
    inline int           CurRec()                 const { return m_curRec; }
    inline TStreamInfo*  CurRecS()                const { return m_curRecS; }
    inline int           CurTieVal()              const { return m_curTieVal; }
    inline bool          Distincts()              const { return m_distincts; }
    inline TStreamInfo** ExtremeRecSArray()       const { return m_extremeRecSArray; }
    inline TStreamInfo*  FrameRecS()              const { return m_frameRecS; }
    inline TWindowFrame* Frames()                 const { return m_frames; }
    inline bool*         FvRecNull()              const { return m_fvRecNull; }
    inline int*          FvRecNum()               const { return m_fvRecNum; }
    inline bool          FullPartitionFrames()    const { return m_fullPartitionFrames; }
    inline bool          GroupExclusions()        const { return m_groupExclusions; }
    inline TWindowAggrNodeHelper* Helper()        const { return m_helper; }
    inline int           HiFrameEnd()             const { return m_hiFrameEnd; }
    inline int           LoBlock()                const { return m_loBlock; }
    inline int           LoFrameStart()           const { return m_loFrameStart; }
    inline int           LogicalFrameTblBlocks()  const { return m_logicalFrameTblBlocks; }
    inline TFrameDelta   MaxFollowDelta()         const { return m_maxFollowDelta; }
    inline TFrameDelta   MaxPrecedeDelta()        const { return m_maxPrecedeDelta; }
    inline EPartitionMode Mode()                  const { return m_mode; }
    inline bool          NonRunners()             const { return m_nonRunners; }
    inline int*          NullsLoList()            const { return m_nullsLoList; }
    inline int           NumAggrs()               const { return m_numAggrs; }
    inline int           NumFrames()              const { return m_numFrames; }
    inline int           NumOrderCols()           const { return m_numOrderCols; }
    inline int           NumPartitionCols()       const { return m_numPartitionCols; }
    inline int*          OrderColList()           const { return m_orderColList; }
    inline int*          PartitionColList()       const { return m_partitionColList; }
    inline TStreamInfo*  PartRecS()               const { return m_partRecS; }
          inline TStreamInfo*  PinnedFrameS()     const { return m_pinnedFrameS; }
    inline CTable*       PinnedFrameTbl()         const { return m_pinnedFrameTbl; }
    inline TableShape*   PinnedFrameTblShape()    const { return m_pinnedFrameTblShape; }
    inline TStreamInfo*  PrevResultRecS()         const { return m_prevResultRecS; }
    inline void*         PrevWindowAggrCopy()     const { return m_prevWindowAggrCopy; }
    inline bool          PrevWindowAggrDeepCopy() const { return m_prevWindowAggrDeepCopy; }
    inline int           RangeFrameCnt()          const { return m_rangeFrameCnt; }
    inline bool          Ranks()                  const { return m_ranks; }
    inline void**        ResultRow()                    { return m_resultRow; }
    inline RowSpec*      Row()                    const { return m_row; }
    inline int           RowArrayCapacity()       const { return m_rowArrayCapacity; }
    inline int           RowArrayRow()            const { return m_rowArrayRow; }
    inline int           RowFrameCnt()            const { return m_rowFrameCnt; }
    inline bool          Runners()                const { return m_runners; }
    inline int           SrcMaxRecLen()           const { return m_srcMaxRecLen; }
    inline TStreamInfo*  TmpRecS()                const { return m_tmpRecS; }

    // Setters
    inline void SetAggrs( TWindowAggrField* val )       { m_aggrs = val; }
    inline void SetAscendingList( int* val )            { m_ascendingList = val; }
    inline void SetCurBuf( int val )                    { m_curBuf = val; }
    inline void SetCurRec( int val )                    { m_curRec = val; }
    inline void SetCurRecS( TStreamInfo* val )          { m_curRecS = val; }
    inline void SetCurTieVal( int val )                 { m_curTieVal = val; }
    inline void SetFrameRecS( TStreamInfo* val )        { m_frameRecS = val; }
    inline void SetFrames( TWindowFrame* val )          { m_frames = val; }
    inline void SetHelper( TWindowAggrNodeHelper* val ) { m_helper = val; }
    inline void SetLoBlock( int val )                   { m_loBlock = val; }
    inline void SetLoFrameStart( int val )              { m_loFrameStart = val; }
    inline void SetHiFrameEnd( int val )                { m_hiFrameEnd = val; }
    inline void SetMode( EPartitionMode val )           { m_mode = val; }
    inline void SetNullsLoList( int* val )              { m_nullsLoList = val; }
    inline void SetNumAggrs( int val )                  { m_numAggrs = val; }
    inline void SetNumFrames( int val )                 { m_numFrames = val; }
    inline void SetNumOrderCols( int val )              { m_numOrderCols = val; }
    inline void SetNumPartitionCols( int val )          { m_numPartitionCols = val; }
    inline void SetOrderColList( int* val )             { m_orderColList = val; }
    inline void SetPartitionColList( int* val )         { m_partitionColList = val; }
    inline void SetPartRecS( TStreamInfo* strm )        { m_partRecS = strm; }
    inline void SetPinnedFrameS( TStreamInfo* strm )    { m_pinnedFrameS = strm; }
    inline void SetPrevResultRecS( TStreamInfo* strm )  { m_prevResultRecS = strm; }
    inline void SetRowArrayRow( int val )               { m_rowArrayRow = val; }
    inline void SetSrcMaxRecLen( int len )              { m_srcMaxRecLen = len; }
    inline void SetTmpRecS( TStreamInfo* strm )         { m_tmpRecS = strm; }

    // Non-inlines
    const char *getExtremeValStr(
        TStreamInfo *sP,
        int dstCol,
        int dataType,
        int aggrType);
    void getFrameDelta(
        bool startBound,
        const EWindowFrameBound &bound,
        const TFrameDelta &delta,
        TFrameDelta *maxDelta);
    void getPinningRequirement( void );
    void growRows( void );
    int pagesNeeded(bool useTblShape);
    void pinnedFrameNewBlock(bool lastCall);
    void freeRow();

    // runtime state
    // code gen runtime state
    bool hasNonNullFrame() const;

};

#endif // INCLUDED_WINDOWAGGR
