#ifndef ZMSTRN_H
#define ZMSTRN_H

//========================================================================
///\file: zmstrn.h
//
// (C) Copyright IBM Corp. 2010, 2011  All Rights Reserved.
//========================================================================

#include "nzstd.h"

namespace zm {

// Dense containers for strings from 1 to 8 bytes irrespective of encoding.
#define _PACKED     __attribute__(( packed ))
#define _FIELDS(X)  struct { X; } _PACKED

struct str1 { union { uint8 c[1]; uint64 u :  8;                                              } _PACKED; };
struct str2 { union { uint8 c[2]; uint64 u : 16;                                              } _PACKED; };
struct str3 { union { uint8 c[3]; uint64 u : 24;                                              } _PACKED; };
struct str4 { union { uint8 c[4]; uint64 u : 32;         uint32 u32;                          } _PACKED; };
struct str5 { union { uint8 c[5]; uint64 u : 40; _FIELDS(uint32 u32;             uint8 u8) p; } _PACKED; };
struct str6 { union { uint8 c[6]; uint64 u : 48; _FIELDS(uint32 u32; uint16 u16          ) p; } _PACKED; };
struct str7 { union { uint8 c[7]; uint64 u : 56; _FIELDS(uint32 u32; uint16 u16; uint8 u8) p; } _PACKED; };

struct str8 { union { uint8 c[8]; uint64 u;              uint32 u32[2];                       } _PACKED;
    uint64 top () const;
    uint64 ustr() const;
};


unsigned bswap(unsigned);


// Reverse bytes without consideration of any larger containers.  This
// allows use of unsigned integer comparison is to achieve lexical
// comparison instead of the more costly byte at a time formulation.
inline uint32 reverseBytes(str1 const& s) { return                                   (                                 (s.c[0] << 0)); }
inline uint32 reverseBytes(str2 const& s) { return                                   (                 (s.c[0] << 8) | (s.c[1] << 0)); }
inline uint32 reverseBytes(str3 const& s) { return                                   ((s.c[0] << 16) | (s.c[1] << 8) | (s.c[2] << 0)); }
inline uint32 reverseBytes(str4 const& s) { return                                   (                                  bswap(s.u32)); }
inline uint64 reverseBytes(str5 const& s) { return ((uint64)bswap(s.p.u32)  <<  8) | (                                 (s.c[4] << 0)); }
inline uint64 reverseBytes(str6 const& s) { return ((uint64)bswap(s.p.u32)  << 16) | (                 (s.c[4] << 8) | (s.c[5] << 0)); }
inline uint64 reverseBytes(str7 const& s) { return ((uint64)bswap(s.p.u32)  << 24) | ((s.c[4] << 16) | (s.c[5] << 8) | (s.c[6] << 0)); }
inline uint64 reverseBytes(str8 const& s) { return ((uint64)bswap(s.u32[0]) << 32) | (                               bswap(s.u32[1])); }

// When evaluating relationals on strings wider than 4 bytes the result
// depends on later characterss only when the first 4 characters match.
inline uint32 laterChars(str5 const& s) { return s.p.u8                  ; }
inline uint32 laterChars(str6 const& s) { return s.p.u16                 ; }
inline uint32 laterChars(str7 const& s) { return s.p.u16 | (s.p.u8 << 16); }
inline uint32 laterChars(str8 const& s) { return s.u32[1]                ; }


// Homogeneous comparison operations needed by std::sort and deduplication
#define STRN_CMP_OP(N, OP) \
inline \
bool operator OP \
    ( str##N const& lhs \
    , str##N const& rhs \
    ) \
{ \
    return lhs.u OP rhs.u; \
} \

#define STRN_REL_OP(N, OP) \
inline \
bool operator OP \
    ( str##N const& lhs \
    , str##N const& rhs \
    ) \
{ \
    return bswap((uint32)lhs.u) OP bswap((uint32)rhs.u); \
} \

#define STRN_OPS(N) \
STRN_CMP_OP(N, ==)  \
STRN_CMP_OP(N, !=)  \
STRN_REL_OP(N, < )  \
STRN_REL_OP(N, > )  \

STRN_OPS(1)
STRN_OPS(2)
STRN_OPS(3)
STRN_OPS(4)

#undef STRN_REL_OP

#define STRN_REL_OP(N, OP) \
inline \
bool operator OP \
    ( str##N const& lhs \
    , str##N const& rhs \
    ) \
{ \
    uint32 l4 = *(uint32 const*)&lhs; \
    uint32 r4 = *(uint32 const*)&rhs; \
    if (l4 == r4) { \
        l4 = laterChars(lhs); \
        r4 = laterChars(rhs); \
    } \
    return bswap(l4) OP bswap(r4); \
} \


STRN_OPS(5)
STRN_OPS(6)
STRN_OPS(7)
STRN_OPS(8)


// Derive padding-aware strNa and strNe types

#define DEF_STRN_WITH_ENCODING(SUFFIX, LEN, SPACE) \
struct str##LEN##SUFFIX : public str##LEN { \
uint64 top () const { return ((uint64)                  u << (64 - LEN*8)) | (SPACE * (0x0101010101010101ULL >> (LEN*8))); } \
uint64 ustr() const { return ((uint64)reverseBytes(*this) << (64 - LEN*8)) | (SPACE * (0x0101010101010101ULL >> (LEN*8))); } \
} _PACKED \

#define DEF_STRN_WITH_ENCODING_PAIR(LEN) \
    DEF_STRN_WITH_ENCODING(A, LEN,  ' '); \
    DEF_STRN_WITH_ENCODING(E, LEN, 0x40)  \

DEF_STRN_WITH_ENCODING_PAIR(1);
DEF_STRN_WITH_ENCODING_PAIR(2);
DEF_STRN_WITH_ENCODING_PAIR(3);
DEF_STRN_WITH_ENCODING_PAIR(4);
DEF_STRN_WITH_ENCODING_PAIR(5);
DEF_STRN_WITH_ENCODING_PAIR(6);
DEF_STRN_WITH_ENCODING_PAIR(7);

#undef _PACKED
#undef _FIELDS
#undef STRN_CMP_OP
#undef STRN_REL_OP
#undef STRN_OPS
#undef STRN_REL_OP
#undef DEF_STRN_WITH_ENCODING
#undef DEF_STRN_WITH_ENCODING_PAIR


// Convenient constants
static const uint64 SPACES8_ASCII  =  ' ' * 0x0101010101010101LL;
static const uint64 SPACES8_EBCDIC = 0x40 * 0x0101010101010101LL;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


inline
unsigned bswap
    ( unsigned val
    )
{
    unsigned xval;

asm("bswap   %[val]\n"
    : [xval] "=r" (xval)
    : [val]   "0" (val)
    );
    return xval;
}

inline
uint64 str8::top
    /// Return the 8 byte contents as a wide integer.
    () const
{
    return u;
}

inline
uint64 str8::ustr
    /// Return the 8 byte contents in an order that supports wide
    /// integer comparison.
    () const
{
    return (uint64)reverseBytes(*this);
}

}; // namespace zm
#endif // ZMSTRN_H
